/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass <dama@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iostream>
#include "nmm/base/connect.hpp"
#include "nmm/utils/utils.hpp"

// for remote registries
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

#include "nmm/interfaces/video/osd/manager/IOSDManager.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/IBufferManager.hpp"
#include "nmm/interfaces/av/mpeg/IMPEGDemux.hpp"
#include "nmm/interfaces/base/IBufferManager.hpp"
#include "nmm/interfaces/base/sync/ISwitch.hpp"
#include "nmm/interfaces/base/registry/IPluginManager.hpp"
#include "nmm/interfaces/base/IReferenceCount.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"

#include "camera.hpp"

using namespace NMM;

CameraState::CameraState( string id,  MMBoxApplication* _app ) : MMBoxState( id, _app ),
                                                                 move_up(false),
                                                                 move_down(false),
                                                                 move_left(false),
                                                                 move_right(false),
                                                                 zoom_in(false),
                                                                 zoom_out(false)
{
  /* get pointer to global nodes */
  osd_node   = app->getOsdNode();
  display    = app->getDisplayNode();

  camera      = NULL;

  camera_interface = NULL;
}

CameraState::~CameraState() {
}

Result CameraState::doCreate( string session_location ) {
  GLOBAL_DEBUG_STREAM( "Camera: initialize: " << endl);

  /* Begin> Create default graph */
  if (FAILURE == app->createDefaultGraph(background)) {
    return FAILURE;
  }
  if (FAILURE == app->startDefaultGraph()) {
    return FAILURE;
  }
  /* End< Create default graph */

  /* Begin> Request nodes */
  try {
    app->disconnectDefaultGraph();
    requestNodes(session_location);
    camera_interface = camera->getParentObject()->getCheckedInterface<IVISCACamera>();

    app->setSynchronized( true );
    app->getOsdManager()->setPause(false);

    /* Disconnect the application's synchronizer from audio and video sink */
    disconnectSinkSynchronizer( app->getSynchronizer() );

    /* Use the given synchronizer from graph description ... */
    sink_synchronizer.reset( graph.getISession()->getSinkSynchronizer() );

    /* and connect it with our audio and video sink */
    connectSinkSynchronizer( sink_synchronizer.get() );

    /* realize the graph */
    graph.realizeGraph();

  } catch (const Exception& e){
    GLOBAL_ERROR_STREAM("RealizeGraph failed: " << e << endl);
    showAlert("Input Error", "Failed to initialize CameraState");
    app->getOsdManager()->update();
    cleanup();
    return SUCCESS;
  }

  try{
    graph.activateGraph();
    app->getDisplayNode()->reachStarted();
    graph.startGraph();
  } catch (...) {
    showAlert("Input Error", "Failed to start CameraState");
    app->getOsdManager()->update();
    cleanup();
    return FAILURE;
  }
  camera_interface->PowerOn(1);
  
  updateColorBar();
  return( SUCCESS );
}

Result CameraState::doKill() {
  cleanup();
  if (app->getOsdNode()->isInputConnected("stream1")) {
    // if OSD's video input is still connected after cleanup, the
    // default graph must be connected
    app->disconnectDefaultGraph();
    // this leaves the osd in INITIALIZED where the other MMBoxStates
    // expect it.
  }
  app->getOsdManager()->hideColorBar();
  return( SUCCESS );
}

void CameraState::requestNodes(string session_location) {
  if (session_location != "") {
    ClientRegistry& reg = app->getNMMApplication()->getRegistry();
    reg.addRegistry(TCPAddress(session_location, ClientRegistry::default_port));
  }
  
  /*Begin> Create the request for the VISCACameraNode */
  NodeDescription camera_req;
  camera_req.setNodeName("VISCACameraReadNode");
  camera_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);

  if (!session_location.empty())
    camera_req.disableLocation();
  
  string location;
  MMBoxConfig* config = app->getMMBoxConfig();
  config->getValue( "camera_host", location );
  
  camera_req.setLocation( location);
  /*End< */
  
  NodeDescription demux_req;
  demux_req.setNodeName("MPEGDemuxNode");
  demux_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  
  NodeDescription decoder_req;
  decoder_req.setNodeName("MPEGVideoDecodeNode");
  decoder_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  decoder_req.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(200), new TInValue<string>("default")), INode::CONSTRUCTED);
  decoder_req.addNodeEvent(Event(ISwitch::setSwitchOutputsEnabled_event, new TInValue<bool>(true)), INode::ACTIVATED);

  //  NodeDescription adecoder_req;
  //  adecoder_req.setNodeName("MPEGAudioDecodeNode");
  //  adecoder_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  //  adecoder_req.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(200), new TInValue<string>("default")), INode::CONSTRUCTED);
  //  adecoder_req.addNodeEvent(Event(ISwitch::setSwitchOutputsEnabled_event, new TInValue<bool>(true)), INode::ACTIVATED);
  
  osdNode_nd = new NodeDescription("OSDManagerNode");
  osdNode_nd->setSharingType(NodeDescription::EXCLUSIVE);
  
  //  playback_nd = new NodeDescription("PlaybackNode");
  //  playback_nd->setSharingType(NodeDescription::EXCLUSIVE);

  /*Begin> Get the ClientRegistry and request Nodes */
  ClientRegistry& registry = app->getNMMApplication()->getRegistry();
  
  //RegistryLock lock(registry);
  graph.clear();
  graph.addEdge(camera_req, demux_req);
  graph.addEdge(demux_req,"mpeg_video0", decoder_req);
  //  graph.addEdge(demux_req,"mpeg_audio0", adecoder_req);
  graph.addEdge(decoder_req,*osdNode_nd,"stream1");
  //  graph.addEdge(adecoder_req,*playback_nd);
  graph.setINode(*osdNode_nd,*(app->getOsdNode()));
  //  graph.setINode(*playback_nd,*(app->getAudioNode()));

  graph.setReleaseFlag(*osdNode_nd,false);
  //  graph.setReleaseFlag(*playback_nd,false);

  registry.requestGraph(graph, RegistryName(session_location, ClientRegistry::default_port));
  camera = graph.getINode(camera_req);
}

void CameraState::cleanup() {
  /* Begin> Flushing the Display  */
  if (app->getDisplayNode()->isStarted()) {
    app->getDisplayNode()->stop();
    app->getDisplayNode()->flush();
  }
  /* End< */


  /* Begin> Release interfaces */
  if ( camera_interface ) {
    delete camera_interface;
    camera_interface = NULL;
  }
  
  if(osdNode_nd){
    delete osdNode_nd;
    osdNode_nd = NULL;
  }
  /* End< Release interfaces */
  try{
    app->getNMMApplication()->getRegistry().releaseGraph(graph);
  } catch(const Exception& e){
    GLOBAL_ERROR_STREAM( e << endl);
  }
  GLOBAL_DEBUG_STREAM("CameraStateCleanup done"<<endl);
}

Result CameraState::doInitialize( string ) {
  app->getOsdManager()->displayColorBar();
  return( SUCCESS );
}

Result CameraState::doUninitialize() {
  app->getOsdManager()->hideColorBar();
  return( SUCCESS );
}

Result CameraState::receiveMessage( string ) {
  app->getOsdManager()->deleteMessage();
  app->disconnectDefaultGraph();
  
  deactivate();
  goToLastState();
  return SUCCESS;
}

/* Switch to last state */
Result CameraState::Back() {
  GLOBAL_DEBUG_STREAM( "Camera: Back" << endl );
  GLOBAL_DEBUG_STREAM( "Switch to stateid:" << endl);
  
  return( goToLastState() );
}

void CameraState::showAlert(string title, string message) {
  app->getOsdManager()->displayMessage( title, message , "", "CANCEL" );
  changeState("OSDInputState");
}

Result CameraState::Stop() {
  camera_interface->ZoomStop(1);
  camera_interface->PanTiltDriveStop(1 /*camera_number*/, 0x18, 0x14);
  move_down = move_up = move_right = move_left = zoom_in = zoom_out = false;
  return SUCCESS;
}

Result CameraState::Down() {
  if (move_down) {
    Stop();
  }
  else {
    move_down = move_up = move_right = move_left = zoom_in = zoom_out = false;
    move_down = true;
    camera_interface->PanTiltDriveDown(1 /*camera_number*/,
                                       0x04 /*camera_state[camera_number].motion_speed*/,
                                       0x01);
  }
  return SUCCESS;
}
Result CameraState::Up() {
  if (move_up)
    Stop();
  else {
    move_down = move_up = move_right = move_left = zoom_in = zoom_out = false;
    move_up = true;
    camera_interface->PanTiltDriveUp(1 /*camera_number*/,
                                     0x04 /*camera_state[camera_number].motion_speed*/,
                                     0x01);
  }
  return SUCCESS;
}
Result CameraState::Left() {
  if (move_left)
    Stop();
  else {
    move_down = move_up = move_right = move_left = zoom_in = zoom_out = false;
    move_left = true;
    camera_interface->PanTiltDriveLeft(1 /*camera_number*/,
                                       0x04 /*camera_state[camera_number].motion_speed*/,
                                       0x01);
  }
  return SUCCESS;
}
Result CameraState::Right() {
  if (move_right)
    Stop();
  else {
    move_down = move_up = move_right = move_left = zoom_in = zoom_out = false;
    move_right = true;
    camera_interface->PanTiltDriveRight(1 /*camera_number*/,
                                        0x04 /*camera_state[camera_number].motion_speed*/,
                                        0x01);
  }
  return SUCCESS;
}
Result CameraState::colorKeyRed() {
  if (zoom_in)
    Stop();
  else {
    move_down = move_up = move_right = move_left = zoom_in = zoom_out = false;
    zoom_in = true;
    camera_interface->ZoomTeleVar(1, 0x03);
  }
  return SUCCESS;
}
Result CameraState::colorKeyGreen() {
  if (zoom_out) 
    Stop();
  else {
    move_down = move_up = move_right = move_left = zoom_in = zoom_out = false;
    zoom_out = true;
    camera_interface->ZoomWideVar(1, 0x03);
  }
  return SUCCESS;
}

void CameraState::updateColorBar() {
  app->getOsdManager()->displayColorBar(); 
  app->getOsdManager()->setColorBarText("Zoom in","Zoom out","","");
}

void CameraState::connectSinkSynchronizer(ISinkSynchronizer * tmp_sync) {
  IMultiAudioSinkSynchronizer_var ia_sync(tmp_sync->getParentObject()->getInterface<IMultiAudioSinkSynchronizer>());
  IMultiVideoSinkSynchronizer_var iv_sync(tmp_sync->getParentObject()->getInterface<IMultiVideoSinkSynchronizer>());

  /* Begin> Add the controller to the synchronizer */
  //    ISynchronizedSink_var av(app->getAudioNode()->getParentObject()->getInterface<ISynchronizedSink>());
  //    if (av.get()) {
  //      ISinkController_var controller (av->getController());
  //      av->setSynchronized(true);
  //      if(ia_sync.get()) { // add audio controller ...
  //        ia_sync->addAudioSinkController(controller.get());
  //      }
  //    }
  
    //av.reset(app->getDisplayNode()->getParentObject()->getInterface<ISynchronizedSink>());
  ISynchronizedSink_var av(app->getDisplayNode()->getParentObject()->getInterface<ISynchronizedSink>());
  if(iv_sync.get()) { // ... or video (unbuffered)
    ISinkController_var controller (av->getController());
    av->setSynchronized(true);
    iv_sync->addVideoSinkController(controller.get());
  }
}


void CameraState::disconnectSinkSynchronizer(ISinkSynchronizer* tmp_sync) {
  ISinkSynchronizer_var isync(tmp_sync->getParentObject()->getCheckedInterface<ISinkSynchronizer>());

  if (isync.get()) {
    //    ISynchronizedSink_var av(app->getAudioNode()->getParentObject()->getInterface<ISynchronizedSink>());
    //    if (av.get()) {
    //      ISinkController_var controller (av->getController());
    //      isync->removeSinkController(controller.get());
    //    }
    ISynchronizedSink_var av( app->getDisplayNode()->getParentObject()->getInterface<ISynchronizedSink>() );
    //    av.reset( app->getDisplayNode()->getParentObject()->getInterface<ISynchronizedSink>() );
    if (av.get()) {
      ISinkController_var controller (av->getController());
      isync->removeSinkController(controller.get());
    }
  }
}
