/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef MMBOX_DVB_HPP
#define MMBOX_DVB_HPP MMBOX_DVB_HPP

// the app
#include "../../main/application.hpp"
#include "nmm/base/registry/GraphDescription.hpp"

#include "nmm/interfaces/base/IUsageNotifyImpl.hpp"

// stored Interfaces
#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/interfaces/file/ITimeshifting.hpp"
#include "nmm/interfaces/video/osd/manager/ITextView.hpp"
#include "nmm/interfaces/video/tv/IIVTV.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuideProducer.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuide.hpp"

// need a local clock for timeshifting
#include "nmm/base/sync/TimedElement.hpp"

#include "nmm/misc/timer/TimerEvents.hpp"
#include "nmm/misc/tvtimer/TVTimerEntry.hpp"
#include "nmm/misc/tvmanager/TVManager.hpp"
#include "nmm/plugins/video/osd/manager/OSDElements/listunit.hpp"
#include "nmm/plugins/video/osd/manager/OSDElements/tableunit.hpp"

namespace NMM {

  class DvbState; //forward for DDVB-State

  class DvbStateEPG 
    : public ThreadedObject
  {
  private:

    enum mode_t{UNKNOWN = -1, SHORT=0, NOW, NEXT};
    MMBoxApplication *app;

    ProgramDescription currentShow;    
    string cName;

    list<ProgramDescription> pdList;

    DvbState* state; //pointer to DvbState to access its widgets
 
    mode_t currentMode;

    string tz;

    TEDObject1<DvbStateEPG, string> newShow_listener;
    ITVGuide* iepg;

    ThreadMutex mutex;
  public:
    DvbStateEPG(MMBoxApplication *application, DvbState *dvbState);
    ~DvbStateEPG();

    Result updateShort(string channelName);
    Result updateShort(ProgramDescription pd);
    Result updateNow();
    Result updateNext();
    Result updateOverviewDescription(unsigned int index);
    Result updateOverview(ProgramDescription &pd);

    ProgramDescription& getCurrentShow(){return currentShow;}
    ProgramDescription  getShow(unsigned int index);
  private:
    Result updateShortWindows();
    Result updateNowWindows();
    Result updateNextWindows();
    void run();
    string formatTimeString(string s);

    Result newShow(string& channel);

  };



  /** MMBox DVB State.
   * @bug Seeking to the begin of the progressbar in timeshifting mode can cause a segfault.
   * @bug Switching from timeshifting mode to live mode shows the epg window at the bottom of the screen
   * @bug Playback in timehsifting mode is out of sync.
   * @bug Switching from timeshifting mode to live mode does not update the OSD.
   * @bug exiting TS-Mode to Live-Mode causes the MMBox to quit without error-message
   */
  class DvbState
    : public MMBoxState,
      public TimedElement,
      public ThreadedObject,
      public IUsageNotifyImpl
  {
    friend class DvbStateEPG;
  private:
    enum epgDisplayMode_t{NOW=0, NEXT, BACK};
    enum stateMode_t{LIVETV=0, TIMERPLAYBACK, SHAREDTV};
  public:

    DvbState(string, MMBoxApplication*);
    ~DvbState();

    Result doCreate(string);
    Result doKill();

    Result doInitialize(string);
    Result doUninitialize();

    Result Up();
    Result Down();
    Result Left();
    Result Right();

    Result Back();
    Result Pause();
    Result Rewind();
    Result Forward();
    Result Return(); // Return means Play
    Result Stop();  // switch to LIVE-mode  (and clear filebuffer)
    Result specialKey01();

    Result colorKeyRed();
    Result colorKeyGreen();
    Result colorKeyYellow();
    Result colorKeyBlue();
    
    Result Record();
    Result OSDKey();

    Result numKey0();
    Result numKey1();
    Result numKey2();
    Result numKey3();
    Result numKey4();
    Result numKey5();
    Result numKey6();
    Result numKey7();
    Result numKey8();
    Result numKey9();

    // EventListener methods
    Result live();
    Result timeshiftingBegin();
    Result startTrack(string&);
    Result setProgress(Time&);
    Result setChannel(int &i);
    Result switchingCard();
    Result sharedCard();
    Result skippingChannel();

    void synchronize();
    void disconnectSinkSynchronizer(ISinkSynchronizer* tmp_sync);
    void connectSinkSynchronizer(ISinkSynchronizer* tmp_sync);

    Result setTvChannel(int &c);
    Result setTVTimerPlayback(TVTimerEntry &entry);
    Result setLiveTV();
    Result receiveMessage( string message );

    Result becomeMaster();
    Result becomeSlave();

    TVChannelList* getChannelConfig();
    //override function from MMBoxstate-class
    string getSessionLocation();

  private:
    stateMode_t state_mode;    
    epgDisplayMode_t overview_mode;

    int timer_id;

    bool switchingCards;
    // FF & REW
    int fast_forward_speed;
    int rewind_speed;
    
    ///! flag for storing if OSD was enabled when uninitialized()
    bool osd_enabled;

    // EventListener objects
    TEDObject1<DvbState, Time> setProgress_listener;
    TEDObject0<DvbState> live_listener;
    TEDObject0<DvbState> timeshiftingBegin_listener;
    
    TEDObject1<DvbState, int> setChannel_listener;
    TEDObject0<DvbState> switchingCard_listener;
    TEDObject0<DvbState> sharedCard_listener;
    TEDObject0<DvbState> skippingChannel_listener;

    GraphDescription graph;
    NodeDescription* video_mpeg_nd;
    NodeDescription* ts_nd;

    NodeDescription *mpa_dec_nd;
    NodeDescription *last_video_nd;

    // Interfaces
    ITimeshifting*    itimeshift;
    
    // widgets
    ITextView* text;
    IWindow* win;

    
    //EPG-widgets
    TableUnit* epgWin;
    ListUnit* detailWin;
    
    TableUnit* overviewTable;
    ListUnit*  overviewDetails;

    ISinkSynchronizer_var sink_synchronizer;
    
    /** Timeshifting offset to local time
     */
    Interval ts_offset;
    ProgramDescription currentShow;    

    TVManager *tvmanager;

    unsigned int currentChannel;

    DvbStateEPG *epg;

    bool isMaster;

    bool channelSkipped;
    bool overview_table_selected;

    bool deinterlace;

    ThreadMutex eventMutex;
  private:

    // private helpers
    ///! this method cleans up by releasing and deleting all allocated
    //   stuff
    void cleanup();
    void updateOSD();

    ///! these to methods allocate and deallocate the OSD window
    void createWindow();
    void deleteWindow();
    ///! these to methods allocate and deallocate the boot window
    void createBootWindow();
    void deleteBootWindow();

    void disconnectMPEG();
    void connectMPEG();

    //funtions for showing/hiding window with current and following show
    Result showShortEPGInfo();
    void hideShortEPGInfo();
    Result switchChannelTo(int i);

    time_t stringToTime(string s);    

    void run();

    void switchOverview(epgDisplayMode_t mode);

    Result connectGlobalSinks();
    void restartGraph();
  };

}

#endif //MMBOX_DVB_HPP
