/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass <dama@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iostream>
#include "nmm/base/connect.hpp"
#include "nmm/utils/utils.hpp"

// for remote registries
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

#include "nmm/interfaces/video/osd/manager/IOSDManager.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/IBufferManager.hpp"
#include "nmm/interfaces/av/mpeg/IMPEGDemux.hpp"
#include "nmm/interfaces/base/IBufferManager.hpp"
#include "nmm/interfaces/base/sync/ISwitch.hpp"
#include "nmm/interfaces/base/registry/IPluginManager.hpp"
#include "nmm/interfaces/base/IReferenceCount.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"

#include "dvd.hpp"

using namespace NMM;

DvdState::DvdState( string id,  MMBoxApplication* _app ) : MMBoxState( id, _app ) {
  /* get pointer to global nodes */
  osd_node   = app->getOsdNode();
  display    = app->getDisplayNode();
  audio_play = app->getAudioNode();

  /* nodes that are needed for DVD playing */
  dvd      = NULL;
  demux    = NULL;
  mpg2yuv  = NULL;
  ac3_dec  = NULL;
  overlay  = NULL;
  spudec   = NULL;

  dvd_interface = NULL;
  dvdnav = NULL;

  ac3_passthrough = false;
  pause = false;

  device = "/dev/cdrom";
}

DvdState::~DvdState() {
}

Result DvdState::doCreate( string session_location ) {
  GLOBAL_DEBUG_STREAM( "DVD: initialize: " << endl);
  
  MMBoxConfig* config = app->getMMBoxConfig();
  config->getValue( "useac3", ac3_passthrough );
  config->getValue( "dvd_device", device );
  
  pause = false;
  
  /* Begin> Create default graph */
  if (FAILURE == app->createDefaultGraph(background)) {
    return FAILURE;
  }
  
  if (FAILURE == app->startDefaultGraph()) {
    return FAILURE;
  }
  /* End< Create default graph */

  /* Begin> Request nodes */
  try {
    app->disconnectDefaultGraph();
    requestNodes(session_location);

    dvd_interface = dvd->getParentObject()->getCheckedInterface<IDVD>();
    dvdnav = dvd->getParentObject()->getCheckedInterface<IDVDNav>();
    
    app->setSynchronized( true );
    app->getOsdManager()->setPause(false);
    /* Disconnect the application's synchronizer from auio and video sink */
    disconnectSinkSynchronizer( app->getSynchronizer() );
    /* Use the given synchronizer from graph description ... */
    sink_synchronizer.reset( graph.getISession()->getSinkSynchronizer() );
    /* and connect it with our audio and video sink */
    connectSinkSynchronizer( sink_synchronizer.get() );

    /* realize the graph */
    graph.realizeGraph();
  } catch (const Exception& e){
    GLOBAL_ERROR_STREAM("RealizeGraph failed: " << e << endl);
    showAlert("Input Error", "Failed to initialize DVD Player");
    app->getOsdManager()->update();
    cleanup();
    return SUCCESS;
  }

  try{
    graph.activateGraph();
    app->getDisplayNode()->reachStarted();
    graph.startGraph();
  } catch (...) {
    showAlert("Input Error", "Failed to start DVD Player");
    app->getOsdManager()->update();
    cleanup();
    return FAILURE;
  }
  //app->getOsdManager()->update();
  /* Begin> Start nodes */
  
  /* specify that graph is the currently used graph description */
  setCurrentGraphDescription(graph);
  
  return( SUCCESS );
}

Result DvdState::doKill() {
  cleanup();
  if (app->getOsdNode()->isInputConnected("stream1")) {
    // if OSD's video input is still connected after cleanup, the
    // default graph must be connected
    app->disconnectDefaultGraph();
    // this leaves the osd in INITIALIZED where the other MMBoxStates
    // expect it.
  }
  
  return( SUCCESS );
}

Result DvdState::doInitialize( string ) {
  return( SUCCESS );
}

Result DvdState::doUninitialize() {
  return( SUCCESS );
}

Result DvdState::receiveMessage( string ) {
  app->getOsdManager()->deleteMessage();
  app->disconnectDefaultGraph();
  
  deactivate();
  goToLastState();
  return SUCCESS;
}

/* Switch to last state */
Result DvdState::Back() {
  GLOBAL_DEBUG_STREAM( "DVD: Back" << endl );
  GLOBAL_DEBUG_STREAM( "Switch to stateid:" << endl);
  
  return( goToLastState() );
}

Result DvdState::Stop() {
  kill();
  Back();
  
  return( SUCCESS );
}

Result DvdState::Pause() {
  if (!pause) {
    graph.getISession()->getSinkSynchronizer()->pause();
    pause = true;
  } else {
    graph.getISession()->getSinkSynchronizer()->wakeup();
    pause = false;
  }
  
  return( SUCCESS );
}

Result DvdState::Up() {
  GLOBAL_DEBUG_STREAM( "DVD: Up" << endl );

  if (dvdnav) {
    dvdnav->buttonUp();
  }
  
  return(SUCCESS);
}

Result DvdState::Down() {
  GLOBAL_DEBUG_STREAM( "DVD: Down" << endl );
  
  if (dvdnav) {
    dvdnav->buttonDown();
  }
  
  return(SUCCESS);
}

Result DvdState::Right() {
  GLOBAL_DEBUG_STREAM( "DVD: Right" << endl );
  
  if (dvdnav) {
    dvdnav->buttonRight();
  }
  
  return(SUCCESS);
}

Result DvdState::Left() {
  GLOBAL_DEBUG_STREAM( "DVD: Left" << endl );
  
  if (dvdnav) {
    dvdnav->buttonLeft();
  }

  return( SUCCESS );
}

Result DvdState::Return() {
  GLOBAL_DEBUG_STREAM( "DVD: Return" << endl );
  
  if (dvdnav) {
    dvdnav->activateSelectedButton();
  }
  
  return(SUCCESS);
}

Result DvdState::Forward() {
  int title, chapter, angel;
  angel = 1;
  title = dvd_interface->getCurrentTitle();
  chapter = dvd_interface->getCurrentChapter();
  chapter++;
  dvd_interface->sendFlush();
  dvd_interface->select( title, chapter, angel );

  return SUCCESS;
}

Result DvdState::Rewind() {
  int title, chapter, angel;
  angel = 1;
  title = dvd_interface->getCurrentTitle();
  chapter = dvd_interface->getCurrentChapter();
  chapter--;
  dvd_interface->sendFlush();
  dvd_interface->select( title, chapter, angel );

  return SUCCESS;
}

Result DvdState::specialKey01() {
  GLOBAL_DEBUG_STREAM( "DVD: goUp" << endl );

  if (dvdnav) {
    dvdnav->goUp();
  }

  return(SUCCESS);
}

void DvdState::requestNodes(string session_location) {
  if (session_location != "") {
    ClientRegistry& reg = app->getNMMApplication()->getRegistry();
    reg.addRegistry(TCPAddress(session_location, ClientRegistry::default_port));
  }

  /*Begin> Create the request for the DVDReadNode */
  NodeDescription dvd_req;
  dvd_req.setNodeName("DVDNavReadNode");
  dvd_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  dvd_req.addNodeEvent(Event(IDVD::setDevice_event, new TInValue<string>(device)),INode::CONSTRUCTED);

  //
  dvd_req.setLocation(session_location);
  /*End< */

  /*Begin> Create the request for the MPEGDemuxNode */
  NodeDescription demux_req;
  demux_req.setNodeName("MPEGDemuxNode");
  demux_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  //demux_req.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(100), new TInValue<string>("default")), INode::CONSTRUCTED);
  demux_req.addNodeEvent(Event(IMPEGDemux::setDVDMode_event, new TInValue<bool>(true)),INode::INITIALIZED);
  /*End< */

  /*Begin> Create the request for the MPEGVideoDecodeNode */
  NodeDescription mpg2yuv_req;
  mpg2yuv_req.setNodeName("MPEGVideoDecodeNode");
  mpg2yuv_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  mpg2yuv_req.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(200), new TInValue<string>("default")), INode::CONSTRUCTED);
  /*End< */

  /*Begin> Create the request for the AC3DecodeNode */
  NodeDescription ac3_dec_req;
  ac3_dec_req.setNodeName("AC3DecodeNode");
  ac3_dec_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  //ac3_dec_req.addNodeEvent(Event(ISwitch::setSwitchOutputsEnabled_event, new TInValue<bool>(true)), INode::ACTIVATED);
  ac3_dec_req.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(200), new TInValue<string>("default")), INode::CONSTRUCTED);
  /*End< */

  // Begin> Hack for synchronized playback : repplix, Cebit 2005
   NodeDescription mpeg_audio_encode_nd;
  mpeg_audio_encode_nd.setNodeName("MPEGAudioEncodeNode");
  mpeg_audio_encode_nd.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  mpeg_audio_encode_nd.addNodeEvent(Event(ISwitch::setSwitchOutputsEnabled_event, new TInValue<bool>(true)), INode::ACTIVATED);
   mpeg_audio_encode_nd.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(100), new TInValue<string>("default")), INode::CONSTRUCTED);
  mpeg_audio_encode_nd.setLocation("localhost");

  NodeDescription mpeg_audio_decode_nd;
  mpeg_audio_decode_nd.setNodeName("MPEGAudioDecodeNode");
  mpeg_audio_decode_nd.setSharingType(NodeDescription::EXCLUSIVE);
  mpeg_audio_decode_nd.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(1), new TInValue<string>("default")), INode::CONSTRUCTED);

  /* Output format of MP3 Encoder */
  Format mp3_encode_out_format;
  mp3_encode_out_format.addIntParamValue(Format::bitrate_param, 64000);
  // End > Hack

  /*Begin> Create the request for the YUVDeInterlaceNode */
  NodeDescription deinter_req;
  deinter_req.setNodeName("YUVDeInterlaceNode");
  deinter_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  deinter_req.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(3), new TInValue<string>("default")), INode::CONSTRUCTED);
  /*End< */

  /*Begin> Create the request for the OverlayNode */
  NodeDescription overlay_req;
  overlay_req.setNodeName("OverlayNode");
  overlay_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  overlay_req.addNodeEvent(Event(ISwitch::setSwitchOutputsEnabled_event, new TInValue<bool>(true)), INode::ACTIVATED);
  overlay_req.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(5), new TInValue<string>("default")), INode::CONSTRUCTED);
  /*End< */

  /*Begin> Create the request for the SPUDecodeNode */
  NodeDescription spudec_req;
  spudec_req.setNodeName("SPUDecodeNode");
  spudec_req.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
  spudec_req.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<unsigned int>(200), new TInValue<string>("default")), INode::CONSTRUCTED);
  /*End< */

  osdNode_nd = new NodeDescription("OSDManagerNode");
  osdNode_nd->setSharingType(NodeDescription::EXCLUSIVE);

  playback_nd = new NodeDescription("PlaybackNode");
  playback_nd->setSharingType(NodeDescription::EXCLUSIVE);
  
  /*Begin> Get the ClientRegistry and request Nodes */
  ClientRegistry& registry = app->getNMMApplication()->getRegistry();

  //RegistryLock lock(registry);
  graph.clear();
  graph.addEdge(dvd_req,demux_req);
  graph.addEdge(demux_req,"mpeg_video0",mpg2yuv_req);
  if (!ac3_passthrough) {
    cerr<<"no ac3 passthrough"<<endl;
    graph.addEdge(demux_req,"ac3_audio0",ac3_dec_req);

    graph.addEdge(ac3_dec_req, *playback_nd);

  } else {
    graph.addEdge(demux_req,"ac3_audio0",*playback_nd);
  }
  graph.addEdge(demux_req,"spu0",spudec_req);
  graph.addEdge(mpg2yuv_req,deinter_req);
  graph.addEdge(deinter_req,overlay_req,"background");
  //  graph.addEdge(mpg2yuv_req,overlay_req,"background");
  graph.addEdge(spudec_req,overlay_req,"foreground");
  graph.addEdge(overlay_req,*osdNode_nd,"stream1");



  graph.setINode(*osdNode_nd,*osd_node);
  graph.setINode(*playback_nd,*audio_play);

  graph.setReleaseFlag(*osdNode_nd,false);
  graph.setReleaseFlag(*playback_nd,false);

  registry.requestGraph(graph, RegistryName(session_location, ClientRegistry::default_port));

  dvd = graph.getINode(dvd_req);
  //  deinter = graph.getINode(deinter_req);
  mpg2yuv = graph.getINode(mpg2yuv_req);
  demux = graph.getINode(demux_req);
  if (!ac3_passthrough) {
    ac3_dec = graph.getINode(ac3_dec_req);
  }
  overlay = graph.getINode(overlay_req);
  spudec = graph.getINode(spudec_req);
}

  /* ***** REMOVED FOR CEBIT 2004 ***************************************************************
  {
    IBufferManager_var bm1(display->getParentObject()->getCheckedInterface<IBufferManager>());
    IBufferManager_var bm2(mpg2yuv->getParentObject()->getCheckedInterface<IBufferManager>());
    bm2->useBufferManager(*bm1.get());
  }
  ********************************************************************************************* */


void DvdState::cleanup() {
  if (sink_synchronizer.get()) {
    disconnectSinkSynchronizer( sink_synchronizer.get() );
    connectSinkSynchronizer( app->getSynchronizer() );
  }

  /* Begin> Flushing the Display and the PlaybackNode */
  if (app->getDisplayNode()->isStarted()) {
    app->getDisplayNode()->stop();
    app->getDisplayNode()->flush();
  }
  if (app->getAudioNode()->isStarted()) {
    app->getAudioNode()->stop();
    app->getAudioNode()->flush();
  }
  /* End< */


  /* Begin> Release interfaces */
  if ( dvd_interface ) {
    delete dvd_interface;
    dvd_interface = NULL;
  }

  if ( dvdnav ) {
    delete dvdnav;
    dvdnav = NULL;
  }

  if(playback_nd){
    delete playback_nd;
    playback_nd = NULL;
  }

  if(osdNode_nd){
    delete osdNode_nd;
    osdNode_nd = NULL;
  }
  /* End< Release interfaces */
  try{
    app->getNMMApplication()->getRegistry().releaseGraph(graph);
  } catch(const Exception& e){
    GLOBAL_ERROR_STREAM( e << endl);
  }
  GLOBAL_DEBUG_STREAM("DVDStateCleanup done"<<endl);
}

void DvdState::showAlert(string title, string message) {
  app->getOsdManager()->displayMessage( title, message , "", "CANCEL" );
  changeState("OSDInputState");
}


void DvdState::connectSinkSynchronizer(ISinkSynchronizer * tmp_sync) {
  IMultiAudioSinkSynchronizer_var ia_sync(tmp_sync->getParentObject()->getInterface<IMultiAudioSinkSynchronizer>());
  IMultiVideoSinkSynchronizer_var iv_sync(tmp_sync->getParentObject()->getInterface<IMultiVideoSinkSynchronizer>());

  /* Begin> Add the controller to the synchronizer */
  ISynchronizedSink_var av(app->getAudioNode()->getParentObject()->getInterface<ISynchronizedSink>());
  if (av.get()) {
    ISinkController_var controller (av->getController());
    av->setSynchronized(true);
    if(ia_sync.get()) { // add audio controller ...
      ia_sync->addAudioSinkController(controller.get());
    }
  }

  av.reset(app->getDisplayNode()->getParentObject()->getInterface<ISynchronizedSink>());
  if(iv_sync.get()) { // ... or video (unbuffered)
    ISinkController_var controller (av->getController());
    av->setSynchronized(true);
    iv_sync->addVideoSinkController(controller.get());
  }
}

void DvdState::disconnectSinkSynchronizer(ISinkSynchronizer* tmp_sync) {
  ISinkSynchronizer_var isync(tmp_sync->getParentObject()->getCheckedInterface<ISinkSynchronizer>());

  if (isync.get()) {
    ISynchronizedSink_var av(app->getAudioNode()->getParentObject()->getInterface<ISynchronizedSink>());
    if (av.get()) {
      ISinkController_var controller (av->getController());
      isync->removeSinkController(controller.get());
    }
    av.reset( app->getDisplayNode()->getParentObject()->getInterface<ISynchronizedSink>() );
    if (av.get()) {
      ISinkController_var controller (av->getController());
      isync->removeSinkController(controller.get());
    }
  }
}
