/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <stdio.h>
#include <iostream>
#include "grabdvd.hpp"
#include "nmm/utils/utils.hpp"

#include "nmm/base/connect.hpp"

#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/stats/IEncodeStats.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/IThreadedNode.hpp"
#include "nmm/interfaces/av/IAVMux.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"
#include "nmm/interfaces/general/progress/IProgress.hpp"

#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"
#include "nmm/base/registry/PluginManager.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

static int dbg_serial=0;

namespace NMM {
  
  DVDGrabState::DVDGrabState( string id, MMBoxApplication* _app )
    : MMBoxState( id, _app ),
      dvd(NULL),
      demux(NULL),
      video_dec(NULL),
      spu_dec(NULL),
      overlay(NULL),
      deinterlacer(NULL),
      scaler(NULL),
      video_enc(NULL),
      ac3_dec(NULL),
      audio_enc(NULL),
      av_mux(NULL),
      writer(NULL),
      dvd_interface(NULL),
      startTrack_listener(this, &DVDGrabState::startTrack),
      endTrack_listener(this, &DVDGrabState::endTrack),
      setFrameCount_listener(this, &DVDGrabState::setFrameCount),
      setProgress_listener(this, &DVDGrabState::setProgress),
      device("/dev/cdrom"),
      filename("dvddump.avi"),
      root_path("/"),
      dest_dir(""),
      audio_track(0),
      title(1),
      chapter(1),
      subtitle_stream(0),
      video_bitrate(1200),
      audio_bitrate(129),
      audio_quality(2),
      ac3_writethrough(false),
      no_subtitles(true),
      video_deinterlace(false),
      video_scale(true),
      auto_aspect(true),
      use_ffmpeg(false),
      state(TITLE_SELECT),
      osd_node(NULL),
      manager(NULL),
      text(NULL),
      plane(NULL),
      border(NULL),
      scroll(NULL),
      header(NULL),
      win(NULL),
      dlunit(NULL),
      current_progress(0),
      config(NULL)
  {
    setState(TITLE_SELECT);
    
    osd_node = app->getOsdNode();
    manager = app->getOsdManager();
    config = app->getMMBoxConfig();
  }
  
  DVDGrabState::~DVDGrabState() {    
    if ( dlunit ) {
      delete dlunit;
      dlunit = NULL;
    }
  }

  Result DVDGrabState::setFrameCount(unsigned int& count, float& rate) {
    string text = string("Encoding frame: ")+count+string(" Framerate: ")+rate;
    header->setText(text);
    win->update();
    manager->update();
    return SUCCESS;
  }
  
  Result DVDGrabState::setProgress(u_int64_t& val, u_int64_t& max) {
    current_progress = Rational(val, max);
    
    if (hasFocus()) {
      manager->setProgress(current_progress.getNumerator());
      manager->setMaxProgress(current_progress.getDenominator());
      manager->updateProgress();
      manager->update();
    }
    
    return SUCCESS;
  }
  
  Result DVDGrabState::doCreate(string) {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": BEGIN> create" << endl );
  
    title = 1;
    chapter = 1;
  
    MMBoxConfig* config = app->getMMBoxConfig();
    if (!(config->getValue("grabdvdmp3bitrate",audio_bitrate))) {
      audio_bitrate = 128;
    }
    
    /* Begin> options window */
    if (dlunit) {
      delete dlunit;
      dlunit = NULL;
    }
    IWidgetFactory& factory = app->getOSDWidgetFactory();
    dlunit = new DoubleListUnit( manager, &factory, "medium" );
    dlunit->setFirstColumnWidth( 80 );
    
    dlunit->setDimension( 520, 400 );
    dlunit->setPosition( 70, 50 );
    dlunit->setColors( 0x80ff0000,
                       0x80ff0000,
                       0x800000ff,
                       0x800000ff,
                       0x800000ff );
    dlunit->selectWindow();
    /* End< options window */
    
    /* fill up dlistunit with added values */
    for (map<string,string>::const_iterator s = shown_options.begin();
         s != shown_options.end();
         ++s) {
      const MMBoxConfig::Setting* setting;
      if (!config->getSetting(s->first, setting)) {
        GLOBAL_ERROR_STREAM("Unknown setting: " << s->first << endl);
        continue;
      }
      string val(s->second.empty() ? setting->getDescription() : s->second);
      string shortform(s->first);
      const list<string>& values(setting->getValues());
      
      DoubleListUnit::line_t line;
      
      /* add all values */
      string default_value;
      for (list<string>::const_iterator i = values.begin();
           i != values.end();
           ++i) {
        string value = *i;
        line.push_back( value );
        if (default_value.empty()) {
          default_value = value;
        }
      }
      
      /* add human readable configuration text to list */
      dlunit->addLine( val, line, shortform );
    }
  
    if ( isBackgroundFree() ) {
      if (FAILURE == app->createDefaultGraph( background )) {
        GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< create (FAILURE)" << endl );
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< create (FAILURE)" << endl );
        return FAILURE;
      }
    }
  
    createWindow();
  
    GLOBAL_DEBUG_STREAM( "DVDGrabState " << stateid << ": open DVD" << endl );
    if (FAILURE == initDVD()) {
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< create (SUCCESS)" << endl );
      return SUCCESS;
    }
  
    if (FAILURE == startTitleMenu()) {
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< create (FAILURE)" << endl );
      return FAILURE;
    } else {
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< create (SUCCESS)" << endl );
      return SUCCESS;
    }
  }

  Result DVDGrabState::startGrabbing() {  
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": BEGIN> startGrabbing" << endl );
    
    header->setText( "Initializing..." );
    text->deleteAllLines();
    string buffer = string("Title ") + title + " Chapter " + chapter;
    text->addLine( buffer );
    text->selectLine(1);
    win->setVisible(true);
    dlunit->hide();
    win->update();
    manager->update();
    
    /* read configuration */
    readConfig();
    
    /* select title and chapter */
    dvd_interface->select(title, chapter, 1);
    
    /* create directory from current date */
    time_t today = time(NULL);
    tm today_tm;
    localtime_r(&today, &today_tm);
    char buffer2[128];
    snprintf(buffer2, 128, "%04d-%02d-%02d", today_tm.tm_year + 1900, today_tm.tm_mon + 1, today_tm.tm_mday);
    string today_str = buffer2;
    
    string base_dir = root_path + string("/") + dest_dir;
    string today_dir = base_dir + string("/") + today_str;
    
    /* create filename (title and chapter) */
    filename = string(buffer) + ".avi";
    filename = today_dir + "/" + filename;
    
    if (!DirReader::makeDir(base_dir)) {
      GLOBAL_ERROR_STREAM("DVDGrabState::doCreate() cannot make dir " << base_dir << endl);
      return FAILURE;
    }
    
    if (!DirReader::makeDir(today_dir)) {
      GLOBAL_ERROR_STREAM("DVDGrabState::doCreate() cannot make dir " << today_dir << endl);
      return FAILURE;
    }
    
    /* Begin> Create nodes */
    try {
    
      createNodes();

    } catch(const Exception& e) {
    
      GLOBAL_ERROR_STREAM("DVDGrabState: " << stateid << ": " << e << " while creating nodes" << endl);
    
      deleteNodes();

      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startGrabbing (FAILURE)" << endl );
      return FAILURE;
    
    } catch(...) {
    
      GLOBAL_ERROR_STREAM("DVDGrabState: " << stateid << ": something strange happened while creating nodes" << endl);
    
      // don't know how to cleanup in this situation
    
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startGrabbing (FAILURE)" << endl );
      return FAILURE;
    
    }
    /* End< Create nodes */
  
    /* Begin> Init and configure nodes */
    try {
    
      initNodes();
    
    } catch(const Exception& e) {
    
      GLOBAL_ERROR_STREAM("DVDGrabState: " << stateid << ": " << e << " while initializing nodes" << endl);
    
      deinitNodes();
      deleteNodes();
    
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startGrabbing (FAILURE)" << endl );
      return FAILURE;
    
    } catch(...) {
    
      GLOBAL_ERROR_STREAM("DVDGrabState: " << stateid << ": something strange happened while initializing nodes" << endl);
    
      // don't know how to cleanup in this situation
    
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startGrabbing (FAILURE)" << endl );
      return FAILURE;
    
    }
    /* End< Init and configure nodes */
    
    /* Begin> Connect nodes */
    try {
    
      connectNodes();
      lowerNodesPriority();
    
    } catch(const Exception& e) {
    
      GLOBAL_ERROR_STREAM("DVDGrabState: " << stateid << ": " << e << " while connecting nodes" << endl);
    
      disconnectNodes();
      deinitNodes();
      deleteNodes();
    
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startGrabbing (FAILURE)" << endl );
      return FAILURE;
    
    } catch(...) {
    
      GLOBAL_ERROR_STREAM("DVDGrabState: " << stateid << ": something strange happened while connecting nodes" << endl);
    
      // don't know how to cleanup in this situation
    
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startGrabbing (FAILURE)" << endl );
      return FAILURE;
    
    }
    /* End< Connect nodes */
  
    /* Begin> Start nodes */
    try {
    
      startNodes();
    
    } catch(const Exception& e) {
    
      GLOBAL_ERROR_STREAM("DVDGrabState: " << stateid << ": " << e << " while starting nodes" << endl);
    
      stopNodes();
      disconnectNodes();
      deinitNodes();
      deleteNodes();
    
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startGrabbing (FAILURE)" << endl );
      return FAILURE;
    
    } catch(...) {
    
      GLOBAL_ERROR_STREAM("DVDGrabState: " << stateid << ": something strange happened while starting nodes" << endl);
    
      // don't know how to cleanup in this situation
    
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startGrabbing (FAILURE)" << endl );
      return FAILURE;
    
    }
    /* End< Start nodes */
    
    header->setText( "Grabbing..." );
    win->update();
    manager->displayProgressBar( "Grabbing DVD Title" );
    manager->setProgress(current_progress.getNumerator());
    manager->setMaxProgress(current_progress.getDenominator());
    manager->updateProgress();
    manager->update();
    setState(GRABBING);
    
    writer->getParentObject()->registerEventListener(ITrack::endTrack_event, &endTrack_listener);
    writer->getParentObject()->registerEventListener(ITrack::startTrack_event, &startTrack_listener);
    
    registerProgressListener();
    
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startGrabbing (SUCCESS)" << endl );
    return( SUCCESS );
  }

  Result DVDGrabState::startTitleMenu() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": BEGIN> startTitleMenu [" << dbg_serial++ << "]" << endl );
  
    setState(TITLE_SELECT);
  
    text->deleteAllLines();
    for(int i=1; i<=num_titles; i++) {
      string buffer = string("Title ") + i + " (" + num_chapters[i] + " chapters)";
      text->addLine(buffer);
    }
    text->selectLine(title);
  
    header->setText( "Select Title" );
    win->setVisible(true);
    dlunit->hide();
    win->update();
    manager->update();
  
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startTitleMenu (SUCCESS) [" << dbg_serial++ << "]" << endl );
    return SUCCESS;
  }

  Result DVDGrabState::startChapterMenu() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": BEGIN> startChapterMenu" << endl );
  
    setState(CHAPTER_SELECT);
  
    if (chapter < 1 || chapter > num_chapters[title]) {
      chapter = 1;
    }
  
    text->deleteAllLines();
    for(int i=1; i<=num_chapters[title]; i++) {
      string buffer = string("Chapter ") + i;
      text->addLine(buffer);
    }
    text->selectLine( chapter );
  
    header->setText( "Select Chapter" );
    win->setVisible(true);
    dlunit->hide();
    win->update();
    manager->update();
  
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startChapterMenu (SUCCESS)" << endl );
    return SUCCESS;
  }
  
  Result DVDGrabState::startOptionsMenu() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": BEGIN> startOptionsMenu" << endl );
    
    setState(OPTIONS);
    
    // XXX: implement options
    win->setVisible(false);
    
    dlunit->setHeader( "DVD Grab Options" );
    
    initOptionsList();
    
    dlunit->show();
    dlunit->update();
    
    manager->update();
    
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< startOptionsMenu (SUCCESS)" << endl );
    return SUCCESS;
  }
  
  Result DVDGrabState::doKill() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": BEGIN> kill" << endl );
    
    if ( writer ) {
      writer->getParentObject()->removeEventListener(&startTrack_listener);
      writer->getParentObject()->removeEventListener(&endTrack_listener);
    }
    
    manager->hideProgressBar();
    
    if (dlunit) {
      dlunit->hide();
      delete dlunit;
      dlunit = NULL;
    }
    
    deRegisterProgressListener();
    
    stopNodes();
    flushNodes();
    disconnectNodes();
    deinitNodes();
    deleteNodes();
  
    deinitDVD();
  
    deleteWindow();

    if ( isBackgroundFree() ) app->disconnectVideoGraph();  
  
    setState( TITLE_SELECT );

    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< kill" << endl );  

    return SUCCESS;
  }

  Result DVDGrabState::doInitialize( string ) {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": BEGIN> initialize" << endl );
  
    if ( dvd == NULL ) {
      GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< initialize (SUCCESS)" << endl );
      return( SUCCESS );
    }
    
    if ( isBackgroundFree() ) {
      if (FAILURE == app->createDefaultGraph( background )) {
        GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< initialize (FAILURE)" << endl );
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< initialize (FAILURE)" << endl );
        return FAILURE;
      }
    }
    
    manager->displayProgressBar( "Grabbing DVD Title" );
    manager->setProgress(current_progress.getNumerator());
    manager->setMaxProgress(current_progress.getDenominator());
    manager->updateProgress();
    
    win->setVisible(true);
    win->update();
    manager->update();
  
    return( SUCCESS );
  }

  Result DVDGrabState::doUninitialize() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": BEGIN> uninitialize" << endl );
    
    manager->hideTime();
    if ( win ) win->setVisible(false);
  
    if ( isBackgroundFree() && dvd ) { app->disconnectVideoGraph(); }; // XXX
    
    manager->hideProgressBar();
    
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< uninitialize (SUCCESS)" << endl );
    return( SUCCESS );
  }

  Result DVDGrabState::receiveMessage( string ) {
    //cerr << "Grab message: " << message << endl;

    manager->deleteMessage();

    deactivate();
    goToLastState();

    return SUCCESS;
  }

  void DVDGrabState::registerProgressListener() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": register progress listener" << endl );
    
    if ( video_enc ) {
      if (writer) {
        writer->getParentObject()->registerEventListener(IEncodeStatsListener::setFrameCount_event, &setFrameCount_listener);
      }
      
      IEncodeStats_var encode_stats(video_enc->getParentObject()->getInterface<IEncodeStats>());
      if (encode_stats.get()) {
        encode_stats->sendFrameCount(true);
      }
    }
    
    if ( dvd ) {
      //if (writer) {
        dvd->getParentObject()->registerEventListener(IProgressListener::setProgress_event, &setProgress_listener);
        //}
      
      IProgress_var progress(dvd->getParentObject()->getInterface<IProgress>());
      if (progress.get()) {
        progress->sendProgressInformation(true);
      }
    }
  }


  void DVDGrabState::deRegisterProgressListener() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": deregister progress listener" << endl );
    
    if ( video_enc ) {
      if (writer) {
        writer->getParentObject()->removeEventListener(&setFrameCount_listener);
      }
      
      IEncodeStats_var encode_stats(video_enc->getParentObject()->getInterface<IEncodeStats>());
      if (encode_stats.get()) {
        encode_stats->sendFrameCount(false);
      }
    }
    
    if ( dvd ) {
      //if (writer) {
        dvd->getParentObject()->removeEventListener(&setProgress_listener);
        //}
      
      IProgress_var progress(dvd->getParentObject()->getInterface<IProgress>());
      if (progress.get()) {
        progress->sendProgressInformation(false);
      }
    }
  }

  Result DVDGrabState::Back() {
    switch (state) {
    case OPTIONS:
      startChapterMenu();
      break;
      
    case CHAPTER_SELECT:
      startTitleMenu();
      break;
    
    case TITLE_SELECT:
      kill();
      goToLastState();
      break;
    
    default:
      GLOBAL_DEBUG_STREAM("Switch to stateid:" << lastid << endl);
      goToLastState();
    }
  
    return SUCCESS;
  }

  Result DVDGrabState::Stop() {
    kill();
    goToLastState();
  
    return SUCCESS;
  }

  Result DVDGrabState::Up() {
    switch (state) {
    case CHAPTER_SELECT:
      if (chapter > 1) {
        --chapter;
        text->selectLine( chapter );
        win->update();
        manager->update();
      }
      break;
    
    case TITLE_SELECT:
      if (title > 1) {
        --title;
        text->selectLine( title );
        win->update();
        manager->update();
      }
      break;

    case OPTIONS:
      dlunit->selectUp();
      dlunit->update();
      manager->update();
      break;
    
    default:
      break;
    }
  
    return SUCCESS;
  }

  Result DVDGrabState::Down() {
    switch (state) {
    case CHAPTER_SELECT:
      if ( chapter < num_chapters[title] ) {
        ++chapter;
        text->selectLine( chapter );
        win->update();
        manager->update();
      }
      break;
    
    case TITLE_SELECT:
      if ( title < num_titles ) {
        ++title;
        text->selectLine( title );
        win->update();
        manager->update();
      }
      break;
      
    case OPTIONS:
      dlunit->selectDown();
      dlunit->update();
      manager->update();
      break;
      
    default:
      break;
    }
  
    return SUCCESS;
  }
  
  Result DVDGrabState::Left() {
    switch(state) {
    case OPTIONS:
      dlunit->selectLeft();
      dlunit->update();
      manager->update();
      
    default:
      break;
    }
    
    return SUCCESS;
  }

  Result DVDGrabState::Right() {
    switch(state) {
    case OPTIONS:
      dlunit->selectRight();
      dlunit->update();
      manager->update();
      
    default:
      break;
    }
    
    return SUCCESS;
  }
  
  Result DVDGrabState::Return() {
    switch (state) {
    case TITLE_SELECT:
      if (num_chapters[title] > 1) {
        startChapterMenu();
      } else {
        chapter = 1;
        startOptionsMenu();
      }
      break;
    
    case CHAPTER_SELECT:
      startOptionsMenu();
      break;
      
    case OPTIONS:
      {
        int n = dlunit->getNumLines();
        for (int i=1; i<=n; ++i) {
          string val   = dlunit->getLineUrl( i );
          string value = dlunit->getLineContent( i );
          config->setValue( val, value );
        }
        
        if (FAILURE == startGrabbing()) {
          manager->displayMessage( "Error", "Failed to initialize the grabber" , "", "CANCEL" );
          delete dvd;
          dvd = NULL;
          
          changeState("OSDInputState");
          manager->update();
        }
      }
      break;
      
    default:
      break;
    }
  
    return SUCCESS;
  }

  Result DVDGrabState::startTrack(string& name) {
    GLOBAL_ERROR_STREAM("Got Event >>   start   << with Value \"" 
                        << name << "\"" << endl);
    
    current_progress = Rational(0);
    
    if ( hasFocus() ) {
      manager->displayProgressBar( "Grabbing DVD" );
      manager->setProgress(current_progress.getNumerator());
      manager->setMaxProgress(current_progress.getDenominator());
      manager->updateProgress();
      
      win->update();
      manager->update();
    }
  
    return SUCCESS;
  }


  Result DVDGrabState::endTrack() {
    header->setText("Grabbing finished.");
    
    current_progress = Rational(1);
    
    if ( hasFocus() ) {
      manager->setProgress(current_progress.getNumerator());
      manager->setMaxProgress(current_progress.getDenominator());
      manager->updateProgress();
      
      win->update();
      manager->update();
    }
    return( SUCCESS );
    
    /* GLOBAL_ERROR_STREAM("Got Event >>   alldone   << with Value \""
       << f << "\"" << endl);

       if (!f) {
       return FAILURE;
       }

       GLOBAL_ERROR_STREAM("returning..." << endl);
       return( goToLastState() ); */
  }

  void DVDGrabState::createWindow() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": create window" << endl );
  
    deleteWindow();
  
    IWidgetFactory& factory = app->getOSDWidgetFactory();
    
    /* Begin> main window */
    text = factory.createTextView();
    text->setFont("medium");
    text->setSelectionColor(0xff0080ff);
  
    scroll = factory.createScrollDecorator(*text);
  
    plane  = factory.createPlaneDecorator(*scroll);
    plane->setPlaneColor(0x80ff0000);
  
    border = factory.createBorderDecorator(*plane);
  
    header = factory.createHeaderDecorator(*border);
    header->setFont("medium");
    header->setText( "Accessing DVD..." );
  
    win = manager->createWindow();
    win->setDimension( 620, 350 );
    win->setPosition( 50, 50 );
    win->setContent( *header );
    
    win->update();
    /* End< main window */
    
    manager->update();
  }

  void DVDGrabState::deleteWindow() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": delete window" << endl );
    
    if ( win ) {
      manager->deleteWindow( *win );
      delete win;
      win = NULL;
    }
  
    if ( text ) {
      manager->deleteWidget(*text);
      delete text;
      text = NULL;
    }
  
    if ( plane ) {
      manager->deleteWidget(*plane);
      delete plane;
      plane = NULL;
    }
  
    if ( border ) {
      manager->deleteWidget(*border);
      delete border;
      border = NULL;
    }

    if ( scroll ) {
      manager->deleteWidget(*scroll);
      delete scroll;
      scroll = NULL;
    }

    if ( header ) {
      manager->deleteWidget(*header);
      delete header;
      header = NULL;
    }
  }

  Result DVDGrabState::initDVD() {
    try {
    
      /* Begin> Request DVDReadNode */
      NodeDescription dvd_req;
      dvd_req.setNodeName("DVDReadNode");
      ClientRegistry& registry = app->getNMMApplication()->getRegistry();
      {
        RegistryLock lock(registry);

        list<Response> dvd_res = registry.initRequest(dvd_req);

        if (dvd_res.empty() ) {
          GLOBAL_ERROR_STREAM("Request for DVDReadNode failed" << endl);
          throw Exception("Request for DVDReadNode failed");
        }

        dvd = registry.requestNode(dvd_res.front());
      }
      /* End< Request DVDReadNode */
    
      /* Begin> Init DVD */
      dvd_interface = dvd->getParentObject()->getCheckedInterface<IDVD>();
      dvd_interface->setDevice(device);
    
      dvd->init();
      dvd->initOutput();
      /* End< Init DVD */
    
    } catch(const Exception &e) {
      GLOBAL_ERROR_STREAM("Couldn\'t init DVD Read Node, returning to upper menu" << endl);
      GLOBAL_ERROR_STREAM(e << endl);
    
      manager->displayMessage( "Input Error", "Could not find a valid DVD" , "", "CANCEL" );
    
      deinitDVD();
    
      changeState("OSDInputState");
      manager->update();
    
      return FAILURE;
    }
  
    // get number of titles and number of chapters for each title
    num_titles = dvd_interface->getTitles();
    for(int i=1; i<=num_titles; i++) {
      num_chapters[i] = dvd_interface->getChapters(i);
    }
  
    GLOBAL_DEBUG_STREAM( "DVDGrabState " << stateid << ": analyze stream format" << endl );

    // analyze dvd stream format
    Property_var dvd_wout(dvd->getWorkingOutputProperty());
    const Format* dvd_file_format = dvd_wout->getDefaultFormat();
    if ( dvd_file_format ) {
      if ( !(dvd_file_format->getFormatType() == "av/mpegps") ) {
        manager->displayMessage( "Input Error", "No PS/PES stream was found, seems to be an elementary stream!", "", "CANCEL");
        changeState("OSDInputState");
        GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": END< create (SUCCESS)" << endl );
        return SUCCESS;
      }
    } else {
      manager->displayMessage( "Input Error", "Can't identify DVD data stream!", "", "CANCEL");
      changeState("OSDInputState");
      return FAILURE;
    }

    return SUCCESS;
  }

  void DVDGrabState::deinitDVD() {
    ClientRegistry& registry = app->getNMMApplication()->getRegistry();
    if (dvd_interface) {
      delete dvd_interface;
      dvd_interface = NULL;
    }
    if (dvd) {
      if (dvd->isOutputInitialized()) {
        dvd->deinitOutput();
      }
      if (dvd->isInitialized()) {
        dvd->deinit();
      }
      registry.releaseNode(*dvd);
      delete dvd;
      dvd = NULL;
    }
  }

  void DVDGrabState::createNodes() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": createNodes" << endl );

    /* Begin> Create requests for needed Nodes */
    NodeDescription demux_req;
    demux_req.setNodeName("MPEGDemuxNode");
    NodeDescription video_dec_req;
    video_dec_req.setNodeName("MPEGVideoDecodeNode");
    NodeDescription spu_dec_req;
    spu_dec_req.setNodeName("SPUDecodeNode");
    NodeDescription overlay_req;
    overlay_req.setNodeName("OverlayNode");
    NodeDescription deinterlacer_req;
    deinterlacer_req.setNodeName("YUVDeInterlaceNode");
    NodeDescription scaler_req;
    scaler_req.setNodeName("VideoScalerNode");
    NodeDescription video_enc_req;
    video_enc_req.setNodeName(use_ffmpeg ? "FFMpegEncodeNode" : "DivX4EncodeNode");
    NodeDescription ac3_dec_req;
    ac3_dec_req.setNodeName("AC3DecodeNode");
    NodeDescription audio_enc_req;
    audio_enc_req.setNodeName("MPEGAudioEncodeNode");
    NodeDescription av_mux_req;
    av_mux_req.setNodeName("AVMuxNode");
    NodeDescription writer_req;
    writer_req.setNodeName("AVIWriteNode");
    /* End< */

    /* Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getNMMApplication()->getRegistry();
    {
      RegistryLock lock(registry);

      list<string>::const_iterator iter = registry_list.begin();
      while(true) {

	string host_and_port = "localhost";
	if(iter != registry_list.end()) { // remote
	  host_and_port = *iter;
	}

        demux_req.setLocation(host_and_port);
        video_dec_req.setLocation(host_and_port);
        spu_dec_req.setLocation(host_and_port);
        overlay_req.setLocation(host_and_port);
        deinterlacer_req.setLocation(host_and_port);
        scaler_req.setLocation(host_and_port);
        video_enc_req.setLocation(host_and_port);
        ac3_dec_req.setLocation(host_and_port);
        audio_enc_req.setLocation(host_and_port);
        av_mux_req.setLocation(host_and_port);
        writer_req.setLocation(host_and_port);

	GLOBAL_DEBUG_STREAM("Trying to get nodes from " << host_and_port << endl);

	list<Response> demux_res = registry.initRequest(demux_req);
	list<Response> video_dec_res = registry.initRequest(video_dec_req);
	list<Response> spu_dec_res = registry.initRequest(spu_dec_req);
	list<Response> overlay_res = registry.initRequest(overlay_req);
	list<Response> deinterlacer_res = registry.initRequest(deinterlacer_req);
	list<Response> scaler_res = registry.initRequest(scaler_req);
	list<Response> video_enc_res = registry.initRequest(video_enc_req);
	list<Response> ac3_dec_res = registry.initRequest(ac3_dec_req);
	list<Response> audio_enc_res = registry.initRequest(audio_enc_req);
	list<Response> av_mux_res = registry.initRequest(av_mux_req);

	if (demux_res.empty() ||
	    video_dec_res.empty() ||
	    spu_dec_res.empty() ||
	    overlay_res.empty() ||
	    deinterlacer_res.empty() ||
	    scaler_res.empty() ||
	    video_enc_res.empty() ||
	    ac3_dec_res.empty() ||
	    audio_enc_res.empty() ||
	    av_mux_res.empty() ) {
	  GLOBAL_ERROR_STREAM("Size of demux response " << demux_res.size() << endl);
	  GLOBAL_ERROR_STREAM("Size of video_dec response " << video_dec_res.size() << endl);
	  GLOBAL_ERROR_STREAM("Size of spu_dec response " << spu_dec_res.size() << endl);
	  GLOBAL_ERROR_STREAM("Size of overlay response " << overlay_res.size() << endl);
	  GLOBAL_ERROR_STREAM("Size of deinterlacer response " << deinterlacer_res.size() << endl);
	  GLOBAL_ERROR_STREAM("Size of scaler response " << scaler_res.size() << endl);
	  GLOBAL_ERROR_STREAM("Size of video_enc response " << video_enc_res.size() << endl);
	  GLOBAL_ERROR_STREAM("Size of ac3_dec response " << ac3_dec_res.size() << endl);
	  GLOBAL_ERROR_STREAM("Size of audio_enc response " << audio_enc_res.size() << endl);
	  GLOBAL_ERROR_STREAM("Size of av_mux response " << av_mux_res.size() << endl);

	  if(iter != registry_list.end()) { // try with next remote registry
	    ++iter;
	    continue;
	  }
	  else { // even local nodes could not be obtained
	    throw Exception("A required node is not available");
	  }
	}

	demux = registry.requestNode(demux_res.front());
	video_dec = registry.requestNode(video_dec_res.front());
	spu_dec = registry.requestNode(spu_dec_res.front());
	overlay = registry.requestNode(overlay_res.front());
	deinterlacer = registry.requestNode(deinterlacer_res.front());
	scaler = registry.requestNode(scaler_res.front());
	video_enc = registry.requestNode(video_enc_res.front());
	ac3_dec = registry.requestNode(ac3_dec_res.front());
	audio_enc = registry.requestNode(audio_enc_res.front());
	av_mux = registry.requestNode(av_mux_res.front());

	// if we come here, we have found the needed nodes
	break;
      } // end while(true)

      list<Response> writer_res = registry.initRequest(writer_req);
      if(writer_res.empty()) {
        GLOBAL_ERROR_STREAM("Size of writer response " << writer_res.size() << endl);
	throw Exception("A required node is not available");
      }
      writer = registry.requestNode(writer_res.front());
    }
    /* End< Get the ClientRegistry and request Nodes */

    /* Begin> Set queue sizes */
    IBiDirQueuedNode_var video_dec_q(video_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    IBiDirQueuedNode_var spu_dec_q(spu_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    IBiDirQueuedNode_var ac3_dec_q(ac3_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    IBiDirQueuedNode_var audio_enc_q(audio_enc->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    IBiDirQueuedNode_var video_enc_q(video_enc->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    IBiDirQueuedNode_var av_mux_q(av_mux->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    IBiDirQueuedNode_var writer_q(writer->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    video_dec_q->setDownstreamMaxSize(200);
    spu_dec_q->setDownstreamMaxSize(100);
    ac3_dec_q->setDownstreamMaxSize(200);
    audio_enc_q->setDownstreamMaxSize(100);
    video_enc_q->setDownstreamMaxSize(100);
    av_mux_q->setDownstreamMaxSize(100);
    writer_q->setDownstreamMaxSize(100);
    /* End< Set queue sizes */
  }

  void DVDGrabState::deleteNodes() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": deleteNodes" << endl );

    ClientRegistry& registry = app->getNMMApplication()->getRegistry();
    if (writer) {
      registry.releaseNode(*writer);
      delete writer;
      writer = NULL;
    }

    if (av_mux) {
      registry.releaseNode(*av_mux);
      delete av_mux;
      av_mux = NULL;
    }

    if (audio_enc) {
      registry.releaseNode(*audio_enc);
      delete audio_enc;
      audio_enc = NULL;
    }

    if (ac3_dec) {
      registry.releaseNode(*ac3_dec);
      delete ac3_dec;
      ac3_dec = NULL;
    }

    if (video_enc) {
      registry.releaseNode(*video_enc);
      delete video_enc;
      video_enc = NULL;
    }
  
    if (scaler) {
      registry.releaseNode(*scaler);
      delete scaler;
      scaler = NULL;
    }
  
    if (deinterlacer) {
      registry.releaseNode(*deinterlacer);
      delete deinterlacer;
      deinterlacer = NULL;
    }

    if (overlay) {
      registry.releaseNode(*overlay);
      delete overlay;
      overlay = NULL;
    }

    if (spu_dec) {
      registry.releaseNode(*spu_dec);
      delete spu_dec;
      spu_dec = NULL;
    }
  
    if (video_dec) {
      registry.releaseNode(*video_dec);
      delete video_dec;
      video_dec = NULL;
    }
  
    if (demux) {
      registry.releaseNode(*demux);
      delete demux;
      demux = NULL;
    }
  }

  void DVDGrabState::initNodes() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": initNodes" << endl );
  
    if (demux) {
      demux->init();
    }
  
    if (video_dec) {
      video_dec->init();
    }
  
    if (spu_dec) {
      spu_dec->init();
    }

    if (overlay) {
      overlay->init();
    }
  
    if (deinterlacer) {
      deinterlacer->init();
    }
  
    if (scaler) {
      scaler->init();
    }
  
    if (video_enc) {
      video_enc->init();
    }
  
    if (ac3_dec) {
      ac3_dec->init();
    }
  
    if (audio_enc) {
      audio_enc->init();
    }
    
    if (av_mux) {
      /* Turn off sync mode to avoid timestamp problems around angle blocks (temporary fix!) */
      IAVMux_var av_mux_interface(av_mux->getParentObject()->getCheckedInterface<IAVMux>());
      av_mux_interface->setSyncMode(false);
      av_mux->init();
    }
    
    if (writer) {
      ISinkFileHandler_var file_sink(writer->getParentObject()->getCheckedInterface<ISinkFileHandler>());
      file_sink->setFilename(filename);
      writer->init();
    }
  }

  void DVDGrabState::deinitNodes() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": deinitNodes" << endl );
  
    if (dvd && dvd->isInitialized()) {
      dvd->deinit();
    }
  
    if (demux && demux->isInitialized()) {
      demux->deinit();
    }
  
    if (video_dec && video_dec->isInitialized()) {
      video_dec->deinit();
    }
  
    if (spu_dec && spu_dec->isInitialized()) {
      spu_dec->deinit();
    }
  
    if (overlay && overlay->isInitialized()) {
      overlay->deinit();
    }
  
    if (deinterlacer && deinterlacer->isInitialized()) {
      deinterlacer->deinit();
    }
  
    if (scaler && scaler->isInitialized()) {
      scaler->deinit();
    }
  
    if (ac3_dec && ac3_dec->isInitialized()) {
      ac3_dec->deinit();
    }
  
    if (audio_enc && audio_enc->isInitialized()) {
      audio_enc->deinit();
    }
  
    if (video_enc && video_enc->isInitialized()) {
      video_enc->deinit();
    }
  
    if (av_mux && av_mux->isInitialized()) {
      av_mux->deinit();
    }
  
    if (writer && writer->isInitialized()) {
      writer->deinit();
    }
  }

  void DVDGrabState::startNodes() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": startNodes" << endl );
  
    if (dvd && dvd->isActivated()) {
      dvd->start();
    }
  
    if (demux && demux->isActivated()) {
      demux->start();
    }
  
    if (ac3_dec && ac3_dec->isActivated()) {
      ac3_dec->start();
    }
  
    if (audio_enc && audio_enc->isActivated()) {
      audio_enc->start();
    }
  
    if (video_dec && video_dec->isActivated()) {
      video_dec->start();
    }
  
    if (spu_dec && spu_dec->isActivated()) {
      spu_dec->start();
    }
  
    if (overlay && overlay->isActivated()) {
      overlay->start();
    }
  
    if (deinterlacer && deinterlacer->isActivated()) {
      deinterlacer->start();
    }
  
    if (scaler && scaler->isActivated()) {
      scaler->start();
    }
  
    if (video_enc && video_enc->isActivated()) {
      video_enc->start();

    }
  
    if (av_mux && av_mux->isActivated()) {
      av_mux->start();
    }
  
    if (writer && writer->isActivated()) {
      writer->start();

    }
  }

  void DVDGrabState::stopNodes() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": stopNodes" << endl );
  
    if (dvd && dvd->isStarted()) {
      dvd->stop();
    }
  
    if (demux && demux->isStarted()) {
      demux->stop();
    }
  
    if (video_dec && video_dec->isStarted()) {
      video_dec->stop();
    }
  
    if (spu_dec && spu_dec->isStarted()) {
      spu_dec->stop();
    }
  
    if (overlay && overlay->isStarted()) {
      overlay->stop();
    }

    if (deinterlacer && deinterlacer->isStarted()) {
      deinterlacer->stop();
    }
  
    if (scaler && scaler->isStarted()) {
      scaler->stop();
    }
  
    if (ac3_dec && ac3_dec->isStarted()) {
      ac3_dec->stop();
    }
  
    if (audio_enc && audio_enc->isStarted()) {
      audio_enc->stop();
    }
  
    if (video_enc && video_enc->isStarted()) {
      video_enc->stop();
    }
  
    if (av_mux && av_mux->isStarted()) {
      av_mux->stop();
    }
  
    if (writer && writer->isStarted()) {
      writer->stop();
    }
  }

  void DVDGrabState::flushNodes() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": flushNodes" << endl );
  
    if (dvd && dvd->isActivated()) {
      dvd->flush();
    }
  
    if (demux && demux->isActivated()) {
      demux->flush();
    }
  
    if (video_dec && video_dec->isActivated()) {
      video_dec->flush();
    }
  
    if (spu_dec && spu_dec->isActivated()) {
      spu_dec->flush();
    }
  
    if (overlay && overlay->isActivated()) {
      overlay->flush();
    }
  
    if (deinterlacer && deinterlacer->isActivated()) {
      deinterlacer->flush();
    }

    if (scaler && scaler->isActivated()) {
      scaler->flush();
    }
  
    if (ac3_dec && ac3_dec->isActivated()) {
      ac3_dec->flush();
    }
  
    if (audio_enc && audio_enc->isActivated()) {
      audio_enc->flush();
    }
  
    if (video_enc && video_enc->isActivated()) {
      video_enc->flush();
    }
  
    if (av_mux && av_mux->isActivated()) {
      av_mux->flush();
    }
  
    if (writer && writer->isActivated()) {
      writer->flush();
    }
  }

  void DVDGrabState::connectNodes() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": connectNodes" << endl );

    connect(dvd, demux);
    dvd->activate();
    demux->initOutput();
  
    /* Check MPEG format */
    string spu_jack_string = string("spu") + 0 /*spu_stream*/;
    string audio_jack_string = string("ac3_audio") + audio_track;
    string video_jack_string = string("mpeg_video") + 0;
    {
      Property_var demux_wout_property(demux->getWorkingOutputProperty());
    
      const Format* mpeg_video_format = demux_wout_property->getDefaultFormat(video_jack_string);
      if ( mpeg_video_format != NULL && mpeg_video_format->getFormatType() == "video/mpeg" ) {
        cout << "Found MPEG video stream version " << mpeg_video_format->getIntValue("version") << endl;
      } else {
        throw Exception("No video stream found");
      }
    
      const Format* ac3_audio_format = demux_wout_property->getDefaultFormat(audio_jack_string);
      if ( ac3_audio_format != NULL && ac3_audio_format->getFormatType() == "audio/ac3" ) {
        cout << "Found AC3 audio stream" << endl;
      } else {
        throw Exception("No audio stream found");
      }
    
      if (!no_subtitles) {
        const Format* spu_format = demux_wout_property->getDefaultFormat(spu_jack_string);
        if ( spu_format != NULL && spu_format->getFormatType() == "video/spu" ) {
          cout << "Found subtitle picture unit" << endl;
        } else {
          no_subtitles = true;
          GLOBAL_WARNING_STREAM("No SPU stream found");
        }
      }
    }
  
    /* Connect audio */
    if (!ac3_writethrough) {
      connect(demux, audio_jack_string, ac3_dec);
    } else {
      connect(demux, audio_jack_string, av_mux, string("audio"));
    }
  
    /* Connect SPU */
    if (!no_subtitles) {
      connect(demux, spu_jack_string, spu_dec);
    }
  
    /* Connect video */
    connect(demux, video_jack_string, video_dec);
  
    demux->activate();
  
    video_dec->initOutput();
  
    /* Add deinterlacer */
    INode* deinterlaced_video_source = video_dec;
    if (video_deinterlace) {
      connect(video_dec, deinterlacer);
      deinterlaced_video_source = deinterlacer;
    
      video_dec->activate();
      deinterlacer->initOutput();
    }
  
    /* Add overlay */
    INode* raw_video_source = deinterlaced_video_source;
    if (!no_subtitles) {
      spu_dec->initOutput();
      connect(spu_dec, overlay, string("foreground"));
      connect(deinterlaced_video_source, overlay, string("background"));
    
      raw_video_source = overlay;
    
      spu_dec->activate();
      deinterlaced_video_source->activate();
      overlay->initOutput();
    }
  
    /* Add scaler */
    if (video_scale) {
      connect(raw_video_source, scaler);
    
      raw_video_source->activate();
      scaler->initOutput();
    
      Property_var scaler_wout(scaler->getWorkingOutputProperty());
      const Format* scaler_def_format = scaler_wout->getDefaultFormat();
      Format* scaled_video_format = new Format(scaler_def_format->getFormatType());
      int x_resolution = scaler_def_format->getIntValue(Format::x_resolution_param);
      int y_resolution = scaler_def_format->getIntValue(Format::y_resolution_param);
//       if (width != 0) {
//         x_resolution = width;
//       }
//       if (height != 0) {
//         y_resolution = height;
//       }
      int ratio_x = scaler_def_format->getIntValue(Format::ratio_x_param);
      int ratio_y = scaler_def_format->getIntValue(Format::ratio_y_param);
      float framerate = scaler_def_format->getFloatValue(Format::framerate_param);
      scaled_video_format->addIntParamValue(Format::x_resolution_param, x_resolution);
      scaled_video_format->addIntParamValue(Format::y_resolution_param, y_resolution);
      scaled_video_format->addIntParamValue(Format::ratio_x_param, ratio_x);
      scaled_video_format->addIntParamValue(Format::ratio_y_param, ratio_y);
      scaled_video_format->addFloatParamValue(Format::framerate_param, framerate);
      scaled_video_format->addStringParamValue(Format::colorspace_param, "yv12");
      scaled_video_format->addStringParamValue(Format::channelorder_param, "yvu");
      scaled_video_format->addStringParamValue(Format::format_param, "planar");
      scaled_video_format->addStringParamValue(Format::endian_param, "little");
      scaled_video_format->addIntParamValue(Format::bitperpixel_param, 12);
    
      connect(scaler, video_enc, scaled_video_format);
      scaler->activate();
    } else {
      connect(raw_video_source, video_enc);
      raw_video_source->activate();
    }
  
    // create divx format (with bitrate)
    video_enc->initOutput();
    {
      Property_var video_enc_wout(video_enc->getWorkingOutputProperty());
      const Format* video_enc_def_format = video_enc_wout->getDefaultFormat();
      Format* divx_video_format = new Format(video_enc_def_format->getFormatType());
      int x_resolution = video_enc_def_format->getIntValue(Format::x_resolution_param);
      int y_resolution = video_enc_def_format->getIntValue(Format::y_resolution_param);
      int ratio_x = video_enc_def_format->getIntValue(Format::ratio_x_param);
      int ratio_y = video_enc_def_format->getIntValue(Format::ratio_y_param);
      float framerate = video_enc_def_format->getFloatValue(Format::framerate_param);
      string compressor = video_enc_def_format->getStringValue(Format::compressor_param);
      divx_video_format->addIntParamValue(Format::x_resolution_param, x_resolution);
      divx_video_format->addIntParamValue(Format::y_resolution_param, y_resolution);
      divx_video_format->addIntParamValue(Format::ratio_x_param, ratio_x);
      divx_video_format->addIntParamValue(Format::ratio_y_param, ratio_y);
      divx_video_format->addFloatParamValue(Format::framerate_param, framerate);
      divx_video_format->addIntParamValue(Format::bitrate_param, video_bitrate * 1000);
      divx_video_format->addStringParamValue(Format::compressor_param, compressor);
    
      /* Video to Mux */
      connect(video_enc, av_mux, string("video"), divx_video_format);
    }
    video_enc->activate();
  
    /* Add audio encoder */
    if (!ac3_writethrough) {
      ac3_dec->initOutput();
      connect(ac3_dec, audio_enc);
    
      ac3_dec->activate();
      audio_enc->initOutput();
    
      Property_var audio_enc_wout(audio_enc->getWorkingOutputProperty());
      const Format* audio_enc_def_format = audio_enc_wout->getDefaultFormat();
      Format* mpeg_audio_format = new Format(audio_enc_def_format->getFormatType());
      int sample_rate = audio_enc_def_format->getIntValue(Format::sample_rate_param);
      int channels = audio_enc_def_format->getIntValue(Format::channels_param);
      int audio_codec = audio_enc_def_format->getIntValue(Format::audio_codec_param);
      mpeg_audio_format->addIntParamValue(Format::layer_param, 3);
      mpeg_audio_format->addIntParamValue(Format::sample_rate_param, sample_rate);
      mpeg_audio_format->addIntParamValue(Format::channels_param, channels);
      mpeg_audio_format->addIntParamValue(Format::bitpersample_param, 16);
      mpeg_audio_format->addIntParamValue(Format::audio_codec_param, audio_codec);
      mpeg_audio_format->addIntParamValue(Format::bitrate_param, audio_bitrate * 1000);
    
      connect(audio_enc, av_mux, string("audio"), mpeg_audio_format);
      audio_enc->activate();
    }
  
    /* Connect Mux to writer */
    av_mux->initOutput();
    connect(av_mux, writer);
    av_mux->activate();
    writer->initOutput();
    writer->activate();  
  }

  void DVDGrabState::disconnectNodes() {
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": disconnectNodes" << endl );
  
    if (writer) {
      if (writer->isActivated()) {
        writer->deactivate();
      }
      if (writer->isOutputInitialized()) {
        writer->deinitOutput();
      }
      IInputJack_var writer_in_jack(writer->getInputStream("default"));
      if (writer_in_jack.get() && writer_in_jack->isConnected()) {
        if (av_mux && av_mux->isActivated()) {
          av_mux->deactivate();
        }
        disconnect_input(writer);
      }
    }
    if (av_mux) {
      if (av_mux->isActivated()) {
        av_mux->deactivate();
      }
      if (av_mux->isOutputInitialized()) {
        av_mux->deinitOutput();
      }
      IInputJack_var av_mux_audio_in_jack(av_mux->getInputStream("audio"));
      if (av_mux_audio_in_jack.get() && av_mux_audio_in_jack->isConnected()) {
        if (ac3_writethrough) {
          if (demux && demux->isActivated()) {
            demux->deactivate();
          }
        } else {
          if (audio_enc && audio_enc->isActivated()) {
            audio_enc->deactivate();
          }
        }
        disconnect_input(av_mux, "audio");
      }
      IInputJack_var av_mux_video_in_jack(av_mux->getInputStream("video"));
      if (av_mux_video_in_jack.get() && av_mux_video_in_jack->isConnected()) {
        if (video_enc && video_enc->isActivated()) {
          video_enc->deactivate();
        }
        disconnect_input(av_mux, "video");
      }
    }
    if (!ac3_writethrough) {
      if (audio_enc) {
        if (audio_enc->isActivated()) {
          audio_enc->deactivate();
        }
        if (audio_enc->isOutputInitialized()) {
          audio_enc->deinitOutput();
        }
        IInputJack_var audio_enc_in_jack(audio_enc->getInputStream("default"));
        if (audio_enc_in_jack.get() && audio_enc_in_jack->isConnected()) {
          if (ac3_dec && ac3_dec->isActivated()) {
            ac3_dec->deactivate();
          }
          disconnect_input(audio_enc);
        }
      }
      if (ac3_dec) {
        if (ac3_dec->isActivated()) {
          ac3_dec->deactivate();
        }
        if (ac3_dec->isOutputInitialized()) {
          ac3_dec->deinitOutput();
        }
        IInputJack_var ac3_dec_in_jack(ac3_dec->getInputStream("default"));
        if (ac3_dec_in_jack.get() && ac3_dec_in_jack->isConnected()) {
          if (demux && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(ac3_dec);
        }
      }
    }
    if (video_enc) {
      if (video_enc->isActivated()) {
        video_enc->deactivate();
      }
      if (video_enc->isOutputInitialized()) {
        video_enc->deinitOutput();
      }
      IInputJack_var video_enc_in_jack(video_enc->getInputStream("default"));
      if (video_enc_in_jack.get() && video_enc_in_jack->isConnected()) {
        if (video_scale) {
          if (scaler && scaler->isActivated()) {
            scaler->deactivate();
          }
        } else {
          if (!no_subtitles) {
            if (overlay && overlay->isActivated()) {
              overlay->deactivate();
            }
          } else {
            if (video_deinterlace) {
              if (deinterlacer && deinterlacer->isActivated()) {
                deinterlacer->deactivate();
              }
            } else {
              if (video_dec && video_dec->isActivated()) {
                video_dec->deactivate();
              }
            }
          }
        }
        disconnect_input(video_enc);
      }
    }
    if (video_scale && scaler) {
      if (scaler->isActivated()) {
        scaler->deactivate();
      }
      if (scaler->isOutputInitialized()) {
        scaler->deinitOutput();
      }
      IInputJack_var scaler_in_jack(scaler->getInputStream("default"));
      if (scaler_in_jack.get() && scaler_in_jack->isConnected()) {
        if (!no_subtitles) {
          if (overlay && overlay->isActivated()) {
            overlay->deactivate();
          }
        } else {
          if (video_deinterlace) {
            if (deinterlacer && deinterlacer->isActivated()) {
              deinterlacer->deactivate();
            }
          } else {
            if (video_dec && video_dec->isActivated()) {
              video_dec->deactivate();
            }
          }
        }
        disconnect_input(scaler);
      }
    }
    if (!no_subtitles) {
      if (overlay) {
        if (overlay->isActivated()) {
          overlay->deactivate();
        }
        if (overlay->isOutputInitialized()) {
          overlay->deinitOutput();
        }
        IInputJack_var overlay_background_in_jack(overlay->getInputStream("background"));
        if (overlay_background_in_jack.get() && overlay_background_in_jack->isConnected()) {
          if (video_deinterlace) {
            if (deinterlacer && deinterlacer->isActivated()) {
              deinterlacer->deactivate();
            }
          } else {
            if (video_dec && video_dec->isActivated()) {
              video_dec->deactivate();
            }
          }
          disconnect_input(overlay, "background");
        }
        IInputJack_var overlay_foreground_in_jack(overlay->getInputStream("foreground"));
        if (overlay_foreground_in_jack.get() && overlay_foreground_in_jack->isConnected()) {
          if (spu_dec && spu_dec->isActivated()) {
            spu_dec->deactivate();
          }
          disconnect_input(overlay, "foreground");
        }
      }
      if (spu_dec) {
        if (spu_dec->isActivated()) {
          spu_dec->deactivate();
        }
        if (spu_dec->isOutputInitialized()) {
          spu_dec->deinitOutput();
        }
        IInputJack_var spu_dec_in_jack(spu_dec->getInputStream("default"));
        if (spu_dec_in_jack.get() && spu_dec_in_jack->isConnected()) {
          if (demux && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(spu_dec);
        }
      }
    }
    if (video_deinterlace && deinterlacer) {
      if (deinterlacer->isActivated()) {
        deinterlacer->deactivate();
      }
      if (deinterlacer->isOutputInitialized()) {
        deinterlacer->deinitOutput();
      }
      IInputJack_var deinterlacer_in_jack(deinterlacer->getInputStream("default"));
      if (deinterlacer_in_jack.get() && deinterlacer_in_jack->isConnected()) {
        if (video_dec && video_dec->isActivated()) {
          video_dec->deactivate();
        }
        disconnect_input(deinterlacer);
      }
    }
    if (video_dec) {
      if (video_dec->isActivated()) {
        video_dec->deactivate();
      }
      if (video_dec->isOutputInitialized()) {
        video_dec->deinitOutput();
      }
      IInputJack_var video_dec_in_jack(video_dec->getInputStream("default"));
      if (video_dec_in_jack.get() && video_dec_in_jack->isConnected()) {
        if (demux && demux->isActivated()) {
          demux->deactivate();
        }
        disconnect_input(video_dec);
      }
    }
    if (demux) {
      if (demux->isActivated()) {
        demux->deactivate();
      }
      if (demux->isOutputInitialized()) {
        demux->deinitOutput();
      }
      IInputJack_var demux_in_jack(demux->getInputStream("default"));
      if (demux_in_jack.get() && demux_in_jack->isConnected()) {
        if (dvd && dvd->isActivated()) {
          dvd->deactivate();
        }
        disconnect_input(demux);
      }
    }
    if (dvd) {
      if (dvd->isActivated()) {
        dvd->deactivate();
      }
      if (dvd->isOutputInitialized()) {
        dvd->deinitOutput();
      }
    }
  }

  void DVDGrabState::setState(State state) {
    this->state = state;
  
    string state_name;
    switch (state) {
    case TITLE_SELECT:
      state_name = "TITLE_SELECT";
      break;
    
    case CHAPTER_SELECT:
      state_name = "CHAPTER_SELECT";
      break;
      
    case OPTIONS:
      state_name = "OPTIONS";
      break;
      
    case GRABBING:
      state_name = "GRABBING";
      break;
    
    default:
      GLOBAL_ERROR_STREAM("INVALID SUBSTATE ID" << endl);
    }
  
    GLOBAL_DEBUG_STREAM( "DVDGrabState: " << stateid << ": set state: " << state_name << endl );
  }

  void DVDGrabState::lowerNodesPriority() {
    
    if (dvd) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (demux) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (video_dec) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (spu_dec) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (overlay) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (deinterlacer) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (scaler) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (video_enc) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (ac3_dec) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (audio_enc) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (av_mux) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
    if (writer) {
      IThreadedNode_var threaded_node(dvd->getParentObject()->getCheckedInterface<IThreadedNode>());
      threaded_node->lowerPriority(15);
    }
  }
  
  Result DVDGrabState::showOption( string id, string shortform ) {
    shown_options[shortform] = id;
    return SUCCESS;
  }
  
  void DVDGrabState::initOptionsList() {
    // XXX: fw: code duplicated from apps/mmbox/apps/config/config.cpp
    
    unsigned int size = dlunit->getNumLines();

    for( unsigned int i=1; i<=size; i++ ) {
      string val = dlunit->getLineUrl( i );
      string value;
      config->getValue( val, value );
      dlunit->chooseLineContent( i, value );
    }
    
    dlunit->update();
    manager->update();
  }
  
  void DVDGrabState::readConfig() {
    if (!(config->getValue("rootpath", root_path))) {
      root_path="/";
    }

    if (!(config->getValue("grabdvddestinationdirectory",dest_dir))) {
      dest_dir="DVD-Recordings";
    }

    if (!(config->getValue("grabdvdac3writethrough",ac3_writethrough))) {
      ac3_writethrough = true;
    }

    if (!(config->getValue("grabdvdautoaspect",auto_aspect))) {
      auto_aspect = true;
    }

    if (!(config->getValue("grabdvddeinterlace",video_deinterlace))) {
      video_deinterlace = false;
    }

    if (!(config->getValue("grabdvduseffmpeg",use_ffmpeg))) {
      use_ffmpeg = false;
    }

    if (!(config->getValue("grabdvdaudioquality",audio_quality))) {
      audio_quality = 2;
    }

    if (!(config->getValue("grabdvdaudiobitrate",audio_bitrate))) {
      audio_bitrate = 128;
    }

    if (!(config->getValue("grabdvdvideobitrate",video_bitrate))) {
      video_bitrate = 1200;
    }

    int  transcodinghosts_number = 0;
    if(config->getValue("transcodinghosts_number", transcodinghosts_number)) {
      if(transcodinghosts_number > 0) {
	char tmp[128];
	for(int i = 0; i < transcodinghosts_number; ++i) {
	  sprintf(tmp, "transcodinghost_%u", i);
	  string current_host_string(tmp);
	  string host_and_port = "";
	  string host = "";
	  int port = 0;
	  if(config->getValue(current_host_string, host_and_port)) {
	    string::size_type colon_pos = host_and_port.find_first_of(':');
	    if (colon_pos != string::npos) {
	      host = host_and_port.substr(0, colon_pos);
	      port = atoi(host_and_port.substr(colon_pos+1, string::npos).c_str());
	    }
	    else {
	      host = host_and_port;
	      port = ClientRegistry::default_port;
	    }
	    host_and_port = host + ":" + port;

	    try {
	      GLOBAL_MESSAGE_STREAM("Adding transcoding host " << host << ":" << port << "... " << endl);
	      app->getNMMApplication()->getRegistry().addRegistry(TCPAddress(host, port));
	      registry_list.push_back(host);
	    }
	    catch(const Exception& e) {
	      GLOBAL_ERROR_STREAM("Failed to add transcoding host" << endl);
	    }
	  }
	}
      }
    }
  }
}
