/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iostream>
#include"mp3.hpp"

#include "nmm/base/connect.hpp"
#include "nmm/utils/utils.hpp"

// include used interfaces
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/IBufferSize.hpp"
#include "nmm/interfaces/general/ICopy.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"
#include "nmm/interfaces/general/ITrackDuration.hpp"
#include "nmm/interfaces/base/registry/IPluginManager.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"


using namespace NMM;

MP3State::MP3State( string id, MMBoxApplication* _app )
  : MMBoxState( id, _app ),
    state(STOPPED),
    osd_node(NULL),
    audio_play(NULL),
    readfile(NULL),
    mp3dec(NULL),
    copy(NULL),
    text(NULL),
    plane(NULL),
    border(NULL),
    scroll(NULL),
    header(NULL),
    win(NULL),
    progress(NULL),
    time_progress(NULL),
    file_source(NULL),
    iseek(NULL),
    manager(NULL),
    factory(NULL),
    actual_playing_line(0),
    mp3suffix("mp3"),
    MP3suffix("MP3"),
    first_file(true),
    loaded_m3u(false),
    current_progress(0),
    endTrack_listener(this, &MP3State::endTrack),
    syncReset_listener(this, &MP3State::syncReset),
    setProgress_listener(this, &MP3State::setProgress),
    setTimeProgress_listener(this, &MP3State::setTimeProgress),
    trackDuration_listener(this, &MP3State::trackDuration)  
{
  MMBoxConfig* config = _app->getMMBoxConfig();
  if (!(config->getValue("rootpath", mp3rootpath))) {
    mp3rootpath="/";
  }
  
  dirreader.setNiceDir( true );
  
  /*Begin> Get OSD manager node */
  osd_node = app->getOsdNode();
  manager = app->getOsdManager();
  factory = &(app->getOSDWidgetFactory());
  /*End< */
}

MP3State::~MP3State() {
}

Result MP3State::doCreate( string ) {
  GLOBAL_DEBUG_STREAM( "MP3State: doCreate: " << stateid << /*" mp3: " << mp3dec->getState() <<*/  endl );
  
  if ( isBackgroundFree() ) {
    if (FAILURE == app->createDefaultGraph( background )) {
      GLOBAL_ERROR_STREAM("MP3State::doCreate(): Could not create the default graph!" << endl);
      return FAILURE;
    }

    if (FAILURE == app->startDefaultGraph()) {
      GLOBAL_ERROR_STREAM("MP3State::doCreate(): Could not start the default graph!" << endl);
      return FAILURE;
    }
  }

  first_file = true;

  if (initSound() != SUCCESS) {
    GLOBAL_ERROR_STREAM("MP3State::doCreate(): Could not init sound!" << endl);
    return FAILURE;
  }

  manager->displaySymbol( "STOP" );

  createWindow();
  readDir();
  win->update();
  playing_titel = "Playback stopped";
  manager->displayProgressBar(playing_titel, 80, 420, 450);
  manager->displayTime("", 640 - manager->getMaxWidth("medium")*4, 420);

  manager->resetTime();
  app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
  app->getOsdManager()->setProgress(current_progress.getNumerator());
  manager->updateProgress();
  manager->update();
  if (progress) {
    audio_play->getParentObject()->registerEventListener(IProgressListener::setProgress_event, &setProgress_listener);
    progress->sendProgressInformation(true);
  }
  if (time_progress) {
    audio_play->getParentObject()->registerEventListener(ITimeProgressListener::setTimeProgress_event, &setTimeProgress_listener);
    time_progress->setProgressUpdateEnabled(true);
  }

  state = STOPPED;
  
  actual_playing_line = 0;

  track_duration = 0;

  return( SUCCESS );
}

Result MP3State::doKill() {
  GLOBAL_DEBUG_STREAM( "MP3State: doKill" << " mp3: " << mp3dec->getState() << endl );
  stopmp3();
  
  if (progress) {
    audio_play->getParentObject()->removeEventListener(&setProgress_listener);
    progress->sendProgressInformation(false);
  }

  if (time_progress) {
    audio_play->getParentObject()->removeEventListener(&setTimeProgress_listener);
    time_progress->setProgressUpdateEnabled(false);
  }

  /* remove trackDuration listener */
  app->getDispatcher()->removeEventListener(&trackDuration_listener);
  audio_play->getParentObject()->removeEventListener(ITrackDuration::trackDuration_event, app->getDispatcher());

  manager->hideProgressBar();
  manager->hideTime();
  manager->hideSymbol();
  
  // VERY IMPORTANT: unregister MP3State as Listener, because
  // otherwise it would get Event for example from cd-state
  app->getDispatcher()->removeEventListener(&endTrack_listener);
  app->getDispatcher()->removeEventListener(&syncReset_listener);
  audio_play->getParentObject()->removeEventListener(ITrack::endTrack_event, app->getDispatcher());
  audio_play->getParentObject()->removeEventListener(ISyncReset::syncReset_event, app->getDispatcher());
  
  GLOBAL_DEBUG_STREAM("Deiniting Sound" << endl);
  deInitSound();
  GLOBAL_DEBUG_STREAM("Deinited Sound" << endl);
  deleteWindow();
  
  if (isBackgroundFree()) {
    GLOBAL_DEBUG_STREAM("Disconnecting Default graph" << endl);
    app->disconnectDefaultGraph();
    GLOBAL_DEBUG_STREAM("Disconnected Default graph" << endl);
  }

  loaded_m3u = false;
  return(SUCCESS);
}

Result MP3State::doInitialize( string ) {
    // init the background
    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph( background )) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }    
    if (copy->getState()>INode::INITIALIZED) {
      /* stop and flush nodes */
      copy->reachOutputInitialized();
      osd_node->reachInitialized();
      connect(copy, string("1"), osd_node, string("stream2"));
      
      /* reactivate copy node */
      copy->activate();
      if (state == PLAYING) {
        // only start copy node (again) if we are in state PLAYING!
        copy->start();
      }
      osd_node->reachStarted();
    }
    
    win->setVisible(true); 
    win->update();
    manager->displayTime("", 640 - manager->getMaxWidth("medium")*4, 420);
    manager->displayProgressBar(playing_titel, 80, 420, 450);
    manager->setMaxProgress(current_progress.getDenominator());
    manager->setProgress(current_progress.getNumerator());
    manager->updateProgress();
    // we choose the right symbol upon our stored state
    switch (state) {
    case PLAYING:
      manager->displaySymbol("PLAY");
      break;
    case PAUSED:
      manager->displaySymbol("PAUSE");
      break;
    case STOPPED:
      manager->displaySymbol("STOP");
      break;
    }
    manager->update();
    
    GLOBAL_DEBUG_STREAM("CdState::doInitialize() == SUCCESS" << endl);
    
    return SUCCESS;
}

Result MP3State::doUninitialize() {
  manager->hideProgressBar();
  manager->hideTime();
  manager->hideSymbol();

  win->setVisible(false);

  if(copy->getState()>INode::INITIALIZED) {
    copy->reachOutputInitialized();
    osd_node->reachInitialized();
    
    IInputJack_var osdin_jack(osd_node->getInputStream("stream2"));
    if (osdin_jack.get() && osdin_jack->isConnected()) {
      disconnect_input(osd_node,string("stream2"));
    }
    
    
    /* reactivate copy node */
    copy->activate();
    if (state == PLAYING) {
      // only start copy node (again) if we are in state PLAYING!
      copy->start();
    }
    osd_node->reachStarted();
  }
    if (isBackgroundFree()) { 
      // we go in the background so:
      app->disconnectVideoGraph();
    }

  return( SUCCESS );
}


Result MP3State::setProgress( u_int64_t& val, u_int64_t& max ) {
  current_progress = Rational(val, max);

  if (hasFocus()) {
    app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
    app->getOsdManager()->setProgress(current_progress.getNumerator());
    manager->updateProgress();
    manager->update();
  }
  
  return SUCCESS;
}

Result MP3State::setTimeProgress(Time& current_media_time) {
  manager->setTime(current_media_time.sec);
  manager->update();
  return SUCCESS;
}

Result MP3State::trackDuration(Interval& duration) {
  track_duration = duration.sec;

  return DELETE;
}

void MP3State::createWindow() {
  deleteWindow();
  
  /*Begin> create widgets */
  text = factory->createTextView();
  text->setFont( "medium" );
  text->setSelectionColor( 0xff0080ff );
  
  scroll = factory->createScrollDecorator( *text );
  
  plane = factory->createPlaneDecorator( *scroll );
  plane->setPlaneColor( 0x80ff0000 );
  
  border = factory->createBorderDecorator( *plane );
  
  header = factory->createHeaderDecorator( *border );
  header->setFont( "medium" );
  header->setText( "MP3Player" );
  /*End< */
  
  /*Begin> create new window */
  win = manager->createWindow();
  win->setDimension(560, 350);
  
  win->setPosition(80, 50);
  win->setContent( *header );
  
  win->update();
  manager->update();
  /*End> */
}

void MP3State::deleteWindow() {
  if ( win ) {
    manager->deleteWindow( *win );
    delete win;
    win = NULL;
  }
  
  if ( text ) {
    manager->deleteWidget(*text);
    delete text;
    text = NULL;
  }

  if ( plane ) {
    manager->deleteWidget(*plane);
    delete plane;
    plane = NULL;
  }

  if ( border ) {
    manager->deleteWidget(*border);
    delete border;
    border = NULL;
  }

  if ( scroll ) {
    manager->deleteWidget(*scroll);
    delete scroll;
    scroll = NULL;
  }

  if ( header ) {
    manager->deleteWidget(*header);
    delete header;
    header = NULL;
  }
}


void MP3State::dirToText() {
  dir = dirreader.getDirEntries();
  
  text->deleteAllLines();
  for ( unsigned int i=0; i < dir.size(); i++ ) {
    if ( dir[i].isdir ) text->addLine( "DIR: " + dirreader.getNiceFileName( dir[i].name, 40 ) );
    else {
      string suffix(dirreader2.getSuffix(dir[i].name));
      if ((suffix == string("mp3")) || (suffix == string("MP3")) || (suffix == string("Mp3"))) {
          //(suffix == string("m3u")) || (suffix == string("M3U"))) {
        text->addLine(dirreader.getNiceFileName(dir[i].name, 40), dir[i].path+dir[i].name);
      }
      else if ((suffix == string("m3u")) || (suffix == string("M3U"))) {
        text->addLine(dirreader.getFileName(dir[i].name, 40), dir[i].path+dir[i].name);
      }
    }
  }
}

void MP3State::readDir() {
  dirreader.setRootDir( mp3rootpath );
  dirreader.changeDir( "." );
  
  header->setText( dirreader.getNiceDirName( dirreader.getCurDirWithoutRoot(), 40 ) );
  
  dirToText();
  
  text->selectLine( 1 );
}

/* Switch to last state */
Result MP3State::Back() {
  GLOBAL_DEBUG_STREAM( "MP3State: Back" << endl );
  GLOBAL_DEBUG_STREAM( "Switch to stateid:" << lastid << endl );
  
  return( goToLastState() );
}

Result MP3State::Up() {
  GLOBAL_DEBUG_STREAM( "MP3State: Up" << endl );
  
  text->selectUp();
  
  win->update();
  manager->update();
  
  return SUCCESS;
}

Result MP3State::Down() {
  GLOBAL_DEBUG_STREAM( "MP3State: Down" << endl );

  text->selectDown();

  win->update();
  manager->update();
  
  return SUCCESS;
}

Result MP3State::Left() {
  Rewind();

  return SUCCESS;
}

Result MP3State::Right() {
  Forward();

  return SUCCESS;
}

Result MP3State::Return() {
  GLOBAL_DEBUG_STREAM( "MP3State: Return" << endl );
  
  unsigned int nr = text->getSelectedLine();
  if ( nr ) {
    if ( dir[nr-1].isdir ) {

      if (!loaded_m3u) {
        dirreader.changeDir( dir[nr-1].name );
        string dirname=dir[nr-1].name;
        header->setText( dirreader.getNiceDirName( dirreader.getCurDirWithoutRoot(), 40 ) );
        
        dirreader2 = dirreader;
        dir2 = dirreader2.getRegFiles();
        
        actual_playing_line = nr;//-d_size;
        
        dirToText();
        text->reset();
        if (dirname=="Go Up") {
          text->selectLine(dir_position.top());
          dir_position.pop();
        } else {
          text->selectLine( 1 );
          dir_position.push(nr);
        }
      } else {
        dirreader.changeDir(dirreader.getCurDirWithoutRoot());
        loaded_m3u = false;
        header->setText( dirreader.getNiceDirName( dirreader.getCurDirWithoutRoot(), 40 ) );
        
        dirreader2 = dirreader;
        dir2 = dirreader2.getRegFiles();
        
        actual_playing_line = nr;//-d_size;
        dirToText();
        text->reset();
        text->selectLine( 1 );
      }
    } else {
      
      dirreader2 = dirreader;
      dir2 = dirreader2.getRegFiles();
      
      actual_playing_line = nr;//-d_size;

      if ( dirreader.getSuffix( dir[nr-1].name ) == "m3u" ) {
        loadm3u(dirreader.getCurDir()+dir[nr-1].name);
        loaded_m3u=true;
        vector<DirReader::directory_t> temp;
        temp.clear();
        dirreader.importDirectory( temp, dir2 );
        
        dirToText();
        text->reset();
        text->selectLine(1);

      } else {
        playLine( actual_playing_line );
      }
    }
  }
  
  GLOBAL_DEBUG_STREAM("MP3State::Return OSD stuff" << endl);
  win->update();
  manager->update();
  GLOBAL_DEBUG_STREAM("MP3State::Return OSD stuff finished" << endl);
  
  return SUCCESS;
}

void MP3State::loadm3u( string filename ) {
  ifstream playlist;
  
  playlist.open( filename.c_str() );
  if ( playlist.fail() ) return;
  
  string path = dir2[0].path;
  dir2.clear();
  unsigned int i=0;
  string temp;
  while( getline(playlist, temp) ) {
    dir2.resize( i+1 );
    dir2[i].isdir = false;
    dir2[i].path = path;
    dir2[i].name = temp;
    i++;
  }
  
  playlist.close();
}

Result MP3State::playLine( unsigned int line ) {
  GLOBAL_DEBUG_STREAM("MP3State::playLine(" << line << ")" << endl);

  stopmp3();
  if (line <= text->getNumLines()) {
    playmp3(text->getLineURL(line));
    GLOBAL_DEBUG_STREAM("MP3State::playLine: OSD stuff" << endl);
    playing_titel = dirreader2.getNiceFileName(text->getLineContent(line), 46);
    if (hasFocus()) {
      manager->displayProgressBar(playing_titel, 80, 420, 450);
      manager->displaySymbol("PLAY");
      manager->setMaxProgress(current_progress.getDenominator());
      manager->setProgress(current_progress.getNumerator());
      manager->updateProgress();
      manager->update();
    }
    GLOBAL_DEBUG_STREAM("MP3State::playLine: OSD stuff finished" << endl);
  }
  else {
    playing_titel = "Playback stopped";
    manager->resetTime();
    app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
    app->getOsdManager()->setProgress(current_progress.getNumerator());
    manager->updateProgress();
    if (hasFocus()) {
      manager->displayProgressBar(playing_titel, 80, 420, 450);
      manager->hideSymbol();
      manager->update();
    }
  }

  return( SUCCESS );

}

Result MP3State::Pause() {
  if (state == PLAYING) {
    pausemp3();
    state = PAUSED;
    manager->displaySymbol( "PAUSE" );
    manager->update();
  }
  else if (state == PAUSED) {
    unpausemp3();
    state = PLAYING;
    manager->displaySymbol( "PLAY" );
    manager->update();
  }
  return SUCCESS;
}

Result MP3State::Rewind() {
  if ( state == PLAYING ) {
    manager->displaySymbol( "REW" );
    manager->update();
    
    seek(-10);
  }
  return SUCCESS;
}

Result MP3State::Forward() {
  if ( state == PLAYING ) {
    manager->displaySymbol( "FF" );
    manager->update();
    
    seek(5);
  }
  return SUCCESS;
}

Result MP3State::seek(int seektime) {
  if (state == PAUSED) return SUCCESS;
  if (0 == seektime)   return SUCCESS;
  
  iseek->seekPercent(Rational(seektime,100));
  
  return SUCCESS;
}


void MP3State::playmp3(string filename) {
  GLOBAL_DEBUG_STREAM("Begin> MP3State::playmp3" << endl);
  if (state == STOPPED) {
    try {
      
      file_source->setFilename(filename);
      
      if (first_file) {
        first_file = false;
        
        readfile->initOutput();
        GLOBAL_DEBUG_STREAM("connect(readfile, mp3dec);" << endl);
        connect(readfile, mp3dec);
        readfile->activate();
        mp3dec->initOutput();
        GLOBAL_DEBUG_STREAM("connect(mp3dec, copy);" << endl);
        connect(mp3dec, copy);
        mp3dec->activate();
        copy->initOutput();
        
        if (hasFocus()) {
          /* copy audio to OSD for timedisplay */
          if (osd_node->isStarted()) {
            osd_node->stop();
          }
          // fw: flushing the OSD at this point has strange effects
          //         if (osd_node->isActivated()) {
          //           osd_node->flush();
          //         }
          if (osd_node->isActivated()) {
            osd_node->deactivate();
          }
          if (osd_node->isOutputInitialized()) {
            osd_node->deinitOutput();
          }
          connect(copy, string("1"), osd_node, string("stream2"));
        }
        
        connect(copy, string("0"), audio_play);
        copy->activate();
        
        if (hasFocus()) {
          /* copy audio to OSD for timedisplay */
          osd_node->initOutput();
          osd_node->activate();
        }
        
        audio_play->initOutput();
        audio_play->activate();
        
        manager->update();
      }
      
      if (readfile->isActivated()) {
        readfile->start();
      }
      if (mp3dec->isActivated()) {
        mp3dec->start();
      }
      
      if (hasFocus()) {
        if (osd_node->isActivated()) {
          osd_node->start();
        }
      }

      if (copy->isActivated()) {
        copy->start();
      }
      if (audio_play->isActivated()) {
        audio_play->start();
      }
      
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM("Exception in MP3State::playmp3" << endl);
      GLOBAL_ERROR_STREAM(e << endl);
      
      if (e.getParent()) {
        GLOBAL_ERROR_STREAM("Parent of Exception is " << e.getParent() << endl);
      }      
      throw e;
    }
  
    state = PLAYING;
  }
  GLOBAL_DEBUG_STREAM("End< MP3State::playmp3" << endl);
}

void MP3State::stopmp3() {
  if (state == PLAYING || state == PAUSED) {
    try {
      if (readfile->isStarted()) {
	readfile->stop();
      }
      if (mp3dec->isStarted()) {
	mp3dec->stop();
      }
      mp3dec->flush();
      ISyncReset_var mp3sync(mp3dec->getParentObject()->getCheckedInterface<ISyncReset>());
      mp3sync->syncReset();
      if (copy->isStarted()) {
         copy->stop();
      }
      copy->flush();
      
      if (hasFocus()) {
        /* copy audio to OSD for timedisplay */
        if (osd_node->isStarted()) {
          osd_node->stop();
        }
        if (osd_node->isActivated()) {
          // disconnectVideoGraph already deinited OSD, skip
          // fw: flushing the OSD has strange effects
          osd_node->flush();
        }
      }
      
      if (audio_play->isStarted()) {
	audio_play->stop();
      }
      audio_play->flush();

      if (hasFocus())
        osd_node->reachStarted();

      state = STOPPED;
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM("Exception in MP3State::stopmp3" << endl);
      GLOBAL_ERROR_STREAM(e << endl);
      if (e.getParent()) {
        GLOBAL_ERROR_STREAM("Parent of Exception: " << e.getParent()->getName() << endl);
      }
    }
  }
  // playback is stopped, reset progress
  current_progress = Rational(0);
}

void MP3State::pausemp3() {
  if (state == PLAYING) {
    app->setSynchronized( false );
    audio_play->stop();
  }
}

void MP3State::unpausemp3() {
  if (state == PAUSED) {
    audio_play->start();
  }
}

void MP3State::replaceToken(string* textStr, string token, string replacement) {
  while (textStr->find(token) != string::npos) {
    textStr->replace(textStr->find(token),token.length(),replacement);
  }
}

void MP3State::replaceTokens(string* textStr) {
  // replace '_' and german chars, french chars
  // because they can't be displayed by our font
  replaceToken(textStr, "_", " ");
  replaceToken(textStr, "", "ae");
  replaceToken(textStr, "", "oe");
  replaceToken(textStr, "", "ue");
  replaceToken(textStr, "", "ss");
  replaceToken(textStr, "", "Ae");
  replaceToken(textStr, "", "Oe");
  replaceToken(textStr, "", "Ue");
  replaceToken(textStr, "", "a");
  replaceToken(textStr, "", "a");
  replaceToken(textStr, "", "e");
  replaceToken(textStr, "", "e");
  replaceToken(textStr, "", "o");
  replaceToken(textStr, "", "o");
  replaceToken(textStr, "", "u");
  replaceToken(textStr, "", "u");
}

// Event Handling:
Result  MP3State::endTrack() {
  GLOBAL_DEBUG_STREAM ("receiveEndTrack" << endl);
  
  actual_playing_line++;
  playLine( actual_playing_line );
  
  return SUCCESS;
};

Result  MP3State::syncReset() {
  manager->displaySymbol( "PLAY" );
  manager->update();

  return SUCCESS;
};

Result MP3State::initSound() {
  
  // NMM-operations may throw exception:
  // catch them, output them, clean up and return FAILURE
  try {
    /*Begin> Create the request for the MP3ReadNode */
    NodeDescription readfile_req;
    readfile_req.setNodeName("MP3ReadNode");
    /*End< */
    
    /*Begin> Create the request for the MPEGAudioDecodeNode  */
    NodeDescription mp3dec_req;
    mp3dec_req.setNodeName("MPEGAudioDecodeNode");
    /*End< */
    
    /*Begin> Create the request for the CopyNode  */
    NodeDescription copy_req;
    copy_req.setNodeName("CopyNode");
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getNMMApplication()->getRegistry();
    {
      RegistryLock lock(registry);

      list<Response> readfile_response = registry.initRequest(readfile_req);
      list<Response> mp3dec_response = registry.initRequest(mp3dec_req);
      list<Response> copy_response = registry.initRequest(copy_req);
      
      if (readfile_response.empty() ||
          mp3dec_response.empty() ||
          copy_response.empty()) {
        GLOBAL_ERROR_STREAM("Getting Nodes failed" << endl);
        GLOBAL_ERROR_STREAM("Size of readfile_response " << readfile_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of mp3dec_response " << mp3dec_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of copy_response " << copy_response.size() << endl);
        throw Exception(string("A required node is not available\n") +
                        string("Size of readfile_response ") + readfile_response.size() + string("\n") +
                        string("Size of mp3dec_response ") + mp3dec_response.size() + string("\n") +
                        string("Size of copy_response ") + copy_response.size());
      }
      
      readfile = registry.requestNode(readfile_response.front());
      mp3dec = registry.requestNode(mp3dec_response.front());
      copy = registry.requestNode(copy_response.front());
    }
    /*End< */

  } catch (const Exception& e) {
    GLOBAL_ERROR_STREAM("MP3State::initSound(): requesting failed!" << endl);
    GLOBAL_ERROR_STREAM(e << endl);

    // cleanup allocated stuff
    if (readfile) {
      app->getNMMApplication()->getRegistry().releaseNode(*readfile);
      delete readfile;
      readfile = NULL;
    }

    if (mp3dec) {
      app->getNMMApplication()->getRegistry().releaseNode(*mp3dec);
      delete mp3dec;
      mp3dec = NULL;
    }

    if (copy) {
      app->getNMMApplication()->getRegistry().releaseNode(*copy);
      delete copy;
      copy = NULL;
    }

    return FAILURE;
  } catch (...) {
    cerr << "Something strang went wrong while requesting in MMBoxApplication::createGlobalNodes()" << endl;

    // no idea, how to clean up, if this happens!
    return FAILURE;
  }

  // step 2: configure & initialize the nodes
  try {
    file_source = readfile->getParentObject()->getCheckedInterface<IFileHandler>();
    iseek = readfile->getParentObject()->getCheckedInterface<ISeekable>();
    progress = readfile->getParentObject()->getInterface<IProgress>();
    time_progress = app->getAudioNode()->getParentObject()->getInterface<ITimeProgress>();

    IBufferSize_var buffersize(readfile->getParentObject()->getCheckedInterface<IBufferSize>());
    if (buffersize.get()) {
        buffersize->setBufferSize(1024);
    }

    readfile->init();

    mp3dec->init();

    ICopy_var copy_iface(copy->getParentObject()->getCheckedInterface<ICopy>());
    copy_iface->setNumberOfOutputs(2);
    copy->init();
  } catch (const Exception& e) {
    GLOBAL_ERROR_STREAM("MP3State::initSound(): configuring and initializing failed!" << endl);
    GLOBAL_ERROR_STREAM(e << endl);

    if (readfile->isInitialized()) {
      readfile->deinit();
    }

    if (mp3dec->isInitialized()) {
      mp3dec->deinit();
    }

    if (copy->isInitialized()) {
      copy->deinit();
    }

    if (file_source) {
      delete file_source;
      file_source = NULL;
    }

    if (iseek) {
      delete iseek;
      iseek = NULL;
    }

    if (progress) {
      delete progress;
      progress = NULL;
    }

    if (time_progress) {
      delete time_progress;
      time_progress = NULL;
    }

    return FAILURE;
  } catch (...) {
    cerr << "Something strang went wrong while configuring & initializing in MMBoxApplication::createGlobalNodes()" << endl;

    // no idea, how to clean up, if this happens!
    return FAILURE;
  }
  /*End< */

  // take audio_play from mmbox-app...
  audio_play = app->getAudioNode();
  if (audio_play->isStarted()) {
    audio_play->stop();
  }
  if (audio_play->isActivated()) {
    audio_play->deactivate();
  }
  if (audio_play->isOutputInitialized()) {
    audio_play->deinitOutput();
  }
  
  app->getDispatcher()->registerEventListener(ITrack::endTrack_event, &endTrack_listener);
  audio_play->getParentObject()->registerEventListener(ITrack::endTrack_event, app->getDispatcher());

  app->getDispatcher()->registerEventListener(ISyncReset::syncReset_event, &syncReset_listener);
  audio_play->getParentObject()->registerEventListener(ISyncReset::syncReset_event, app->getDispatcher());

  app->getDispatcher()->registerEventListener(ITrackDuration::trackDuration_event, &trackDuration_listener);
  audio_play->getParentObject()->registerEventListener(ITrackDuration::trackDuration_event, app->getDispatcher());
  
  return SUCCESS;
}

void MP3State::deInitSound() {
  try {

    /* BEGIN> Disconnect audio graph */
    if (audio_play->isActivated()) {
      audio_play->deactivate();
    }
    if (audio_play->isOutputInitialized()) {
      audio_play->deinitOutput();
    }
    
    osd_node->reachInitialized();
    
    if (copy->isActivated()) {
      copy->deactivate();
    }
    
    {
      IInputJack_var audioin_jack(audio_play->getInputStream("default"));
      if (audioin_jack.get() && audioin_jack->isConnected()) {
        disconnect_input(audio_play);
      }
    }
    
    {
      IInputJack_var osdin_jack(osd_node->getInputStream("stream2"));
      if (osdin_jack.get() && osdin_jack->isConnected()) {
        disconnect_input(osd_node, string("stream2"));
      }
    }
    
    if (copy->isOutputInitialized()) {
      copy->deinitOutput();
    }
    
    if (mp3dec->isActivated()) {
      mp3dec->deactivate();
    }
    
    {
      IInputJack_var copyin_jack(copy->getInputStream("default"));
      if (copyin_jack.get() && copyin_jack->isConnected()) {
        disconnect_input(copy);
      }
    }
    
    if (mp3dec->isOutputInitialized()) {
      mp3dec->deinitOutput();
    }
    
    if (readfile->isActivated()) {
      readfile->deactivate();
    }
    
    {
      IInputJack_var mp3decin_jack(mp3dec->getInputStream("default"));
      if (mp3decin_jack.get() && mp3decin_jack->isConnected()) {
        disconnect_input(mp3dec);
      }
    }
    
    if (readfile->isOutputInitialized()) {
      readfile->deinitOutput();
    }
    /* END< Disconnect audio graph */
    
    if (readfile) {
      app->getNMMApplication()->getRegistry().releaseNode(*readfile);
      delete readfile;
      readfile = NULL;
    
      if (progress != NULL) {
        delete progress;
        progress = NULL;
      }

      if (time_progress != NULL) {
        delete time_progress;
        time_progress = NULL;
      }
    
      if (file_source != NULL) {
        delete file_source;
        file_source = NULL;
      }
      if (iseek != NULL) {
        delete iseek;
        iseek = NULL;
      }
    }
  
    if (mp3dec) {
      app->getNMMApplication()->getRegistry().releaseNode(*mp3dec);
      delete mp3dec;
      mp3dec = NULL;
    }
  
    if (copy) {
      app->getNMMApplication()->getRegistry().releaseNode(*copy);
      delete copy;
      copy = NULL;
    }

  } catch (const Exception& e) {
    GLOBAL_ERROR_STREAM("Exception in MP3State::deInitSound" << endl);
    GLOBAL_ERROR_STREAM(e << endl);
    
    if (e.getParent()) {
      GLOBAL_ERROR_STREAM("Parent of Exception: " << e.getParent()->getName() << endl);
    }
  }
  osd_node->reachStarted();
}

Result MP3State::Stop() {
  
  stopmp3();

  if (hasFocus()) {
    playing_titel = "Playback stopped";
    manager->displayProgressBar(playing_titel, 80, 420, 450);
    //    manager->displayTime("", 640 - manager->getMaxWidth("medium")*4, 420);
    manager->resetTime();
    app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
    app->getOsdManager()->setProgress(current_progress.getNumerator());
    manager->updateProgress();
    manager->hideSymbol();
    manager->update();
  }
  
  return( SUCCESS );
}
