/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/utils/utils.hpp"

#include "playlist.hpp"

#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"
#include "nmm/interfaces/general/progress/IProgress.hpp"
#include "nmm/interfaces/general/progress/ITimeProgress.hpp"
#include "nmm/interfaces/general/progress/ITimeProgressListener.hpp"
#include "nmm/interfaces/general/ITrackDuration.hpp"
#include "nmm/interfaces/device/cdda/ICDDA.hpp"
#include "nmm/interfaces/device/cdda/ICDDB.hpp"
#include "nmm/interfaces/file/IBufferSize.hpp"

#include "nmm/base/utils/GraphURL.hpp"

#include <locale.h>
#include <libintl.h>
#include <iostream>

namespace NMM {
  
  Playlist::Playlist(string id, MMBoxApplication* _app)
    : MMBoxState(id, _app),
      state(STOPPED),
      progress_x(0),
      progress_y(0),
      with_video(false),
      with_audio(false),
      rootpath("/"),
      endTrack_listener(this, &Playlist::endTrack),
      setProgress_listener(this, &Playlist::setProgress),
      trackDuration_listener(this, &Playlist::trackDuration),
      setTimeProgress_listener(this, &Playlist::setTimeProgress),
      osd_node(NULL),
      manager(NULL),
      audio_node(NULL),
      display(NULL),
      hide_win(false),
      show_play(true),
      play_single_url(false),
      cn(NULL),
      cn_node(NULL),
      seeker(NULL),
      use_image_scaler(false),
      image_rotation(0)
  {
    string cdrompath;
    MMBoxConfig* config = app->getMMBoxConfig();
    if (!(config->getValue("rootpath", rootpath))) {
      rootpath="/";
    }
    if (!(config->getValue("cdrompath",cdrompath))) {
      cdrompath="/cdrom";
    }

    osd_node   = app->getOsdNode();
    audio_node = app->getAudioNode();
    display    = app->getDisplayNode();
    manager    = app->getOsdManager();

    playlist_gui = new WindowSelectorPlaylist(app->getOsdManager(),  &app->getOSDWidgetFactory(), dynamic_cast<PlaylistTemplateFunctions*>(this));
    playlist_gui->setRootpath(rootpath);
    playlist_gui->setApplication(app);
    playlist_gui->setCDRomPath(cdrompath);
   
    progress_x = 80;
    progress_y = 420;

    if (!app->getMMBoxConfig()->getValue("slideshowscaler", use_image_scaler)) {
      use_image_scaler = true;
    }
  }
  
  Playlist::~Playlist(){
    delete playlist_gui;
  }
  
  void Playlist::setWinProperties( int winnr, OSDElementComposit::win_properties_t* prop ) {
    playlist_gui->setWinProperties(winnr,prop);
  }
  
  
  Result Playlist::doCreate( string file = "") {
    GLOBAL_DEBUG_STREAM("Playlist: initialize: " << stateid << endl);
    
    if (FAILURE == app->createDefaultGraph( background )) {
      return FAILURE;
    }
    
    if (FAILURE == app->startDefaultGraph()) {
      return FAILURE;
    }
    
    playlist_gui->createWindows();

    /* register end track listener */
    app->getDispatcher()->registerEventListener(ITrack::endTrack_event, &endTrack_listener);

    /* register track duration listener */
    app->getDispatcher()->registerEventListener(ITrackDuration::trackDuration_event, &trackDuration_listener);

    resetOSD();
    hide_win = true;
    if(!file.empty()){
      hideOSD();
      play_single_url=true;
      playUrl(file);
    }
    playlist_gui->update();

    return( SUCCESS );
  }
  
  Result Playlist::doKill() {
    /* hide all on-screen items if state is left */
    manager->hideProgressBar();
    manager->hideSymbol();
    manager->hideTime();
    manager->hideColorBar();
    
    /* disconnect all nodes */
    if ( state != STOPPED ) {
      disconnectNodes();
      state = STOPPED;
    }
    else app->disconnectDefaultGraph();
    
    /* delete composite node */
    deleteCnNode();
    
    /* remove endtrack listener */
    app->getDispatcher()->removeEventListener(&endTrack_listener);

    /* remove trackDuration listener */
    app->getDispatcher()->removeEventListener(&trackDuration_listener);

    /* delete all windows */
    playlist_gui->deleteWindows();
    
    return( SUCCESS );
  }

  Result Playlist::doInitialize( string ) {
    GLOBAL_DEBUG_STREAM("Playlist: initialize: " << stateid << endl);

    /* if the state is reentered, only the on-screen items must be refreshed */
    playlist_gui->show();

    registerProgressListener();
    manager->displayProgressBar(playing_titel, progress_x, progress_y, 450 );
    manager->displayTime("", progress_x + 560 - manager->getMaxWidth("medium")*4, progress_y);
    manager->displayColorBar();    
    playlist_gui->updateColorBar();
    manager->updateProgress();
    if (state==PLAYING)
      playlist_gui->displaySymbol("PLAY");

    return( SUCCESS );
  }

  Result Playlist::doUninitialize() {
    GLOBAL_DEBUG_STREAM("Playlist: initialize: " << stateid << endl);

    /* if state is left, no progress information is needed */
    deRegisterProgressListener();
    manager->hideProgressBar();
    manager->hideTime();
    manager->hideColorBar();

    /* hide all on-screen items */
    playlist_gui->hide();

    return( SUCCESS );
  }

  void Playlist::registerProgressListener() {
    /* listen progress event on PlaybackNode */
    if (with_audio) {
      audio_node->getParentObject()->registerEventListener(IProgressListener::setProgress_event, &setProgress_listener);
    } else {
      app->getDisplayNode()->getParentObject()->registerEventListener(IProgressListener::setProgress_event, &setProgress_listener);
    }

    if (with_video) {
      app->getDisplayNode()->getParentObject()->registerEventListener(ITimeProgressListener::setTimeProgress_event,
								      &setTimeProgress_listener);
      app->getDisplayNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(true);
      app->getAudioNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(false);
    }
    else {
      audio_node->getParentObject()->registerEventListener(ITimeProgressListener::setTimeProgress_event,
							   &setTimeProgress_listener);
      app->getDisplayNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(false);
      app->getAudioNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(true);
    }


    /* tell node to produce progress informations */
    if (cn) {
      IProgress_var progress(cn->getInterface<IProgress>());
      if (progress.get()) {
        progress->sendProgressInformation(true);
      }
    }
  }


  void Playlist::deRegisterProgressListener() {
    if (with_audio) {
      audio_node->getParentObject()->removeEventListener(&setProgress_listener);
    } else {
      app->getDisplayNode()->getParentObject()->removeEventListener(&setProgress_listener);
    }

    if (with_video) {
      app->getDisplayNode()->getParentObject()->removeEventListener(&setTimeProgress_listener);
    }
    else {
      audio_node->getParentObject()->removeEventListener(&setTimeProgress_listener);
    }

    app->getDisplayNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(false);
    app->getAudioNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(false);
    
    /* tell node to produce no more progress informations */
    if (cn) {
      IProgress_var progress(cn->getInterface<IProgress>());
      if (progress.get()) {
        progress->sendProgressInformation(true);
      }
    }
  }
  
  Result Playlist::setProgress( u_int64_t& val, u_int64_t& max ) {
    /* set progress bar according to event information */
    manager->setMaxProgress( max );
    manager->setProgress( val);

    /* redraw progress bar */
    manager->updateProgress();
    manager->update();
    
    if (show_play) {
      playlist_gui->displaySymbol( "PLAY" );
      show_play=false;
    }
    return( SUCCESS );
  }

  Result Playlist::setTimeProgress(Time& current_media_time) {
    manager->setTime(current_media_time.sec);
    manager->update();
    return SUCCESS;
  }

  Result Playlist::trackDuration(Interval& duration) {
    track_duration = duration.sec;
    return DELETE;
  }
  
  Result Playlist::receiveMessage( string message ) {
    manager->deleteMessage();
    manager->update();
    
     /* the only message we can get is "DEL:filename" that indicates to delete
        the given file */
     unsigned int pos = message.find_first_of(':');
     if ( pos != string::npos ) {
       string temp;
       temp = message.substr( 0, pos );
       if ( temp == "DEL" ) {
         temp = ("rm -rf '" + message.substr( pos+1, message.length()))+"'";
         system( temp.c_str() );

         playlist_gui->refreshFiles();
       }
    }
     return SUCCESS;
  }
  
  Result Playlist::Down() {
    return playlist_gui->selectDown();
  }
  
  Result Playlist::Up() {
    return playlist_gui->selectUp();
  }

  Result Playlist::Left() {
    return playlist_gui->selectLeft();
  }

  Result Playlist::Right() {
    return playlist_gui->selectRight();
  }

  Result Playlist::Back() {
    GLOBAL_DEBUG_STREAM( "Playlist: Back" << endl );

    //deactivate();

    return( goToLastState() );
  }

  Result Playlist::colorKeyRed() {
    return playlist_gui->colorKeyRed();
  }

  Result Playlist::colorKeyGreen() {
    return playlist_gui->colorKeyGreen();
  }

  Result Playlist::colorKeyYellow() {
    return playlist_gui->colorKeyYellow();
  }

  Result Playlist::colorKeyBlue() {
    return playlist_gui->colorKeyBlue();
  }

  Result Playlist::Return() {
    playlist_gui->select();
    return(SUCCESS);
  }

  Result Playlist::doSelectPlaylistWindow(string url) {
    play_single_url = false;
    return playUrl(url);
  }
  
  Result Playlist::doSelectTrackWindow(string url) {
    play_single_url = true;
    return playUrl(url);
  }
  
  Result Playlist::resetOSD() {
    playing_titel = gettext("Playback stopped");
    manager->displayProgressBar( playing_titel, progress_x, progress_y, 450);
    manager->setProgress(0);
    manager->updateProgress();
    manager->displayTime("", progress_x + 560 - manager->getMaxWidth("medium")*4, progress_y);
    manager->setTime(0);
    manager->displayColorBar();
    playlist_gui->updateColorBar();
    return( SUCCESS );
  }
  
  Result Playlist::playUrl(string url)  {
    /* parse graph URL */
    /* if URL is empty (after rotating an image), use last URL without options */
    GraphURL graph_url;
    if(url.empty()) {
      graph_url = *gb.getURL();
      graph_url.clearOptions();
    } else {
      graph_url = GraphURL(url);
      if(!graph_url.isValid()) {
        GLOBAL_ERROR_STREAM("Invalid URL given" << std::endl);
        return FAILURE;
      }
    }
    
    /* set image rotation */
    {
      string rot_file = getRotationFileName(graph_url.getURIPath());
      ifstream in(rot_file.c_str());
      if (!in.fail()) {
        in >> image_rotation;
      }
    }    
    char rot_char = image_rotation ? '1' + image_rotation - 1 : '0';
    graph_url.setOption("image_rotation", string() + rot_char);
    graph_url.setOption("image_scaler", use_image_scaler ? "1" : "0");
    
    /* set URL */
    GLOBAL_DEBUG_STREAM("Playing \"" << url << "\"" << endl);
    string demux_audio_jack_tag = "audio"; // e.g. set this to audio0 or audio1 to explicitly choose audio track for demux nodes
    string demux_video_jack_tag = "video";
    gb.setURL(graph_url);
    
    // XXX: fw: ???
//     app->disconnectDefaultGraph();
    
    /* stop old Graph */
    if (state != STOPPED) {
      disconnectNodes();
      
      // XXX: fw: ???
//       app->createDefaultGraph(background);
//       app->startDefaultGraph();
//       resetOSD();
      
      state = STOPPED;
    } else {
      app->disconnectDefaultGraph();
    }
    
    gb.setAudioSink(app->getAudioNode());
    gb.setVideoSink(app->getOsdNode(), "stream1");
    gb.setDemuxAudioJackTag(demux_audio_jack_tag);
    gb.setDemuxVideoJackTag(demux_video_jack_tag);
    
    deleteCnNode();
    
    /* create new graph */
    cn = gb.createGraph(*app->getNMMApplication());
    
    if(NULL == cn) {
      GLOBAL_ERROR_STREAM("Could not greate Graph for URL (" << url << ")!" << endl);
      
      if (hasFocus()) {
        resetOSD();
        manager->update();
      }
      
      manager->displayMessage("Error", "Could not create Fileplayer","OK");
      changeState("OSDInputState");
      return FAILURE;
    }
    
    // XXX: fw: ???
//     deleteCnNode();
//     cn = new_cn;
    
    cn_node = cn->getCheckedInterface<INode>();
    seeker = cn->getInterface<ISeekable>();
    
    /* check if URL has video */
    with_video = (app->getOsdNode())->isInputConnected("stream1");
    with_audio = (app->getAudioNode())->isInputConnected();

    if (with_video) {
      app->getDisplayNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(true);
      app->getAudioNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(false);
    }
    else {
      app->getAudioNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(true);
      app->getDisplayNode()->getParentObject()->getCheckedInterface<ITimeProgress>()->setProgressUpdateEnabled(false);
    }
    
    if (!with_video && !with_audio) {
      GLOBAL_ERROR_STREAM("Could not find a video nor audio jack in Graph for URL (" << url << ")!" << endl);
      
      if (hasFocus()) {
        resetOSD();
        manager->update();
      }
      
      delete cn;
      cn = NULL;
      delete cn_node;
      cn_node = NULL;
      if (seeker) {
        delete seeker;
        seeker = NULL;
      }
      manager->displayMessage("Error","Could not find video/audio stream","OK");
      changeState("OSDInputState");
      return FAILURE;
    }
    
    if (with_video) {
      playlist_gui->hide();
      resetOSD();
      //if we have video but no seeker -> we view a picture -> put OSD into pause mode
      if (!seeker) 
        manager->setPause(true);
      else
        manager->setPause(false);
      app->setSynchronized(true);
      playing_titel = playlist_gui->getActualFile();
      manager->hideProgressBar();
      manager->hideTime();
      manager->hideColorBar();
      if (!with_audio) {
        app->setAudioSynchronized(false);
      }
    } else {
      app->createDefaultGraph(background);
      app->startDefaultGraph();
      playing_titel = playlist_gui->getActualFile();
      if ( hasFocus() ) {
        playlist_gui->show();
        manager->displayProgressBar( playing_titel, progress_x, progress_y, 450 );
        manager->updateProgress();
        
        playlist_gui->displaySymbol("PLAY");
      }
    }
    
    /* choose a small buffer size to make update progressbar frequently */
    IBufferSize_var buffersize(cn->getInterface<IBufferSize>());
    if (buffersize.get()) {
      if (with_video) {
        // XXX: fw: this causes artifacts with MPEG2 video
//         buffersize->setBufferSize(32*1024);
      } else {
        buffersize->setBufferSize(1024); 
      }
    }

    // register endTrack and trackDuration listener
    if (with_audio) {
      app->getAudioNode()->getParentObject()->registerEventListener(ITrack::endTrack_event, app->getDispatcher());
    } else {
      app->getDisplayNode()->getParentObject()->registerEventListener(ITrack::endTrack_event, app->getDispatcher());
    }
    
    app->getAudioNode()->getParentObject()->registerEventListener(ITrackDuration::trackDuration_event, &trackDuration_listener);
    app->getDisplayNode()->getParentObject()->registerEventListener(ITrackDuration::trackDuration_event, &trackDuration_listener);
    
    cn->reachStarted();
    if (with_audio)
      app->getAudioNode()->reachStarted();
    if (with_video) {
      app->getOsdNode()->reachStarted();
      display->reachStarted();
    }

    if (!seeker) {
      app->startDefaultGraph();
    }
    registerProgressListener();

    state = PLAYING;

    return( SUCCESS );
  }
  
  Result Playlist::endTrack() {
     GLOBAL_DEBUG_STREAM("Playlist::endTrack()" << endl);
     if (!seeker) {
       return SUCCESS;
     }

     if (play_single_url) {
       disconnectNodes();
       app->createDefaultGraph( background );
       app->startDefaultGraph();
       deleteCnNode();
       state = STOPPED;
       hide_win = false;
       playlist_gui->displaySymbol("STOPPED");
       resetOSD();
       showOSD();
       manager->update();
     } else {
       string temp = playlist_gui->getNextPlaylistfile();

       if (temp!="") {
         playUrl(temp);
       } else {
         disconnectNodes();
	 app->createDefaultGraph( background );
	 app->startDefaultGraph();
         deleteCnNode();
         state = STOPPED;
	 playlist_gui->show();
         playlist_gui->displaySymbol("STOPPED");
         resetOSD();
         showOSD();
         manager->update();
       }
     }
     return SUCCESS;
  }

  void Playlist::stopNodes() {
    if (display && display->isStarted()) {
      display->stop();
    }
    if (audio_node && audio_node->isStarted()) {
      audio_node->stop();
    }
    if (osd_node && osd_node->isStarted()) {
      osd_node->stop();
    }
    if (cn_node && cn_node->isStarted()) {
      cn_node->stop();
    }
  }

  void Playlist::flushAll() {
    if (display && display->isActivated()) {
      display->flush();
    }
    if (audio_node && audio_node->isActivated()) {
      audio_node->flush();
    }
    if (osd_node && osd_node->isActivated()) {
      osd_node->flush();
    }
    if (cn_node && cn_node->isActivated()) {
      cn_node->flush();
    }
  }
  
  void Playlist::deleteCnNode() {
    deRegisterProgressListener();
    
    if (seeker) {
      delete seeker;
      seeker = NULL;
    }
    if (cn_node) {
      delete cn_node;
      cn_node = NULL;
    }
    if (cn) {
      delete cn;
      cn = NULL;
    }
  }
  
  void Playlist::disconnectNodes() {    
    // de-register endTrack and trackDuration listener
    if (with_audio) {
      app->getAudioNode()->getParentObject()->removeEventListener(ITrack::endTrack_event, app->getDispatcher());
    } else {
      app->getDisplayNode()->getParentObject()->removeEventListener(ITrack::endTrack_event, app->getDispatcher());
    }
    app->getAudioNode()->getParentObject()->removeEventListener(ITrackDuration::trackDuration_event, app->getDispatcher());
    app->getDisplayNode()->getParentObject()->removeEventListener(ITrackDuration::trackDuration_event, app->getDispatcher());
    deRegisterProgressListener();
    stopNodes();
    flushAll();
    
    /* disconnect audio */
    if (with_audio) {
      GLOBAL_DEBUG_STREAM("Disconnecting audio" << endl);
      app->getAudioNode()->reachInitialized();
      cn_node->reachOutputInitialized();
      disconnect_input(app->getAudioNode());
      GLOBAL_DEBUG_STREAM("Disconnected audio" << endl);
    }
    
    /* disconnect video */
    if (with_video) {
      GLOBAL_DEBUG_STREAM("Disconnecting video" << endl);
      app->getOsdNode()->reachInitialized();
      cn_node->reachOutputInitialized();
      //may already be disconnected or not even connected
      try{
        disconnect_input(app->getOsdNode(), string("stream1"));
      } catch (const Exception& e) {
        GLOBAL_ERROR_STREAM(e<<endl);
      }
      GLOBAL_DEBUG_STREAM("Disconnected video" << endl);
    } else {
      GLOBAL_DEBUG_STREAM("Disconnecting DefaultGraph" << endl);
      app->disconnectDefaultGraph();
      GLOBAL_DEBUG_STREAM("Disconnected DefaultGraph" << endl);
    }

    /* deiniting composite node */
    GLOBAL_DEBUG_STREAM("deiniting CN" << endl);
    cn->reachConstructed();
    GLOBAL_DEBUG_STREAM("deinited CN" << endl);
    
    app->setSynchronized(false);
  }

  Result Playlist::Pause() {
    if (seeker) { 
      // source is seekable, so pausing should be possible

      if (state == PLAYING) {
        playlist_gui->displaySymbol( "PAUSE" );
	state = PAUSED;
        app->waitForSymbolChange( "PAUSE" );
        manager->update();
        app->getSynchronizer()->pause();
	manager->setPause(true);
      }
      else if (state == PAUSED) {
        if (with_video) {
          manager->setPause(false);
        }
	app->getSynchronizer()->wakeup();
	state = PLAYING;
	playlist_gui->displaySymbol( "PLAY" );
      }
    }
    else {
      // Source is not seekable, thus it is an image.
      // The image is rotated by 90 degrees, when the pause button is pressed. For each image, the new rotation is written
      // to a special rotation file.
      string rot_file = getRotationFileName(gb.getURL()->getURIPath());
      if (rot_file != "") {
	ofstream out(rot_file.c_str());
	if (!out.fail()) {
	  out << (image_rotation + 1) % 4;
	}
	else {
	  // If we can't write to the rotation file, simply rotate the image.
	  ++image_rotation;
	  GLOBAL_ERROR_STREAM("Could not save image rotation in " << rot_file << endl);
	}
      }

      playUrl("");
    }

    return( SUCCESS );
  }

  Result Playlist::Rewind() {
     if (state == PLAYING) {
       playlist_gui->displaySymbol( "REW" );

       if (seeker) {
         seeker->seekPercent(Rational(-5,100));
       } else {
	 string temp = playlist_gui->getPreviousPlaylistfile();

	 if (temp!="") {
	   playUrl(temp);	 
	 }
       }

       show_play=true;
     }

    return SUCCESS;
  }

  Result Playlist::Forward() {
     if (state == PLAYING) {
       playlist_gui->displaySymbol( "FF" );

       if (seeker) {
         seeker->seekPercent(Rational(5,100));
       } else {
	 string temp = playlist_gui->getNextPlaylistfile();

	 if (temp!="") {
	   playUrl(temp);
	 }
       }

       show_play=true;
     }

    return SUCCESS;
  }

  Result Playlist::Stop() {
    if ( state != STOPPED ) {
      disconnectNodes();
      app->createDefaultGraph( background );
      app->startDefaultGraph();
      deleteCnNode();
      state = STOPPED;
      playlist_gui->show();
      playlist_gui->displaySymbol("STOPPED");
      resetOSD();
    }
    return( SUCCESS );
  }

  Result Playlist::OSDKey() {

    if ( playlist_gui->isHidden() ) {
       showOSD();
     } else {
       if ( with_video && state==PLAYING ) {
         hideOSD();
       }
     }
    manager->update();
    return( SUCCESS );
  }

   Result Playlist::showOSD() {
     if (!hide_win) {
       manager->displayColorBar();
       playlist_gui->updateColorBar();
       
       playlist_gui->show();
       hide_win = true;
     } else hide_win = false;
     manager->update();
     manager->displayProgressBar( playing_titel, progress_x, progress_y, 450);
     manager->updateProgress();
     manager->displayTime("", progress_x + 560 - manager->getMaxWidth("medium")*4, progress_y);

     if ( state == PLAYING ) {
       playlist_gui->displaySymbol( "PLAY" );
     }
     if ( state == PAUSED ) {
       playlist_gui->displaySymbol( "PAUSE" );
     }
     manager->update();
     return( SUCCESS );
   }
  
   Result Playlist::hideOSD() {
     manager->hideProgressBar();
     manager->hideSymbol();
     manager->hideTime();
     manager->hideColorBar();
     manager->update();
     if ( hide_win ) {
       playlist_gui->hide();
       hide_win = false;
     }
     hide_win=true;
     return( SUCCESS );
   }

  void Playlist::doChangeState(string state) {
    changeState(state);
  }

  const string Playlist::getRotationFileName(const string& path) {
    string dir(path, 0, path.rfind("/"));
    string file(path, path.rfind("/") + 1);
    return dir + "/." + file + ".mmbox";
  }
}

