/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002, 2003
 *                    Network-Multimedia Workgroup,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass (dama@graphics.cs.uni-sb.de) 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TranscodingGroup.hpp"
#include "nmm/utils/utils.hpp"
#include <sys/stat.h>

namespace NMM {
  TranscodingGroup::TranscodingGroup(GraphDescription* g, list<NodeConfiguration>* n, const string& p, ClientRegistry& reg, IPluginManager& pm, JobManager* jm, TranscodeTemplateFunctions* f) :
    _protocol(p),
    _registry(reg),
    _pm(pm),
    _jobmanager(jm),
    _state_template(f),
    _running_jobs(0),
    _jobs_done(0),
    _jobs_error(0),
    _jobs_sum(0),
    _start_time(0),
    _end_time(0)
  {
    _graph = new GraphDescription(*g);
    _node_configuration = new list<NodeConfiguration>(*n),
    _input_files.clear();
    _input_file_names.clear();
  }

   TranscodingGroup::~TranscodingGroup()
  {
    for (uint i=0;i<_transcoding_job_list.size();i++)
      if (_transcoding_job_list[i])
        delete _transcoding_job_list[i];
    delete _graph;
    delete _node_configuration;
  }

  void TranscodingGroup::addInputFiles(const vector<GraphURL>& files)
  { 
    for (uint i=0;i<files.size();i++) {
      _input_files.push_back(GraphURL::urldecode(files[i].getURI()));
      _jobs_sum++;
    }
  }
  
  void TranscodingGroup::addInputFileNames(const vector<string>& files)
  {
    for (uint i=0;i<files.size();i++)
      _input_file_names.push_back(files[i]);
  }

  void TranscodingGroup::addInputFile(const string& file)
  {
    _input_files.push_back(file); 
    _jobs_sum++;
  }

  void TranscodingGroup::addInputFileName(const string& file)
  {
    _input_file_names.push_back(file);
  }
  
  void TranscodingGroup::setDescription(const string& des)
  { 
    _description = des; 
  }

  string TranscodingGroup::getDescription()
  {
    return _description;
  }

  void TranscodingGroup::setDistribute(const bool& dist)
  { 
    _distribute = dist; 
  }
  
  void TranscodingGroup::setPriorize(const bool& prio)
  {
    _priorize = prio; 
  }
  
  void TranscodingGroup::setParallel(const int par)
  {
    _parallel = par;
  }
  
  void TranscodingGroup::useRegistrys (const list<pair<string,int> >& reg_list)
  {
    _registry_list = reg_list;
  }
  
  void TranscodingGroup::useLocalhost(const  bool& use) 
  {
    _use_localhost = use;
  }
  
  void TranscodingGroup::setWorkingDirectory (const string& dir)
  { 
    _working_directory = dir;
  }
  
  list<list<string> > TranscodingGroup::getStatus() 
  { 
    list<list<string> > slist;
    for (uint i=0;i<_transcoding_job_list.size(); i++) {
      if (_transcoding_job_list[i]){
        list<string> slist2;
        slist2.push_back(itoa(i));
        slist2.push_back(_transcoding_job_list[i]->getActiveInputFile());
        slist.push_back(slist2);
      }
    }
    return slist;
  }
  
  list<string> TranscodingGroup::getJobStatus(const int& i) {
    list<string> slist;
    if (_transcoding_job_list[i]) {
      slist.push_back("File: "+_transcoding_job_list[i]->getActiveInputFile());
      slist.push_back("Host: "+_transcoding_job_list[i]->getHost());
      slist.push_back("Status: "+_transcoding_job_list[i]->getStatus());
      if (_transcoding_job_list[i]->getStatus()=="done")
	slist.push_back("used Time: "+_transcoding_job_list[i]->getRemaining());
      else 
	slist.push_back("approx. remaining: "+_transcoding_job_list[i]->getRemaining());
    }
    return slist;
  }

  void TranscodingGroup::getJobProgress(const int& jobid, int& val, int& max) {
    if (_transcoding_job_list[jobid]) {
      _transcoding_job_list[jobid]->getProgress(val,max);
    }
  }

  list<string> TranscodingGroup::getGroupStatus() {
    list<string> slist;
    slist.push_back("Total Jobs: "+itoa(_jobs_sum));
    slist.push_back("Jobs done: "+itoa(_jobs_done));
    slist.push_back("Jobs with errors: "+itoa(_jobs_error));
    int running_time;
    if (allDone()) {
      running_time = _end_time-_start_time;
    } else {
      running_time = _clock.getTime() - _start_time;
    }
    strstream time;
    if (running_time>=3600) {
      time<<running_time/3600<<" h ";
      running_time = running_time%3600;
    }
    if (running_time>=60) {
      time<<running_time/60<<" min ";
      running_time = running_time%60;
    }
    time<<running_time<<" secs";
  
    slist.push_back(string("Running time: ")+time.str());
    
    return slist;
  }

  void TranscodingGroup::deleteJob(const int& i) {
    if (_transcoding_job_list[i]) {
      if (_transcoding_job_list[i]->getStatus()=="running") {
        processJobs();
      }
      delete _transcoding_job_list[i];
      _transcoding_job_list[i] = NULL;
      _jobs_sum--;
    }
  }

  void TranscodingGroup::pauseJob(const int& i) {
    if (_transcoding_job_list[i]) 
      _transcoding_job_list[i]->pause();
  }

  void TranscodingGroup::setup() 
  {
    //first of all: check if we have a global _pre script we have to run
    struct stat info;
    if (stat((_script+"_pre.sh").c_str(),&info)==0) {
      system(("sh "+_script+"_pre.sh").c_str());
      GLOBAL_DEBUG_STREAM("Running a global _pre script for TranscodingGroup"<<endl);
    }

    //store starttime
    _start_time = _clock.getTime();

    //build up the TranscodingJob-list
    for (uint i=0;i<_input_files.size();i++) {
      TranscodingJob* job = new TranscodingJob(_protocol,_registry,_pm,_jobmanager);
      job->useRegistrys(_registry_list);
      job->setInputFile(GraphURL::urldecode(_input_files[i]));
      if (i<_input_file_names.size()){
        job->setFileName(_input_file_names[i]);
      }
      job->setTranscodingGroup(this);
      job->setDistribute(_distribute);
      job->setPriorize(_priorize);
      job->setWorkingDirectory(_working_directory);
      job->useLocalhost(_use_localhost);
      job->useScript(_script);
      _transcoding_job_list.push_back(job);
    }

    //start up as much jobs as specified by _parallel
    processJobs();
  }

  void TranscodingGroup::processJobs() 
  {
    //we could be called by external jobmanager AND internal from delete job -> we need mutex
    _mutex.lock();

    if (allDone()) {
      _end_time = _clock.getTime();
      int running_time = _end_time-_start_time;

      strstream time;
      if (running_time>=3600) {
        time<<running_time/3600<<" h ";
        running_time = running_time%3600;
      }
      if (running_time>=60) {
        time<<running_time/60<<" min ";
        running_time = running_time%60;
      }
      time<<running_time<<" secs";
      cerr<<"This TranscodingGroup was running for "<<_end_time-_start_time<<endl;
      cerr<<"In this time, "<<_jobs_done<<" of a total of "<<_jobs_sum<<" were done successful, "<<_jobs_error<<" gave an error."<<endl;
      //all Jobs are done. If available: run the global _post script
      struct stat info;
      if (stat((_script+"_post.sh").c_str(),&info)==0) {
        system(("sh "+_script+"_post.sh").c_str());
        GLOBAL_DEBUG_STREAM("Running a global _post script for TranscodingGroup"<<endl);
      }
    } else {
      //there must be a waiting job - find and start it
      for (uint i=0;i<_transcoding_job_list.size();i++) {
        if (_transcoding_job_list[i]) {
          if (_transcoding_job_list[i]->getStatus() == "waiting" && _running_jobs<_parallel) {
            _transcoding_job_list[i]->setGraphDescription(_graph);
            _transcoding_job_list[i]->setNodeConfiguration(_node_configuration);
            if (_transcoding_job_list[i]->setup())
              _running_jobs++;
	    else {
              _jobs_error++;
              //if the last job of the group produced an error, we need to set the endtime and run the post script
              //here, as processjobs will never be triggered again
              if (allDone()) {
                _end_time = _clock.getTime();
                //all Jobs are done. If available: run the global _post script
                struct stat info;
                if (stat((_script+"_post.sh").c_str(),&info)==0) {
                  system(("sh "+_script+"_post.sh").c_str());
                  GLOBAL_DEBUG_STREAM("Running a global _post script for TranscodingGroup"<<endl);
                }
              }
            }
          }
        }
      }
    }
    _mutex.unlock();
  }

  void TranscodingGroup::jobDone() {
    _jobs_done++;
    //if we get triggered by the JobManager we need to decrement the number of running jobs!
    if (_running_jobs>0) _running_jobs--;
  }

  bool TranscodingGroup::allDone() {
    for (uint i=0;i<_transcoding_job_list.size();i++) {
      if (_transcoding_job_list[i]) {
	if (_transcoding_job_list[i]->getStatus() == "running" || (_transcoding_job_list[i]->getStatus() == "waiting" ))
	  return false;
      }
    }
    return true;
  }

  string TranscodingGroup::mapFormat(const string& format) {
    return _state_template->mapFormat(format);
  }

  bool TranscodingGroup::spymode(const int& i) {
    if (_transcoding_job_list[i])
      return _transcoding_job_list[i]->spymode();
    else return false;
  }

  Result TranscodingGroup::spymodeRaw(const int& i, INode* node, const string& stream) {
    if (_transcoding_job_list[i])
      return _transcoding_job_list[i]->spymodeRaw(node,stream);
    else return FAILURE;
  }

  Result TranscodingGroup::spymodeTrans(const int& i, INode* node, const string& stream) {
    if (_transcoding_job_list[i])
      return _transcoding_job_list[i]->spymodeTrans(node,stream);
    else return FAILURE;
  }
  
  void TranscodingGroup::endSpymode(const int& i) {
    if (_transcoding_job_list[i])
      _transcoding_job_list[i]->endSpymode();
  }
}
