/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002, 2003
 *                    Network-Multimedia Workgroup,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass (dama@graphics.cs.uni-sb.de) 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TranscodingJob.hpp"

#include "nmm/utils/utils.hpp"
#include "nmm/utils/DirReader.hpp"
#include "nmm/utils/gdparse/IClicValueStream.hpp"

#include "nmm/interfaces/general/progress/IProgressListener.hpp"
#include "nmm/interfaces/general/progress/IProgress.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/base/resourcemgr/IResourceManager.hpp"
#include "nmm/interfaces/base/IThreadedNode.hpp"
#include "nmm/interfaces/device/cdda/ICDDA.hpp"
#include "nmm/interfaces/device/dvd/IDVD.hpp"

#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/connect.hpp"

#include "nmm/interfaces/base/INamed.hpp"
#include "TranscodingGroup.hpp"

#include <sstream>
#include <stdlib.h> //system
#include <sys/stat.h> //stat
#include <stdio.h>

namespace NMM {
  TranscodingJob::TranscodingJob(const string& _protocol, ClientRegistry& registry, IPluginManager& pm, JobManager* j) : 
    _graph(NULL),
    _graph_copy(NULL),
    _raw_graph(NULL),
    _trans_graph(NULL),
    _registry(registry),
    _pm(pm),
    act_progress(0),
    max_progress(0),
    _description(""),
    input_file(""),
    file_name(""),
    output_file(""),
    working_directory(""),
    _status(waiting),
    _spymode(NONE),
    _node_configuration(NULL),
    _distribute(true),
    _priori(false),
    _use_localhost(false),
    _jobmanager(j),
    protocol(_protocol),
    eof_counter(0),
    first_video_edge(0),
    trans_video_edge(0),
    first_video_node(0),
    trans_video_node(0),
    first_video_tag("default"),
    trans_video_tag("default"),
    raw_init(false),
    trans_init(false)
  {
  }

  TranscodingJob::~TranscodingJob() {
    if (_graph) {
      try{
        _registry.releaseGraph(*_graph);
      } catch (const Exception&) {
        GLOBAL_ERROR_STREAM("Transcoding graph already released"<<endl);
      }
      delete _graph;
      _graph = NULL;
    }

    if (_graph_copy) {
      delete _graph_copy;
      _graph_copy = NULL;
    }

    if (_raw_graph) {
      try {
        _registry.releaseGraph(*_raw_graph);
      } catch (const Exception& e) {
        GLOBAL_ERROR_STREAM("Error releasing raw-graph: "<<e<<endl);
      }
      delete _raw_graph;
      _raw_graph = NULL;
    }
    if (_trans_graph) {
      try {
        _registry.releaseGraph(*_trans_graph);
      } catch (const Exception& e) {
        GLOBAL_ERROR_STREAM("Error releasing trans-graph: "<<e<<endl);
      }
      delete _trans_graph;
      _trans_graph = NULL;
    }

    if (_node_configuration) {
      delete _node_configuration;
      _node_configuration = NULL;
    }
  }

  void TranscodingJob::useRegistrys(const list< pair<string,int> >& reg_list) {
    registry_list = reg_list;
  }

  Result TranscodingJob::setup() {
    GLOBAL_DEBUG_STREAM("Entering TranscodingJob setup()"<<endl);
    if (input_file==""){
      GLOBAL_ERROR_STREAM("No inputfile! Can't start TranscodingJob"<<endl);
      _status = error;
      return FAILURE;
    }

    //run per-file _pre script
    struct stat info;
    if (stat((_script+"_prefile.sh").c_str(),&info)==0) {
      system(("sh "+_script+"_prefile.sh"+" '"+input_file+"' '"+output_file+"'").c_str());
      GLOBAL_DEBUG_STREAM("Running a global _prefile script for TranscodingJob"<<endl);
    }

    //Resouce Manager
    IServerRegistry* reg;
    IResourceManager* res_mgr;

    string cpu_type;
    uint cpu_speed,cpu_num;
    float cpu_free;

    multimap<float,pair<string,int>,greater<float> > rating_map;

    if (_distribute) {
      GLOBAL_DEBUG_STREAM("getting remote hostratings"<<endl);
      for(list<pair<string,int> >::iterator it = registry_list.begin(); it != registry_list.end(); it++) {
        reg = _registry.getServerRegistry(RegistryName((*it).first, (*it).second));
        if (!reg) {
          continue;
        }
        res_mgr = reg->getResourceManager();
        if (res_mgr) {
          //update the cpuinfo more often
          res_mgr->setCPUUpdateIntervall(1);
          res_mgr->getCPUInfo(cpu_type,cpu_speed,cpu_num,cpu_free);
          rating_map.insert(make_pair(cpu_speed*cpu_free,*(it)));
        }
      }
    }

    //get local rating
    if (_use_localhost || rating_map.empty()) {
      GLOBAL_DEBUG_STREAM("getting localhost rating"<<endl);
      res_mgr = _registry.getRegistryList()[RegistryName(UserInfo().getHostname(), ClientRegistry::default_port)].server_registry->getResourceManager();
      //update the cpuinfo more often
      res_mgr->setCPUUpdateIntervall(1);
      res_mgr->getCPUInfo(cpu_type,cpu_speed,cpu_num,cpu_free);
      rating_map.insert(make_pair(cpu_speed*cpu_free,RegistryName(UserInfo().getHostname(), ClientRegistry::default_port)));
    }
    
    //distribute plugins
    if (_distribute) {
      string host = (*rating_map.begin()).second.first;
      int    port = (*rating_map.begin()).second.second;
      opt_host = (*rating_map.begin()).second.first;
      cerr<<"distribute to: "<<opt_host<<endl;

      SequenceValue<NodeDescription>& nd_list(_graph->getNodeDescriptionList());
      for (uint i=0;i<nd_list.size();++i) {
	if(_graph->getInEdgeList(i).size()>0 && _graph->getOutEdgeList(i).size()>0) {
	  nd_list[i]->setLocation(host);
          nd_list[i]->setPort(port);
	} else nd_list[i]->setLocation(getHostname());
      }
    }
   
    //create preview
    //    _raw_graph = new GraphDescription(*_graph);
    //_trans_graph = new GraphDescription(*_graph);
    _graph_copy = new GraphDescription(*_graph);
    SequenceValue<NodeDescription>& nd_list1(_graph->getNodeDescriptionList());
    SequenceValue<NodeDescription>& nd_list2(_graph_copy->getNodeDescriptionList());
    for (uint i=0;i<nd_list1.size();++i) {
      nd_list1[i]->setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);
      nd_list2[i]->setSharingType(NodeDescription::SHARED);
      //      nd_list3[i]->setSharingType(NodeDescription::SHARED);
    }

    //request graph
    try{
      _registry.requestGraph(*_graph);
    } catch (const Exception& e) {
      cout<<"couldn't request graph because of: "<<e<<endl;
      _status=error;
      return FAILURE;
    }

    SequenceValue<INode>& inode_list = _graph->getINodeList();
    
    //handling input
    for (unsigned int i = 0; i< inode_list.size(); ++i) {
      if(_graph->getInEdgeList(i).size() == 0) {
        inode_list[i]->getParentObject()->registerEventListener(IProgressListener::setProgress_event, new  TEDObject2<TranscodingJob, u_int64_t, u_int64_t>(this,&TranscodingJob::progressHandler));
        
        IProgress_var prog(inode_list[i]->getParentObject()->getInterface<IProgress>());
        if (prog.get()) {
          prog->sendProgressInformation(true);
	  //	  prog->setProgressInterval(50);
        }
        if (protocol == "audiocd") {
          int nr = atoi(input_file.substr(input_file.find_last_of("=")+1,input_file.size()).c_str());
          Event* event = new Event(ICDDA::setTrack_event,new TInValue<int>(nr));
          _graph->getNodeDescription(i).addNodeEvent(*event, INode::ACTIVATED);
          delete event;
        } else if (protocol == "dvd") {
          int nr = atoi(input_file.substr(input_file.find_first_of("=")+1,input_file.size()).c_str());
          Event* event = new Event(IDVD::select_event, new TInValue<unsigned int>(nr), new TInValue<unsigned int>(1), new TInValue<unsigned int>(1));
          _graph->getNodeDescription(i).addNodeEvent(*event, INode::OUTPUT_INITIALIZED);
          delete event;
        } else {
          IFileHandler_var i_handler(inode_list[i]->getParentObject()->getInterface<IFileHandler>());
          if (i_handler.get()) {
            i_handler->setFilename(input_file);
            break; // currently only works for a single source
          }
        }
      }
    }

    //handling output
    for (unsigned int i = 0; i< inode_list.size(); ++i) {
      if(_graph->getOutEdgeList(i).size() == 0) { 
        sink_list.push_back(i);
        inode_list[i]->getParentObject()->registerEventListener(ITrack::endTrack_event, 
								new TEDObject0<TranscodingJob, Result (TranscodingJob:: * )(), Result >
								(this, &TranscodingJob::eofHandler));
        inode_list[i]->getParentObject()->registerEventListener(ITrack::finished_event, 
								new TEDObject0<TranscodingJob, Result (TranscodingJob:: * )(), Result >
								(this, &TranscodingJob::eofHandler));
      }
    }

    /* Begin> Configure the nodes according to its event-description and store the events in the NodeDescription */
    for (list<NodeConfiguration>::iterator it = _node_configuration->begin(); it != _node_configuration->end(); ++it) {
      if (!it->getEventList().empty()) {
        for (list<NCElement>::iterator it1 = it->getEventList().begin(); it1 != it->getEventList().end(); ++it1) {
          /* create an ClicValueStream */
          IClicValueStream iv(it1->event_key, it1->argument_list);
          
          /* get the inode */
          INode* inode = _graph->getINode(*it->getNodeDescription());

          /* request the event from the node */
          Event* event= inode->getParentObject()->matchEventPrototype( iv.getEventKey() );

          if (!event) {
            throw Exception ("Unknown event with key " + iv.getEventKey() + " in node " + it->getNodeDescription()->getNodeName() );
          }

          /* read the parameters from the string */
          iv.readEvent(*event);

          /* and add the event to the NodeDescription */
          _graph->getNodeDescription(it->getNodeDescription()->getInternalID()).addNodeEvent(*event, it1->state);
          /* delete the old event */
          delete event;
        }
      }
    }
    /* End< */
    
    try{
      _graph->realizeGraph();

      //look for first video/raw edge
      for (unsigned int i = 0; i< inode_list.size(); ++i) {
        list<int> temp = _graph->getOutEdgeList(i);
        if (temp.size() != 0) {
          for (list<int>::iterator iter=temp.begin();iter!=temp.end();iter++) {
            string format;
            if ( _graph->getEdge(*iter).getFormat()) {
              format = _graph->getEdge(*iter).getFormat()->getFormatType();
            }
            else continue;
//             if (format.find("video") != string::npos) {
//               if (!first_video_edge) {
//                 //We have found the first edge with video
//                 first_video_edge = (*iter);
//                 first_video_tag  = _graph->getEdge(*iter).out_tag.getValue();
//                 //                cerr<<_graph->getEdge(*iter).out_tag.getValue()<<endl;
//                 first_video_node = i;
//               }
//               if ((*iter) > trans_video_edge) {
//                 trans_video_edge = (*iter);
//                 trans_video_tag  = _graph->getEdge(*iter).out_tag.getValue();
//                 trans_video_node = i;
//               }
//             }
            
            if (format == "video/raw") {
              if (!first_video_edge) {
                first_video_edge = *iter;
                first_video_node = i;
              }
              break;
            }
            else if (format.find("video") != string::npos) {
              //find LAST non-raw edge -> this is the transcoded videostream
              if ((*iter) > trans_video_edge) {
                trans_video_edge = (*iter);
                trans_video_node = i;
              }
            }
          }
        }
      }

      for (list<int>::iterator iter = sink_list.begin(); iter!=sink_list.end();iter++) {
        list<int> temp = _graph->getInEdgeList(*iter);
        string outformat = "";
        if (temp.size() == 1) 
          outformat = _graph->getEdge(temp.front()).getFormat()->getFormatType();
        
        ISinkFileHandler_var o_handler(inode_list[*iter]->getParentObject()->getInterface<ISinkFileHandler>());
        string oformat = _group->mapFormat(outformat);
        if (oformat=="")
          oformat="out";

        if (o_handler.get()) {
          if (!working_directory.empty()) {
            DirReader tempreader;
            if (protocol == "audiocd") {

              /* create directory from current date */
              time_t today = time(NULL);
              tm today_tm;
              localtime_r(&today, &today_tm);
              char buffer2[128];
              snprintf(buffer2, 128, "%04d-%02d-%02d, %02d:%02d", today_tm.tm_year + 1900, today_tm.tm_mon + 1, today_tm.tm_mday, today_tm.tm_hour, today_tm.tm_min);
              string today_str = buffer2;
              string today_dir = working_directory + string("/CD-Grab ") + today_str;

              if (!DirReader::makeDir(today_dir)) {
                GLOBAL_ERROR_STREAM("TranscodingState: Could not create directory "+today_dir
                                    + ". File will be put into " + working_directory<<endl);
                output_file = working_directory + string("/") + file_name + "." + oformat;
              } else {
                output_file = today_dir + string("/") + file_name + "." + oformat;
              }
            } 
            else if (protocol == "dvd") {
              /* create directory from current date */
              time_t today = time(NULL);
              tm today_tm;
              localtime_r(&today, &today_tm);
              char buffer2[128];
              snprintf(buffer2, 128, "%04d-%02d-%02d, %02d:%02d", today_tm.tm_year + 1900, today_tm.tm_mon + 1, today_tm.tm_mday, today_tm.tm_hour, today_tm.tm_min);
              string today_str = buffer2;
              string today_dir = working_directory + string("/DVD-Grab ") + today_str;

              int nr = atoi(input_file.substr(input_file.find_first_of("=")+1,input_file.size()).c_str());
              input_file = (string("DVD-Title ")+nr);

              if (!DirReader::makeDir(today_dir)) {
                GLOBAL_ERROR_STREAM("TranscodingState: Could not create directory "+today_dir
                                    + ". File will be put into " + working_directory<<endl);
                output_file = working_directory + string("/") + "DVD-Grab Title " + nr + "." + oformat;
              } else {
                output_file = today_dir + string("/") + "DVD-Grab Title " + nr + "." + oformat;
              }
            } else {
              string ifile = tempreader.getFileName(input_file);
              ifile = ifile.substr(0,ifile.find_last_of("."));

              //look for directory
              string temp = input_file.substr(0,input_file.find_last_of("/"));
              string temp2 = temp.substr(temp.find_last_of("/"));

              if (!DirReader::makeDir(working_directory + string("/") + temp2)) {
                GLOBAL_ERROR_STREAM("TranscodingState: Could not create directory "+working_directory+string("/")+temp2
                                    + ". File will be put into " + working_directory<<endl);
                output_file = working_directory + string("/") + ifile + "." + oformat;
              } else {
                output_file = working_directory + string("/") + temp2 + string("/") + ifile + "." + oformat;
              }
            }
            o_handler->setFilename(output_file);
          }
        }
      }

      _graph->activateGraph();

      //lower priority
      if (_priori) {
        GLOBAL_DEBUG_STREAM("Lowering prioriy of the nodes"<<endl);
        for (uint i=0;i<inode_list.size();++i) {
          IThreadedNode_var thread(inode_list[i]->getParentObject()->getCheckedInterface<IThreadedNode>());
          thread->lowerPriority(15);
        }
      }

      _graph->startGraph();
      _start_time = clock.getTime();
      _status = running;
    } catch(const Exception&) {
      _status = error;
      return FAILURE;
    }
    return SUCCESS;
  }
  
  Result TranscodingJob::eofHandler() {
    //we have to wait for as many eof-events as we have sinks in the graph
    eof_counter++;

    if (eof_counter == sink_list.size()) {
      _processing_time = (clock.getTime()-_start_time).sec;
      act_progress=max_progress;
      _jobmanager->stopJob(this,_group);
    }
    return SUCCESS;
  }
  
  Result TranscodingJob::progressHandler(u_int64_t& val, u_int64_t& max) {
    act_progress = val;
    max_progress = max;
    _jobmanager->update();
    return SUCCESS;
  }
    

  void TranscodingJob::eof() {
    _status = done;
    if (_node_configuration) {
      delete _node_configuration;
      _node_configuration = NULL;
    }
    if (_graph) {
      try{
	_registry.releaseGraph(*_graph);
      } catch (const Exception& e) {
	GLOBAL_ERROR_STREAM(e<<endl);
      }
      delete _graph;
      _graph = NULL;
    }
    if (_raw_graph) {
      try{
	_registry.releaseGraph(*_raw_graph);
      } catch (const Exception& e) {
	GLOBAL_ERROR_STREAM(e<<endl);
      }
      delete _raw_graph;
      _raw_graph = NULL;
    }
    if (_trans_graph) {
      try{
      _registry.releaseGraph(*_trans_graph);
      } catch (const Exception& e) {
	GLOBAL_ERROR_STREAM(e<<endl);
      }
      delete _trans_graph;
      _trans_graph = NULL;
    }


    
    //run per-file _post script
    struct stat info;
    if (stat((_script+"_postfile.sh").c_str(),&info)==0) {
      system(("sh "+_script+"_postfile.sh"+" '"+input_file+"' '"+output_file+"'").c_str());
      GLOBAL_DEBUG_STREAM("Running a global _postfile script for TranscodingJob"<<endl);
    }
  }
    
  void TranscodingJob::pause() {
    switch (_status) {
    case running:
      _graph->stopGraph();
      _status = paused;
      break;
    case paused:
      _graph->startGraph();
      _status = running;
      break;
    default:
      return;
    }
  }

  string TranscodingJob::getStatus() {
    switch (_status) {
    case done:
      return "done";
      break;
    case waiting:
      return "waiting";
      break;
    case paused:
      return "paused";
      break;
    case running:
      return "running";
      break;
    case error:
      return "error";
      break;
    }
    return "";
  }
  
  string TranscodingJob::getDescription() {
    return _description;
  }

  string TranscodingJob::getRemaining() {
    if (_status != running && _status != done)
      return "---";
    else {
      int rem;
      if (_status==done)
	rem = _processing_time;
      else {
	float progress = (((double)act_progress/max_progress)*100);
	Interval processing = clock.getTime()-_start_time;
	rem = (int)((processing/progress)*100 - processing.sec);
      }
      strstream time;
      if (rem>=3600) {
        time<<rem/3600<<" h ";
        rem = rem%3600;
      }
      if (rem>=60) {
        time<<rem/60<<" min ";
        rem = rem%60;
      }
      time<<rem<<" secs";
      return time.str();
    }
  }
  
  void TranscodingJob::getProgress(int& val, int& max) {
    val = act_progress;
    max = max_progress;
  }

  Result TranscodingJob::spymodeRaw(INode* node, const string& stream) {
    GLOBAL_DEBUG_STREAM("Entering SpymodeRaw"<<endl);
    if (_spymode != NONE)
      endSpymode();

    _raw_graph = _graph_copy->copy();

    if (first_video_node) {
      GLOBAL_DEBUG_STREAM("Creating edges for SpymodeRaw"<<endl);
      NodeDescription nd2;
      nd2.setNodeName(node->getParentObject()->getInterface<INamed>()->getName());
      nd2.setSharingType(NodeDescription::EXCLUSIVE);
      
      _raw_graph->addEdge(_raw_graph->getNodeDescription(first_video_node),nd2,stream);
      _raw_graph->setINode(nd2,*node);
      _raw_graph->setReleaseFlag(nd2,false);
      
      try {
        GLOBAL_DEBUG_STREAM("Requesting Spymode-Raw graph"<<endl);
        _registry.requestGraph(*_raw_graph);
        _raw_graph->realizeGraph();
        _raw_graph->startGraph();
      } catch (const Exception& e) {
        GLOBAL_ERROR_STREAM("Error requesting Spymode-Raw graph: "<<e<<endl);
	return FAILURE;
      }

      _spymode = RAW;
      GLOBAL_DEBUG_STREAM("Started SpymodeRaw"<<endl);
      return SUCCESS;
    }
    GLOBAL_ERROR_STREAM("No video for spymode"<<endl);
    return FAILURE;
  }

  Result TranscodingJob::spymodeTrans(INode* node, const string& stream) {
    GLOBAL_DEBUG_STREAM("Entering Spymode-Trans"<<endl);
    if (_spymode != NONE) 
      endSpymode();

    _trans_graph = _graph_copy->copy();

    if (trans_video_node)  {
      GLOBAL_DEBUG_STREAM("Creating edges for Spymode-Trans"<<endl);
      Format* trans = _graph->getEdge(trans_video_edge).getFormat();
      NodeDescription nd;
      nd.addFormat(*trans,Format("video/raw"));
      nd.setNodeType(INode::CONVERTER);
      
      NodeDescription nd2;
      nd2.setNodeName(node->getParentObject()->getInterface<INamed>()->getName());
      nd2.setSharingType(NodeDescription::EXCLUSIVE);
      
      _trans_graph->addEdge(_trans_graph->getNodeDescription(trans_video_node),trans_video_tag,nd);
      _trans_graph->addEdge(nd,nd2,stream);
      _trans_graph->setINode(nd2,*node);
      _trans_graph->setReleaseFlag(nd2,false);

      try {
        GLOBAL_DEBUG_STREAM("Requesting Spymode-Trans"<<endl);
        _registry.requestGraph(*_trans_graph);
        _trans_graph->realizeGraph();
        _trans_graph->startGraph();
      } catch (const Exception& e) {
        GLOBAL_ERROR_STREAM("Error requesting Spymode-Trans: "<<e<<endl);
	return FAILURE;
      }
      
      _spymode = TRANS;
      GLOBAL_DEBUG_STREAM("Started Spymode-Trans"<<endl);
      return SUCCESS;
    } else {
      GLOBAL_ERROR_STREAM("No video for spymode"<<endl);
      return FAILURE;
    }
  }
  void TranscodingJob::endSpymode() {
    if (_spymode==RAW) {
      GLOBAL_DEBUG_STREAM("Releasing raw-preview spymode graph"<<endl);
      try {
        _registry.releaseGraph(*_raw_graph);
      } catch (const Exception&) {
        GLOBAL_ERROR_STREAM("Preview graph already released"<<endl);
      }
      delete _raw_graph;
      _raw_graph = NULL;
      GLOBAL_DEBUG_STREAM("Raw-preview graph released"<<endl);
      _spymode=NONE;
    }
    else if (_spymode==TRANS) {
      GLOBAL_DEBUG_STREAM("Releasing transcode-preview spymode graph"<<endl);
      try{
        _registry.releaseGraph(*_trans_graph);
      } catch (const Exception&) {
        GLOBAL_ERROR_STREAM("Transcode graph already released"<<endl);
      }
      delete _trans_graph;
      _trans_graph = NULL;
      GLOBAL_DEBUG_STREAM("Released Spymode-Trans graph"<<endl);
      _spymode=NONE;
    }
  }
}
