/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef MMBOXAPPLICATION_HPP
#define MMBOXAPPLICATION_HPP

// for MMBoxEventDispatcher
#include <queue>
#include "nmm/base/EventDispatcher.hpp"
#include "nmm/base/IReceiver.hpp"
#include "nmm/utils/thread/ThreadedObject.hpp"

// include used interfaces
#include "nmm/interfaces/video/osd/manager/IOSDManager.hpp"
#include "nmm/interfaces/video/osd/manager/IWidgetFactory.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/base/sync/IMultiAudioVideoSynchronizer.hpp"
#include "nmm/interfaces/video/osd/manager/ISymbolChange.hpp"
// include implemented interfaces
#include "nmm/interfaces/apps/mmbox/IMMBoxApplicationImpl.hpp"

#include "mmboxstate.hpp"

#include "nmm/utils/NMMConfig.hpp"

namespace NMM {

  class NMMApplication;
  class BitmapFont;
  class MultiAudioVideoSynchronizer;
  class INode;
  class TimerProducer;

  class MMBoxConfig : public NMMConfig {
  public:  
    MMBoxConfig();
    MMBoxConfig(const string);
  };
  
  class MMBoxEventDispatcher :
    public EventDispatcher,
    public IEventReceiver,
    public ThreadedObject {

  private:
    Result internal_dispatch(const Event&, const IEventSender* = 0);

  protected:
    queue<pair<Event*, const IEventSender*> > event_queue;
    ThreadMutex event_queue_mutex;
    ThreadCondition event_queue_cv;
    const IEventSender* current_sender;
    bool dispatching;
    bool is_running; // we need our own flag!

    void run();

  public:

    MMBoxEventDispatcher();
    ~MMBoxEventDispatcher();

    // IEventReceiver Interface
    virtual Result receiveEvent(const CEvent&, const IEventSender* = 0 );
    virtual Result receiveEvent(const Event&, const IEventSender* = 0 );

    virtual Result receiveEvent(const CEvent&, const IEventSender* = 0 ) const { return FAILURE; } // don't know why these methods are there
    virtual Result receiveEvent(const Event&, const IEventSender* = 0 )  const { return FAILURE; } // don't know why these methods are there

    //   void start() { activateThread(); }
    //   void stop() { deactivateThread(); }

    /** This method returns the current EventSender
     */
    inline const IEventSender* getCurrentSender() const { return event_queue.front().second; }
    
  };


  typedef map<string, MMBoxState*> string_to_state_t;

  class MMBoxApplication : public IMMBoxApplicationImpl {
    typedef map<string, BitmapFont*> fonts_t;

  protected:
    // nmm application
    NMMApplication* nmmapp;
  
    // own nodes
    INode* pngread_node;
    INode* converter_node;
    INode* display_node;
    INode* audio_node;
    INode* osd_node;
  
    // own objects
    MultiAudioVideoSynchronizer* sync;
  
    // other interfaces
    IMultiAudioVideoSynchronizer* isync;
    ISynchronizedSink* audio_sync;
    ISynchronizedSink* video_sync;
    IOSDManager* manager;
    IWidgetFactory* factory;
  
    string                    config_filename;
    MMBoxConfig*              config;
  
    TimerProducer*		   timerproducer;
    MMBoxEventDispatcher*          dispatcher;
  
    bool                      fullscreen;
    bool                      use_mga;
    
    /** Using an external synchronizer.
     * This flag is set when the audio controller is connected to an external synchronizer (such as the handover synchronizer
     * of an audio handover client).
     * This is a temporary hack for CeBit 2004 and will be removed when it is no longer needed.
     */
    bool                      __use_external_synchronizer;
    
    int                       width;
    int                       height;
    double                    framerate;
    
    /* thread condition for waiting for symbol change */
    ThreadMutex               symbol_mutex;
    ThreadCondition           symbol_cond;
    string                    current_symbol;
    
    /* actual state */
    MMBoxState*               _state;
  
    /* active state that needs audio node */
    MMBoxState*               audio_state;
  
    /* active state that needs video node (background) */
    MMBoxState*               video_state;
  
    /* Mapping of string ids to MMBoxState instances */
    string_to_state_t         stateinstance;
  
    /* list of all active states */
    list<MMBoxState*>         active_states;
  
    /**
     * Last state id.
     * here the different MMBoxState instances
     * can read the last stateid, the instances themselves are resbonsible
     * to update this stateid, normally the instance should read out this id
     * when they are initialized, save it and then overwrite with its own stateid string, so
     * that the child state can read out it (this is done in the initialize() method automatically)
     * this mechanism is needed to switch from a state to its parent state
     */
    string last_stateid;
  
    /** Delete all registered state instances */
    void deleteStateInstances();
  
    friend class MMBoxState;
    Result changeState( string stateid, string message="" );
  
    void addToActiveStates( MMBoxState* );
    void delFromActiveStates( MMBoxState* );
  
  public:
    MMBoxApplication();
    ~MMBoxApplication();

    /**
     *  Set use external synchronizer flag.
     *  Set whether the audio controller is connected to a a external synchronizer.
     */
    inline void setUseExternalSynchronizer(bool flag) {
      __use_external_synchronizer = flag;
    }
    
    /**
     *  Get use external synchronizer flag.
     *  Get whether the audio controller is connected to a an external synchronizer.
     */
    inline bool getUseExternalSynchronizer() const {
      return __use_external_synchronizer;
    }

    /**
     * Kills the given state and goes to last state
     * \param state_id the state to be killed
     */
    void killState(const string& state_id);

    Result Up();
    Result Down();

    Result VCR_Up();
    Result VCR_Down();

    Result Left();
    Result Right();

    Result Return();
    Result Back();

    Result Pause();
    Result Rewind();
    Result Forward();
    Result Record();
    Result Stop();

    Result OSDKey();

    Result specialKey01();
    Result specialKey02();

    Result colorKeyRed();
    Result colorKeyGreen();
    Result colorKeyYellow();
    Result colorKeyBlue();

    Result numKey0();
    Result numKey1();
    Result numKey2();
    Result numKey3();
    Result numKey4();
    Result numKey5();
    Result numKey6();
    Result numKey7();
    Result numKey8();
    Result numKey9();

    Result REVEAL();
    Result reInitActualState();
    Result receiveMessage( string message );
    
    bool isStateActive( MMBoxState* );
    bool isStateActive( string stateid );
  
    typedef list<MMBoxState*> mmbox_state_list_t;
  
    const mmbox_state_list_t* getActiveStates();
    MMBoxState* getState( string stateid );
    MMBoxState* getActualState();
    string getActualStateID();
  
    /** Some startup stuff: setting font, ... */
    Result init();
  
    /** Delete unused nodes, etc. */
    Result deinit();
  
    /**
     * Put state with given id into stateinstance map
     * @param state instance of the state
     */
    void registerState( MMBoxState* );
  
    /** Set actual state that needs audio node */
    void registerAudioState( MMBoxState* );
  
    /** Set actual state that needs video node (background) */
    void registerVideoState( MMBoxState* );
  
    /** Get actual state that needs audio node */
    MMBoxState* getAudioState();
  
    /** Get actual state that needs video node (background) */
    MMBoxState* getVideoState();
  
    /**
     * Initialize of menu instances.
     * @param id menu id of first menu
     */
    Result initStates( string );
  
  
    /** Set DisplayNode. */
    void setDisplayNode( INode* _display );
  
    /** Set AudioNode. */
    void setAudioNode( INode* _audio );
  
    /** Set OSDNode. */
    void setOsdNode( INode* _osd );
  
    /** Set Configuration file name. */
    void setConfigFile(const string _filename);
  
    /** Turn synchronization on or off */
    void setSynchronized( bool s );

    /** Turn audio synchronization on or off */
    void setAudioSynchronized( bool s );

    /** Turn video synchronization on or off */
    void setVideoSynchronized( bool s );
  
  
    /** audio/video node are reset. */
    void syncReset();

    /**
     * get DisplayNode
     */
    INode* getDisplayNode();
  
    /**
     * get AudioNode
     */
    INode* getAudioNode();

    /**
     * get OSDNode
     */
    INode* getOsdNode();
  
    /**
     * get OSDManager
     */
    IOSDManager* getOsdManager();
  
    /**
     * get WidgetFactory
     */
    IWidgetFactory& getOSDWidgetFactory();
  
    /**
     * get Config-Object
     */
    MMBoxConfig* getMMBoxConfig();

    /**
     * get NMMAplication
     */
    NMMApplication* getNMMApplication();
  
    /**
     * is NMMAplication a ProxyApp
     */
    bool hasProxyApplication() const;
  
    /**
     * get Audio/Video Synchronizer
     */
    IMultiAudioVideoSynchronizer* getSynchronizer();
  
    /**
     * Add font to list.
     * @param id font identifier
     * @param xmfilename filename of the XML file where the font is described
     * @param fontname as a XML font file can describe several fonts, this parameter describes
     *        the desired fontname, if this name is empty, the first font is loaded
     */
    void setFont( string id, string xmlfilename, string fontname );
  
    /** Set timer producer. */
    void setTimerProducer( TimerProducer* tp );
  
    /** Get timer producer. */
    TimerProducer* getTimerProducer();
  
    /** Set event dispatcher. */
    void setDispatcher( MMBoxEventDispatcher* );
  
    /** Get event dispatcher. */
    MMBoxEventDispatcher* getDispatcher();
  
    /** Set NMMApplication */
    void setNMMApplication(  NMMApplication* );
  
    /** Creates Display node, OSDManager node, Playbacknode. */
    Result createGlobalNodes();
  
    /**
     * Creates a default graph ( pngread_node -> converter_node -> osd_node -> display_node )
     * @param _background png file that should be used for background. an empty string indicates
     *                    that the png file should be taken from the config file.
     */
    Result createDefaultGraph( string _background = "" );
  
    /** Starts the default graph. */
    Result startDefaultGraph();
  
    /** Stops the default graph and disconnects all its nodes. */
    Result disconnectDefaultGraph();
  
    /** Stops the video graph and disconnects all its nodes. */
    Result disconnectVideoGraph();
    
    /** Wait for symbol change at sink */
    void waitForSymbolChange(const string& symbol);
    
  protected:     /* Methods from IMMBoxApplicationImpl */
    virtual string getStateSessionLocation(const string &id);
    virtual IMMBoxState* createDistState(const string& id);
    virtual void deleteDistState(const string& id);
    
    Result symbolChanged(const string& in);
  };

}

#endif //MMBOXAPPLICATION_HPP
