/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:       
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <getopt.h>
#include <iostream>

#include "main/application.hpp"
#include "main/mmboxstate.hpp"
#include "XMLParser/XMLParser.hpp"

#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/ProxyApplication.hpp"

#include "nmm/misc/lcd/LCDWriter.hpp"

#ifdef HAVE_LIRCEVENTS
#include "nmm/misc/lirc/LircEvents.hpp"
#endif

#include "nmm/misc/timer/TimerEvents.hpp"

#include "nmm/utils/utils.hpp"
#include "nmm/utils/eventutils/EventRecorder.hpp"
#include "nmm/utils/eventutils/EventPlayer.hpp"
#include "nmm/utils/fork/Fork.hpp"

#include "nmm/base/proxy/tcp/TCPAcceptor.hpp"
#include "nmm/base/registry/ReferenceManager.hpp"

// used interfaces
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/video/display/IDisplayEventSender.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/apps/mmbox/IMMBoxEventReceiverImpl.hpp"
#include "nmm/interfaces/video/display/IVideoDisplay.hpp"

using namespace NMM;

EventRecorder* event_recorder = 0;
EventPlayer*   event_player = 0;
bool           enable_server_registry = false;
int            server_registry_port = ClientRegistry::default_port;



class MMBox {

  typedef list<MMBoxState*> state_list_t;

private:
  bool fullscreen;
  bool use_mga;
  bool uselirc;

  MMBoxApplication*  app;

  state_list_t*      state_list;

  string             configfile;
  string             root_path;

  IDisplayEventSender* display_events;
  InputProducer*        lirc_producer;
  TimerProducer*       timer_producer;

  MMBoxEventDispatcher*   dispatcher;

  LCDWriter*         lcdWriter;
  unsigned int       screenshot_nr;

  Result Quit();
  Result makeScreenShot();
  Result killActualState();
  Result setFullscreen();

  Result test01() {
    timer_producer->registerTimer( time(0)+2, new Event("KEY_Right") );
    return( SUCCESS );
  }

  /*Result test02() {
  //timer_e->registerTimer( time(0)-5, new Event("KEY_Left") );
  timer_producer->registerTimer( 00, 43, 15, 15, 10, 2002, new Event("KEY_Left") );
  return( SUCCESS );
  }*/

public:
  MMBox(int, char*[]);
  ~MMBox();

  /**
   * Starts the main application
   *\param conf configuration file
   */
  void start(string conf);

  MMBoxApplication* getMMBoxApplication() {  return app;  }
  void setTimerProducer(TimerProducer* tp) { timer_producer = tp; }
}; /* class MMBox */


/*********************************************
  Event Functions
*********************************************/

Result MMBox::Quit() {
  GLOBAL_DEBUG_STREAM ("quit" << endl);

  if ( event_recorder )
    event_recorder->receiveEvent( Event ("KEY_EXIT")  );


  delete event_recorder;

  if (event_player) {
    app->setTimerProducer(0);
  }

  delete app;

  SFork::getInstance().killall();

  GLOBAL_DEBUG_STREAM("Exit" << endl);

  exit(0);
}

Result MMBox::killActualState() {
  app->killState(app->getActualStateID());

  return( SUCCESS );
}

Result MMBox::setFullscreen() {
  if ( !use_mga ) {
    IVideoDisplay_var display(app->getDisplayNode()->getParentObject()->getInterface<IVideoDisplay>());
    app->getDisplayNode()->reachActivated();
    if (fullscreen) {
      display->setFullscreen(0);
      fullscreen = false; 
    } else {
      display->setFullscreen(1);
      fullscreen = true;
    }
    app->getDisplayNode()->reachStarted();
    app->getOsdManager()->update();
  }
  return SUCCESS;
}

Result MMBox::makeScreenShot() {
  if ( !use_mga ) {
    IVideoDisplay_var display(app->getDisplayNode()->getParentObject()->getInterface<IVideoDisplay>());
    
    char temp[255];
    sprintf( temp, "Screenshot-%i.png", screenshot_nr );
    
    display->makeScreenshot( temp );
    screenshot_nr++;
  }

  return( SUCCESS );
}

MMBox::MMBox(int argc, char* argv[])  {
  app = new MMBoxApplication();

  NMMApplication* nmmapp = ProxyApplication::getApplication(argc, argv);

  if (enable_server_registry) {
    nmmapp->enableServerRegistry(server_registry_port);
    cerr << "Start a server-registry with port " << server_registry_port << endl;
  }


  app->setNMMApplication(nmmapp);

  dispatcher = new MMBoxEventDispatcher();
  dispatcher->registerEvent( "KEY_w",    new TEDObject0<MMBox>(this, &MMBox::test01) );
  //dispatcher->registerEvent( "KEY_e",    new TEDObject0<MMBox>(this, &MMBox::test02) );

  dispatcher->registerEvent( "KEY_EXIT",    new TEDObject0<MMBox>(this, &MMBox::Quit) );
  dispatcher->registerEvent( "KEY_ESC",      new TEDObject0<MMBox>(this, &MMBox::killActualState) );
  dispatcher->registerEvent( "KEY_Up",     new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Up ) );
  dispatcher->registerEvent( "KEY_Down",   new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Down ) );
  dispatcher->registerEvent( "KEY_VCR_Up",     new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::VCR_Up ) );
  dispatcher->registerEvent( "KEY_VCR_Down",   new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::VCR_Down ) );
  dispatcher->registerEvent( "KEY_Left",   new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Left ) );
  dispatcher->registerEvent( "KEY_Right",  new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Right ) );
  dispatcher->registerEvent( "KEY_Return", new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Return) );
  dispatcher->registerEvent( "KEY_Back",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Back) );
  dispatcher->registerEvent( "KEY_Pause",  new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Pause) );
  dispatcher->registerEvent( "KEY_FF",     new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Forward) );
  dispatcher->registerEvent( "KEY_REW",    new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Rewind) );
  dispatcher->registerEvent( "KEY_Record", new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Record) );
  dispatcher->registerEvent( "KEY_Stop",   new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::Stop) );
  dispatcher->registerEvent( "KEY_a",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::specialKey01) );
  dispatcher->registerEvent( "KEY_s",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::specialKey02) );
  dispatcher->registerEvent( "REVEAL",     new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::REVEAL) );
  dispatcher->registerEvent( "KEY_0",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey0) );
  dispatcher->registerEvent( "KEY_1",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey1) );
  dispatcher->registerEvent( "KEY_2",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey2) );
  dispatcher->registerEvent( "KEY_3",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey3) );
  dispatcher->registerEvent( "KEY_4",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey4) );
  dispatcher->registerEvent( "KEY_5",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey5) );
  dispatcher->registerEvent( "KEY_6",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey6) );
  dispatcher->registerEvent( "KEY_7",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey7) );
  dispatcher->registerEvent( "KEY_8",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey8) );
  dispatcher->registerEvent( "KEY_9",      new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::numKey9) );
  dispatcher->registerEvent( "KEY_F1",     new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::colorKeyRed) );
  dispatcher->registerEvent( "KEY_F2",     new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::colorKeyGreen) );
  dispatcher->registerEvent( "KEY_F3",     new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::colorKeyYellow) );
  dispatcher->registerEvent( "KEY_F4",     new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::colorKeyBlue) );
  dispatcher->registerEvent( "KEY_F5",     new TEDObject0<MMBox>( this, &MMBox::makeScreenShot) );
  dispatcher->registerEvent( "KEY_F6",     new TEDObject0<MMBox>( this, &MMBox::setFullscreen) );
  dispatcher->registerEvent( "KEY_OSD",    new TEDObject0<MMBoxApplication>( app, &MMBoxApplication::OSDKey) );


  state_list = NULL;
  lirc_producer  = NULL;
  timer_producer = NULL;
  screenshot_nr = 0;

  app->setDispatcher( dispatcher );
}

MMBox::~MMBox() {
  if (display_events) {
    display_events->disableDisplayEvents();
    delete display_events;
  }

  if (app->getNMMApplication()) {
    delete app->getNMMApplication();
  }
  delete app;
  delete dispatcher;
  if (lirc_producer != NULL) delete lirc_producer;
  if (lcdWriter != NULL) delete lcdWriter;
  if (timer_producer != NULL ) delete timer_producer;
}

void MMBox::start( string mmboxrcfile_ ) {
  lcdWriter = new LCDWriter();
  lcdWriter->init();

  app->setConfigFile(mmboxrcfile_);

  /* If a timer producer is set use this one, otherwise create it */
  if ( ! timer_producer ) 
    timer_producer = new TimerProducer();

  timer_producer->connectTo( dispatcher );
  app->setTimerProducer( timer_producer );
  timer_producer->start();

  app->init();
  MMBoxConfig* config = app->getMMBoxConfig();
  if (!(config->getValue("fullscreen",fullscreen))) {
    fullscreen = false;
  }
  if (!(config->getValue("uselirc",uselirc))) {
    uselirc = false;
  }
  string resourcepath("");
  config->getValue("resourcepath",resourcepath);
  if(!(config->getValue("configfile",configfile))) {
    configfile = "blue.xml";
  }
  configfile = "mmbox/" + configfile;
  configfile = getNMMResource(configfile);
  cerr << "Using configfile " << configfile << endl;

  if (!(config->getValue("matrox",use_mga))) {
    use_mga = false;
  }
  if (!(config->getValue("rootpath", root_path))) {
    root_path="/";
  } else {
    root_path+="/";
  }

  /**************************************************************************************************************/
  XmlParser     xmlparser;
  if ( xmlparser.open( configfile ) == false ) {
    cerr << "XML configuration file <" << configfile << "> could not be read" << endl;
    exit(1);
  }

  if (config->getValue("resourcepath",resourcepath)) {
    cerr << "Overriding ResourcePath to: " << resourcepath << endl;
    xmlparser.setResourcePath(resourcepath);
  }

  if ( app->createGlobalNodes()  == FAILURE ) exit(1);
  //fw: migration of mmbox
  //if ( app->createDefaultGraph() == FAILURE ) exit(1);
  //if ( app->startDefaultGraph()  == FAILURE ) exit(1);

  // XXX: error checking?
  app->setFont("medium", xmlparser.getResourcePath() + "font/font.xml", "Test");

  IBitmapReader_var bmreader(app->getOsdManager()->getBitmapReader());
  state_list = xmlparser.parse( bmreader.get(), app );
  state_list_t::iterator iter, stop;
  iter = state_list->begin();
  stop = state_list->end();

  while( iter != stop ) {
    MMBoxState* state = *iter;
    iter++;
    //cout << "Registering State: " << state->getID() << endl;
    app->registerState( state );
  }

  app->initStates( "MainMenu" );

  /* Keyboard producer */
  display_events = app->getDisplayNode()->getParentObject()->getInterface<IDisplayEventSender>();
  display_events->getParentObject()->registerEventListener(IDisplayEvents::keyReleased_event, dispatcher);
  if ( display_events ) {
    display_events->enableDisplayEvents( configfile );
  }

  /* IR producer */
  if (uselirc) {
#ifdef HAVE_LIRCEVENTS
    lirc_producer = new LircProducer();
    if ( dynamic_cast<LircProducer *>(lirc_producer)->readConfig( getNMMResource("mmbox/remote.conf") ) == true ) {
      lirc_producer->connectTo( dispatcher );
      //dynamic_cast<LircProducer *>(lirc_producer)->setSendExtendedEvents(true);
      lirc_producer->start();
    } else {
      GLOBAL_ERROR_STREAM("Lirc is enabled in mmboxrc, but could not read remote.conf" << endl);
      delete lirc_producer;
      lirc_producer = NULL;
    }
#else
    GLOBAL_ERROR_STREAM("Lirc is enabled in mmboxrc, but was /nmm/misc/lirc/ was not built" << endl);
#endif
  }

  /* LCD Listener */
  INode* audio_node = app->getAudioNode();
  if (audio_node) {
    audio_node->getParentObject()->registerEventListener(ITrack::startTrack_event, dispatcher);
    dispatcher->registerEventListener(ITrack::startTrack_event, new TEDObject1<LCDWriter, string>(lcdWriter, &LCDWriter::receiveStartTrack));
  }

  /* sleep until quit event */
  while(1)
    sleep(100000);
}

int usage(const char *name) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options]" << endl
       << endl
       << "    -c --config          Set configuration file" << endl
       << "    -s --serverregistry  Starts a server-regegistry that can be access from other applications. " << endl
       << "    -l --registry_port   Set a specific port for the server registry" << endl
       << "    -p --port            Set port for remote access (enables remote access)" << endl
       << "    -v --verbose:        increase verbosity (can be given up to four times -vvvv to inrease verbosity)" << endl
       << "    -r --record:         records the user interaction and stores the into a file " << endl
       << "    -d --demo:           plays a recorded user interation " << endl
       << "    -h --help:           Shows this help and exits." << endl;
  return 1;
}

class  MMBoxEventReceiver : public IMMBoxEventReceiverImpl {
public:
  MMBoxEventReceiver(MMBoxEventDispatcher* dispatcher) : __dispatcher(dispatcher) {
  }
  virtual ~MMBoxEventReceiver() {};

protected:
  Result receiveEventID(const string& id) 
  {
    cerr << "DIS " << __dispatcher << " " << id << endl;
    return __dispatcher->receiveEvent(Event(id));
  }

private:
  MMBoxEventDispatcher* __dispatcher;
};

int main(int argc, char *argv[]) {
  /* global options */
  const char* home(getenv("HOME"));
  string mmboxrcfile(string(home) + string("/.mmboxrc"));
  string record_file;
  string demo_file;
  unsigned int verbose(0);
  int port(0);
  
  const char *optstring = "r:d:c:vsl:h";
  struct option longopts[] = {
    {"config",         1, NULL, 'c'},
    {"verbose",        0, NULL, 'v'},
    {"help",           0, NULL, 'h'},
    {"record",         1, NULL, 'r'},
    {"demo",           1, NULL, 'd'},
    {"serverregistry", 0, NULL, 's'},
    {"registry_port ", 1, NULL, 's'},
    {NULL,             0, NULL, 0  } /* last row must be zero */
  };

  int ch, optionindex;

  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'c':
      mmboxrcfile = string(optarg);
      break;
    case 'v':
      verbose++;
      break;
    case 'r':
      record_file = string(optarg);
      event_recorder = new EventRecorder( record_file);
      break;
    case 'd':
      demo_file = string(optarg);
      //use a delay of 10 seconds. After this delay the XDisplay is available and can receive events
      event_player = new EventPlayer(demo_file, 10); 
      break;
    case 'l':
      server_registry_port = atoi(optarg);
      break;
    case 's':
      enable_server_registry = true;
      break;
    default: //includes 'h'
      return usage(argv[0]);
      break;
    }
  }

  /* handle *Stream output */
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }

  SFork::getInstance();

  if (event_player) {

    MMBox mmbox(argc, argv);

    /* set the TimerProducer from the event_player */
    mmbox.setTimerProducer(&event_player->getTimerProducer());

    mmbox.start(mmboxrcfile);

    /* reset the TimerProducer, because it is managed by the event_player */
    mmbox.getMMBoxApplication()->setTimerProducer(0);
  } else {
      MMBox mmbox(argc, argv);

      /* register the event_recorder */
      if ( event_recorder )
        mmbox.getMMBoxApplication()->getDispatcher()->registerEventListener(event_recorder);

      mmbox.start(mmboxrcfile);
    
  }

  delete event_recorder;
  delete event_player;
  return 0;
}


