/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "qcamctrlview.h"
#include "qgraphlist.h"
#include "qgraphwidget.h"
#include <qmessagebox.h>
#include <qvalidator.h>
#include <qlineedit.h>
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

QCamCtrlView::QCamCtrlView(QWidget *parent, NMM::NMMApplication& app)
  : QWidget(parent), zoom_label(0), focus_label(0), /*view_style(0),*/ toplevel(0), camera(0),
    zoom(0), focus(0), technical_layout(0), technical_box(0), paint_box(0), one_push(0),
    start_(0), stop_(0), __graph_list(0), __display(0), __current_sender(0), __application(app),
    __is_display_active(false)
{
  __camera_format.setFormatType("video/raw");
  __camera_format.addIntParamValue(NMM::Format::x_resolution_param,320);
  __camera_format.addIntParamValue(NMM::Format::y_resolution_param,240);
  __camera_format.addWildcard(NMM::Format::ratio_x_param);
  __camera_format.addWildcard(NMM::Format::ratio_y_param);
  __camera_format.addStringParamValue(NMM::Format::channelorder_param,"uyvy");
  __camera_format.addStringParamValue(NMM::Format::endian_param, "little");
  __camera_format.addStringParamValue(NMM::Format::colorspace_param,"yuv422");
  __camera_format.addIntParamValue(NMM::Format::bitperpixel_param,16);
  __camera_format.addStringParamValue(NMM::Format::format_param, "packet");
  __camera_format.addFloatParamValue(NMM::Format::framerate_param, 30.0);
                 

  this->camera = camera ;

  toplevel = new QHBoxLayout(this,4);

  initPaint();
  initTechnical();
  initSlider();
  initRegistry();
  initConnect();
  buttonConnect();
  toplevel->activate();

  if (camera == NULL)  {
    setCamera(0);
    cout << "no cam connected" << endl;
  } else   {
    update();
  }
  scanCameras("localhost", NMM::ClientRegistry::default_port);
  scanDisplay();
}

QCamCtrlView::~QCamCtrlView()
{
  cerr << ">Delete" << endl;
  if (__current_sender) {
    __current_sender->stop();
    __display->stop();
    __current_sender->deactivate();
    __display->deactivate();

    __display->deinitOutput();
    disconnect_output(__current_sender);
    __current_sender->deinitOutput();

    __current_sender->deinit();
    __display->deinit();
  }
  for (list<NMM::INode*>::iterator it = __reserved_node_list.begin(); it != __reserved_node_list.end(); ++it)
  {
    __application.getRegistry().releaseNode(**it);
    delete *it;
  }
  __reserved_node_list.clear();

  __application.getRegistry().releaseNode(*__display);
  delete __display;
  cerr << "<Delete" << endl;
}



void QCamCtrlView::setCamera(NMM::ICamera * new_camera)
{
  if (!new_camera) {
    paint_box->setDisabled(true);
    technical_box->setDisabled(true);
    zoom->setDisabled(true);
    focus->setDisabled(true);

    for (unsigned int i = 0; i< paint_button.size(); ++i)
      paint_button[i]->setDisabled(true);

    for (unsigned int i = 0; i< technical_button.size(); ++i)
      technical_button[i]->setDisabled(true);
  }

  if(!new_camera || (new_camera == camera)) {
    return;
  }

  if (new_camera) {
    paint_box->setDisabled(false);
    technical_box->setDisabled(false);
    camera = new_camera;
    update();
  }
}

void QCamCtrlView::initRegistry() {
  QGridLayout *registry_layout = new QGridLayout( 6, 5, 15);
  toplevel->addLayout(registry_layout);

  QLabel* host_label = new QLabel("Hostname", this);
  QLabel* port_label = new QLabel("Port", this);
  port_label->setAlignment(AlignCenter);
  host_label->setAlignment(AlignCenter);
//  port_label->setFixedSize(40,20);
//  host_label->setFixedSize(50,20);


  __host_name = new QLineEdit(this);
  __port_line = new QLineEdit(this);
  __port_line->setMinimumWidth(60);
  __int_validator = new QIntValidator(0, 65500, __port_line );
  __port_line->setValidator(__int_validator);
  __add_button = new QPushButton("Add", this);  


  std::list<QString> tmp_list;
  tmp_list.push_back("Registry-Host");
  tmp_list.push_back("Port");

  __graph_list = new QGraphList(this, 0, tmp_list);
  registry_layout->addMultiCellWidget(host_label, 0, 0, 0, 2);
  registry_layout->addWidget(port_label, 0, 3);
  registry_layout->addMultiCellWidget(__host_name,  1, 1, 0, 2);
  registry_layout->addMultiCellWidget(__port_line, 1, 1, 3, 3);
  registry_layout->addWidget(__add_button, 1, 4);
  registry_layout->addMultiCellWidget(__graph_list, 2, 5, 0, 4);
}

void QCamCtrlView::initPaint()
{
  QGridLayout *paint_layout = new QGridLayout( 7, 6, 15);
  toplevel->addLayout(paint_layout);

  paint_box = new QGroupBox(this);
  paint_box->setTitle("Painting");

  paint_layout->addMultiCellWidget(paint_box, 0,6,0,5);

  paint_button.resize(4);

  for (int i = 0; i<4; i++)
    paint_button[i] = new MyCamKnob(this);

  one_push = new QPushButton(this);
  one_push->setFixedSize(10,10);

  paint_button[0]->setLabel("Hue");
  paint_layout->addWidget(paint_button[0],1,1);

  paint_button[1]->setLabel("Saturation");
  paint_layout->addWidget(paint_button[1],1,3);

  QLabel *wb = new QLabel("White Balance", this);

  wb->setAlignment(AlignCenter);
  paint_layout->addMultiCellWidget(wb,2,2,1,3);

  paint_button[2]->setLabel("Red");
  paint_layout->addWidget(paint_button[2],4,1);

  paint_layout->addWidget(one_push,3,2);
  paint_layout->addWidget(new QLabel("Auto",this) , 3,3);

  paint_button[3]->setLabel("Blue");
  paint_layout->addWidget(paint_button[3],4,3);
}



void QCamCtrlView::initTechnical()
{
  technical_layout = new QGridLayout(5,5,15);
  toplevel->addLayout(technical_layout);

  technical_box = new QGroupBox(this);
  technical_box->setTitle("Technical");

  technical_button.resize(6);

  for (int i = 0; i<6; i++)
    technical_button[i] = new MyCamKnob(this);

  technical_button[0]->setLabel("Brightness");
  technical_layout->addWidget(technical_button[0],1,1);

  technical_button[1]->setLabel("Sharpness");
  technical_layout->addWidget(technical_button[1],1,2);

  technical_button[2]->setLabel("AutoExposure");
  technical_layout->addWidget(technical_button[2],1,3);

  technical_button[3]->setLabel("Iris");
  technical_layout->addWidget(technical_button[3],2,1);

  technical_button[4]->setLabel("Shutter");
  technical_layout->addWidget(technical_button[4],2,2);

  technical_button[5]->setLabel("Gain");
  technical_layout->addWidget(technical_button[5],2,3);

  technical_layout->addMultiCellWidget(technical_box, 0,4,0,4);
}

/*void QCamCtrlView::toggleAdvancedControl()
{
  if(technical_button[0]->isVisible())
    for(int i = 0; i<5 ;i++)
      technical_button[i]->hide();

  else
    for(int i = 0; i<5; i++)
      technical_button[i]->show();

}*/
void QCamCtrlView::initSlider()
{
  QGridLayout *slider_layout = new QGridLayout(3,2);
  toplevel->addLayout(slider_layout);
  slider_layout->setColStretch(0,2);

  zoom  = new QSlider(this);
  //  zoom->setStyle(view_style);

  focus = new QSlider(this);
  //  focus->setStyle(view_style);
  zoom->setMaximumWidth(40);
  focus->setMaximumWidth(40);


  zoom_label  = new QLabel(this);
  focus_label = new QLabel(this);
  zoom_label->setText("Zoom");
  zoom_label->setAlignment(AlignCenter);
  focus_label->setText("Focus");
  focus_label->setAlignment(AlignCenter);

//  zoom_label->setFixedSize(40,20);
//  focus_label->setFixedSize(40,20);

  slider_layout->addWidget(zoom_label,0,0);
  slider_layout->addWidget(zoom,1,0);
  slider_layout->addWidget(focus_label,0,1);
  slider_layout->addWidget(focus,1,1);

//  start_ = new QPushButton("start", this);
//  stop_  = new QPushButton("stop", this);
//  slider_layout->addWidget(start_, 2,0);
//  slider_layout->addWidget(stop_, 2,1);

}

void QCamCtrlView::update()
{
  pair<int, int> range;
  if(camera->hasHue())
  {
    range = camera->getHueRange();

    paint_button[0]->setEnabled(true);
    paint_button[0]->setRange(range.first, range.second, 1);
    paint_button[0]->setValue(camera->getHue());
  }

  else
    paint_button[0]->setEnabled(false);

  if(camera->hasSaturation())
  {
    range = camera->getSaturationRange();

    paint_button[1]->setEnabled(true);
    paint_button[1]->setRange(range.first, range.second, 1);
    paint_button[1]->setValue(camera->getSaturation());
  }

  else
    paint_button[1]->setEnabled(false);

  if(camera->hasWhiteBalance())
  {
    range = camera->getWhiteBalanceRange();

    paint_button[2]->setEnabled(true);
    paint_button[2]->setRange(range.first, range.second,1);
    paint_button[2]->setValue(camera->getWhiteBalance());
  }

  else
    paint_button[2]->setEnabled(false);

  if(camera->hasWhiteBalance())
  {
    range = camera->getWhiteBalanceRange();

    paint_button[3]->setEnabled(true);
    paint_button[3]->setRange(range.first, range.second, 1);
    paint_button[3]->setValue(camera->getWhiteBalance());
  }
    else
    paint_button[3]->setEnabled(false);

  //technical part

  if(camera->hasBrightness())
  {
    range = camera->getBrightnessRange();

    technical_button[0]->setEnabled(true);
    technical_button[0]->setRange(range.first, range.second,1);
    technical_button[0]->setValue(camera->getBrightness());
  }
  else
    technical_button[0]->setEnabled(false);

  if(camera->hasSharpness())
  {
    range = camera->getSharpnessRange();

    technical_button[1]->setEnabled(true);
    technical_button[1]->setRange(range.first ,range.second,1);
    technical_button[1]->setValue(camera->getSharpness());
  }
  else
    technical_button[1]->setEnabled(false);

  if(camera->hasAutoExposure())
  {
    range = camera->getAutoExposureRange();

    technical_button[2]->setEnabled(true);
    technical_button[2]->setRange(range.first, range.second,1);
    technical_button[2]->setValue(camera->getAutoExposure());
  }

  else
    technical_button[2]->setEnabled(false);


  if(camera->hasIris())
  {
    range = camera->getIrisRange();

    technical_button[3]->setEnabled(true);
    technical_button[3]->setRange(range.first, range.second, 1);
    technical_button[3]->setValue(camera->getIris());
  }

  else
    technical_button[3]->setEnabled(false);

  if(camera->hasShutter())
  {
    range = camera->getShutterRange();

    technical_button[4]->setEnabled(true);
    technical_button[4]->setRange(range.first, range.second,1);
    technical_button[4]->setValue(camera->getShutter());
  }
  else
    technical_button[4]->setEnabled(false);

  if(camera->hasGain())
  {
    range = camera->getGainRange();
    technical_button[5]->setEnabled(true);
    technical_button[5]->setRange(range.first, range.second,1);
    technical_button[5]->setValue(camera->getGain());
  }

  else
    technical_button[5]->setEnabled(false);

  if(camera->hasZoom())
  {
    range = camera->getZoomRange();

    zoom->setEnabled(true);
    zoom_label->setEnabled(true);
    zoom->setRange(range.first,range.second);
    zoom->setValue(camera->getZoom());
  }

  else
  {
    zoom->setEnabled(false);
    zoom_label->setEnabled(false);
  }

  if(camera->hasFocus())
  {
    range = camera->getFocusRange();

    focus->setEnabled(true);
    focus_label->setEnabled(true);
    focus->setRange(range.first, range.second);
    focus->setValue(camera->getFocus());
  }

  else
  {
    focus->setEnabled(false);
    focus_label->setEnabled(false);
  }



}

void QCamCtrlView::buttonConnect()
{
  //connect technical button
  connect(technical_button[0], SIGNAL(valueChanged(int)), this, SLOT(changeBrightness(int)));
  connect(technical_button[1], SIGNAL(valueChanged(int)), this, SLOT(changeSharpness(int)));
  connect(technical_button[2], SIGNAL(valueChanged(int)), this, SLOT(changeAutoExposure(int)));
  connect(technical_button[3], SIGNAL(valueChanged(int)), this, SLOT(changeIris(int)));
  connect(technical_button[4], SIGNAL(valueChanged(int)), this, SLOT(changeShutter(int)));
  connect(technical_button[5], SIGNAL(valueChanged(int)), this, SLOT(changeGain(int)));
  connect(one_push, SIGNAL(clicked()), this, SLOT(onePush()));
  //connect paint_button
  connect(paint_button[0], SIGNAL(valueChanged(int)), this, SLOT(changeHue(int)));
  connect(paint_button[1], SIGNAL(valueChanged(int)), this, SLOT(changeSaturation(int)));
  connect(paint_button[2], SIGNAL(valueChanged(int)), this, SLOT(changeRedBalance(int)));
  connect(paint_button[3], SIGNAL(valueChanged(int)), this, SLOT(changeBlueBalance(int)));

  //connect slider
  connect(zoom,  SIGNAL(valueChanged(int)), this, SLOT(changeZoom(int)));
  connect(focus, SIGNAL(valueChanged(int)), this, SLOT(changeFocus(int)));
//  connect(start_, SIGNAL(clicked()), this, SLOT(start()));
//  connect(stop_,  SIGNAL(clicked()), this, SLOT(stop()));

}

// Slots

void QCamCtrlView::changeBrightness(int i)
{
   if (i!= camera->getBrightness()) {
     camera->setBrightness(i);
     technical_button[0]->setValue(camera->getBrightness());
   }
}
void QCamCtrlView::changeAutoExposure(int i)
{
  if (i!= camera->getAutoExposure()) {
    camera->setAutoExposure(i);
    technical_button[2]->setValue(camera->getAutoExposure());
  }
}

void QCamCtrlView::changeHue(int i)
{
  if (i!= camera->getHue()) {
    camera->setHue(i);
    paint_button[0]->setValue(camera->getHue());
  }
}

void QCamCtrlView::changeSharpness(int i)
{
  if (i!= camera->getSharpness()) {
    camera->setSharpness(i);
    technical_button[1]->setValue(camera->getSharpness());
  }
}

void QCamCtrlView::changeRedBalance(int i)
{
  if (i!= camera->getRedBalance()) {
    camera->setRedBalance(i);
    paint_button[2]->setValue(i);
  }
}

void QCamCtrlView::changeBlueBalance(int i)
{
  if (i!=camera->getBlueBalance()) {
    camera->setBlueBalance(i);
    paint_button[3]->setValue(i);
  }
}


void QCamCtrlView::changeSaturation (int i)
{
  if (i != camera->getSaturation())
    camera->setSaturation(i);

  paint_button[1]->setValue(camera->getSaturation());
}

void QCamCtrlView::changeGamma (int i)
{
  if (i != camera->getGamma()) {
    camera->setGamma(i);
    technical_button[3]->setValue(camera->getGamma());
  }
}

void QCamCtrlView::changeShutter (int i)
{
  if (i != camera->getShutter()) {
    camera->setShutter(i);
    technical_button[4]->setValue(camera->getShutter());
  }
}

void QCamCtrlView::changeGain(int i)
{
  if (i != camera->getGain()) {
    camera->setGain(i);
    technical_button[5]->setValue(camera->getGain());
  }
}


void QCamCtrlView::changeFocus(int i)
{
  if (i != camera->getFocus()) {
    camera->setFocus(i);
    focus->setValue(camera->getFocus());
  }
}


void QCamCtrlView::changeZoom(int i)
{
  if ( i != camera->getZoom() ) {
    camera->setZoom(i);
    zoom->setValue(camera->getZoom());
  }
}

void QCamCtrlView::changeIris(int i)
{
  if (i != camera->getIris()) {
    camera->setIris(i);
    technical_button[3]->setValue(camera->getIris());
  }
}

void QCamCtrlView::onePush()
{
  camera->onePush();
  /* Wait up to five seconds. */
  struct timeval tv;
  tv.tv_sec = 2;
  tv.tv_usec = 0;
  select(0, NULL, NULL, NULL, &tv);
  paint_button[2]->setValue(camera->getRedBalance());
  paint_button[3]->setValue(camera->getBlueBalance());
}

void QCamCtrlView::start() {
  cout << "Stop pressed" << endl;
//  for (list<QEventHandler*>::iterator it1 = event_handler.begin(); it1 != event_handler.end(); it1++)
//    (*it1)->start();
}
void QCamCtrlView::stop() {
  cout << "Start pressed" << endl;
//  for (list<QEventHandler*>::iterator it1 = event_handler.begin(); it1 != event_handler.end(); it1++)
//    (*it1)->stop();
}


void QCamCtrlView::scanCameras(const string& host, int port) {
  NMM::NodeDescription nd;
  nd.addInterface(NMM::ICamera::ICamera_type);
    try {
      nd.setLocation(host);
      list<NMM::Response>  tmp_list = __application.getRegistry().initRequest(nd);

      list<NMM::INode*> node_list = requestAllNodes(tmp_list);

      if (!node_list.empty())
        __reserved_node_list.insert(__reserved_node_list.end(), node_list.begin(), node_list.end());

      /* Update the requested camreas */
      std::list<QString> header_list;
      header_list.push_back(host.c_str());
      header_list.push_back(QString().setNum(port));
      QGraphWidget* widget = new QGraphWidget(header_list);
      for (list<NMM::INode*>::iterator it = node_list.begin(); it != node_list.end(); ++it) {
        widget->addNode(*it);
      }
      __graph_list->addNodeGraph(widget);
      /**/
    } catch (const NMM::Exception& e) {
      QMessageBox::critical(this, "Can not request Camera", e.getComment().c_str());
    }
}

list<NMM::INode*> QCamCtrlView::requestAllNodes(const list<NMM::Response>& resp) {
  list<NMM::INode*> result_list;
  for (list<NMM::Response>::const_iterator it = resp.begin(); it != resp.end(); ++it) {
    for (list<NMM::ObjectID>::const_iterator int_it = it->second.begin(); int_it != it->second.end(); ++int_it) {
      try {
        NMM::INode* tmp_node = __application.getRegistry().requestNode(it->first, *int_it);
        tmp_node->init();
        result_list.push_back(tmp_node);
      } catch (const NMM::Exception& e) {
        QMessageBox::critical(this, "Can not request Node", e.getComment().c_str());
      }
    }
  }
  return result_list;
}

void QCamCtrlView::scanDisplay() {
  try {
    NMM::NodeDescription nd;
    nd.setNodeName("XDisplayNode");
    nd.setNodeType(NMM::INode::SINK);

    list<NMM::Response>  tmp_list = __application.getRegistry().initRequest(nd);

    if (!tmp_list.empty()) {
      __display = __application.getRegistry().requestNode(tmp_list);
      __display->init();
    } else {
      cerr << "NO Display" << endl;
    }
  } catch (const NMM::Exception& e) {
    QMessageBox::critical(this, "Can not request Display", e.getComment().c_str());
  }
}

void QCamCtrlView::initConnect() {
  connect(__graph_list, SIGNAL(clicked(NMM::INode*)), this, SLOT(changeCamera(NMM::INode* )));
  connect(__add_button, SIGNAL(clicked()), this, SLOT(addRegistry()));
}


void QCamCtrlView::changeCamera(NMM::INode* new_camera) {
  NMM::MutexGuard guard(__change_camera_mutex);

  if (! new_camera )
    return;

  if ( __current_sender == new_camera ) {
    cerr << "The same camera" << endl;
    return;
  }

  /* Begin> If a camera is running, stop it and disconnect it from our window */
  if (__current_sender) {
    __current_sender->stop();
    __display->stop();
    __current_sender->flush();
    __display->flush();

    __current_sender->deactivate();
    __display->deactivate();

    __display->deinitOutput();
    disconnect_output(__current_sender);
    __current_sender->deinitOutput();
    __current_sender = 0;
  }
  /* End< */

  __current_sender = new_camera;
  /*Connect the current_sender with a display*/
  try {
    __current_sender->initOutput();
    NMM::connect(__current_sender, __display, &__camera_format);
    __display->initOutput();
  } catch (NMM::Exception& e) {
    QMessageBox::critical(this, "Connection Error", e.getComment().c_str());        
  }
  /*End< */

  __current_sender->activate();
  __display->activate();
  __display->start();
  __current_sender->start();
  __current_camera.reset(__current_sender->getParentObject()->getInterface<NMM::ICamera>());
  setCamera(__current_camera.get());
}


void QCamCtrlView::addRegistry() {
  try {
    if (__host_name->text().isEmpty() || __port_line->text().isEmpty())
      return;
    string host = __host_name->text().latin1();
    int port    = __port_line->text().toInt();
    __application.getRegistry().addRegistry(NMM::TCPAddress(host, port));
    scanCameras(host, port);       
  } catch (const NMM::Exception& e) {
    QMessageBox::critical(this, "Invalid address ", e.getComment().c_str());
  }
}

