/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Eric Peters <epeters@graphics.cs.uni-sb.de>
 *                    Christian Gerstner <gerstner@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
 

#include <qfiledialog.h>
#include <qcolordialog.h>
#include <qmessagebox.h>
#include "sigreader.hpp"
#include <unistd.h>
#include <iostream>



// general includes
#include "nmm/utils/utils.hpp"
#include "nmm/utils/xml/xml++.hpp"
#include "nmm/base/connect.hpp"
extern "C" {
#include "nmm/utils/passwdtools/passwdtools.h"
}

// include the plugin manager
#include "nmm/base/sync/MasterSlaveSourceSynchronizer.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/IThreadedNode.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSource.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/file/ISendFilenameAndTimestamp.hpp"
#include "nmm/interfaces/file/IFileModeHandler.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/video/IFrameCount.hpp"
#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/interfaces/video/display/IXDisplayEventSender.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/net/INetClient.hpp"
#include "nmm/interfaces/video/display/IXWindowID.hpp"

class SigReader::VcoreEventHandler {
private:
  string pressed_key;
  string released_key;
  IVISCACamera* vcam;
  IVNC* ivnc;
  ISVGEncode* isvgencode;
  ISVGPaint* isvgpaint;
  IMouseCursor* imousecursor;
  int write_fd;
  int vnc_write_fd;
  map<string, Timestamp> timestamps;
  enum eventtype{KEYPRESSED, KEYRELEASED, BUTTONPRESSED, BUTTONRELEASED, MOUSEMOVED, RFBUPDATEREQUEST, COLOR, NONE};
  bool annotationMode;
  ISVGEncode::SVGDrawMode svgDrawType;
  int svgDrawColorRed;
  int svgDrawColorGreen;
  int svgDrawColorBlue;

  struct vncevents{
    eventtype type;
    unsigned long keyReleased;
    unsigned long keyPressed;
    int button, x, y, buttonmask,color;
    bool shift;
  }vncevent;

public:
  VcoreEventHandler() :
    pressed_key(""),
    released_key(""),
    vcam(NULL),
    ivnc(NULL),
    isvgencode(NULL),
    isvgpaint(NULL),
    imousecursor(NULL),
    write_fd(-1),
    vnc_write_fd(-1),
    timestamps(),
    annotationMode(true),
    svgDrawType(ISVGEncode::pointer),
    svgDrawColorRed(0),
    svgDrawColorGreen(0),
    svgDrawColorBlue(255)
  {
    vncevent.type=NONE;
  }

  void setIVISCACamera(IVISCACamera* vcam_) {
    vcam = vcam_;
  }

  void setIVNC(IVNC* ivnc_) {
    ivnc = ivnc_;
  }

  void setISVGEncode(ISVGEncode* isvgencode_) {
    isvgencode = isvgencode_;
  }

  void setISVGPaint(ISVGPaint* isvgpaint_) {
    isvgpaint = isvgpaint_;
  }

  void setIMouseCursor(IMouseCursor* imousecursor_) {
    imousecursor = imousecursor_;
  }

  void setWriteFD(int write_fd_) {
    if (write_fd != write_fd_) {
      write_fd = write_fd_;
    }
  }

  void setVNCWriteFD(int write_fd_) {
    if (vnc_write_fd != write_fd_) {
      vnc_write_fd = write_fd_;
    }
  }

  map<string, Timestamp> getTimestamps() {
    return timestamps;
  }

  Result keyPressed(string& key) {
    pressed_key = key;
    char c = 'x';
    write(write_fd, &c, 1);
    return SUCCESS;
  }
  
  Result keyReleased(string& key) {
    released_key = key;
    char c = 'x';
    write(write_fd, &c, 1);
    return SUCCESS;
  }

  Result sendFilenameAndTimestamp(string& fn, Timestamp& ts) {
    timestamps[fn] = ts;
    return SUCCESS;
  }

  void clearTimestamps() {
    timestamps.clear();
  }

  void stopCamera() {
    if (!vcam) {
      cerr << "VcoreEventHandler: No IVISCACamera interface set!" << endl;
      return;
    }
    vcam->ZoomStop(1);
    vcam->PanTiltDriveStop(1 /*camera_number*/, 0x18, 0x14);
  }

  void handleEvents() {
    if (pressed_key != string("")) {
      if (!vcam) {
	cerr << "VcoreEventHandler: No IVISCACamera interface set!" << endl;
	return;
      } 

      if (pressed_key == "KEY_Left") { // 'KEY_Left' to turn camera left
	vcam->PanTiltDriveLeft(1 /*camera_number*/,
			       0x04 /*camera_state[camera_number].motion_speed*/,
			       0x01);
      } else if (pressed_key == "KEY_Right") { // 'KEY_Right' to turn camera left
	vcam->PanTiltDriveRight(1 /*camera_number*/,
				0x04 /*camera_state[camera_number].motion_speed*/,
				0x01);
      } else if (pressed_key == "KEY_Up") { // 'KEY_Up' to turn camera up
	vcam->PanTiltDriveUp(1 /*camera_number*/,
			     0x01,
			     0x04 /*camera_state[camera_number].motion_speed*/);
      } else if (pressed_key == "KEY_Down") { // 'KEY_Down' to turn camera down
	vcam->PanTiltDriveDown(1 /*camera_number*/,
			       0x01,
			       0x04 /*camera_state[camera_number].motion_speed*/);
      } else if (pressed_key == "KEY_Pageup") { // 'KEY_Pageup' to zoom in
	vcam->ZoomTeleVar(1, 0x03);
      } else if (pressed_key == "KEY_Pagedown") { // 'KEY_Pagedown' to zoom out
	vcam->ZoomWideVar(1, 0x03);
      } else if (pressed_key == "KEY_space") {
	stopCamera();
      }
    }
    pressed_key = string("");

    if (released_key != string("")) {
      stopCamera();
      released_key = string("");
    }
  }

  Result sendKeyReleasedEvent(unsigned long& keysym){
    GLOBAL_DEBUG_STREAM("keysym = " << keysym << " State = Released"  <<endl);
    vncevent.type=KEYRELEASED;
    vncevent.keyReleased=keysym;
    char c = 'x';
    write(vnc_write_fd, &c, 1);
    return SUCCESS;
  }

  Result sendKeyPressedEvent(unsigned long& keysym){
    GLOBAL_DEBUG_STREAM("keysym = " << keysym << " State = Pressed" <<endl);
    vncevent.type=KEYPRESSED;
    vncevent.keyPressed=keysym;
    char c = 'x';
    write(vnc_write_fd, &c, 1);
    return SUCCESS;
  }

  Result sendButtonPressedEvent(int& button, int& x, int& y, int& buttonmask, bool& shift)
  {
    GLOBAL_DEBUG_STREAM("(x,y)=("<< x << "," << y << ") buttonmask=" << (((buttonmask & 0x1f00) >> 8) |
									 (1 << (button - 1))) << " shift=" << shift << endl);
    vncevent.type=BUTTONPRESSED;
    vncevent.button=button;
    vncevent.x=x;
    vncevent.y=y;
    vncevent.buttonmask=buttonmask;
    vncevent.shift=shift;
    char c = 'x';
    write(vnc_write_fd, &c, 1);
    return SUCCESS;
  }

  Result sendButtonReleasedEvent(int& button, int& x, int& y, int& buttonmask, bool& shift)
  {
    GLOBAL_DEBUG_STREAM("(x,y)=("<< x << "," << y << ") buttonmask=" << (((buttonmask & 0x1f00) >> 8) &
									 ~(1 << (button - 1))) << endl);
    vncevent.type=BUTTONRELEASED;
    vncevent.button=button;
    vncevent.x=x;
    vncevent.y=y;
    vncevent.buttonmask=buttonmask;
    vncevent.shift=shift;
    char c = 'x';
    write(vnc_write_fd, &c, 1);
    return SUCCESS;
  }

  Result sendMouseMotionEvent(int& x, int& y, int& buttons, bool& shift)
  {
    GLOBAL_DEBUG_STREAM("(x,y)=("<< x << "," << y << ") buttonmask=" << ((buttons & 0x1f00) >> 8) << endl);
    vncevent.type=MOUSEMOVED;
    vncevent.x=x;
    vncevent.y=y;
    vncevent.buttonmask=buttons;
    vncevent.shift=shift;
    char c = 'x';
    write(vnc_write_fd, &c, 1);
    return SUCCESS;
  }

  Result sendRFBUpdateRequest() {
    GLOBAL_DEBUG_STREAM("RfbUpdate requested..." << endl);
    vncevent.type=RFBUPDATEREQUEST;
    char c = 'x';
    write(vnc_write_fd, &c, 1);
    return SUCCESS;
  }

  Result setSVGDrawType(ISVGEncode::SVGDrawMode svgdt){
    svgDrawType = svgdt;
    return SUCCESS;
  };

  Result setSVGDrawColor(int red, int green, int blue){
    svgDrawColorRed = red;
    svgDrawColorGreen = green;
    svgDrawColorBlue = blue;
    isvgencode->setColor(red, green, blue);
    return SUCCESS;
  };

  Result setAnnotationMode(bool _am) {
    if(_am!=annotationMode){
      if(annotationMode == false){
        annotationMode = true;
        isvgpaint->setShowSVGAnnotation(true);
	 
        return SUCCESS;
      }else{
	annotationMode = false;
	isvgpaint->setShowSVGAnnotation(false);
	
	return SUCCESS;
      }
    }
    return SUCCESS;
  };

  Result setVNCWorkingMode(IVNC::Mode vncmode) {
    ivnc->setMode(vncmode);
    return SUCCESS;
  };

  Result newAnnotation(){
    isvgencode->newAnnotation();
    return SUCCESS;
  }

  Result removeAnnotation(){
    isvgencode->clearAnnotation();
    return SUCCESS;
  }

  Result undoAnnotation(){
    isvgencode->undo();
    return SUCCESS;
  }

  Result redoAnnotation(){
    isvgencode->redo();
    return SUCCESS;
  }

  void handleVNCEvents() {
    if (!isvgencode || !ivnc) {
      cerr << "VcoreEventHandler: No ISVG and/or IVNC interface set!" << endl;
      return;
    }
    if(annotationMode == false){ //VNC-Mode - send Events to VNCSourceNode
      switch(vncevent.type){

      case KEYPRESSED:
	ivnc->sendKeyEvent(vncevent.keyPressed,1);
	break;

      case KEYRELEASED:
	ivnc->sendKeyEvent(vncevent.keyReleased,0);
	break;

      case BUTTONPRESSED:
	ivnc->sendPointerEvent(vncevent.x,vncevent.y,(((vncevent.buttonmask & 0x1f00) >> 8) |
						      (1 << (vncevent.button - 1))));
	break;

      case BUTTONRELEASED:
	ivnc->sendPointerEvent(vncevent.x,vncevent.y,(((vncevent.buttonmask & 0x1f00) >> 8) &
						      ~(1 << (vncevent.button - 1))));
	break;

      case MOUSEMOVED:
	ivnc->sendPointerEvent(vncevent.x,vncevent.y,((vncevent.buttonmask & 0x1f00) >> 8));
	break;

      case RFBUPDATEREQUEST:
	ivnc->sendFrameBufferUpdateRequest();
	break;

      default:
	break;
      }

    }else{ //SVG-Mode - send (all necessery) Events to SVGEncodeNode

      switch(vncevent.type){

      case BUTTONPRESSED:
	if((((vncevent.buttonmask & 0x1f00) >> 8) | (1 << (vncevent.button - 1)))!=2){
	  switch (svgDrawType) {
	  case ISVGEncode::line:
	    isvgencode->drawSVGLine(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) | (1 << (vncevent.button - 1))));
	    break;
	  case ISVGEncode::rectangle:
	    isvgencode->drawSVGRectangle(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) | (1 << (vncevent.button - 1))));
	    break;
	  case ISVGEncode::freehand:
	    isvgencode->drawSVGFreehand(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) | (1 << (vncevent.button - 1))));
	    break;
	  case ISVGEncode::pointer:
	    imousecursor->useTriangleCursor(true);
	    isvgencode->drawSVGPointer(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) | (1 << (vncevent.button - 1))));
	    break;
	  default:
	    GLOBAL_DEBUG_STREAM("Unknown draw type" << endl);
	    break;
	  }
	}else{  //if middle mousebutton ist pressed
	  imousecursor->useTriangleCursor(true);
	  isvgencode->drawSVGPointer(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) | (1 << (vncevent.button - 1))));
	} 
	break;

      case BUTTONRELEASED:
	if(((vncevent.buttonmask & 0x1f00) >> 8 & ~(1 << (vncevent.button - 1)))!=2){
	  switch (svgDrawType) {
	  case ISVGEncode::line:
	    isvgencode->drawSVGLine(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) & ~(1 << (vncevent.button - 1))));
	    break;
	  case ISVGEncode::rectangle:
	    isvgencode->drawSVGRectangle(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) & ~(1 << (vncevent.button - 1))));
	    break;
	  case ISVGEncode::freehand:
	    isvgencode->drawSVGFreehand(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) & ~(1 << (vncevent.button - 1))));
	    break;
          case ISVGEncode::pointer:
	    imousecursor->useTriangleCursor(false);
	    isvgencode->drawSVGPointer(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) & ~(1 << (vncevent.button - 1))));
	    break;
	  default:
	    GLOBAL_DEBUG_STREAM("Unknown draw type" << endl);
	    break;
	  }
	}else{  //if middle mousebutton ist pressed
	  imousecursor->useTriangleCursor(false);
	  isvgencode->drawSVGPointer(vncevent.x, vncevent.y, (((vncevent.buttonmask & 0x1f00) >> 8) &  ~(1 << (vncevent.button - 1))));
	}
	break;

      case MOUSEMOVED:
	if(((vncevent.buttonmask & 0x1f00) >> 8)!=2){
	  switch (svgDrawType) {
	  case ISVGEncode::line:
	    isvgencode->drawSVGLine(vncevent.x, vncevent.y, ((vncevent.buttonmask & 0x1f00) >> 8));
	    break;
	  case ISVGEncode::rectangle:
	    isvgencode->drawSVGRectangle(vncevent.x, vncevent.y, ((vncevent.buttonmask & 0x1f00) >> 8));
	    break;
	  case ISVGEncode::freehand:
	    isvgencode->drawSVGFreehand(vncevent.x, vncevent.y, ((vncevent.buttonmask & 0x1f00) >> 8));
	    break;
	  case ISVGEncode::pointer:
	    //imousecursor->useTriangleCursor(true);
	    isvgencode->drawSVGPointer(vncevent.x, vncevent.y, ((vncevent.buttonmask & 0x1f00) >> 8));
	    break;
	  default:
	    GLOBAL_DEBUG_STREAM("Unknown draw type" << endl);
	    break;
	  }
	}else{ //if middle mousebutton ist pressed
	  imousecursor->useTriangleCursor(true);
	  isvgencode->drawSVGPointer(vncevent.x, vncevent.y, ((vncevent.buttonmask & 0x1f00) >> 8));
	}
	break;

      case KEYPRESSED:
        if(vncevent.keyPressed == 0x020  ||   //XK_space = 0x020
           vncevent.keyPressed == 0xFF08 ||   //XK_Backspace = 0xFF08
           vncevent.keyPressed == 0xFF0D ||   //XK_Return = 0xFF0D
           vncevent.keyPressed == 0xFF51 ||   //XK_Left = 0xFF51
           vncevent.keyPressed == 0xFF52 ||   //XK_Up = 0xFF52
           vncevent.keyPressed == 0xFF53 ||   //XK_Right = 0xFF53
           vncevent.keyPressed == 0xFF54 ||   //XK_Down = 0xFF54
           vncevent.keyPressed == 0xFF55 ||   //XK_Page_Up = 0xFF55
           vncevent.keyPressed == 0xFF56) {   //XK_Page_Down = 0xFF56
          ivnc->sendKeyEvent(vncevent.keyPressed,1);
        }
	break;

      case KEYRELEASED:
	if(vncevent.keyReleased == 0x020  ||   //XK_space = 0x020
           vncevent.keyReleased == 0xFF08 ||   //XK_Backspace = 0xFF08
           vncevent.keyReleased == 0xFF0D ||   //XK_Return = 0xFF0D
           vncevent.keyReleased == 0xFF51 ||   //XK_Left = 0xFF51
           vncevent.keyReleased == 0xFF52 ||   //XK_Up = 0xFF52
           vncevent.keyReleased == 0xFF53 ||   //XK_Right = 0xFF53
           vncevent.keyReleased == 0xFF54 ||   //XK_Down = 0xFF54
           vncevent.keyReleased == 0xFF55 ||   //XK_Page_Up = 0xFF55
           vncevent.keyReleased == 0xFF56){    //XK_Page_Down = 0xFF56
          ivnc->sendKeyEvent(vncevent.keyReleased,0);
          isvgencode->newAnnotation();
        }
	break;

      case RFBUPDATEREQUEST:
	ivnc->sendFrameBufferUpdateRequest();
	break;

      default:
	break;
      }
    }
  }
};





// ### End of eventhandler ###





SigReader::SigReader(GuiWidget* guiWidget_, NMM::NMMApplication& app_, MDIWindow* mdiWindow_, ContainerWidget* containerWidget_) :
  state(CONSTRUCTED),
  guiWidget(guiWidget_),
  mdiWindow(mdiWindow_),
  containerWidget(containerWidget_),
  app(app_),
  startDelay(3,0),
  vd("/dev/video0"),
  width(384),
  height(288),
  deinterlace(false),
  vc(2), // video channel
  vnc_host_and_port(""),
  vnc_password(""),
  vnctolerance(0.1),
  vncinterval(1.0),
  vncwidth(640),
  vncheight(480),
  found_intersect(false),
  serial_dev("/dev/ttyS0"),
  video_bitrate(400),
  audio_bitrate(128),
  vidout_filename(""),
  use_visca(false),
  use_display(false),
  display_events_enabled(false),
  vnc_display_events_enabled(false),
  vncMode(-1),
  svgDrawType(-1),
  svgDrawColorRed(-1),
  svgDrawColorGreen(-1),
  svgDrawColorBlue(-1),
  display(NULL),
  av_writer(NULL),
  vncnode(NULL),
  vnc_display(NULL),
  pngwrite(NULL),
  synchronizer(NULL),
  registry(app_.getRegistry()),
  veh(new VcoreEventHandler()),
  write_fd(-1),
  vnc_write_fd(-1),
  visca_cam(NULL),
  isvgencode(NULL),
  isvgpaint(NULL),
  ipngwrite(NULL),
  imousecursor(NULL),
  timestamp_listener(veh, &VcoreEventHandler::sendFilenameAndTimestamp),
  keyPressed_listener(veh, &VcoreEventHandler::keyPressed),
  keyReleased_listener(veh, &VcoreEventHandler::keyReleased),
  avi_frame_rate(0.0)
{
}

SigReader::~SigReader() {
  if (visca_cam != NULL) delete visca_cam;
  if (synchronizer != NULL) delete synchronizer;
  if (veh != NULL) delete veh;
}

void SigReader::preview() {
  switch (state) {
  case CONSTRUCTED:
    initPreview();
    state = INITPREVIEW; // fall through!
  case INITPREVIEW:
    startPreview();
    state = PREVIEW;
    break;
  default:
    break;
  }
}

void SigReader::start() {
  switch (state) {
  case PREVIEW:
    stopPreview();
    deinitPreview(); // fall through!!!
  case CONSTRUCTED:
    init();
    if (testFileName()) {
      initav();
      initvnc(); // fall through!!!
    } else {
      break;
    }
  case INITIALIZED:
    startav();
    startvnc();
    guiWidget->getButtonStart()->hide();
    guiWidget->getButtonStop()->show();
    state = STARTED;
    break;
  default:
    break;
  }
}

void SigReader::stop() {
  switch (state) {
  case PREVIEW:
    stopPreview();
    deinitPreview();
    state = STOPPED;
    break;
  case STARTED:
    stopav();
    stopvnc();
    guiWidget->getButtonStop()->hide();
    guiWidget->getButtonStart()->show();
    state = STOPPED;
    break;
  default:
    break;
  }
}

void SigReader::deinit() {
  // deinit means to eventually stop and then deinit
  switch (state) {
  case PREVIEW:
    stopPreview();
    deinitPreview();
    break;
  case STARTED:
    stop();        // fall through!
  case INITIALIZED:
  case STOPPED:
    internalDeinit();
    break;
  default:
    break;
  }
  state = CONSTRUCTED;
}

void SigReader::switchVNCMode(int vcoremode){
  if(vncMode != vcoremode){
    vncMode = vcoremode;      
      
    switch(vcoremode){
    case 0:  //annotate
      mdiWindow->getToolBarVNCMenu()->setDisabled(false);
      ipngwrite->setDiscardImages(false);
      veh->setAnnotationMode(true);
      veh->setVNCWorkingMode(IVNC::REQUESTED);
      isvgencode->repaint();
      break;
    case 1:  //present
      mdiWindow->getToolBarVNCMenu()->setDisabled(true);
      ipngwrite->setDiscardImages(false);
      veh->setAnnotationMode(false);
      veh->setVNCWorkingMode(IVNC::REQUESTED);
      break;
    case 2:  //write
      mdiWindow->getToolBarVNCMenu()->setDisabled(false);
      ipngwrite->setDiscardImages(false);
      veh->setAnnotationMode(true);
      veh->setVNCWorkingMode(IVNC::BLANK_SHEET);
      isvgencode->newAnnotation();
      break;
    case 3:  //prepare
      mdiWindow->getToolBarVNCMenu()->setDisabled(true);
      ipngwrite->setDiscardImages(true);
      veh->setAnnotationMode(false);
      veh->setVNCWorkingMode(IVNC::REAL_TIME);
      break;
    default:
      break;
    }
    emit vncModeChanged(vcoremode);
  }
}


void SigReader::setSVGDrawType(int svge) {
  if(svgDrawType != svge){
    svgDrawType = svge;
    ISVGEncode::SVGDrawMode svgevent;
    switch(svge){
    case 0:
      svgevent = ISVGEncode::pointer;
      mdiWindow->getToolButtonBlue()->setDisabled(true);
      mdiWindow->getToolButtonRed()->setDisabled(true);
      mdiWindow->getToolButtonGreen()->setDisabled(true);
      mdiWindow->getToolButtonBlack()->setDisabled(true);
      mdiWindow->getToolButtonCustom()->setDisabled(true);
      break;
    case 1:
      svgevent = ISVGEncode::freehand;
      mdiWindow->getToolButtonBlue()->setDisabled(false);
      mdiWindow->getToolButtonRed()->setDisabled(false);
      mdiWindow->getToolButtonGreen()->setDisabled(false);
      mdiWindow->getToolButtonBlack()->setDisabled(false);
      mdiWindow->getToolButtonCustom()->setDisabled(false);
      break;
    case 2:
      svgevent = ISVGEncode::rectangle;
      mdiWindow->getToolButtonBlue()->setDisabled(false);
      mdiWindow->getToolButtonRed()->setDisabled(false);
      mdiWindow->getToolButtonGreen()->setDisabled(false);
      mdiWindow->getToolButtonBlack()->setDisabled(false);
      mdiWindow->getToolButtonCustom()->setDisabled(false);
      break;
    case 3:
      svgevent = ISVGEncode::line;
      mdiWindow->getToolButtonBlue()->setDisabled(false);
      mdiWindow->getToolButtonRed()->setDisabled(false);
      mdiWindow->getToolButtonGreen()->setDisabled(false);
      mdiWindow->getToolButtonBlack()->setDisabled(false);
      mdiWindow->getToolButtonCustom()->setDisabled(false);
      break;
    default:
      svgevent = ISVGEncode::freehand;
      mdiWindow->getToolButtonBlue()->setDisabled(false);
      mdiWindow->getToolButtonRed()->setDisabled(false);
      mdiWindow->getToolButtonGreen()->setDisabled(false);
      mdiWindow->getToolButtonBlack()->setDisabled(false);
      mdiWindow->getToolButtonCustom()->setDisabled(false);
      break;
    }
    veh->setSVGDrawType(svgevent);
    emit svgDrawTypeChanged(svgevent);
  }
}

void SigReader::setSVGDrawColor(int _svgdrawcolor) {
  QColor custom_color; 
  switch(_svgdrawcolor){
  case 0: // blue
    svgDrawColorRed = 0;
    svgDrawColorGreen = 0;
    svgDrawColorBlue = 255;
    break;
  case 1: // green
    svgDrawColorRed = 0;
    svgDrawColorGreen = 255;
    svgDrawColorBlue = 0;
    break;
  case 2: // red
    svgDrawColorRed = 255;
    svgDrawColorGreen = 0;
    svgDrawColorBlue = 0;
    break;
  case 3: // black
    svgDrawColorRed = 0;
    svgDrawColorGreen = 0;
    svgDrawColorBlue = 0;
    break;
  default:  // lets user choose a color
    custom_color = QColorDialog::getColor(Qt::black, mdiWindow,"custom_color");
    if(custom_color.isValid()){
      svgDrawColorRed = custom_color.red();
      svgDrawColorGreen = custom_color.green();
      svgDrawColorBlue = custom_color.blue();
    }
    break;
  }
  
    
  veh->setSVGDrawColor(svgDrawColorRed, svgDrawColorGreen, svgDrawColorBlue);
  emit svgDrawColorChanged(svgDrawColorRed, svgDrawColorGreen, svgDrawColorBlue);
}

void SigReader::removeSVGAnnotation(){
  veh->removeAnnotation();
  emit svgAnnotationRemoved();
}

void SigReader::newSVGAnnotation(){
  veh->newAnnotation();
  emit svgAnnotationNew();
}

void SigReader::undoSVGAnnotation(){
  veh->undoAnnotation();
  emit svgAnnotationUndone();
}

void SigReader::redoSVGAnnotation(){
  veh->redoAnnotation();
  emit svgAnnotationRedone();
}

void SigReader::setVideoBitrate(int val) {
  if (val != video_bitrate) {
    video_bitrate = val;
    emit videoBitrateChanged(val);
  }
}

void SigReader::setAudioBitrate(int val) {
  if (val != audio_bitrate) {
    audio_bitrate = val;
    emit audioBitrateChanged(val);
  }
}

void SigReader::setVNCHostAndPort(const QString& host_and_port_) {
  if (vnc_host_and_port != host_and_port_) {
    vnc_host_and_port = host_and_port_;
    emit vncHostAndPortChanged(vnc_host_and_port);
  }
}

void SigReader::setVNCPassword(const QString& password_) {
  if (vnc_password != password_) {
    vnc_password = password_;
    emit vncPasswordChanged(vnc_password);
  }
}

void SigReader::setOutputFileName() {
  QString f = QFileDialog::getSaveFileName( QString::null, "*.avi", guiWidget );

  if ( !f.isEmpty() ) {
    vidout_filename = f;
    emit outputFileNameChanged(vidout_filename);
  } else {
    // the user cancelled the dialog
  }
}

void SigReader::setOutputFileName(const QString& outfilename_) {
  if (vidout_filename != outfilename_) {
    vidout_filename = outfilename_;
    emit outputFileNameChanged(vidout_filename);
  }
}

bool SigReader::testFileName() {
  if ( !vidout_filename.isEmpty() ) {
    FILE* avifile = fopen(vidout_filename.ascii (),"r");

    if (avifile != NULL) { // file exists
      switch( QMessageBox::warning(guiWidget, "File exists",
                                   "File " + vidout_filename + " already exists.\n" +
                                   "Should I overwrite it?\n\n",
                                   "Yes",
                                   "No", 0, 0, 1 ) ) {
      case 0: // The user clicked the Yes button or pressed Enter
        // nothing to be done
        return true;
        break;
      case 1: // The user clicked the No or pressed Escape
        // the file dialog window will be shown
        return false;
        break;
      }
    }
    return true;
  }
  return false;
}

void SigReader::setUseDisplay(bool usedisplay_) {
  if (use_display != usedisplay_) {
    use_display = usedisplay_;
  }
}

void SigReader::setUseVisca(bool usevisca_) {
  if (use_visca != usevisca_) {
    use_visca = usevisca_;
  }
}

void SigReader::setUseDeinterlacer(bool usedeinterlacer_) {
  if (deinterlace != usedeinterlacer_) {
    deinterlace = usedeinterlacer_;
  }
}

void SigReader::setWriteFD(int write_fd_) {
  if (write_fd != write_fd_) {
    write_fd = write_fd_;
    veh->setWriteFD(write_fd);
  }
}

void SigReader::setReadFD(int read_fd_) {
  if (read_fd != read_fd_) {
    read_fd = read_fd_;
  }
}

void SigReader::setVNCWriteFD(int write_fd_) {
  if (vnc_write_fd != write_fd_) {
    vnc_write_fd = write_fd_;
    veh->setVNCWriteFD(vnc_write_fd);
  }
}

void SigReader::setVNCReadFD(int read_fd_) {
  if (vnc_read_fd != read_fd_) {
    vnc_read_fd = read_fd_;
  }
}

void SigReader::setWidth(int width_) {
  if (width != width_) {
    width = width_;
  }
}

void SigReader::setHeight(int height_) {
  if (height != height_) {
    height = height_;
  }
}

void SigReader::setChannel(int channel_) {
  if (vc != channel_) {
    vc = channel_;
  }  
}

void SigReader::init() {
  try {
    /*Begin> Init requests for nodes */
    if (use_visca) {
      req_av_tvread.setNodeName("VISCACameraReadNode");
    } else {
      req_av_tvread.setNodeName("TVCardReadNode");
    }
    req_av_deinterlacer.setNodeName("YUVDeInterlaceNode");
    req_av_display.setNodeName("XDisplayNode");
    req_av_video_enc.setNodeName("FFMpegEncodeNode");
    req_av_audio_rec.setNodeName("RecordNode");
    req_av_audio_enc.setNodeName("MPEGAudioEncodeNode");
    req_av_avmux.setNodeName("AVMuxNode");
    req_av_avwrite.setNodeName("AVIWriteNode");

    req_vnc.setNodeName("VNCSourceNode");
    req_vnc_svgencode.setNodeName("SVGEncodeNode");
    req_vnc_svgpaint.setNodeName("SVGPaintNode");
    req_vnc_conv_vnc_svg.setNodeName("RGBtoRGBConverterNode");
    req_vnc_conv_vnc_png.setNodeName("RGBtoRGBConverterNode");
    req_vnc_conv_svg_dis.setNodeName("RGBtoRGBConverterNode");
    req_vnc_pngwrite.setNodeName("PNGWriteNode");
    req_vnc_display.setNodeName("XDisplayNode");
    /*End< */
    
    // Create the source synchronizer
    synchronizer = new MasterSlaveSourceSynchronizer("AVSourceSync");
  } catch (const NMM::Exception& e) {
    cerr << "init: " << e << endl;
    exit(1);
  }
}

void SigReader::initPreview() {
  
  try {

    if (use_visca) {
      req_pv_tvread.setNodeName("VISCACameraReadNode");
    } else {
      req_pv_tvread.setNodeName("TVCardReadNode");
    }
    req_pv_display.setNodeName("XDisplayNode");
    req_pv_deinterlacer.setNodeName("YUVDeInterlaceNode");
        
    req_pv_tvread.addNodeEvent(Event(ITVCard::setChannel_event, new TInValue<int>(vc)), INode::INITIALIZED);
    req_pv_tvread.addNodeEvent(Event(ITVCard::setResolution_event, new TInValue<int>(width), new TInValue<int>(height)), INode::INITIALIZED);
    req_pv_display.addNodeEvent(Event(ISynchronizedSink::setSynchronized_event, new TInValue<bool>(false)), INode::CONSTRUCTED);

    if (deinterlace) {
      pv_graph.addEdge(req_pv_tvread, req_pv_deinterlacer);
      pv_graph.addEdge(req_pv_deinterlacer, req_pv_display);
    } else {
      pv_graph.addEdge(req_pv_tvread, req_pv_display);
    }
    
    registry.requestGraph(pv_graph);
    pv_graph.realizeGraph();
    
    display.reset(pv_graph.getINode(req_pv_display));
    
  } catch (const NMM::Exception& e) {
    cerr << "initPreview: " << e << endl;
    exit(1);
  }
}

void SigReader::enableDisplayEvents() {
  IXDisplayEventSender_var xdisplay_event_sender;
  display_events_enabled = true;
  IVISCACamera* visca_cam(pv_graph.getINode(req_pv_tvread)->getParentObject()->getCheckedInterface<IVISCACamera>());
  veh->setIVISCACamera(visca_cam);
  xdisplay_event_sender.reset(display->getParentObject()->getInterface<IXDisplayEventSender>());
  if (display->isStarted() && xdisplay_event_sender.get()) {
    xdisplay_event_sender->enableDisplayEvents(getNMMResource("vcorerecorder_configuration.xml"));
    xdisplay_event_sender->setXAutoRepeat(false);
  }
  display->getParentObject()->registerEventListener(IDisplayEvents::keyPressed_event, &keyPressed_listener);
  display->getParentObject()->registerEventListener(IDisplayEvents::keyReleased_event, &keyReleased_listener);
}

void SigReader::enableVNCDisplayEvents() {
  IXDisplayEventSender_var vnc_xdisplay_event_sender;
  vnc_display_events_enabled = true;
  vnc_xdisplay_event_sender.reset(vnc_display->getParentObject()->getInterface<IXDisplayEventSender>());
  if (vnc_display->isStarted() && vnc_xdisplay_event_sender.get()) {
    vnc_xdisplay_event_sender->enableDisplayEvents(getNMMResource("vncviewer_configuration.xml"));
    vnc_xdisplay_event_sender->setSendingMouseEventsDirect(true);
    vnc_xdisplay_event_sender->setUsingKeySymKeyEvents(true);
  }
  vnc_display->getParentObject()->
    registerEventListener(IDisplayEvents::keysym_keyPressed_event, new TEDObject1<VcoreEventHandler, unsigned long>(veh, &VcoreEventHandler::sendKeyPressedEvent));

  vnc_display->getParentObject()->
    registerEventListener(IDisplayEvents::keysym_keyReleased_event, new TEDObject1<VcoreEventHandler, unsigned long>(veh, &VcoreEventHandler::sendKeyReleasedEvent));

  vnc_display->getParentObject()->
    registerEventListener(IDisplayEvents::windowExposed_event, new TEDObject0<VcoreEventHandler>(veh, &VcoreEventHandler::sendRFBUpdateRequest));

  vnc_display->getParentObject()->
    registerEventListener(IDisplayEvents::mouseButtonPressed_event, new TEDObject5<VcoreEventHandler, int, int, int, int, bool>(veh, &VcoreEventHandler::sendButtonPressedEvent));

  vnc_display->getParentObject()->
    registerEventListener(IDisplayEvents::mouseButtonReleased_event, new TEDObject5<VcoreEventHandler, int, int, int, int, bool>(veh, &VcoreEventHandler::sendButtonReleasedEvent));

  vnc_display->getParentObject()->
    registerEventListener(IDisplayEvents::mouseMoved_event, new TEDObject4<VcoreEventHandler, int, int, int, bool
			  >(veh, &VcoreEventHandler::sendMouseMotionEvent));
}

void SigReader::disableDisplayEvents() {
  display->getParentObject()->removeEventListener(&keyPressed_listener);
  display->getParentObject()->removeEventListener(&keyReleased_listener);
  display_events_enabled = false;
}

void SigReader::disableVNCDisplayEvents() {
  vnc_display->getParentObject()->removeEventListener(&keyPressed_listener);
  vnc_display->getParentObject()->removeEventListener(&keyReleased_listener);
  vnc_display_events_enabled = false;
}

void SigReader::startPreview() {
    pv_graph.startGraph();
    if (use_visca) {
	enableDisplayEvents();
    }
}

void SigReader::handleEvent() {
  char c;
  read(read_fd, &c, 1);
  veh->handleEvents();
}

void SigReader::handleVNCEvent() {
  char c;
  read(vnc_read_fd, &c, 1);
  veh->handleVNCEvents();
}

void SigReader::stopPreview() {
    pv_graph.stopGraph();
    
    if (display_events_enabled) {
	disableDisplayEvents();
    }
}

void SigReader::deinitPreview() {
  registry.releaseGraph(pv_graph);

  // workaround for possible bug in TVCardReadNode:
  // do not deinit tvread again (if graph has been started and completely downgraded,
  // and then tvread->init() is called again, it will fail.)
  //   tvread->deinit();
  
}

void SigReader::initav() {
  try {
    vncnode.reset(vnc_graph.getINode(req_vnc));
        
    // set queue size and mode for display
    if (use_display) {
      req_av_display.addNodeEvent(Event(IBiDirQueuedNode::setDownstreamMaxSize_event, new TInValue<int>(5)), INode::CONSTRUCTED);
      req_av_display.addNodeEvent(Event(IBiDirQueuedNode::setMode_event, new TInValue<IQueued::Mode>(IQueued::MODE_KEEP_NEWEST)), INode::CONSTRUCTED);
    }
    
    req_av_avwrite.addNodeEvent(Event(ISinkFileHandler::setFilename_event, new TInValue<string>(vidout_filename.ascii())), INode::INITIALIZED);
    req_av_tvread.addNodeEvent(Event(ITVCard::setResolution_event, new TInValue<int>(width), new TInValue<int>(height)), INode::INITIALIZED);
    req_av_tvread.addNodeEvent(Event(ITVCard::setChannel_event, new TInValue<int>(vc)), INode::INITIALIZED);

    // connect the nodes
    // video part
    if (deinterlace) {
      av_graph.addEdge(req_av_tvread, req_av_deinterlacer);
      if (use_display) {
	av_graph.addEdge(req_av_deinterlacer, req_av_display);
      }
      av_graph.addEdge(req_av_deinterlacer, req_av_video_enc);
    }
    else {
      if (use_display) {
	av_graph.addEdge(req_av_tvread, req_av_display);
      }
      av_graph.addEdge(req_av_tvread, req_av_video_enc);
    }

    Format* enc_out_format = new Format("video/mpeg4");
    enc_out_format->addIntParamValue(Format::bitrate_param, video_bitrate * 1000);
    av_graph.addEdge(req_av_video_enc, req_av_avmux, "video", enc_out_format);

    // audio part
   
    // record mono only
    Format* rawaucon_format = new Format("audio/raw");
    rawaucon_format->addIntParamValue(Format::channels_param, 1);
    rawaucon_format->addStringParamValue(Format::endian_param, "little");
    rawaucon_format->addStringParamValue(Format::encoding_param, "linear");
    rawaucon_format->addStringParamValue(Format::sign_param, "signed");
    
    av_graph.addEdge(req_av_audio_rec, req_av_audio_enc, rawaucon_format);

    
    Format* aucon_format = new Format("audio/mpeg");
    aucon_format->addIntParamValue(Format::layer_param, 3);
    aucon_format->addIntParamValue(Format::bitrate_param, audio_bitrate * 1000);

    av_graph.addEdge(req_av_audio_enc, req_av_avmux, "audio", aucon_format);
    av_graph.addEdge(req_av_avmux, req_av_avwrite);

    registry.requestGraph(av_graph);
    av_graph.realizeGraph();

    av_writer.reset(av_graph.getINode(req_av_avwrite));

    // remember framerate for calculation of duration of AVI file
    Format* videnc_out_format = av_graph.getINode(req_av_video_enc)->getOutputStreamFormat();
    avi_frame_rate = videnc_out_format->getFloatValue(Format::framerate_param);

  } catch (const NMM::Exception& e) {
    cerr << "initav: " << e << endl;
    exit(1);
  }
}

void SigReader::initvnc() {
  try {
    //////////////
    // VNC PART //
    //////////////

    // Request vncnode sperately. You need information about the configuration
    // of the vnc server for building the graph correctly. This information
    // about the colorspace is available in state OUTPUT_INITIALIZED
    if (!vncnode.get()) {
      RegistryLock lock(registry);

      list<Response> vnc_response = registry.initRequest(req_vnc);

      if (vnc_response.empty()) {
	GLOBAL_ERROR_STREAM("Size of vnc_response " << vnc_response.size() << endl);
	throw Exception(string("vnc node is not available\n") +
			string("Size of vnc_response ") + vnc_response.size() + string("\n"));
      }
      
      vncnode.reset(registry.requestNode(vnc_response.front()));
    }

    ivncvnc = vncnode->getParentObject()->getCheckedInterface<IVNC>();
    veh->setIVNC(ivncvnc);
    INetClient_var ivncnet(vncnode->getParentObject()->getCheckedInterface<INetClient>());

    string vnc_hap = string(vnc_host_and_port.ascii());
    uint lpos = vnc_hap.rfind(":");
    string host, portstr;
    int port;
    if (lpos != string::npos) {
      host = string(vnc_hap, 0, lpos);
      portstr = string(vnc_hap, lpos + 1, vnc_hap.length() - lpos + 1);
      port = atoi(portstr.c_str());
      ivncnet->setHostname(host);
      ivncnet->setPort(port);
    } else {
      cerr << "Wrong format of host and port. Use host:port, e.g. foo:1!" << endl;
      exit (-1);
    }
    
    ivncvnc->setMode(IVNC::REQUESTED);

    // set password and encrypt it
    unsigned char encoded_password[CHALLENGESIZE];
    encryptPasswd(encoded_password, vnc_password.ascii());
    encoded_password[8]=0;

    ivncvnc->setPasswd(string(reinterpret_cast<char*>(encoded_password)));
    ivncvnc->setTolerance(vnctolerance);
    ivncvnc->setInterval(vncinterval);
    ivncvnc->setShared(true);

    // As it is not known which RGB format the foreign VNC server has,
    // it might be necessary to use a RGBtoRGBConverterNode.
    vncnode->init();
    vncnode->initOutput();

    int vncdepth;
    Property_var  vncout_property(vncnode->getWorkingOutputProperty());
    const Format* def_vncout_format = vncout_property->getDefaultFormat();
    vncwidth    = def_vncout_format->getIntValue(Format::x_resolution_param);
    vncheight   = def_vncout_format->getIntValue(Format::y_resolution_param);
    vncdepth    = def_vncout_format->getIntValue(Format::bitperpixel_param);

    found_intersect = true;
    if ( ((vncdepth == 24) || (vncdepth == 32)) &&
	 def_vncout_format->getStringValue(Format::colorspace_param) == "rgb24" &&
	 def_vncout_format->getStringValue(Format::channelorder_param) == "rgb") {
      found_intersect = true;
    } else {
      found_intersect = false;
    }

    req_vnc_pngwrite.addNodeEvent(Event(IFileModeHandler::setFilenameMode_event, new TInValue<IFileModeHandler::FilenameMode>(IFileModeHandler::TIMESTAMP_AS_NAME)), INode::CONSTRUCTED);
    // write the slides in the same directory as the AVI file:
    string vidfilename = string(vidout_filename.ascii());
    unsigned int lastPos = vidfilename.find_last_of('/');
    if (lastPos != string::npos) {
      string workdir = vidfilename.substr(0,lastPos);
      req_vnc_pngwrite.addNodeEvent(Event(ISinkFileHandler::setWorkingDirectory_event, new TInValue<string>(workdir)), INode::CONSTRUCTED);
    }
    
    if (found_intersect) {
      vnc_graph.addEdge(req_vnc, req_vnc_pngwrite);
    } else {
      GLOBAL_DEBUG_STREAM("Using converter before PNGWrite" << endl);
      vnc_graph.addEdge(req_vnc, req_vnc_conv_vnc_png);
      vnc_graph.addEdge(req_vnc_conv_vnc_png, req_vnc_pngwrite);
    }
    
    // lower priority of the pngwrite thread, it is assumed that slides come
    // time by time and not too many at one time. To not take away to much resources
    // lower this nodes priority.
    req_vnc_pngwrite.addNodeEvent(Event(IThreadedNode::lowerPriority_event, new TInValue<int>(10)), INode::ACTIVATED);
			      
    req_vnc_svgencode.addNodeEvent(Event(ISVGEncode::setDimension_event, new TInValue<int>(vncwidth), new TInValue<int>(vncheight)), INode::INITIALIZED);
    
    vnc_graph.addEdge(req_vnc_svgencode, req_vnc_svgpaint, "svgtextstream");
    
    if (found_intersect) {
      vnc_graph.addEdge(req_vnc, req_vnc_svgpaint, "imagestream");
    } else {
      GLOBAL_DEBUG_STREAM("Using converter before SVGPaint " << endl);
      vnc_graph.addEdge(req_vnc, req_vnc_conv_vnc_svg);
      vnc_graph.addEdge(req_vnc_conv_vnc_svg, req_vnc_svgpaint, "imagestream");
    }
    
    vnc_graph.addEdge(req_vnc_svgpaint, req_vnc_display);

    // setting explicit vncnode which is requested above
    vnc_graph.setINode(req_vnc, *vncnode);

    registry.requestGraph(vnc_graph);
    vnc_graph.realizeGraph();

    vnc_display.reset(vnc_graph.getINode(req_vnc_display));
    pngwrite.reset(vnc_graph.getINode(req_vnc_pngwrite));

    ipngwrite  = pngwrite->getParentObject()->getCheckedInterface<IPNGWrite>();
    isvgencode = vnc_graph.getINode(req_vnc_svgencode)->getParentObject()->getCheckedInterface<ISVGEncode>();
    isvgpaint  = vnc_graph.getINode(req_vnc_svgpaint)->getParentObject()->getCheckedInterface<ISVGPaint>();
    imousecursor =  vnc_display->getParentObject()->getInterface<IMouseCursor>();
    veh->setISVGEncode(isvgencode);
    veh->setISVGPaint(isvgpaint);
    veh->setIMouseCursor(imousecursor);
    
  } catch (const NMM::Exception& e) {
    cerr << "initvnc: " << e << endl;
    exit(1);
  }
}

void SigReader::startav() {

  ISynchronizedSink_var video_sync;
  if (use_display) {
    video_sync.reset(av_graph.getINode(req_av_display)->getParentObject()->getCheckedInterface<ISynchronizedSink>());
  }
  IMasterSlaveSourceSynchronizer_var
    isynchronizer(synchronizer->getCheckedInterface<IMasterSlaveSourceSynchronizer>());
  ISynchronizedSource_var sync_audio_rec(av_graph.getINode(req_av_audio_rec)->getParentObject()->getCheckedInterface<ISynchronizedSource>());
  if (!sync_audio_rec.get()) cerr << "Could not get ISynchronizedSource from audio_rec" << endl;
  ISynchronizedSource_var sync_vncnode(vncnode->getParentObject()->getCheckedInterface<ISynchronizedSource>());
  if (!sync_vncnode.get()) cerr << "Could not get ISynchronizedSource from vncnode" << endl;
  ISynchronizedSource_var sync_tvread(av_graph.getINode(req_av_tvread)->getParentObject()->getCheckedInterface<ISynchronizedSource>());
  if (!sync_tvread.get()) cerr << "Could not get ISynchronizedSource from tvread->getParentObject()" << endl;
  
  sync_audio_rec->setSynchronized(true);
  sync_tvread->setSynchronized(true);
  sync_vncnode->setSynchronized(true);
  if (use_display) {
    // do not use intra stream sync on surveillance display
    // because the timestamps of the video frames are corrected
    // by source sync and may confuse sink sync.
    video_sync->setSynchronized(false);
  }
  
  isynchronizer->setMasterController(sync_audio_rec->getController());
  isynchronizer->addSlaveController(sync_tvread->getController());
  isynchronizer->addSlaveController(sync_vncnode->getController());
  isynchronizer->setStartDelay(startDelay);
  
  av_graph.startGraph();
}

void SigReader::startvnc() {

  
  if (vnc_display.get()) {
    ISynchronized_var ivncdisplay_synchronized(vnc_display->getParentObject()->getInterface<ISynchronized>());
    // display does not have to be synchronized
    // and without synchronization presentation is much faster
    ivncdisplay_synchronized->setSynchronized(false);
  } else {
    cerr << "No vnc display found. Synchronisation for vnc display couldn't be disabled" << endl;
  }
  
  // empty the list of timestamps at each start
  veh->clearTimestamps();

  pngwrite->getParentObject()->
    registerEventListener(ISendFilenameAndTimestamp::sendFilenameAndTimestamp_event, &timestamp_listener);

  vnc_graph.startGraph();
  enableVNCDisplayEvents();
  
  // Embed XDisplay-Window into containerWidget
  IXWindowID_var vnc_display_ixwindowid;
  vnc_display_ixwindowid.reset(vnc_display->getParentObject()->getInterface<IXWindowID>());
  containerWidget->embedWindow(vnc_display_ixwindowid->getWindowID());
  containerWidget->setMinimumSize(vncwidth,vncheight);
  containerWidget->showMaximized();
  mdiWindow->getWorkspace()->setMinimumSize(vncwidth,vncheight);
  mdiWindow->getWorkspace()->adjustSize();
  mdiWindow->show();
}

void SigReader::stopav() {
  av_graph.stopGraph();

  // disable display events
  if (display_events_enabled) {
    disableDisplayEvents();
  }

  createSMIL();
}

void SigReader::stopvnc() {
  vnc_graph.stopGraph();

  pngwrite->getParentObject()->removeEventListener(&timestamp_listener);

  // disable VNC display events
  if (vnc_display_events_enabled) {
    disableVNCDisplayEvents();
  }
}

void SigReader::internalDeinit() {
  
  // detach XDisplay-Window from containerWidget
  containerWidget->discardWindow(); 

  registry.releaseGraph(av_graph);
  registry.releaseGraph(vnc_graph);
 
}

void SigReader::createSMIL() {
  xmlDocPtr docptr;
  docptr = xmlNewDoc((xmlChar*)"1.0");
  docptr->encoding = xmlStrdup((xmlChar*)"UTF-8");
  docptr->xmlRootNode = xmlNewDocNode(docptr, NULL, (xmlChar*)"smil", NULL);
  xmlNodePtr root = docptr->xmlRootNode;

  // add head and body sections
  xmlNodePtr head = xmlNewChild(root, NULL, (xmlChar*)"head", NULL);
  xmlNodePtr body = xmlNewChild(root, NULL, (xmlChar*)"body", NULL);

  // add layout section to head
  xmlNodePtr layout = xmlNewChild(head, NULL, (xmlChar*)"layout", NULL);
  xmlNodePtr rootlayout = xmlNewChild(layout, NULL, (xmlChar*)"root-layout", NULL);
  xmlSetProp(rootlayout, (xmlChar*)"width", (xmlChar*)"1024");
  xmlSetProp(rootlayout, (xmlChar*)"height", (xmlChar*)"480");
  xmlSetProp(rootlayout, (xmlChar*)"background-color", (xmlChar*)"white");
  xmlNodePtr regionvideo = xmlNewChild(layout, NULL, (xmlChar*)"region", NULL);
  xmlSetProp(regionvideo, (xmlChar*)"id", (xmlChar*)"video");
  xmlSetProp(regionvideo, (xmlChar*)"left", (xmlChar*)"0");
  xmlSetProp(regionvideo, (xmlChar*)"top", (xmlChar*)"0");
  xmlSetProp(regionvideo, (xmlChar*)"width", (xmlChar*)"384");
  xmlSetProp(regionvideo, (xmlChar*)"height", (xmlChar*)"288");
  xmlNodePtr regionslide = xmlNewChild(layout, NULL, (xmlChar*)"region", NULL);
  xmlSetProp(regionslide, (xmlChar*)"id", (xmlChar*)"slide");
  xmlSetProp(regionslide, (xmlChar*)"left", (xmlChar*)"384");
  xmlSetProp(regionslide, (xmlChar*)"top", (xmlChar*)"0");
  xmlSetProp(regionslide, (xmlChar*)"width", (xmlChar*)"640");
  xmlSetProp(regionslide, (xmlChar*)"height", (xmlChar*)"480");

  // add "par" section to body
  xmlNodePtr par = xmlNewChild(body, NULL, (xmlChar*)"par", NULL);

  // add video and seq to par section
  xmlNodePtr video = xmlNewChild(par, NULL, (xmlChar*)"video", NULL);
  // extract the filename from the path to the AVI file
  string vidsrc = string(vidout_filename.ascii());
  unsigned int lastPos = vidsrc.find_last_of('/');
  if (lastPos != string::npos) {
    vidsrc = vidsrc.substr(lastPos + 1, vidsrc.length());
  }
  xmlSetProp(video, (xmlChar*)"src", (xmlChar*)vidsrc.c_str());
  xmlSetProp(video, (xmlChar*)"region", (xmlChar*)"video");
  xmlNodePtr seq = xmlNewChild(par, NULL, (xmlChar*)"seq", NULL);

  xmlNodePtr img;
  map<string, Timestamp> timestamps = veh->getTimestamps();
  map<string, Timestamp>::iterator iter = timestamps.begin();

  while (iter != timestamps.end()) {
    string strsrc = iter->second.sync_time.getPrettyString();
    strsrc += ".png";
    const char* src = strsrc.c_str();
    string strdur;
    const char* dur;

    Interval duration;

    map<string, Timestamp>::iterator next = iter;
    next++;

    if (next != timestamps.end()) {
      duration = next->second.sync_time - iter->second.sync_time;
    } else {
      // special handling for the last slide
      IFrameCount_var iframecount(av_writer->getParentObject()->getCheckedInterface<IFrameCount>());
      int framecount = iframecount->getFrameCount();
      duration = NMM::Time(Rational(framecount * 1000, static_cast<int>(avi_frame_rate * 1000))) - iter->second.sync_time;
    }
    strdur = duration.getPrettyString();
    dur = strdur.c_str();

    img = xmlNewChild(seq, NULL, (xmlChar*)"img", NULL);
    xmlSetProp(img, (xmlChar*)"src", (xmlChar*)src);
    xmlSetProp(img, (xmlChar*)"region", (xmlChar*)"slide");
    xmlSetProp(img, (xmlChar*)"duration", (xmlChar*)dur);
    iter++;
  }

  string filename = string(vidout_filename.ascii());
  filename.replace(filename.find(".avi"), 4, ".smil");

  int ret = xmlSaveFormatFileEnc (filename.c_str(), docptr, "UTF-8", 1);//pretty_formated_xml ? 1 : 0);
  if (ret == -1) cerr << "Could not save SMIL File!!!" << endl;
}
