/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Eric Peters <epeters@graphics.cs.uni-sb.de>
 *                    Christian Gerstner <gerstner@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_VCORE_SIGREADER_HPP_
#define _NMM_VCORE_SIGREADER_HPP_ _NMM_VCORE_SIGREADER_HPP_

#include "GuiWidget.hpp"
#include "MDIWindow.hpp"
#include "ContainerWidget.hpp"
#include "nmm/base/EDObject.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/interfaces/video/viscacamera/IVISCACamera.hpp"
#include "nmm/interfaces/text/ISVGEncode.hpp"
#include "nmm/interfaces/text/ISVGPaint.hpp"
#include "nmm/interfaces/file/IPNGWrite.hpp"
#include "nmm/interfaces/net/IVNC.hpp"
#include "nmm/interfaces/video/display/IMouseCursor.hpp"

using namespace NMM;

class SigReader : public QObject
{
  Q_OBJECT
public:
  SigReader(GuiWidget*, NMM::NMMApplication&, MDIWindow* = 0, ContainerWidget* = 0);
  ~SigReader();

  int getVideoBitrate() { return video_bitrate; }
  int getAudioBitrate() { return audio_bitrate; }
  QString getVNCHostAndPort() { return vnc_host_and_port; }
  QString getVNCPassword() { return vnc_password; }
  QString getOutputFileName() { return vidout_filename; }

  /**
   * init - Request all needed nodes from registry
   */
  void init();

  /**
   * initav - initialize the audio/video part of the graph
   * bring all nodes to ACTIVATED
   */
  void initav();

  /**
   * initvnc - initialize the vnc part of the graph
   * bring all nodes to ACTIVATED
   */
  void initvnc();

  /**
   * set the channel for the tvcard. This should be improved by
   * setting it from the GUI.
   */
  void setChannel(int);

  enum vncmode{_annotate,_present,_write,_prepare};

public slots:
  /**
   * start the recording. This is called when the Start Button is pressed
   */
  void start();
  /**
   * stop the preview or the recording. This is called when the Stop button is pressed
   */
  void stop();
  /**
   * start the preview. This is called when the Preview Button is pressed
   */
  void preview();
  /**
   * change the video bitrate, called if the slider in the GUI is moved
   */
  void setVideoBitrate(int);
  /**
   * change the audio bitrate, called if the slider in the GUI is moved
   */
  void setAudioBitrate(int);
  /**
   * set VNC host and port from the editline in the GUI
   */
  void setVNCHostAndPort(const QString&);
  /**
   * set VNC password from the password field in the GUI
   */
  void setVNCPassword(const QString&);
  /**
   * set the output filename by opening a file dialog
   */
  void setOutputFileName();
  /**
   * set the output filename from the editline in the GUI
   */
  void setOutputFileName(const QString&);
  /**
   * test wether the file already exists
   */
  bool testFileName();
  /**
   * deinitialize the nodes
   */
  void deinit();
  /**
   * set whether to use a surveillance display or not,
   * called if the switchbox in the GUI changes its state
   */
  void setUseDisplay(bool);
  /**
   * set whether to use the VISCA control or not,
   * called if the switchbox in the GUI changes its state
   */
  void setUseVisca(bool);
  /**
   * set the width for video, called if the editline in the GUI changes
   */
  void setWidth(int);
  /**
   * set the height for video, called if the editline in the GUI changes
   */
  void setHeight(int);
  /**
   * set whether to deinterlace the video or not,
   * called if the switchbox in the GUI changes its state
   */
  void setUseDeinterlacer(bool);
  /**
   * set the write file descriptor of the event pipe
   */
  void setWriteFD(int);
  /**
   * set the read file descriptor of the event pipe
   */
  void setReadFD(int);
  /**
   * called if any event arrives at the application. Calls the handleEvents method of
   * the VcoreEventHandler
   */
  void handleEvent();
  /**
   * create the SMIL file depending on the filenames and timestamps got from the
   * PNGWriteNode and the AVIWriteNode
   */
  void createSMIL();
  void setVNCWriteFD(int);
  void setVNCReadFD(int);
  void switchVNCMode(int);
  void setSVGDrawType(int);
  void setSVGDrawColor(int);
  void removeSVGAnnotation();
  void newSVGAnnotation();
  void undoSVGAnnotation();
  void redoSVGAnnotation();
  void handleVNCEvent();

signals:
  /**
   * signal to emit if the video bitrate was changed (qt-specific)
   */
  void videoBitrateChanged(int);
  /**
   * signal to emit if the audio bitrate was changed (qt-specific)
   */
  void audioBitrateChanged(int);
  /**
   * signal to emit if the host and port for VNC were changed (qt-specific)
   */
  void vncHostAndPortChanged(const QString&);
  /**
   * signal to emit if the password for VNC was changed (qt-specific)
   */
  void vncPasswordChanged(const QString&);
  /**
   * signal to emit if the filename was changed (qt-specific)
   */
  void outputFileNameChanged(const QString&);
  void vncsvgModeChanged(bool);
  void vncModeChanged(int);
  void pointerStateChanged(int);
  void vncWorkingModeChanged(IVNC::Mode);
  void svgDrawTypeChanged(ISVGEncode::SVGDrawMode);
  void svgDrawColorChanged(int, int, int);
  void svgAnnotationRemoved();
  void svgAnnotationNew();
  void svgAnnotationUndone();
  void svgAnnotationRedone();

private:
  /**
   * internal states of the application
   */
  enum states {CONSTRUCTED, INITIALIZED, INITPREVIEW, PREVIEW, STARTED, STOPPED} state;
  /**
   * the main widget
   */
  GuiWidget* guiWidget;
  MDIWindow* mdiWindow;
  ContainerWidget* containerWidget;
  NMM::NMMApplication& app;
    
  NMM::Interval startDelay; //<! time to wait before starting the recording
  string vd; //<! the video device
  int width; //<! width of the video
  int height; //<! height of the video
  bool deinterlace; //<! flag: use YUVDeinterlaceNode
  int vc; //<! video channel
  QString vnc_host_and_port; //<! Host and Port for the VNC server
  QString vnc_password; //<! the VNC password
  float vnctolerance; //<! tolerance to decide when to create a new image
  float vncinterval; //<! timeout for VNC
  int vncwidth; //<!width which the VNCServer delivers
  int vncheight; //<!heigth which the VNCServer delivers
  bool found_intersect; //<! test whether to use a RGBtoRGBConverter
  string serial_dev; //<! the serial device for the VISCA control

  int video_bitrate; //<! bitrate for video
  int audio_bitrate; //<! bitrate for audio
  QString vidout_filename; //<! video filename
  bool use_visca; //<! flag: use the VISCA control
  bool use_display; //<! flag: use a surveillance display
  bool display_events_enabled; //<! enable events from XDisplayNode
  bool vnc_display_events_enabled;
  bool annotationMode;
  int vncWorkingMode;
  int vncWorkingMode_Backup;
  int vncMode;
  int svgDrawType;
  int svgDrawColorRed;
  int svgDrawColorGreen;
  int svgDrawColorBlue;
  int pointerState;

  /**
   * internal initialization of the nodes for the preview mode
   */
  void initPreview();
  /**
   * start all nodes needed in preview mode
   */
  void startPreview();
  /**
   * stop nodes needed in preview mode
   */
  void stopPreview();
  /**
   * deinitialize all nodes needed in preview mode
   */
  void deinitPreview();
  /**
   * start the audio/video part of the graph
   */
  void startav();
  /**
   * start the VNC part of the graph
   */
  void startvnc();
  /**
   * stop the audio/video part of the graph
   */
  void stopav();
  /**
   * stop the VNC part of the graph
   */
  void stopvnc();
  /**
   * internal deinitialization of the nodes, called from deinit(),
   * depending on the applications state
   */
  void internalDeinit();
  /**
   * enable the display events arriving from the XDisplayNode
   */
  void enableDisplayEvents();
  /**
   * disable the display events arriving from the XDisplayNode
   */
  void disableDisplayEvents();
  void enableVNCDisplayEvents();
  void disableVNCDisplayEvents();

  //NodeDescriptions for AV-Graph
  NodeDescription req_av_tvread;
  NodeDescription req_av_display;
  NodeDescription req_av_deinterlacer;
  NodeDescription req_av_video_enc;
  NodeDescription req_av_audio_rec;
  NodeDescription req_av_audio_enc;
  NodeDescription req_av_avmux;
  NodeDescription req_av_avwrite;

  //NodeDescriptions for Preview-Graph
  NodeDescription req_pv_tvread;
  NodeDescription req_pv_display;
  NodeDescription req_pv_deinterlacer;
  
  //NodeDescriptions for VNC-Graph
  NodeDescription req_vnc_display;
  NodeDescription req_vnc;
  NodeDescription req_vnc_svgencode;
  NodeDescription req_vnc_svgpaint;
  NodeDescription req_vnc_conv_vnc_svg;
  NodeDescription req_vnc_conv_vnc_png;
  NodeDescription req_vnc_conv_svg_dis;
  NodeDescription req_vnc_pngwrite;

  INode_var display;
  INode_var av_writer;
  INode_var vncnode;
  INode_var vnc_display;
  INode_var pngwrite;
  

  // Source-Synchronizer
  MasterSlaveSourceSynchronizer *synchronizer;

  // The registry
  ClientRegistry& registry;
  GraphDescription pv_graph;
  GraphDescription av_graph;
  GraphDescription vnc_graph;

  class VcoreEventHandler;
  /**
   * The event handler for all events arriving at the application
   */
  VcoreEventHandler* veh;
  int write_fd;
  int read_fd;
  int vnc_write_fd;
  int vnc_read_fd;
  IVISCACamera* visca_cam;
  IVNC* ivncvnc;
  ISVGEncode* isvgencode;
  ISVGPaint* isvgpaint;
  IPNGWrite* ipngwrite;
  IMouseCursor* imousecursor;

  TEDObject2<VcoreEventHandler, string, Timestamp> timestamp_listener;
  TEDObject1<VcoreEventHandler, string> keyPressed_listener;
  TEDObject1<VcoreEventHandler, string> keyReleased_listener;
  
  float avi_frame_rate;
};
#endif //_NMM_VCORE_SIGREADER_HPP_
