/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"

// the plugin manager
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/ProxyApplication.hpp"

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <getopt.h>

/* Global options */
bool verbose = false;
string filename;

using namespace NMM;

void usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options] [filename]" << endl
       << endl
       << "    filename:            ac3 file" << endl
       << "    -v --verbose:        Verbose mode (lots of messages!)" << endl
       << "    -h --help:           Shows this help and exits." << endl;
  exit(0);
}

void turnOffAllMessages() {
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
}

int main(int argc, char ** argv) {
  // create the NMM application
  NMMApplication* app = ProxyApplication::getApplication(argc, argv);

  // get the registry
  ClientRegistry& registry = app->getRegistry();
  
  const char *optstring = "vh";
  struct option longopts[] = {
    {"verbose",        0, NULL, 'v'},
    {"help",           0, NULL, 'h'},
    {NULL,             0, NULL, 0  } /* last row must be zero */
  };
  
  int ch, optionindex;
  
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'v':
      verbose = true;
      break;
    default: //includes 'h'
      usage(argv[0]);
    }
  }
  
  if (argc <= optind) { // no file specified
    usage(argv[0]);
  } else {
    filename = string(argv[optind]);
  }
    
  // maybe turn off debug messages
  if (!verbose) {
    turnOffAllMessages();
  }
  
  INode_var readfile;
  INode_var ac3parse;
  
  int error_level = 0;
  try {

    /* create requests for nodes */
    NodeDescription readfile_req("GenericReadNode");
    NodeDescription ac3parse_req("AC3ParseNode");

    /* request nodes */
    {
      RegistryLock lock(registry);
      
      list<Response> readfile_resp = registry.initRequest(readfile_req);
      list<Response> ac3parse_resp = registry.initRequest(ac3parse_req);
      
      if (readfile_resp.empty() || ac3parse_resp.empty()) {
        throw Exception("A required node is not available");
      }
      
      readfile.reset(registry.requestNode(readfile_resp.front()));
      ac3parse.reset(registry.requestNode(ac3parse_resp.front()));
    }
    
    /* initialize file reader */
    readfile->init();
    IFileHandler_var file_source(readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(filename);
    
    ac3parse->init();
    
    /* setup source file format
       (needed because we are reading raw files) */
    Format* file_format = new Format(Format::audio_ac3_type);
    
    /* connect and activate the nodes */
    readfile->initOutput();
    connect(readfile.get(), ac3parse.get(), file_format);
    readfile->activate();
    ac3parse->initOutput();
    
    /* analyse format */
    Property_var ac3parse_wout(ac3parse->getWorkingOutputProperty());
    const Format* ac3_format = ac3parse_wout->getDefaultFormat();
    
    cout << "AC3 format:" << endl;
    cout << *ac3_format << endl;
    
  } catch( const Exception& e ) {
    
    cerr << e << endl;
    error_level = 1;
    
  } catch( ... ) {
    
    cerr << "Something strange happened..." << endl;
    return 2;
    
  }
  
  try {
    
    if (ac3parse.get()) {
      if (ac3parse->isOutputInitialized()) {
        ac3parse->deinitOutput();
      }
      if (ac3parse->isInitialized()) {
        IInputJack_var ac3parse_in(ac3parse->getInputStream("default"));
        if (ac3parse_in.get() && ac3parse_in->isConnected()) {
          if (readfile->isActivated()) {
            readfile->deactivate();
          }
          disconnect_input(ac3parse, "default");
        }
        ac3parse->deinit();
      }
    }
    
    if (readfile.get()) {
      if (readfile->isActivated()) {
        readfile->deactivate();
      }
      if (readfile->isOutputInitialized()) {
        readfile->deinitOutput();
      }
      if (readfile->isInitialized()) {
        readfile->deinit();
      }
    }
    
    if (ac3parse.get()) {
      registry.releaseNode(*ac3parse);
    }
    if (readfile.get()) {
      registry.releaseNode(*readfile);
    }
    
    return error_level;
    
  } catch( const Exception& e ) {
    
    cerr << "Exception during cleanup:" << endl;
    cerr << e << endl;
    return 2;
    
  } catch( ... ) {
    
    cerr << "Something strange happened during cleanup..." << endl;
    return 2;
    
  }
}
