/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


// C includes
#include <getopt.h>
#include <iostream>

// general includes
#include <set>
#include "nmm/base/connect.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/utils/utils.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/general/progress/IProgress.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"


using namespace NMM;


/**
 * Class for handling events.
 * An instance of this class is registered at some nodes as an event listener.
 */
class EventCatcher {
private:
  bool exit;
  ThreadMutex exit_mutex;
  ThreadCondition exit_cv;

  void signalExit() {
    MutexGuard em(exit_mutex);
    exit = true;
    exit_cv.notifyAll();
  }

public:
  EventCatcher() 
    : exit(false),
      exit_mutex(),
      exit_cv(exit_mutex) {}
  
  Result endTrack() {
    cout << "100% extracted - finish" << endl;
    signalExit();
    return SUCCESS;
  }

  Result setProgress(u_int64_t& actual, u_int64_t& total) {
    int progress = static_cast<int>(actual * 100.0 / static_cast<float>(total));
    if (10 > progress) {
      cout << "0";
    }
    cout << progress << "% extracted\r";
    cout.flush();
    return SUCCESS;
  }

  void handleEvents() {
    exit_mutex.lock();
    while (!exit) {
      exit_cv.wait();
    }
    exit_mutex.unlock();
  }

};


void beVerbose(unsigned int verbose) {
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }
}

int usage(char* name) {
  cerr << "Audio Extractor" << endl
       << endl
       << "Usage:" << endl
       << name << " [options] [input filename] [output filename]" << endl
       << endl
       << "    input filename:          mpeg program file" << endl
       << "    output filename:         filename(without suffix) of extracted audio stream(s)" << endl
       << "    -t       --track:        Set audio track number to extract (0), multible tracks are possible" << endl
       << "    -m       --mpeg:         Extract mpeg audio instead of ac3" << endl
       << "    -v       --verbose:      Verbose mode (lots of messages!)" << endl
       << "    -h       --help:         Shows this help and exits." << endl;
  return 0;
}

int main(int argc, char ** argv) {
  /* Global options */
  int verbose = 0;
  bool use_ac3 = true;
  bool force_mp3 = false;
  set<int> track_list;
  string input_filename;
  string output_filename;

  const char *optstring = "t:mvh";

  struct option longopts[] = {
    {"track",        1, NULL, 't'},
    {"mpeg",         0, NULL, 'm'},
    {"verbose",      0, NULL, 'v'},
    {"help",         0, NULL, 'h'},
    {NULL,           0, NULL, 0  } /* last row must be zero */
  };

  int ch, optionindex;
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 't':
      if (optarg) {
	track_list.insert( atoi(optarg) );
      }
      break;
    case 'm':
      force_mp3 = true;
      break;
    case 'v':
      verbose++;
      break;
    default: //includes 'h'
      return usage(argv[0]);
    }
  }

  if (track_list.empty()) {
    track_list.insert(0);
  }

  if (argc <= optind+1) { // no input/output file specified
    return usage(argv[0]);
  } else {
    input_filename = string(argv[optind]);
    output_filename = string(argv[optind+1]);
  }

  // eventually turn off debug messages
  beVerbose(verbose);


  NMMApplication* app = 0;
  
  /*Begin> Allocate _var variables for the INodes */
  INode_var readfile;
  INode_var demux;
  vector<INode*> raw_write(0);
  /*End< */
  
  /* Allocate an event listener */
  EventCatcher ec;
  
  try {
    app = ProxyApplication::getApplication(argc, argv);

    /*Begin> Init requests for nodes */
    NodeDescription readfile_req;
    readfile_req.setNodeName("MPEGReadNode");

    NodeDescription demux_req;
    demux_req.setNodeName("MPEGDemuxNode");
    
    NodeDescription raw_write_req;
    raw_write_req.setNodeName("GenericWriteNode");
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    {
      RegistryLock lock(registry);
      
      list<Response> readfile_resp = registry.initRequest(readfile_req);
      list<Response> demux_resp = registry.initRequest(demux_req);
      list<Response> raw_write_resp = registry.initRequest(raw_write_req);
      
      if (readfile_resp.empty() ||
          demux_resp.empty() ||
          raw_write_resp.empty()) {
        GLOBAL_ERROR_STREAM("Size of readfile_resp " << readfile_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of demux_resp " << demux_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of raw_write_resp " << raw_write_resp.size() << endl);
        throw Exception("A required node is not available");
      }
      
      readfile.reset(registry.requestNode(readfile_resp.front()));
      demux.reset(registry.requestNode(demux_resp.front()));
      raw_write = vector<INode*>(track_list.size());
      for (unsigned int i = 0; i < track_list.size(); i++) {
	raw_write[i] = registry.requestNode(raw_write_resp.front());
      } 
    }
    /*End< */
    
    /*Begin> Initialize and connect Nodes */
    IFileHandler_var file_source(readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(input_filename);
    IProgress_var progress(readfile->getParentObject()->getInterface<IProgress>());
    progress->sendProgressInformation(true);
    readfile->init();
    
    // XXX: assume we have a program stream stream and need a demux
    demux->init();
    
    readfile->initOutput();
    
    // XXX: here assume we have an mpeg program stream
    // XXX: if we had an elementary stream the connection would fail and we had to connect directly to mpg2yuv
    GLOBAL_DEBUG_STREAM("Connecting readfile to demux..." << endl);
    connect(readfile, demux);
    
    readfile->activate();
    demux->initOutput();
    
    list<string> demux_streams = demux->getOutputStreamTags();
    demux_streams.sort();
    cout << "Available Output streams:" << endl;
    for (list<string>::const_iterator i = demux_streams.begin();
	 i != demux_streams.end();
	 i++) {
      cout << "\t" << *i << endl;
      const Format* format = demux->getWorkingOutputProperty()->getDefaultFormat(*i);
      if (format->getFormatType() == Format::audio_ac3_type) {
	use_ac3 = true;
      }
    }
    if (force_mp3) {
      use_ac3 = false;
    }
    string audio_stream;
    int audio_track;
    for (set<int>::iterator i = track_list.begin(); i != track_list.end(); i++) {
      audio_track = *i;
      audio_stream = (use_ac3 ? string("ac3_audio") : string("mpeg_audio")) + audio_track;
      list<string>::iterator stream;
      for (stream = demux_streams.begin();
	   stream != demux_streams.end();
	   stream++) {
	if (*stream == audio_stream) {
	  break;
	}
      }
      if (demux_streams.end() == stream) {
	cerr << "Audio Stream " << audio_stream << " not found. Not using it." << endl;
	track_list.erase(i);
      }
    }

    string audio_jack_string;
    int i = 0;
    string fileext;
    for(set<int>::iterator iter=track_list.begin(); iter!=track_list.end(); iter++) {
      if (use_ac3) {
	audio_jack_string  = string("ac3_audio");
	fileext = string(".ac3");
      }
      else {
	audio_jack_string  = string("mpeg_audio");
	fileext = string(".mp3");
      }

      if (raw_write[i]) {
	ISinkFileHandler_var file_sink(raw_write[i]->getParentObject()->getCheckedInterface<ISinkFileHandler>());
	file_sink->setFilename(output_filename + *iter + fileext);
      }
      raw_write[i]->init();
      connect(demux.get(), (audio_jack_string + *iter).c_str(), raw_write[i]);

      i++;
    }

    demux->activate();

    if (!track_list.empty()) { // main loop only if there is a stream to extract
      readfile->start();
      demux->start();

      for(unsigned int i = 0; i < track_list.size(); i++) {
	raw_write[i]->initOutput();
	raw_write[i]->activate();
	raw_write[i]->start();
      }

      cout << "Extracting audio stream(s) (" << (use_ac3 ? "AC3" : "MPEG") << ") ";
      for(set<int>::iterator iter = track_list.begin(); iter != track_list.end(); iter++) {
	cout << *iter << " ";
      }
      cout << ":" << endl;
      raw_write[0]->getParentObject()->registerEventListener(ITrack::endTrack_event, new TEDObject0<EventCatcher>(&ec, &EventCatcher::endTrack));
      raw_write[0]->getParentObject()->registerEventListener(IProgressListener::setProgress_event, new TEDObject2<EventCatcher, u_int64_t, u_int64_t>(&ec, &EventCatcher::setProgress));
      ec.handleEvents();
    } else {
      cerr << "No streams to extract!" << endl;
    }

    /*Begin> Stop nodes */
    for (unsigned int i = 0; i < track_list.size(); i++) {
      raw_write[i]->stop();
    }
    if (demux->isStarted()) {
      demux->stop();
    }
    if (readfile->isStarted()) {
      readfile->stop();
    }
    /*End< */
  
    /*Begin> Flush nodes */
    for (unsigned int i = 0; i < track_list.size(); i++) {
      raw_write[i]->flush();
    }
    demux->flush();
    readfile->flush();
    /*End< */

    /*Begin> Deactivate nodes */
    for (unsigned int i = 0; i < track_list.size(); i++) {
      raw_write[i]->deactivate();
    }
    demux->deactivate();
    readfile->deactivate();
    /*End< */

    /*Begin> DeinitOutput nodes */
    for (unsigned int i = 0; i < track_list.size(); i++) {
      raw_write[i]->deinitOutput();
    }
    demux->deinitOutput();
    readfile->deinitOutput();
    /*End< */

    /*Begin> Deinit nodes */
    for (unsigned int i = 0; i < track_list.size(); i++) {
      raw_write[i]->deinit();
    }
    demux->deinit();
    readfile->deinit();
    /*End< */
    
    /*Begin> release nodes */
    if (readfile.get()) {
      app->getRegistry().releaseNode(*readfile);
    }
    if (demux.get()) {
      app->getRegistry().releaseNode(*demux);
    }
    for (unsigned int i = 0; i < raw_write.size(); i++) {
      if (raw_write[i]) {
	app->getRegistry().releaseNode(*raw_write[i]);
	delete raw_write[i];
      }
    }
    /*End< */
    
  } catch(Exception e) {
    cerr << e << endl;

    /*Begin> Release nodes */
    if (app != NULL) {
      if (readfile.get()) {
        app->getRegistry().releaseNode(*readfile);
      }
      if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
      }
      for (unsigned int i = 0; i < raw_write.size(); i++) {
	if (raw_write[i]) {
	  app->getRegistry().releaseNode(*raw_write[i]);
	  delete raw_write[i];
	}
      }
    }
    /*End< */
    
    return 1;
  } catch(...) {
    cerr << "Something went wrong..." << endl;

    /*Begin> Release nodes */
    if (app != NULL) {
      if (readfile.get()) {
        app->getRegistry().releaseNode(*readfile);
      }
      if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
      }
      for (unsigned int i = 0; i < raw_write.size(); i++) {
	if (raw_write[i]) {
	  app->getRegistry().releaseNode(*raw_write[i]);
	  delete raw_write[i];
	}
      }
    }
    /*End< */
    
    return 1;
  }
  
  return 0;
}
