/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


// general includes
#include "nmm/base/connect.hpp"  
#include "nmm/utils/utils.hpp"

// include the plugin manager
#include "nmm/base/NMMApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/base/INamed.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/general/IEncode.hpp"
#include "nmm/interfaces/video/IFramerateConverter.hpp"
#include "nmm/interfaces/video/IScaler.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <getopt.h>
#include <X11/Xlib.h>
#include <iostream>

using namespace NMM;

int usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options] [input-filename] [output-filename]" << endl
       << endl
       << "    filename:            mpeg ps file" << endl
       << "    -b --video_bitrate   Video bitrate (KBit/s)" << endl
       << "    -w --audio_bitrate   Audio bitrate (KBit/s)" << endl
       << "    -a --audiostream     Select audio stream number" << endl
       << "    -p --autoaspect      Scale video to adjust to aspect" << endl
       << "    -D --deinterlace     Deinterlace video" << endl
       << "    -t --writethrough    Write through audio" << endl
       << "    -d --delay           Set audio delay relative to video" << endl
       << "    -F --ffmpeg          Use ffmpeg library" << endl
       << "    -v --verbose         Verbose mode (lots of messages!)" << endl
       << "    -h --help            Shows this help and exits." << endl;
  return 1;
}

void turnOffAllMessages() {
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
}

int main(int argc, char ** argv) {
  string filename;
  string vidoutfilename;
  bool no_audio = false;
  bool use_ac3 = false;
  bool ffmpeg = false;
  bool verbose = false;
  bool use_deinterlacer = false;
  bool use_scaler = false;
  bool use_fpsconv = false;
  bool writethrough = false;
  bool mpeg_video_only = false;
  int video_bitrate = 900; // default value in kBit/s
  int audio_track = 0;
  int mp3_bitrate = 128;    // default value for mp3-bitrate in kBit/s
  float delay = 0;
  
  const char *optstring = "b:w:a:d:pDtFvh";
  struct option longopts[] = {
    {"vid_bitrate" , 1, NULL, 'b'},
    {"aud_bitrate" , 1, NULL, 'w'},
    {"audiostream" , 1, NULL, 'a'},
    {"delay"       , 1, NULL, 'd'},
    {"deinterlace" , 0, NULL, 'D'},
    {"autoaspect"  , 0, NULL, 'p'},
    {"writethrough", 0, NULL, 't'},
    {"ffmpeg"      , 0, NULL, 'F'},
    {"verbose"     , 0, NULL, 'v'},
    {"help"        , 0, NULL, 'h'},
    {NULL          , 0, NULL, 0  } /* last row must be zero */
  };
  
  int ch, optionindex;
  
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'b':
      if (optarg) {
        video_bitrate = atoi(optarg);
      }
      break;
    case 'w':
      if (optarg) {
        mp3_bitrate = atoi(optarg);
      }
      break;
    case 'a':
      if (optarg) {      
        audio_track = atoi(optarg);
      }
      break;
    case 'd':
      if (optarg) {      
        delay = atof(optarg);
        use_fpsconv = true;
      }
      break;
    case 'p':
      use_scaler = true;
      break;
    case 'D':
      use_deinterlacer = true;
      break;
    case 'F':
      ffmpeg = true;
      break;
    case 't':
      writethrough = true;
      break;
    case 'v':
      verbose = true;
      break;
    default: //includes 'h'
      return usage(argv[0]);
    }
  }
  
  // assure that 2 filenames are given
  if (argc != optind + 2) {
    cerr << "Make sure to provide input and output filename!" << endl;
    cerr << " argc = " << argc << "; optind = " << optind << endl;
    while (optind < argc) {
      cerr << "argv[" << optind << "] = " << argv[optind++] << endl;
    }
    return usage(argv[0]); 
  }
  
  // here argc is equal optind + 2
  filename= "";
  filename.append(argv[optind++]);
  vidoutfilename = "";
  vidoutfilename.append(argv[optind]);
  
  // eventually turn off debug messages
  if (!verbose) turnOffAllMessages();
  
  NMMApplication* app = 0;
  
  /*Begin> Allocate _var variables for the INodes */
  INode_var readfile;
  INode_var demux;
  INode_var mpg2yuv;
  INode_var deinterlacer;
  INode_var scaler;
  INode_var fpsconv;
  INode_var divx4enc;
  INode_var ac3_dec;
  INode_var mp3_dec;
  INode_var mp3_enc;
  INode_var av_mux;
  INode_var writefile;
  INode* audio_dec = NULL;
  /*End< */
  
  int error_level = 0;
  try {
    /*Begin> create the NMM application */
    app = ProxyApplication::getApplication(argc, argv);
    /*End< */
    
    /*Begin> Init requests for nodes */
    NodeDescription readfile_req("MPEGReadNode");
    NodeDescription demux_req("MPEGDemuxNode");
    NodeDescription videodec_req("MPEGVideoDecodeNode");
    NodeDescription deinterlacer_req("YUVDeInterlaceNode");
    NodeDescription scaler_req("VideoScalerNode");
    NodeDescription fpsconv_req("FramerateConverterNode");
    NodeDescription divx4enc_req(ffmpeg ? "FFMpegEncodeNode" : "DivX4EncodeNode");
    NodeDescription ac3dec_req(writethrough ? "AC3ParseNode" : "AC3DecodeNode");
    NodeDescription mp3dec_req("MPEGAudioDecodeNode");
    NodeDescription mp3enc_req("MPEGAudioEncodeNode");
    NodeDescription avmux_req("AVMuxNode");
    NodeDescription writer_req("AVIWriteNode");
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    {
      RegistryLock lock(registry);
      
      list<Response> readfile_resp = registry.initRequest(readfile_req);
      list<Response> demux_resp = registry.initRequest(demux_req);
      list<Response> videodec_resp = registry.initRequest(videodec_req);
      list<Response> deinterlacer_resp = registry.initRequest(deinterlacer_req);
      list<Response> scaler_resp = registry.initRequest(scaler_req);
      list<Response> fpsconv_resp = registry.initRequest(fpsconv_req);
      list<Response> divx4enc_resp = registry.initRequest(divx4enc_req);
      list<Response> ac3dec_resp = registry.initRequest(ac3dec_req);
      list<Response> mp3dec_resp = registry.initRequest(mp3dec_req);
      list<Response> mp3enc_resp = registry.initRequest(mp3enc_req);
      list<Response> avmux_resp = registry.initRequest(avmux_req);
      list<Response> writer_resp = registry.initRequest(writer_req);
      
      if (readfile_resp.empty() ||
          demux_resp.empty() ||
          videodec_resp.empty() ||
          deinterlacer_resp.empty() ||
          scaler_resp.empty() ||
          fpsconv_resp.empty() ||
          divx4enc_resp.empty() ||
          ac3dec_resp.empty() ||
          mp3dec_resp.empty() ||
          mp3enc_resp.empty() ||
          avmux_resp.empty() ||
          writer_resp.empty()) {
        GLOBAL_ERROR_STREAM("Size of readfile_resp " << readfile_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of demux_resp " << demux_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of videodec_resp " << videodec_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of deinterlacer_resp " << deinterlacer_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of scaler_resp " << scaler_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of fpsconv_resp " << fpsconv_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of divx4enc_resp " << divx4enc_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of ac3dec_resp " << ac3dec_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of mp3dec_resp " << mp3dec_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of mp3enc_resp " << mp3enc_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of avmux_resp " << avmux_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of writer_resp " << writer_resp.size() << endl);
        throw Exception("A required node is not available");
      }
      
      readfile.reset(registry.requestNode(readfile_resp.front()));
      demux.reset(registry.requestNode(demux_resp.front()));
      mpg2yuv.reset(registry.requestNode(videodec_resp.front()));
      deinterlacer.reset(registry.requestNode(deinterlacer_resp.front()));
      scaler.reset(registry.requestNode(scaler_resp.front()));
      fpsconv.reset(registry.requestNode(fpsconv_resp.front()));
      divx4enc.reset(registry.requestNode(divx4enc_resp.front()));
      ac3_dec.reset(registry.requestNode(ac3dec_resp.front()));
      mp3_dec.reset(registry.requestNode(mp3dec_resp.front()));
      mp3_enc.reset(registry.requestNode(mp3enc_resp.front()));
      av_mux.reset(registry.requestNode(avmux_resp.front()));
      writefile.reset(registry.requestNode(writer_resp.front()));
    }
    /*End< */
    
    
    /*Begin> init nodes */
    IFileHandler_var file_source(readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(filename);
    readfile->init();
    
    readfile->initOutput();
    
    string video_stream;
    string audio_stream;
    list<string> demux_video_streams;
    list<string> demux_audio_streams;
    
    no_audio = false;
    use_ac3 = false;
    
    if (readfile->getWorkingOutputProperty()->getDefaultFormat()->getFormatType() == Format::video_mpeg_type) {
      mpeg_video_only = true;
      no_audio = true;  // video only implies no audio!
      cout << "Pure MPEG Video stream. No Audio!" << endl;
      video_stream = string("default");
    } else {
      demux->init();
      GLOBAL_DEBUG_STREAM("Connecting readfile to demux..." << endl);
      connect(readfile, demux);
      
      readfile->activate();
      demux->initOutput();
      
      list<string> demux_streams = demux->getOutputStreamTags();
      demux_streams.sort();
      cout << "Available Output streams:" << endl;
      for (list<string>::const_iterator i = demux_streams.begin();
           i != demux_streams.end();
           i++) {
        cout << "\t" << *i << endl;
        const Format* format = demux->getWorkingOutputProperty()->getDefaultFormat(*i);
        if (audio_stream.empty()) {
          if (format->getFormatType() == Format::audio_ac3_type) {
            use_ac3 = true;
          }
        }
        if ((format->getFormatType() == Format::audio_ac3_type) || (format->getFormatType() == Format::audio_mpeg_type)) {
          demux_audio_streams.push_back(*i);
        }
        if (format->getFormatType() == Format::video_mpeg_type) {
          demux_video_streams.push_back(*i);
        }
      }
      
      if (demux_audio_streams.empty()) {
        no_audio = true;
      }
    }
    
    if (use_ac3) {
      audio_stream = string("ac3_audio") + audio_track;
    } else {
      audio_stream = string("mpeg_audio") + audio_track;
    }
    
    if (!mpeg_video_only) {
      IBiDirQueuedNode_var mpg2yuv_q(mpg2yuv->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
      mpg2yuv_q->setDownstreamMaxSize(100);
      mpg2yuv->init();
      connect(demux, string("mpeg_video0"), mpg2yuv);
    }
    
    if (!no_audio) {
      if (use_ac3) {
        if (!writethrough) {
          audio_dec = ac3_dec.get();
        } else {
          audio_dec = ac3_dec.get();
        }
      } else {
        if (!writethrough) {
          audio_dec = mp3_dec.get();
        }
      }
    }  
    
    IBiDirQueuedNode_var av_mux_q(av_mux->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    av_mux_q->setDownstreamMaxSize(100);
    av_mux->init();
    /*End< */
    
    /* Begin> Connect nodes */
    if (!no_audio) {
      if (audio_dec != NULL) {
        audio_dec->init();
        
        connect(demux.get(), audio_stream, audio_dec);
      } else {
        connect(demux, audio_stream, av_mux, string("audio"), NULL);
      }
    }
    
    if (!mpeg_video_only) {
      demux->activate();
    }
    
    if (!writethrough) {
      audio_dec->initOutput();
      mp3_enc->init();
      connect(audio_dec, mp3_enc.get());
      audio_dec->activate();
      
      mp3_enc->initOutput();
      Property_var mpa_wout_property(mp3_enc->getWorkingOutputProperty());
      const Format* def_mpa_format = mpa_wout_property->getDefaultFormat();
      Format* mpa_format = new Format("audio/mpeg");
      mpa_format->addIntParamValue(Format::layer_param, 3);
      mpa_format->addIntParamValue(Format::audio_codec_param, def_mpa_format->getIntValue(Format::audio_codec_param));
      mpa_format->addIntParamValue(Format::sample_rate_param, def_mpa_format->getIntValue(Format::sample_rate_param));
      mpa_format->addIntParamValue(Format::channels_param, def_mpa_format->getIntValue(Format::channels_param));
      mpa_format->addIntParamValue(Format::bitpersample_param, def_mpa_format->getIntValue(Format::bitpersample_param));
      mpa_format->addIntParamValue(Format::bitrate_param, mp3_bitrate * 1000);
      
      connect(mp3_enc, av_mux, string("audio"), mpa_format);
      mp3_enc->activate();
    } else if (audio_dec) {
      audio_dec->initOutput();
      connect(audio_dec, av_mux.get(), string("audio"));
      audio_dec->activate();
    }
    
    INode* raw_video_out = mpg2yuv.get();
    if (use_deinterlacer) {
      raw_video_out->initOutput();
      deinterlacer->init();
      connect(raw_video_out, deinterlacer.get());
      raw_video_out->activate();
      raw_video_out = deinterlacer.get();
    }
    
    if (use_fpsconv) {
      raw_video_out->initOutput();
      
      delay = -delay;
      int sec = static_cast<int>(delay);
      int msec = static_cast<int>(((delay - sec) * 1000000000));
      
      IFramerateConverter_var ifpsconv(fpsconv->getParentObject()->getCheckedInterface<IFramerateConverter>());
      ifpsconv->setTimeOffset(Interval(sec, msec));
      fpsconv->init();
      
      connect(raw_video_out, fpsconv.get());
      raw_video_out->activate();
      raw_video_out = fpsconv.get();      
    }
    
    if (use_scaler) {
      raw_video_out->initOutput();
      scaler->init();
      connect(raw_video_out, scaler.get());
      raw_video_out->activate();
      raw_video_out = scaler.get();
    }
    
    IBiDirQueuedNode_var divx4enc_q(divx4enc->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    divx4enc_q->setDownstreamMaxSize(100);
    divx4enc->init();
    raw_video_out->initOutput();
    connect(raw_video_out, divx4enc.get());
    raw_video_out->activate();
    
    divx4enc->initOutput();
    Property_var divx4_wout_property(divx4enc->getWorkingOutputProperty());
    const Format* def_divx4_format = divx4_wout_property->getDefaultFormat();
    Format* divx4_format = new Format("video/mpeg4");
    divx4_format->addStringParamValue(Format::compressor_param, def_divx4_format->getStringValue(Format::compressor_param));
    divx4_format->addFloatParamValue(Format::framerate_param, def_divx4_format->getFloatValue(Format::framerate_param));
    divx4_format->addIntParamValue(Format::x_resolution_param, def_divx4_format->getIntValue(Format::x_resolution_param));
    divx4_format->addIntParamValue(Format::y_resolution_param, def_divx4_format->getIntValue(Format::y_resolution_param));
    divx4_format->addIntParamValue(Format::ratio_x_param, def_divx4_format->getIntValue(Format::ratio_x_param));
    divx4_format->addIntParamValue(Format::ratio_y_param, def_divx4_format->getIntValue(Format::ratio_y_param));
    divx4_format->addIntParamValue(Format::bitrate_param, video_bitrate * 1000);
    
    connect(divx4enc, av_mux, string("video"), divx4_format);
    divx4enc->activate();
    
    ISinkFileHandler_var file_sink(writefile->getParentObject()->getCheckedInterface<ISinkFileHandler>());
    file_sink->setFilename(vidoutfilename);
    writefile->init();
    
    av_mux->initOutput();
    connect(av_mux, writefile);
    av_mux->activate();
    
    writefile->initOutput();
    writefile->activate();
    /*End< */
    
    /*Begin> Start nodes */
    if (readfile.get() && readfile->isActivated()) {
      readfile->start();
    }
    if (demux.get() && demux->isActivated()) {
      demux->start();
    }
    if (mpg2yuv.get() && mpg2yuv->isActivated()) {
      mpg2yuv->start();
    }
    if (divx4enc.get() && divx4enc->isActivated()) {
      divx4enc->start();
    }
    if (deinterlacer.get() && deinterlacer->isActivated()) {
      deinterlacer->start();
    }
    if (scaler.get() && scaler->isActivated()) {
      scaler->start();
    }
    if (fpsconv.get() && fpsconv->isActivated()) {
      fpsconv->start();
    }
    if (audio_dec && audio_dec->isActivated()) {
      audio_dec->start();
    }
    if (mp3_enc.get() && mp3_enc->isActivated()) {
      mp3_enc->start();
    }
    if (av_mux.get() && av_mux->isActivated()) {
      av_mux->start();
    }
    if (writefile.get() && writefile->isActivated()) {
      writefile->start();
    }
    /*End< */
    
    /* Run until a key - and enter - is pressed */
    getchar();
    
  } catch( const Exception& e ) {
    
    cerr << e << endl;
    error_level = 1;
    
  } catch( ... ) {
    
    cerr << "Something strange happened..." << endl;
    return 2;
    
  }
  
  try {
    /*Begin> Stop nodes */
    if (readfile.get() && readfile->isStarted()) {
      readfile->stop();
    }
    if (demux.get() && demux->isStarted()) {
      demux->stop();
    }
    if (mpg2yuv.get() && mpg2yuv->isStarted()) {
      mpg2yuv->stop();
    }
    if (divx4enc.get() && divx4enc->isStarted()) {
      divx4enc->stop();
    }
    if (deinterlacer.get() && deinterlacer->isStarted()) {
      deinterlacer->stop();
    }
    if (scaler.get() && scaler->isStarted()) {
      scaler->stop();
    }
    if (fpsconv.get() && fpsconv->isStarted()) {
      fpsconv->stop();
    }
    if (audio_dec && audio_dec->isStarted()) {
      audio_dec->stop();
    }
    if (mp3_enc.get() && mp3_enc->isStarted()) {
      mp3_enc->stop();
    }
    if (av_mux.get() && av_mux->isStarted()) {
      av_mux->stop();
    }
    if (writefile.get() && writefile->isStarted()) {
      writefile->stop();
    }
    /*End< */

    /*Begin> Flush nodes */
    if (readfile.get() && readfile->isActivated()) {
      readfile->flush();
    }
    if (demux.get() && demux->isActivated()) {
      demux->flush();
    }
    if (mpg2yuv.get() && mpg2yuv->isActivated()) {
      mpg2yuv->flush();
    }
    if (divx4enc.get() && divx4enc->isActivated()) {
      divx4enc->flush();
    }
    if (deinterlacer.get() && deinterlacer->isActivated()) {
      deinterlacer->flush();
    }
    if (scaler.get() && scaler->isActivated()) {
      scaler->flush();
    }
    if (fpsconv.get() && fpsconv->isActivated()) {
      fpsconv->flush();
    }
    if (audio_dec && audio_dec->isActivated()) {
      audio_dec->flush();
    }
    if (mp3_enc.get() && mp3_enc->isActivated()) {
      mp3_enc->flush();
    }
    if (av_mux.get() && av_mux->isActivated()) {
      av_mux->flush();
    }
    if (writefile.get() && writefile->isActivated()) {
      writefile->flush();
    }
    /*End< */
    
    /*Begin> Disconnect nodes */
    if (writefile.get()) {
      switch(writefile->getState()) {
      case INode::ACTIVATED:
        writefile->deactivate();
      case INode::OUTPUT_INITIALIZED:
        writefile->deinitOutput();
      case INode::INITIALIZED:
        {
          if (writefile->isInputConnected("default")) {
            if (av_mux.get() && av_mux->isActivated()) {
              av_mux->deactivate();
            }
            disconnect_input(writefile, "default");
          }
          writefile->deinit();
        }
      default:
        break;
      }
    }
    if (av_mux.get()) {
      switch(av_mux->getState()) {
      case INode::ACTIVATED:
        av_mux->deactivate();
      case INode::OUTPUT_INITIALIZED:
        av_mux->deinitOutput();
      case INode::INITIALIZED:
        {
          if (av_mux->isInputConnected("audio")) {
            if (!writethrough) {
              if (mp3_enc.get() && mp3_enc->isActivated()) {
                mp3_enc->deactivate();
              }
            } else {
              if (audio_dec && audio_dec->isActivated()) {
                audio_dec->deactivate();
              }
            }
            disconnect_input(av_mux, "audio");
          }
          if (av_mux->isInputConnected("video")) {
            if (divx4enc.get() && divx4enc->isActivated()) {
              divx4enc->deactivate();
            }
            disconnect_input(av_mux, "video");
          }
          av_mux->deinit();
        }
      default:
        break;
      }
    }
    if (mp3_enc.get()) {
      switch(mp3_enc->getState()) {
      case INode::ACTIVATED:
        mp3_enc->deactivate();
      case INode::OUTPUT_INITIALIZED:
        mp3_enc->deinitOutput();
      case INode::INITIALIZED:
        {
          if (mp3_enc->isInputConnected("default")) {
            if (audio_dec && audio_dec->isActivated()) {
              audio_dec->deactivate();
            }
            disconnect_input(mp3_enc, "default");
          }
          mp3_enc->deinit();
        }
      default:
        break;
      }
    }
    if (divx4enc.get()) {
      switch(divx4enc->getState()) {
      case INode::ACTIVATED:
        divx4enc->deactivate();
      case INode::OUTPUT_INITIALIZED:
        divx4enc->deinitOutput();
      case INode::INITIALIZED:
        {
          if (divx4enc->isInputConnected("default")) {
            if (use_scaler) {
              if (scaler.get() && scaler->isActivated()) {
                scaler->deactivate();
              }
            } else if (use_fpsconv) {
              if (fpsconv.get() && fpsconv->isActivated()) {
                fpsconv->deactivate();
              }
            } else if (use_deinterlacer) {
              if (deinterlacer.get() && deinterlacer->isActivated()) {
                deinterlacer->deactivate();
              }
            } else {
              if (mpg2yuv.get() && mpg2yuv->isActivated()) {
                mpg2yuv->deactivate();
              }
            }
            disconnect_input(divx4enc, "default");
          }
          divx4enc->deinit();
        }
      default:
        break;
      }
    }
    if (scaler.get()) {
      switch(scaler->getState()) {
      case INode::ACTIVATED:
        scaler->deactivate();
      case INode::OUTPUT_INITIALIZED:
        scaler->deinitOutput();
      case INode::INITIALIZED:
        {
          if (scaler->isInputConnected("default")) {
            if (use_fpsconv) {
              if (fpsconv.get() && fpsconv->isActivated()) {
                fpsconv->deactivate();
              }
            } else if (use_deinterlacer) {
              if (deinterlacer.get() && deinterlacer->isActivated()) {
                deinterlacer->deactivate();
              }
            } else {
              if (mpg2yuv.get() && mpg2yuv->isActivated()) {
                mpg2yuv->deactivate();
              }
            }
            disconnect_input(scaler, "default");
          }
          scaler->deinit();
        }
      default:
        break;
      }
    }
    if (fpsconv.get()) {
      switch(fpsconv->getState()) {
      case INode::ACTIVATED:
        fpsconv->deactivate();
      case INode::OUTPUT_INITIALIZED:
        fpsconv->deinitOutput();
      case INode::INITIALIZED:
        {
          if (fpsconv->isInputConnected("default")) {
            if (use_deinterlacer) {
              if (deinterlacer.get() && deinterlacer->isActivated()) {
                deinterlacer->deactivate();
              }
            } else {
              if (mpg2yuv.get() && mpg2yuv->isActivated()) {
                mpg2yuv->deactivate();
              }
            }
            disconnect_input(fpsconv, "default");
          }
          fpsconv->deinit();
        }
      default:
        break;
      }
    }
    if (deinterlacer.get()) {
      switch(deinterlacer->getState()) {
      case INode::ACTIVATED:
        deinterlacer->deactivate();
      case INode::OUTPUT_INITIALIZED:
        deinterlacer->deinitOutput();
      case INode::INITIALIZED:
        {
          if (deinterlacer->isInputConnected("default")) {
            if (mpg2yuv.get() && mpg2yuv->isActivated()) {
              mpg2yuv->deactivate();
            }
            disconnect_input(deinterlacer, "default");
          }
          deinterlacer->deinit();
        }
      default:
        break;
      }
    }
    if (mpg2yuv.get()) {
      switch(mpg2yuv->getState()) {
      case INode::ACTIVATED:
        mpg2yuv->deactivate();
      case INode::OUTPUT_INITIALIZED:
        mpg2yuv->deinitOutput();
      case INode::INITIALIZED:
        {
          if (mpg2yuv->isInputConnected("default")) {
            if (!mpeg_video_only) {
              if (demux.get() && demux->isActivated()) {
                demux->deactivate();
              }
            } else {
              if (readfile.get() && readfile->isActivated()) {
                readfile->deactivate();
              }
            }
            disconnect_input(mpg2yuv, "default");
          }
          mpg2yuv->deinit();
        }
      default:
        break;
      }
    }
    if (audio_dec) {
      switch(audio_dec->getState()) {
      case INode::ACTIVATED:
        audio_dec->deactivate();
      case INode::OUTPUT_INITIALIZED:
        audio_dec->deinitOutput();
      case INode::INITIALIZED:
        {
          if (audio_dec->isInputConnected("default")) {
            if (demux.get() && demux->isActivated()) {
              demux->deactivate();
            }
            disconnect_input(audio_dec, "default");
          }
          audio_dec->deinit();
        }
      default:
        break;
      }
    }
    if (demux.get()) {
      switch(demux->getState()) {
      case INode::ACTIVATED:
        demux->deactivate();
      case INode::OUTPUT_INITIALIZED:
        demux->deinitOutput();
      case INode::INITIALIZED:
        {
          if (demux->isInputConnected("default")) {
            if (readfile.get() && readfile->isActivated()) {
              readfile->deactivate();
            }
            disconnect_input(demux, "default");
          }
          demux->deinit();
        }
      default:
        break;
      }
    }
    if (readfile.get()) {
      switch(demux->getState()) {
      case INode::ACTIVATED:
        readfile->deactivate();
      case INode::OUTPUT_INITIALIZED:
        readfile->deinitOutput();
      case INode::INITIALIZED:
        readfile->deinit();
      default:
        break;
      }
    }
    /*End< */
    
    /*Begin> Release nodes */
    if (app != NULL) {
      if (readfile.get()) {
        app->getRegistry().releaseNode(*readfile);
      }
      if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
      }
      if (mpg2yuv.get()) {
        app->getRegistry().releaseNode(*mpg2yuv);
      }
      if (divx4enc.get()) {
        app->getRegistry().releaseNode(*divx4enc);
      }
      if (deinterlacer.get()) {
        app->getRegistry().releaseNode(*deinterlacer);
      }
      if (ac3_dec.get()) {
        app->getRegistry().releaseNode(*ac3_dec);
      }
      if (mp3_dec.get()) {
        app->getRegistry().releaseNode(*mp3_dec);
      }
      if (mp3_enc.get()) {
        app->getRegistry().releaseNode(*mp3_enc);
      }
      if (av_mux.get()) {
        app->getRegistry().releaseNode(*av_mux);
      }
      if (writefile.get()) {
        app->getRegistry().releaseNode(*writefile);
      }
    }
    /*End< */
    
    return error_level;
    
  } catch( const Exception e ) {
    
    cerr << "Exception during cleanup: " << endl;
    cerr << e << endl;
    return 2;
    
  } catch(...) {
    
    cerr << "Something strange happened during cleanup..." << endl;
    return 2;
    
  }
}
