/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

// C includes
#include <getopt.h>

#include <iostream>

#include <iostream>

// general includes
#include "nmm/base/connect.hpp"

// include the plugin manager
#include "nmm/base/NMMApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

// include the synchronizer
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"

// include used interfaces
#include "nmm/interfaces/video/IVideoDelay.hpp"
#include "nmm/interfaces/video/display/IVideoDisplay.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/video/display/IDisplayEventSender.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/ISeekable.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/base/IBufferManager.hpp"


using namespace NMM;

class EventCatcher {
private:
  bool finished;
  bool pause;
  bool fullscreen;
  int sec_pos;

  ThreadMutex finished_mutex;
  ThreadCondition finished_cv;
  
  IVideoDisplay* video_display;
  IMultiAudioVideoSynchronizer* sync;
  ISeekable* seekable;
  
public:
  EventCatcher() : finished(false),
                   pause(false),
                   fullscreen(false),
		   sec_pos(0),
                   finished_cv(finished_mutex),
                   video_display(NULL),
                   sync(NULL),
		   seekable(NULL) {
  }
  
  void setVideoDisplay(IVideoDisplay* value) {
    video_display = value;
    if (video_display) {
      fullscreen = video_display->getFullscreen() && video_display->hasFullscreen();
    }
  }
  
  void setSynchronizer(IMultiAudioVideoSynchronizer* value) {
    sync = value;
  }

  void setSeekable(ISeekable* value) {
    seekable = value;
  }

  void setSecPos(int sec) {
    sec_pos = sec;
  }
  
  Result endTrack() {
    cout << "End of file" << endl;
    signalExit();
    return SUCCESS;
  }
  
  Result keyReleased(string& key) {
    
    // if paused, any key should wakeup
    if (pause && sync) {
      pause = false;
      sync->wakeup();
      return SUCCESS;
    }

    if (key == "KEY_q") { // 'q' to exit
      cerr << "Quit" << endl;
      signalExit();

    } else if (key == "KEY_f") { // 'f' to toggle fullscreen
      if (!video_display) {
	cerr << "EventHandler: No VideoDisplay set!" << endl;
	return FAILURE;
      }
      if (video_display->hasFullscreen()) {
	fullscreen = !fullscreen;
	video_display->setFullscreen(fullscreen);
      } else {
	cerr << "Fullscreen not supported with current display!" << endl;
      }

    } else if ((key == "KEY_space") || (key == "KEY_p")) { // <space> or 'p' to pause, any key wakes up, see above
      if (!sync) {
	cerr << "EventHandler: No Synchronizer set!" << endl;
	return FAILURE;
      }
      pause = true;
      sync->pause();

    } else if (key == "KEY_Right") {
      sec_pos += 10;
      seekable->seekTimeTo(Time(sec_pos,0));
    } else if (key == "KEY_Left") {
      sec_pos -= 10;
      seekable->seekTimeTo(Time(sec_pos,0));
    } else {
      cerr << "Unsupported key pressed: " << key << endl;
    }

    return SUCCESS;
  }
  
  void signalExit() {
    MutexGuard guard(finished_mutex);
    finished = true;
    finished_cv.notifyAll();
  }
  
  void handleEvents() {
    finished_mutex.lock();
    while (!finished) {
      finished_cv.wait();
    }
    finished_mutex.unlock();
  }
};


int usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options] [filename]" << endl
       << endl
       << "    filename:            mpeg ps file" << endl
       << "    -f --fullscreen      Fullscreen mode" << endl
       << "    -a --ac3passthrough  send ac3 data directly to soundcard" << endl
       << "    -s --silent          Disable audio output" << endl
       << "    -j --jumptosecond    begin playback at given second" << endl
       << "    -d --delay:          Set video delay" << endl
       << "    -F --ffmpeg:         Use ffmpeg library" << endl
       << "    -l --local           Force using local registry" << endl
       << "    -r --registry        Add a registry host" << endl
       << "    -v --verbose:        Verbose mode (lots of messages!)" << endl
       << "    -h --help:           Shows this help and exits." << endl;
  return 1;
}


int main(int argc, char ** argv) {
  const int DEFAULT_REGISTRY_PORT = ClientRegistry::default_port;

  // global options
  bool fullscreen = false;
  bool ac3_passthrough = false;
  int verbose = 0;
  bool silent = false;
  bool ffmpeg = false;
  bool local = false;
  string filename;
  int delay = 0;
  int begin_second = 0;
  vector<string> reghosts;
  vector<int> regports;

  const char *optstring = "fasj:d:Flr:vh";
  struct option longopts[] = {
    {"fullscreen"    , 0, NULL, 'f'},
    {"ac3passthrough", 0, NULL, 'a'},
    {"silent"        , 0, NULL, 's'},
    {"jumptosecond"  , 1, NULL, 'j'},
    {"delay"         , 1, NULL, 'd'},
    {"ffmpeg"        , 0, NULL, 'F'},
    {"local"         , 0, NULL, 'l'},
    {"registry"      , 1, NULL, 'r'},
    {"verbose"       , 0, NULL, 'v'},
    {"help"          , 0, NULL, 'h'},
    {NULL            , 0, NULL, 0  } // last row must be zero
  };
  
  int ch, optionindex;
  
  if (argc == optind) { return usage(argv[0]); } // if no filename given, print help
  
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'F':
      ffmpeg = true;
      break;
    case 's':
      silent = true;
      break;
    case 'j':
      if (optarg) {
        begin_second = atoi(optarg);
      }
      else begin_second = 0;      
      break;
    case 'f':
      fullscreen = true;
      break;
    case 'a':
      ac3_passthrough = true;
      break;
    case 'r':
      if (optarg) {
        const char *split = strchr(optarg, ':');
        if (split) {
          reghosts.push_back(string(optarg, split-optarg));
          regports.push_back(atoi(split+1));
        } else {
          reghosts.push_back(string(optarg));
          regports.push_back(DEFAULT_REGISTRY_PORT);
        }
      }
      break;
    case 'l':
      local = true;
      break;
    case 'v':
      verbose++;
      break;
    case 'd':
      if (optarg) {
        delay = atoi(optarg);
      }
      else delay = 0;
      break;
    default:
      return usage(argv[0]);
    }
  }
  
  if (optind < argc) {
    filename= "";
    filename.append(argv[optind]);  
  } else {
    return usage(argv[0]);
  }
  
  /* handle *Stream output */
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }

  NMMApplication* app = 0;
  
  /*Begin> Allocate _var variables for the INodes */
  INode_var readfile;
  INode_var demux;
  INode_var divx4dec;
  INode_var display;
  INode_var mp3_dec;
  INode_var ac3_dec;
  INode_var audio_play;
  /*End< */
  
  INode* audio_dec = NULL;
  
  /* Allocate an event listener */
  EventCatcher ec;
  
  int error_level = 0;
  try {
    /*Begin> create the NMM application */
    app = ProxyApplication::getApplication(argc, argv);
    
    /*Begin> Add remote registry servers */
    ClientRegistry& registry = app->getRegistry();
    for (unsigned int i=0; i<reghosts.size(); ++i) {
      GLOBAL_DEBUG_STREAM("Add registry host " << reghosts[i] << ":" << regports[i] << endl);
      registry.addRegistry(TCPAddress(reghosts[i], regports[i]));
    }
    /*End< */
    
    /*Begin> Init requests for nodes */
    NodeDescription request1;
    request1.setNodeName("AVIReadNode");
    
    NodeDescription request2;
    request2.setNodeName("AVDemuxNode");
    
    NodeDescription request3;
    request3.setNodeName(ffmpeg ? "FFMpegDecodeNode" : "DivX4DecodeNode");
    
    NodeDescription request4;
    request4.addInterface(IVideoDisplay::IVideoDisplay_type);
    request4.addInterface(IDisplayEventSender::IDisplayEventSender_type);
    
    NodeDescription request5;
    request5.setNodeName("MPEGAudioDecodeNode");
    
    NodeDescription request6;
    request6.setNodeName("AC3DecodeNode");
    
    NodeDescription request7;
    request7.setNodeType(INode::SINK);
    request7.addInterface(IAudioDevice::IAudioDevice_type);
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    {
      RegistryLock lock(registry);
      
      list<Response> response1 = registry.initRequest(request1);
      list<Response> response2 = registry.initRequest(request2);
      list<Response> response3 = registry.initRequest(request3);
      list<Response> response4 = registry.initRequest(request4);
      list<Response> response5 = registry.initRequest(request5);
      list<Response> response6 = registry.initRequest(request6);
      list<Response> response7 = registry.initRequest(request7);
      
      if (response1.empty() ||
          response2.empty() ||
          response3.empty() ||
          response4.empty() ||
          response5.empty() ||
          response6.empty() ||
          response7.empty()) {
        GLOBAL_ERROR_STREAM("Size of response1 " << response1.size() << endl);
        GLOBAL_ERROR_STREAM("Size of response2 " << response2.size() << endl);
        GLOBAL_ERROR_STREAM("Size of response3 " << response3.size() << endl);
        GLOBAL_ERROR_STREAM("Size of response4 " << response4.size() << endl);
        GLOBAL_ERROR_STREAM("Size of response5 " << response5.size() << endl);
        GLOBAL_ERROR_STREAM("Size of response6 " << response6.size() << endl);
        GLOBAL_ERROR_STREAM("Size of response7 " << response7.size() << endl);
        throw Exception("A required node is not available");
      }
      
      readfile.reset(registry.requestNode(response1.front()));
      demux.reset(registry.requestNode(response2.front()));
      divx4dec.reset(registry.requestNode(response3.front()));
      display.reset(registry.requestNode(response4.front()));
      mp3_dec.reset(registry.requestNode(response5.front()));
      ac3_dec.reset(registry.requestNode(response6.front()));
      audio_play.reset(registry.requestNode(response7.front()));
    }
    /*End< */
    
    /*Begin> Init nodes */
    IFileHandler_var file_source(readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(filename);
    readfile->init();
    
    demux->init();
    
    IVideoDelay_var video_delay(divx4dec->getParentObject()->getInterface<IVideoDelay>());
    if (video_delay.get()) {
      // ffmpeg doesn't have this interface
      video_delay->setVideoDelay(delay);
    }
    IBiDirQueuedNode_var divx4dec_q(divx4dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    divx4dec_q->setDownstreamMaxSize(100);
    divx4dec->init();
    
    IVideoDisplay_var ivideo_display(display->getParentObject()->getInterface<IVideoDisplay>());
    if (ivideo_display.get()) {
      ivideo_display->setFullscreen(fullscreen);
    }
    IBiDirQueuedNode_var display_queue(display->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    display_queue->setDownstreamMaxSize(1);
    display->init();
    
    IAudioDevice_var audio_device(audio_play->getParentObject()->getCheckedInterface<IAudioDevice>());
    audio_device->setDevice("/dev/dsp");
    /*End< */
    
    /*Begin> Register event catcher */
    display->getParentObject()->
      registerEventListener(ITrack::endTrack_event, new TEDObject0<EventCatcher>(&ec, &EventCatcher::endTrack));
    display->getParentObject()->
      registerEventListener(IDisplayEvents::keyReleased_event,
			    new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    /*End< */
    
    /*Begin> Init synchronization */
    MultiAudioVideoSynchronizer av_sync;
    IMultiAudioVideoSynchronizer_var sync(av_sync.getCheckedInterface<IMultiAudioVideoSynchronizer>());
    
    ISynchronizedSink_var audio_sync(audio_play->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    ISynchronizedSink_var video_sync(display->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    ISinkController_var audio_controller(audio_sync->getController());
    ISinkController_var video_controller(video_sync->getController());
    sync->addAudioSinkController(audio_controller.get());
    sync->addVideoSinkController(video_controller.get());
    /*End< */
    
    /*Begin> Check audio format */
    readfile->initOutput();

    ISeekable_var read_seekable;
    read_seekable.reset(readfile->getParentObject()->getCheckedInterface<ISeekable>());
    ec.setSeekable(read_seekable.get());
    if (begin_second != 0) {
      ec.setSecPos(begin_second);
      read_seekable->seekTimeTo(Time(begin_second,0));
    }

    bool no_audio = silent;
    bool use_ac3 = false;
    if (!no_audio) {
      Property_var file_property(readfile->getWorkingOutputProperty());
      const Format* file_format = file_property->getDefaultFormat();
      if (file_format->isParameterIncluded(Format::audio_format_param)) {
        Format* audio_format = file_format->getFPointerValue(Format::audio_format_param);
        if (audio_format->getFormatType() == "audio/ac3") {
          use_ac3 = true;
        }
      } else {
        no_audio = true;
      }
    }
    
    if (!no_audio) {
      if (use_ac3) {
        if (!ac3_passthrough) {
          audio_dec = ac3_dec.get();
        }
      } else {
        audio_dec = mp3_dec.get();
      }
    }
    /*End< */
    
    /*Begin> Connect nodes */
    connect(readfile, demux);
    
    readfile->activate();
    demux->initOutput();
    if (audio_dec != NULL) {
      IBiDirQueuedNode_var audio_dec_q(audio_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
      audio_dec_q->setDownstreamMaxSize(100);
      audio_dec->init();
      connect(demux.get(), string("audio"), audio_dec);
    } else {
      IBiDirQueuedNode_var audio_play_q(audio_play->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
      audio_play_q->setDownstreamMaxSize(100);
      audio_play->init();
      connect(demux, string("audio"), audio_play);
    }
    
    connect(demux, string("video"), divx4dec);
    
    demux->activate();
    if (audio_dec != NULL) {
      audio_dec->initOutput();
      audio_play->init();
      connect(audio_dec, audio_play.get());
      
      audio_dec->activate();
    }
    
    if (!no_audio) {
      audio_play->initOutput();
      audio_play->activate();
    }
    
    divx4dec->initOutput();
    connect(divx4dec, display);
    divx4dec->activate();
    
    display->initOutput();
    display->activate();
    /*End< */
    
    IBufferManager_var bm1(display->getParentObject()->getCheckedInterface<IBufferManager>());
    IBufferManager_var bm2(divx4dec->getParentObject()->getCheckedInterface<IBufferManager>());
    bm2->useBufferManager(*bm1.get());

    /*Begin> Start nodes */
    readfile->start();
    demux->start();
    divx4dec->start();
    display->start();
    if (!no_audio) {
      if (audio_dec != NULL) {
        audio_dec->start();
      }
      audio_play->start();
    }
    /*End< */
    
    /*Begin> Event loop */
    IDisplayEventSender_var idisplay_event_sender(display->getParentObject()->getInterface<IDisplayEventSender>());
    if (display->isStarted() && idisplay_event_sender.get()) {
      idisplay_event_sender->enableDisplayEvents();
    }
    ec.setVideoDisplay(ivideo_display.get());
    ec.setSynchronizer(sync.get());
    ec.handleEvents();
    if (display->isStarted() && idisplay_event_sender.get()) {
      idisplay_event_sender->disableDisplayEvents();
    }
    /*End< */
    
  } catch(const Exception& e) {
    
    GLOBAL_ERROR_STREAM(e << endl);
    error_level = 1;
    
  } catch(...) {
    
    cerr << "Something strange went wrong.." << endl;
    return 2;
    
  }
  
  try {
    
    /*Begin> Stop nodes */
    if (readfile.get() && readfile->isStarted()) {
      readfile->stop();
    }
    if (demux.get() && demux->isStarted()) {
      demux->stop();
    }
    if (divx4dec.get() && divx4dec->isStarted()) {
      divx4dec->stop();
    }
    if (display.get() && display->isStarted()) {
      display->stop();
    }
    if (audio_dec && audio_dec->isStarted()) {
      audio_dec->stop();
    }
    if (audio_play.get() && audio_play->isStarted()) {
      audio_play->stop();
    }
    /*End< */
    
    /*Begin> Flush nodes */
    if (readfile.get() && readfile->isActivated()) {
      readfile->flush();
    }
    if (demux.get() && demux->isActivated()) {
      demux->flush();
    }
    if (divx4dec.get() && divx4dec->isActivated()) {
      divx4dec->flush();
    }
    if (display.get() && display->isActivated()) {
      display->flush();
    }
    if (audio_dec && audio_dec->isActivated()) {
      audio_dec->flush();
    }
    if (audio_play.get() && audio_play->isActivated()) {
      audio_play->flush();
    }
    /*End< */
    
    /*Begin> Disconnect nodes */
    if (audio_play.get()) {
      if (audio_play->isActivated()) {
        audio_play->deactivate();
      }
      if (audio_play->isOutputInitialized()) {
        audio_play->deinitOutput();
      }
      if (audio_play->isInitialized()) {
        IInputJack_var audio_play_in_jack(audio_play->getInputStream("default"));
        if (audio_play_in_jack.get() && audio_play_in_jack->isConnected()) {
          if (audio_dec) {
            if (audio_dec->isActivated()) {
              audio_dec->deactivate();
            }
          } else {
            if (demux.get() && demux->isActivated()) {
              demux->deactivate();
            }
          }
          disconnect_input(audio_play, "default");
        }
        audio_play->deinit();
      }
    }
    if (audio_dec) {
      if (audio_dec->isActivated()) {
        audio_dec->deactivate();
      }
      if (audio_dec->isOutputInitialized()) {
        audio_dec->deinitOutput();
      }
      if (audio_dec->isInitialized()) {
        IInputJack_var audio_dec_in_jack(audio_dec->getInputStream("default"));
        if (audio_dec_in_jack.get() && audio_dec_in_jack->isConnected()) {
          if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(audio_dec, "default");
        }
        audio_dec->deinit();
      }
    }
    if (display.get()) {
      if (display->isActivated()) {
        display->deactivate();
      }
      if (display->isOutputInitialized()) {
        display->deinitOutput();
      }
      if (display->isInitialized()) {
        IInputJack_var display_in_jack(display->getInputStream("default"));
        if (display_in_jack.get() && display_in_jack->isConnected()) {
          if (divx4dec.get() && divx4dec->isActivated()) {
            divx4dec->deactivate();
          }
          disconnect_input(display, "default");
        }
        display->deinit();
      }
    }
    if (divx4dec.get()) {
      if (divx4dec->isActivated()) {
        divx4dec->deactivate();
      }
      if (divx4dec->isOutputInitialized()) {
        divx4dec->deinitOutput();
      }
      if (divx4dec->isInitialized()) {
        IInputJack_var divx4dec_in_jack(divx4dec->getInputStream("default"));
        if (divx4dec_in_jack.get() && divx4dec_in_jack->isConnected()) {
          if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(divx4dec, "default");
        }
        divx4dec->deinit();
      }
    }
    if (demux.get()) {
      if (demux->isActivated()) {
        demux->deactivate();
      }
      if (demux->isOutputInitialized()) {
        demux->deinitOutput();
      }
      if (demux->isInitialized()) {
        IInputJack_var demux_in_jack(demux->getInputStream("default"));
        if (demux_in_jack.get() && demux_in_jack->isConnected()) {
          if (readfile.get() && readfile->isActivated()) {
            readfile->deactivate();
          }
          disconnect_input(demux, "default");
        }
        demux->deinit();
      }
    }
    if (readfile.get()) {
      if (readfile->isActivated()) {
        readfile->deactivate();
      }
      if (readfile->isOutputInitialized()) {
        readfile->deinitOutput();
      }
      if (readfile->isInitialized()) {
        readfile->deinit();
      }
    }
    /*End< Disconnect nodes */
    
    /*Begin> Release nodes */
    if (readfile.get()) {
      app->getRegistry().releaseNode(*readfile);
    }
    if (demux.get()) {
      app->getRegistry().releaseNode(*demux);
    }
    if (divx4dec.get()) {
      app->getRegistry().releaseNode(*divx4dec);
    }
    if (display.get()) {
      app->getRegistry().releaseNode(*display);
    }
    if (mp3_dec.get()) {
      app->getRegistry().releaseNode(*mp3_dec);
    }
    if (ac3_dec.get()) {
      app->getRegistry().releaseNode(*ac3_dec);
    }
    if (audio_play.get()) {
      app->getRegistry().releaseNode(*audio_play);
    }
    /*End< */
    
  } catch(Exception e) {
    
    cerr << "Exception during cleanup" << endl;
    cerr << e << endl;
    return 2;
    
  } catch(...) {
    
    cerr << "Something strange went wrong during cleanup..." << endl;
    return 2;
    
  }
  
  return error_level;
}
