/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <getopt.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <iostream>

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/device/dvd/IDVD.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/video/IScaler.hpp"
#include "nmm/interfaces/av/mpeg/IMPEGDemux.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"

// include the registry
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/utils/utils.hpp"

#include "nmm/base/ProxyObject.hpp"
// console producer
#include "nmm/misc/console/ConsoleProducer.hpp"

// import namespace
using namespace NMM;


/**
 * Class for receiving events.
 */
class AppEventReceiver : public IEventReceiver {
private:
  EventDispatcher* dispatcher;
  
public:
  AppEventReceiver() {
    dispatcher = new EventDispatcher();
  }
  
  virtual ~AppEventReceiver() {
    delete dispatcher;
  }
  
  EventDispatcher& getEventDispatcher() const {
    return *dispatcher;
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) const {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) const {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
protected:
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) {
    return dispatcher->dispatch(event, sender);
  }
  
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) const {
    return dispatcher->dispatch(event, sender);
  }
};


/**
 * Class for handling events.
 * An instance of this class is registered at some nodes as an event listener.
 */
class EventCatcher {
private:
  bool exit;
  ThreadMutex exit_mutex;
  ThreadCondition exit_cv;
  
  void signalExit() {
    MutexGuard em(exit_mutex);
    exit = true;
    exit_cv.notifyAll();
  }
  
public:
  EventCatcher() 
    : exit(false),
      exit_mutex(),
      exit_cv(exit_mutex)
  {}
  
  Result endTrack() {
    signalExit();
    return SUCCESS;
  }
  
  Result keyReleased(string& key) {
    if (key == "KEY_q") {
      signalExit();
    } else {
      GLOBAL_ERROR_STREAM("Unsupported key pressed: " << key << endl);
    }
    
    return SUCCESS;
  }
  
  void handleEvents() {
    exit_mutex.lock();
    while (!exit) {
      exit_cv.wait();
    }
    exit_mutex.unlock();
  }
};


/* Global options */
bool verbose          = false;
bool show_information = false;
bool ac3_writethrough = false;
bool deinterlace      = false;
bool scale            = false;
bool auto_aspect      = false;
bool no_subtitles     = false;
bool use_ffmpeg       = false;
int audio_track       = 0;
int spu_stream        = 0;
string device         = "/dev/cdrom";
string filename       = "dvddump.avi";
unsigned int width    = 0;
unsigned int height   = 0;
int ratio_x           = 0;
int ratio_y           = 0;
int chapter           = 1;
int title             = 1;
int video_bitrate     = 1200;
int audio_bitrate     = 128;
int audio_quality     = 2;
int sample_rate       = 48000;
int num_channels      = 2;
int bits_per_sample   = 16;
double framerate      = 25;

void usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options]" << endl
       << endl
       << "    -d --device:         Set the DVD device [" << device << "]" << endl
       << "    -o --filename:       Set output filename [" << filename << "]" << endl
       << "    -b --vid-bitrate:    Set video bitrate (kBit/s) [" << (video_bitrate / 1000) << "]" << endl
       << "    -w --aud-bitrate:    Set audio bitrate (kBit/s) [" << audio_bitrate << "]" << endl
       << "    -q --quality:        Set audio quality [" << audio_quality << "]" << endl
       << "    -x --ratio-x:        Set the horizontal component of the video aspect [auto]" << endl
       << "    -y --ratio-y:        Set the vertical component of the video aspect [auto]" << endl
       << "    -a --audiostream:    Set the audio stream number [" << audio_track << "]" << endl
       << "    -s --spu:            Set the subtitle stream number [" << spu_stream << "]" << endl
       << "    -t --title:          Select title [" << title << "]" << endl
       << "    -c --chapter:        Select chapter [" << chapter << "]" << endl
       << "    -P --ac3writethrough Write ac3 data directly to output" << endl
       << "    -D --deinterlace     Deinterlace video frames" << endl
       << "    -n --no-scale        Do not scale video" << endl
       << "    -p --auto-aspect     Automatically adjust size to aspect ratio" << endl
       << "    -F --ffmpeg          Use ffmpeg library for encoding" << endl
       << "    -S --no-subtitles    Disable subtitles" << endl
       << "    -v --verbose:        Verbose mode (lots of messages!)" << endl
       << "    -i --information:    Shows number of title, chapters and stream informations" << endl
       << "    -h --help:           Shows this help and exits." << endl;
  exit(0);
}

void turnOffAllMessages() {
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
}


int fderr = -1;
int fd    = -1;

void setStderr( bool on ) {
  if ( !on ) {
    // get copy of stderr desciptor
    fderr = dup( 2 );
    // close stderr
    close( 2 );
    // open /dev/null directly, so that it has fd=2 (stderr)
    fd = open( "/dev/null", O_WRONLY );
  } else {
    // close fd(=2=/dev/null), and set it to stderr again
    dup2( fderr, fd );
  }
}


int main(int argc, char ** argv) {
  const char *optstring = "d:o:x:y:a:s:t:c:b:w:q:pFPSvDnih";
  struct option longopts[] = {
    {"vid-bitrate",    1, NULL, 'b'},
    {"aud-bitrate",    1, NULL, 'w'},
    {"quality",        1, NULL, 'q'},
    {"device",         1, NULL, 'd'},
    {"filename",       1, NULL, 'o'},
    {"ratio-x",        1, NULL, 'x'},
    {"ratio-y",        1, NULL, 'y'},
    {"audiostream",    1, NULL, 'a'},
    {"audiostream",    1, NULL, 's'},
    {"title",          1, NULL, 't'},
    {"chapter",        1, NULL, 'c'},
    {"auto-aspect",    0, NULL, 'p'},
    {"ffmpeg",         0, NULL, 'F'},
    {"ac3writethrough",0, NULL, 'P'},
    {"no-subtitles",   0, NULL, 'S'},
    {"deinterlace",    0, NULL, 'D'},
    {"no-scale",       0, NULL, 'n'},
    {"verbose",        0, NULL, 'v'},
    {"information",    0, NULL, 'i'},
    {"help",           0, NULL, 'h'},
    {NULL         ,    0, NULL, 0  } /* last row must be zero */
  };

  int ch, optionindex;

  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'b':
      if (optarg) {
	video_bitrate = atoi(optarg);
      }
      break;
    case 'w':
      if (optarg) {
	audio_bitrate = atoi(optarg);
      }
      break;
    case 'q':
      if (optarg) {
	audio_quality = atoi(optarg);
      }
      break;
    case 'd':
      if (optarg) {
	device = optarg;
      }
      break;
    case 'o':
      if (optarg) {
	filename = optarg;
      }
      break;
    case 'x':
      if (optarg) {
	ratio_x = atoi(optarg);
      }
      break;
    case 'y':
      if (optarg) {
	ratio_y = atoi(optarg);
      }
      break;
    case 'a':
      if (optarg) {
	audio_track = atoi(optarg);
      }
      break;
    case 's':
      if (optarg) {
	spu_stream = atoi(optarg);
      }
      break;
    case 'F':
      use_ffmpeg = true;
      break;
    case 'S':
      no_subtitles = true;
      break;
    case 'P':
      ac3_writethrough = true;
      break;
    case 't':
      if (optarg) {
	title = atoi(optarg);
      }
      break;
    case 'c':
      if (optarg) {
	chapter = atoi(optarg);
      }
      break;
    case 'D':
      deinterlace = true;
      break;
    case 'p':
      auto_aspect = true;
      break;
    case 'n':
      scale = false;
      break;
    case 'i':
      show_information = true;
      break;
    case 'v':
      verbose = true;
      break;
    default: //includes 'h'
      usage(argv[0]);
    }
  }
  
  // maybe turn off debug messages
  if (!verbose) {
    turnOffAllMessages();
  }
  
  if (width != 0 || height != 0 || auto_aspect) {
    scale = true;
  }
  
  NMMApplication* app = 0;
  
  /*Begin> Allocate _var variables for the INodes */
  INode_var dvd;
  INode_var demux;
  INode_var video_dec;
  INode_var deinterlacer;
  INode_var scaler;
  INode_var spu_dec;
  INode_var overlay;
  INode_var ac3_dec;
  INode_var audio_enc;
  INode_var video_enc;
  INode_var av_mux;
  INode_var writer;
  /*End< */
  
  /* Allocate an event listener */
  EventCatcher ec;
  AppEventReceiver receiver;
  
  int error_level = 0;
  try {
    app = ProxyApplication::getApplication(argc, argv);
    
    /*Begin> Init requests for nodes */
    NodeDescription dvd_request;
    dvd_request.setNodeName("DVDReadNode");
    
    NodeDescription demux_request;
    demux_request.setNodeName("MPEGDemuxNode");
    
    NodeDescription video_dec_request;
    video_dec_request.setNodeName("MPEGVideoDecodeNode");
    
    NodeDescription deinterlacer_request;
    deinterlacer_request.setNodeName("YUVDeInterlaceNode");
    
    NodeDescription scaler_request;
    scaler_request.setNodeName("VideoScalerNode");
    
    NodeDescription spu_dec_request;
    spu_dec_request.setNodeName("SPUDecodeNode");
    
    NodeDescription overlay_request;
    overlay_request.setNodeName("OverlayNode");
    
    NodeDescription ac3_dec_request;
    ac3_dec_request.setNodeName("AC3DecodeNode");
    
    NodeDescription audio_enc_request;
    audio_enc_request.setNodeName("MPEGAudioEncodeNode");
    
    NodeDescription video_enc_request;
    video_enc_request.setNodeName(use_ffmpeg ? "FFMpegEncodeNode" : "DivX4EncodeNode");

    NodeDescription av_mux_request;
    av_mux_request.setNodeName("AVMuxNode");
    
    NodeDescription writer_request;
    writer_request.setNodeName("AVIWriteNode");
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    {
      RegistryLock lock(registry);
      
      list<Response> dvd_response = registry.initRequest(dvd_request);
      list<Response> demux_response = registry.initRequest(demux_request);
      list<Response> video_dec_response = registry.initRequest(video_dec_request);
      list<Response> deinterlacer_response = registry.initRequest(deinterlacer_request);
      list<Response> scaler_response = registry.initRequest(scaler_request);
      list<Response> spu_dec_response = registry.initRequest(spu_dec_request);
      list<Response> overlay_response = registry.initRequest(overlay_request);
      list<Response> ac3_dec_response = registry.initRequest(ac3_dec_request);
      list<Response> audio_enc_response = registry.initRequest(audio_enc_request);
      list<Response> video_enc_response = registry.initRequest(video_enc_request);
      list<Response> av_mux_response = registry.initRequest(av_mux_request);
      list<Response> writer_response = registry.initRequest(writer_request);
      
      if (dvd_response.empty() ||
          demux_response.empty() ||
          video_dec_response.empty() ||
          deinterlacer_response.empty() ||
          scaler_response.empty() ||
          spu_dec_response.empty() ||
          overlay_response.empty() ||
          ac3_dec_response.empty() ||
          audio_enc_response.empty() ||
          video_enc_response.empty() ||
          av_mux_response.empty() ||
          writer_response.empty()) {
        GLOBAL_ERROR_STREAM("Size of dvd_response " << dvd_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of demux_response " << demux_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of video_dec_response " << video_dec_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of deinterlacer_response " << deinterlacer_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of scaler_response " << scaler_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of spu_dec_response " << spu_dec_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of overlay_response " << overlay_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of ac3_dec_response " << ac3_dec_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of audio_enc_response " << audio_enc_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of video_enc_response " << video_enc_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of av_mux_response " << av_mux_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of writer_response " << writer_response.size() << endl);
        throw Exception("A required node is not available");
      }
      
      dvd.reset(registry.requestNode(dvd_response.front()));
      demux.reset(registry.requestNode(demux_response.front()));
      video_dec.reset(registry.requestNode(video_dec_response.front()));
      deinterlacer.reset(registry.requestNode(deinterlacer_response.front()));
      scaler.reset(registry.requestNode(scaler_response.front()));
      spu_dec.reset(registry.requestNode(spu_dec_response.front()));
      overlay.reset(registry.requestNode(overlay_response.front()));
      ac3_dec.reset(registry.requestNode(ac3_dec_response.front()));
      audio_enc.reset(registry.requestNode(audio_enc_response.front()));
      video_enc.reset(registry.requestNode(video_enc_response.front()));
      av_mux.reset(registry.requestNode(av_mux_response.front()));
      writer.reset(registry.requestNode(writer_response.front()));
    }
    /*End< */
    
    /*Begin> Initialize DVD */
    IDVD_var dvd_interface(dvd->getParentObject()->getCheckedInterface<IDVD>());
    dvd_interface->setDevice(device);
    
    //setStderr( false );
    dvd->init();
    dvd->initOutput();
    //setStderr( true );
    
    dvd_interface->select(1,1,1);
    
    {
      Property_var dvd_wout(dvd->getWorkingOutputProperty());
      const Format* file_format = dvd_wout->getDefaultFormat();
      if (file_format) {
        if (file_format->getFormatType() != "av/mpegps") {
          throw Exception("No PS/PES stream was found, seems to be an elementary stream!");
        }
      } else {
        throw Exception("Can't identify data stream!");
      }
    }
    
    int chapters[100];
    int titles = dvd_interface->getTitles();
    for(int i=1; i<=titles; i++) {
      dvd_interface->select(i,1,1);
      chapters[i] = dvd_interface->getChapters(i);
    }
    if ( (title > titles) || (title < 1) ) {
      throw Exception(string("You selected title ") + title + string(", but there are only ") + titles + string(" titles on this DVD"));
    }
    if ( (chapter > chapters[title]) || (chapter < 1) ) {
      throw Exception(string("You selected chapter ") + chapter + string(", but there are only ") + chapters[title] + string(" chapters in this title"));
    }
    
    if ( show_information ) {
      
      cout << "DVD-Contents:" << endl;
      cout << "-------------" << endl;
      cout << "There are " << titles << " titles on this DVD." << endl;
      
      for(int i=1; i<=titles; i++) {
        cout << "Title " << i << " has " << chapters[i] << " chapters" << endl;
      }
      
    } else {
      
      dvd_interface->select(title, chapter, 1);
      
      if ( title > 0 && chapter > 0 ) {
        dvd_interface->select( title, chapter, 1 );
      }
      /*End< Initialize DVD */
    
      /*Begin> Initialize transcoder graph */
      demux->init();
      IMPEGDemux_var mpeg_demux(demux->getParentObject()->getCheckedInterface<IMPEGDemux>());
      mpeg_demux->setDVDMode(true);
      {
        IBiDirQueuedNode_var q(video_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
        q->setDownstreamMaxSize(200);
      }
      video_dec->init();
      if (deinterlace) {
        deinterlacer->init();
      }
      if (scale) {
        // fw: not needed. only affects instream format changes. default working output format is always aspect corrected
//         IScaler_var scaler_interface(scaler->getParentObject()->getCheckedInterface<IScaler>());
//         scaler_interface->setAutoAspect(auto_aspect);
        scaler->init();
      }
      if (!no_subtitles) {
        {
          IBiDirQueuedNode_var q(spu_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
          q->setDownstreamMaxSize(100);
        }
        spu_dec->init();
        overlay->init();
      }
      if (!ac3_writethrough) {
        {
          IBiDirQueuedNode_var q(ac3_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
          q->setDownstreamMaxSize(200);
        }
        ac3_dec->init();
        {
          IBiDirQueuedNode_var q(audio_enc->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
          q->setDownstreamMaxSize(100);
        }
        audio_enc->init();
      }
      {
        IBiDirQueuedNode_var q(video_enc->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
        q->setDownstreamMaxSize(100);
      }
      video_enc->init();
      {
        IBiDirQueuedNode_var q(av_mux->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
        q->setDownstreamMaxSize(100);
      }
      av_mux->init();
      {
        IBiDirQueuedNode_var q(writer->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
        q->setDownstreamMaxSize(100);
      }
    
      ISinkFileHandler_var file_sink(writer->getParentObject()->getCheckedInterface<ISinkFileHandler>());
      file_sink->setFilename(filename);
      writer->init();
      /*End< Initialize transcoder graph */
      
      /*Begin> Connect transcoder graph */
      connect(dvd, demux);
      dvd->activate();
      demux->initOutput();

      /* Check MPEG format */
      string spu_jack_string = string("spu") + spu_stream;
      string audio_jack_string = string("ac3_audio") + audio_track;
      string video_jack_string = string("mpeg_video") + 0;
      {
        Property_var demux_wout_property(demux->getWorkingOutputProperty());
        
        const Format* mpeg_video_format = demux_wout_property->getDefaultFormat(video_jack_string);
        if ( mpeg_video_format != NULL && mpeg_video_format->getFormatType() == "video/mpeg" ) {
          cout << "Found MPEG video stream version " << mpeg_video_format->getIntValue("version") << endl;
        } else {
          throw Exception("No video stream found");
        }
        
        const Format* ac3_audio_format = demux_wout_property->getDefaultFormat(audio_jack_string);
        if ( ac3_audio_format != NULL && ac3_audio_format->getFormatType() == "audio/ac3" ) {
          cout << "Found AC3 audio stream" << endl;
        } else {
          throw Exception("No audio stream found");
        }
        
        if (!no_subtitles) {
          const Format* spu_format = demux_wout_property->getDefaultFormat(spu_jack_string);
          if ( spu_format != NULL && spu_format->getFormatType() == "video/spu" ) {
            cout << "Found subtitle picture unit" << endl;
          } else {
            no_subtitles = true;
            GLOBAL_WARNING_STREAM("No SPU stream found");
          }
        }
      }
      
      /* Connect audio */
      if (!ac3_writethrough) {
        connect(demux, audio_jack_string, ac3_dec);
      } else {
        connect(demux, audio_jack_string, av_mux, string("audio"));
      }
      
      /* Connect SPU */
      if (!no_subtitles) {
        connect(demux, spu_jack_string, spu_dec);
      }
      
      /* Connect video */
      connect(demux, video_jack_string, video_dec);
      
      demux->activate();
      
      video_dec->initOutput();
      
      /* Add deinterlacer */
      INode* deinterlaced_video_source = video_dec.get();
      if (deinterlace) {
        connect(video_dec, deinterlacer);
        deinterlaced_video_source = deinterlacer.get();
      
        video_dec->activate();
        deinterlacer->initOutput();
      }
    
      /* Add overlay */
      INode* raw_video_source = deinterlaced_video_source;
      if (!no_subtitles) {
        spu_dec->initOutput();
        connect(spu_dec, overlay, string("foreground"));
        connect(deinterlaced_video_source, overlay.get(), string("background"));
      
        raw_video_source = overlay.get();
      
        spu_dec->activate();
        deinterlaced_video_source->activate();
        overlay->initOutput();
      }
      
      /* Add scaler */
      if (scale) {
        connect(raw_video_source, scaler.get());
        
        raw_video_source->activate();
        scaler->initOutput();
        
        Property_var scaler_wout(scaler->getWorkingOutputProperty());
        const Format* scaler_def_format = scaler_wout->getDefaultFormat();
        Format* scaled_video_format = new Format(scaler_def_format->getFormatType());
        int x_resolution = scaler_def_format->getIntValue(Format::x_resolution_param);
        int y_resolution = scaler_def_format->getIntValue(Format::y_resolution_param);
        if (width != 0) {
          x_resolution = width;
        }
        if (height != 0) {
          y_resolution = height;
        }
        int ratio_x = scaler_def_format->getIntValue(Format::ratio_x_param);
        int ratio_y = scaler_def_format->getIntValue(Format::ratio_y_param);
        float framerate = scaler_def_format->getFloatValue(Format::framerate_param);
        scaled_video_format->addIntParamValue(Format::x_resolution_param, x_resolution);
        scaled_video_format->addIntParamValue(Format::y_resolution_param, y_resolution);
        scaled_video_format->addIntParamValue(Format::ratio_x_param, ratio_x);
        scaled_video_format->addIntParamValue(Format::ratio_y_param, ratio_y);
        scaled_video_format->addFloatParamValue(Format::framerate_param, framerate);
        scaled_video_format->addStringParamValue(Format::colorspace_param, "yv12");
        scaled_video_format->addStringParamValue(Format::channelorder_param, "yvu");
        scaled_video_format->addStringParamValue(Format::format_param, "planar");
        scaled_video_format->addStringParamValue(Format::endian_param, "little");
        scaled_video_format->addIntParamValue(Format::bitperpixel_param, 12);
        
        connect(scaler, video_enc, scaled_video_format);
        scaler->activate();
      } else {
        connect(raw_video_source, video_enc.get());
        raw_video_source->activate();
      }
      
      // create divx format (with bitrate)
      video_enc->initOutput();
      {
        Property_var video_enc_wout(video_enc->getWorkingOutputProperty());
        const Format* video_enc_def_format = video_enc_wout->getDefaultFormat();
        Format* divx_video_format = new Format(video_enc_def_format->getFormatType());
        int x_resolution = video_enc_def_format->getIntValue(Format::x_resolution_param);
        int y_resolution = video_enc_def_format->getIntValue(Format::y_resolution_param);
        int ratio_x = video_enc_def_format->getIntValue(Format::ratio_x_param);
        int ratio_y = video_enc_def_format->getIntValue(Format::ratio_y_param);
        float framerate = video_enc_def_format->getFloatValue(Format::framerate_param);
        string compressor = video_enc_def_format->getStringValue(Format::compressor_param);
        divx_video_format->addIntParamValue(Format::x_resolution_param, x_resolution);
        divx_video_format->addIntParamValue(Format::y_resolution_param, y_resolution);
        divx_video_format->addIntParamValue(Format::ratio_x_param, ratio_x);
        divx_video_format->addIntParamValue(Format::ratio_y_param, ratio_y);
        divx_video_format->addFloatParamValue(Format::framerate_param, framerate);
        divx_video_format->addIntParamValue(Format::bitrate_param, video_bitrate * 1000);
        divx_video_format->addStringParamValue(Format::compressor_param, compressor);
        
        /* Video to Mux */
        connect(video_enc, av_mux, string("video"), divx_video_format);
      }
      video_enc->activate();
      
      /* Add audio encoder */
      if (!ac3_writethrough) {
        ac3_dec->initOutput();
        connect(ac3_dec, audio_enc);
        
        ac3_dec->activate();
        audio_enc->initOutput();
        
        Property_var audio_enc_wout(audio_enc->getWorkingOutputProperty());
        const Format* audio_enc_def_format = audio_enc_wout->getDefaultFormat();
        Format* mpeg_audio_format = new Format(audio_enc_def_format->getFormatType());
        int sample_rate = audio_enc_def_format->getIntValue(Format::sample_rate_param);
        int channels = audio_enc_def_format->getIntValue(Format::channels_param);
        int audio_codec = audio_enc_def_format->getIntValue(Format::audio_codec_param);
        mpeg_audio_format->addIntParamValue(Format::layer_param, 3);
        mpeg_audio_format->addIntParamValue(Format::sample_rate_param, sample_rate);
        mpeg_audio_format->addIntParamValue(Format::channels_param, channels);
        mpeg_audio_format->addIntParamValue(Format::bitpersample_param, 16);
        mpeg_audio_format->addIntParamValue(Format::audio_codec_param, audio_codec);
        mpeg_audio_format->addIntParamValue(Format::bitrate_param, audio_bitrate * 1000);
        
        connect(audio_enc, av_mux, string("audio"), mpeg_audio_format);
        audio_enc->activate();
      }
      
      /* Connect Mux to writer */
      av_mux->initOutput();
      connect(av_mux, writer);
      av_mux->activate();
      writer->initOutput();
      writer->activate();
      /*End< Connect transcoder graph */
    
      /*Begin> Start graph */
      if (dvd.get() && dvd->isActivated()) {
        dvd->start();
      }
      if (demux.get() && demux->isActivated()) {
        demux->start();
      }
      if (video_dec.get() && video_dec->isActivated()) {
        video_dec->start();
      }
      if (deinterlacer.get() && deinterlacer->isActivated()) {
        deinterlacer->start();
      }
      if (scaler.get() && scaler->isActivated()) {
        scaler->start();
      }
      if (spu_dec.get() && spu_dec->isActivated()) {
        spu_dec->start();
      }
      if (overlay.get() && overlay->isActivated()) {
        overlay->start();
      }
      if (ac3_dec.get() && ac3_dec->isActivated()) {
        ac3_dec->start();
      }
      if (audio_enc.get() && audio_enc->isActivated()) {
        audio_enc->start();
      }
      if (video_enc.get() && video_enc->isActivated()) {
        video_enc->start();
      }
      if (av_mux.get() && av_mux->isActivated()) {
        av_mux->start();
      }
      if (writer.get() && writer->isActivated()) {
        writer->start();
      }
      /*End< Start graph */

      /* Begin> Add console producer */
      ConsoleProducer console;
      console.readConfig( getNMMResource("XProducer.xml") );
      console.connectTo(&receiver);
      console.start();
      /* End< Add console producer */
      
      cout << "\nGrabbing now... (press 'q' to stop grabbing)" << endl;
      receiver.getEventDispatcher().registerEventListener(IDisplayEvents::keyReleased_event,
                                                          new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
      writer->getParentObject()->registerEventListener(ITrack::endTrack_event,
                                                       new TEDObject0<EventCatcher>(&ec, &EventCatcher::endTrack));
      ec.handleEvents();
      cout << "Grabbing finished successfully" << endl;
      
    } // if (show_information)
    
  } catch(const Exception& e) {
    cerr << e << endl;
    error_level = 1;
  } catch(...) {
    cerr << "Something strange went wrong!" << endl;
    error_level = 1;
  }
  
  /*Begin> Clean up */
  try {
    /*Begin> Stop graph */
    if (dvd.get() && dvd->isStarted()) {
      dvd->stop();
    }
    if (demux.get() && demux->isStarted()) {
      demux->stop();
    }
    if (video_dec.get() && video_dec->isStarted()) {
      video_dec->stop();
    }
    if (deinterlacer.get() && deinterlacer->isStarted()) {
      deinterlacer->stop();
    }
    if (scaler.get() && scaler->isStarted()) {
      scaler->stop();
    }
    if (spu_dec.get() && spu_dec->isStarted()) {
      spu_dec->stop();
    }
    if (overlay.get() && overlay->isStarted()) {
      overlay->stop();
    }
    if (ac3_dec.get() && ac3_dec->isStarted()) {
      ac3_dec->stop();
    }
    if (audio_enc.get() && audio_enc->isStarted()) {
      audio_enc->stop();
    }
    if (video_enc.get() && video_enc->isStarted()) {
      video_enc->stop();
    }
    if (av_mux.get() && av_mux->isStarted()) {
      av_mux->stop();
    }
    if (writer.get() && writer->isStarted()) {
      writer->stop();
    }
    /*End< Stop graph */
    
    /*Begin> Flush graph */
    if (dvd.get() && dvd->isActivated()) {
      dvd->flush();
    }
    if (demux.get() && demux->isActivated()) {
      demux->flush();
    }
    if (video_dec.get() && video_dec->isActivated()) {
      video_dec->flush();
    }
    if (deinterlacer.get() && deinterlacer->isActivated()) {
      deinterlacer->flush();
    }
    if (scaler.get() && scaler->isActivated()) {
      scaler->flush();
    }
    if (spu_dec.get() && spu_dec->isActivated()) {
      spu_dec->flush();
    }
    if (overlay.get() && overlay->isActivated()) {
      overlay->flush();
    }
    if (ac3_dec.get() && ac3_dec->isActivated()) {
      ac3_dec->flush();
    }
    if (audio_enc.get() && audio_enc->isActivated()) {
      audio_enc->flush();
    }
    if (video_enc.get() && video_enc->isActivated()) {
      video_enc->flush();
    }
    if (av_mux.get() && av_mux->isActivated()) {
      av_mux->flush();
    }
    if (writer.get() && writer->isActivated()) {
      writer->flush();
    }
    /*End< Flush graph */
    
    /*Begin> Disconnect graph */
    if (writer.get()) {
      if (writer->isActivated()) {
        writer->deactivate();
      }
      if (writer->isOutputInitialized()) {
        writer->deinitOutput();
      }
      IInputJack_var writer_in_jack(writer->getInputStream("default"));
      if (writer_in_jack.get() && writer_in_jack->isConnected()) {
        if (av_mux.get() && av_mux->isActivated()) {
          av_mux->deactivate();
        }
        disconnect_input(writer);
      }
    }
    if (av_mux.get()) {
      if (av_mux->isActivated()) {
        av_mux->deactivate();
      }
      if (av_mux->isOutputInitialized()) {
        av_mux->deinitOutput();
      }
      IInputJack_var av_mux_audio_in_jack(av_mux->getInputStream("audio"));
      if (av_mux_audio_in_jack.get() && av_mux_audio_in_jack->isConnected()) {
        if (ac3_writethrough) {
          if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
        } else {
          if (audio_enc.get() && audio_enc->isActivated()) {
            audio_enc->deactivate();
          }
        }
        disconnect_input(av_mux, "audio");
      }
      IInputJack_var av_mux_video_in_jack(av_mux->getInputStream("video"));
      if (av_mux_video_in_jack.get() && av_mux_video_in_jack->isConnected()) {
        if (video_enc.get() && video_enc->isActivated()) {
          video_enc->deactivate();
        }
        disconnect_input(av_mux, "video");
      }
    }
    if (!ac3_writethrough) {
      if (audio_enc.get()) {
        if (audio_enc->isActivated()) {
          audio_enc->deactivate();
        }
        if (audio_enc->isOutputInitialized()) {
          audio_enc->deinitOutput();
        }
        IInputJack_var audio_enc_in_jack(audio_enc->getInputStream("default"));
        if (audio_enc_in_jack.get() && audio_enc_in_jack->isConnected()) {
          if (ac3_dec.get() && ac3_dec->isActivated()) {
            ac3_dec->deactivate();
          }
          disconnect_input(audio_enc);
        }
      }
      if (ac3_dec.get()) {
        if (ac3_dec->isActivated()) {
          ac3_dec->deactivate();
        }
        if (ac3_dec->isOutputInitialized()) {
          ac3_dec->deinitOutput();
        }
        IInputJack_var ac3_dec_in_jack(ac3_dec->getInputStream("default"));
        if (ac3_dec_in_jack.get() && ac3_dec_in_jack->isConnected()) {
          if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(ac3_dec);
        }
      }
    }
    if (video_enc.get()) {
      if (video_enc->isActivated()) {
        video_enc->deactivate();
      }
      if (video_enc->isOutputInitialized()) {
        video_enc->deinitOutput();
      }
      IInputJack_var video_enc_in_jack(video_enc->getInputStream("default"));
      if (video_enc_in_jack.get() && video_enc_in_jack->isConnected()) {
        if (scale) {
          if (scaler.get() && scaler->isActivated()) {
            scaler->deactivate();
          }
        } else {
          if (!no_subtitles) {
            if (overlay.get() && overlay->isActivated()) {
              overlay->deactivate();
            }
          } else {
            if (deinterlace) {
              if (deinterlacer.get() && deinterlacer->isActivated()) {
                deinterlacer->deactivate();
              }
            } else {
              if (video_dec.get() && video_dec->isActivated()) {
                video_dec->deactivate();
              }
            }
          }
        }
        disconnect_input(video_enc);
      }
    }
    if (scale && scaler.get()) {
      if (scaler->isActivated()) {
        scaler->deactivate();
      }
      if (scaler->isOutputInitialized()) {
        scaler->deinitOutput();
      }
      IInputJack_var scaler_in_jack(scaler->getInputStream("default"));
      if (scaler_in_jack.get() && scaler_in_jack->isConnected()) {
        if (!no_subtitles) {
          if (overlay.get() && overlay->isActivated()) {
            overlay->deactivate();
          }
        } else {
          if (deinterlace) {
            if (deinterlacer.get() && deinterlacer->isActivated()) {
              deinterlacer->deactivate();
            }
          } else {
            if (video_dec.get() && video_dec->isActivated()) {
              video_dec->deactivate();
            }
          }
        }
        disconnect_input(scaler);
      }
    }
    if (!no_subtitles) {
      if (overlay.get()) {
        if (overlay->isActivated()) {
          overlay->deactivate();
        }
        if (overlay->isOutputInitialized()) {
          overlay->deinitOutput();
        }
        IInputJack_var overlay_background_in_jack(overlay->getInputStream("background"));
        if (overlay_background_in_jack.get() && overlay_background_in_jack->isConnected()) {
          if (deinterlace) {
            if (deinterlacer.get() && deinterlacer->isActivated()) {
              deinterlacer->deactivate();
            }
          } else {
            if (video_dec.get() && video_dec->isActivated()) {
              video_dec->deactivate();
            }
          }
          disconnect_input(overlay, "background");
        }
        IInputJack_var overlay_foreground_in_jack(overlay->getInputStream("foreground"));
        if (overlay_foreground_in_jack.get() && overlay_foreground_in_jack->isConnected()) {
          if (spu_dec.get() && spu_dec->isActivated()) {
            spu_dec->deactivate();
          }
          disconnect_input(overlay, "foreground");
        }
      }
      if (spu_dec.get()) {
        if (spu_dec->isActivated()) {
          spu_dec->deactivate();
        }
        if (spu_dec->isOutputInitialized()) {
          spu_dec->deinitOutput();
        }
        IInputJack_var spu_dec_in_jack(spu_dec->getInputStream("default"));
        if (spu_dec_in_jack.get() && spu_dec_in_jack->isConnected()) {
          if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(spu_dec);
        }
      }
    }
    if (deinterlace && deinterlacer.get()) {
      if (deinterlacer->isActivated()) {
        deinterlacer->deactivate();
      }
      if (deinterlacer->isOutputInitialized()) {
        deinterlacer->deinitOutput();
      }
      IInputJack_var deinterlacer_in_jack(deinterlacer->getInputStream("default"));
      if (deinterlacer_in_jack.get() && deinterlacer_in_jack->isConnected()) {
        if (video_dec.get() && video_dec->isActivated()) {
          video_dec->deactivate();
        }
        disconnect_input(deinterlacer);
      }
    }
    if (video_dec.get()) {
      if (video_dec->isActivated()) {
        video_dec->deactivate();
      }
      if (video_dec->isOutputInitialized()) {
        video_dec->deinitOutput();
      }
      IInputJack_var video_dec_in_jack(video_dec->getInputStream("default"));
      if (video_dec_in_jack.get() && video_dec_in_jack->isConnected()) {
        if (demux.get() && demux->isActivated()) {
          demux->deactivate();
        }
        disconnect_input(video_dec);
      }
    }
    if (demux.get()) {
      if (demux->isActivated()) {
        demux->deactivate();
      }
      if (demux->isOutputInitialized()) {
        demux->deinitOutput();
      }
      IInputJack_var demux_in_jack(demux->getInputStream("default"));
      if (demux_in_jack.get() && demux_in_jack->isConnected()) {
        if (dvd.get() && dvd->isActivated()) {
          dvd->deactivate();
        }
        disconnect_input(demux);
      }
    }
    if (dvd.get()) {
      if (dvd->isActivated()) {
        dvd->deactivate();
      }
      if (dvd->isOutputInitialized()) {
        dvd->deinitOutput();
      }
    }
    /*End< Disconnect graph */
    
    /*Begin> Deinit graph */
    if (dvd.get() && dvd->isInitialized()) {
      dvd->deinit();
    }
    if (demux.get() && demux->isInitialized()) {
      demux->deinit();
    }
    if (video_dec.get() && video_dec->isInitialized()) {
      video_dec->deinit();
    }
    if (deinterlacer.get() && deinterlacer->isInitialized()) {
      deinterlacer->deinit();
    }
    if (scaler.get() && scaler->isInitialized()) {
      scaler->deinit();
    }
    if (spu_dec.get() && spu_dec->isInitialized()) {
      spu_dec->deinit();
    }
    if (overlay.get() && overlay->isInitialized()) {
      overlay->deinit();
    }
    if (ac3_dec.get() && ac3_dec->isInitialized()) {
      ac3_dec->deinit();
    }
    if (audio_enc.get() && audio_enc->isInitialized()) {
      audio_enc->deinit();
    }
    if (video_enc.get() && video_enc->isInitialized()) {
      video_enc->deinit();
    }
    if (av_mux.get() && av_mux->isInitialized()) {
      av_mux->deinit();
    }
    if (writer.get() && writer->isInitialized()) {
      writer->deinit();
    }
    /*End< Deinit graph */
    
    ClientRegistry& registry = app->getRegistry();  
    if (writer.get()) {
      registry.releaseNode(*writer);
    }
    if (av_mux.get()) {
      registry.releaseNode(*av_mux);
    }
    if (audio_enc.get()) {
      registry.releaseNode(*audio_enc);
    }
    if (ac3_dec.get()) {
      registry.releaseNode(*ac3_dec);
    }
    if (video_enc.get()) {
      registry.releaseNode(*video_enc);
    }
    if (scaler.get()) {
      registry.releaseNode(*scaler);
    }
    if (overlay.get()) {
      registry.releaseNode(*overlay);
    }
    if (spu_dec.get()) {
      registry.releaseNode(*spu_dec);
    }
    if (deinterlacer.get()) {
      registry.releaseNode(*deinterlacer);
    }
    if (video_dec.get()) {
      registry.releaseNode(*video_dec);
    }
    if (demux.get()) {
      registry.releaseNode(*demux);
    }
    if (dvd.get()) {
      registry.releaseNode(*dvd);
    }
  } catch(const Exception& e) {
    cerr << "Exception during cleanup" << endl;
    cerr << e << endl;
    error_level = 1;
  } catch(...) {
    cerr << "Something strange went wrong during cleanup" << endl;
    error_level = 1;
  }
  /*End< Clean up */
  
  return error_level;
}
