/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *                     
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iostream>
#include <unistd.h>

#include <getopt.h>

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/utils/utils.hpp"

using namespace NMM;

// Listener class, catches events from the nodes
class MyEventListener {
private:
  bool finished;
public:
  MyEventListener() { finished = false; }
  Result setfinishedflag() {
    finished = true;
    return SUCCESS;
  }
  bool getfinishedflag() { return finished;}
};

void Usage( char *name ) {
  cout<<"\nUsage:\n";
  cout<<"%s inputfilename outputfilename [-b bitrate (kBit/s)] [-v] [-h]\n"<<name<<endl;
  cout<<"-b\t\tSet bitrate (kBit/s)\n" ;
  cout<<"-v\t\tverbose\n" ;
  cout<<"-h\t\tPrint this help\n" ;
  exit(0);
}

void turnOffAllMessages()
{
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
}

int main(int argc, char ** argv) {
  MyEventListener* l = new MyEventListener();     
  string infilename;
  string outfilename;
  bool verbose = false;
  //  int quality = 2;      // set default value for mp3-quality (2=high  5=medium  7=low)
  int bitrate = 128;    // default value for mp3-bitrate in kBit/s

  const char *optstring = "b:vh";
  struct option longopts[] = {
    {"bitrate", 1, NULL, 'b'},
    {"verbose", 0, NULL, 'v'},
    {"help"   , 0, NULL, 'h'},
    {NULL     , 0, NULL, 0  } // last row must be zero
  };

  int ch, optionindex;
  
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'b':
      bitrate = atoi(optarg);
      break;
    case 'v':
      verbose = true;
      break;
    default: //includes 'h'
      Usage(argv[0]);
    }
  }

  // assure that 2 filenames are given
  if (argc != optind + 2) {
    cerr << "Make sure to provide input and output filename!" << endl;
    cerr << " argc = " << argc << "; optind = " << optind << endl;
    while (optind < argc) {
      cerr << "argv[" << optind << "] = " << argv[optind++] << endl;
    }
    Usage(argv[0]); 
  }
  
  // here argc is equal optind + 2
  infilename= "";
  infilename.append(argv[optind++]);
  outfilename = "";
  outfilename.append(argv[optind]);

  // eventually turn off debug messages
  if (!verbose) turnOffAllMessages();
  //////////////////////////////////////////////////////////////////////////
  NMMApplication* app = ProxyApplication::getApplication(argc, argv);
  
  INode_var read,decode,encode,write;

  try{
    ClientRegistry& registry = app->getRegistry();
    
    NodeDescription request1;
    request1.setNodeName("MP3ReadNode");
    NodeDescription request2;
    request2.setNodeName("MPEGAudioDecodeNode");
    NodeDescription request3;
    request3.setNodeName("MPEGAudioEncodeNode");
    NodeDescription request4;
    request4.setNodeName("GenericWriteNode");
    
    list<Response> response1 = registry.initRequest(request1);
    list<Response> response2 = registry.initRequest(request2);
    list<Response> response3 = registry.initRequest(request3);
    list<Response> response4 = registry.initRequest(request4);
    
    if (response1.empty() || response2.empty() || response3.empty() || response4.empty()) {
      GLOBAL_ERROR_STREAM("Size of 1. response " << response1.size() << endl);
      GLOBAL_ERROR_STREAM("Size of 2. response " << response2.size() << endl);
      GLOBAL_ERROR_STREAM("Size of 3. response " << response3.size() << endl);
      GLOBAL_ERROR_STREAM("Size of 4. response " << response4.size() << endl);
      throw Exception ("A required node is not available");
    }

    read.reset(registry.requestNode(response1.front()));
    decode.reset(registry.requestNode(response2.front()));
    encode.reset(registry.requestNode(response3.front()));
    write.reset(registry.requestNode(response4.front()));

    IFileHandler_var iread(read->getParentObject()->getInterface<IFileHandler>());
    ISinkFileHandler_var iwrite(write->getParentObject()->getInterface<ISinkFileHandler>());
    
    /* -- setting Parameters in the Nodes using the Interfaces -- */
    iread->setFilename(infilename);
    //    iwrite->setFileExtension("mp3");
    iwrite->setFilename(outfilename);

    /* -- registering ITrack::endTrack_event -- */
    write->getParentObject()->
      registerEventListener(ITrack::endTrack_event, new TEDObject0<MyEventListener>(l, &MyEventListener::setfinishedflag));

    /* -- try to init and activate the nodes -- */
    try {
      read->init();
      decode->init();
      encode->init();
      write->init();
      
      read->initOutput();
      connect(read,decode);
      read->activate();

      const Format *mp3_format = read->getWorkingOutputProperty()->getDefaultFormat();

      decode->initOutput();
      connect(decode,encode);
      decode->activate();

      encode->initOutput();
      Format* convert_out_format = new Format("audio/mpeg");
      convert_out_format->addIntParamValue(Format::layer_param, 3);
      convert_out_format->addIntParamValue(Format::sample_rate_param, mp3_format->getIntValue(Format::sample_rate_param));
      convert_out_format->addIntParamValue(Format::channels_param, mp3_format->getIntValue(Format::channels_param));
      convert_out_format->addIntParamValue(Format::bitpersample_param, mp3_format->getIntValue(Format::bitpersample_param));
      convert_out_format->addIntParamValue(Format::audio_codec_param, mp3_format->getIntValue(Format::audio_codec_param));
      convert_out_format->addIntParamValue(Format::bitrate_param, bitrate * 1000);
      connect(encode,write,convert_out_format);
      encode->activate();

      write->initOutput();
      write->activate();
    }
    catch(const Exception& e) {
      GLOBAL_ERROR_STREAM(e<<endl);
      GLOBAL_ERROR_STREAM("FAILED to init/activate the needed notes - I'm sorry"<<endl);
      return 1;
    }
    
    /* -- Starting Nodes -- */
    try {
      write->start();
      decode->start();
      encode->start();
      read->start();
    } 
    catch (...) {
      GLOBAL_ERROR_STREAM("FAILED to start the nodes" << endl);
      return 1;
    }
    
    while (!(l->getfinishedflag())) { sleep(1); }
    
    /* -- Stopping Nodes -- */
    try {
      read->stop();
      decode->stop();
      encode->stop();
      write->stop();
    } 
    catch (...) {
      GLOBAL_ERROR_STREAM("FAILED to stop the nodes" << endl);
      return 1;
    }
    
    /* -- Deactivate/Deinit Nodes -- */
    try {
      
      read->deactivate();
      decode->deactivate();
      encode->deactivate();
      write->deactivate();
      read->deinitOutput();
      decode->deinitOutput();
      encode->deinitOutput();
      write->deinitOutput();
      
      read->deinit();
      decode->deinit();
      encode->deinit();
      write->deinit();
    } 
    catch (...) {
      GLOBAL_ERROR_STREAM("FAILED to deactivate/deinit the nodes" << endl);
      return 1;
    }
    registry.releaseNode(*read);
    registry.releaseNode(*decode);
    registry.releaseNode(*encode);
    registry.releaseNode(*write);
    GLOBAL_DEBUG_STREAM("\033[0;31m" << "We did it! ... Enjoy your MP3-File" << "\033[0m" << endl); 
  }     
  catch(...) {
    if (read.get())
      app->getRegistry().releaseNode(*read);
    if (decode.get())
      app->getRegistry().releaseNode(*decode);
    if (encode.get())
      app->getRegistry().releaseNode(*encode);
    if (write.get())
      app->getRegistry().releaseNode(*write);
    cerr<<"So sorry - but something went wront..."<<endl;
    return 1; 
  }
  return 0;
}
