/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "nmm/NMM.hpp"
#include "nmm/utils/net/SockAcceptor.hpp"
#include "nmm/utils/net/SockConnector.hpp"
#include "nmm/utils/net/SockStream.hpp"
#include "nmm/utils/net/NetReactor.hpp"
#include "nmm/utils/net/INetAddr.hpp"

#include "nmm/utils/thread/Thread.hpp"

#include "nmm/base/ISender.hpp"
#include "nmm/base/IReceiver.hpp"

#include "nmm/base/registry/ClientRegistry.hpp"

#include "nmm/base/memorymgr/BufferManager.hpp"
#include "nmm/base/memorymgr/IExternalBufferManager.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"

#include "nmm/base/serialize/net/UnbufferedNetStream.hpp"
#include "nmm/base/serialize/net/BufferedNetStream.hpp"
#include "nmm/base/serialize/net/PacketNetStream.hpp"

#include "nmm/base/serialize/MagicNumberValueStream.hpp"

#include <getopt.h>
#include <sys/time.h>

#include <iostream>
using std::cerr;
using std::endl;

#include <cstdio>

using namespace NMM;

enum { MAX_BUFFER_SIZE = 16384 };

/*======================================== Generic Classes ========================================*/
/** Buffer sender.
 * This class sends buffers to an IMessageReceiver and counts the number of bytes.
 */
class BufferSender :
  public IMessageSender
{
public:
  BufferSender(IMessageReceiver* receiver)
    : __total_bytes(0),
      __receiver(receiver)
  {}
  
  inline ssize_t getTotalBytes() const { return __total_bytes; }
  
  virtual Result sendMessage(Message* message) {
    Buffer* buffer = static_cast<Buffer*>(message);
    if (buffer) {
      __total_bytes += buffer->getUsed();
      return __receiver->receiveMessage(message, this);
    }
    return FAILURE;
  }
  
private:
  ssize_t __total_bytes;
  IMessageReceiver* __receiver;
};


/** Buffer receiver.
 * This class receives buffers and counts the number of bytes.
 */
class BufferReceiver :
  public IMessageReceiver
{
public:
  BufferReceiver()
    : __total_bytes(0)
  {}
  
  inline ssize_t getTotalBytes() const { return __total_bytes; }
  
  virtual Result receiveMessage(Message* message, const IMessageSender*) {
    Buffer* buffer = dynamic_cast<Buffer*>(message);
    if (buffer) {
      __total_bytes += buffer->getUsed();
      buffer->release();
      return SUCCESS;
    }
    message->release();
    return FAILURE;
  }
  
private:
  ssize_t __total_bytes;
};

/** Network connection.
 * This class transmits NMM buffers through the network
 */
class Connection :
  public IMessageSender,
  public IMessageReceiver
{
public:
  Connection()
    : __receiver(NULL),
      __buffer_manager(NULL),
      __cond(__mutex),
      __closed(false)
  {}
  
  virtual ~Connection() {}
  
  inline IMessageReceiver* getMessageReceiver() const { return __receiver; }
  inline BufferManager* getBufferManager() const { return __buffer_manager; }
  
  inline void setMessageReceiver(IMessageReceiver* receiver) { __receiver = receiver; }
  inline void setBufferManager(BufferManager* buffer_manager) { __buffer_manager = buffer_manager; }
  
  inline void setBufferSize(ssize_t buffer_size) { __buffer_size = buffer_size; }
  inline ssize_t getBufferSize() const { return __buffer_size; }
  
  virtual void disconnect() {
    MutexGuard mg(__mutex);
    __closed = true;
    __cond.notify();
  }
  
  virtual ssize_t getTotalBytes() const = 0;
  
  void waitForClosing() {
    MutexGuard mg(__mutex);
    if (!__closed) {
      __cond.wait();
    }
  }
  
  Result receiveMessage(Message* message, const IMessageSender*) {
    Buffer* buffer = static_cast<Buffer*>(message);
    if (buffer) {
      sendBuffer(buffer);
      buffer->release();
      return SUCCESS;
    }
    message->release();
    return FAILURE;
  }
  
  Result sendMessage(Message* message) {
    return __receiver->receiveMessage(message, this);
  }
  
  Buffer* getBuffer(ssize_t size = 0) {
    return recvBuffer(size);
  }
  
  virtual void start() = 0;
  
protected:
  virtual void sendBuffer(Buffer* buffer) = 0;
  virtual Buffer* recvBuffer(ssize_t size = 0) = 0;
  
private:
  IMessageReceiver*        __receiver;
  BufferManager*           __buffer_manager;
  
  ThreadMutex              __mutex;
  ThreadCondition          __cond;
  bool                     __closed;
  
  ssize_t                  __buffer_size;
};

/** Network acceptor.
 * This class listens for a network connection and returns a Connection object.
 */
class Acceptor {
public:
  Acceptor(IMessageReceiver* receiver, BufferManager* manager, unsigned int /*port*/, Connection* connection)
    : __cond(__mutex),
      __connection(connection),
      __exception(NULL),
      __connected(false)
  {
    __connection->setBufferManager(manager);
    __connection->setMessageReceiver(receiver);
  }
  
  virtual ~Acceptor() {
    if (__connection) {
      delete __connection;
      __connection = NULL;
    }
  }
  
  inline Connection* getConnection() const { return __connection; }
  
  Connection* accept()
    throw (const Exception&)
  {
    MutexGuard mg(__mutex);
    if (!__connected) {
      __cond.wait();
    }
    
    if (__exception) {
      Exception e(*__exception);
      delete __exception;
      __exception = NULL;
      throw e;
    }
    
    Connection* c = __connection;
    __connection = NULL;
    return c;
  }
  
  virtual void start() = 0;
  
protected:
  void setException(const Exception& e) {
    __exception = e.copy();
  }
  
  void notifyAccepted() {
    MutexGuard mg(__mutex);
    __connected = true;
    __cond.notify();
  }
  
private:
  ThreadMutex       __mutex;
  ThreadCondition   __cond;
  
  Connection*       __connection;
  Exception*        __exception;
  bool              __connected;
};

template<class _Acceptor, class _Connection>
void runBenchmark(unsigned int num, ssize_t bytes, string host, unsigned int port, bool retry, unsigned int repeat, bool sync) {
  double avg_elapsed=0, avg_payload_over_time=0, avg_transmit_over_time=0;
  ssize_t avg_total_bytes=0, avg_total_transmit=0;
  unsigned int iterations = repeat;
  
  printf(" elapsed       | payload       | payload       | transmit      | transmit      \n");
  printf(" (s)           | (bytes)       | (kb/s)        | (bytes)       | (kb/s)        \n");
  printf("===============================================================================\n");
  
  if (repeat > 1) {
    retry = true;
  }
  while (repeat--) {
    ssize_t total_bytes = 0;     ///< Total buffer bytes
    ssize_t total_transmit = 0;  ///< Total bytes transmitted over the network
    timeval start_time;
    timeval end_time;
    
    if (host.empty()) {
      
      BufferReceiver receiver;
      BufferManager* manager = new B_Allocator;
      
      _Acceptor acceptor(&receiver, manager, port);
      if (sync) {
        acceptor.setNoRegister();
      }
      acceptor.start();
      GLOBAL_MESSAGE_STREAM("Listening on port " << port << endl);
      
      Connection* conn = acceptor.accept();
      GLOBAL_MESSAGE_STREAM("Accepted connection" << endl);
      
      gettimeofday(&start_time, NULL);
      conn->setBufferSize(bytes);
      conn->start();
      if (sync) {
        for (unsigned int i=0; i<num; ++i) {
          Buffer* buffer = conn->getBuffer(bytes);
          if (buffer) {
            receiver.receiveMessage(buffer, conn);
          }
        }
      } else {
        conn->waitForClosing();
      }
      gettimeofday(&end_time, NULL);
      GLOBAL_MESSAGE_STREAM("Connection closed" << endl);
      
      total_bytes = receiver.getTotalBytes();
      total_transmit = conn->getTotalBytes();
      
      delete conn;
      
    } else {
      
      _Connection conn;
      conn.setNoRegister();
      BufferSender sender(&conn);
      INetAddr server_addr(host.c_str(), port);
      GLOBAL_MESSAGE_STREAM("Connecting to " << server_addr.toString() << endl);
      
      for( ;; ) {
        try {
          conn.connect(server_addr);
          conn.start();
          GLOBAL_MESSAGE_STREAM("Connected to " << server_addr.toString() << endl);
        } catch(ConnectionException& e) {
          if (!retry) {
            cerr << e << endl;
            return;
          } else {
            continue;
          }
        }
        break;
      }
      
      BufferManager* manager = new B_Allocator;
      gettimeofday(&start_time, NULL);
      for (unsigned int i=0; i<num; ++i) {
        Buffer* buffer = manager->requestBuffer(bytes);
        buffer->setUsed(bytes);
        sender.sendMessage(buffer);
      }
      gettimeofday(&end_time, NULL);
      
      total_bytes = sender.getTotalBytes();
      total_transmit = conn.getTotalBytes();
    }
    
    end_time.tv_usec -= start_time.tv_usec;
    if (end_time.tv_usec < 0) {
      end_time.tv_usec += 1000000;
      end_time.tv_sec--;
    }
    end_time.tv_sec -= start_time.tv_sec;
    double elapsed = end_time.tv_sec + end_time.tv_usec * 0.000001;
//     cerr << "Elapsed time: " << elapsed << "s" << endl
//          << "Payload     : " << total_bytes / 1024 << "kb"
//          << " (" << total_bytes / 1024 / elapsed << "kb/s)" << endl
//          << "Transmitted : " << total_transmit / 1024 << "kb"
//          << " (" << total_transmit / 1024 / elapsed << "kb/s)" << endl
//          << endl;
    double payload_over_time = static_cast<double>(total_bytes) / 1024.0 / elapsed;
    double transmit_over_time = static_cast<double>(total_transmit) / 1024.0 / elapsed;
    printf("%15.4f|%15d|%15.4f|%15d|%15.4f\n", elapsed, total_bytes, payload_over_time, total_transmit, transmit_over_time);

    avg_elapsed += elapsed;
    avg_total_bytes += total_bytes;
    avg_payload_over_time += payload_over_time;
    avg_total_transmit += total_transmit;
    avg_transmit_over_time += transmit_over_time;
  }
  
  printf("-------------------------------------------------------------------------------\n");
  avg_elapsed /= iterations;
  avg_total_bytes /= iterations;
  avg_payload_over_time /= iterations;
  avg_total_transmit /= iterations;
  avg_transmit_over_time /= iterations;
  printf("%15.4f|%15d|%15.4f|%15d|%15.4f\n", avg_elapsed, avg_total_bytes, avg_payload_over_time, avg_total_transmit, avg_transmit_over_time);
}
/*======================================== Generic Classes ========================================*/


/*======================================== Simple TCP Transport ========================================*/
/** Simple network connection.
 * This class transmits NMM buffers directly through a SockStream.
 */
class SimpleConnection :
  public Connection,
  public DispatchedNetHandler
{
public:
  SimpleConnection()
    : __total_bytes(0),
      __no_register(false)
  {}
  
  virtual ~SimpleConnection() {
    disconnect();
  }
  
  virtual void disconnect() {
    if (__sock_stream.getSocket() >= 0) {
      SNetReactor::getInstance().unregisterSocket(__sock_stream.getSocket());
      __sock_stream.close();
    }
    
    Connection::disconnect();
  }
  
  virtual ssize_t getTotalBytes() const {
    return __total_bytes;
  }
  
  inline SockStream& getSocket() { return __sock_stream; }
  inline void setNoRegister() {
    __no_register = true;
  }
  
  void connect(const INetAddr& server_addr)
    throw (const ConnectionException&)
  {
    SockConnector connector;
    connector.connect(__sock_stream, server_addr, "tcp");
  }
  
  virtual void start() {
    if (!__no_register) {
      SNetReactor::getInstance().registerSocket(getSocket().getSocket(), this, NetReactor::NR_POLLIN);
    }
  }
  
protected:
  void countBytes(ssize_t bytes) {
    __total_bytes += bytes;
  }
  
  virtual void sendBuffer(Buffer* buffer) {
    try {
      countBytes(buffer->getUsed());
      __sock_stream.send(buffer->getData(), buffer->getUsed(), MSG_NOSIGNAL);
    } catch(const TransmissionException& e) {
      cerr << e << endl;
    }
  }
  
  virtual Buffer* recvBuffer(ssize_t size = 0) {
    u_int32_t used = size;
    Buffer* buffer = getBufferManager()->requestBuffer(used);
    buffer->setUsed(used);
    try {
      char* offset = buffer->getData();
      while (used) {
        ssize_t bytes = __sock_stream.recv(offset, used, MSG_NOSIGNAL);
        if (!bytes) {
          buffer->release();
          disconnect();
          return NULL;
        }
        countBytes(bytes);
        used -= bytes;
        offset += bytes;
      }
      return buffer;
    } catch(const TransmissionException& e) {
      buffer->release();
      cerr << e << endl;
      return NULL;
    }
  }
  
protected:
  virtual NetResult receiveData(int) {
    Buffer* buffer = recvBuffer(getBufferSize());
    if (buffer) {
      sendMessage(buffer);
    }
    return VOID;
  }
  
private:
  SockStream               __sock_stream;
  ssize_t                  __total_bytes;
  
  bool                     __no_register;
};


/** Simple network acceptor.
 * This class accepts a simple TCP connection.
 */
class SimpleAcceptor :
  public Acceptor,
  public DispatchedNetHandler
{
public:
  SimpleAcceptor(IMessageReceiver* receiver, BufferManager* manager, unsigned int port,
                 Connection* connection = new SimpleConnection)
    : Acceptor(receiver, manager, port, connection),
      __no_register(false)
  {
    INetAddr local_addr;
    local_addr.setTransportProtocol("tcp");
    local_addr.setPort(port);
    __sock_acceptor.open(local_addr, "tcp");
  }
  
  virtual ~SimpleAcceptor() {
    if (__sock_acceptor.getSocket() >= 0) {
      SNetReactor::getInstance().unregisterSocket(__sock_acceptor.getSocket());
      __sock_acceptor.close();
    }
  }
  
  inline void setNoRegister() {
    __no_register = true;
  }
  
  virtual void start() {
    SNetReactor::getInstance().registerSocket(__sock_acceptor.getSocket(), this, NetReactor::NR_POLLIN);
  }
  
protected:
  virtual NetResult receiveData(int) {
    SNetReactor::getInstance().unregisterSocket(__sock_acceptor.getSocket());
    
    try {
      SimpleConnection* connection = dynamic_cast<SimpleConnection*>(getConnection());
      __sock_acceptor.accept(connection->getSocket());
      if (__no_register) {
        connection->setNoRegister();
      }
      notifyAccepted();
    } catch(const ConnectionException& e) {
      cerr << e << endl;
      setException(e);
      notifyAccepted();
    }
    
    __sock_acceptor.close();
    return VOID;
  }
  
private:
  SockAcceptor      __sock_acceptor;
  bool              __no_register;
};
/*======================================== Simple TCP Transport ========================================*/


/*======================================== NetStream Transport ========================================*/
/** NetStream network connection.
 * This class transmits NMM buffers through a NetStream.
 */
template<class _NetStream>
class TNetStreamConnection :
  public SimpleConnection
{
public:
  TNetStreamConnection()
    : __net_stream(getSocket())
  {}
  
  virtual ~TNetStreamConnection()
  {}
  
  inline _NetStream& getNetStream() { return __net_stream; }
  
protected:
  virtual void sendBuffer(Buffer* buffer) {
    try {
      u_int32_t used = buffer->getUsed();
      __net_stream.resetWriteCounter();
      __net_stream << used;
      __net_stream.writeBuffer(buffer->getData(), buffer->getUsed());
      __net_stream.writeData();
      countBytes(__net_stream.getWriteCounter());
    } catch(const TransmissionException& e) {
      cerr << e << endl;
    }
  }
  
  virtual Buffer* recvBuffer(ssize_t /*size*/ = 0) {
    Buffer* buffer = NULL;
    try {
      u_int32_t used = 0;
      __net_stream.resetReadCounter();
      __net_stream.readData();
      __net_stream >> used;
      buffer = getBufferManager()->requestBuffer(used);
      buffer->setUsed(used);
      __net_stream.readBuffer(buffer->getData(), buffer->getUsed());
      countBytes(__net_stream.getReadCounter());
      return buffer;
    } catch(const TransmissionException& e) {
      cerr << e << endl;
      if (buffer) {
        buffer->release();
      }
      disconnect();
      return NULL;
    } catch(const ConnectionClosed& e) {
      if (buffer) {
        buffer->release();
      }
      disconnect();
      return NULL;
    }
  }
  
private:
  _NetStream __net_stream;
};


/** NetStream network acceptor.
 * This class accepts a NetStream TCP connection.
 */
template<class _NetStream>
class TNetStreamAcceptor :
  public SimpleAcceptor
{
public:
  TNetStreamAcceptor(IMessageReceiver* receiver, BufferManager* manager, unsigned int port,
                    Connection* connection = new TNetStreamConnection<_NetStream>)
    : SimpleAcceptor(receiver, manager, port, connection)
  {}
  
  virtual ~TNetStreamAcceptor()
  {}  
};

typedef TNetStreamConnection<UnbufferedNetStream> NetStreamConnection;
typedef TNetStreamAcceptor<UnbufferedNetStream> NetStreamAcceptor;
typedef TNetStreamConnection<BufferedNetStream> BufferedNetStreamConnection;
typedef TNetStreamAcceptor<BufferedNetStream> BufferedNetStreamAcceptor;
/*======================================== NetStream Transport ========================================*/


/*======================================== PacketNetStream Transport ========================================*/
/** PacketNetStream network connection.
 * This class transmits NMM buffers through a PacketNetStream.
 */
class PacketNetStreamConnection :
  public SimpleConnection
{
public:
  PacketNetStreamConnection()
    : __packet_stream(getSocket())
  {
    //__packet_stream.setPacketSize(16384);
  }
  
  virtual ~PacketNetStreamConnection()
  {}
  
  inline PacketNetStream& getPacketNetStream() { return __packet_stream; }
  
protected:
  virtual void sendBuffer(Buffer* buffer) {
    try {
      u_int32_t used = buffer->getUsed();
      __packet_stream << used;
      __packet_stream.writeBuffer(buffer->getData(), buffer->getUsed());
      __packet_stream.writeData();
      countBytes(__packet_stream.getTotalSentBytes());
      __packet_stream.resetStatistics();
    } catch(const TransmissionException& e) {
      cerr << e << endl;
    }
  }
  
  virtual Buffer* recvBuffer(ssize_t /*size*/ = 0) {
    Buffer* buffer = NULL;
    try {
      u_int32_t used = 0;
      __packet_stream.readData();
      __packet_stream >> used;
      buffer = getBufferManager()->requestBuffer(used);
      buffer->setUsed(used);
      __packet_stream.readBuffer(buffer->getData(), buffer->getUsed());
      countBytes(__packet_stream.getTotalReceivedBytes());
      __packet_stream.resetStatistics();
      return buffer;
    } catch(const TransmissionException& e) {
      cerr << e << endl;
      if (buffer) {
        buffer->release();
      }
      disconnect();
      return NULL;
    } catch(const ConnectionClosed& e) {
      if (buffer) {
        buffer->release();
      }
      disconnect();
      return NULL;
    }
  }
  
  virtual NetResult receiveData(int) {
    try {
      if (__packet_stream.receivePacket()) {
        Buffer* buffer = recvBuffer(getBufferSize());
        if (buffer) {
          sendMessage(buffer);
        }
      }
      return VOID;
    } catch(const ConnectionClosed& e) {
      disconnect();
      return VOID;
    }
  }
  
private:
  PacketNetStream __packet_stream;
};


/** PacketNetStream network acceptor.
 * This class accepts a NetStream TCP connection.
 */
class PacketNetStreamAcceptor :
  public SimpleAcceptor
{
public:
  PacketNetStreamAcceptor(IMessageReceiver* receiver, BufferManager* manager, unsigned int port,
                       Connection* connection = new PacketNetStreamConnection)
    : SimpleAcceptor(receiver, manager, port, connection)
  {}
  
  virtual ~PacketNetStreamAcceptor()
  {}  
};
/*======================================== PacketNetStream Transport ========================================*/


/*======================================== ValueStream Transport ========================================*/
/** ValueStream network connection.
 * This class transmits NMM buffers through a ValueStream
 */
template<class _IValueStream, class _OValueStream, class _NetStream>
class TValueStreamConnection :
  public TNetStreamConnection<_NetStream>
{
public:
  TValueStreamConnection() {
    __i_stream.setIPartnerStream(&getNetStream());
    __o_stream.setOPartnerStream(&getNetStream());
  }
  
  virtual ~TValueStreamConnection()
  {}
  
protected:
  virtual void sendBuffer(Buffer* buffer) {
    try {
      u_int32_t used = buffer->getUsed();
      getNetStream().resetWriteCounter();
      __o_stream << used;
//       int f0 = 0;
//       int f1 = 1;
//       for (unsigned i=0; i<buffer->getUsed(); ++i) {
//         buffer->getData()[i] = static_cast<char>(f0);
//         int f = f0+f1;
//         f0=f1;
//         f1=f;
//       }
      __o_stream << *buffer;
      //__o_stream.writeBuffer(buffer->getData(), buffer->getUsed());
      __o_stream.writeData();
      countBytes(getNetStream().getWriteCounter());
    } catch(const TransmissionException& e) {
      cerr << e << endl;
    }
  }
  
  virtual Buffer* recvBuffer(ssize_t /*size*/ = 0) {
    Buffer* buffer = NULL;
    try {
      u_int32_t used = 0;
      getNetStream().resetReadCounter();
      __i_stream.readData();
      __i_stream >> used;
      buffer = getBufferManager()->requestBuffer(used);
      buffer->setUsed(used);
      buffer->createFromStream(__i_stream);
//       int f0 = 0;
//       int f1 = 1;
//       for (unsigned i=0; i<buffer->getUsed(); ++i) {
//         if (buffer->getData()[i] != static_cast<char>(f0)) {
//           cerr << "inconsistent data!" << endl;
//           disconnect();
//           buffer->release();
//           return NULL;
//         }
//         int f = f0+f1;
//         f0=f1;
//         f1=f;
//       }
      //__i_stream.readBuffer(buffer->getData(), buffer->getUsed());
      countBytes(getNetStream().getReadCounter());
      return buffer;
    } catch(const TransmissionException& e) {
      cerr << e << endl;
      if (buffer) {
        buffer->release();
      }
      disconnect();
      return NULL;
    } catch(const ConnectionClosed& e) {
      if (buffer) {
        buffer->release();
      }
      disconnect();
      return NULL;
    }
  }
  
private:
  _IValueStream __i_stream;
  _OValueStream __o_stream;
};

/** ValueStream network acceptor.
 * This class accepts a ValueStream TCP connection.
 */
template<class _IValueStream, class _OValueStream, class _NetStream>
class TValueStreamAcceptor :
  public TNetStreamAcceptor<_NetStream>
{
public:
  typedef TValueStreamConnection<_IValueStream, _OValueStream, _NetStream> ValueStreamConnection;
  
  TValueStreamAcceptor(IMessageReceiver* receiver, BufferManager* manager, unsigned int port,
                      Connection* connection = new ValueStreamConnection)
    : TNetStreamAcceptor<_NetStream>(receiver, manager, port, connection)
  {}
  
  virtual ~TValueStreamAcceptor()
  {}  
};

typedef TValueStreamAcceptor<MagicNumberIValueStream, MagicNumberOValueStream, UnbufferedNetStream> MagicNumberStreamAcceptor;
typedef TValueStreamConnection<MagicNumberIValueStream, MagicNumberOValueStream, UnbufferedNetStream> MagicNumberStreamConnection;
typedef TValueStreamAcceptor<MagicNumberIValueStream, MagicNumberOValueStream, BufferedNetStream> MagicNumberStream2Acceptor;
typedef TValueStreamConnection<MagicNumberIValueStream, MagicNumberOValueStream, BufferedNetStream> MagicNumberStream2Connection;
/*======================================== ValueStream Transport ========================================*/


/*======================================== ValueStream Transport with PacketNetStream ========================================*/
/** ValueStream network connection.
 * This class transmits NMM buffers through a ValueStream
 */
template<class _IValueStream, class _OValueStream>
class TPacketValueStreamConnection :
  public PacketNetStreamConnection
{
public:
  TPacketValueStreamConnection() {
    __i_stream.setIPartnerStream(&getPacketNetStream());
    __o_stream.setOPartnerStream(&getPacketNetStream());
  }
  
  virtual ~TPacketValueStreamConnection()
  {}
  
protected:
  virtual void sendBuffer(Buffer* buffer) {
    try {
      u_int32_t used = buffer->getUsed();
      __o_stream << used;
      __o_stream << *buffer;
      //__o_stream.writeBuffer(buffer->getData(), buffer->getUsed());
      __o_stream.writeData();
      countBytes(getPacketNetStream().getTotalSentBytes());
      getPacketNetStream().resetStatistics();
    } catch(const TransmissionException& e) {
      cerr << e << endl;
    }
  }
  
  virtual Buffer* recvBuffer(ssize_t /*size*/ = 0) {
    Buffer* buffer = NULL;
    try {
      u_int32_t used = 0;
      __i_stream.readData();
      __i_stream >> used;
      buffer = getBufferManager()->requestBuffer(used);
      buffer->setUsed(used);
      buffer->createFromStream(__i_stream);
      //__i_stream.readBuffer(buffer->getData(), buffer->getUsed());
      countBytes(getPacketNetStream().getTotalReceivedBytes());
      getPacketNetStream().resetStatistics();
      return buffer;
    } catch(const TransmissionException& e) {
      cerr << e << endl;
      if (buffer) {
        buffer->release();
      }
      disconnect();
      return NULL;
    } catch(const ConnectionClosed& e) {
      if (buffer) {
        buffer->release();
      }
      disconnect();
      return NULL;
    }
  }
  
private:
  _IValueStream __i_stream;
  _OValueStream __o_stream;
};

/** ValueStream network acceptor.
 * This class accepts a ValueStream TCP connection.
 */
template<class _IValueStream, class _OValueStream>
class TPacketValueStreamAcceptor :
  public PacketNetStreamAcceptor
{
public:
  typedef TPacketValueStreamConnection<_IValueStream, _OValueStream> PacketValueStreamConnection;
  
  TPacketValueStreamAcceptor(IMessageReceiver* receiver, BufferManager* manager, unsigned int port,
                             Connection* connection = new PacketValueStreamConnection)
    : PacketNetStreamAcceptor(receiver, manager, port, connection)
  {}
  
  virtual ~TPacketValueStreamAcceptor()
  {}  
};

typedef TPacketValueStreamAcceptor<MagicNumberIValueStream, MagicNumberOValueStream> PacketMagicNumberStreamAcceptor;
typedef TPacketValueStreamConnection<MagicNumberIValueStream, MagicNumberOValueStream> PacketMagicNumberStreamConnection;
/*======================================== ValueStream Transport with PacketNetStream ========================================*/


/*======================================== System Call Benchmark ========================================*/
void runSystemCallBenchmark(unsigned int num, ssize_t bytes, string host, unsigned int port, bool retry, unsigned int repeat) {
  double avg_elapsed=0, avg_payload_over_time=0, avg_transmit_over_time=0;
  ssize_t avg_total_bytes=0, avg_total_transmit=0;
  unsigned int iterations = repeat;
  
  printf(" elapsed       | payload       | payload       | transmit      | transmit      \n");
  printf(" (s)           | (kb)          | (kb/s)        | (kb)          | (kb/s)        \n");
  printf("===============================================================================\n");
  
  if (repeat > 1) {
    retry = true;
  }
  while (repeat--) {
    ssize_t total_bytes = 0;     ///< Total buffer bytes
    ssize_t total_transmit = 0;  ///< Total bytes transmitted over the network
    timeval start_time;
    timeval end_time;
    
    if (host.empty()) {
      
      BufferReceiver receiver;
      BufferManager* manager = new B_Allocator;
      
      SimpleAcceptor acceptor(&receiver, manager, port);
      acceptor.setNoRegister();
      acceptor.start();
      GLOBAL_MESSAGE_STREAM("Listening on port " << port << endl);
      
      SimpleConnection* conn = static_cast<SimpleConnection*>(acceptor.accept());
      conn->start();
      GLOBAL_MESSAGE_STREAM("Accepted connection" << endl);
      
      Buffer* buffer = manager->requestBuffer(bytes);
      try {  
        buffer->setUsed(bytes);
        gettimeofday(&start_time, NULL);
        char* offset = buffer->getData();
        unsigned int used = buffer->getUsed();
        for (unsigned int i=0; i<num; ) {
          ssize_t bytes = conn->getSocket().recv(offset, used, MSG_NOSIGNAL);
          total_transmit += bytes;
          offset += bytes;
          used -= bytes;
          if (used <= 0) {
            offset -= buffer->getUsed();
            used += buffer->getUsed();
            ++i;
          }
        }
        gettimeofday(&end_time, NULL);
      } catch(const TransmissionException& e) {
        cerr << e << endl;
        break;
      }
      buffer->release();
      GLOBAL_MESSAGE_STREAM("Connection closed" << endl);
      
      total_bytes = total_transmit;
      
      delete conn;
      
    } else {
      
      SimpleConnection conn;
      INetAddr server_addr(host.c_str(), port);
      GLOBAL_MESSAGE_STREAM("Connecting to " << server_addr.toString() << endl);
      
      for( ;; ) {
        try {
          conn.connect(server_addr);
          GLOBAL_MESSAGE_STREAM("Connected to " << server_addr.toString() << endl);
        } catch(ConnectionException& e) {
          if (!retry) {
            cerr << e << endl;
            return;
          } else {
            continue;
          }
        }
        break;
      }
      
      BufferManager* manager = new B_Allocator;
      Buffer* buffer = manager->requestBuffer(bytes);
      buffer->setUsed(bytes);
      gettimeofday(&start_time, NULL);
      for (unsigned int i=0; i<num; ++i) {
        try {
          total_transmit += conn.getSocket().send(buffer->getData(), buffer->getUsed(), MSG_NOSIGNAL);
        } catch(const TransmissionException& e) {
          cerr << e << endl;
          break;
        }
      }
      gettimeofday(&end_time, NULL);
      buffer->release();
      
      total_bytes = total_transmit;
    }
    
    end_time.tv_usec -= start_time.tv_usec;
    if (end_time.tv_usec < 0) {
      end_time.tv_usec += 1000000;
      end_time.tv_sec--;
    }
    end_time.tv_sec -= start_time.tv_sec;
    double elapsed = end_time.tv_sec + end_time.tv_usec * 0.000001;
//     cerr << "Elapsed time: " << elapsed << "s" << endl
//          << "Payload     : " << total_bytes / 1024 << "kb"
//          << " (" << total_bytes / 1024 / elapsed << "kb/s)" << endl
//          << "Transmitted : " << total_transmit / 1024 << "kb"
//          << " (" << total_transmit / 1024 / elapsed << "kb/s)" << endl
//          << endl;
    double payload_over_time = static_cast<double>(total_bytes) / 1024.0 / elapsed;
    double transmit_over_time = static_cast<double>(total_transmit) / 1024.0 / elapsed;
    printf("%15.4f|%15d|%15.4f|%15d|%15.4f\n", elapsed, total_bytes, payload_over_time, total_transmit, transmit_over_time);

    avg_elapsed += elapsed;
    avg_total_bytes += total_bytes;
    avg_payload_over_time += payload_over_time;
    avg_total_transmit += total_transmit;
    avg_transmit_over_time += transmit_over_time;
  }
  
  printf("-------------------------------------------------------------------------------\n");
  avg_elapsed /= iterations;
  avg_total_bytes /= iterations;
  avg_payload_over_time /= iterations;
  avg_total_transmit /= iterations;
  avg_transmit_over_time /= iterations;
  printf("%15.4f|%15d|%15.4f|%15d|%15.4f\n", avg_elapsed, avg_total_bytes, avg_payload_over_time, avg_total_transmit, avg_transmit_over_time);
}
/*======================================== System Call Benchmark ========================================*/


/*======================================== Main Program ========================================*/
void usage(char* name) {
    cerr << "Usage:" << endl
         << "  " << name << " [options] [benchmark number]" << endl
         << endl
         << "Options:" << endl
         << "  -n : Sets the number of buffers to transmit" << endl
         << "  -b : Sets the number of bytes per buffer" << endl
         << "  -H : Address to send the buffers to" << endl
         << "  -p : Listen for connections on this port" << endl
         << "  -r : Repeat benchmark test n times " << endl
         << "  -s : Synchronous receiver (don't use NetReactor)" << endl
         << "  -v : increase verbositiy (may be given multiple times)" << endl
         << "  -h : Show this help" << endl << endl
         << "Benchmark Nr.: " << endl
         << " 0 : system call benchmark" << endl
         << " 1 : send buffers directly through a SockStream" << endl
         << " 2 : send buffers through a NetStream" << endl
         << " 3 : send buffers through a MagicNumberValueStream and NetStream" << endl
         << " 4 : send buffers through a BufferedNetStream" << endl
         << " 5 : send buffers through a MagicNumberValueStream and BufferedNetStream" << endl
         << " 6 : send buffers through a PacketNetStream" << endl
         << " 7 : send buffers through a MagicNumberValueStream and PacketNetStream" << endl
         << "For all benchmarks you need to run one server (using -p option)" << endl
         << "and one client (using -H option) that connects to the server." << endl
         << "If you do not specify a benchmark, all benchmarks are run." << endl
         << endl;
}

int main(int argc, char** argv) {
  try {
    int          verbose = 0;
    unsigned int num     = 1000;
    ssize_t      bytes   = 1024;
    string       host    = "";
    unsigned int port    = 0;
    unsigned int repeat  = 1;
    unsigned int benchmark;
    bool         sync    = false;
    
    const char *opt_string = "vhn:b:H:p:r:R:s";
    struct option options [] = {
      {"num",         1,NULL,'n'},
      {"bytes",       1,NULL,'b'},
      {"host",        1,NULL,'H'},
      {"port",        1,NULL,'p'},
      {"repeat",      1,NULL,'r'},
      {"sync",        0,NULL,'s'},
      {"verbose",     0,NULL,'v'},
      {"help",        0,NULL,'h'},
      {NULL,          0,NULL,0}
    };
    int ch,long_option_index;
    
    while((ch=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
      // which option did we get?
      switch (ch) {
      case 'v':
        verbose++;
        break;
      case 'p':
        if (optarg) {
          port = atoi(optarg);
          if (!port) {
            usage(argv[0]);
            throw Exception("Invalid port number");
          }
        }
        break;
      case 'b':
        if (optarg) {
          bytes = atoi(optarg);
          if (bytes > MAX_BUFFER_SIZE) {
            throw Exception("Buffer size too large");
          }
        }
        break;
      case 'n':
        if (optarg) {
          num = atoi(optarg);
        }
        break;
      case 'r':
        if (optarg) {
          repeat = atoi(optarg);
        }
        break;
      case 'H':
        if(optarg) {
          string server_addr = string(optarg);
          string::size_type colon_pos = server_addr.find_first_of(':');
          if (colon_pos == string::npos) {
            host = server_addr;
            port = ClientRegistry::default_port;
          } else {
            host = server_addr.substr(0, colon_pos);
            port = atoi(server_addr.substr(colon_pos+1, string::npos).c_str());
          }
          if (!port) {
            usage(argv[0]);
            throw Exception("Invalid port number");
          }
        }
        break;
      case 's':
        sync = true;
        break;
      case 'h':
        usage(argv[0]);
        exit(0);
      default:
        usage(argv[0]);
        throw Exception("Invalid option");
      }
    }
    
    /* handle *Stream output */
    if (verbose < 1) {
      NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose < 2) {
      NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose < 3) {
      NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose < 4) {
      NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
    }
    
    /* check for benchmark number */
    bool all_benchmarks;
    if ( argc - optind < 1 ) {
      benchmark = 1;
      all_benchmarks = true;
    } else {
      benchmark = atoi(argv[optind]);
      all_benchmarks = false;
    }
    
    /* run benchmarks */
    /* if all benchmarks are run and the connection to the server fails, the client retries infinite times
       to give the server time enough to initialize itself */
    switch (benchmark) {
    case 0:
      cerr << "0: System call benchmark" << endl;
      runSystemCallBenchmark(num, bytes, host, port, all_benchmarks, repeat);
      if (!all_benchmarks) {
        break;
      }
      if (repeat > 1) {
        cerr << endl;
      }
    case 1:
      cerr << "1: Simple TCP Connection" << endl;
      runBenchmark<SimpleAcceptor, SimpleConnection>(num, bytes, host, port, all_benchmarks, repeat, sync);
      if (!all_benchmarks) {
        break;
      }
      if (repeat > 1) {
        cerr << endl;
      }
    case 2:
      cerr << "2: NetStream over TCP" << endl;
      runBenchmark<NetStreamAcceptor, NetStreamConnection>(num, bytes, host, port, all_benchmarks, repeat, sync);
      if (!all_benchmarks) {
        break;
      }
      if (repeat > 1) {
        cerr << endl;
      }
    case 3:
      cerr << "3: MagicNumberValueStream over NetStream over TCP" << endl;
      runBenchmark<MagicNumberStreamAcceptor, MagicNumberStreamConnection>(num, bytes, host, port, all_benchmarks, repeat, sync);
      if (!all_benchmarks) {
        break;
      }
      if (repeat > 1) {
        cerr << endl;
      }
    case 4:
      cerr << "4: BufferedNetStream over TCP" << endl;
      runBenchmark<BufferedNetStreamAcceptor, BufferedNetStreamConnection>(num, bytes, host, port, all_benchmarks, repeat, sync);
      if (!all_benchmarks) {
        break;
      }
      if (repeat > 1) {
        cerr << endl;
      }
    case 5:
      cerr << "5: MagicNumberValueStream over BufferedNetStream over TCP" << endl;
      runBenchmark<MagicNumberStream2Acceptor, MagicNumberStream2Connection>(num, bytes, host, port, all_benchmarks, repeat, sync);
      if (!all_benchmarks) {
        break;
      }
      if (repeat > 1) {
        cerr << endl;
      }
    case 6:
      cerr << "6: PacketNetStream over TCP" << endl;
      runBenchmark<PacketNetStreamAcceptor, PacketNetStreamConnection>(num, bytes, host, port, all_benchmarks, repeat, sync);
      if (!all_benchmarks) {
        break;
      }
      if (repeat > 1) {
        cerr << endl;
      }
    case 7:
      cerr << "7: MagicNumberValueStream over PacketNetStream over TCP" << endl;
      runBenchmark<PacketMagicNumberStreamAcceptor, PacketMagicNumberStreamConnection>(num, bytes, host, port, all_benchmarks, repeat, sync);
      if (!all_benchmarks) {
        break;
      }
      if (repeat > 1) {
        cerr << endl;
      }
    default:
      if (!all_benchmarks) {
        usage(argv[0]);
        throw Exception("Invalid benchmark number");
      }
    }
    
    return 0;
  } catch(const Exception& e) {
    cerr << e << endl;
    return 1;
  }
}
/*======================================== Main Program ========================================*/
