/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#include "CompositeStrategy.hpp"

#include "ConnectorAddress.hpp"
#include "ConnectorAcceptor.hpp"
#include "ConnectionManager.hpp"
#include "TSFactory.hpp"

#include "nmm/interfaces/base/IConnectionManager.hpp"
#include "nmm/interfaces/net/IObjectAcceptor.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/ProxyObject.hpp"

// XXX: fw: HACK FOR RELEASE
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

#include <cassert>
#include <cstdlib>

namespace NMM {
  
  CompositeAddress::CompositeAddress(const NMMReference& ref)
    : TSAddress(ref, CompositeStrategy::getStaticStrategyType(), TypeInfo("CompositeAddress", "NMM"), 0),
      __master_address(0),
      __buffer_address(0),
      __event_address(0)
  {
    insertArgument(&__master_address);
    insertArgument(&__buffer_address);
    insertArgument(&__event_address);
  }
  
  CompositeAddress::CompositeAddress(const NMMReference& ref, const TSAddress& master_address,
                                     const TSAddress& buffer_address, const TSAddress& event_address)
    : TSAddress(ref, CompositeStrategy::getStaticStrategyType(), TypeInfo("CompositeAddress", "NMM"), 0),
      __master_address(master_address.copy()),
      __buffer_address(buffer_address.copy()),
      __event_address(event_address.copy())
  {
    insertArgument(&__master_address);
    insertArgument(&__buffer_address);
    insertArgument(&__event_address);
    
    setBindingType(__buffer_address.getRValue()->getBindingType() & __event_address.getRValue()->getBindingType());
  }
  
  CompositeAddress::CompositeAddress(const CompositeAddress& address)
    : TSAddress(address),
      __master_address(address.__master_address.getRValue() ? address.__master_address.getRValue()->copy() : 0),
      __buffer_address(address.__buffer_address.getRValue() ? address.__buffer_address.getRValue()->copy() : 0),
      __event_address(address.__event_address.getRValue() ? address.__event_address.getRValue()->copy() : 0)
  {
    insertArgument(&__master_address);
    insertArgument(&__buffer_address);
    insertArgument(&__event_address);
    
    if (__buffer_address.getRValue()) {
      if (__event_address.getRValue()) {
        setBindingType(__buffer_address.getRValue()->getBindingType() & __event_address.getRValue()->getBindingType());
      } else {
        setBindingType(__buffer_address.getRValue()->getBindingType());
      }
    } else {
      if (__event_address.getRValue()) {
        setBindingType(__event_address.getRValue()->getBindingType());
      } else {
        setBindingType(0);
      }
    }
  }
  
  CompositeAddress* CompositeAddress::copy() const {
    return new CompositeAddress(*this);
  }
  
  
  InstreamHandler::InstreamHandler() throw()
    : __parent(NULL),
      __child(NULL),
      __send_timestamp(),
      __blocked(false),
      __send_mutex(),
      __pending(NULL),
      __transmission_delay(),
      __recv_timestamp(),
      __recv_mutex(ThreadMutex::RECURSIVE),
      __is_current(false)
  {
    __transmission_delay.setAverageInterval(10);
  }
  
  void InstreamHandler::connect(CompositeStrategy* parent, TransportStrategy* child) throw() {
    /* disconnect old child first */
    if(__child) {
      disconnect();
    }
    
    /* set parent and child */
    __parent = parent;
    __child = child;
    
    /* set child's signal and message receiver */
    getTransportStrategy().setSignalReceiver(this);
    getTransportStrategy().setMessageReceiver(this);
    getTransportStrategy().setEventReceiver(parent);
    
    /* reset sender state */
    __blocked = false;
    __send_timestamp = Timestamp();
    __send_timestamp.stream_counter = 0;
    
    /* reset receiver state */
    __pending = NULL;
    __recv_timestamp = Timestamp();
    __recv_timestamp.stream_counter = 0;
  }
  
  void InstreamHandler::disconnect() {
    if(__child) {
      /* unset child's signal and message receiver */
      getTransportStrategy().setSignalReceiver(0);
      getTransportStrategy().setMessageReceiver(0);
      
      /* destroy child */
      __child->scheduleDestruction();
      __child = NULL;
    }
    
    if(__pending) {
      __pending->release();
    }
  }
  
  void InstreamHandler::unblock() {
    /* remove pending message */
    __pending = NULL;
    
    /* unblock sub binding */
    if(!__is_current) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Sending FREE_QUEUE_SLOT to sub binding" << std::endl);
      getTransportStrategy().receiveSignal(FREE_QUEUE_SLOT); // XXX: direction
    }
  }
  
  void InstreamHandler::receiveSignal(Signal signal) {
    assert(__parent);
    assert(__child);
    
    /* check if we are no longer blocked */
    if(signal == FREE_QUEUE_SLOT || signal == FREE_UPSTR_QUEUE_SLOT) {
      MutexGuard mg(__send_mutex);
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Received FREE_QUEUE_SLOT signal from sub binding" << std::endl);
      __blocked = false;
    }
    
    /* forward signal to parent */
    getParent().receiveChildSignal(signal, &getTransportStrategy());
  }
  
  Result InstreamHandler::receiveMessage(Message* message, const IMessageSender* sender /*= 0*/) {
    MutexGuard mg(__recv_mutex);
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: Received message from sub binding" << std::endl);
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: Sequence no. " << message->getTimestamp().stream_counter << std::endl);
    
    assert(&getParent());
    assert(&getTransportStrategy());
    assert(sender == &getTransportStrategy());
    
    /* BEGIN> compute average transmission delay */
    // XXX: we can't use the timestamp! replace this with sender reports!
    Time current_time = getParent().getTime();
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: *Current time       : " << current_time << std::endl);
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: *Send time          : " << message->getTimestamp().sync_time << std::endl);
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: *Transmission delay : " << current_time - message->getTimestamp().sync_time << std::endl);
    if(message->getTimestamp().sync_time.sec != 0) { // XXX: this is a hack to avoid mysteriously broken timestamps.
      Interval transmission_delay = current_time - message->getTimestamp().sync_time;
      __transmission_delay.put(transmission_delay);
    }
    /* END< compute average transmission delay */    
    
    /* store receiver timestamp */
    __recv_timestamp = message->getTimestamp();
    
    /* store pending message and return BLOCKED */
    __pending = message;
    
    /* notify the multiplexer */
    __is_current = true;
    getParent().run();
    __is_current = false;
    
    /* if message is still pending, return BLOCKED, otherwise SUCCESS */
    if(__pending) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Message still pending, blocking sub binding" << std::endl);
      return BLOCKED;
    } else {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Unblocking sub binding" << std::endl);
      return SUCCESS;
    }
  }
  
  Result InstreamHandler::forwardMessage(Message* message, const IMessageSender* sender /*= 0*/) {
    MutexGuard mg(__send_mutex);
    
    if(__blocked) {
      /* SHOULD NOT HAPPEN */
      GLOBAL_ERROR_STREAM2("CompositeStrategy: Sub binding already blocked!" << std::endl);
      message->release();
      return FAILURE;
    }
    
    /* forward message */
    Result result = getTransportStrategy().receiveMessage(message, sender);
    if(result == BLOCKED) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Child strategy returned BLOCKED, which means that its StreamQueue is full now" << std::endl);
      __blocked = true;
    }
    
    /* forward result */
    return result;
  }
  
  
  CompositeStrategy::CompositeStrategy()
    : __remote_address(NULL),
      __local_address(NULL),
      __send_mutex(),
      __recv_mutex(),
      __timeout_cond(__recv_mutex),
      __mode(ICompositeStrategy::MODE_RELIABLE_RELIABLE),
      __loss_interval(0),
      __send_seq_no(0),
      __wait_event_ack(false),
      __wait_for_timeout(false),
      __timeout_stream_counter(0),
      __recv_timestamp(),
      __blocked(false),
      __last_timeout_stream_counter(-1)
  {
    GLOBAL_DEBUG_STREAM2("Created a CompositeStrategy" << std::endl);
    
    /* create local address */
    __local_address = new ConnectorAddress("CompositeStrategy", 0);
    __local_address->setConnectionManager(*SConnectionManager::getInstance().getInterface<IConnectionManager>());
    __local_address->setTSReference(getReference());
  }
  
  CompositeStrategy::~CompositeStrategy() {
    GLOBAL_DEBUG_STREAM2("Destroying CompositeStrategy" << std::endl);
    
    /* destroy child strategies */
    setBufferStrategy(0);
    setEventStrategy(0);
    
    /* destroy addresses */
    delete __remote_address;
    delete __local_address;
  }
  
  void CompositeStrategy::setMode(const ICompositeStrategy::Mode& mode) {
    __mode = mode;
  }
  
  void CompositeStrategy::setLossInterval(const unsigned int& loss_interval) {
    __loss_interval = loss_interval;
  }
  
  void CompositeStrategy::scheduleDestruction() {
    /* destroy self */
    delete this;
  }
  
  void CompositeStrategy::setBufferStrategyType(const string& strategy_type)
    throw (const BindingException&)
  {
    GLOBAL_DEBUG_STREAM1("CompositeStrategy: set buffer strategy type: " << strategy_type << std::endl);
    
    /* create strategy */
    TransportStrategy* strategy = STSFactory::getInstance().createStrategy(strategy_type);
    if (!strategy) {
      GLOBAL_ERROR_STREAM("Failed to create child strategy" << endl);
      throw BindingException("Failed to create transport strategy of type " + strategy_type);
    }
    
    /* set as buffer child strategy */
    setBufferStrategy(strategy);
  }
  
  void CompositeStrategy::setEventStrategyType(const string& strategy_type)
    throw (const BindingException&)
  {
    GLOBAL_DEBUG_STREAM1("CompositeStrategy: set event strategy type: " << strategy_type << std::endl);
    
    /* create strategy */    
    TransportStrategy* strategy = STSFactory::getInstance().createStrategy(strategy_type);
    if (!strategy) {
      GLOBAL_ERROR_STREAM("Failed to create child strategy" << endl);
      throw BindingException("Failed to create transport strategy of type " + strategy_type);
    }

    /* set as event child strategy */    
    setEventStrategy(strategy);
  }
  
  void CompositeStrategy::setBufferStrategy(TransportStrategy* strategy) {
    MutexGuard mg1(__send_mutex);
    MutexGuard mg2(__recv_mutex);
    
    /* destroy old buffer child strategy */
    __sub_binding[Message::BUFFER].connect(this, strategy);
  }
  
  void CompositeStrategy::setEventStrategy(TransportStrategy* strategy) {
    MutexGuard mg1(__send_mutex);
    MutexGuard mg2(__recv_mutex);
    
    /* destroy old buffer child strategy */
    __sub_binding[Message::EVENT].connect(this, strategy);
  }
  
  void CompositeStrategy::setAddress(const TSAddress& address)
    throw (const BindingException&)
  {
    GLOBAL_DEBUG_STREAM1("CompositeStrategy: set address: " << address.getTypeName() << std::endl);
    
    /* check if address is a CompositeAddress */
    const CompositeAddress* composite_address = dynamic_cast<const CompositeAddress*>(&address);
    if (composite_address) {
      /* check if master address is given */
      if (!composite_address->getMasterAddress()) {
        GLOBAL_ERROR_STREAM("CompositeStrategy: Composite address has no master address" << endl);
        throw BindingException("Composite address has no master address");
      }
      const TSAddress& master_address = *composite_address->getMasterAddress();
      
      /* the master address must be a ConnectorAddress */
      const ConnectorAddress* connector_address = dynamic_cast<const ConnectorAddress*>(&master_address);
      if(!connector_address) {
        GLOBAL_ERROR_STREAM("CompositeStrategy: Master address is not a ConnectorAddress" << endl);
        throw BindingException("Master address is not a ConnectorAddress");
      }
      
      /* check if both child addresses are given */
      if (!composite_address->getBufferAddress()) {
        GLOBAL_ERROR_STREAM("CompositeStrategy: Composite address has no buffer child address" << endl);
        throw BindingException("Composite address has no buffer child address");
      }
      if (!composite_address->getEventAddress()) {
        GLOBAL_ERROR_STREAM("CompositeStrategy: Composite address has no event child address" << endl);
        throw BindingException("Composite address has no event child address");
      }
      const TSAddress& ts_address1 = *composite_address->getBufferAddress();
      const TSAddress& ts_address2 = *composite_address->getEventAddress();
      
      /* the child addresses may not be CompositeAddresses */
      if(dynamic_cast<const CompositeAddress*>(&ts_address1) ||
         dynamic_cast<const CompositeAddress*>(&ts_address2)) {
        GLOBAL_ERROR_STREAM("CompositeStrategy: Child address of CompositeAddress must not be a CompositeAddress" << std::endl);
        throw BindingException("Child address of CompositeAddress must not be a CompositeAddress");
      }
      
      /* STEP 1: create a partner CompositeStrategy using the ConnectionManager */
      GLOBAL_DEBUG_STREAM1("CompositeStrategy: Create partner CompositeStrategy" << std::endl);
      ConnectorAddress* remote_addr = NULL;
      ConnectorAddress local_addr("CompositeStrategy", 0);
      local_addr.setConnectionManager(*SConnectionManager::getInstance().getInterface<IConnectionManager>());
      local_addr.setTSReference(getReference());
      SConnectionManager::getInstance().connect(&local_addr, remote_addr, connector_address->getConnectionManager());
      __remote_address = remote_addr;
      if(!__remote_address) {
        /* SHOULD NOT HAPPEN */
        GLOBAL_ERROR_STREAM2("CompositeStrategy: Could not get remote address" << endl);
        throw BindingException("Could not get remote address");
      }
      if(!__remote_address->getTSReference()) {
        /* SHOULD NOT HAPPEN */
        GLOBAL_ERROR_STREAM2("CompositeStrategy: Could not get NMM reference of partner strategy" << endl);
        throw BindingException("Could not get NMM reference of partner strategy");
      }
      
      /* STEP 2: connect partner strategy to partner object */
      if(master_address.getNMMReference()) {
        Event e(IObjectAcceptor::connectToObject_event, new TInValue<NMMReference>(master_address.getNMMReference()));
        Result result = sendReliableEvent(e);
        if(result != SUCCESS) {
          GLOBAL_ERROR_STREAM("CompositeStrategy: Failed to connect to object" << std::endl);
          if(e.getException()) {
            GLOBAL_ERROR_STREAM2("CompositeStrategy: " << *e.getException() << endl);
            throw BindingException("Failed to connect to object (" + e.getException()->getComment() + ")");
          } else {
            throw BindingException("Failed to connect to object");
          }
        }
      }
      
      /* STEP 3: create local strategies */
      setBufferStrategyType(ts_address1.getStrategyType());
      setEventStrategyType(ts_address2.getStrategyType());
      
      /* modify ts_address1 and ts_address2 so that the child strategies connect to the partner strategy */
      auto_ptr<TSAddress> tmp_ts_address1(ts_address1.copy());
      tmp_ts_address1->setNMMReference(__remote_address->getTSReference());
      auto_ptr<TSAddress> tmp_ts_address2(ts_address2.copy());
      tmp_ts_address2->setNMMReference(__remote_address->getTSReference());
      
      /* STEP 4: connect the child strategies */
      GLOBAL_DEBUG_STREAM("CompositeStrategy: connect buffer child strategy" << std::endl);
      getBufferStrategy()->setAddress(*tmp_ts_address1);
      GLOBAL_DEBUG_STREAM("CompositeStrategy: connect event child strategy" << std::endl);
      getEventStrategy()->setAddress(*tmp_ts_address2);
      
      /* STEP 5: connect the partner child strategies to the partner CompositeStrategy */
      Event event1(ICompositeStrategy::connectToCompositeBuffer_event);
      if(getBufferStrategy()->receiveReliableEvent(event1) != SUCCESS) {
        GLOBAL_DEBUG_STREAM2("CompositeStrategy: Failed to connect buffer child strategy to partner strategy" << std::endl);
        if(event1.getException()) {
          GLOBAL_ERROR_STREAM2("CompositeStrategy: " << *event1.getException() << std::endl);
          throw BindingException("Failed to connect buffer child strategy to partner strategy (" +
                                 event1.getException()->getComment() + ")");
        } else {
          throw BindingException("Failed to connect buffer child strategy to partner strategy");
        }
      }
      Event event2(ICompositeStrategy::connectToCompositeEvent_event);
      if(getEventStrategy()->receiveReliableEvent(event2) != SUCCESS) {
        GLOBAL_DEBUG_STREAM2("CompositeStrategy: Failed to connect event child strategy to partner strategy" << std::endl);
        if(event2.getException()) {
          GLOBAL_ERROR_STREAM2("CompositeStrategy: " << *event2.getException() << std::endl);
          throw BindingException("Failed to connect event child strategy to partner strategy (" +
                                 event2.getException()->getComment() + ")");
        } else {
          throw BindingException("Failed to connect event child strategy to partner strategy");
        }
      }
      
      /* BEGIN> Set mode */
      // XXX: HACK FOR RELEASE!
      if(!dynamic_cast<TCPStrategy*>(&__sub_binding[Message::BUFFER].getTransportStrategy())) {
        __mode = ICompositeStrategy::MODE_RELIABLE_UNRELIABLE;
      }
      /* END< Set mode */
      
      /* STEP 6: TEMPORARY: set partner mode */
      if(__remote_address) {
        Event mode_event(ICompositeStrategy::setMode_event, new TInValue<ICompositeStrategy::Mode>(__mode));
        __remote_address->getConnectionManager()->sendReliableConfigEvent(__remote_address->getTSReference(), mode_event);
      }
      
      return;
    }
    
    /* check if address is a ConnectorAddress */
    const ConnectorAddress* connector_address = dynamic_cast<const ConnectorAddress*>(&address);
    if(connector_address) {
      /* set connection manager interface */
      __remote_address = dynamic_cast<ConnectorAddress*>(connector_address->copy());
      
      /* establish binding */
      __remote_address->getConnectionManager()->getParentObject()->establishBinding();
      
      return;
    }
    
    /* invalid address type */
    GLOBAL_ERROR_STREAM2("CompositeStrategy: Invalid address type, expected a CompositeAddress or ConnectorAddress" << std::endl);
    throw BindingException("Invalid address type");
  }
  
  void CompositeStrategy::configureForInputJack(const Format& format) {
    /* set buffer manager here */
    getBufferStrategy()->setBufferManager(getBufferManager());
    getEventStrategy()->setBufferManager(getBufferManager());
    
    /* forward to child strategies */
    getBufferStrategy()->configureForInputJack(format);
    getEventStrategy()->configureForInputJack(format);
  }
  
  void CompositeStrategy::configureForOutputJack(const Format& format) {
    /* set buffer manager here */
    getBufferStrategy()->setBufferManager(getBufferManager());
    getEventStrategy()->setBufferManager(getBufferManager());
    
    /* forward to child strategies */
    getBufferStrategy()->configureForOutputJack(format);
    getEventStrategy()->configureForOutputJack(format);
  }
  
  void CompositeStrategy::connectToCompositeBuffer() throw(const BindingException&) {
    /* get sender strategy */
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if (!strategy) {
      throw BindingException("Failed to get current sender");
    }
    
    /* set child strategy */
    GLOBAL_DEBUG_STREAM("CompositeStrategy: Connecting partner for buffer strategy" << std::endl);
    setBufferStrategy(strategy);
  }
  
  void CompositeStrategy::connectToCompositeEvent() throw(const BindingException&) {
    /* get sender strategy */
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if (!strategy) {
      throw BindingException("Failed to get current sender");
    }
    
    /* set child strategy */
    GLOBAL_DEBUG_STREAM("CompositeStrategy: Connecting partner for event strategy" << std::endl);
    setEventStrategy(strategy);
  }
  
  const TSAddress& CompositeStrategy::getLocalAddress() const throw() {
    return *__local_address;
  }
  
  ITransportStrategy* CompositeStrategy::getBufferChildStrategy() const {
    return getBufferStrategy() ? getBufferStrategy()->getInterface<ITransportStrategy>() : NULL;
  }
  
  ITransportStrategy* CompositeStrategy::getEventChildStrategy() const {
    return getEventStrategy() ? getEventStrategy()->getInterface<ITransportStrategy>() : NULL;
  }
  
  Result CompositeStrategy::sendReliableEvent(Event& event) const {
    /* we must have a valid remote address */
    if (!__remote_address) {
      GLOBAL_ERROR_STREAM("CompositeStrategy: sendReliableEvent failed (No remote address)" << endl);
      return FAILURE;
    }
    
    try {
      /* send the event */
      return __remote_address->getConnectionManager()->sendReliableEvent(__remote_address->getTSReference(), event);
    } catch(BindingException& e) {
      GLOBAL_ERROR_STREAM("CompositeStrategy: sendReliableEvent failed" << std::endl);
      GLOBAL_ERROR_STREAM("CompositeStrategy:" << e << endl);
      return FAILURE;
    }
  }
  
  Result CompositeStrategy::receiveReliableEvent(Event& event, const IEventSender* sender) const {
    /* if this TransportStrategy is the sender don't send the event again */
    if(sender == this) {
      return SUCCESS;
    }
    
    /* if the event is a config event, then this strategy must handle it */
    if(find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
      return NMMObject::receiveEvent(event, sender);
    }
    
    /* forward the Event to the partner strategy */
    return sendReliableEvent(event);
  }
  
  Result CompositeStrategy::receiveMessage(Message* message, const IMessageSender* sender /*= 0*/) {
    MutexGuard mg(__send_mutex);
    
    /* adjust sender timestamp */
    if(!message->getTimestamp().is_valid) {
      Timestamp temp = message->getTimestamp();
      temp.stream_counter = ++__send_seq_no;
      message->setTimestamp(temp);
      GLOBAL_DEBUG_STREAM2(__send_seq_no << std::endl);
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Modified sequence number to " << message->getTimestamp().stream_counter << std::endl);
    } else {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Original sequence number is " << message->getTimestamp().stream_counter << std::endl);
      __send_seq_no = message->getTimestamp().stream_counter;
    }
    
    /* BEGIN> Set timestamp to current time of day (this is a temporary hack) */
    Timestamp ts = message->getTimestamp();
    ts.sync_time = getTime();
    message->setTimestamp(ts);
    /* END< */
    
    /* BEGIN> fake data loss (test) */
    if(__loss_interval != 0 && message->getType() == Message::BUFFER) {
      if(message->getTimestamp().stream_counter % __loss_interval == 0) {
        GLOBAL_WARNING_STREAM2("CompositeStrategy: FAKE DATA LOSS -- DROPPING BUFFER!! " << message->getTimestamp().stream_counter << std::endl);
        message->release();
        return SUCCESS;
      }
    }
    /* END< fake data loss (test) */
    
    /* message demultiplexing */
    assert(message->getType() == Message::BUFFER || message->getType() == Message::EVENT);
    InstreamHandler* destination = &__sub_binding[message->getType()];
    
    /* forward message to sub binding */
    Result result = destination->forwardMessage(message, sender);
    
    /* BEGIN> wait for event acknowledgement in unreliable mode */
    if(result != FAILURE && __mode == ICompositeStrategy::MODE_RELIABLE_UNRELIABLE &&
       message->getType() != Message::BUFFER) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Wait for event ACK" << std::endl);
      __wait_event_ack = true;
      
      return BLOCKED;
    }
    /* END< */
    
    /* forward result */
    return result;
  }
  
  Result CompositeStrategy::receiveEvent(const CEvent& event, const IEventSender* sender /*= 0*/ ) {
    /* if this strategy is the sender don't send the event again */
    if(sender == this) {
      return SUCCESS;
    }
    
    /* if the event is a config event, then this strategy must handle it */
    if(find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
      return NMMObject::receiveEvent(event, sender);
    }
    
    /* send this event */
    return sendEvent(event);
  }
  
  Result CompositeStrategy::receiveEvent(const Event& event, const IEventSender* sender /*= 0*/ ) {
    /* if this strategy is the sender don't send the event again */
    if(sender == this) {
      return SUCCESS;
    }
    
    /* if the event is a config event, then this strategy must handle it */
    if(find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
      return NMMObject::receiveEvent(event, sender);
    }
    
    /* send this event */
    return sendEvent(event);
  }
  
  Result CompositeStrategy::receiveEvent(const CEvent& event, const IEventSender* sender /*= 0*/ ) const {
    /* if this strategy is the sender don't send the event again */
    if(sender == this) {
      return SUCCESS;
    }
    
    /* if the event is a config event, then this strategy must handle it */
    if(find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
      return NMMObject::receiveEvent(event, sender);
    }
    
    /* send this event */
    return sendEvent(event);
  }
  
  Result CompositeStrategy::receiveEvent(const Event& event, const IEventSender* sender /*= 0*/ ) const {
    /* if this strategy is the sender don't send the event again */
    if(sender == this) {
      return SUCCESS;
    }
    
    /* if the event is a config event, then this strategy must handle it */
    if(find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
      return NMMObject::receiveEvent(event, sender);
    }
    
    /* send this event */
    return sendEvent(event);
  }
  
  Result CompositeStrategy::sendEvent(const CEvent& event) {
    /* simple event demultiplexing */
    return getEventStrategy()->sendEvent(event);
  }
  
  Result CompositeStrategy::sendEvent(const Event& event) {
    /* simple event demultiplexing */
    return getEventStrategy()->sendEvent(event);
  }
  
  Result CompositeStrategy::sendEvent(const CEvent& event) const {
    /* simple event demultiplexing */
    return getEventStrategy()->sendEvent(event);
  }
  
  Result CompositeStrategy::sendEvent(const Event& event) const {
    /* simple event demultiplexing */
    return getEventStrategy()->sendEvent(event);
  }
  
  Result CompositeStrategy::sendMessage(Message* message) {
    /* send message */
    return receiveMessage(message, 0);
  }
  
  void CompositeStrategy::receiveSignal(Signal signal) {
    if(signal == FREE_QUEUE_SLOT || signal == FREE_UPSTR_QUEUE_SLOT) {
      {
        MutexGuard mg(__recv_mutex);
        GLOBAL_DEBUG_STREAM2("CompositeStrategy: received FREE_QUEUE_SLOT signal from owner" << std::endl);
        __blocked = false;
      }
      
      /* check if we can process messages */
      run();
    } else {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: received unknown signal from owner" << std::endl);
    }
  }
  
  void CompositeStrategy::receiveChildSignal(Signal signal, TransportStrategy* /*child*/) {
    if(signal == FREE_QUEUE_SLOT || signal == FREE_UPSTR_QUEUE_SLOT) {
      MutexGuard mg(__send_mutex);
      
      /* check if both are unblocked, then forward signal to jack */
      if(!__sub_binding[Message::BUFFER].isBlocked() &&
         !__sub_binding[Message::EVENT].isBlocked() &&
         !__wait_event_ack) {
        if(getSignalReceiver()) {
          GLOBAL_ERROR_STREAM("Sending FREE_QUEUE_SLOT to partner" << std::endl);
          getSignalReceiver()->receiveSignal(signal);
        } else {
          GLOBAL_ERROR_STREAM("CompositeStrategy has no signal receiver" << std::endl);
          return;
        }
      } else {
        GLOBAL_DEBUG_STREAM2("CompositeStrategy: At least one sub binding is still blocked" << std::endl);
      }
    } else {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: received unknown signal from sub binding" << std::endl);
    }
  }
  
  Result CompositeStrategy::receiveChildMessage(Message* message, InstreamHandler& source) {
    if(__wait_for_timeout && message->getTimestamp().stream_counter == __timeout_stream_counter+1) {
      return BLOCKED;
    }
    
    /* check message type */
    bool event = &source == &__sub_binding[Message::EVENT];
    if(event) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: received message from event child" << std::endl);
    } else {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: received message from buffer child" << std::endl);
    }
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: sequence number is " << message->getTimestamp().stream_counter << std::endl);
    
    /* display difference between timeout and time of reception */
    // XXX: test code
//     if(__mode == ICompositeStrategy::MODE_RELIABLE_UNRELIABLE && !event) {
//       float value;
//       if(message->getTimestamp().stream_counter > __last_timeout_stream_counter) {
//         value = 0;
//       } else {
//         Interval diff = getTime() - __last_timeout;
//         value = diff;
//       }
//     }
    
    /* BEGIN> check if this message is out of date */
    if(__recv_timestamp.stream_counter >= message->getTimestamp().stream_counter) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Message is out of date, dropping it" << std::endl);
      source.unblock();
      message->release();
      
      return SUCCESS;
    }
    /* END< check if this message is out of date */    
    
    /* if we are in unreliable mode, we must send ACK for events */
    if(__mode == ICompositeStrategy::MODE_RELIABLE_UNRELIABLE && event) {
      sendEventReceived(message->getTimestamp().stream_counter);
    }
    
    /* BEGIN> Handle non-continuous sequence numbers */
    if(__recv_timestamp.stream_counter+1 < message->getTimestamp().stream_counter) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Sequence number is non-continuous" << std::endl);
      
      if(__mode == ICompositeStrategy::MODE_RELIABLE_RELIABLE) {
        /* wait for missing message(s) to arrive */
        return BLOCKED;
      } else {
        
        if(event) {
          /* wait for timeout */
          __wait_for_timeout = true;
          __timeout_stream_counter = message->getTimestamp().stream_counter - 1;
          waitForTimeout();
        } else {
          GLOBAL_DEBUG_STREAM2("CompositeStrategy: Buffers have been lost" << std::endl);
          Timestamp lost_ts;
          for(int i=__recv_timestamp.stream_counter + 1; i<=message->getTimestamp().stream_counter-1; ++i) {
            lost_ts.stream_counter = i;
            lost_ts.is_valid = false;
            setRecvTimestamp(lost_ts, Message::BUFFER, true);
          }
        }
      }
    }
    /* END< Handle non-continuous sequence numbers */
    
    /* BEGIN> Interrupt timeout */
    if(__mode == ICompositeStrategy::MODE_RELIABLE_UNRELIABLE &&
       __wait_for_timeout &&
       __timeout_stream_counter == message->getTimestamp().stream_counter) {
      __timeout_cond.notify();
    }
    /* END< Interrupt timeout */
    
    /* forward message */
    return forwardMessage(message, source);
  }
  
  Result CompositeStrategy::forwardMessage(Message* message, InstreamHandler& source) {
    /* remove pending message */
    source.unblock();
    
    /* store timestamp */
    setRecvTimestamp(message->getTimestamp(), message->getType(), false);
    
    /* forward message */
    Result result;
    if(getMessageReceiver()) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Forward message to partner" << std::endl);
      result = getMessageReceiver()->receiveMessage(message, this);
    } else {
      GLOBAL_WARNING_STREAM2("CompositeStrategy: No message receiver" << std::endl);
      result = SUCCESS;
      message->release();
    }
    
    /* check if partner returned BLOCKED */
    if(result == BLOCKED) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Partner returned BLOCKED which means that its StreamQueue is full now" << std::endl);
      __blocked = true;
    }
    
    return SUCCESS;
  }
  
  void CompositeStrategy::setSendValueStream(const string&) {
    /* do nothing */
  }
  
  void CompositeStrategy::setReceiveValueStream(const string&) {
    /* do nothing */
  }
  
  void CompositeStrategy::eventReceived(const int& stream_counter) {
    MutexGuard mg(__send_mutex);
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: Received event ACK for event " << stream_counter << std::endl);
    
    /* we can unblock now */
    if(!__sub_binding[0].isBlocked() && !__sub_binding[1].isBlocked()) {
      __wait_event_ack = false;
      if(getSignalReceiver()) {
        getSignalReceiver()->receiveSignal(FREE_QUEUE_SLOT); // XXX: direction
      } else {
        GLOBAL_WARNING_STREAM2("CompositeStrategy: No signal receiver" << std::endl);
      }
    }
  }
  
  void CompositeStrategy::waitForTimeout() {
    /* compute time to wait. */
    Time current_time = getTime();
    Time event_send_time = __sub_binding[Message::EVENT].getRecvTimestamp().sync_time;
    Interval buffer_transmission_delay = __sub_binding[Message::BUFFER].getAvgTransmissionDelay();
    Time deadline = event_send_time + buffer_transmission_delay;
    
    __last_timeout = deadline;
    __last_timeout_stream_counter = __timeout_stream_counter;
    
    /* wait for timeout */
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: Transmission delay " << buffer_transmission_delay << std::endl);
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: Current time: " << current_time << std::endl);
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: Send time   : " << event_send_time << std::endl);
    GLOBAL_DEBUG_STREAM2("CompositeStrategy: Deadline    : " << deadline << std::endl);
    Result result = __timeout_cond.timedWait(deadline.sec, deadline.nsec);
    if(result == TIME_OUT) {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Timeout at  : " << getTime() << std::endl);
      int stream_counter = __sub_binding[Message::EVENT].getRecvTimestamp().stream_counter - 1;
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: Buffer " << stream_counter << " timed out" << std::endl);
      
      /* store an artificial timestamp for the lost buffer */
      Timestamp lost_ts;
      for(int i=__recv_timestamp.stream_counter + 1; i<=stream_counter; ++i) {
        lost_ts.stream_counter = i;
        lost_ts.is_valid = false;
        setRecvTimestamp(lost_ts, Message::BUFFER, true);
      }
    } else {
      GLOBAL_DEBUG_STREAM2("CompositeStrategy: NOmeout at  : " << getTime() << std::endl);
    }
    
    __wait_for_timeout = false;
  }
  
  void CompositeStrategy::sendEventReceived(const int& stream_counter) {
    /* send acknowledgement for event upstream */
    GLOBAL_DEBUG_STREAM2("CompositeStrategy> Send event ACK for event " << stream_counter << std::endl);
    Event event(ICompositeStrategy::eventReceived_event,
                new TInValue<int>(stream_counter));
    __remote_address->getConnectionManager()->sendReliableConfigEvent(__remote_address->getTSReference(), event);
  }
  
  void CompositeStrategy::setRecvTimestamp(const Timestamp& timestamp, Message::Type /*type*/, bool /*lost*/) {
    /* check for continuity */
    // XXX: test code
//     std::cout << timestamp.stream_counter;
//     switch(type) {
//     case Message::EVENT:
//       std::cout << " E";
//       break;
//     case Message::BUFFER:
//       std::cout << " B";
//       break;    default:
//       std::cout << " ?";
//     };
//     if(lost) {
//       std::cout << " (lost)";
//     }
//     std::cout << std::endl;
    
    /* set timestamp*/
    __recv_timestamp = timestamp;
  }
  
  void CompositeStrategy::run() {
    MutexGuard mg(__recv_mutex);
    
    for(;;) {
      /* check if we are blocked */
      if(__blocked) {
        GLOBAL_DEBUG_STREAM2("CompositeStrategy: Partner still blocked, not processing new messages" << std::endl);
        return;
      }
      
      /* BEGIN> get the pending message with the lowest timestamp */
      InstreamHandler* next = NULL;
      int stream_counter = INT_MAX;
      for(unsigned int i=0; i<2; ++i) {
        if(__sub_binding[i].getMessage() &&
           __sub_binding[i].getRecvTimestamp().stream_counter < stream_counter) {
          next = &__sub_binding[i];
          stream_counter = __sub_binding[i].getRecvTimestamp().stream_counter;
        }
      }
      if(!next) {
        GLOBAL_DEBUG_STREAM2("CompositeStrategy: No messages to process" << std::endl);
        return;
      }
      /* END< */
      
      /* handle this next message */
      if(next && !__blocked) {
        Message* message = next->getMessage();
        if(BLOCKED == receiveChildMessage(message, *next)) {
          /* we can't continue at the moment (wait for a message or a timeout */
          break;
        }
      }
    }
  }
  
  
  template<>
  class TConstructor<CompositeAddress> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    {}
    virtual CompositeAddress*          createValue()          const {return new CompositeAddress();}
    virtual TValue<CompositeAddress>*  createTValue()         const {return new TValue<CompositeAddress>();}
    
    virtual TValue<TSAddress*>*  createTPointerValue()        const {return new TValue<TSAddress*>();}
    
    virtual TConstructor<CompositeAddress>* copy()            const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };
  
  
  const char CompositeStrategy::__strategy_type[] = "CompositeStrategy";
  
  TConstructor<CompositeAddress> __composite_address(TypeInfo("CompositeAddress", "NMM"));
  
  TTSFactoryRegistrator< TConnectorTSFactoryObject<CompositeStrategy, ConnectorAcceptor> >
  __composite_strategy_prototype("CompositeStrategy", false, false);
  
} // namespace NMM
