/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_INTERFACE_HPP
#define NMM_INTERFACE_HPP

#include <memory>
#include "ISender.hpp"
#include "nmm/base/registry/RegistryStructs.hpp"
#include "nmm/base/serialize/ComplexValue.hpp"
#include "nmm/base/format/Format.hpp"
#include "BaseExceptions.hpp"
#include "TSAddress.hpp"

namespace NMM {

class IValueStream;
class OValueStream;
class IEventReceiver;
class NMMObject;
class TSAddress;
class ProxyObject;


/** \class Interface
  * This class is base class for all interfaces. It strores all addresses to its parent
  * object wich provides the implementation of an interface. The communication between
  * an Interface and its object is done through a proxy object, wich is connected through
  * a communication channel with the receiver.
  * This class performs reference counting. If all interfaces to an object are destroyed.
  * the proxy object is deleted.
  */
class Interface : public SerializedValue, public IEventSender {
  friend class ProxyObject;
  friend class ReferenceManager;
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor, that uses the given NMMObject as parent
  */
  Interface(NMMObject& receiver, const string& interface_type = "Interface", const string& name_space = "NMM");

/** Constructor 
  */
  Interface(const string& interface_type = "Interface", const string& name_space = "NMM");

/** Copy Constructor
  */
  Interface(const Interface& );

/** Destructor
  */
  virtual ~Interface();

/** virtual Copy Constructor
  */
  virtual Interface* copy() const {return new Interface(*this);}
//@}

/** @name Event Sending
  */
//@{
/** This mehtod sends an CEvent to the current event receiver.
  */
  virtual Result sendEvent( const CEvent& ce)         ;

/** This mehtod sends an Event to the current event receiver.
  */
  virtual Result sendEvent( const Event& e)           ;

/** This mehtod sends an CEvent to the current event receiver.
  */
  virtual Result sendEvent( const CEvent& ce)   const ;

/** This mehtod sends an Event to the current event receiver.
  */
  virtual Result sendEvent( const Event& e)     const ;
//@}

/** @name Access Interace information
  */
//@{
/** This method returns a list of all subtypes of this class
  */
  inline const list<string>& getSubTypes() const {return __interfaces;}


/** This method returns true if given type is a subtype of this interface
  */
  bool isSubType(const string& type);

/** This method returns the typename of this interface
  */
  virtual       string getInterfaceType() const { return getTypeName(); }

/** This method returns the used ProxyObject to send events
  */
  virtual       ProxyObject* getParentObject()          ;

/** This method returns the used ProxyObject to send events
  */
  virtual const ProxyObject* getParentObject() const    ;

/** This method returns the used ProxyObject to send events. This method is obsolete, use getParentObject()
  */
  virtual       ProxyObject* getEventReceiver()         ;

/** This method returns the used ProxyObject to send events. This method is obsolete, use getParentObject()
  */
  virtual const ProxyObject* getEventReceiver() const   ;

/** This method returns the NMMReference of the NMMObject that offers this interface
  */
  virtual const NMMReference& getParentReference() const ;

/** This method returns the UserInfo of the NMMObject
  */
  inline const UserInfo& getParentInfo()           const { return __parent_user_info.getRValue(); }

/** This method returns a list of all addresses to connect to the node.
  */
  const SequenceValue<TSAddress>& getAddressList() const;

/** This method returns a list of supported TransportStrategies of the NMMObject
  */
  const list<string> getSupportedStrategies()      const  { return __supported_strategies.getRValue(); }

/** This method returns the TSAddress for a specific Transport Strategy
  */
  TSAddress* getDefaultAddress(const string&); 

/** This method can be used to enforce the estabelishment of a conntection to the Node.
  */
  void establishBinding()
    throw (const BindingException&);
//@}

/** @name Methods to specify the position of interfaces
  */
//@{
/** returns true if the object of this interface and the object of the given interface runs on the same system
  * and in the same process
  */
  virtual bool isLocalTo(const Interface& ) const;

/** returns true if the object of this interface and the object of the given interface runs on the same system
  * but in a different process
  */
  virtual bool isSystemTo(const Interface&) const;

/** returns true if the object of this interface and the object of the given interface runs on different systems.
  */
  virtual bool isNetworkTo(const Interface&) const;
//@}

/** @name Operator
  */
//@{
/** Assignment operator
  */
  Interface& operator=(const Interface& );
//@}

protected:
/** This method sets the EventReceiver
  */
  void setEventReceiver(NMMObject& receiver);

/** This method sets the ParentObject
  */
  void setParentObject(ProxyObject* receiver) { __receiver = receiver; }

/** This method adds an interface subtype
  */
  void addInterfaceType(const string& interface);

/** Prepare the attributes for serialization
  */
  void prepareAttributes();

/** Update the attributes after deserialization
  */
  void updateAttributes();
private:
/** register all attributes
  */
  void registerAttributes();

/** request the ProxyObject
  */
  void requestParentObject() const;
private:
  list<string>                    __interfaces;             ///< List with subtypes
  mutable NMMObject*              __parent_object;          ///< Pointer to the NMMObject
  mutable ProxyObject*            __receiver;               ///< Pointer to the ProxyObject, that receives the events
  mutable TValue<NMMReference>    __parent_reference;       ///< NMMReference of the NMMObject
  TValue<UserInfo>                __parent_user_info;       ///< UserInfo of the NMMObject
  SequenceValue<TSAddress>        __address_list;           ///< List with all TSAddresses
  TValue< list<string> >          __supported_strategies;   ///< Stores the supported TransportStrategies
  bool                            __is_local;               ///< Is false if the Interface was serialized.
};

typedef auto_ptr<Interface> Interface_var;

OValueStream& operator<<(OValueStream& ov, const list<Interface*>& interfaceptr_list);
IValueStream& operator>>(IValueStream& iv, list<Interface*>& interfaceptr_list);



}


#endif
