/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef __NMM_OBJECTCHANNEL_HPP
#define __NMM_OBJECTCHANNEL_HPP

#include "CommunicationChannel.hpp"
#include "TransportStrategy.hpp"

namespace NMM {
  
  /** This Communication Channel represents the connection between a ProxyObject and a NMMObject.
   */
  class OutOfBandChannel : public CommunicationChannel {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    OutOfBandChannel(NMMObject& parent);

    /** Destructor
     */
    virtual ~OutOfBandChannel();

    /** Copy Constructor
     */
    OutOfBandChannel(const OutOfBandChannel&);

    /** virtual Copy Constructor
     */
    virtual      OutOfBandChannel*     copy()                          const { return new OutOfBandChannel(*this); }
    //@}

    /** @name Configure the Communication Channel
     */
    //@{
    /** This method sets the TransportStrategy that should be used
     */
    virtual void setTransportStrategy(const string& strategy_type);

    /** This method sets the TransportStrategy that should be used
     */
    virtual void setTransportStrategy(TransportStrategy* strategy);

    /** This method returns a list of all TransportStrategies availble for the connection
     */
    virtual void setValueStream(const IValueStream&, const OValueStream& );

    /** This method adds an connection address
     */
    virtual      void               addAddress(const TSAddress&);

    /** This method returns the TSAddress for specific TransportStrategy strategy
     */
    virtual      const TSAddress*   getAddress(const string& strategy)      const;

    /** This method returns the Default address for the default TransportStrategy
     */
    inline       const TSAddress*   getDefaultAddress()                     const { return __default_address; }

    /** This method sets the Sefault address for the default TransportStrategy
     */
    virtual void setDefaultAddress(const TSAddress& address);

    /** This method updates the list of available transport strategies.
     */
    virtual void updateStrategyList(const Interface& interface);
    //@}
    
    /** @name Access Communication Channel information
     */
    //@{
    /** This method returns a pointer to the current TransportStrategy
     */
    inline       TransportStrategy* getTransportStrategy()                { return __strategy; }
    
    /** This method returns a pointer to the current TransportStrategy
     */
    inline const TransportStrategy* getTransportStrategy()          const { return __strategy; }
    
    /** This method returns a reference to the NMMObject that uses this communicatio channel
     */
    inline       NMMObject&         getParent()                           { return __parent; }
    
    /** This method returns a reference to the NMMObject that uses this communicatio channel
     */
    inline const NMMObject&         getParent()                     const { return __parent; }
    
    /** This method retursn the Interfaces of the ProxyObject's TranportStrategy
     */
    virtual      list<Interface*>&  getInterfaceList();
    
    /** This method retursn the Interfaces of the NMMObject's TranportStrategy
     */
    virtual      list<Interface*>&  getPartnerInterfaceList();
    //@}
    
    /** @name Obsolete method
     */
    //@{
    virtual void deinit();
    //@}
  private:
    void updateIStrategies();
    
  private:
    NMMObject&              __parent;               ///< The Parent object which uses this NMMObject
    TransportStrategy*      __strategy;             ///< Pointer to the local Transport-Strategy
    ITransportStrategy_var  __i_strategy;           ///< Interface to the local Transport-Stratgey
    ITransportStrategy_var  __i_partner_strategy;   ///< Interface to the distributed partner Stratgey
    bool                    __i_strategies_valid;   ///< Flag to indicate if the interfaces are valid
    list<TSAddress*>        __address_list;         ///< List of connection address
    TSAddress*              __default_address;      ///< Pointer to the used Address
    
  public:
    /** This method returns the TypeInfo for this class.
     */
    static TypeInfo getStaticTypeInfo();
  };
  
}
  
#endif
