/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>
#include "QueuedOutputJack.hpp"
#include "InputJack.hpp"
#include "Node.hpp"
#include "Message.hpp"
#include "EventDispatcher.hpp"
#include "TransportStrategy.hpp"

namespace NMM {

  QueuedOutputJack::QueuedOutputJack(Node* parent, 
				     const string& jack_tag /* = "default */,
				     const string& name /* = "QueuedOutputJack" */,
				     const StreamQueue::Mode queue_mode /* = StreamQueue::DEFAULT_MODE */,
				     const unsigned int size /* = StreamQueue::DEFAULT_SIZE */)
    : OutputJack(parent, jack_tag, name)
  {
    stream_queue  = 
      new StreamQueue(queue_mode, size, name.c_str(), FREE_UPSTR_QUEUE_SLOT, OUTPUT_QUEUE_NOT_EMPTY);

    stream_queue->setParentSignalReceiver(dynamic_cast<ISignalReceiver *>(parent));
  }

  QueuedOutputJack::~QueuedOutputJack()
  {
    stream_queue->setPartnerSignalReceiver(0);
    delete stream_queue;
  }

  void QueuedOutputJack::connectTo(const InputJack* in_jack) const
    throw (const BindingException&)
  {
    OutputJack::connectTo(in_jack);
    stream_queue->setPartnerSignalReceiver(const_cast<Jack*>(getPartner()));
  }

  void QueuedOutputJack::disconnect()
  {
    OutputJack::disconnect();
    stream_queue->setPartnerSignalReceiver(0);
  }

  Result QueuedOutputJack::receiveMessage(Message* message, const IMessageSender*  )
  {
    DEBUG_STREAM2("Forwarding buffer to parent" << endl);

    DEBUG_STREAM2("receiveMessage UPSTREAM" << endl);

    return stream_queue->receiveMessage(message);
  }

  Result QueuedOutputJack::sendMessage(Message *message)
  {
    if(getMessageReceiver() != 0) {
      DEBUG_STREAM2("Send Buffer to Jackpartner" << endl);
      Result r = getMessageReceiver()->receiveMessage(message, this);
      if(r == BLOCKED) {
        DEBUG_STREAM2("Jackpartner returned BLOCKED which means its stream_queue is full now" << endl);
      }
      return r;
    } // end: if (getMessageReceiver() != 0)
    
    else { // no message receiver
      // no partner, must discard buffer
      DEBUG_STREAM2("Partner-Jack is 0" << endl);
      message->release();
      return FAILURE;
    }
  }

  void QueuedOutputJack::setTransportStrategy(TransportStrategy* strategy) {
    OutputJack::setTransportStrategy(strategy);
    stream_queue->setPartnerSignalReceiver(strategy);
    strategy->setSignalReceiver(this);
  }

  Result QueuedOutputJack::flush()
  { 
    return stream_queue->flush(); 
  }

  unsigned int QueuedOutputJack::getSize() const 
  { 
    return stream_queue->getSize(); 
  }

  unsigned int QueuedOutputJack::getMaxSize() const 
  { 
    return stream_queue->getMaxSize(); 
  }

  Result QueuedOutputJack::setMaxSize(unsigned int size) 
  { 
    return stream_queue->setMaxSize(size); 
  }

  StreamQueue::Mode QueuedOutputJack::getMode() const 
  { 
    return stream_queue->getMode(); 
  }

  Result QueuedOutputJack::setMode(StreamQueue::Mode queue_mode) 
  { 
    return stream_queue->setMode(queue_mode); 
  }

  void QueuedOutputJack::receiveSignal(Signal sig) {
    getSignalReceiver()->receiveSignal(sig); // forward signal to own Receiver
  }

}
