/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "StreamQueue.hpp"
#include "Message.hpp"
#include "Event.hpp"
#include "OutputJack.hpp"
#include "ISignalReceiver.hpp"
#include "Node.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"

namespace NMM {

  StreamQueue::StreamQueue(const Mode mode,
			   const unsigned int size,
			   const char* name,
			   const Signal free_queue_slot_sig,
			   const Signal queue_not_empty_sig) 
    : NamedObject(name), 
      queue_mode(mode), 
      queue_length(size), 
      sig_free_queue_slot(free_queue_slot_sig), 
      sig_queue_not_empty(queue_not_empty_sig),
      limited_size(size),
      pre_fetch_queue_size(0),
      pre_fetch_flag(false),
      pre_fetch_condition(pre_fetch_mutex)
  {
    queue_mutex = new ThreadMutex(ThreadMutex::RECURSIVE);
    setName(getName()+"> Queue");
    abort_receive_exception_count = 0;

    parent_of_jack = 0;
    parent_of_partner_jack = 0;
  };

  StreamQueue::~StreamQueue()
  {
    flush(); // I guess, we should not call this method... BufferManagers should delete their Buffers
    delete queue_mutex;
  };

  Result StreamQueue::setParentSignalReceiver(ISignalReceiver* parent)
  {
    MutexGuard mg(*queue_mutex);
    parent_of_jack = parent;
    return SUCCESS;
  }

  Result StreamQueue::setPartnerSignalReceiver(ISignalReceiver* partner)
  {
    MutexGuard mg(*queue_mutex);
    parent_of_partner_jack = partner;
    return SUCCESS;
  }

  Result StreamQueue::setParentSignalReceiver(ISignalReceiver* parent, const Signal queue_not_empty_sig)
  {
    MutexGuard mg(*queue_mutex);
    parent_of_jack = parent;
    sig_queue_not_empty = queue_not_empty_sig;
    return SUCCESS;
  }

  Result StreamQueue::setPartnerSignalReceiver(ISignalReceiver* partner, const Signal free_queue_slot_sig)
  {
    MutexGuard mg(*queue_mutex);
    parent_of_partner_jack = partner;
    sig_free_queue_slot = free_queue_slot_sig;
    return SUCCESS;
  }

  Result StreamQueue::signalFreeQueueSlot() 
  {
    MutexGuard mg(*queue_mutex);
    if(message_queue.size() >= queue_length) {
      if(parent_of_partner_jack) {
	parent_of_partner_jack->receiveSignal(sig_free_queue_slot);
	DEBUG_STREAM2("Signal sig_FREE_QUEUE_SLOT" << endl);
	return SUCCESS;
      }
      else {
	WARNING_STREAM2("Could not signal sig_FREE_QUEUE_SLOT, no parent_of_partner_jack" << endl);
	return FAILURE;
      }
    }
    else {
      DEBUG_STREAM2("Queue not full, no signal sig_FREE_QUEUE_SLOT needed" << endl);
      return SUCCESS;
    }
  }

  Result StreamQueue::flush(const Timestamp& flush_ts)
  {
    DEBUG_STREAM2("Clear Queue." << endl);

    MutexGuard mg(*queue_mutex);

    queue<Message*> tmp_queue;
    Message* msg = 0;

    while (!message_queue.empty()) {
      msg = message_queue.front();
      message_queue.pop();

      if(msg->getMandatoryBit()) { // keep message
	tmp_queue.push(msg);
      }
      else { // message should be discarded
	// if timestamp for flushing was given -> flush all messages with 'older' timestamp only
	if(flush_ts.is_set && flush_ts.is_valid) {
	  Timestamp msg_ts = msg->getTimestamp();
	  if(msg_ts.is_set && msg_ts.is_valid) {
	    if(flush_ts.sync_time < msg_ts.sync_time) { // if flush_ts is 'younger' than msg_ts -> keep message
	      tmp_queue.push(msg);
	      continue;
	    }
	  }
	}

	// code taken from StreamQueue::dequeueMessage() :
	// signal if at least one buffer can be inserted, if we had signalled that
	// the queue was full.
	if (abort_receive_exception_count > 0) {
	  abort_receive_exception_count--;

	  // if partner_of_parent_jack is available, send internal event
	  if(parent_of_partner_jack) {
	    parent_of_partner_jack->receiveSignal(sig_free_queue_slot);
	    DEBUG_STREAM2("Signal sig_FREE_QUEUE_SLOT" << endl);
	  }
	  else {
	    WARNING_STREAM2("No subclass of ISignalReceiver, cannot send FREE_QUEUE_SLOT" << endl);
	  }
	}
	DEBUG_STREAM2("Releasing message of type " << msg->getType() << endl);
	msg->release();
	DEBUG_STREAM2("Releasing message done." << endl);
      }
    }
    DEBUG_STREAM2("Queue cleared." << endl);

    while(!tmp_queue.empty()) { // re-insert mandatory message from tmp_queue to message_queue
      msg = tmp_queue.front();
      tmp_queue.pop();
      message_queue.push(msg);
      DEBUG_STREAM("Re-insert mandatory message " << msg->getType() << endl);
    }

    return SUCCESS;
  };


  Result StreamQueue::receiveMessage(Message* message)
  {
    MutexGuard mg(*queue_mutex);

    if (message == 0) {
      ERROR_STREAM2("BUG: Received 0-Message" << endl);
      return FAILURE;
    }

    // check for FlushEvent
    if ( message->getType() == Message::FLUSH_EVENT) {
      DEBUG_STREAM2("received FlushEvent -> clearing queue" << endl);
      flush(message->getTimestamp());
    }

    Message* oldest = 0;
    int insert_this_buffer = 1;

    if(message_queue.size() >= queue_length) {
      // debug test
      if(message_queue.size() > queue_length) {
        ERROR_STREAM2("ARGH. QUEUE MORE THAN FULL, THIS SHOULD NOT HAPPEN!" << endl);
      }

      switch(queue_mode) {

      case MODE_SUSPEND:
	// Queue will suspend, but message will be inserted
	insert_this_buffer = true;
	break;

      case MODE_KEEP_NEWEST:
	// new message is newer than all enqueued messages, so
	// it has to be inserted because stream queue is in
	// mode MODE_KEEP_NEWEST
	insert_this_buffer = true;

	// remove oldest message in queue
	oldest = message_queue.front();
	oldest->release();
	message_queue.pop();

	break;

      default:
	// this is MODE_KEEP_OLDEST (see header file), or anything else

	// release message, since older messages in queue should
	// be kept in MODE_KEEP_OLDEST
	message->release();
	insert_this_buffer = false;

	break;
      }
    }

    if(insert_this_buffer) {

      // if queue was empty before, signal that a new Buffer arrived and can be dequeued
      bool e = message_queue.empty();

      // flush events with valid timestamp will 'outrun' other messages in stream queue
      // -> this is necessary since the timestamp of the flush event should also be used for dropping all messages
      // in the successor of this node if 'older' (e.g. smaller) than timestamp of flush event
      // (just imagine: the flush event has to be inserted at the 'right' position within the queue!)
      if ( message->getType() == Message::FLUSH_EVENT) {
	// re-order entries in queue: put timed flush event to the beginning of the queue
	Timestamp ts = message->getTimestamp();
	if(ts.is_valid && ts.is_set) {
	  queue<Message*> tmp_queue;
	  Message* msg = 0;
	  while (!message_queue.empty()) {
	    msg = message_queue.front();
	    message_queue.pop();
	    tmp_queue.push(msg);
	  }
	  // re-insert messages to queue, put in 'flush_event' first!
	  message_queue.push(message);
	  while(!tmp_queue.empty()) { 
	    msg = tmp_queue.front();
	    tmp_queue.pop();
	    message_queue.push(msg);
	  }
	  DEBUG_STREAM2("reordered FlushEvent with timestamp " << message->getTimestamp() << endl);
	}
	else { // no timestamp set for flus event -> simply insert into queue
	  message_queue.push(message);
	}
      }
      else { // no flush event -> simply insert into queue
	message_queue.push(message);
      }

      if(e || pre_fetch_flag) {

	if(pre_fetch_flag) {

	  // if partner_of_jack is available, send internal event
	  if(message_queue.size() == pre_fetch_queue_size) {
	    if(parent_of_jack) {
	      MutexGuard mg(pre_fetch_mutex);
	      pre_fetch_flag = false;
	      parent_of_jack->receiveSignal(sig_queue_not_empty);
              pre_fetch_condition.notify();
	    }
	    else {
	      DEBUG_STREAM2("No subclass of ISignalReceiver, cannot send QUEUE_NOT_EMPTY");
	    }
	  }

	}
	else {

	  // if partner_of_jack is available, send internal event
	  if(parent_of_jack) {
	    parent_of_jack->receiveSignal(sig_queue_not_empty);
	  }
	  else {
	    DEBUG_STREAM2("No subclass of ISignalReceiver, cannot send QUEUE_NOT_EMPTY");
	    
	  }

	}
      }

      // if queue is "more than" full now, return BLOCKED
      if((message_queue.size() >= queue_length) &&   // changed from > to >= (mlohse 13-3-2002)
	 (queue_mode == MODE_SUSPEND)) {
	abort_receive_exception_count++;
	return BLOCKED;
      }
    }

    return SUCCESS;
  };

  Message* StreamQueue::dequeueMessage()
  {
    MutexGuard mg(*queue_mutex);

    // empty queue -> return 0
    if (message_queue.empty()) {
      return 0; 
    }

    Message* msg = message_queue.front();
    message_queue.pop();

    // signal if at least one buffer can be inserted, if we had signalled that
    // the queue was full.
    // (code also in StreamQueue::clear() )
    if ((abort_receive_exception_count > 0)) {

      // if partner_of_parent_jack is available, send internal event only if has empty slot
      if(parent_of_partner_jack && (message_queue.size() < queue_length)) {
	while(abort_receive_exception_count > 0) {
	  abort_receive_exception_count--;
	  parent_of_partner_jack->receiveSignal(sig_free_queue_slot);
	  DEBUG_STREAM2("Signal sig_FREE_QUEUE_SLOT" << endl);
	}
      }
      else {
	DEBUG_STREAM2("No subclass of ISignalReceiver, or queue more than full" << std::endl);
      }
    }

    return msg;
  };

  unsigned int StreamQueue::getSize() const 
  {
    MutexGuard mg(*queue_mutex);
    return message_queue.size();
  }

  unsigned int StreamQueue::getMaxSize() const 
  {
    MutexGuard mg(*queue_mutex);
    return queue_length;
  }

  bool StreamQueue::empty() const 
  { 
    MutexGuard mg(*queue_mutex); 
    return message_queue.empty(); 
  }

  bool StreamQueue::full() const 
  { 
    MutexGuard mg(*queue_mutex); 
    return (message_queue.size() >= queue_length); 
  }

  Result StreamQueue::setMaxSize(unsigned int size)
  {
    MutexGuard mg(*queue_mutex); 
    queue_length = size;
    return SUCCESS;
  }

  void StreamQueue::unlimitSize()
  {
    MutexGuard mg(*queue_mutex); 
    limited_size = queue_length;
    DEBUG_STREAM2("unlimitSize() current queue_length " << queue_length << ", new 20000 " << endl);
    queue_length = 20000;
  }

  void StreamQueue::limitSize()
  {
    MutexGuard mg(*queue_mutex); 
    DEBUG_STREAM2("limitSize() current queue_length " << queue_length << ", new " << queue_length << endl);
    queue_length = limited_size;
  }

  StreamQueue::Mode StreamQueue::getMode() const
  {
    MutexGuard mg(*queue_mutex); 
    return queue_mode;
  }

  Result StreamQueue::setMode(StreamQueue::Mode mode)
  {
    MutexGuard mg(*queue_mutex); 
    queue_mode = mode;
    return SUCCESS;
  }

  long StreamQueue::dropBuffers(long drop_bytes)
  {
    // need to lock whole action, otherwise predecessor might insert more buffers
    MutexGuard mg(*queue_mutex);

    long bytes_dropped = 0;
    queue<Message*> tmp_queue;

    while(!message_queue.empty()) {
      Message* msg = message_queue.front();
      message_queue.pop();

      if((msg)->getType() == Message::BUFFER) {
	// still need to drop buffer
	if(bytes_dropped < drop_bytes) {
	  // DEBUG_STREAM2("SEEK Dropping buffer" << endl);
	  bytes_dropped += dynamic_cast<Buffer*>(msg)->getUsed();
	  (msg)->release();

	  // just removed buffer from queue -> code for internal signalling FREE_QUEUE_SLOT
	  if ((abort_receive_exception_count > 0)) {
	    abort_receive_exception_count--;

	    if(parent_of_partner_jack) {
	      parent_of_partner_jack->receiveSignal(sig_free_queue_slot);
	    }
	    else {
	      DEBUG_STREAM2("No subclass of ISignalReceiver, cannot send FREE_QUEUE_SLOT");
	    }
	  }
	}
	// already dropped enough buffers, keep buffers in tmp
	else {
	  // DEBUG_STREAM2("SEEK storing buffer in tmp" << endl);
	  tmp_queue.push(msg);
	}
      }
      // no buffer: keep all other messages ( = events) in tmp
      else {
	// DEBUG_STREAM2("SEEK storing event in tmp" << endl);
	tmp_queue.push(msg);
      }
    } // end while

    // check bytes_dropped
    //if(bytes_dropped >= drop_bytes) {
    //DEBUG_STREAM2("SEEK enough buffers dropped " << bytes_dropped << " >= " << drop_bytes << endl);
    //}
    //else {
    //DEBUG_STREAM2("SEEK _not_ enough buffers were dropped " << bytes_dropped << " < " << drop_bytes << endl);
    //}

    // rebuild queue
    message_queue = tmp_queue;

    return bytes_dropped;
  }
  
  void StreamQueue::lock() const {
    queue_mutex->lock();
  }

  void StreamQueue::unlock() const {
    queue_mutex->unlock();
  }

  void StreamQueue::setPrefetchQueueSize(unsigned int pre_fetch_queue_size) 
  {
    if(pre_fetch_queue_size > 0) {
      if(pre_fetch_queue_size > queue_length) {
	WARNING_STREAM2("Tried to set too large prefetch queue size " << pre_fetch_queue_size << ", using max. queue size of " << queue_length << " instead." << endl);
	this->pre_fetch_queue_size = queue_length;
      }
      else {
	this->pre_fetch_queue_size = pre_fetch_queue_size;
      }
      pre_fetch_flag = true;
    }
    else {
      this->pre_fetch_queue_size = 0;
      pre_fetch_flag = false;
    }
  }

  void StreamQueue::waitPrefetchFinished()
  {
    MutexGuard mg(pre_fetch_mutex);
    while( pre_fetch_flag == true ) {
      pre_fetch_condition.wait();
    }
  }

}
