/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TSFactory.hpp"
#include "TransportStrategy.hpp"

#include "ConnectionManager.hpp"
#include "FormatStrategyMap.hpp"
#include "ConnectorAddress.hpp"

namespace NMM {

  TSFactory::TSFactory()
    : NamedObject("TSFactory")
  {
  }

  TSFactory::~TSFactory() {
    for (InternalIt it = __factory_map.begin(); it != __factory_map.end(); ++it) {
      /* Delete the address */
      delete it->second.default_address;
    }

    __factory_map.clear();
  }

  void TSFactory::registerStrategy(const TSFactoryObject& factory, const string& strategy_type,
                                   const list<Format>& format_list, bool no_events) {
    InternalIt it = __factory_map.find(strategy_type);
    /* If the strategy type is konwn, return */
    if (it != __factory_map.end()) {
      ERROR_STREAM("A TransportStrategy of type " << strategy_type << " is already registered "<< endl);
      return;
    }
  
    /* Otherwise store it in the list */
    InternalStruct is;
    is.factory_object  = &factory;
    is.default_address = NULL;
    is.no_events       = no_events;
    __factory_map.insert(pair<string, InternalStruct>(strategy_type, is));
  
    /* Register formats */
    for (list<Format>::const_iterator i(format_list.begin());
         i != format_list.end();
         ++i) {
      SConnectionManager::getInstance().getFormatStrategyMap().addSupportedFormat(strategy_type, *i);
    }
  }

  void TSFactory::unregisterStrategy(const string& strategy_type) {
    InternalIt it = __factory_map.find(strategy_type);  
  
    /* If the strategy type is konwn, return */
    if (it == __factory_map.end()) {
      // fw: this message is annoying if the TSFactory is destroyed before the TSFactoryRegistrators
      // ERROR_STREAM("A TransportStrategy of type " << strategy_type << " doesn't exist "<< endl);
      return;
    }
  
    /* Unregister formats */
    SConnectionManager::getInstance().getFormatStrategyMap().removeStrategyType(strategy_type);
  
    /* Delete the address */
    delete it->second.default_address;
  
    __factory_map.erase(it);
  }
  

  TransportStrategy* TSFactory::createStrategy(const string& interface_type) {
    InternalIt it = __factory_map.find(interface_type);
    /* If the type name is unknown, return 0 */
    if (it == __factory_map.end()) {
      if (interface_type == "NULL")
        return 0;

      ERROR_STREAM("No TransportStrategy of type " << interface_type << " avaialable "<< endl);
      return 0;
    }

    /* Create the instance*/
    return it->second.factory_object->createInstance();
  }

  bool TSFactory::isStrategyAvailable(const string& strategy_type) const {
    if (__factory_map.find(strategy_type) == __factory_map.end()) {
      return false;
    }

    return true;
  }
  
  bool TSFactory::canStrategySendEvents(const string& strategy_type) const {
    CInternalIt found = __factory_map.find(strategy_type);
    if (found == __factory_map.end()) {
      return false;
    }
    if (found->second.no_events) {
      return false;
    }
    
    return true;
  }
  
  list<string> TSFactory::getStrategyList() const {
    list<string> strategy_list;
    for (CInternalIt it = __factory_map.begin(); it != __factory_map.end(); ++it)
      {
        strategy_list.push_back(it->first);
      }

    return strategy_list;
  }

  TSAddress* TSFactory::getDefaultAddress(const string& strategy) const {
    /* If the startegy type is unkown return NULL */
    CInternalIt it = __factory_map.find(strategy);
    if (it == __factory_map.end())
      return 0;

    /* If a default address exist, return a copy to it.*/
    if (it->second.default_address)  {
      return it->second.default_address->copy();
    }
    
    DEBUG_STREAM("No default address for strategy " << strategy << " available" << endl);
    return 0;
  }

  void TSFactory::setDefaultAddress(const string& strategy, const TSAddress* address) {
    InternalIt it = __factory_map.find(strategy);
    if (it == __factory_map.end()) {
      ERROR_STREAM("No strategy " << strategy << " available" << endl);
      return;
    }

    delete it->second.default_address;

    if (!address) {
      it->second.default_address= 0;
    }

    it->second.default_address = address->copy();
  }

  ConnectorAcceptor* TSFactory::requestConnectorAcceptor(const std::string& strategy) const {
    CInternalIt it = __factory_map.find(strategy);
  
    /* if the type name is unknown, return 0 */
    if (it == __factory_map.end()) {
      if (strategy == "NULL")
        return 0;

      ERROR_STREAM("No TransportStrategy of type " << strategy << " avaialable "<< endl);
      return 0;
    }

    /* create the instance*/
    ConnectorAcceptor* connector_acceptor = it->second.factory_object->requestConnectorAcceptor();
    if(!connector_acceptor) {
      ERROR_STREAM("No ConnectorAcceptor for strategy " << strategy << " avaialable "<< endl);
    }
    
    return connector_acceptor;
  }

  void TSFactory::addConnectorAddress(const string& strategy_type) {
    /* Get IConnectionManager interface */
    IConnectionManager_var interface(SConnectionManager::getInstance().getInterface<IConnectionManager>());
    if (!interface.get()) {
      ERROR_STREAM("Could not get local connection manager interface" << endl);
      return;
    }
    ConnectorAddress address(strategy_type);
    address.setConnectionManager(*interface);
  
    /* Register default address for Strategy */
    setDefaultAddress(strategy_type, &address);
  }


  TSFormatList::TSFormatList() 
  {}
  
  DefaultTSFormatList::DefaultTSFormatList() {
    Format top_format("*/*");
    addFormat(top_format);
  }


}//end namespace
