/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include "nmm/base/ThreadedNode.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"
#include "BaseExceptions.hpp"
#include "EventDispatcher.hpp"

#include <sys/resource.h>
#include <sys/time.h>

namespace NMM {

  ThreadedNode::ThreadedNode(const char* name) : Node(name)
  {
    thread_owner = true;
    activated_before = false;
    thread_pid = 0;
    thread_priority = 0;
  }

  ThreadedNode::~ThreadedNode()
  {
  }

  void ThreadedNode::initOutput()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    Node::initOutput();

    if(thread_owner) {
      if(!activated_before) {
	activated_before = true;
	try {
	  thread.start(run_static, (void *)this); // create thread
	}
	catch(const Exception& e) {
	  ERROR_STREAM2("Starting thread failed : " << e << endl);
	  throw StateTransitionFailedException("Starting thread failed");
	}
	DEBUG_STREAM2("Thread created with ID : " << thread.getID() << endl);
      }
      else {
	DEBUG_STREAM2("Already created thread with ID : " << thread.getID() << " reusing this one... " << endl);
      }
    }
    else {
      DEBUG_STREAM2("ThreadedNode does not own a thread and has to be scheduled manually." << endl);
    }
    if (thread_priority) {
      setpriority(PRIO_PROCESS,thread_pid,thread_priority);
    }
  }

  void ThreadedNode::deinitOutput()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    Node::deinitOutput();

    DEBUG_STREAM2("Joining Thread in deinitOutput..." << endl);
    thread.join();
    activated_before = false;
    DEBUG_STREAM2("Joined!" << endl);

  }

  void* ThreadedNode::run_static(void *me) {
    try {
      static_cast<ThreadedNode*>(me)->setPID(getpid());
      static_cast<ThreadedNode*>(me)->assignDefaultResourceContainer(); // the default container must be assigned in the thread the node runs in


      NODE_DEBUG_STREAM2(static_cast<Node*>(me),
			"Thread PID : " << getpid() << endl);
      (static_cast<ThreadedNode*>(me))->run();

    } catch (const Exception& e) {
      // All exceptions should be caught before this. See ThreadedNode.hpp
      cerr << NamedObject::color_array[NamedObject::RED_COLOR];
      cerr << "Node \"" << static_cast<ThreadedNode*>(me)->getName()
           << "\" (pid: " << getpid() 
           << ") caused a uncaught Exception in its internal thread:" << endl;
      cerr << e << endl;
      cerr << NamedObject::color_array[NamedObject::DEFAULT_COLOR];
    } catch (...) {
      cerr << NamedObject::color_array[NamedObject::RED_COLOR];
      cerr << "Node \"" << static_cast<ThreadedNode*>(me)->getName() 
           << "\" (pid: " << getpid() 
           << ") caused a uncaught Exception in its internal thread:" << endl;
      cerr << "Unknown Exception!" << endl;
      cerr << NamedObject::color_array[NamedObject::DEFAULT_COLOR];
    }
    return 0; // to satisfy void* requirements, only checked with pthread_join
  }

  pid_t ThreadedNode::getPID() {
    return thread_pid;
  }

  void ThreadedNode::setPID(pid_t new_pid) {
        thread_pid = new_pid;
  }

  void ThreadedNode::lowerPriority(const unsigned int& amount) {
    if (thread_pid == 0) {
        return;
    }
    int new_priority = getpriority(PRIO_PROCESS,thread_pid) + amount;
    if (new_priority > 20) {
        new_priority = 20;
    }
    if (setpriority(PRIO_PROCESS,thread_pid,new_priority) != 0) {
        cerr << "setting priority failed!";
    }
  }

  Result ThreadedNode::setPriority(unsigned int level) {
    if (!isOutputInitialized()) {
      ERROR_STREAM2("Wrong state to set priority level! (state != INITIALIZED)" << endl);
      return FAILURE;
    }
    if ((level == 0) || (level > 20)) {
      ERROR_STREAM2("Priority level out of range! (1..20)" << endl);
      return FAILURE;
    }
    thread_priority = level;
    return SUCCESS;
  }


//------ pthread scheduling methods (only allowed for root)
  void ThreadedNode::enableCustomScheduling(Thread::Scheduling_Policy policy) {
    thread.enableCustomScheduling(policy);
  }
  void ThreadedNode::setMinPriority() {
    thread.setMinPriority();
  }
  void ThreadedNode::setMidPriority() {
    thread.setMidPriority();
  }
  void ThreadedNode::setMaxPriority() {
    thread.setMaxPriority();
  }

}
