/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_TRANSPORTSTRATEGY_HPP
#define NMM_TRANSPORTSTRATEGY_HPP

#include "ISignalReceiver.hpp"
#include "TSAddress.hpp"
#include "nmm/interfaces/base/ITransportStrategyImpl.hpp"
#include "nmm/interfaces/base/ITransportStrategyConfigImpl.hpp"

namespace NMM {
class Message;
class CEvent;
class Event;
class CommunicationChannel;
class IExternalBufferManager;
class IValueStream;
class OValueStream;

/** This TSAddress is used for a local connection
  */
class LocalAddress : public TSAddress {
public:
/** @name Constructor & Destructor
  */
//@{
/** Default constructor
  */
  LocalAddress(const NMMReference& = 0);

/** Virtual copy constructor
  */
  virtual LocalAddress* copy() const { return new LocalAddress(*this); }
//@}
};

class TransportStrategy :
    public ITransportStrategyImpl,
    public ITransportStrategyConfigImpl,
    public IMessageReceiver,
    public IMessageSender,
    public IEventSender,
    public ISignalReceiver
{
public:
/** @name Constructor & Destructor
  */
//@{
/** Default constructor
  */
  TransportStrategy();

/** Destructor
  */
  virtual ~TransportStrategy();

/** This method is called if the TS is no longer need. This method must not blocked
  */
  virtual  void scheduleDestruction()                            {  }

//@}

/** @name Set & Get  the Receiver
  */
//@{
/** forwards any received event to event_receiver
  */
  virtual void setEventReceiver(IEventReceiver* event_receiver)        ;

/** forwards any received message to message_receiver
  */
  virtual void setMessageReceiver(IMessageReceiver* message_receiver)  { __message_receiver = message_receiver; }

/** sets a new SignalReceiver
  */
  virtual void setSignalReceiver(ISignalReceiver* signal_receiver)     { __signal_receiver  = signal_receiver; }

/** returns the current SignalReceiver
  */
  virtual ISignalReceiver* getSignalReceiver()                         { return __signal_receiver; }

/** returns the current SignalReceiver
  */
  virtual const ISignalReceiver* getSignalReceiver()             const { return __signal_receiver; }

/** returns the current event receiver
  */
  virtual       IEventReceiver* getEventReceiver()                     { return __event_receiver; }

/** returns the current event receiver
  */
  virtual const IEventReceiver* getEventReceiver()               const { return __event_receiver; }

/** returns the current message receiver
  */
  virtual       IMessageReceiver* getMessageReceiver()                 { return __message_receiver; }

/** returns the current message receiver
  */
  virtual const IMessageReceiver* getMessageReceiver()           const { return __message_receiver; }

/** Incoming Events from this sender are not frowarded, they were processed by the TransportStrategy.
  * This is need if the TransportStrategy is accessed through interfaces.
  */
  inline  void addEventSender(IEventSender* sender)                    { __sender_list.push_back(sender); }

/** This method removes an EventSender from the internal list of EventSender, that are directly \
  * processed.
  */
  inline  void removeEventSender(IEventSender* sender)                 { __sender_list.remove(sender); }
//@}

/** @name TransportStratgegy information
  */
//@{
/** This method returns the typename of the TransportStrategy
  */
  virtual string getStrategyType() const = 0;

/** This method returns an event prototype for the event with kgiven key
  */
  virtual Event* getEventPrototype(const string&)                const { return 0;}

/** This mehtod sets the BufferManager that should used to allocate Buffers.
  */
  inline void setBufferManager(IExternalBufferManager* buffer_manager) {__buffer_manager = buffer_manager; }

/** This mehtod sets the CommunicationChannel that manages this TransportStrategy .
  */
  inline void setBindingChannel(CommunicationChannel* com_channel)     { __com_channel = com_channel; }

/** This method uses the current TransportStrategy
  */
  inline CommunicationChannel* getBindingChannel()                     { return __com_channel; }
//@}

/** @name Connection & Configure
  */
//@{
/** This mehtod connects the TransportStrategy to the given TSAddress and throws an exception
  * if an error occurs.
  */
  virtual void setAddress(const TSAddress& address)
    throw (const BindingException&) = 0;

/** This method set the ValueStream that should be used to serialize/deserialize incoming messages
  */
  virtual void setValueStream(const IValueStream& , const OValueStream&) = 0;

/** This method set the ValueStream that should be used to serialize/deserialize incoming messages
  */
  virtual void setValueStream(const string& , const string&) {};

/** This method returns the Interface to the partner Transport-Strategy. The TransportStrategy must be
  * connected.
  */
  ITransportStrategy* getPartnerStrategy()             const;

/** Reliably send a configuration event.
  * This method does the same as receiveEvent with the exception that the event is delievered reliably.
  * Unreliable transport strategies must deliver the event through a connection manager.
  */
  virtual Result sendReliableEvent(Event&) const;

/** Reliably send a configuration event.
  * This method does the same as sendEvent with the exception that the event is delievered reliably.
  * Unreliable transport strategies must deliver the event through a connection manager.
  */
  virtual Result receiveReliableEvent(Event&, const IEventSender* = 0) const;
//@}
  
/** @name ITransportStrategyConfig interface
 */
//@{
/** This method configures the transport strategy for an input jack.
  * Some protocols need special configuration that depends on the stream format and whether data is sent or
  * received. This method configures transport strategies for such special protocols for use with input jacks.
  * The method must be called again whenever the stream format changes.
  * @param format Stream format
  */
  virtual void configureForInputJack(const Format& format);
  
/** This method configures the transport strategy for an output jack.
  * Some protocols need special configuration that depends on the stream format and whether data is sent or
  * received. This method configures transport strategies for such special protocols for use with output jacks.
  * The method must be called again whenever the stream format changes.
  * @param format Stream format
  */
  virtual void configureForOutputJack(const Format& format);
//@}
protected:

/** @name Internal Transport Strategy methods
  */
//@{
/** This method returns the current BufferManager that should be used to allocate Buffer
  */
  inline IExternalBufferManager* getBufferManager()             { return __buffer_manager; }

/** This method returns the current BufferManager that should be used to allocate Buffer
  */
  inline IExternalBufferManager* getBufferManager()      const  { return __buffer_manager; }

/** This method the list of specific EventSender. Incoming Events of these EventSender
  * are not forwarded, they were processed by this object.
  */
  inline       list<IEventSender*>& getEventSenderList()        { return __sender_list;}

/** This method the list of specific EventSender. Incoming Events of these EventSender
  * are not forwarded, they were processed by this object.
  */
  inline const list<IEventSender*>& getEventSenderList() const  { return __sender_list;}
//@}
private:
  IMessageReceiver*               __message_receiver;      ///< Pointer to the MessageReceiver, that receives incoming messages
  IEventReceiver*                 __event_receiver;        ///< Pointer to the EventReceiver, that receives incoming events
  ISignalReceiver*                __signal_receiver;       ///< Pointer to the SignalReceiver, that is notfied, if messages can be send
  mutable IExternalBufferManager* __buffer_manager;        ///< Pointer to the BufferManager that should be used
  CommunicationChannel*           __com_channel;           ///< Pointer to the CommuunicationChannel that manages the Strategy
  TransportStrategy*              __is_strategy;           ///< is != 0 if eventreceiver is an TransportStratgey
  list<IEventSender*>             __sender_list;           ///< Incoming Events of these EventSender are precessed
};



class LocalStrategy : public TransportStrategy {
public:
/** @name Constructor & Destructor
  */
//@{
/** Default constructor
  */
  LocalStrategy() : __partner_strategy(0) {}

/** Destructor
  */
  virtual ~LocalStrategy();

/** This method is called if the TS is no longer need. This method must not blocked
  */
  virtual void scheduleDestruction()                            { delete this; }
//@}

/** @name Sender & Receiver Interface
  */
//@{
/** Incoming messages are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result receiveMessage(Message*, const IMessageSender* = 0);

/** Incoming events are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result receiveEvent(const CEvent&, const IEventSender* = 0 );

/** Incoming CEvent's are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result receiveEvent(const Event&, const IEventSender* = 0 );

/** Incoming CEvent's are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result receiveEvent(const CEvent&, const IEventSender* = 0 ) const;

/** Incoming Event's are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result receiveEvent(const Event&, const IEventSender* = 0 )  const;

/** Incoming CEvent's are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result sendEvent(const CEvent&);

/** Incoming Event's are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result sendEvent(const Event&);

/** Incoming CEvent's are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result sendEvent(const CEvent&) const;

/** Incoming Event's are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result sendEvent(const Event&)  const;

/** Incoming messages are forwarded to the partner strategy wich sends it to the other object
  */
  virtual Result sendMessage(Message*);
//@}

/** @name Connection & Configure
  */
//@{
/** This mehtod connects the TransportStrategy to the given TSAddress and throws an exception
  * if an error occurs.
  */
  virtual void setAddress(const TSAddress& addr)
    throw (const BindingException&);

/** This methods sets the ValueStream. This TransportStrtategy does not use any ValueStreams
  * Its implementation prints only a warning if it is called.
  */
  virtual void setValueStream(const IValueStream& , const OValueStream&);

/** This method sets the partner strategy which is used for the bidirectrional communication
  */
  void setPartnerStrategy(LocalStrategy* partner_strategy);

/** incoming signals are ignored
  */
  virtual void receiveSignal(Signal ) {}

/** This method returns the typename of this class
  */
  virtual string getStrategyType()          const { return getStaticStrategyType(); }
//@}

/** @name Transmission methods
  */
//@{
/** forwards the message to the other object
  */
  Result forwardMessage(Message* message)   const;

/** forwards the event to the other object
  */
  Result forwardEvent(const Event&)         const;

/** forwards the CEvent to the other object
  */
  Result forwardEvent(const CEvent&)        const;
//@}

/** @name Static type information.
 */
//@{
/** This method returns the type name of strategy.
 */
  static inline string getStaticStrategyType() { return __strategy_type; }
//@}
  
protected: //ITransportStrategy methods
  virtual void setSendValueStream(const string& );
  virtual void setReceiveValueStream(const string&);

private:
  mutable LocalStrategy* __partner_strategy;    ///< pointer to the other LocalStrategy

private:
  static const char __strategy_type[];          ///< Static type name of this strategy
};


template<>
class TConstructor<LocalAddress> : public ValueConstructor {
public:
  TConstructor(const TypeInfo& t_info)
    : ValueConstructor(t_info)
  {}

  virtual LocalAddress*          createValue()     const {return new LocalAddress();}
  virtual TValue<LocalAddress>*  createTValue()    const {return new TValue<LocalAddress>();}

  virtual TValue<TSAddress*>* createTPointerValue()  const {return new TValue<TSAddress*>();}

  virtual TConstructor<LocalAddress>* copy()  const {return new TConstructor(*this); }
};

}
#endif


