/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


// #include <stl.h>
#include "IParameter.hpp"
#include "nmm/base/serialize/ValueStream.hpp"

namespace NMM {
  
  IParameter::IParameter(const string & pName) : Parameter(pName) 
  {
    m_parameter_type = Parameter::INT32;
    m_value_counter = m_value_list.begin();
    m_pair_counter  = m_value_list.begin();
  }
  
  IParameter::IParameter(string pName, int32_t pValue, float pWeight) : Parameter(pName) 
  {
    m_value_list.insert(make_pair(pValue,pWeight));
    m_parameter_type  = Parameter::INT32;
    m_value_counter = m_value_list.begin();
    m_pair_counter  = m_value_list.begin();
    m_default_value = pValue;
  }
  
  IParameter::~IParameter()
  {
  }
  
  void IParameter::addValue(void *pValue)
  {
    this->addValue(pValue, DEFAULT_WEIGHT);
  }

  void IParameter::addValue(void * pValue, float pWeight) 
  {
    if (m_value_list.empty()) m_default_value = *(static_cast<int32_t *>(pValue));
    m_value_list.insert(make_pair( *((int32_t *)pValue) , pWeight ) );
    m_value_counter = m_value_list.begin();
    m_pair_counter  = m_value_list.begin();
  }

  void IParameter::addValue(int32_t value, float weight)
  {
    if (m_value_list.empty()) m_default_value = value;
    m_value_list.insert(make_pair(value,weight));
    m_value_counter = m_value_list.begin();
    m_pair_counter  = m_value_list.begin();
  }

  void IParameter::showParameter() const
  {
    GLOBAL_DEBUG_STREAM2( "Parameter-Name  : " << m_parameter_name << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter-Type  : INT32" << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter-Weight: " << m_parameter_weight << endl);
    map<int32_t, float>::const_iterator i;
    for (i = m_value_list.begin();i != m_value_list.end();i++)  
      {
	GLOBAL_DEBUG_STREAM2( "Value: " << (*i).first << " - Weight: " << (*i).second << endl);
      }
  }

  ostream& IParameter::showParameter(ostream& os) const
  {
    os << m_parameter_name << "[weight:" << m_parameter_weight << "]=int(";

    map<int32_t, float>::const_iterator i;
    for (i = m_value_list.begin();i != m_value_list.end();i++)  
      {
        if (i != m_value_list.begin()) {
          os << ", ";
        }
	os << (*i).first << "[weight:" << (*i).second << "]";
      }
    os << ")";
    return os;
  }

  ostream& IParameter::showParameterVerbose(ostream& os) const
  {
    os << "Parameter-Name  : " << m_parameter_name << endl;
    os << "Parameter-Type  : INT32" << endl;
    os << "Parameter-Weight: " << m_parameter_weight << endl;
    map<int32_t, float>::const_iterator i;
    for (i = m_value_list.begin();i != m_value_list.end();i++)  
      {
 	os << "Value: " << (*i).first << " - Weight: " << (*i).second << endl;
      }
    return os;
  }
  
  bool IParameter::isValueIncluded(void * pValue) const
  {
    map<int32_t, float>::const_iterator i;
    for (i = m_value_list.begin(); i != m_value_list.end(); i++) 
      {
	if ( *((int32_t *)pValue) == (*i).first ) 
	  {
	    return true;
	  }
      }
    return false;
  }

  bool IParameter::isValueIncluded(int32_t value) const
  {
    map<int32_t, float>::const_iterator itr = m_value_list.find(value);
    if ( itr == m_value_list.end())
      return false;
    else
      return true;
  }
  
  void *IParameter::nextValue() const
  {
    map<int32_t, float>::const_iterator tempitr;
    if ( m_value_counter != m_value_list.end() ) 
      {
	tempitr = m_value_counter;
	m_value_counter++;
	return ((int32_t*)&((*tempitr).first));
      }
    else 
      {
	m_value_counter = m_value_list.begin();
	return NULL;
      }
  }

  map<int32_t,float>::const_iterator IParameter::nextPair() const
  {
    map<int32_t,float>::const_iterator tempitr;
    if (m_pair_counter != m_value_list.end() ) 
      {
	tempitr = m_pair_counter;
	m_pair_counter ++;
	return tempitr;
      }
    else 
      {
	m_pair_counter = m_value_list.begin();
	return m_value_list.end();
      }
  }

  Parameter * IParameter::duplicate() const
  {
    Parameter *tempParameter = new IParameter(m_parameter_name);
    int32_t tempint;
    map<int32_t, float>::const_iterator itr;
    for (itr = m_value_list.begin() ; itr != m_value_list.end() ; itr ++)
      {
	tempint = (*itr).first;
	tempParameter->addValue( & tempint, (*itr).second );
      }
    return (Parameter *)tempParameter;
  }
  
  float IParameter::getWeight(void *pValue) const
  {
    map<int32_t, float>::const_iterator itr = m_value_list.find(*(int32_t *)pValue);
    return (*itr).second;
  }

  bool IParameter::isParameterIdentical(const Parameter *param) const
  {
    // if the two parameter do not have the same type, return false.
    if (param->getParameterType() != Parameter::INT32)
      return false;
    // if the number of the values is not equal, the co-domain cannot be equal
    if (m_value_list.size() != param->getNumberOfValues() )
      return false;
    for (map<int32_t,float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if (!param->isValueIncluded((int32_t*)&((*itr).first)))
	  return false;
      }
    return true;
  }

  bool IParameter::isParameterMatching(const Parameter * reference) const
  {
    for (map<int32_t,float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if (reference->isValueIncluded((*itr).first))
	  return true;
      }
    return false;
  }

  uint IParameter::getNumberOfValues() const
  {
    return m_value_list.size();
  }

  Parameter *IParameter::getIntersection(const Parameter *reference) const
  {
    IParameter *new_parameter = new IParameter(m_parameter_name);
    int32_t tempint;
    for (map<int32_t, float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	tempint = (*itr).first;
	if (reference->isValueIncluded(&tempint))
	  {
	    new_parameter->addValue(&tempint, ((*itr).second + reference->getWeight(&tempint))/2);
	  }
      }
    return new_parameter;
  }

  void * IParameter::getFirstValue() const
  {
    m_value_counter = (++m_value_list.begin());
    return (int32_t*)&((*m_value_list.begin()).first);
  }

  float IParameter::getStandardizedWeight() const
  {
    float local_weight=0;
    for (map<int32_t, float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	local_weight += (*itr).second;
      }
    return ((local_weight/m_value_list.size())*m_parameter_weight);
  }

  void IParameter::setDefaultValue(int32_t param)
  {
    m_default_value = param;
  }

  int32_t IParameter::getDefaultValue() const
  {
    return m_default_value;
  }

  int32_t IParameter::getLowestValue() const
  {
    return (*m_value_list.begin()).first;
  }

  int32_t IParameter::getHighestValue() const
  {
    return (*(--m_value_list.end())).first;
  }

  const int32_t_list IParameter::getIntValueList() const
  {
    int32_t_list result;
    for (map<int32_t, float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	result.push_back((*itr).first);
      }
    return result;
  }

  void IParameter::fromStream(IValueStream& iv) {
    if ( (iv.getCurrentType() != "IParameter") || (iv.getCurrentNamespace() != "NMM") ) {
      GLOBAL_ERROR_STREAM ("Invalid type or namespace:" << endl);
      GLOBAL_ERROR_STREAM ("Current type : " << iv.getCurrentType() << "                Expected type: IParameter" << endl);
      GLOBAL_ERROR_STREAM ("Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: NMM " << endl);

      iv.nextElement();
      return;
    }
    if (!iv.goDown()) {
      GLOBAL_ERROR_STREAM ("IParameter is empty" << endl);
      iv.nextElement();
      return;
    }

    string par_name;
    float weight =0.0;
    iv >> par_name;
    iv >> weight;

    setParameterName(par_name);
    setParameterWeight(weight);

    int32_t first = 0;
    float second  =0.0;
    while(!iv.isEmpty()) {
      iv >> first;
      iv >> second;
      addValue(first, second);
    }

    iv.goUp();
  }

  void IParameter::toStream(OValueStream& ov) const {
    ov.beginType("IParameter", "NMM");

    ov << getParameterName();
    ov << getParameterWeight();

    for (map<int32_t, float>::const_iterator it = m_value_list.begin(); it != m_value_list.end(); ++it)
    {
      ov << (*it).first;
      ov << (*it).second;
    }
    ov.endType();
  }
}
