/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "Buffer.hpp"
#include "BufferManager.hpp"
#include "nmm/utils/thread/Thread.hpp"

#include <iostream>

namespace NMM
{
  Buffer::Buffer (const size_t newsize,  const BufferManager* master)
    : Message(Message::BUFFER)
  {
    size.setValue(newsize);
    manager = master;
    realsize.setValue(size.getRValue());
    data.setValue(new char[size.getRValue()], size.getRValue());
    own_data.setValue(true); // we allocated the memory

    registerAttributes();

    setTypeName("Buffer");
    setNamespace("NMM");
  }

  Buffer::Buffer (const size_t newsize, char* initial_data, const BufferManager* master)
    : Message(Message::BUFFER)
  {
    size.setValue(newsize);
    manager = master;
    realsize.setValue(size.getRValue());
    data.setValue(initial_data, newsize);
    own_data.setValue(false); // memory was allocated for us.

    registerAttributes();

    setTypeName("Buffer");
    setNamespace("NMM");
  }

  Buffer::Buffer(const Buffer& buffer)
    : Message(buffer)
  {
    registerAttributes();
    *this = buffer;
  }

  Buffer::~Buffer ()
  { 
    //GLOBAL_DEBUG_STREAM2("Buffer Destructor, size " << size << " used " << realsize << endl);
    if (own_data.getValue()) delete [] data.getValue();
  }

  void Buffer::reset()
  {
    setMandatoryBit(false);
    realsize.setValue(size.getRValue());
  }

  Buffer *Buffer::getWriteableInstance()
  {
    MutexGuard mg(op_mutex);

    // Is Buffer referenced more than once?
    if(getCount() > 1) {
      
      Buffer *writeable = manager->requestBuffer(size.getRValue());
      if(writeable != 0) {
        // decrease reference count of this-Buffer:
        // do not call this->release(); it will block!
        // count-1 is still bigger than 0 (see if(count > 1))
        release();

        // copy data to writable Buffer
        *writeable = *this;

        return writeable;
      } 
      else {
        GLOBAL_ERROR_STREAM2( " * COULD NOT GET WRITEABLE BUFFER * " << endl);
        return 0;
      }
    } 
    else { // Buffer is used only once, and can be written
      // increaseCount(); // 2001-11-08 mlohse: no, do not! otherwise the ref-count would be 2 and will never reach 0 again!
      return this;
    }
  }

  void Buffer::erase() const  {
    manager->releaseBuffer(const_cast<Buffer*>(this));
  }

  char* Buffer::getData () const {
    return data.getRValue();
  }

  Result Buffer::setData (const size_t newsize, char *newdata) {
    if (getCount() > 1)
      return FAILURE;
    if (data.getValue() && own_data.getRValue() ) {
      delete [] data.getValue();
    }

    data.setValue(newdata, newsize);
    size.setValue(newsize);
    realsize.setValue(newsize);
    own_data.setValue(0);
    return SUCCESS;
  }

  size_t Buffer::getSize () const
  { return size.getRValue(); }


  size_t Buffer::getUsed () const
  { return realsize.getRValue() ; }

  void Buffer::setUsed (const size_t actused)
  {
    
    if(actused) {
      if(actused <= size.getRValue()) {
        realsize.setValue(actused) ;
        data.setValue(data.getValue(), realsize.getRValue());
      }
      else {
      realsize.setValue( size.getRValue() );
      data.setValue(data.getValue(), realsize.getRValue());
      GLOBAL_WARNING_STREAM2("Buffer::setUsed("<< actused << "), but only "
         << size.getRValue() << " available, set to " << size.getRValue() << endl);
      }
    }
    else {
      realsize.setValue(0);
      data.setValue(data.getValue(), realsize.getRValue());
    }
  }

  const Buffer& Buffer::operator=(const Buffer& b) 
  {
    MutexGuard mg(op_mutex);
    
    if(this == &b) {
      return *this;
    }

    this->setMandatoryBit(b.getMandatoryBit());
    this->setTimestamp(b.getTimestamp());

    // if used area in source buffer is too large,
    // copy as much data as possible
    if(b.getUsed() > this->getSize()) {
      memcpy(data.getRValue(), b.getData(), size.getRValue());
      this->setUsed(size.getRValue());
    }
    else { // copy everything
      memcpy(data.getRValue(), b.getData(), b.getUsed());
      this->setUsed(b.getUsed());
    }
    
    return *this;
  }

  void Buffer::registerAttributes() {
    insertArgument(&realsize);
    insertArgument(&data);
  }

  Buffer* Buffer::copy() const {
    increaseCount();
    return const_cast<Buffer*>(this);
  }

}
      










