/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_BUFFEREDSOURCE_HPP
#define NMM_BUFFEREDSOURCE_HPP NMM_BUFFEREDSOURCE_HPP

#include "nmm/base/ISignalReceiver.hpp"
#include "nmm/base/StreamQueue.hpp"

#include "nmm/interfaces/base/IQueued.hpp"

#include "nmm/utils/thread/Thread.hpp"
#include "nmm/utils/NamedObject.hpp"

#include <live/FramedSource.hh>

namespace NMM {
  
  class Buffer;
  
  /**
   *  @brief Buffered media source
   *
   *  This class is a liveMedia FramedSource which reads data from an NMM buffer.
   *  It has an internal StreamQueue to which buffers can be passed.
   */
  class BufferedSource :
    public FramedSource,
    public ISignalReceiver,
    public NamedObject
  {
  public:
    /**
     *  @brief Create a new BufferedSource
     *  @param env liveMedia UsageEnvironment
     *  @param parent Signal receiver for parent that provides the buffers
     *  @param sec_per_frame Specifies how long it takes to play a frame in seconds
     *  @param queue_mode Queue mode of operation
     *  @param queue_size Maximum number of entries in the queue
     *  @param no_merge Do not merge NMM buffers
     */
    static BufferedSource* createNew(UsageEnvironment& env, ISignalReceiver* parent, float sec_per_frame = 1.0,
                                     IQueued::Mode queue_mode = IQueued::MODE_SUSPEND,
                                     unsigned int queue_size = 10, bool no_merge = false);
    
    /**
     *  @brief Write buffer
     *  @return @c BLOCKED if the internal queue is full, otherwise @c SUCCESS.
     *
     *  This method puts the buffer into the queue. If the previous call to this method returned
     *  @c BLOCKED, then it may not be called again until the parent has received @c FREE_QUEUE_SLOT.
     */
    Result writeBuffer(Buffer* buffer);
    
    /**
     *  @brief Flush the stream
     *
     *  This method flushes the queue and unblocks the thread.
     */
    void flush();
    
    /**
     *  @brief Get stream queue
     */
    StreamQueue& getStreamQueue() throw();
    
  protected:
    /**
     *  @brief Constructor
     *
     *  This constructor is called by createNew.
     */
    BufferedSource(UsageEnvironment&, ISignalReceiver*, float = 1.0,
                   IQueued::Mode = IQueued::MODE_SUSPEND, unsigned int = 1, bool = false);
    
    /**
     *  @brief Destructor
     */
    virtual ~BufferedSource();
    
    /**
     *  @brief Handler called upon stopPlaying
     */
    virtual void doStopGettingFrames();
    
    /**
     *  @brief Receive a signal
     */
    virtual void receiveSignal(Signal signal);
    
  private:
    /**
     *  @name FramedSource functions
     */
    //@{
    /**
     *  @brief Get next frame
     *
     *  This method copies the next frame to the destination buffer
     */
    virtual void doGetNextFrame();
    
    /**
     *  @brief Get playing duration
     *
     *  This method seems to be obsolete...
     */
    virtual float getPlayTime(unsigned int) const;
    //@}
    
    /**
     *  @brief Fetch new data
     *
     *  This method fetches new data from the send queue. It ensures that a buffer is available for reading
     *  data from. If there is no buffer, the method tries to dequeue a buffer from the send queue. Messsages
     *  which are not buffers are discarded immediately. If the queue is empty, then this method waits until
     *  the queue becomes non-empty or BufferedSource::flush is called.
     */
    void fetchData();
    
  private:
    /** The internal queue.
     */
    StreamQueue __stream_queue;
    
    /** Current buffer to read from.
     */
    Buffer* __current_buffer;
    
    /** Current byte offset within the buffer.
     */
    size_t __buffer_offset;
    
    /** Seconds per frame (used by getPlayTime).
     */
    float __sec_per_frame;
    
    /** Internal thread mutex.
     */
    ThreadMutex __internal_mutex;
    
    /** Internal thread condition.
     */
    ThreadCondition __internal_cond;

    /** Flush method was called.
     */
    bool __flush;
    
    /** We are waiting for buffers.
     */
    bool __waiting;
    
    /** Don't merge buffers to frames.
     */
    bool __no_merge;
  };
  
  inline StreamQueue& BufferedSource::getStreamQueue() throw() {
    return __stream_queue;
  }
  
}

#endif
