/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "RTPAddress.hpp"
#include "RTPStrategy.hpp"


namespace NMM {
  
  void IPAddress::setIPAddress(const unsigned long& ip) {
    __ip_address = ip;
    __net_addr.s_addr = htonl(ip);
  }
  
  void IPAddress::setHostname(const string& host)
    throw(const BindingException&)
  {
    if ((__net_addr.s_addr = inet_addr(host.c_str())) == INADDR_NONE) {
      struct hostent* he;
      if (!(he = gethostbyname(host.c_str()))) {
        throw BindingException("Failed to resolve host " + host);
      }
      __net_addr.s_addr = *((uint32_t*) he->h_addr);
    }
    
    __ip_address = ntohl(__net_addr.s_addr);
  }
  
  void IPAddress::updateAttributes() {
    setIPAddress(__ip_address.getRValue());
  }
  
  IPAddress::IPAddress()
    : SerializedValue(TypeInfo("IPAddress", "NMM"))
  {
    setIPAddress(INADDR_NONE);
    insertArgument(&__ip_address);
  }
  
  IPAddress::IPAddress(const TypeInfo& typeinfo)
    : SerializedValue(typeinfo)
  {
    setIPAddress(INADDR_NONE);
    insertArgument(&__ip_address);
  }
  
  IPAddress::IPAddress(const TypeInfo& typeinfo, const string& hostname)
    throw (const BindingException&)
    : SerializedValue(typeinfo)
  {
    setHostname(hostname);
    insertArgument(&__ip_address);
  }
  
  IPAddress::IPAddress(const IPAddress& addr)
    : SerializedValue(addr)
  {
    setIPAddress(addr.getIPAddress());
    insertArgument(&__ip_address);
  }
  
  string IPAddress::toString() const {
    return string(inet_ntoa(getINetAddr()));
  }
  
  
  UnicastIPAddress::UnicastIPAddress()
    : IPAddress(getStaticTypeInfo())
  {}
  
  UnicastIPAddress::UnicastIPAddress(const string& hostname)
    throw (const BindingException&)
    : IPAddress(getStaticTypeInfo(), hostname)
  {}
  
  UnicastIPAddress::UnicastIPAddress(const UnicastIPAddress& addr)
    : IPAddress(addr)
  {}
  
  TConstructor<UnicastIPAddress> __unicast_ip_address(UnicastIPAddress::getStaticTypeInfo());
  TypeInfo UnicastIPAddress::getStaticTypeInfo() { return TypeInfo("UnicastIPAddress", "NMM"); }
  
  
  MulticastIPAddress::MulticastIPAddress()
    : IPAddress(getStaticTypeInfo())
  {}
  
  MulticastIPAddress::MulticastIPAddress(const string& hostname)
    throw (const BindingException&)
    : IPAddress(getStaticTypeInfo(), hostname)
  {}
  
  MulticastIPAddress::MulticastIPAddress(const MulticastIPAddress& addr)
    : IPAddress(addr)
  {}

  TConstructor<MulticastIPAddress> __multicast_ip_address(MulticastIPAddress::getStaticTypeInfo());
  TypeInfo MulticastIPAddress::getStaticTypeInfo() { return TypeInfo("MulticastIPAddress", "NMM"); }
  
  
  RTPAddress::RTPAddress(const NMMReference& reference /*= 0*/)
    : ConnectorAddress(getStaticTypeInfo(), TSAddress::LOCAL | TSAddress::SYSTEM | TSAddress::NETWORK, RTPStrategy::getStaticStrategyType(), reference),
      __rtp_port(0),
      __rtcp_port(0),
      __ip_address(NULL)
  {
    insertArgument(&__rtp_port);
    insertArgument(&__rtcp_port);
    insertArgument(&__ip_address);
  }
  
  RTPAddress::RTPAddress(const IPAddress& addr, unsigned int rtp_port, unsigned int rtcp_port, const NMMReference& reference /*= 0*/)
    : ConnectorAddress(getStaticTypeInfo(), TSAddress::LOCAL | TSAddress::SYSTEM | TSAddress::NETWORK, RTPStrategy::getStaticStrategyType(), reference),
      __rtp_port(rtp_port),
      __rtcp_port(rtcp_port),
      __ip_address(NULL)
  {
    setIPAddress(addr);
    
    insertArgument(&__rtp_port);
    insertArgument(&__rtcp_port);
    insertArgument(&__ip_address);
  }
  
  RTPAddress::RTPAddress(const RTPAddress& addr)
    : ConnectorAddress(addr),
      __rtp_port(addr.__rtp_port),
      __rtcp_port(addr.__rtcp_port),
      __ip_address(NULL)
  {
    setIPAddress(addr.getIPAddress());
    
    insertArgument(&__rtp_port);
    insertArgument(&__rtcp_port);
    insertArgument(&__ip_address);
  }
  
  void RTPAddress::setIPAddress(const IPAddress& ip_address) {
    if (__ip_address.getRValue()) {
      delete __ip_address.getRValue();
    }
    
    __ip_address = ip_address.copy();
  }
  
  TConstructor<RTPAddress> __rtp_address(RTPAddress::getStaticTypeInfo());
  TypeInfo RTPAddress::getStaticTypeInfo() { return TypeInfo("RTPAddress", "NMM"); }
  
} // namespace NMM
