/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_RTP_ADDRESS_HPP
#define NMM_RTP_ADDRESS_HPP

#include "nmm/base/ConnectorAddress.hpp"
#include "nmm/base/BaseExceptions.hpp"

#include <arpa/inet.h>
#include <netdb.h>

namespace NMM {
  
  /** This structure represents an IP Address.
   * It can not be instantiated directly. Create a UnicastIPAddress or a MulticastIPAddress instead.
   */
  class IPAddress : public SerializedValue {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Dummy constructor.
     * XXX: Needed because of TValue<T*> limitations. Do not use.
     */
    IPAddress();
    
    /** Virtual Copy Constructor
     */
    inline virtual IPAddress* copy() const {
      return new IPAddress(*this);
    }
    //@}
    
    /** @name Access Methods
     */
    //@{
    inline unsigned long getIPAddress() const { return __ip_address.getRValue(); }
    
    /** This method sets the IP address
     */
    void setIPAddress(const unsigned long& ip);
    
    /** This method sets the host name
     * @exception BindingException if the host was not found
     */
    void setHostname(const string& host)
      throw(const BindingException&);
    //@}
    
    /** @name Access Internal Structures
     */
    //@{
    /** This method returns the INetAddr of the remote host.
     */
    inline const in_addr& getINetAddr() const { return __net_addr; }
    
    /** This method returns the INetAddr of the remote host.
     */
    inline       in_addr& getINetAddr()       { return __net_addr; }
    //@}
    
    /** @name Debugging
     */
    //@{
    string toString() const;
    //@}
    
  protected:
    /** Prepare all members for serialization
     */
    void updateAttributes();
    
  protected:
    /** @name Private Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    IPAddress(const TypeInfo& typeinfo);
    
    /** Constructor with remote hostname
     * @exception BindingException if the host was not found
     */
    IPAddress(const TypeInfo& typeinfo, const string& hostname)
      throw (const BindingException&);
    
    /** Copy Constructor
     */
    IPAddress(const IPAddress&);
    
  private:
    in_addr __net_addr;                       ///< Internal C struct for IP address
    TValue<unsigned long> __ip_address;       ///< Stores the IP address of the remote host
  };
  
  /** This structure represents an IP address for unicast.
   */
  class UnicastIPAddress : public IPAddress {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    UnicastIPAddress();
    
    /** Constructor with remote hostname
     * @exception BindingException if the host was not found
     */
    UnicastIPAddress(const string& hostname)
      throw (const BindingException&);
    
    /** Copy Constructor
     */
    UnicastIPAddress(const UnicastIPAddress&);
    
    /** Virtual Copy Constructor
     */
    inline virtual UnicastIPAddress* copy() const {
      return new UnicastIPAddress(*this);
    }
    //@}
    
  public:
    /** This method returns the static type information for this class.
     */
    static TypeInfo getStaticTypeInfo();
  };
  
  /** This structure represents an IP address for multicast.
   */
  class MulticastIPAddress : public IPAddress {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    MulticastIPAddress();
    
    /** Constructor with remote hostname
     * @exception BindingException if the host was not found
     */
    MulticastIPAddress(const string& hostname)
      throw (const BindingException&);
    
    /** Copy Constructor
     */
    MulticastIPAddress(const MulticastIPAddress&);

    /** Virtual Copy Constructor
     */
    inline virtual MulticastIPAddress* copy() const {
      return new MulticastIPAddress(*this);
    }
    //@}
    
  public:
    /** This method returns the static type information for this class.
     */
    static TypeInfo getStaticTypeInfo();
  };
  
  /** This class implements an RTP address which stores all information to establish a connection.
   */
  class RTPAddress : public ConnectorAddress {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    RTPAddress(const NMMReference& = 0);
    
    /** Constructor, with IP address, RTP port and RTCP port
     */
    RTPAddress(const IPAddress& ip_address, unsigned int rtp_port, unsigned int rtcp_port, const NMMReference& = 0);
    
    /** Copy Constructor
     */
    RTPAddress(const RTPAddress&);
    
    /** Virtual copy Constructor
     */
    RTPAddress* copy() const {
      return new RTPAddress(*this);
    }
    //@}
    
    /** @name Access Methods
     */
    //@{
    /** This method sets the RTP port
     */
    inline void setRTPPort(unsigned int port) {
      __rtp_port = port;
    }

    /** This method sets the RTCP port
     */
    inline void setRTCPPort(unsigned int port) {
      __rtcp_port = port;
    }
    
    /** This method sets the IP address
     */
    void setIPAddress(const IPAddress& ip_address);
    
    /** This method returns the RTP port
     */
    inline unsigned int getRTPPort() const {
      return __rtp_port.getRValue(); 
    }
    
    /** This method returns the RTCP port
     */
    inline unsigned int getRTCPPort() const {
      return __rtcp_port.getRValue();
    }
    
    /** This method returns the IP address
     */
    inline const IPAddress& getIPAddress() const {
      return *__ip_address.getValue();
    }
    //@}
    
  private:
    TValue<unsigned int>  __rtp_port;         ///< Stores the RTP port
    TValue<unsigned int>  __rtcp_port;        ///< Stores the RTCP port
    TValue<IPAddress*>    __ip_address;       ///< Stores the IP address of the remote host
    
  public:
    /** This method returns the static type information for this class.
     */
    static TypeInfo getStaticTypeInfo();
  };
  
  
  template<>
  class TConstructor<RTPAddress> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual RTPAddress*          createValue()          const {return new RTPAddress();}
    virtual TValue<RTPAddress>*  createTValue()         const {return new TValue<RTPAddress>();}
    
    virtual TValue<TSAddress*>*  createTPointerValue()  const {return new TValue<TSAddress*>();}
    
    virtual TConstructor<RTPAddress>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };
  
} // namespace NMM

#endif // NMM_RTP_ADDRESS_HPP
