/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "TCPGeneric.hpp"

#include <iostream>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>    
#include <netinet/tcp.h>
#include <errno.h>

namespace NMM {


TCPGeneric::TCPGeneric()
{

}


TCPGeneric::~TCPGeneric() {
}

void TCPGeneric::createException(int __errno) const {
  switch(__errno) {
    case EPIPE:
      cerr << "The other end closed the socket unexpectedly or a read is executed on a shut down socket." << endl;
      break;
    case ETIMEDOUT:
      cerr << "The other end didn't acknowledge retransmitted data after some time." << endl;
      break;  
    case EAFNOSUPPORT:
      cerr << "Passed socket address type in sin_family was not AF_INET." << endl;
      break;
    default:
      cerr << "Unknown error" << endl;
  }
}

void TCPGeneric::enableNoDelay(const bool& value) {
  int v1 = value;
  if (setsockopt (getStreamSocket().getSocket(), SOL_TCP, TCP_NODELAY, &v1, sizeof (v1))== -1)
    createException(errno);
}

bool TCPGeneric::isDelayEnabled() const {
  int response = false;
  unsigned int size = sizeof(response);
  if (getsockopt(getStreamSocket().getSocket(), SOL_TCP, TCP_NODELAY, reinterpret_cast<char*>(&response), &size) == -1)  {
    createException(errno);
  }
  return response;
}

void TCPGeneric::setMaxSegment(const int& value){
  if ( setsockopt(getStreamSocket().getSocket(), SOL_TCP, TCP_MAXSEG, reinterpret_cast<const char*>(&value), sizeof(value)) == -1) {
    createException(errno);
  }
}

int  TCPGeneric::getMaxSegment() const {
  int retval = 0;
  unsigned int size = sizeof(retval);
  if (getsockopt(getStreamSocket().getSocket(), SOL_TCP, TCP_MAXSEG, reinterpret_cast<char*>(&retval), &size) == -1)  {
    createException(errno);
  }
  return size;
}

void TCPGeneric::enablePartialFrames(const bool& value){
  if ( setsockopt(getStreamSocket().getSocket(), SOL_TCP,  TCP_CORK, reinterpret_cast<const char*>(&value), sizeof(value)) == -1) {
    createException(errno);
  }
}

bool TCPGeneric::isPartialFramesEnabled(){
  int retval = 0;
  unsigned int size = sizeof(retval);
  if (getsockopt(getStreamSocket().getSocket(), SOL_TCP, TCP_CORK, reinterpret_cast<char*>(&retval), &size) == -1)  {
    createException(errno);
  }
  return retval;
}

int  TCPGeneric::getUnreadQueueCount() const{
  int value = 0;
  if ( ioctl(getStreamSocket().getSocket(), FIONREAD, &value) == -1) {
    createException(errno);
  }
  return value;
}

bool TCPGeneric::allUrgentDataSend()   const{
  int value = 0;
  if ( ioctl(getStreamSocket().getSocket(), SIOCATMARK, &value) == -1) {
    createException(errno);
  }
  return value;
}

int  TCPGeneric::getUnsendQueueCount() const{
 int value = 0;
  if ( ioctl(getStreamSocket().getSocket(), TIOCOUTQ, &value) == -1) {
    createException(errno);
  }
  return value;
}

void TCPGeneric::setRecvBufferSize(const int& value){
  if (setsockopt (getStreamSocket().getSocket(), SOL_SOCKET, SO_RCVBUF, &value, sizeof (value))== -1)
    createException(errno);
}

int TCPGeneric::getRecvBufferSize() const{
  int retval = 0;
  unsigned int size = sizeof(retval);
  if (getsockopt(getStreamSocket().getSocket(), SOL_SOCKET, SO_RCVBUF, reinterpret_cast<char*>(&retval), &size) == -1)  {
    createException(errno);
  }
  return retval;
}

void TCPGeneric::setSendBufferSize(const int& value){
  if (setsockopt (getStreamSocket().getSocket(), SOL_SOCKET, SO_SNDBUF, &value, sizeof (value))== -1)
    createException(errno);
}

int TCPGeneric::getSendBufferSize() const{
  int retval = 0;
  unsigned int size = sizeof(retval);
  if (getsockopt(getStreamSocket().getSocket(), SOL_SOCKET, SO_SNDBUF, reinterpret_cast<char*>(&retval), &size) == -1)  {
    createException(errno);
  }
  return retval;
}

void TCPGeneric::enableKeepAlive(const bool& value) {
  if (setsockopt (getStreamSocket().getSocket(), SOL_SOCKET, SO_KEEPALIVE, &value, sizeof (value))== -1)
    createException(errno);
}

bool TCPGeneric::getKeepAlive() const{
  int retval = 0;
  unsigned int size = sizeof(retval);
  if (getsockopt(getStreamSocket().getSocket(), SOL_SOCKET, SO_KEEPALIVE, reinterpret_cast<char*>(&retval), &size) == -1)  {
    createException(errno);
  }
  return retval;
}

void TCPGeneric::setPriority(const int& value){
  if (setsockopt (getStreamSocket().getSocket(), SOL_SOCKET, SO_PRIORITY, &value, sizeof (value))== -1)
    createException(errno);
}

int TCPGeneric::getPriority() const{
  int retval = 0;
  unsigned int size = sizeof(retval);
  if (getsockopt(getStreamSocket().getSocket(), SOL_SOCKET, SO_RCVBUF, reinterpret_cast<char*>(&retval), &size) == -1)  {
    createException(errno);
  }
  return retval;
}

}




