/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TCPStrategy.hpp"

#include "nmm/base/serialize/net/BufferedNetStream.hpp"
#include "nmm/base/serialize/net/UnbufferedNetStream.hpp"

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Event.hpp"
#include "nmm/base/NMMObject.hpp"

#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/memorymgr/IExternalBufferManager.hpp"

#include "nmm/base/serialize/Value.hpp"
#include "nmm/base/serialize/XMLValueStream.hpp"
#include "nmm/base/serialize/MagicNumberValueStream.hpp"
#include "nmm/base/serialize/ValueStream.hpp"
#include "nmm/utils/NamedObject.hpp"

#include "nmm/base/TSFactory.hpp"
#include "nmm/interfaces/net/IObjectAcceptor.hpp"

#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>    
#include <netinet/tcp.h>
#include <errno.h>

namespace NMM {


TCPAddress::TCPAddress (const NMMReference& reference)
  : TSAddress(reference, TCPStrategy::getStaticStrategyType(), TypeInfo("TCPAddress", "NMM"), TSAddress::LOCAL|TSAddress::SYSTEM|TSAddress::NETWORK)
{
  insertArgument(&__net_addr);
}

TCPAddress::TCPAddress (const string& name, int port, const NMMReference& reference)
  : TSAddress(reference, TCPStrategy::getStaticStrategyType(), TypeInfo("TCPAddress", "NMM"), TSAddress::LOCAL|TSAddress::SYSTEM|TSAddress::NETWORK)
{
  insertArgument(&__net_addr);
  setPort(port);
  setHostname(name);
}

TCPAddress::TCPAddress (const INetAddr& net_addr, const NMMReference& reference)
  : TSAddress(reference, TCPStrategy::getStaticStrategyType(), TypeInfo("TCPAddress", "NMM"), TSAddress::LOCAL|TSAddress::SYSTEM|TSAddress::NETWORK),
    __net_addr(net_addr)
{
  insertArgument(&__net_addr);
}

TCPAddress::TCPAddress (const TCPAddress& addr)
  : TSAddress(addr)
{
  insertArgument(&__net_addr);
  *this = addr;
}

void TCPAddress::setPort(int port) {
  __net_addr.setPort(port);
  TSAddress::setPort(port);
}

void TCPAddress::setHostname(const string& host)      {
  __net_addr.setAddress(host.c_str());
  TSAddress::setHostname(host);
}


TCPStrategy::TCPStrategy()
  :
    /* Begin> Initialize all the Attributes with sensfull values */
    __event_condition(__event_mutex), __event_sending_mutex(ThreadMutex::RECURSIVE), __current_value(0), __current_result(FAILURE),
    __o_stream(0), __i_stream(0),
    __message_stream(0),
    __is_running(true), __thread_condition(__condition_thread_mutex),
    __send_message(0), __blocked_message(0), __send_message_mutex(ThreadMutex::RECURSIVE),
    __current_address(0), 
    __blocked(false), __blocked_condition(__blocked_mutex)
    /* End <*/
{
  /* Set the TCP net stream */
  setNetStream(BufferedNetStream(getStreamSocket()));

  /* Set the default stream */
  MagicNumberIValueStream istream;
  MagicNumberOValueStream ostream;
  setValueStream(istream, ostream);

  /* activate the thread */
  activateThread();
}

TCPStrategy::~TCPStrategy()
{
  if(getSignalReceiver() && __send_message) {
    if(__send_message->getDirection() == Message::DOWNSTREAM) {
      GLOBAL_DEBUG_STREAM("getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT); in ~TCPStrategy" << endl);
      getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
    }
    else if(__send_message->getDirection() == Message::UPSTREAM) {
      GLOBAL_DEBUG_STREAM("getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT); in ~TCPStrategy" << endl);
      getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
    }
    else {
      GLOBAL_ERROR_STREAM("Unknown message direction in ~TCPStrategy" << endl);
    }
  }
  if(__send_message) {
    __send_message->release();
  }

  setSignalReceiver(0);
  setEventReceiver(0);
  setMessageReceiver(0);

  closeConnection(getStreamSocket().getSocket());

  /* That the running flag to false */
  __condition_thread_mutex.lock();
  __is_running = false;
  __condition_thread_mutex.unlock();

  /* Notify the thread to leave the while loop */
  conditionNotify();

  __blocked_mutex.lock();
  __blocked_condition.notify();
  __blocked_mutex.unlock();

  /* Deactivate the thread */
  deactivateThread();
  /* Destroy value streams */
  if (__i_stream) {
    delete __i_stream;
    __i_stream = 0;
  }
  if (__o_stream) {
    delete __o_stream;
    __o_stream = 0;
  }
  
  /* Destroy net streams */
  delete __message_stream;
  __message_stream = 0;
}


/* Begin>------The sendEvent methods ------*/
Result TCPStrategy::sendEvent( const CEvent& cevent) {
  /* forward the Event to send value */
  return sendValue(cevent, CEVENT_REQUEST);
}

Result TCPStrategy::sendEvent( const CEvent& cevent) const {
  /* forward the Event to send value */
  return sendValue(cevent, CEVENT_REQUEST);
}

Result TCPStrategy::sendEvent(const Event& event) {
  /* forward the Event to send value */
  return sendValue(event, EVENT_REQUEST);
}

Result TCPStrategy::sendEvent(const Event& event) const{
  /* forward the Event to send value */
  return sendValue(event, EVENT_REQUEST);
}

Result TCPStrategy::receiveEvent(const CEvent& event, const IEventSender* sender) {
  /* If this TCPStrategy is the sender don't send the event again */
  if (sender == this) {
    return SUCCESS;
  }
  if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
    return NMMObject::receiveEvent(event, sender);
  }

  return sendValue(event, CEVENT_REQUEST);
}

Result TCPStrategy::receiveEvent(const Event& event, const IEventSender* sender) {
  /* If this TCPStrategy is the sender don't send the event again */
  if (sender == this) {
    return SUCCESS;
  }

  if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
    return NMMObject::receiveEvent(event, sender);
  }

  /* forward the Event to send value */
  return sendValue(event, EVENT_REQUEST);
}

Result TCPStrategy::receiveEvent(const CEvent& event, const IEventSender* sender ) const {
  /* If this TCPStrategy is the sender don't send the event again */
  if (sender == this) {
    return SUCCESS;
  }

  if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end())
    return NMMObject::receiveEvent(event, sender);

  /* forward the Event to send value */
  return   sendValue(event, CEVENT_REQUEST);
}

Result TCPStrategy::receiveEvent(const Event& event, const IEventSender* sender )  const {
  /* If this TCPStrategy is the sender don't send the event again */
  if (sender == this) {
    return SUCCESS;
  }


  if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end())
  {
    return NMMObject::receiveEvent(event, sender);
  }

  /* forward the Event to send value */
  return sendValue(event, EVENT_REQUEST);
}

/* End<------The sendEvent methods ------*/

Result TCPStrategy::sendValue( const Value& value, TCPMessage message_type) const {
  /* If the thread is no longer running return FAILURE */
  if (!__is_running)
    return FAILURE;
  
  try {
    /* Lock the __event_mutex to ensure that no other event can be send. Use a MutexGuard to unlock
     * the mutex after method is returned and return value is stored by the sender.*/
    MutexGuard guard(__event_sending_mutex);
    
    /* Store a pointer to the given value. So the internal thread can wirte the response to it*/
    __current_value = const_cast<Value*>(&value);

    /* Begin> create a copy from the stream and set the TCPStrategy as partner */
    __o_stream->setOPartnerStream(const_cast<NetStream*>(__message_stream));
    /* End< */

    /* Before writing, lock the condition to ensure that the dispatchResponse method can not
     * access the event until we have called wait*/
    MutexGuard event_guard(__event_mutex);
    
    /*write the message type to the stream. The receiver needs this information to know
     * which receiveEvent-method must be called; the receiveEvent(const Event&) or the
     * receiveEvent(const CEvent&)
     */
    *__o_stream << static_cast<int>(message_type);

    /* Disable the out_values */
    __current_value->disableMemoryType(Value::OUTVALUE);

    /* write the value to the stream */
    *__o_stream << value;
    
    /* ensure that all data is written (similar to a 'flush')*/
    __o_stream->writeData();

    /* wait for the response ...*/
    __event_condition.wait();
    /* reset the internal pointer ...*/
    __current_value = 0;

    /* return the result */
    return __current_result;
  }
  /* Begin> Error handling */
  catch (const TransmissionException& te) {
    const_cast<TCPStrategy*>(this)->closeConnection(getStreamSocket().getSocket());
    throw BindingException(te.getComment());
  } catch(const ConnectionClosed& ce){
    const_cast<TCPStrategy*>(this)->closeConnection(getStreamSocket().getSocket());
    throw BindingException(ce.getComment());
  }
  /* End< */
}


void TCPStrategy::sendResponse(Result result, const Value* ce, TCPMessage message_type) {
  /* No other events can be send */
  MutexGuard event_guard(__event_sending_mutex);
  
  try {
    /* Begin> Create the response */
    __o_stream->setOPartnerStream(__message_stream);

    /* write the message_type */
    (*__o_stream) << static_cast<int>(message_type);

    /* write the result */
    (*__o_stream) << static_cast<const Result&>(result);

    /* If given value != 0*/
    if (ce)
      /* write the value */
      (*__o_stream) << *ce;

    /* call writeData to finish the transmission */
    __o_stream->writeData();

    /* and return */
    return;
  }
  /* Begin> Error handling */
  catch (const TransmissionException& te) {
    GLOBAL_ERROR_STREAM(te << endl);
    //throw te;
  } catch(const ConnectionClosed& ce){
    GLOBAL_ERROR_STREAM(ce << endl);
    //throw ce;
  } catch (const Exception& e) {
    GLOBAL_ERROR_STREAM(ce << endl);
  }
  closeConnection(getStreamSocket().getSocket());
  /* End */
}
/* -------------------------------------- */

Result TCPStrategy::sendMessage(Message* message) {
  return receiveMessage(message, 0);
}

Result TCPStrategy::receiveMessage(Message* message, const IMessageSender* ) {
  if (!message) {
    return FAILURE;
  }


  /* Begin> Store the message pointer */
  __condition_thread_mutex.lock();
  __send_message_mutex.lock();

  if (__send_message) {
    GLOBAL_ERROR_STREAM( "ARGH ERROR IN TCPSTARTEGY __send_message != 0 !!!!!!!!!!!!!" << endl);
    cerr << "ARGH ERROR IN TCPSTARTEGY __send_message != 0 !!!!!!!!!!!!! EVENT? " << (message->getType() == Message::EVENT) << endl;
    __blocked_message = message;
  } else {
    __send_message = message;
  }

  __send_message_mutex.unlock();
  __condition_thread_mutex.unlock();
  /* End< */

  /* Notify the Thread */
  conditionNotify();

  /* Return blocked*/
  return BLOCKED;
}

NetHandler::NetResult TCPStrategy::receiveData(int) {
  /* If the transport strategy is not runinngm return remove */
  if (!__is_running ) {
    return REMOVE;
  }

  /* Dispatch the message */
  dispatchMessage();

  /* After this there may be other data*/
  while (__is_running) {
    if ( !__message_stream->isEmpty() && !__blocked && __is_running )  {
      dispatchMessage();
    } else {
      if (__message_stream->isEmpty()  && !__blocked && __is_running ) {
        return VOID;
      }
      __blocked_mutex.lock();
      if ( __blocked ) {
        __blocked_condition.wait();
      }
     __blocked_mutex.unlock();
    }
  }
  return REMOVE;
}



void TCPStrategy::dispatchMessage() {
  try {
    __i_stream->setIPartnerStream(__message_stream);
    __i_stream->readData();
    int tmp_message;

    *__i_stream >> tmp_message;
    TCPMessage message_type = static_cast<TCPMessage>(tmp_message);
    switch(message_type) {
      case EVENT_RESPONSE :
        dispatchEventResponse(*__i_stream);
        break;

      case EVENT_REQUEST  :
        dispatchEventRequest(*__i_stream);
        break;

      case CEVENT_REQUEST  :
        dispatchCEventRequest(*__i_stream);
        break;

      case INSTREAM_BUFFER :
        dispatchInstreamMessage(*__i_stream, message_type);
        break;

      case INSTREAM_MESSAGE :
        dispatchInstreamMessage(*__i_stream, message_type);
        break;

      case CONFIG_EVENT :
        dispatchConfigEvent(*__i_stream);
        break;

      default: GLOBAL_ERROR_STREAM("Unknown Type" << endl); break;
    }

  } catch (const Exception& te) {
    GLOBAL_ERROR_STREAM( "TCPStrategy dispatch Message " << te << endl);
    __is_running = false;
    closeConnection(getStreamSocket().getSocket());
    return;
  }
}

void TCPStrategy::setValueStream(const IValueStream& i_stream, const OValueStream& o_stream) {
  delete __i_stream;
  delete __o_stream;

  __i_stream = i_stream.copy();
  __o_stream = o_stream.copy();
}

void TCPStrategy::setNetStream(const NetStream& stream) {
  delete __message_stream;
  
  __message_stream = stream.copy();
}

void TCPStrategy::setSendValueStream(const string& input_stream) {
  /* No other events can be send */
  MutexGuard event_guard(__event_sending_mutex);
  /* No other messages can be send */
  MutexGuard message_guard(__send_message_mutex);

  IValueStream* new_prototype = SValueFactory::getInstance().createIValueStream(input_stream);
  string partner_stream = SValueFactory::getInstance().getIValueStreamPartner(input_stream);

  if (!new_prototype || partner_stream.empty())
    throw Exception("Could not change to new ValueStream " + input_stream);

  Event e(ITransportStrategy::setReceiveValueStream_event, new TInValue<string>(partner_stream));

  if  ( sendValue( e, TCPStrategy::CONFIG_EVENT ) != FAILURE ) {
    delete __i_stream;
    __i_stream = new_prototype;
  } else {
    if (e.getException()) {
      throw *e.getException();
    } else {
      throw Exception("Could not change to new ValueStream in Partner Strategy");
    }
  }
}

void TCPStrategy::setReceiveValueStream(const string& out_stream) {
  /* No other events can be send */
  MutexGuard event_guard(__event_sending_mutex);
  /* No other messages can be send */
  MutexGuard message_guard(__send_message_mutex);

  OValueStream* new_prototype = SValueFactory::getInstance().createOValueStream(out_stream);
  if ( !new_prototype )
    throw Exception("Could not change reveiving OValueStream to " + out_stream);

  delete __o_stream;
  __o_stream = new_prototype;
}

void TCPStrategy::dispatchEventResponse(IValueStream& iv) {
  /* Read the result */
  try {
    iv >> static_cast<Result&>(__current_result);

    /* If the event is processed successfully read the value*/
    if (__current_result == SUCCESS) {
      __current_value->resetMemoryType();
      __current_value->disableMemoryType(Value::INVALUE);
      iv >> *__current_value;
      __current_value->resetMemoryType();
    } else {
      /* If the return value is FAILURE read the data if the stream is not empty */
      if (!iv.isEmpty()) {
        __current_value->resetMemoryType();
        __current_value->disableMemoryType(Value::INVALUE);
        iv >> *__current_value;
        __current_value->resetMemoryType();
      }
    }
  }
  /* Begin> Error Handling */
  catch (const ConnectionClosed& e) {
    __is_running = false;
    closeConnection(getStreamSocket().getSocket());
  }
  catch (const TransmissionException& e) {
    __is_running = false;
    closeConnection(getStreamSocket().getSocket());
  }
  catch (const Exception& e) {
    Event * tmp_event = dynamic_cast<Event*>(__current_value);
    cerr << "Exception during deserialization of event response (method dsipatchEventResponse())" << endl;
    if (tmp_event) {
      cerr << "Could not deserialize response for Event " << tmp_event->getKey() << endl;
    } else {
      cerr << "Could not deserialize response for other Value"  << endl;
    }
    GLOBAL_ERROR_STREAM(e << endl);
  }
  /* End< */

  /* Begin> Notify the condtion variable */
  __event_mutex.lock();
  __event_condition.notify();
  __event_mutex.unlock();
  /* End< */
}


void TCPStrategy::dispatchEventRequest(IValueStream& iv) {
 /* Begin> If no receiver available  the Event can not handled and return UNDEF*/
  if (!getEventReceiver())
    return sendResponse(UNDEF, 0, EVENT_RESPONSE);
  /* End< */

  /* Create the Event*/
  Event* event = SValueFactory::getInstance().createType<Event>(iv.getCurrentTypeInfo());

  /* If the ValueFactory can not create an instance of the object the event can not be
   *  handled. So return UNDEF
   */
  if (!event)
    return sendResponse(UNDEF, 0, EVENT_RESPONSE);

  try {
    /* Otherwise read the event from stream */
    event->createFromStream(iv);
  } catch (const Exception& e) {
    cerr << "Exception during deserialization of event request (method dsipatchEventRequest())" << endl;
    GLOBAL_ERROR_STREAM( "DispatchEventRequest for event " << event->getKey() << endl << e << endl);
  }

  /* send it to the receiver */
  Result result = getEventReceiver()->receiveEvent(*event, this);

  /* disable the in-values */
  event->disableMemoryType(Value::INVALUE);

  /* and send the response back to the sender */
  sendResponse(result, event, EVENT_RESPONSE);

  /* reset the Memory type befor destruction */
  event->resetMemoryType();

  /* delete the event */
  delete event;
  /*delete the out and inoutvalues and the return value*/
}


void TCPStrategy::dispatchConfigEvent(IValueStream& iv) {
  /* Create the Event*/
  Event* event = SValueFactory::getInstance().createType<Event>(iv.getCurrentTypeInfo());

  /* If the ValueFactory can not create an instance of the object the event can not be
   *  handled. So return UNDEF
   */
  if (!event)
    return sendResponse(UNDEF, 0, EVENT_RESPONSE);

  try {
    /* Otherwise read the event from stream */
    event->createFromStream(iv);
  } catch (const Exception& e) {
    cerr << "Exception during deserialization of a internal config event (method dispatchConfigEvent()): " << event->getKey() << endl;
    GLOBAL_ERROR_STREAM( "DispatchConfigEvent for event " << event->getKey() << endl << e << endl);
  }

  /* The event is dispatched in the own EventDispatcher */
  Result result = getEventDispatcher().dispatch(*event, this);

  /* and send the response back to the sender */
  sendResponse(result, event, EVENT_RESPONSE);

  /* delete the event */
  delete event;
}


void TCPStrategy::dispatchCEventRequest(IValueStream& iv) {
  /* Begin> If no receiver available  the Cevent can not handled and return UNDEF*/
  if (!getEventReceiver())
    return sendResponse(UNDEF, 0, EVENT_RESPONSE);
  /* End< */

  /* Create the CEvent*/
  CEvent* cevent = SValueFactory::getInstance().createType<CEvent>(iv.getCurrentTypeInfo());

  /* If the ValueFactory can not create an instance of the object the cevent can not be
   *  handled. So return UNDEF
   */
  if (!cevent) {
    cerr << "Could not create Object with type " << iv.getCurrentType() << endl;
    return sendResponse(UNDEF, 0, EVENT_RESPONSE);
  }


  try {
    /* Otherwise read the cevent from stream */
    cevent->createFromStream(iv);
  } catch (const Exception& e ) {
    cerr << "Exception during deserialization of an CEvent (method dispatchCEventRequest()): "  << endl;
    cerr << e << endl;
  }

  /* send it to the receiver */
  Result result = getEventReceiver()->receiveEvent(*cevent, this);

  /* and send the response back to the sender */
  sendResponse(result, cevent, EVENT_RESPONSE);

  /* delete the event */
  delete cevent;
  /*delete the out and inoutvalues and the return value*/
}



void TCPStrategy::waitForConnect() const {
}

NetHandler::NetResult TCPStrategy::closeConnection(int socket) {
  /* That the running flag to false */
  __condition_thread_mutex.lock();
  __is_running = false;
  __condition_thread_mutex.unlock();

  /* Notify the thread to leave the while loop */
  conditionNotify();
  return internalCloseConnection(socket);
}


NetHandler::NetResult TCPStrategy::internalCloseConnection(int) {
  /* Begin> Reset the Receiver if connection is closed */
  __socket_mutex.lock();
  if (getStreamSocket().getSocket() == -1) {
    __socket_mutex.unlock();
    return REMOVE;
  }

  try {
    getStreamSocket().close();
  } catch (const Exception& e) {
    GLOBAL_ERROR_STREAM( e << endl );
  }
  /* End< */
  __socket_mutex.unlock();

  /* Begin >Notify a waiting event sender */
  __event_mutex.lock();
  __current_result = FAILURE;
  __event_condition.notify();
  __event_mutex.unlock();

  /* End< */
  return REMOVE;
}


void TCPStrategy::connectTo()
  throw (const ConnectionException&)
{
  /* Begin> Use the connector to connect to the given address*/
  try {
    __connector.connect(getStreamSocket(), __current_address->getINetAddr());
    
    /* Begin> Set the connection flag to true and register the socket by the NetReactor */
    SNetReactor::getInstance().registerSocket(getStreamSocket().getSocket(), this);
    /* End< */

    /* Begin> Enable TCP NoDelay */
    enableNoDelay(true);
    /* End< */
    
    /* Begin> If the given address points to an NMMObject connect to it */
    if (__current_address->getNMMReference()) {
      Event e(IObjectAcceptor::connectToObject_event, new TInValue<NMMReference>(__current_address->getNMMReference()));
      if (sendEvent(e) == FAILURE)
        throw ConnectionException("Could not connect to object",errno, ConnectionException::connect);
    }
    /*End< */

  } catch (const ConnectionException& ce) {
    GLOBAL_ERROR_STREAM(ce << endl);
    throw ce;
  }
  /* End< */

}


void TCPStrategy::setAddress(const TSAddress& addr)
    throw (const BindingException&)
{
  const TCPAddress* address = dynamic_cast<const TCPAddress*>(&addr);
  if (!address) {
    throw BindingException("Invalid address type. Expect a TCPAddr");
  }

  delete __current_address;
  __current_address = address->copy();

  try {
    connectTo();
  } catch (const Exception& e) {
    throw BindingException(e.getComment());
  }
}

void TCPStrategy::dispatchInstreamMessage(IValueStream& iv, TCPMessage message_type) {
  Message* message = 0;
  try {
    /* Begin> Check which kind of message was received */
    if (message_type == INSTREAM_BUFFER) {
      /* Begin> If it is a Buffer we must use the BufferManager to allocate the Message object */
      unsigned int size;
      iv >> size;
      if (!getBufferManager()) {
        GLOBAL_ERROR_STREAM("No Buffer Manager set" << std::endl);
        return;
      }
      Buffer* tmp_buffer = getBufferManager()->getNewBuffer(size);
      if (tmp_buffer)
        tmp_buffer->setUsed(size);   //this should be done in the BufferManager!!

      message = tmp_buffer;
      /* End< */
    }

    if (message_type == INSTREAM_MESSAGE) {
      /* If its no Buffer, the ValueFactory allocates the object. Currently only CEvents are send */
      message = SValueFactory::getInstance().createType<Message>(iv.getCurrentTypeInfo());
    }
    /* End< */

    /* Begin> If the message object is successfully created, create it from the network and send it to our receiver */
    if (message) {
      try {
        message->createFromStream(iv);
      } catch (const Exception& e ) {
        if (message_type == INSTREAM_BUFFER) {
          cerr << "Exception during deserialization of instream buffer (method dispatchInstreamBuffer()): " << endl;
        } else {
          cerr << "Exception during deserialization of instream cevent (method dispatchInstreamBuffer()): " << endl;
        }
      }

      /* > Print some debug messages */
      if(message->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM(" Received Downstream Message " << endl);
      } else if(message->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM(" Received Upstream Message " << endl);
      } else {
        GLOBAL_WARNING_STREAM(" Received Message with UNKNOWN DIRECTION !!!!!!!" << endl);
      }

      if (getMessageReceiver()) {
        __blocked_mutex.lock();
        if (getMessageReceiver()->receiveMessage(message, this) == BLOCKED) {
          __blocked = true;
        }
        __blocked_mutex.unlock();
      } else {
        /* If no message receiver is set, release the buffer*/
        message->release();
      }
    } else {
      if (message_type == INSTREAM_BUFFER) {
        cerr << "Could not allocate Buffer" << endl;
      } else {
        cerr << "Could not allocate Event" << endl;        
      }
      throw Exception("Could not create message. No data from network read" );
    }
    /* End< */
  /* Begin> Error handling */
  } catch ( const Exception& e){
    GLOBAL_ERROR_STREAM ( e << endl);
    closeConnection(getStreamSocket().getSocket());
    return;
  } catch ( ... ) {
    GLOBAL_ERROR_STREAM ( "Unknown exception occured " << endl);    
  }
  /* End< */
}



void TCPStrategy::run() {
  while( __is_running ) {
   __blocked_mutex.lock();
   if ( (!__send_message && __is_running) ||
        (!__send_message && __is_running && __blocked) ||
        (!__blocked_message && __is_running) //HACK!!!!!!
      )
   {
     __blocked_mutex.unlock();
     /* If nothing to do wait for jobs */
     conditionWait();
   } else {
     __blocked_mutex.unlock();
   }

    /* Begin> HACK FOR INCORRECT BLOCKED HANDLING */
    if ( __blocked_message ) {
      __send_message_mutex.lock();
      cerr << "SEND BLOCKED MESSAGE " << endl;
      Message* tmp_message = __blocked_message;
      /* Begin> reset the __send_message_flag */
      __blocked_message = 0;
      /* Notify the Parent about FREE_QUEUE_SLOT*/
      if (getSignalReceiver()) {
        if(tmp_message->getDirection() == Message::DOWNSTREAM) {
          getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
        }
        else if(tmp_message->getDirection() == Message::UPSTREAM) {
        getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
      } else {
          GLOBAL_ERROR_STREAM("Unknown message direction" << endl);
        }
      } else {
        GLOBAL_ERROR_STREAM("TCPStrategy has no SignalReceiver" << endl);
      }
      __send_message_mutex.unlock();
      /* End< */

      /* Send message */
      internalMessageSend(tmp_message);
    } 
    /* END< HACK !!! */

    __send_message_mutex.lock();
    /*Begin> Handle a message that should be send */
    if ( __send_message ) {
      Message* tmp_message = __send_message;
      /* Begin> reset the __send_message_flag */
      __send_message = 0;
      /* Notify the Parent about FREE_QUEUE_SLOT*/
      if (getSignalReceiver()) {
        if(tmp_message->getDirection() == Message::DOWNSTREAM) {
          getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
        }
        else if(tmp_message->getDirection() == Message::UPSTREAM) {
        getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
      } else {
          GLOBAL_ERROR_STREAM("Unknown message direction" << endl);
        }
      } else {
        GLOBAL_ERROR_STREAM("TCPStrategy has no SignalReceiver" << endl);
      }
      __send_message_mutex.unlock();
      /* End< */

      /* Send message */
      internalMessageSend(tmp_message);
    } else {
      __send_message_mutex.unlock();
    }
    /* End< */
  }
}

void TCPStrategy::receiveSignal(Signal signal) {
  switch(signal) {
    case FREE_QUEUE_SLOT:
      __blocked_mutex.lock();
      __blocked = false;
      __blocked_condition.notify();
      __blocked_mutex.unlock();
      break;
    default:
      break;
  }
}

void TCPStrategy::internalMessageSend(Message* message) {
  if (!message) {
    return;
  }
  if (getStreamSocket().getSocket() == -1) {
    GLOBAL_WARNING_STREAM( "TCPStrategy is not connected " << endl );
    message->release();
    return;
  }

  /* Begin> If the buffer stores no data, release it */
  if (message->getType() == Message::BUFFER) {
    if (static_cast<Buffer*>(message)->getUsed() == 0) {
      GLOBAL_WARNING_STREAM( "Try to send Buffer without data " << endl);
      message->release();
      return;
    }
  }
  /* End< */

  /* use the __message_stream to send instream messages */
  __o_stream->setOPartnerStream(__message_stream);
  try {
    if (message->getType() == Message::BUFFER) {
      *__o_stream << static_cast<int>(INSTREAM_BUFFER);
      *__o_stream << static_cast<Buffer*>(message)->getUsed();
    } else {
      *__o_stream << static_cast<int>(INSTREAM_MESSAGE);
    }

    if(message->getDirection() == Message::DOWNSTREAM) {
      GLOBAL_DEBUG_STREAM( "Send DownStream Message " << endl);
    } else if(message->getDirection() == Message::UPSTREAM) {
      GLOBAL_DEBUG_STREAM( "Send UpStream Message " << endl );
    } else {
      GLOBAL_WARNING_STREAM( "Send Message with UNKNOWN DIRECTION !!!!!!!" << endl);
    }

    *__o_stream << *message;

    __o_stream->writeData();
//     GLOBAL_DEBUG_STREAM( "sent " << __message___o_stream->getWriteCounter() << " over TCP" << endl);
  }
  /* Begin> Error handling */
  catch (const Exception& te) {
    GLOBAL_WARNING_STREAM(te << endl);
    internalCloseConnection(getStreamSocket().getSocket());
    __is_running = false;
  }
  /* End< */
  /* Release the message */
  message->release();
}

void TCPStrategy::conditionWait() {
  __condition_thread_mutex.lock();
  if ( !__send_message && __is_running) {
    __thread_condition.wait();
  }
  __condition_thread_mutex.unlock();
}

void TCPStrategy::conditionNotify() {
  __condition_thread_mutex.lock();
  __thread_condition.notify();
  __condition_thread_mutex.unlock();
}

void TCPStrategy::scheduleDestruction() {
  int socket = getStreamSocket().getSocket();
  __is_running = false;
  disableNotification();
  SNetReactor::getInstance().unregisterSocket(socket);
  closeConnection(socket);
  delete this;
}


const char TCPStrategy::__strategy_type[] = "TCPStrategy";

TTSFactoryRegistrator< TTSFactoryObject<TCPStrategy> > __tcp_strategy_prototype("TCPStrategy");
TConstructor<TCPAddress>      __tcp_address(TypeInfo("TCPAddress", "NMM"));

}




