/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_TCPRECEIVER_HPP
#define NMM_TCPRECEIVER_HPP

#include "TCPGeneric.hpp"
#include "nmm/base/serialize/net/NetStream.hpp"
#include "nmm/utils/net/NetReactor.hpp"
#include "nmm/utils/net/SockConnector.hpp"
#include "nmm/interfaces/stats/IByteCounterImpl.hpp"
#include "nmm/base/TransportStrategy.hpp"

namespace NMM {
  class Node;
  class IValueStream;
  class OValueStream;


  /** This class implements an TCP address which stores all information to establish a connection.
   */
  class TCPAddress : public TSAddress {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    TCPAddress (const NMMReference& = 0);

    /** Constructor, with hostname and port of the address
     */
    TCPAddress (const string& name, int port, const NMMReference& = 0);
    
    /** Constructor, with INetAddr
     */
    TCPAddress (const INetAddr& net_addr, const NMMReference& = 0);
    
    /** Copy Constructor
     */
    TCPAddress (const TCPAddress&);

    /** Virtual copy Constructor
     */
    TCPAddress* copy()                          const { return new TCPAddress(*this); }
    //@}

    /** @name Access information
     */
    //@{
    /** This method sets the port
     */
    virtual void setPort(int port);

    /** This method sets the host name
     */
    virtual void setHostname(const string& host);

    /** This method sets the complete INetAddr
     */
    void setINetAddr(const INetAddr& net_addr)        { __net_addr = net_addr; }

    /** This method returns the INetAddr.
     */
    const INetAddr& getINetAddr()               const { return __net_addr; }

    /** This method returns the INetAddr.
     */
    INetAddr& getINetAddr()                     { return __net_addr; }
    //@}
  private:
    INetAddr __net_addr;                              ///< The network address
  };


  template<>
  class TConstructor<TCPAddress> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual TCPAddress*          createValue()          const {return new TCPAddress();}
    virtual TValue<TCPAddress>*  createTValue()         const {return new TValue<TCPAddress>();}

    virtual TValue<TSAddress*>*  createTPointerValue()  const {return new TValue<TSAddress*>();}

    virtual TConstructor<TCPAddress>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };


  class TCPStrategy
    : public TransportStrategy,
      public TCPGeneric,
      public ThreadedNetHandler,
      public IByteCounterImpl,
      protected ThreadedObject
  {
    enum TCPMessage { EVENT_REQUEST = 0, EVENT_RESPONSE , CEVENT_REQUEST, INSTREAM_BUFFER, INSTREAM_MESSAGE, CONFIG_EVENT};
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    TCPStrategy() ;

    /** Destructor
     */
    virtual ~TCPStrategy();

    /** virtual copy constructor
     */
    virtual TCPStrategy* copy() const { return new TCPStrategy(); }

    /** Destroy the TCPStrategy
     */
    virtual void scheduleDestruction();
    //@}

    /** @name Sender & Receiver Interface
     */
    //@{
    /** This method sends an incoming CEvent over a tcp connection
     */
    virtual Result receiveEvent(const CEvent&, const IEventSender* = 0 );

    /** This method sends an incoming Event over a tcp connection
     */
    virtual Result receiveEvent(const Event&, const IEventSender* = 0 );

    /** This method sends an incoming CEvent over a tcp connection
     */
    virtual Result receiveEvent(const CEvent&, const IEventSender* = 0 ) const;

    /** This method sends an incoming Event over a tcp connection
     */
    virtual Result receiveEvent(const Event&, const IEventSender* = 0 )  const;

    /** This method sends an incoming Message over a tcp connection
     */
    virtual Result receiveMessage(Message* , const IMessageSender* sender);

    /** This method sends an incoming Message over a tcp connection
     */
    virtual Result sendMessage(Message* );

    /** This method sends an incoming CEvent over a tcp connection
     */
    virtual Result sendEvent( const CEvent& ce);

    /** This method sends an incoming Event over a tcp connection
     */
    virtual Result sendEvent( const Event& e);

    /** This method sends an incoming CEvent over a tcp connection
     */
    virtual Result sendEvent( const CEvent& ce)   const;

    /** This method sends an incoming Event over a tcp connection
     */
    virtual Result sendEvent( const Event& e)     const;
    //@}


    /** @name Configure the Transport-Strategy
     */
    //@{
    /** This method sets the ValueStreams that should be used
     */
    virtual void setValueStream(const IValueStream& , const OValueStream&);

    /** This method sets the connection address
     */
    virtual void setAddress(const TSAddress& addr)
      throw (const BindingException&);

    /** This method processes internal NMM-Signals from Nodes or jacks
     */
    virtual void   receiveSignal(Signal signal);

    /** This method returns the typename of the Transport Strategy
     */
    virtual string getStrategyType() const { return getStaticStrategyType(); }
    
    /** This method sets the TCP net stream
     */
    virtual void setNetStream(const NetStream& stream);

    /** This method returns the net stream, used for instream and out of band messages
     */
    virtual NetStream& getNetStream()                     { return *__message_stream; }

    /** This method returns the net stream, used for instream and out of band messages
     */
    virtual const NetStream& getNetStream()         const { return *__message_stream; }
    //@}

    /** @name DispatchedNetHandler methods
     */
    //@{
    /** This method is called if the event POLLHUP occurs.
     * \param socket The socket on which an event occurs.
     */
    virtual NetResult errorOccured(int socket)     { return closeConnection(socket); }

    /** This method is called if the event POLLINVAL occurs.
     * \param socket The socket on which an event occurs.
     */
    virtual NetResult invalidSocket(int socket)      { return closeConnection(socket); }

    /** This method is called if the event POLLERR occurs.
     * \param socket The socket on which an event occurs.
     */
    virtual NetResult closeConnection(int socket);
  
    /** This method handles incoming data
     */
    virtual NetResult receiveData(int socket);
    //@}
    
    /** @name Static type information.
     */
    //@{
    /** This method returns the type name of strategy.
     */
    static inline string getStaticStrategyType() { return __strategy_type; }
    //@}
    
  protected:
    /** @name Internal dispatch methods
     */
    //@{
    /** This method dispatch incoming messgages to the other methods in this section
     */
    virtual void dispatchMessage();

    /** This method handles an incoming Event.
     */
    void dispatchEventRequest(IValueStream&);

    /** This method handles an Event response.
     */
    void dispatchEventResponse(IValueStream& );

    /** This method handles an incoming CEvent.
     */
    void dispatchCEventRequest(IValueStream&);

    /** This method handles an incoming Config Event.
     */
    void dispatchConfigEvent(IValueStream& );

    /** This method handles an incoming instream message.
     */
    void dispatchInstreamMessage(IValueStream& iv, TCPMessage);

    /** This method handles an incoming upstream message.
     */
    void dispatchUpstreamMessage(IValueStream& iv, TCPMessage);
    //@}

    /** @name Methods to send Data
     */
    //@{
    /** This method sends an response to an received Event or CEvent
     */
    virtual void sendResponse(Result, const Value* , TCPMessage);

    /** This method sends an Event or CEvent
     */
    virtual Result sendValue(const Value& e, TCPMessage)     const;

    /** This method sends the message in an own thread.
     */
    void internalMessageSend(Message* message);

    /** This method waits until the connection is established
     */
    void waitForConnect() const;

    /** This method establish a connection to given address.
     */
    virtual void connectTo()
      throw (const ConnectionException& );
    //@}

    /** @name Methods to send Data
     */
    //@{
    /** This thread handles incoming data and is used to send messages
     */
    virtual void run();

    /** Wait for the internal thread.
     */
    void conditionWait();

    /** Notify waiting threads.
     */
    void conditionNotify();
    //@}
    /** @name ValueStreams
     */
    //@{
    /** This method sets the ValueStream that should be used to send data
     */
    virtual void setSendValueStream(const string& );

    /** This method sets the ValueStream that should be used for incoming data
     */
    virtual void setReceiveValueStream(const string&);
    //@}

    /**
     * Get number of bytes send.
     * Returns the current value of the sending byte counter.
     */
    int getSendingCounter()        { return __message_stream->getWriteCounter(); }

    /**
     * Get number of bytes received.
     * Returns the current value of the receiving byte counter.
     */
    int getReceivingCounter()      { return __message_stream->getReadCounter(); }

    /**
     * Reset the sending byte counter.
     * Resets the sending counter to zero bytes processed.
     */
    void resetSendingCounter()     { __message_stream->resetWriteCounter(); }

    /**
     * Reset the receiving byte counter.
     * Resets the sending counter to zero bytes processed.
     */
    void resetReceivingCounter()   { __message_stream->resetReadCounter(); }
  private:
    NetResult internalCloseConnection(int );
  
  private:
    /* Begin> Attributes for send/receiveEvent */
    ThreadMutex                   __event_mutex;                ///< Mutex for __event_condition
    ThreadCondition               __event_condition;            ///< Condition to be notified about event response
    ThreadMutex                   __event_sending_mutex;        ///< This mutex ensures that only one event can be send
    mutable Value*                __current_value;              ///< Pointer to the current Value that is send
    Result                        __current_result;             ///< Result for the returning Value
    /* End< */

    /* Begin> Pointers to the used ValueStreams */
    OValueStream*                 __o_stream;                   ///< OValueStream which is used for serialization
    IValueStream*                 __i_stream;                   ///< IValueStream which is used for deserialization
    /* End< */
    
    /* Begin> Pointers to the used NetStreams */
    NetStream*                    __message_stream;             ///< NetStream used for instream and out of band transmissions
    /* End< */
    
    /* Begin> Thread specific attributes */
    bool                          __is_running;                 ///< Threads running flag
    ThreadMutex                   __condition_thread_mutex;     ///< Mutex for Threads condition __thread_condition
    ThreadCondition               __thread_condition;           ///< Thread condition to wait if its nothing to do
    /* End< */

    /* Begin> Message sending attributes */
    Message*                      __send_message;               ///< Temporary pointer to the message that should be send
    Message*                      __blocked_message;            ///< Hack to avoid problems if BLOCKED is ignored
    ThreadMutex                   __send_message_mutex;         ///< Mutex for save message sending
    /* End< */

    /* Begin> Connection specific attributes */
    SockConnector                 __connector;                  ///< Used SockConnector to establish a connection
    TCPAddress*                   __current_address;            ///< Pointer to the Current Address
    /* End< */

    /* Begin> Attributes for incoming messages from network */
    bool                          __blocked;                    ///< Flag that is true if the parent returns blocked
    ThreadMutex                   __incoming_message_mutex;     ///< Mutex for the incoming messages
    ThreadMutex                   __socket_mutex;
    ThreadMutex                   __blocked_mutex;
    ThreadCondition               __blocked_condition;
    /* End< */
    
  private:
    static const char __strategy_type[];                        ///< Static type name of this strategy
  };

}
#endif





