/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "UDPStrategy.hpp"

#include "nmm/base/serialize/net/PacketNetStream.hpp"
#include "nmm/base/serialize/net/UnbufferedNetStream.hpp"

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Event.hpp"
#include "nmm/base/NMMObject.hpp"

#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/memorymgr/IExternalBufferManager.hpp"

#include "nmm/base/serialize/Value.hpp"
#include "nmm/base/serialize/XMLValueStream.hpp"
#include "nmm/base/serialize/MagicNumberValueStream.hpp"
#include "nmm/base/serialize/ValueStream.hpp"

#include "nmm/base/ConnectionManager.hpp"
#include "nmm/base/ConnectorAcceptor.hpp"
#include "nmm/base/TSFactory.hpp"

#include "nmm/interfaces/net/IObjectAcceptor.hpp"

#include "nmm/base/ConnectorAddress.hpp"

#include "nmm/utils/net/SockAcceptor.hpp"
#include "nmm/utils/net/SockConnector.hpp"

#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

namespace NMM {


  /*============================== UDPAddress ==============================*/
  UDPAddress::UDPAddress (const NMMReference& reference)
    : ConnectorAddress(TypeInfo("UDPAddress", "NMM"), TSAddress::LOCAL|TSAddress::SYSTEM|TSAddress::NETWORK, UDPStrategy::getStaticStrategyType(), reference)
  {
    getINetAddr().setTransportProtocol("udp");
    insertArgument(&__net_addr);
  }


  UDPAddress::UDPAddress (const string& name, int port, const NMMReference& reference)
    : ConnectorAddress(TypeInfo("UDPAddress", "NMM"), TSAddress::LOCAL|TSAddress::SYSTEM|TSAddress::NETWORK, UDPStrategy::getStaticStrategyType(), reference)
  {
    insertArgument(&__net_addr);
    getINetAddr().setTransportProtocol("udp");
    setPort(port);
    setHostname(name);
  }


  UDPAddress::UDPAddress (const UDPAddress& addr)
    : ConnectorAddress(addr)
  {
    insertArgument(&__net_addr);
    *this = addr;
  }

  void UDPAddress::setPort(int port)  {
    __net_addr.setPort(port);
    TSAddress::setPort(port);
  }

  void UDPAddress::setHostname(const string& host)              {
    __net_addr.setAddress(host.c_str());
    TSAddress::setHostname(host);
  }

  /*============================== UDPAddress ==============================*/
  
  
  UDPStrategy::UDPStrategy()
    :
    /* Begin> Initialize all the Attributes with meaningful values */
    __event_condition(__event_mutex), __event_sending_mutex(ThreadMutex::RECURSIVE), __current_value(0), __current_result(FAILURE),
    __o_stream(0), __i_stream(0),
    __udp_stream(0), __use_packet_stream(true),
    __is_running(true), __thread_condition(__condition_thread_mutex),
    __send_message(0), __send_message_mutex(ThreadMutex::RECURSIVE),
    __local_address(0), __remote_address(0), __icm(0), __remote_ts_ref(0),
    __incoming_message(false), __blocked(false)
    /* End <*/
  {
    /* Set the default net stream */
    setNetStream(PacketNetStream(getDatagramSocket()));
  
    /* Set the default stream */
    MagicNumberIValueStream istream;
    MagicNumberOValueStream ostream;
    setValueStream(istream, ostream);

    /* activate the thread */
    activateThread();
  }

  UDPStrategy:: ~UDPStrategy() {
    if(getSignalReceiver() && __send_message) {
      if(__send_message->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM("getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT); in ~UDPStrategy" << endl);
        getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
      }
      else if(__send_message->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM("getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT); in ~UDPStrategy" << endl);
        getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
      } 
      else {
        GLOBAL_ERROR_STREAM("Unknown message direction in ~UDPStrategy" << endl);
      }
    }
    if(__send_message) {
      __send_message->release();
    }

    setSignalReceiver(0);
    setEventReceiver(0);
    setMessageReceiver(0);

    closeConnection(getDatagramSocket().getSocket());

    /* Set the running flag to false */
    __condition_thread_mutex.lock();
    __is_running = false;
    __condition_thread_mutex.unlock();
    
    /* Notify the thread to leave the while loop */
    conditionNotify();

    /* Deactivate the thread */
    deactivateThread();
  
    /* Destroy value streams */
    if (__i_stream) {
      delete __i_stream;
      __i_stream = 0;
    }
    if (__o_stream) {
      delete __o_stream;
      __o_stream = 0;
    }
  
    /* Destroy net streams */
    MutexGuard g(__net_stream_mutex);
    if (__udp_stream) {
      delete __udp_stream;
      __udp_stream = 0;
    }
  }


  /*============================== IEventSender ==============================*/
  Result UDPStrategy::sendEvent( const CEvent& cevent) {
    /* forward the Event to send value */
    return sendValue(cevent, CEVENT_REQUEST);
  }

  Result UDPStrategy::sendEvent( const CEvent& cevent) const {
    /* forward the Event to send value */
    return sendValue(cevent, CEVENT_REQUEST);
  }

  Result UDPStrategy::sendEvent(const Event& event) {
    /* forward the Event to send value */
    return sendValue(event, EVENT_REQUEST);
  }

  Result UDPStrategy::sendEvent(const Event& event) const{
    /* forward the Event to send value */
    return sendValue(event, EVENT_REQUEST);
  }
  /*============================== IEventSender ==============================*/
  
  
  /*============================== IEventReceiver ==============================*/
  Result UDPStrategy::receiveEvent(const CEvent& event, const IEventSender* sender) {
    /* If this UDPStrategy is the sender don't send the event again */
    if (sender == this) {
      return SUCCESS;
    }
    if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
      return NMMObject::receiveEvent(event, sender);
    }

    return sendValue(event, CEVENT_REQUEST);
  }

  Result UDPStrategy::receiveEvent(const Event& event, const IEventSender* sender) {
    /* If this UDPStrategy is the sender don't send the event again */
    if (sender == this) {
      return SUCCESS;
    }

    if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
      return NMMObject::receiveEvent(event, sender);
    }

    /* forward the Event to send value */
    return sendValue(event, EVENT_REQUEST);
  }

  Result UDPStrategy::receiveEvent(const CEvent& event, const IEventSender* sender ) const {
    /* If this UDPStrategy is the sender don't send the event again */
    if (sender == this) {
      return SUCCESS;
    }

    if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end())
      return NMMObject::receiveEvent(event, sender);

    /* forward the Event to send value */
    return   sendValue(event, CEVENT_REQUEST);
  }

  Result UDPStrategy::receiveEvent(const Event& event, const IEventSender* sender )  const {
    /* If this UDPStrategy is the sender don't send the event again */
    if (sender == this) {
      return SUCCESS;
    }


    if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end())
      {
        return NMMObject::receiveEvent(event, sender);
      }

    /* forward the Event to send value */
    return sendValue(event, EVENT_REQUEST);
  }
  /*============================== IEventReceiver ==============================*/
  
  
  Result UDPStrategy::receiveReliableEvent(Event& event, const IEventSender* sender )  const {
    /* If this UDPStrategy is the sender don't send the event again */
    if (sender == this) {
      return SUCCESS;
    }
    
    if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
      return NMMObject::receiveEvent(event, sender);
    }
    
    /* forward the Event to sendReliableEvent */
    return sendReliableEvent(event);
  }

  Result UDPStrategy::sendValue( const Value& value, UDPMessage message_type) const {
    /* If the thread is no longer running return FAILURE */
    if (!__is_running) {
      GLOBAL_ERROR_STREAM("UDP sendValue failed (UDPStrategy not running)" << endl);
      return FAILURE;
    }
    
    try {
      /* Lock the __event_mutex to ensure that no other event can be send. Use a MutexGuard to unlock
       * the mutex after method is returned and return value is stored by the sender.*/
      MutexGuard guard(__event_sending_mutex);
      
      /* Store a pointer to the given value. So the internal thread can wirte the response to it*/
      __current_value = const_cast<Value*>(&value);
      
      /* Begin> Serialize the event and send it */
      {
        /* Lock the net stream while we use it */
        MutexGuard g(__net_stream_mutex);
        if (!__udp_stream) {
          GLOBAL_ERROR_STREAM("UDP strategy has no NetStream - sendValue failed!");
          throw Exception("No NetStream");
        }
        
        /* Create a copy of the value stream and set the net stream as its partner */
        auto_ptr<OValueStream> xov(__o_stream->copy());
        xov->setOPartnerStream(const_cast<NetStream*>(__udp_stream));
        
        /* Before writing, lock the condition to ensure that the dispatchResponse method can not
         * access the event until we have called wait*/
        MutexGuard event_guard(__event_mutex);
        
        /*write the message type to the stream. The receiver needs this information to know
         * which receiveEvent-method must be called; the receiveEvent(const Event&) or the
         * receiveEvent(const CEvent&)
         */
        *xov << static_cast<int>(message_type);
        
        /* Disable the out_values */
        __current_value->disableMemoryType(Value::OUTVALUE);
        
        /* write the value to the stream */
        *xov << value;
        
        /* ensure that all data are written (similar to a 'flush')*/
        xov->writeData();
      }
      
      /* wait for the response ...*/
      __event_condition.wait();
      /* reset the internal pointer ...*/
      __current_value = 0;

      /* return the result */
      return __current_result;
    }
    /* End< */
    
    /* Begin> Error handling */
    catch (const TransmissionException& te) {
      GLOBAL_ERROR_STREAM(te << endl);
      const_cast<UDPStrategy*>(this)->closeConnection(getDatagramSocket().getSocket());
      throw BindingException(te.getComment());
    } catch(const ConnectionClosed& ce){
      GLOBAL_ERROR_STREAM(ce << endl);
      const_cast<UDPStrategy*>(this)->closeConnection(getDatagramSocket().getSocket());
      throw BindingException(ce.getComment());
    } catch(const Exception& e){
      GLOBAL_ERROR_STREAM(e << endl);
      const_cast<UDPStrategy*>(this)->closeConnection(getDatagramSocket().getSocket());
      throw BindingException(e.getComment());
    }
    /* End< */
  }
  
  Result UDPStrategy::sendReliableEvent(Event& event) const {
    /* If we have no IConnectionManager, return FAILURE */
    if (!__icm) {
      GLOBAL_ERROR_STREAM("UDP sendReliableEvent failed (No IConnectionManager)" << endl);
      return FAILURE;
    }
    
    /* If we have no partner reference, return FAILURE */
    if (!__remote_ts_ref) {
      GLOBAL_ERROR_STREAM("UDP sendReliableEvent failed (No partner reference)" << endl);
      return FAILURE;
    }
    
    try {
      return __icm->sendReliableEvent(__remote_ts_ref, event);
    } catch(BindingException& e) {
      GLOBAL_ERROR_STREAM("UDP sendReliableEvent failed (" << e << ")" << endl);
      return FAILURE;
    }
  }
  
  Result UDPStrategy::sendReliableConfigEvent(Event& event) const {
    /* If we have no IConnectionManager, return FAILURE */
    if (!__icm) {
      GLOBAL_ERROR_STREAM("UDP sendReliableConfigEvent failed (No IConnectionManager)" << endl);
      return FAILURE;
    }
    
    /* If we have no partner reference, return FAILURE */
    if (!__remote_ts_ref) {
      GLOBAL_ERROR_STREAM("UDP sendReliableConfigEvent failed (No partner reference)" << endl);
      return FAILURE;
    }
    
    try {
      return __icm->sendReliableConfigEvent(__remote_ts_ref, event);
    } catch(BindingException& e) {
      GLOBAL_ERROR_STREAM("UDP sendReliableConfigEvent failed (" << e << ")" << endl);
      return FAILURE;
    }
  }
  
  
  void UDPStrategy::sendResponse(Result result, const Value* ce, UDPMessage message_type) {
    try {
      /* Begin> Create and serialize the response */
      {
        /* lock the net stream while we use it */
        MutexGuard g(__net_stream_mutex);
        if (!__udp_stream) {
          GLOBAL_ERROR_STREAM("UDP strategy has no NetStream - sendResponse failed!");
          throw Exception("No NetStream");
        }
        
        /* create a copy of the value stream and set the net stream as its partenr */
        auto_ptr<OValueStream> ov(__o_stream->copy());
        ov->setOPartnerStream(__udp_stream);
        
        /* write the message_type */
        (*ov) << static_cast<int>(message_type);
        
        /* write the result */
        (*ov) << static_cast<const Result&>(result);
        
        /* If given value != 0, write the value*/
        if (ce) {
          (*ov) << *ce;
        }
        
        /* call writeData to finish the transmission */
        ov->writeData();
      }
      
      /* and return */
      return;
    }
    /* End< */
    
    /* Begin> Error handling */
    catch (const TransmissionException& te) {
      GLOBAL_ERROR_STREAM(te << endl);
    } catch(const ConnectionClosed& ce){
      GLOBAL_ERROR_STREAM(ce << endl);
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM(ce << endl);
    }
    closeConnection(getDatagramSocket().getSocket());
    /* End */
  }
  
  Result UDPStrategy::sendMessage(Message* message) {
    return receiveMessage(message, 0);
  }

  Result UDPStrategy::receiveMessage(Message* message, const IMessageSender* ) {
    if (!message) {
      return FAILURE;
    }

    if (__send_message) {
      GLOBAL_ERROR_STREAM( "ARGH ERROR IN UDPSTARTEGY __send_message != 0 !!!!!!!!!!!!!" << endl);
      return BLOCKED;
    }

    /* Begin> Store the message pointer */
    __condition_thread_mutex.lock();
    __send_message_mutex.lock();
    __send_message = message;
    __send_message_mutex.unlock();
    __condition_thread_mutex.unlock();
    /* End< */

    /* Notify the Thread */
    conditionNotify();

    /* Return blocked*/
    return BLOCKED;
  }

  NetHandler::NetResult UDPStrategy::receiveData(int /*socket*/) {
    if (__incoming_message) {
      return VOID; //old message was not processed
    }
    
    __incoming_message_mutex.lock();
    __incoming_message = true;
    __incoming_message_mutex.unlock();
    
    /* Notify the Thread about changes */
    conditionNotify();

    /* and return */
    return REMOVE;
  }
  
  NetHandler::NetResult UDPStrategy::writeData(int) {
    return VOID;
  }
  
  void UDPStrategy::dispatchMessage(NetStream* stream) {
    try {
      /* Create a copy of the value stream and set the net stream as its partner */
      auto_ptr<IValueStream> iv(__i_stream->copy());
      iv->setIPartnerStream(stream);
      
      /* Read data from the value stream */
      iv->readData();
      int tmp_message;
      
      *iv >> tmp_message;
      
      UDPMessage message_type = static_cast<UDPMessage>(tmp_message);
      switch(message_type) {
      case EVENT_RESPONSE :
        dispatchEventResponse(*iv);
        break;

      case EVENT_REQUEST  :
        dispatchEventRequest(*iv);
        break;

      case CEVENT_REQUEST  :
        dispatchCEventRequest(*iv);
        break;

      case INSTREAM_BUFFER :
        dispatchInstreamMessage(*iv, message_type);
        break;

      case INSTREAM_MESSAGE :
        dispatchInstreamMessage(*iv, message_type);
        break;

      case CONFIG_EVENT :
        dispatchConfigEvent(*iv);
        break;

      default:
        GLOBAL_ERROR_STREAM("Unknown Type" << endl);
        break;
      }
      
    } catch (const Exception& te) {
      GLOBAL_ERROR_STREAM( "UDPStrategy dispatch Message " << te << endl);
      __is_running = false;
      closeConnection(getDatagramSocket().getSocket());
    }
  }

  void UDPStrategy::setValueStream(const IValueStream& i_stream, const OValueStream& o_stream) {
    delete __i_stream;
    delete __o_stream;

    __i_stream = i_stream.copy();
    __o_stream = o_stream.copy();
  }

  void UDPStrategy::setNetStream(const NetStream& udp_stream) {
    delete __udp_stream;
    __udp_stream = udp_stream.copy();
  }

  void UDPStrategy::setSendValueStream(const string& input_stream) {
    /* No other events can be send */
    MutexGuard event_guard(__event_sending_mutex);
    /* No other messages can be send */
    MutexGuard message_guard(__send_message_mutex);

    IValueStream* new_prototype = SValueFactory::getInstance().createIValueStream(input_stream);
    string partner_stream = SValueFactory::getInstance().getIValueStreamPartner(input_stream);

    if (!new_prototype || partner_stream.empty())
      throw Exception("Could not change to new ValueStream " + input_stream);

    Event e(ITransportStrategy::setReceiveValueStream_event, new TInValue<string>(partner_stream));

    if  ( sendReliableConfigEvent(e) != FAILURE ) {
      delete __i_stream;
      __i_stream = new_prototype;
    } else {
      if (e.getException()) {
        throw *e.getException();
      } else {
        throw Exception("Could not change to new ValueStream in Partner Strategy");
      }
    }
  }

  void UDPStrategy::setReceiveValueStream(const string& out_stream) {
    /* No other events can be send */
    MutexGuard event_guard(__event_sending_mutex);
    /* No other messages can be send */
    MutexGuard message_guard(__send_message_mutex);

    OValueStream* new_prototype = SValueFactory::getInstance().createOValueStream(out_stream);
    if ( !new_prototype )
      throw Exception("Could not change reveiving OValueStream to " + out_stream);

    delete __o_stream;
    __o_stream = new_prototype;
  }


  void UDPStrategy::dispatchEventResponse(IValueStream& iv) {
    /* Read the result */
    try {
      iv >> static_cast<Result&>(__current_result);

      /* If the event is processed successfully read the value*/
      if (__current_result == SUCCESS) {
        __current_value->resetMemoryType();
        __current_value->disableMemoryType(Value::INVALUE);
        iv >> *__current_value;
        __current_value->resetMemoryType();
      } else {
        /* If the return value is FAILURE read the data if the stream is not empty */
        if (!iv.isEmpty()) {
          __current_value->resetMemoryType();
          __current_value->disableMemoryType(Value::INVALUE);
          iv >> *__current_value;
          __current_value->resetMemoryType();
        }
      }
    }
    /* Begin> Error Handling */
    catch (const ConnectionClosed& e) {
      __is_running = false;
      GLOBAL_ERROR_STREAM(e << endl);
      closeConnection(getDatagramSocket().getSocket());
    }
    catch (const TransmissionException& e) {
      __is_running = false;
      GLOBAL_ERROR_STREAM(e << endl);
      closeConnection(getDatagramSocket().getSocket());
    }
    catch (const Exception& e) {
      GLOBAL_ERROR_STREAM(e << endl);
    }
    /* End< */
    
    /* Begin> Notify the condtion variable */
    __event_mutex.lock();
    __event_condition.notify();
    __event_mutex.unlock();
    /* End< */
  }


  void UDPStrategy::dispatchEventRequest(IValueStream& iv) {
    /* Begin> If no receiver available  the Event can not handled and return UNDEF*/
    if (!getEventReceiver())
      return sendResponse(UNDEF, 0, EVENT_RESPONSE);
    /* End< */

    /* Create the Event*/
    Event* event = SValueFactory::getInstance().createType<Event>(iv.getCurrentTypeInfo());
    
    /* If the ValueFactory can not create an instance of the object the event can not be
     *  handled. So return UNDEF
     */
    if (!event)
      return sendResponse(UNDEF, 0, EVENT_RESPONSE);

    /* Otherwise read the event from stream */
    event->createFromStream(iv);

    /* send it to the receiver */
    Result result = getEventReceiver()->receiveEvent(*event, this);

    /* disable the in-values */
    event->disableMemoryType(Value::INVALUE);

    /* and send the response back to the sender */
    sendResponse(result, event, EVENT_RESPONSE);

    /* reset the Memory type befor destruction */
    event->resetMemoryType();

    /* delete the event */
    delete event;
    /*delete the out and inoutvalues and the return value*/
  }


  void UDPStrategy::dispatchConfigEvent(IValueStream& iv) {
    /* Create the Event*/
    Event* event = SValueFactory::getInstance().createType<Event>(iv.getCurrentTypeInfo());

    /* If the ValueFactory can not create an instance of the object the event can not be
     *  handled. So return UNDEF
     */
    if (!event)
      return sendResponse(UNDEF, 0, EVENT_RESPONSE);

    /* Otherwise read the event from stream */
    event->createFromStream(iv);

    /* The event is dispatched in the own EventDispatcher */
    Result result = getEventDispatcher().dispatch(*event, this);

    /* and send the response back to the sender */
    sendResponse(result, event, EVENT_RESPONSE);

    /* delete the event */
    delete event;
  }


  void UDPStrategy::dispatchCEventRequest(IValueStream& iv) {
    /* Begin> If no receiver available  the Cevent can not handled and return UNDEF*/
    if (!getEventReceiver())
      return sendResponse(UNDEF, 0, EVENT_RESPONSE);
    /* End< */

    /* Create the CEvent*/
    CEvent* cevent = SValueFactory::getInstance().createType<CEvent>(iv.getCurrentTypeInfo());

    /* If the ValueFactory can not create an instance of the object the cevent can not be
     *  handled. So return UNDEF
     */
    if (!cevent)
      return sendResponse(UNDEF, 0, EVENT_RESPONSE);
  
    /* Otherwise read the cevent from stream */
    cevent->createFromStream(iv);

    /* send it to the receiver */
    Result result = getEventReceiver()->receiveEvent(*cevent, this);

    /* and send the response back to the sender */
    sendResponse(result, cevent, EVENT_RESPONSE);

    /* delete the event */
    delete cevent;
    /*delete the out and inoutvalues and the return value*/
  }



  void UDPStrategy::waitForConnect() const {
  }

  NetHandler::NetResult UDPStrategy::closeConnection(int /*socket*/) {
    /* Begin> Reset the Receiver if connection is closed */
    __socket_mutex.lock();
    if (getDatagramSocket().getSocket() != -1) {
      try {
        getDatagramSocket().close();
      } catch (const Exception& e) {
        GLOBAL_ERROR_STREAM( e << endl );
      }
    }
    __socket_mutex.unlock();
    /* End< */
    
    /* Begin >Notify a waiting event sender */
    __event_mutex.lock();
    __current_result = FAILURE;
    __event_condition.notify();
    __event_mutex.unlock();
    /* End< */
    
    return REMOVE;
  }


  void UDPStrategy::createLocalSocket()
    throw (const ConnectionException&)
  {
    try {
      
      /* Get local host name */
      char buffer[256];
      ::gethostname(buffer, 256);
      string hostname(buffer);
      
      /* Set up local address and find a free port*/
      INetAddr udp_addr;
      udp_addr.setTransportProtocol("udp");
      udp_addr.setPort(__udp_min_port);
      udp_addr.setAddress(hostname.c_str());
      SockAcceptor udp_acceptor;
      udp_acceptor.tryOpen(udp_addr, __udp_min_port, __udp_max_port, "udp");
      udp_acceptor.accept(getDatagramSocket());
      
      INetAddr temp_addr(getDatagramSocket().getLocalAddress());
      GLOBAL_DEBUG_STREAM2("<<< UDP *LOCAL* ADDRESS IS " << temp_addr << " >>>" << endl);
      
      /* Register socket at NetReactor */
      SNetReactor::getInstance().registerSocket(getDatagramSocket().getSocket(), this);
      __udp_socket_events = NetReactor::NR_POLLIN | NetReactor::NR_POLLERR;
      
    } catch (const ConnectionException& ce) {
      
      GLOBAL_ERROR_STREAM(ce << endl);
      throw ce;
      
    }
  }
  
  void UDPStrategy::connectToPeer()
    throw (const ConnectionException&)
  {
    try {
      
      /* Set up peer address and connect */
      INetAddr udp_addr;
      udp_addr = __remote_address->getINetAddr();
      udp_addr.setTransportProtocol("udp");
      SockConnector udp_connector;
      udp_connector.connect(getDatagramSocket(), udp_addr, "udp");
      
      GLOBAL_DEBUG_STREAM2("<<< UDP *PEER* ADDRESS IS " << udp_addr << " >>>" << endl);
      
    } catch (const ConnectionException& ce) {
      
      GLOBAL_ERROR_STREAM(ce << endl);
      throw ce;
      
    }
  }
  
  void UDPStrategy::setAddress(const TSAddress& addr)
    throw (const BindingException&)
  {
    GLOBAL_DEBUG_STREAM2("UDPStrategy: Set address type=" << addr.getTypeName() << std::endl);
    
    /* create a local socket */
    try {
      createLocalSocket();
    } catch (const Exception& e) {
      throw BindingException(e.getComment());
    }
    
    /* set up the local address */
    // XXX: because we don't know the local object, we can't set the NMM reference properly!
    GLOBAL_DEBUG_STREAM("UDPStrategy: Setup local address" << std::endl);
    delete __local_address;
    __local_address = new UDPAddress(0);
    __local_address->setINetAddr(getDatagramSocket().getLocalAddress());    
    __local_address->setConnectionManager(*SConnectionManager::getInstance().getInterface<IConnectionManager>());
    __local_address->setTSReference(getReference());
    
    /* check if we have a UDPAddress */
    const UDPAddress* udp_address = dynamic_cast<const UDPAddress*>(&addr);
    if(udp_address) {
      /* store this address as remote address */
      delete __remote_address;
      __remote_address = udp_address->copy();
      
      /* connect socket to peer address */
      try {
        connectToPeer();
      } catch (const Exception& e) {
        throw BindingException(e.getComment());
      }
      
      return;
    }
    
    /* check if we have a ConnnectorAddress */
    const ConnectorAddress* connector_address = dynamic_cast<const ConnectorAddress*>(&addr);
    if(connector_address) {      
      /* store the IConnectionManager interface */
      if (!connector_address->getConnectionManager()) {
        /* SHOULD NOT HAPPEN */
        GLOBAL_ERROR_STREAM2("UDPStrategy: ConnectorAddress has no IConnectionManager interface" << endl);
        throw BindingException("Shared address has no IConnectionManager interface");
      }
      delete __icm;
      __icm = dynamic_cast<IConnectionManager*>(connector_address->getConnectionManager()->copy());
      
      /* request a connection */
      ConnectorAddress* remote_addr = NULL;
      SConnectionManager::getInstance().connect(__local_address, remote_addr, __icm);
      
      /* the remote address must be a UDPAddress */
      if(!remote_addr) {
        /* SHOULD NOT HAPPEN */
        GLOBAL_ERROR_STREAM2("UDPStrategy: Could not get remote address" << endl);
        throw BindingException("Could not get remote address");
      }
      const UDPAddress* udp_addr = dynamic_cast<UDPAddress*>(remote_addr);
      if (!udp_addr) {
        /* SHOULD NOT HAPPEN */
        GLOBAL_ERROR_STREAM2("UDPStrategy: Remote address is not a UDPAddress" << endl);
        throw BindingException("Remote address is not a UDPAddress");
      }
      
      /* store the remote address */
      __remote_address = udp_addr->copy();
      
      /* check if the returned NMM reference is valid */
      if (!__remote_address->getTSReference()) {
        GLOBAL_WARNING_STREAM("UDPStrategy: Could not get NMM reference of remote transport strategy" << endl);
      }
      __remote_ts_ref = __remote_address->getTSReference();
      
      /* connect socket to peer address */
      try {
        connectToPeer();
      } catch (const Exception& e) {
        throw BindingException(e.getComment());
      }
      
      /* if the address points to an NMM object, connect to it */
      if (connector_address->getNMMReference()) {
        Event e(IObjectAcceptor::connectToObject_event, new TInValue<NMMReference>(connector_address->getNMMReference()));
        if (sendReliableEvent(e) == FAILURE) {
          throw BindingException("Could not connect to object");
        }
      }
      
      return;
    }
    
    throw BindingException(getStrategyType() + " can't handle address of type " + addr.getTypeName());
  }

  void UDPStrategy::dispatchInstreamMessage(IValueStream& iv, UDPMessage message_type) {
    Message* message = 0;
    try {
      /* Begin> Check which kind of message was received */
      if (message_type == INSTREAM_BUFFER) {
        /* Begin> If it is a Buffer we must use the BufferManager to allocate the Message object */
        unsigned int size;
        iv >> size;
        if (!getBufferManager()) {
          GLOBAL_ERROR_STREAM("No Buffer Manager set");
          return;
        }
        Buffer* tmp_buffer = getBufferManager()->getNewBuffer(size);
        if (tmp_buffer)
          tmp_buffer->setUsed(size);   //this should be done in the BufferManager!!
      
        message = tmp_buffer;
        /* End< */
      }
      
      if (message_type == INSTREAM_MESSAGE) {
        /* If its no Buffer, the ValueFactory allocates the object. Currently only CEvents are send */
        message = SValueFactory::getInstance().createType<Message>(iv.getCurrentTypeInfo());
      }
      /* End< */

      /* Begin> If the message object is successfully created, create it from the network and send it to our receiver */
      if (message) {
        message->createFromStream(iv);
        if (getMessageReceiver()) {
          __internal_thread_mutex.lock();
          if (getMessageReceiver()->receiveMessage(message, this) == BLOCKED) {
            __blocked = true;
          }
          __internal_thread_mutex.unlock();
        }
      } else {
        //throw something
      }
      /* End< */
      /* Begin> Error handling */
    } catch (...){
      GLOBAL_ERROR_STREAM("Something strange went wrong in distapchInstreamMessage" << endl);
      closeConnection(getDatagramSocket().getSocket());
      return;
    }
    /* End< */
  }



  void UDPStrategy::run() {
    while( __is_running ) {

      if ( !__send_message && !__incoming_message && __is_running ) {
        /* If nothing to do wait for jobs */
        conditionWait();
      }

      if (!__is_running) {
        closeConnection(getDatagramSocket().getSocket());
        return;
      }

      /* Begin> Handle an incoming message */
      __incoming_message_mutex.lock();
      if ( __incoming_message ) {
        /* Begin> reset the __incoming_message flag */
        __incoming_message_mutex.unlock();
        /* End< */
        
        if (__incoming_message) {
          
          /* Receive a packet on the UDP stream and check if it is a complete message */
          MutexGuard g(__net_stream_mutex);
          if (!__udp_stream) {
            GLOBAL_ERROR_STREAM("Data received on UDP socket, but no UDP stream has been set!" << endl);
          } else {
            if (__udp_stream->receivePacket()) {
              dispatchMessage(__udp_stream);
            }
          }
        }
        
        __incoming_message_mutex.lock();
        __incoming_message = false;
        __incoming_message_mutex.unlock();
        
        /* Register the socket again, becaus we return REMOVE now */
        if ( !__send_message && !__incoming_message && __is_running && !__blocked) {
          SNetReactor::getInstance().registerSocket(getDatagramSocket().getSocket(), this);
        }
        
        continue; // FIXME: attention: this continue will block all traffic in other direction (e.g. upstream events) ! maybe it should be removed?
      } else {
        __incoming_message_mutex.unlock();
      }
      /* End< */

      /*Begin> Handle a message that should be send */
      __send_message_mutex.lock();
      if ( __send_message ) {
        Message* tmp_message = __send_message;
        /* Begin> reset the __send_message_flag */
        __send_message = 0;
        /* Notify the Parent about FREE_QUEUE_SLOT*/
        if (getSignalReceiver()) {
          if(tmp_message->getDirection() == Message::DOWNSTREAM) {
            getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
          }
          else if(tmp_message->getDirection() == Message::UPSTREAM) {
            getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
          } else {
            GLOBAL_ERROR_STREAM("Unknown message direction" << endl);
          }
        } else {
          GLOBAL_ERROR_STREAM("UDPStrategy has no SignalReceiver" << endl);
        }
        __send_message_mutex.unlock();
        /* End< */

        /* Send message */
        internalMessageSend(tmp_message);
        
        /* If nothing to do, register the socket again, because we returned REMOVE */
        if ( !__send_message && !__incoming_message && __is_running && !__blocked) {
          SNetReactor::getInstance().registerSocket(getDatagramSocket().getSocket(), this);
        }
      
        continue;
      } else {
        __send_message_mutex.unlock();
      }
      /* End< */
    }
  }
  
  void UDPStrategy::receiveSignal(Signal signal) {
    switch(signal) {
    case FREE_QUEUE_SLOT:
      __internal_thread_mutex.lock();
      __blocked = false;
      if ( !__send_message && !__incoming_message && __is_running && !__blocked) {
        SNetReactor::getInstance().registerSocket(getDatagramSocket().getSocket(), this);
      }
      __internal_thread_mutex.unlock();
      break;
    default:
      break;
    }
  }

  void UDPStrategy::internalMessageSend(Message* message) {
    if (!message) {
      return;
    }
    if (getDatagramSocket().getSocket() == -1) {
      GLOBAL_WARNING_STREAM( "UDPStrategy is not connected " << endl );
      message->release();
    }
    
    /* Begin> Serialize and send the message */
    try {
      /* Lock the net stream while we use it */
      MutexGuard g(__net_stream_mutex);
      if (!__udp_stream) {
        GLOBAL_ERROR_STREAM("No NetStream -- internalMessageSend failed" << endl);
        throw Exception("No NetStream");
      }
      
      /* Create a copy of the value stream and set the net stream as its partner */
      auto_ptr<OValueStream> stream(__o_stream->copy());
      stream->setOPartnerStream(__udp_stream);
      if (message->getType() == Message::BUFFER) {
        *stream << static_cast<int>(INSTREAM_BUFFER);
        *stream << static_cast<Buffer*>(message)->getUsed();
      } else {
        *stream << static_cast<int>(INSTREAM_MESSAGE);
      }
      
      *stream << *message;
      
      stream->writeData();
    }
    /* End< */
    
    /* Begin> Error handling */
    catch (const Exception& te) {
      GLOBAL_WARNING_STREAM(te << endl);
      closeConnection(getDatagramSocket().getSocket());
      __is_running = false;
    }
    /* End< */
    
    /* Release the message */
    message->release();
  }

  void UDPStrategy::conditionWait() {
    __condition_thread_mutex.lock();
    if ( !__send_message && !__incoming_message && __is_running) {
      __thread_condition.wait();
    }
    __condition_thread_mutex.unlock();
  }

  void UDPStrategy::conditionNotify() {
    __condition_thread_mutex.lock();
    __thread_condition.notify();
    __condition_thread_mutex.unlock();
  }

  void UDPStrategy::scheduleDestruction() {
    closeConnection(getDatagramSocket().getSocket());
    SNetReactor::getInstance().unregisterSocket(getDatagramSocket().getSocket());
    delete this;
  }
  
  /*============================== IUDPControl Interface ==============================*/
  void UDPStrategy::setUsePacketStream(const bool& use_packet_stream) {
    MutexGuard g(__net_stream_mutex);
    
    /* Check if we are already using the right stream */
    if (use_packet_stream == __use_packet_stream && __udp_stream != NULL) {
      return;
    }
    
    /* Create and use the new stream */
    __use_packet_stream = use_packet_stream;
    if (use_packet_stream) {
      /* Use the packet stream */
      setNetStream(PacketNetStream(getDatagramSocket()));
    } else {
      /* Use the net stream */
      setNetStream(UnbufferedNetStream(getDatagramSocket()));
    }
  }
  
  bool UDPStrategy::getUsePacketStream() const {
    MutexGuard g(__net_stream_mutex);
    return __use_packet_stream;
  }
  
  void UDPStrategy::setPacketSize(const ssize_t& size) {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return;
    }
    dynamic_cast<PacketNetStream*>(__udp_stream)->setPacketSize(size);
  }
  
  ssize_t UDPStrategy::getPacketSize() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketNetStream*>(__udp_stream)->getPacketSize();
  }
  
  void UDPStrategy::setMaxPartialMessages(const unsigned int& max_partial) {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return;
    }
    dynamic_cast<PacketNetStream*>(__udp_stream)->setMaxPartialMessages(max_partial);
  }
  
  unsigned int UDPStrategy::getMaxPartialMessages() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketNetStream*>(__udp_stream)->getMaxPartialMessages();
  }
  
  void UDPStrategy::resetStatistics() {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return;
    }
    dynamic_cast<PacketNetStream*>(__udp_stream)->resetStatistics();
  }
  
  ssize_t UDPStrategy::getTotalSentBytes() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      /* This is the only information we can get from a net stream */
      return __udp_stream->getWriteCounter();
    }
    return dynamic_cast<const PacketNetStream*>(__udp_stream)->getTotalSentBytes();
  }
  
  ssize_t UDPStrategy::getTotalReceivedBytes() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketNetStream*>(__udp_stream)->getTotalReceivedBytes();
  }
  
  unsigned int UDPStrategy::getTotalSentPackets() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketNetStream*>(__udp_stream)->getTotalSentPackets();
  }
  
  unsigned int UDPStrategy::getTotalReceivedPackets() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketNetStream*>(__udp_stream)->getTotalReceivedPackets();
  }
  
  unsigned int UDPStrategy::getTotalDiscardedPackets() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketNetStream*>(__udp_stream)->getTotalDiscardedPackets();
  }
  
  unsigned int UDPStrategy::getTotalLostPackets() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketNetStream*>(__udp_stream)->getTotalLostPackets();
  }
  
  /*============================== IUDPControl Interface ==============================*/
  

  const char UDPStrategy::__strategy_type[] = "UDPStrategy";

  int UDPStrategy::__udp_min_port = 20000;
  int UDPStrategy::__udp_max_port = 40000;
  
  TTSFactoryRegistrator< TConnectorTSFactoryObject<UDPStrategy, ConnectorAcceptor> >
  __udp_strategy_prototype("UDPStrategy", false, false);
  TConstructor<UDPAddress>      __udp_address(TypeInfo("UDPAddress", "NMM"));
}




