/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iostream>
#include "GraphDescription.hpp"
#include "nmm/base/connect.hpp"
#include "nmm/interfaces/base/registry/ISession.hpp"
#include "nmm/base/CommunicationChannel.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/interfaces/base/sync/IMultiAudioVideoSynchronizer.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/base/IQueued.hpp"

#include "nmm/base/InstreamChannel.hpp"

#include <cassert>

namespace NMM {

GraphDescription::GraphDescription()
  : SerializedValue(GraphDescription_type_name, GraphDescription_type_ns), __internal_counter(0), __i_session(0), __sink_synch_descr()
{
  __interface_list.useTValue(false);
  __node_list.useTValue(false);
  registerAttributes();
}


GraphDescription::~GraphDescription()
{
  delete __i_session.getValue();

  for(SequenceValue< NodeDescription >::iterator i = __node_list.begin(); i != __node_list.end(); ++i) {
    delete (*i);
  }
  for(SequenceValue< INode >::iterator j = __interface_list.begin(); j != __interface_list.end(); ++j) {
    delete (*j);
  }
}

bool GraphDescription::isIdenticalOverlap(const GraphDescription& gd) const
{
  if(getNumEdges() == gd.getNumEdges()) {
    size_t num_matching_edges = 0;
    for(size_t i = 0; i < getNumEdges(); ++i) {
      const GraphEdge& edge = getEdge(i);
      for(size_t j = 0; j < gd.getNumEdges(); ++j) {
	const GraphEdge& other = gd.getEdge(j);
	if(edge.isIdenticalOverlap(other)) {
	  ++num_matching_edges;
	  break;
	}
      }
    }
    if(num_matching_edges == getNumEdges()) {
      return true;
    }
    else {
      return false;
    }
  }
  else {
    return false;
  }
}

void GraphDescription::clear()
{
  /* Begin> Delete and reset the session */
  delete __i_session.getValue();
  __i_session.setValue(0);
  /* End< */

  /* Clear the object description of the sink synchronizer */
  __sink_synch_descr = ObjectDescription();

  /* Begin> Delete the NodeDescriptions and clear the vector */
  for(SequenceValue< NodeDescription >::iterator i = __node_list.begin(); i != __node_list.end(); ++i) {
    delete (*i);
  }
  __node_list.clear();
  /*End< */

  /* Begin> Delete the INodes and clear the vector */
  for(SequenceValue< INode >::iterator j = __interface_list.begin(); j != __interface_list.end(); ++j) {
    delete (*j);
  }
  __interface_list.clear();
  /* End< */
  
  /* Begin> Clear the remaing container */  
  __empty_list.clear();
  __edges.getRValue().clear();
  __out_edges.getRValue().clear();
  __in_edges.getRValue().clear();
  __request_flags.getRValue().clear();
  __release_flags.getRValue().clear();
  __internal_counter = 0;
  /* End< */
}

GraphDescription::GraphDescription(const GraphDescription& nd)
  :  SerializedValue(nd)
{
  __interface_list.useTValue(false);
  __node_list.useTValue(false);
  registerAttributes();
  *this = nd;
}


void GraphDescription::registerAttributes()
{
  insertArgument(&__edges);
  insertArgument(&__out_edges);
  insertArgument(&__in_edges);
  insertArgument(&__node_list);
  insertArgument(&__interface_list);
  insertArgument(&__request_flags);
  insertArgument(&__release_flags);
  insertArgument(&__i_session);
  insertArgument(&__sink_synch_descr);
}

void GraphDescription::setINode(const NodeDescription& nd, const INode& inode) 
{
  if (nd.getInternalID() < 0) {
    GLOBAL_ERROR_STREAM2("gd: "<< "Illegal NodeDescription" << endl);
    return;
  }

  int current_size = __interface_list.size();

  if (static_cast<unsigned int>(nd.getInternalID()) >= __interface_list.size()) {
    __interface_list.resize(nd.getInternalID()+1);

    /* Initialize the new fields */
    for (unsigned int i = current_size; i < __interface_list.size(); ++i) {
      __interface_list[i] = 0;
    }
  }

  if (__interface_list[nd.getInternalID()]) return;
  __interface_list[nd.getInternalID()] = dynamic_cast<INode*>(inode.copy());
}

const INode* GraphDescription::getINode(const NodeDescription& nd) const
{
  if (nd.getInternalID() < 0 )
    return 0;

  if (static_cast<unsigned int>(nd.getInternalID()) < __interface_list.size())
    return __interface_list[nd.getInternalID()];

  return 0;
}

INode* GraphDescription::getINode(const NodeDescription& nd) 
{
  if (nd.getInternalID() < 0 ) {
    return 0;
  }

  if (static_cast<unsigned int>(nd.getInternalID()) < __interface_list.size()) {
    return __interface_list[nd.getInternalID()];
  }

  return 0;
}

INode* GraphDescription::getINode(const NodeDescription* nd)
{
  return getINode(*nd);
}
 
const INode* GraphDescription::getINode(const NodeDescription* nd) const
{
  return getINode(*nd);
}

const list<int>& GraphDescription::getOutEdgeList(int i) const 
{ 
  CInternalIt iter = __out_edges.getRValue().find(i);
  if(iter == __out_edges.getRValue().end()) {
    return __empty_list;
  }
  else {
    return iter->second;  
  }
}

const list<int>& GraphDescription::getInEdgeList(int i) const 
{ 
  CInternalIt iter = __in_edges.getRValue().find(i);
  if(iter == __in_edges.getRValue().end()) {
    return __empty_list;
  }
  else {
    return iter->second;  
  }
}

const NodeDescription& GraphDescription::getFirstNodeDescription() const 
{ 
  if( (__node_list.size() > 0) && __node_list[0]) {
    return *(__node_list[0]); 
  }
  else {
    GLOBAL_ERROR_STREAM2("gd: "<< "GraphDescription::getFirstNodeDescription() called with size " << __node_list.size() << endl);
    throw Exception(" GraphDescription::getFirstNodeDescription() > Graph description is empty or 1. element is NULL");
  }
}

const NodeDescription& GraphDescription::getNodeDescription(unsigned int i) const 
{ 
  if(__node_list.size() > i) {
    return *(__node_list[i]); 
  }
  else {
    GLOBAL_ERROR_STREAM2("gd: "<< "GraphDescription::getNodeDescription( " << i << " ) const called with size " << __node_list.size() << endl);
    exit(-1);
  }
}

NodeDescription& GraphDescription::getNodeDescription(unsigned int i)
{
  if(__node_list.size() > i) {
    return *(__node_list[i]); 
  }
  else {
    GLOBAL_ERROR_STREAM2("gd: "<< "GraphDescription::getNodeDescription( " << i << " ) called with size " << __node_list.size() << endl);
    exit(-1);
  }
}

void GraphDescription::addNode(NodeDescription& nd) {
 // node descr has no id yet -> add to end 
  if(nd.getInternalID() < 0) {
    GLOBAL_DEBUG_STREAM2("gd: "<< "Adding new NodeDescription " << nd.getNodeName() << " with new ID " << __internal_counter << endl);

    nd.setInternalID(__internal_counter);
    NodeDescription* new_nd = new NodeDescription(nd);

    /* resize the node list */
    __node_list.resize(nd.getInternalID() + 1);

    /* resize the request flag */
    __request_flags.getRValue().resize(nd.getInternalID() + 1);
    /* resize the release flag */
    __release_flags.getRValue().resize(nd.getInternalID() + 1);

    __node_list[nd.getInternalID()] = new_nd;
    __request_flags.getRValue()[nd.getInternalID()] = true;
    __release_flags.getRValue()[nd.getInternalID()] = true;

    ++__internal_counter;
  }
  // node descr already has id >= size -> increase size and add node descr
  else if(static_cast<size_t>(nd.getInternalID()) >= __node_list.size()) { 
    size_t current_size = __node_list.size();

    /* resize the node list */
    __node_list.resize(nd.getInternalID() + 1);

    /* resize the request flag */
    __request_flags.getRValue().resize(nd.getInternalID() + 1);
    __release_flags.getRValue().resize(nd.getInternalID() + 1);

    /* Initialize the node list and request flags */
    for(size_t i = current_size; i < __node_list.size(); ++i) {
      __node_list[i] = 0;
      __request_flags.getRValue()[i] = true;
      __release_flags.getRValue()[i] = true;
    }

    NodeDescription* new_nd = new NodeDescription(nd);
    __node_list[nd.getInternalID()] = new_nd;
    __internal_counter = __node_list.size();
    GLOBAL_DEBUG_STREAM2("gd: "<< "Adding new NodeDescription " << nd.getNodeName() << " with preset ID " << nd.getInternalID() << " >= size." << endl);
  }
  // node descr already has id < size -> add node desrc at position of id if no other node descr exists
  else { 
    if(__node_list[nd.getInternalID()] == 0) {
      NodeDescription* new_nd = new NodeDescription(nd);
      __node_list[nd.getInternalID()] = new_nd;
      GLOBAL_DEBUG_STREAM2("gd: "<< "Adding new NodeDescription " << nd.getNodeName() << " with preset ID " << nd.getInternalID() << " < size." << endl);
    }
    else {
      GLOBAL_DEBUG_STREAM2("A NodeDescription " << __node_list[nd.getInternalID()]->getNodeName() << " with ID " << nd.getInternalID() << " already exists for " << nd.getNodeName() << endl);
    }
  }
}

size_t GraphDescription::addEdge(const NodeDescription& nd_from, 
			       const NodeDescription& nd_to  , const string& in_tag,
			       const Format* format, const string& buffer_strategy, const string& event_strategy)
{
  return addEdge(nd_from, "default", nd_to, in_tag, format, buffer_strategy, event_strategy);
}

size_t GraphDescription::addEdge(const NodeDescription& nd_from, const string& out_tag, 
			       const NodeDescription& nd_to  ,  
			       const Format* format, const string& buffer_strategy, const string& event_strategy)
{
  return addEdge(nd_from, out_tag, nd_to, "default", format, buffer_strategy, event_strategy);
}

size_t GraphDescription::addEdge(const NodeDescription& nd_from,  
			       const NodeDescription& nd_to  ,  
			       const Format* format, const string& buffer_strategy, const string& event_strategy)
{
  return addEdge(nd_from, "default", nd_to, "default", format, buffer_strategy, event_strategy);
}

size_t GraphDescription::addEdge(const NodeDescription& nd_from, const string& out_tag, 
			       const NodeDescription& nd_to  , const string& in_tag,
			       const Format* format, const string& buffer_strategy, const string& event_strategy)
{
  addNode(const_cast<NodeDescription& >(nd_from));
  addNode(const_cast<NodeDescription& >(nd_to));

  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(GraphEdge(nd_from.getInternalID(), out_tag, nd_to.getInternalID(), in_tag, format, current_id, buffer_strategy, event_strategy));
  __out_edges.getRValue()[nd_from.getInternalID()].push_back(__edges.getRValue().size() - 1);
  __in_edges.getRValue() [nd_to.getInternalID()]  .push_back(__edges.getRValue().size() - 1);
  
  GLOBAL_DEBUG_STREAM2("gd: "<< "addEdge " << __edges.getRValue().size() - 1 << " " << nd_from.getInternalID() << "::" << out_tag << " -> " << nd_to.getInternalID() << "::" << in_tag<< ":: " << buffer_strategy << endl);
  
  return ( __edges.getRValue().size() - 1 );
}

size_t GraphDescription::addEdgeFromIDs(int nd_from, const string& out_tag, 
					int nd_to  , const string& in_tag,
					const Format* format, const string& buffer_strategy, const string& event_strategy)
{
  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(GraphEdge(nd_from, out_tag, nd_to, in_tag, format, current_id, buffer_strategy, event_strategy));
  __out_edges.getRValue()[nd_from].push_back(__edges.getRValue().size() - 1);
  __in_edges.getRValue() [nd_to]  .push_back(__edges.getRValue().size() - 1);
  
  GLOBAL_DEBUG_STREAM2("gd: "<< "addEdge " << __edges.getRValue().size() - 1 << " " << nd_from << "::" << out_tag << " -> " << nd_to << "::" << in_tag<< ":: " << buffer_strategy << endl);
  
  return ( __edges.getRValue().size() - 1 );
}

size_t GraphDescription::addInEdge(const NodeDescription& nd_to, const string& in_tag, 
				 const Format* format)
{
  addNode(const_cast<NodeDescription& >(nd_to));

  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(GraphEdge(GraphEdge::UNCONNECTED_IN_ID, "default", nd_to.getInternalID(), in_tag, format, current_id));
  __in_edges.getRValue() [nd_to.getInternalID()].push_back(__edges.getRValue().size() - 1);
		    
  GLOBAL_DEBUG_STREAM2("gd: "<< "addInEdge " << __edges.getRValue().size() - 1 << " UNCONNECTED_IN_ID::default -> " << nd_to.getInternalID() << "::" << in_tag<< endl);

  return ( __edges.getRValue().size() - 1 );
}

size_t GraphDescription::addInEdge(const NodeDescription& nd_to,  
				 const Format* format)
{
  return addInEdge(nd_to, "default", format);
}

size_t GraphDescription::addOutEdge(const NodeDescription& nd_from, const string& out_tag, 
				  const Format* format)
{
  addNode(const_cast<NodeDescription& >(nd_from));

  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(GraphEdge(nd_from.getInternalID(), out_tag, GraphEdge::UNCONNECTED_OUT_ID, "default", format, current_id));
  __out_edges.getRValue()[nd_from.getInternalID()].push_back(__edges.getRValue().size() - 1);

  GLOBAL_DEBUG_STREAM2("gd: "<< "addEdge " << __edges.getRValue().size() - 1 << " " << nd_from.getInternalID() << "::" << out_tag << " -> UNCONNECTED_OUT_ID::default" << endl);

  return ( __edges.getRValue().size() - 1 );
}

size_t GraphDescription::addOutEdge(const NodeDescription& nd_from,  
				  const Format* format)
{
  return addOutEdge(nd_from, "default", format);
}

size_t GraphDescription::addOutHelpEdge(const NodeDescription& nd_from, const string& out_tag, 
					const Format* format)
{
  addNode(const_cast<NodeDescription& >(nd_from));

  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(GraphEdge(nd_from.getInternalID(), out_tag, GraphEdge::HELP_EDGE_ID, "default", format, current_id));
  __out_edges.getRValue()[nd_from.getInternalID()].push_back(__edges.getRValue().size() - 1);

  GLOBAL_DEBUG_STREAM2("gd: "<< "addEdge " << __edges.getRValue().size() - 1 << " " << nd_from.getInternalID() << "::" << out_tag << " -> UNCONNECTED_OUT_ID::default" << endl);

  return ( __edges.getRValue().size() - 1 );
}

void GraphDescription::setOutStrategyConfig(const NodeDescription& nd_from, const string& out_tag,
					    const NodeDescription& nd_to  , const string&  in_tag,
					    const CEvent& out_config)
{
  for(size_t i = 0; i < __edges.getRValue().size(); ++i) {
    const GraphEdge& e = __edges.getRValue()[i];
    if( (e.out_id.getValue()  == nd_from.getInternalID()) && 
	(e.in_id.getValue() == nd_to.getInternalID()) && 
	(e.out_tag.getValue() == out_tag) && 
	(e.in_tag.getValue() == in_tag) ) 
    {
      (const_cast<GraphEdge& >(e)).out_strategy_config = out_config;
    }
  }
}

void GraphDescription::setOutStrategyConfig(const NodeDescription& nd_from, 
					    const NodeDescription& nd_to  , const string&  in_tag,
					    const CEvent& out_config)
{
  setOutStrategyConfig(nd_from, "default", nd_to, in_tag, out_config);
}

void GraphDescription::setOutStrategyConfig(const NodeDescription& nd_from, const string& out_tag,
					    const NodeDescription& nd_to  , 
					    const CEvent& out_config)
{
  setOutStrategyConfig(nd_from, out_tag, nd_to, "default", out_config);
}

void GraphDescription::setOutStrategyConfig(const NodeDescription& nd_from, 
					    const NodeDescription& nd_to  ,
					    const CEvent& out_config)
{
  setOutStrategyConfig(nd_from, "default", nd_to, "default", out_config);
}

void GraphDescription::setInStrategyConfig(const NodeDescription& nd_from, const string& out_tag,
					   const NodeDescription& nd_to  , const string&  in_tag,
					   const CEvent& in_config)
{
  for(size_t i = 0; i < __edges.getRValue().size(); ++i) {
    const GraphEdge& e = __edges.getRValue()[i];
    if( (e.out_id.getValue()  == nd_from.getInternalID()) && 
	(e.in_id.getValue() == nd_to.getInternalID()) && 
	(e.out_tag.getValue() == out_tag) && 
	(e.in_tag.getValue() == in_tag) ) 
    {
      (const_cast<GraphEdge& >(e)).in_strategy_config = in_config;
    }
  }
}
  
void GraphDescription::setInStrategyConfig(const NodeDescription& nd_from, 
					   const NodeDescription& nd_to  , const string&  in_tag,
					   const CEvent& in_config)
{
  setInStrategyConfig(nd_from, "default", nd_to, in_tag, in_config);
}

void GraphDescription::setInStrategyConfig(const NodeDescription& nd_from, const string& out_tag,
					   const NodeDescription& nd_to  , 
					   const CEvent& in_config)
{
  setInStrategyConfig(nd_from, out_tag, nd_to, "default", in_config);
}

void GraphDescription::setInStrategyConfig(const NodeDescription& nd_from, 
					   const NodeDescription& nd_to  ,
					   const CEvent& in_config)
{
  setInStrategyConfig(nd_from, "default", nd_to, "default", in_config);
}

void GraphDescription::showAll()
{
  for(size_t i = 0; i < __edges.getRValue().size(); ++i) {
    const GraphEdge& e = __edges.getRValue()[i];
    GLOBAL_DEBUG_STREAM2("gd: "<< "edge " << i << " " << e.out_id.getRValue() << "::" << e.out_tag.getRValue() << " -> " << e.in_id.getRValue() << "::" << e.in_tag.getRValue() << endl);
  }

  for(size_t j = 0; j < __internal_counter; ++j) {
    GLOBAL_DEBUG_STREAM2("gd: "<< "node " << j << endl);
    GLOBAL_DEBUG_STREAM2("gd: "<< "  out" << endl);
    const list<int>& out_l = __out_edges.getRValue()[j];
    for(list<int>::const_iterator iter = out_l.begin(); iter != out_l.end(); ++iter) {
      GLOBAL_DEBUG_STREAM2("gd: "<< "    " << *iter << endl);
    }
    GLOBAL_DEBUG_STREAM2("gd: "<< "  in" << endl);
    const list<int>& in_l = __in_edges.getRValue()[j];
    for(list<int>::const_iterator iter = in_l.begin(); iter != in_l.end(); ++iter) {
      GLOBAL_DEBUG_STREAM2("gd: "<< "    " << *iter << endl);
    }
  }
}

void GraphDescription::showEdges()
{
  for(size_t i = 0; i < __edges.getRValue().size(); ++i) {
    const GraphEdge& e = __edges.getRValue()[i];
    GLOBAL_DEBUG_STREAM2("gd: "<< "edge " << i << " " << e.out_id.getRValue() << "::" << e.out_tag.getRValue() << " -> " << e.in_id.getRValue() << "::" << e.in_tag.getRValue() << endl);
  }

}


void GraphDescription::removeEdge(const int edge_num)
{
  // get edge to remove
  const GraphEdge& e = __edges.getRValue()[edge_num];
  //GLOBAL_DEBUG_STREAM2("gd: "<< "removeEdge " << edge_num << " " << e.out_id.getRValue() << "::" << e.out_tag.getRValue() << " -> " << e.in_id.getRValue() << "::" << e.in_tag.getRValue() << endl);

  // get id of node description for outgoing edge
  int out_num = e.out_id.getRValue();
  // check if out_num represents a connected id
  if(out_num != GraphEdge::UNCONNECTED_OUT_ID) {
    // try to get list of outgoing edges for node descr with id out_num
    InternalIt out_iter = __out_edges.getRValue().find(out_num);
    // delete edge
    if(out_iter != __out_edges.getRValue().end()) {
      (*out_iter).second.remove(edge_num);
    }
    else {
      GLOBAL_ERROR_STREAM2("gd: "<< "ERROR out_iter == __out_edges.getRValue().end()" << endl);
    }
  }
  
  // get id of node description for ingoing edge
  int in_num  = e.in_id.getRValue();
  // check if in_num represents a connected id
  if(in_num != GraphEdge::UNCONNECTED_IN_ID) {
    // try to get list of ingoing edges for node descr with id in_num
    InternalIt in_iter = __in_edges.getRValue().find(in_num);
    // delete edge
    if(in_iter != __in_edges.getRValue().end()) {
      (*in_iter).second.remove(edge_num);
    }
    else {
      GLOBAL_ERROR_STREAM2("gd: "<< "ERROR in_iter == __in_edges.getRValue().end()" << endl);
    }
  }

  // erase edge (do not, otherwise indices also have to be updated!)
  // __edges.getRValue().erase( ((__edges.getRValue()).begin() + edge_num) );
}

void GraphDescription::removeSubgraph(const int edge_num)
{
  // get edge to remove
  const GraphEdge& e = __edges.getRValue()[edge_num];

  // get id of node description for ingoing edge
  int in_num  = e.in_id.getRValue();
  // check if in_num represents a connected id
  if(in_num != GraphEdge::UNCONNECTED_IN_ID) {
    // try to get list of ingoing edges for node descr with id in_num
    InternalIt in_iter = __in_edges.getRValue().find(in_num);
    // delete all incoming edges 
    if(in_iter != __in_edges.getRValue().end()) {
      list<int> l = (*in_iter).second;
      for(list<int>::const_iterator i = l.begin(); i != l.end(); ++i) {
	removeEdge(*i);
	GLOBAL_DEBUG_STREAM2("gd: "<< "edge_num " << edge_num << " removeEdge " << *i << endl);
      }
    }
    else {
      GLOBAL_ERROR_STREAM2("gd: "<< "ERROR in_iter == __in_edges.getRValue().end()" << endl);
    }

    // call removeSubgraph for all outgoing edges
    // try to get list of outgoing edges for node descr with id in_num
    InternalIt out_iter = __out_edges.getRValue().find(in_num);
    if(out_iter != __out_edges.getRValue().end()) {
      list<int> l = (*out_iter).second;
      for(list<int>::const_iterator i = l.begin(); i != l.end(); ++i) {
        removeSubgraph(*i);
      }
    }
    else {
      // GLOBAL_ERROR_STREAM2("gd: "<< "No more outgoing edges" << endl);
    }
  }
}

void GraphDescription::addEdges(const NodeDescription* n1, const NodeDescription* n2, const NodeDescription* n3, 
				const NodeDescription* n4, const NodeDescription* n5, const NodeDescription* n6,
				const NodeDescription* n7, const NodeDescription* n8, const NodeDescription* n9,
				const NodeDescription* n10,const NodeDescription* n11,const NodeDescription* n12)
{
  if(n1 && n2) addEdge(*n1, *n2);
  if(n2 && n3) addEdge(*n2, *n3);
  if(n3 && n4) addEdge(*n3, *n4);
  if(n4 && n5) addEdge(*n4, *n5);
  if(n5 && n6) addEdge(*n5, *n6);
  if(n6 && n7) addEdge(*n6, *n7);
  if(n7 && n8) addEdge(*n7, *n8);
  if(n8 && n9) addEdge(*n8, *n9);
  if(n9 && n10) addEdge(*n9, *n10);
  if(n10 && n11) addEdge(*n10, *n11);
  if(n11 && n12) addEdge(*n11, *n12);
}

GraphEdge::GraphEdge()
  : SerializedValue(GraphEdge_type_name, GraphEdge_type_ns)
{
  registerAttributes();
  this->out_id.setValue(UNCONNECTED_OUT_ID);
  this->out_tag.setValue("");
  this->in_id.setValue(UNCONNECTED_IN_ID);
  this->in_tag.setValue("");
  this->format.setValue(0);
  this->overlap_type.setValue(NONE);
  this->q_out_id.setValue(UNCONNECTED_OUT_ID);
  this->q_in_id.setValue(UNCONNECTED_IN_ID);
  this->id.setValue(UNDEF_EDGE_ID);
  this->q_id.setValue(UNDEF_EDGE_ID);
  this->r_id.setValue(UNDEF_EDGE_ID);
  instream_channel.setValue(0);
}

GraphEdge::GraphEdge(int out_id, const string& out_tag, int in_id, const string& in_tag, const Format* format, int id,
                     const string& _c_buffer_strategy, const string& _c_event_strategy)
  : SerializedValue(GraphEdge_type_name, GraphEdge_type_ns)
{
  registerAttributes();
  this->out_id.setValue(out_id);
  this->out_tag.setValue(out_tag);
  this->in_id.setValue(in_id);
  this->in_tag.setValue(in_tag);
  if (format) {
    this->format.setValue(new TValue<Format>(*format));
  } else {
    this->format.setValue(0);
  }
  this->overlap_type.setValue(NONE);
  this->q_out_id.setValue(UNCONNECTED_OUT_ID);
  this->q_in_id.setValue(UNCONNECTED_IN_ID);
  this->id.setValue(id);
  this->q_id.setValue(UNDEF_EDGE_ID);
  this->r_id.setValue(UNDEF_EDGE_ID);
  c_buffer_strategy.setValue(_c_buffer_strategy);
  c_event_strategy.setValue(_c_event_strategy);
  instream_channel.setValue(0);
}

GraphEdge::GraphEdge(const GraphEdge& edge)
  : SerializedValue(edge), instream_channel(0)
{
  registerAttributes();
  *this = edge;
}

GraphEdge::~GraphEdge() {
  delete format.getValue();
  delete instream_channel.getValue();
}

void GraphEdge::setFormat(const Format* fom) {

  delete format.getValue();
  if ( fom ) {
    this->format.setValue(new TValue<Format>(*fom));
  } else {
    this->format.setValue( 0 );
  }
}

GraphEdge& GraphEdge::operator=(const GraphEdge& edge) {
  out_id    = edge.out_id;
  out_tag  = edge.out_tag;
  in_id    = edge.in_id;
  in_tag  = edge.in_tag;

  /* delete the old format */
  delete format.getValue ();

  if (edge.format.getValue() != 0) {
    format.setValue(new TValue<Format>(*edge.format.getValue()));
  }
  else {
    format.setValue(0);
  }

  overlap_type = edge.overlap_type;
  q_out_id     = edge.q_out_id;
  q_in_id      = edge.q_in_id;

  id = edge.id;

  q_id = edge.q_id;
  r_id = edge.r_id;

  c_buffer_strategy = edge.c_buffer_strategy;
  c_event_strategy = edge.c_event_strategy;

  out_strategy_config = edge.out_strategy_config;
  in_strategy_config  = edge.in_strategy_config;

  if ( edge.instream_channel.getValue() )
    instream_channel.setValue( edge.instream_channel.getValue()->copy() );
  else
    instream_channel.setValue(0);

  return *this;
}

bool GraphEdge::isIdenticalOverlap(const GraphEdge& edge) const
{
  if(q_out_id.getValue() != edge.q_out_id.getValue()) {
    return false;
  }
  if(q_in_id.getValue() != edge.q_in_id.getValue()) {
    return false;
  }
  if(q_id.getValue() != edge.q_id.getValue()) {
    return false;
  }
  if(r_id.getValue() != edge.r_id.getValue()) {
    return false;
  }
  return true;
}

void GraphEdge::registerAttributes() {
  insertArgument(&out_id);
  insertArgument(&out_tag);
  insertArgument(&in_id);
  insertArgument(&in_tag);
  insertArgument(&format);
  insertArgument(&overlap_type);
  insertArgument(&q_out_id);
  insertArgument(&q_in_id);
  insertArgument(&id);
  insertArgument(&q_id);
  insertArgument(&r_id);
  insertArgument(&c_buffer_strategy);
  insertArgument(&c_event_strategy);
  insertArgument(&out_strategy_config);
  insertArgument(&in_strategy_config);
  insertArgument(&instream_channel);
}

void GraphDescription::setISession(const ISession& session) {
  delete __i_session.getValue();
  __i_session.setValue(dynamic_cast<ISession*>(session.copy()));
}

void GraphDescription::setSinkSynchronizerDescription(const ObjectDescription& od)
{
  __sink_synch_descr.setValue(od);
}

void GraphDescription::realizeCommunicationChannel(CommunicationChannel* c_channel, GraphEdge& edge)
{
  if(c_channel == 0) {
    GLOBAL_ERROR_STREAM2("gd: CommunicationChannel is 0" << endl);
  }

  /* if no specific strategy is set -> get first one out of list */
  if(edge.c_buffer_strategy.getRValue().empty()) {
    edge.c_buffer_strategy.setValue(c_channel->getStrategyList().front());
  }

  /* if no specific secondary strategy is set -> use first strategy */
  if(edge.c_event_strategy.getRValue().empty()) {
    c_channel->setTransportStrategy(edge.c_buffer_strategy.getRValue());
  } else {
    InstreamChannel* instream_channel = dynamic_cast<InstreamChannel*>(c_channel);
    assert(instream_channel);
    instream_channel->setTransportStrategy(edge.c_buffer_strategy.getRValue(), edge.c_event_strategy.getRValue());
  }

  /* configure out strategy */
  if(edge.out_strategy_config.size() != 0) {
    list<Interface* >& partner_list = c_channel->getPartnerInterfaceList();
    if(partner_list.size() > 0) {
      partner_list.front()->sendEvent(edge.out_strategy_config);
    }
  }
  
  /* configure in strategy */
  if(edge.in_strategy_config.size() != 0) {
    list<Interface* >& own_list = c_channel->getInterfaceList();
    if(own_list.size() > 0) {
      own_list.front()->sendEvent(edge.in_strategy_config);
    }
  }
}

void GraphDescription::realizeGraph() throw(const BindingException&, const Exception&)
{
  /* Begin> Send all events for state CONSTRUCTED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != NodeDescription::SHARED ) {
      if (__node_list[i]->getNodeEvent(INode::CONSTRUCTED)) {
        __interface_list[i]->reachConstructed();
        __interface_list[i]->sendEvent(*__node_list[i]->getNodeEvent(INode::CONSTRUCTED));
      }
    }
  }
  /* End< */

  /* Begin> Send all events for state INITIALIZED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != NodeDescription::SHARED ) {
      if (__node_list[i]->getNodeEvent(INode::INITIALIZED)) {
        __interface_list[i]->reachInitialized();
        __interface_list[i]->sendEvent(*__node_list[i]->getNodeEvent(INode::INITIALIZED));
      }
    }
  }
  /* End< */

  GraphDescription gd(*this);

  list<int> graph_sources;

  // get all sources of query graph
  const SequenceValue< NodeDescription >& nds = gd.getNodeDescriptionList();
  for(size_t n = 0; n < nds.size(); ++n) {
    // node with no incoming edge and some outgoing edges -> source that needs to be connected
    if(gd.getInEdgeList(n).empty() && !gd.getOutEdgeList(n).empty()) {
      graph_sources.push_back(n);
    }
    // node with only unconnected incoming edges and some outgoing edges -> node that needs to be connected
    if(!gd.getOutEdgeList(n).empty()) {
      list<int>::const_iterator in_edge_iter;
      for(in_edge_iter = gd.getInEdgeList(n).begin(); in_edge_iter != gd.getInEdgeList(n).end(); ++in_edge_iter) {
	GraphEdge& in_edge = gd.getEdge(*in_edge_iter);
	if(in_edge.isOutConnected()) { // -> outgoind part is connected -> connected edge -> no 'source' node
	  break;
	}
      }
      if(in_edge_iter == gd.getInEdgeList(n).end()) {
	graph_sources.push_back(n);
      }
    }
  }

  // for all sources in graph
  for(list<int>::iterator i = graph_sources.begin(); i != graph_sources.end(); ++i) {
    INode* left_node = gd.getINode((nds[*i]));
    if (!left_node) {
      throw BindingException("Node is not available in the GraphDescription");
    }

    // used to store all unconnected out jacks and add them to this graph descrition in the end
    // this is needed to connect to unconnected out jacks in shared graph descriptions
    list<string> left_node_unconnected_out_jacks;

    // stores edges that have been connected and that can be removed from copy of this graph description (namely gd)
    list<int> edges_to_remove;

    // try to connect all outgoing edges
    const list<int>& out_list = gd.getOutEdgeList(*i);
    for(list<int>::const_iterator iter = out_list.begin(); iter != out_list.end(); ++iter) {
      GraphEdge& edge = (gd.getEdge(*iter));

      // check if edge is connected at all
      if(edge.isInConnected() && edge.isOutConnected()) {

	if(edge.overlap_type.getValue() == GraphEdge::COMPLETE) { // nothing to do, simply remove edge
	  edges_to_remove.push_back(*iter);
	}
	else if(edge.overlap_type.getValue() == GraphEdge::PARTIAL_COPY || 
		edge.overlap_type.getValue() == GraphEdge::PARTIAL_OUTPUT) { // add new connection

	  INode* right_node = gd.getINode(nds[edge.in_id.getRValue()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  right_node->reachInitialized();
	  left_node->reachOutputInitialized();
	  /* Begin> Check if events in state OUTPUT_INITIALIZED must be send to the left_node */
          if ( nds[edge.in_id.getRValue()]->getSharingType() != NodeDescription::SHARED ) {
	    if (nds[edge.in_id.getRValue()]->getNodeEvent(INode::OUTPUT_INITIALIZED)) {
	      left_node->sendEvent(*(nds[edge.in_id.getRValue()]->getNodeEvent(INode::OUTPUT_INITIALIZED)));
	    }
          }
	  /* End< */
	  // first time: get all unconnected out jacks of left node
	  if(left_node_unconnected_out_jacks.empty()) {
	    left_node_unconnected_out_jacks = left_node->getOutputStreamTags();
	  }

          /*If the edge stores no Format, create a new one */
          if ( getEdge( edge.id.getValue() ).format.getValue() == 0 )
            getEdge( edge.id.getValue() ).format.setValue(new TValue<Format>());

	  // create shared (copied) connection or connection to unconnect output
	  if(edge.overlap_type.getValue() == GraphEdge::PARTIAL_COPY) {
	    if( (edge.c_buffer_strategy.getRValue().empty()) && 
                (edge.c_event_strategy.getRValue().empty()) && 
	        (edge.out_strategy_config.size() == 0) &&
	        (edge.in_strategy_config.size() == 0) )  
            {
	      /* If no specific transport strategy should be used, use the default s_connect method */
              InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );

              cout << "s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " )" << endl;
	    } else {
	      /* Otherwise use the c_s_connect method */
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );

              cout << "c_s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) using " << edge.c_buffer_strategy.getRValue() << endl;
	    }
      realizeCommunicationChannel(getEdge(*iter).instream_channel.getValue(), edge);
	  }
	  else if(edge.overlap_type.getValue() == GraphEdge::PARTIAL_OUTPUT) {
	    Format* f = 0;
	    if(edge.format.getValue() != 0) {
	      f = edge.format.getValue()->getPValue();
	    }
            /* If no connection format is set for this edge, we create it */
            if ( getEdge( edge.id.getValue() ).format.getValue() == 0 )
              getEdge( edge.id.getValue() ).format.setValue(new TValue<Format>());

	    if( (edge.c_buffer_strategy.getRValue().empty()) &&
                (edge.c_event_strategy.getRValue().empty()) &&
	        (edge.out_strategy_config.size() == 0) &&
	        (edge.in_strategy_config.size() == 0) )  
	    {
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), f, getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );

	      cout << "connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) (other output)" << endl;
	    } else {
	      CommunicationChannel_var c_channel(c_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), f, getEdge( edge.id.getValue() ).format.getValue()->getPValue()));

              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel->copy() ));
              cout << "connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) (other output) using " << edge.c_buffer_strategy.getRValue() << endl;
	    }
	    realizeCommunicationChannel(getEdge(*iter).instream_channel.getValue(), edge);
	  }
	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.out_tag.getRValue());
	} // end: if(PARTIAL_COPY or PARTIAL_OUTPUT)
	else if(edge.overlap_type.getValue() == GraphEdge::NONE) { // no overlap -> normal connection
	  INode* right_node = gd.getINode(nds[edge.in_id.getRValue()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  right_node->reachInitialized();

	  left_node->reachOutputInitialized();
	  /* Begin> Check if events in state OUTPUT_INITIALIZED must be send to the node */
          if (nds[*i]->getSharingType() != NodeDescription::SHARED) 
	    if (nds[*i]->getNodeEvent(INode::OUTPUT_INITIALIZED))
	      left_node->sendEvent(*(nds[*i]->getNodeEvent(INode::OUTPUT_INITIALIZED)));
	  /* End< */

	  // first time: get all unconnected out jacks of left node
	  if(left_node_unconnected_out_jacks.empty()) {
	    left_node_unconnected_out_jacks = left_node->getOutputStreamTags();
	  }

	  Format* f = 0;
	  if(edge.format.getValue() != 0) {
	    f = edge.format.getValue()->getPValue();
	  }

          /* If no connection format is set for this edge, we create it */
          if ( getEdge( edge.id.getValue() ).format.getValue() == 0 )
            getEdge( edge.id.getValue() ).format.setValue(new TValue<Format>());

	  if( (edge.c_buffer_strategy.getRValue().empty()) &&
              (edge.c_event_strategy.getRValue().empty()) &&
	      (edge.out_strategy_config.size() == 0) &&
	      (edge.in_strategy_config.size() == 0) )
          {
	    try {
              InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), f, getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );

              realizeCommunicationChannel(getEdge(*iter).instream_channel.getValue(), edge);
              cout << "connect( " << (*i) << ", " << (edge.in_id).getRValue() << " )" << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );

              realizeCommunicationChannel(getEdge(*iter).instream_channel.getValue(), edge);
	      cout << "could not connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ), "
		   << "used s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) instead." << endl;
	    }
	  } else {
	    try {
	      CommunicationChannel_var c_channel(c_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), f, getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
              cout << "c_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) using " << edge.c_buffer_strategy.getRValue() << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      CommunicationChannel_var c_channel(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
	      cout << "could not c_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ), "
		   << "used c_s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) instead "
		   << "using " << edge.c_buffer_strategy.getRValue() << endl;
	    }
	  }

	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.out_tag.getRValue());
	} // end: if(NONE): no overlap -> normal connection
	else {
	  GLOBAL_ERROR_STREAM2("gd: "<< "Unhandled case in realize graph!" << endl);
	  throw BindingException("Unhandled case in realize graph!");
	}
      } // end: check if edge is connected at all
    } // end: try to connect all outgoing edges

    // since all outputs are now connected -> activate , if out_list is not empty() (not empty() means no sink node?)
    if (!out_list.empty()) {
      // need to activate node since its successor(s) needs data
      left_node->reachActivated(); 
    }

    // remove edges from list of edge indices for nodes
    for(list<int>::iterator remover = edges_to_remove.begin(); remover != edges_to_remove.end(); ++remover) {
      // check, if node with in_id becomes new source node
      int n = gd.getEdge(*remover).in_id.getValue();
      // node with no incoming edge (after remove) and some outgoing edges -> source (or unconnected node) that needs to be connected
      if(gd.getInEdgeList(n).size() == 1 && !gd.getOutEdgeList(n).empty()) {
	graph_sources.push_back(n);
      }

      gd.removeEdge(*remover);
    }

    // add all unconnceted out edges to this graph description
    SequenceValue< NodeDescription >& my_nds = this->getNodeDescriptionList();
    for(list<string>::iterator adder = left_node_unconnected_out_jacks.begin();
	adder != left_node_unconnected_out_jacks.end();
	++adder) {
      this->addOutEdge(*(my_nds[*i]), *adder);
      GLOBAL_DEBUG_STREAM2("gd: " << "Adding unconnected out edge " << *adder << " to node " << my_nds[*i]->getNodeName() << endl);
    }
  } // end: for all sources in graph

  /* Finally update the graphdescription stored in the session */
  if (__i_session.getValue()) {
    __i_session.getValue()->setGraphDescription(*this);
  }
}
// END realizeGraph *******************************************************************

void GraphDescription::realizeSharedGraph() throw(const BindingException&,const Exception&)
{
  /* Begin> Send all events for state CONSTRUCTED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != NodeDescription::SHARED ) {
      if (__node_list[i]->getNodeEvent(INode::CONSTRUCTED)) {
        __interface_list[i]->reachConstructed();
        __interface_list[i]->sendEvent(*__node_list[i]->getNodeEvent(INode::CONSTRUCTED));
      }
    }
  }
  /* End< */

  /* Begin> Send all events for state INITIALIZED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != NodeDescription::SHARED ) {
      if (__node_list[i]->getNodeEvent(INode::INITIALIZED)) {
        __interface_list[i]->reachInitialized();
        __interface_list[i]->sendEvent(*__node_list[i]->getNodeEvent(INode::INITIALIZED));
      }
    }
  }
  /* End< */

  GraphDescription gd(*this);

  list<int> graph_sources;

  // get all sources of query graph
  const SequenceValue< NodeDescription >& nds = gd.getNodeDescriptionList();
  for(size_t n = 0; n < nds.size(); ++n) {
    // node with no incoming edge and some outgoing edges -> source that needs to be connected
    if(gd.getInEdgeList(n).empty() && !gd.getOutEdgeList(n).empty()) {
      graph_sources.push_back(n);
    }
    // node with only unconnected incoming edges and some outgoing edges -> node that needs to be connected
    if(!gd.getOutEdgeList(n).empty()) {
      list<int>::const_iterator in_edge_iter;
      for(in_edge_iter = gd.getInEdgeList(n).begin(); in_edge_iter != gd.getInEdgeList(n).end(); ++in_edge_iter) {
	GraphEdge& in_edge = gd.getEdge(*in_edge_iter);
	if(in_edge.isOutConnected()) { // -> outgoind part is connected -> connected edge -> no 'source' node
	  break;
	}
      }
      if(in_edge_iter == gd.getInEdgeList(n).end()) {
	graph_sources.push_back(n);
      }
    }
  }

  // for all sources in graph
  for(list<int>::iterator i = graph_sources.begin(); i != graph_sources.end(); ++i) {
    INode* left_node = gd.getINode((nds[*i]));
    if (!left_node) {
      throw BindingException("Node is not available in the GraphDescription");
    }

    // used to store all unconnected out jacks and add them to this graph descrition in the end
    // this is needed to connect to unconnected out jacks in shared graph descriptions
    list<string> left_node_unconnected_out_jacks;

    // stores edges that have been connected and that can be removed from copy of this graph description (namely gd)
    list<int> edges_to_remove;

    // try to connect all outgoing edges
    const list<int>& out_list = gd.getOutEdgeList(*i);
    for(list<int>::const_iterator iter = out_list.begin(); iter != out_list.end(); ++iter) {
      GraphEdge& edge = (gd.getEdge(*iter));

      // check if edge is connected at all
      if(edge.isInConnected() && edge.isOutConnected()) {

	if(edge.overlap_type.getValue() == GraphEdge::COMPLETE) { // nothing to do, simply remove edge
	  edges_to_remove.push_back(*iter);
	}
	else if(edge.overlap_type.getValue() == GraphEdge::PARTIAL_COPY || 
		edge.overlap_type.getValue() == GraphEdge::PARTIAL_OUTPUT) { // add new connection

	  INode* right_node = gd.getINode(nds[edge.in_id.getRValue()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  // unlimit the queue size for right nodes of all partial shared edges
	  // -> input jacks of these nodes will buffer all downstream messages during realizing of graph
	  // -> no blocking, master graph can continue to work
	  right_node->getParentObject()->getInterface<IBiDirQueuedNode>()->unlimitDownstreamSize();

	  right_node->reachInitialized();

	  left_node->reachOutputInitialized();
	  /* Begin> Check if events in state OUTPUT_INITIALIZED must be send to the left_node */
	  if (nds[edge.in_id.getRValue()]->getSharingType() != NodeDescription::SHARED ) {
	    if (nds[edge.in_id.getRValue()]->getNodeEvent(INode::OUTPUT_INITIALIZED)) {
	      left_node->sendEvent(*(nds[edge.in_id.getRValue()]->getNodeEvent(INode::OUTPUT_INITIALIZED)));
	    }
          }
	  /* End< */
	  // first time: get all unconnected out jacks of left node
	  if(left_node_unconnected_out_jacks.empty()) {
	    left_node_unconnected_out_jacks = left_node->getOutputStreamTags();
	  }

          /* If no connection format is set for this edge, we create it */
          if ( getEdge( edge.id.getValue() ).format.getValue() == 0 )
            getEdge( edge.id.getValue() ).format.setValue(new TValue<Format>());

	  // create shared (copied) connection or connection to unconnect output
	  if(edge.overlap_type.getValue() == GraphEdge::PARTIAL_COPY) {
	    if( (edge.c_buffer_strategy.getRValue().empty()) && 
                (edge.c_event_strategy.getRValue().empty()) && 
	        (edge.out_strategy_config.size() == 0) &&
	        (edge.in_strategy_config.size() == 0) )  
            {
	      /* If no specific transport strategy should be used, use the default s_connect method */
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue() ));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );

              realizeCommunicationChannel(getEdge(*iter).instream_channel.getValue(), edge);
              cout << "s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " )" << endl;
	    } else {
	      /* Otherwise use the c_s_connect method */
	      CommunicationChannel_var c_channel(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue() ));
	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel->copy()));
              cout << "c_s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) using " << edge.c_buffer_strategy.getRValue() << endl;
	    }
	  }
	  else if(edge.overlap_type.getValue() == GraphEdge::PARTIAL_OUTPUT) {
	    Format* f = 0;
	    if(edge.format.getValue() != 0) {
	      f = edge.format.getValue()->getPValue();
	    }

            /* If no connection format is set for this edge, we create it */
            if ( getEdge( edge.id.getValue() ).format.getValue() == 0 )
              getEdge( edge.id.getValue() ).format.setValue(new TValue<Format>());

            if( (edge.c_buffer_strategy.getRValue().empty()) &&
                (edge.c_event_strategy.getRValue().empty()) &&
	        (edge.out_strategy_config.size() == 0) &&
	        (edge.in_strategy_config.size() == 0) )
	    {
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), f, getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );

	      cout << "connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) (other output)" << endl;
	    } else {
	      CommunicationChannel_var c_channel(c_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), f, getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
              cout << "connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) (other output) using " << edge.c_buffer_strategy.getRValue() << endl;
	    }
	    realizeCommunicationChannel(getEdge(*iter).instream_channel.getValue(), edge);
	  }
	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.out_tag.getRValue());
	} // end: if(PARTIAL_COPY or PARTIAL_OUTPUT)
	else if(edge.overlap_type.getValue() == GraphEdge::NONE) { // no overlap -> normal connection
	  INode* right_node = gd.getINode(nds[edge.in_id.getRValue()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  right_node->reachInitialized();

	  left_node->reachOutputInitialized();
	  /* Begin> Check if events in state OUTPUT_INITIALIZED must be send to the node */
          if ( nds[*i]->getSharingType() != NodeDescription::SHARED ) 
	    if (nds[*i]->getNodeEvent(INode::OUTPUT_INITIALIZED))
	      left_node->sendEvent(*(nds[*i]->getNodeEvent(INode::OUTPUT_INITIALIZED)));
	  /* End< */

	  // first time: get all unconnected out jacks of left node
	  if(left_node_unconnected_out_jacks.empty()) {
	    left_node_unconnected_out_jacks = left_node->getOutputStreamTags();
	  }

	  Format* f = 0;
	  if(edge.format.getValue() != 0) {
	    f = edge.format.getValue()->getPValue();
	  }

          /* If no connection format is set for this edge, we create it */
          if ( getEdge( edge.id.getValue() ).format.getValue() == 0 )
            getEdge( edge.id.getValue() ).format.setValue(new TValue<Format>());

	  if( (edge.c_buffer_strategy.getRValue().empty()) && 
              (edge.c_event_strategy.getRValue().empty()) && 
	      (edge.out_strategy_config.size() == 0) &&
	      (edge.in_strategy_config.size() == 0) )  
          {
	    try {
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), f, getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );

	      realizeCommunicationChannel(getEdge(*iter).instream_channel.getValue(), edge);
	      cout << "connect( " << (*i) << ", " << (edge.in_id).getRValue() << " )" << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );
	      realizeCommunicationChannel(getEdge(*iter).instream_channel.getValue(), edge);
	      cout << "could not connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ), "
		   << "used s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) instead." << endl;
	    }
	  } else {
	    try {
	      CommunicationChannel_var c_channel(c_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), f,getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
	      cout << "c_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) using " << edge.c_buffer_strategy.getRValue() << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      CommunicationChannel_var c_channel(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
              cout << "could not c_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ), "
		   << "used c_s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) instead "
		   << "using " << edge.c_buffer_strategy.getRValue() << endl;
	    }
	  }

	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.out_tag.getRValue());
	} // end: if(NONE): no overlap -> normal connection
	else {
	  GLOBAL_ERROR_STREAM2("gd: "<< "Unhandled case in realize graph!" << endl);
	  throw BindingException("Unhandled case in realize graph!");
	}
      } // end: check if edge is connected at all
    } // end: try to connect all outgoing edges

    // stores if left node can already be started after this iteration
    INode* left_node_to_start = 0;

    // remove edges from list of edge indices for nodes
    for(list<int>::iterator remover = edges_to_remove.begin(); remover != edges_to_remove.end(); ++remover) {
      // check, if node with out_id ( = left_node ) has no more unconnected outputs
      int left_node_id = gd.getEdge(*remover).out_id.getValue();
      list<int> left_out_edges = gd.getOutEdgeList(left_node_id);
      // node with no outgoing edge (after remove) -> reach ACTIVATED
      if(left_out_edges.size() == 1) {
	left_node_to_start = gd.getINodeList()[left_node_id];
	left_node_to_start->reachActivated();
      }
      else if(edges_to_remove.size() == 1) { // node with only unconnected outgoing edges (after remove!) -> reach ACTIVATED
	// tricky: count connected edges -> must be equal to size of edges_to_remove list
	size_t connected_out_edges = 0;
	for(list<int>::iterator iter = left_out_edges.begin(); iter != left_out_edges.end(); ++iter) {
	  if(gd.getEdge(*iter).isOutConnected()) {
	    ++connected_out_edges;
	  }
	}
	if(connected_out_edges == edges_to_remove.size()) {
	  left_node_to_start = gd.getINodeList()[left_node_id];
	  left_node_to_start->reachActivated();
	}
      }

      // check, if node with in_id ( = right node ) becomes new source node
      int right_node_id = gd.getEdge(*remover).in_id.getValue();
      // node with no incoming edge (after remove) ... 
      if(gd.getInEdgeList(right_node_id).size() == 1) {
	// ... and some outgoing edges -> source (or unconnected node) that needs to be connected
	if(!gd.getOutEdgeList(left_node_id).empty()) {
	  graph_sources.push_back(right_node_id);
	}
	
	// in any case, node has no more incoming edges that need to be connected -> reach OUTPUT_INITIALIZED
	INode* right_node = gd.getINodeList()[right_node_id];
	right_node->reachOutputInitialized();
      }

      // finally, really remove the edge
      gd.removeEdge(*remover);
    }

    if(left_node_to_start) {
      left_node_to_start->reachStarted();
      cerr << "NOWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW " << endl << endl;
      string s;
      cin >> s;
    }

    // add all unconnceted out edges to this graph description
    SequenceValue< NodeDescription >& my_nds = this->getNodeDescriptionList();
    for(list<string>::iterator adder = left_node_unconnected_out_jacks.begin();
	adder != left_node_unconnected_out_jacks.end();
	++adder) {
      this->addOutEdge(*(my_nds[*i]), *adder);
      GLOBAL_DEBUG_STREAM2("gd: " << "Adding unconnected out edge " << *adder << " to node " << my_nds[*i]->getNodeName() << endl);
    }
  } // end: for all sources in graph


  /* Finally update the graphdescription stored in the session */
  if (__i_session.getValue()) {
    __i_session.getValue()->setGraphDescription(*this);
  }
}
// END realizeSharedGraph *******************************************************************

void GraphDescription::realizeHandoverGraph(const GraphDescription& master_gd) throw(const BindingException&,const Exception&)
{
  /* Begin> Send all events for state CONSTRUCTED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if (__node_list[i]->getNodeEvent(INode::CONSTRUCTED)) {
      __interface_list[i]->reachConstructed();
      __interface_list[i]->sendEvent(*__node_list[i]->getNodeEvent(INode::CONSTRUCTED));
    }
  }
  /* End< */

  /* Begin> Send all events for state INITIALIZED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i)
    if (__node_list[i]->getNodeEvent(INode::INITIALIZED)) {
      __interface_list[i]->reachInitialized();
      __interface_list[i]->sendEvent(*__node_list[i]->getNodeEvent(INode::INITIALIZED));
    }
  /* End< */

  GraphDescription gd(*this);

  // request all connection formats from master and set them in graph description gd
  const SequenceValue<INode >& master_inodes = master_gd.getINodeList();
  for(size_t x = 0; x < master_gd.getNumEdges(); ++x) {
    GraphEdge& gd_edge = gd.getEdge(x);
    if(gd_edge.getFormat() == 0) { // no format set in gd -> take from master
      const GraphEdge& master_edge = master_gd.getEdge(x);
      if(master_edge.isOutConnected()) { // and edge really connected -> get output format of 'left' node
	INode* left  = master_inodes[master_edge.out_id.getValue()];
	Format_var c_format(left->getOutputStreamFormat(master_edge.out_tag.getRValue()));
	TValue<Format>* t_format = new TValue<Format>(*c_format);
	gd_edge.format.setValue(t_format);
	GLOBAL_DEBUG_STREAM2("gd: "<< "Update format x " << *(c_format.get()) << endl);
      }
    }
  }

  list<int> graph_sources;

  // get all sources of query graph
  const SequenceValue< NodeDescription >& nds = gd.getNodeDescriptionList();
  for(size_t n = 0; n < nds.size(); ++n) {
    // node with no incoming edge and some outgoing edges -> source that needs to be connected
    if(gd.getInEdgeList(n).empty() && !gd.getOutEdgeList(n).empty()) {
      graph_sources.push_back(n);
    }
    // node with only unconnected incoming edges and some outgoing edges -> node that needs to be connected
    if(!gd.getOutEdgeList(n).empty()) {
      list<int>::const_iterator in_edge_iter;
      for(in_edge_iter = gd.getInEdgeList(n).begin(); in_edge_iter != gd.getInEdgeList(n).end(); ++in_edge_iter) {
	GraphEdge& in_edge = gd.getEdge(*in_edge_iter);
	if(in_edge.isOutConnected()) { // -> outgoind part is connected -> connected edge -> no 'source' node
	  break;
	}
      }
      if(in_edge_iter == gd.getInEdgeList(n).end()) {
	graph_sources.push_back(n);
      }
    }
  }

  // all shared edges that will first be connected in the very end
  list<int> edges_to_connect_later;

  // for all sources in graph
  for(list<int>::iterator i = graph_sources.begin(); i != graph_sources.end(); ++i) {
    INode* left_node = gd.getINode((nds[*i]));
    if (!left_node) {
      throw BindingException("Node is not available in the GraphDescription");
    }

    // used to store all unconnected out jacks and add them to this graph descrition in the end
    // this is needed to connect to unconnected out jacks in shared graph descriptions
    list<string> left_node_unconnected_out_jacks;

    // stores edges that have been connected and that can be removed from copy of this graph description (namely gd)
    list<int> edges_to_remove;

    // try to connect all outgoing edges
    const list<int>& out_list = gd.getOutEdgeList(*i);
    for(list<int>::const_iterator iter = out_list.begin(); iter != out_list.end(); ++iter) {
      GraphEdge& edge = (gd.getEdge(*iter));

      // check if edge is connected at all
      if(edge.isInConnected() && edge.isOutConnected()) {

	if(edge.overlap_type.getValue() == GraphEdge::COMPLETE) { // nothing to do, simply remove edge
	  edges_to_remove.push_back(*iter);
	}
	else if(edge.overlap_type.getValue() == GraphEdge::PARTIAL_COPY) {
	  // only remember edge for later connection setup
	  edges_to_connect_later.push_back(*iter);
	  edges_to_remove.push_back(*iter);
	} // end: if(PARTIAL_COPY)
	else if(edge.overlap_type.getValue() == GraphEdge::NONE) { // no overlap -> unchecked_connect
	  INode* right_node = gd.getINode(nds[edge.in_id.getRValue()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  left_node->reachInitialized(); 
	  right_node->reachInitialized();

	  Format* f = 0;
	  if(edge.format.getValue() != 0) {
	    f = edge.format.getValue()->getPValue();
	  }

          /* If no connection format is set for this edge, we create it */
          if ( getEdge( edge.id.getValue() ).format.getValue() == 0 )
            getEdge( edge.id.getValue() ).format.setValue(new TValue<Format>());

	  if( (edge.c_buffer_strategy.getRValue().empty()) &&
              (edge.c_event_strategy.getRValue().empty()) &&
	      (edge.out_strategy_config.size() == 0) &&
	      (edge.in_strategy_config.size() == 0) )
          {
	    try {
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(unchecked_c_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), f));
              getEdge( edge.id.getValue() ).format.getValue()->setValue(*f);
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );
              realizeCommunicationChannel(channel, getEdge(*iter));

	      cout << "unchecked_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " )" << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( channel );
              realizeCommunicationChannel(channel, getEdge(*iter));

	      cout << "could not unchecked_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ), "
		   << "used s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) instead." << endl;
	    }
	  } else {
	    try {
	      CommunicationChannel_var c_channel(unchecked_c_connect(left_node, edge.out_tag.getRValue(),
								     right_node, edge.in_tag.getRValue(), f));
              getEdge( edge.id.getValue() ).format.getValue()->setValue(*f);
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel.get()->copy() ));
	      realizeCommunicationChannel(c_channel.get(), edge);
	      cout << "unchecked_c_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) using "
		   << edge.c_buffer_strategy.getRValue() << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      CommunicationChannel_var c_channel(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
              /* Store the returned instream channel */
              getEdge(*iter).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel.get()->copy() ));
	      realizeCommunicationChannel(c_channel.get(), edge);
	      cout << "could not c_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ), "
		   << "used c_s_connect( " << (*i) << ", " << (edge.in_id).getRValue() << " ) instead "
		   << "using " << edge.c_buffer_strategy.getRValue() << endl;
	    }
	  }

	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.out_tag.getRValue());
	} // end: if(NONE): no overlap -> normal connection
	else if(edge.overlap_type.getValue() == GraphEdge::PARTIAL_OUTPUT) {
	  GLOBAL_ERROR_STREAM2("gd: "<< "PARTIAL_OUTPUT overlap not possible for realizeHandoverGraph()" << endl);
	  throw BindingException("PARTIAL_OUTPUT overlap not possible for realizeHandoverGraph()");
	} // end: if(PARTIAL_OUTPUT)
	else {
	  GLOBAL_ERROR_STREAM2("gd: "<< "Unhandled case in realize graph!" << endl);
	  throw BindingException("Unhandled case in realize graph!");
	}
      } // end: check if edge is connected at all
    } // end: try to connect all outgoing edges

    // remove edges from list of edge indices for nodes
    for(list<int>::iterator remover = edges_to_remove.begin(); remover != edges_to_remove.end(); ++remover) {
      // check, if node with in_id becomes new source node
      int n = gd.getEdge(*remover).in_id.getValue();
      // node with no incoming edge (after remove) and some outgoing edges -> source (or unconnected node) that needs to be connected
      if(gd.getInEdgeList(n).size() == 1 && !gd.getOutEdgeList(n).empty()) {
	graph_sources.push_back(n);
      }

      gd.removeEdge(*remover);
    }

    // add all unconnceted out edges to this graph description
    SequenceValue< NodeDescription >& my_nds = this->getNodeDescriptionList();
    for(list<string>::iterator adder = left_node_unconnected_out_jacks.begin();
	adder != left_node_unconnected_out_jacks.end();
	++adder) {
      this->addOutEdge(*(my_nds[*i]), *adder);
      GLOBAL_DEBUG_STREAM2("gd: " << "Adding unconnected out edge " << *adder << " to node " << my_nds[*i]->getNodeName() << endl);
    }
  } // end: for all sources in graph

  // connect all shared edges in the end
  for(list<int>::iterator connector = edges_to_connect_later.begin(); connector != edges_to_connect_later.end(); ++connector) {
    GraphEdge& edge = (gd.getEdge(*connector));
    // check if edge is connected at all
    if(edge.isInConnected() && edge.isOutConnected()) {
      INode* left_node = gd.getINode((nds[edge.out_id.getRValue()]));
      INode* right_node = gd.getINode(nds[edge.in_id.getRValue()]);
      if (!right_node) {
	throw BindingException("Node is not available in the GraphDescription");
      }
      right_node->reachInitialized();
      left_node->reachActivated(); // only reach ACTIVATED

      /* If no connection format is set for this edge, we create it */
      if ( getEdge( edge.id.getValue() ).format.getValue() == 0 )
        getEdge( edge.id.getValue() ).format.setValue(new TValue<Format>());

      if( (edge.c_buffer_strategy.getRValue().empty()) &&
          (edge.c_event_strategy.getRValue().empty()) &&
	  (edge.out_strategy_config.size() == 0) &&
	  (edge.in_strategy_config.size() == 0) )
      {
	/* If no specific transport strategy should be used, use the default s_connect method */
	InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
        /* Store the returned instream channel */
        getEdge(edge.id.getValue()).instream_channel.setValue( channel );
        realizeCommunicationChannel( channel, getEdge(edge.id.getValue()) );
	cout << "s_connect( " << (edge.out_id).getRValue() << ", " << (edge.in_id).getRValue() << " )" << endl;
      } else {
	/* Otherwise use the c_s_connect method */
	CommunicationChannel_var c_channel(c_s_connect(left_node, edge.out_tag.getRValue(), right_node, edge.in_tag.getRValue(), getEdge( edge.id.getValue() ).format.getValue()->getPValue()));
        /* Store the returned instream channel */
        getEdge(edge.id.getValue()).instream_channel.setValue( dynamic_cast<InstreamChannel*>(c_channel.get()->copy() ) );
        realizeCommunicationChannel(c_channel.get(), edge);
	cout << "c_s_connect( " << (edge.out_id).getRValue() << ", " << (edge.in_id).getRValue() << " ) using "
	     << edge.c_buffer_strategy.getRValue() << endl;
      }
      gd.removeEdge(*connector);
    }
  }

  /* Finally update the graphdescription stored in the session */
  if (__i_session.getValue()) {
    __i_session.getValue()->setGraphDescription(*this);
  }
}
// END realizeHandoverGraph ******************************************************

void GraphDescription::activateGraph() {
 for (unsigned int i = 0; i < __interface_list.size(); ++i)
   if (__interface_list[i]) {
     __interface_list[i]->reachActivated();
     /* Begin> Check if events in state ACTIVATED must be send to the node */
     if (__node_list[i]->getNodeEvent(INode::ACTIVATED))
      __interface_list[i]->sendEvent(*__node_list[i]->getNodeEvent(INode::ACTIVATED));
     /* End< */
   }
}

void GraphDescription::startGraph()
{
  list< pair<INode*, size_t> > prefetching_nodes;

  const SequenceValue< NodeDescription >& nds = getNodeDescriptionList();

  // first, start all nodes that have no prefetching enabled 
  for(size_t i = 0; i < nds.size(); ++i) {
    INode* n = getINode(nds[i]);
    if(n) {
      if(n->getPrefetchQueueSize() == 0) { // start nodes that do not prefetch 
	n->reachStarted();
	// Begin> Check if events in state STARTED must be send to the node
	if (__node_list[i]->getNodeEvent(INode::STARTED))
	  __interface_list[i]->sendEvent(*__node_list[i]->getNodeEvent(INode::STARTED));
      }
      else { // add to list of nodes that prefetch 
	prefetching_nodes.push_back(make_pair<INode*, size_t>(n,i));
      }
    }
  }

  // then, wait for prefetching nodes to finish prefetching
  for(list< pair<INode*, size_t> >::iterator j = prefetching_nodes.begin(); j != prefetching_nodes.end(); ++j) {
    (*j).first->waitPrefetchFinished();
  }

  // finally, if all these nodes have finished prefetching, start them 
  for(list< pair<INode*, size_t> >::iterator k = prefetching_nodes.begin(); k != prefetching_nodes.end(); ++k) {
    INode* n = (*k).first;
    size_t i = (*k).second;
    if(n) {
      n->reachStarted();
      // Begin> Check if events in state STARTED must be send to the node 
      if (__node_list[i]->getNodeEvent(INode::STARTED))
	__interface_list[i]->sendEvent(*__node_list[i]->getNodeEvent(INode::STARTED));
      // End< 
    }
  }

}

void GraphDescription::stopGraph() {
  const SequenceValue< NodeDescription >& nds = getNodeDescriptionList();
  for(size_t i = 0; i < nds.size(); ++i) {
    if (getINode(nds[i]) ) {
      getINode(nds[i])->reachActivated();
    }
  }
}

void GraphDescription::unrealizeGraph() {
  const SequenceValue< NodeDescription >& nds = getNodeDescriptionList();
  for(size_t i = 0; i < nds.size(); ++i) {
    if (getINode(nds[i]) ) 
      getINode(nds[i])->reachConstructed();
  }
}

bool GraphDescription::isSharingType(NodeDescription::SharingType s) {
  /* Begin> Check if all NodeDescriptions have sharing type s*/
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != s ) {
      /* If not return false */
      return false;
    }
  }
  /* End< */

  /* Otherwise return true */
  return true;
}

bool GraphDescription::getRequestFlag(const NodeDescription& nd) const {
  return __request_flags.getRValue()[nd.getInternalID()];
}

bool GraphDescription::getReleaseFlag(const NodeDescription& nd) const {
  return __release_flags.getRValue()[nd.getInternalID()];
}

void GraphDescription::setRequestFlag(const NodeDescription& nd, bool flag) {
  __request_flags.getRValue()[nd.getInternalID()] = flag;
}

void GraphDescription::setReleaseFlag(const NodeDescription& nd, bool flag) {
  __release_flags.getRValue()[nd.getInternalID()] = flag;
}

void GraphDescription::configureOutOfBandBindings() {
  /* Begin> Configure the TransportStrategies between ProxyObjects and NMMObjects */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( !__node_list[i]->getTransportStrategy().empty() ) {
      INode* inode = getINode( *__node_list[i] );
      if (! inode ) {
        GLOBAL_ERROR_STREAM2("gd: No INode interface for node " << __node_list[i]->getNodeName() << " in GraphDescription available " << endl ) ;
        continue;
      }
      inode->getParentObject()->getOutOfBandChannel().setTransportStrategy( __node_list[i]->getTransportStrategy() ) ;
      GLOBAL_DEBUG_STREAM2("gd: " << "use TransportStragey " <<        __node_list[i]->getTransportStrategy() << " for node " << __node_list[i]->getNodeName() << endl);
    }
  }
  /* End< */
}

void GraphDescription::connectSinkSynchronizer() {
  if ( __i_session.getValue() ) {
    ISinkSynchronizer_var tmp_sync(__i_session->getSinkSynchronizer());
    IMultiAudioSinkSynchronizer_var ia_sync(tmp_sync->getParentObject()->getInterface<IMultiAudioSinkSynchronizer>());
    IMultiVideoSinkSynchronizer_var iv_sync(tmp_sync->getParentObject()->getInterface<IMultiVideoSinkSynchronizer>());

    /* Begin> Connect all sinks with the synchronizer */
    for (unsigned int i = 0; i< getINodeList().size(); ++i) {
      string node_name = getNodeDescriptionList()[i]->getNodeName();

      /* If the out edge list is empty we assume a sink node. IF NODE_DESCRIPTIONS ARE UPDATED USE THIS INFORMATION TO
       * DECIDE WHETHER A NODE IS A SINK OR NOT.
       */
      if (! getOutEdgeList( getNodeDescriptionList()[i]->getInternalID() ).empty() ){
        continue;
      } 

      /* Begin> Add the controller to the synchronizer */
      ISynchronizedSink_var av(getINodeList()[i]->getParentObject()->getInterface<ISynchronizedSink>());
      if (av.get()) {
        ISinkController_var controller (av->getController());
        av->setSynchronized(true);
        if(av->isBuffered() && ia_sync.get()) { // add audio controller ...
          ia_sync->addAudioSinkController(controller.get());
        }
        else if(iv_sync.get()) { // ... or video (unbuffered)
          iv_sync->addVideoSinkController(controller.get());
        }
      }
      /* End< */
    }
  } else {
    throw Exception("No session available for requesting synchronizer");
  }
}

void GraphDescription::setPrefetchingEnabled(bool enable) 
{
  unsigned int queue_size = 0;
  if(enable) {
    queue_size = 4;
  }

  /* Begin> enable prefetching for all sinks */
  for (unsigned int i = 0; i< getINodeList().size(); ++i) {
    string node_name = getNodeDescriptionList()[i]->getNodeName();

    /* If the out edge list is empty we assume a sink node. IF NODE_DESCRIPTIONS ARE UPDATED USE THIS INFORMATION TO
     * DECIDE WHETHER A NODE IS A SINK OR NOT.
     */
    if (! getOutEdgeList( getNodeDescriptionList()[i]->getInternalID() ).empty() ){
      continue;
    } 
    
    getINodeList()[i]->setPrefetchQueueSize(queue_size);
  }
}

void GraphDescription::disconnectSinkSynchronizer() {
  /* Begin> remove the synchronizer */
  for (unsigned int i = 0; i< getINodeList().size(); ++i) {
    /* If the out edge list is empty we assume a sink node. IF NODE_DESCRIPTIONS ARE UPDATED USE THIS INFORMATION TO
     * DECIDE WHETHER A NODE IS A SINK OR NOT.
     */
    if (! getOutEdgeList( getNodeDescriptionList()[i]->getInternalID() ).empty() ){
      continue;
    }

    if ( __i_session.getValue() ) {
      ISinkSynchronizer_var tmp_sync(__i_session->getSinkSynchronizer());
      ISinkSynchronizer_var isync(tmp_sync->getParentObject()->getCheckedInterface<ISinkSynchronizer>());

      if (isync.get()) {
        ISynchronizedSink_var av(getINodeList()[i]->getParentObject()->getInterface<ISynchronizedSink>());
        if (av.get()) {
          ISinkController_var controller (av->getController());
          isync->removeSinkController(controller.get());
        }
      }
    }
  }
  /*End< */
}

const list<INode*> GraphDescription::getPartialOverlappedNodes() 
{
  list<INode*> l;
  for(size_t j = 0; j < __internal_counter; ++j) {
    const list<int>& out_l = __out_edges.getRValue()[j];
    for(list<int>::const_iterator iter = out_l.begin(); iter != out_l.end(); ++iter) {
      GraphEdge& e = __edges.getRValue()[*iter];
      // found some node with parial overlapped outgoing edge? -> add to list and go to next node until __internal_counter
      if( (e.overlap_type.getValue() == GraphEdge::PARTIAL_COPY) || (e.overlap_type.getValue() == GraphEdge::PARTIAL_OUTPUT) ) {
	INode* n = __interface_list[e.out_id.getValue()];
	l.push_back(n);
	break;
      }
    }
  }
  return l;
}

const GraphDescription& GraphDescription::operator=(const GraphDescription& gd) {
  /* delete our internal attributes */
  clear();

  /* call operator of base class */
  SerializedValue::operator=(gd);

  /*Begin> assign all values */
  __edges             = gd.__edges;
  __out_edges         = gd.__out_edges;
  __in_edges          = gd.__in_edges;

  __node_list         = gd.__node_list;
  __interface_list    = gd.__interface_list;
  __request_flags     = gd.__request_flags;
  __release_flags     = gd.__release_flags;

  __internal_counter  = gd.__internal_counter;

  __sink_synch_descr  = gd.__sink_synch_descr;

  if (gd.getISession())
    setISession(*gd.getISession());

  /* End< */
  return *this;
}


const char* GraphDescription::GraphDescription_type_name = "GraphDescription";
const char* GraphDescription::GraphDescription_type_ns   = "NMM";

const char* GraphEdge::GraphEdge_type_name = "GraphEdge";
const char* GraphEdge::GraphEdge_type_ns   = "NMM";

TConstructor<GraphDescription>    __nmm__GraphDescription_type(GraphDescription::GraphDescription_type_name, GraphDescription::GraphDescription_type_ns );


OValueStream& operator<<(OValueStream& ov, const GraphEdge::OverlapStatus& eo)
{
  ov.beginType(TypeInfo("GraphEdge::OverlapStatus", "NMM"));
  ov << reinterpret_cast<const int&>(eo);
  ov.endType();
  return ov;
}

IValueStream& operator>>(IValueStream& iv, GraphEdge::OverlapStatus& eo)
{
  if(iv.getCurrentTypeInfo() != TypeInfo("GraphEdge::OverlapStatus", "NMM")) {
    throw InvalidTypeException(TypeInfo("GraphEdge::OverlapStatus", "NMM"), iv.getCurrentTypeInfo());
  }
  iv.goDown();
  iv >> reinterpret_cast<int&>(eo);
  iv.goUp();
  return iv;
}



NMM::OValueStream& operator<<(NMM::OValueStream& ov, const vector<NMM::GraphEdge> & string_map) 
{
  return writeVector<NMM::GraphEdge>(ov, string_map, "GraphEdge");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, vector<NMM::GraphEdge> & string_map) 
{
  return readVector<NMM::GraphEdge>(ov, string_map, "GraphEdge");
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const map<int, list<int> >& string_map)
{
  return writeMap<int, list<int> >(ov, string_map, NMM::int32_type_info, NMM::TypeInfo("list_int", "std"));
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, map<int, list<int> > & string_map)
{
  return readMap<int, list<int> >(ov, string_map, NMM::int32_type_info, NMM::TypeInfo("list_int", "std"));
}


} //end namespace
