/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_LOCALNODEFACTORY_HPP
#define NMM_LOCALNODEFACTORY_HPP

#include "nmm/base/registry/NodeFactory.hpp"

namespace NMM {

class FactoryException;
class Node;
class FactoryObject;
class Dll;

/** The LocalNodeFactory implements a NodeFactory for nodes that are managed by the LocalRegistry.
  * This basic implementation offers also the registration of so called FactoryObjects wich can be
  * used to instantiate nodes. These FactoryObjects use the new() operator to create the node and
  * delete() for destruction.
  */
class LocalNodeFactory : public NodeFactory {
friend class CreateUsingNew<LocalNodeFactory>;
public:

/** @name factory methods
  * These methods can be used to create and destroy Nodes
  */
//@{
/** This method creates a Node from a given FactoryTicket. If the node can not be created for some reason
  * an exception is thrown. The FactoryTicket is creates by a Registry, if the node is reserved. After
  * this method the FactoryTicket is invalid and can not be used again.
  * \param ticket The ticket is created by a registry and allows the maintainer to create the node.
  * \return A refernce to the new Node
  */
  virtual  Node& createNode (const FactoryTicket& ticket)
    throw (const FactoryException& );

/** This method releases a creates node. The NodeFactory can delete this node or store it for a later use.
  * \param node The node that should be released
  */
  virtual  void  releaseNode (const Node& node)
    throw (const FactoryException& );
//@}

/** @name registration methods
  * These methods can be used to register/unregister nodes
  */
//@{
/** This method register a FactoryObject with a ObjectID. The FactoryObject is used to create the node.
  * The ObjectID is created from a Registry. If the application requests a Node (which is stored in the
  * FactoryTicket), the LocalNodeFactory searches the corresponding FactoryObject and creates the Node.
  * \param node_id The ObjectID which is used to identify the FactoryObject
  * \param factory_object The FactoryObject which can create the node
  */
  virtual  void registerNode(const ObjectID& node_id, const FactoryObject& factory_object);

/** This method unregister the FactoryObject registered with given ObjectID.
  * \param node_id The ObjectID for the FactoryObject that should be removed
  */
  virtual  void unregisterNode(const ObjectID& node_id);
//@}

private:
/** @name Constructor/Destructor
  * These methods are private
  */
//@{
/** Constructor
  */
  LocalNodeFactory();

/** Destructor
  */
  virtual ~LocalNodeFactory();

/** Disabled Copy-Constructor
  */
  LocalNodeFactory(const LocalNodeFactory&);
//@}
private:
  map<ObjectID, FactoryObject*>   __factory_map;    ///< stores the registered FactoryObjects which create the nodes
  map<Node*, FactoryObject*>    __node_list;      ///< stores the created Nodes with  the corresponding FactoryObjects which can destroy the node
};

/* Typedef to access the LocalNodeFactory with the method call SLocalNodeFactory::getInstance() */
typedef Singleton<LocalNodeFactory, CreateUsingNew> SLocalNodeFactory;

} //end namespace
#endif
