/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_NODEDESCRIPTION_HPP
#define NMM_NODEDESCRIPTION_HPP

#include <string>
#include "ObjectDescription.hpp"
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/base/format/Property.hpp"

namespace NMM {
  /** \class NodeDescription
   * A NodeDescription describes the capabilities of a node.
   */
  class NodeDescription  : public ObjectDescription {
  public:
    /** The SharingType describes if a requested Node should be reserved EXCLUSIVE, so that no
     * other user can access this device. A SHARED Node can be used from several user.
     * EXCLUSIVE_SHARED means that the Node is shared if no EXCLUSIVE access is possible.
     */
    enum SharingType { UNDEFINED = -1, EXCLUSIVE = 0, EXCLUSIVE_THEN_SHARED, EXCLUSIVE_OR_SHARED, SHARED_OR_EXCLUSIVE, SHARED, EXCLUSIVE_THEN_SHARED_OR_SHARED, SHARED_OR_EXCLUSIVE_THEN_SHARED };
    /*! @name Constructor & Destructor
     */
    //@{
    /** constructor
     */ 
    NodeDescription();

    /** constructor
     * \param node_name The name of the node
     */
    explicit NodeDescription(const string& node_name);

    /** destructor
     */
    ~NodeDescription();

    /** copy constructor
     */
    NodeDescription(const NodeDescription& );

    /** virtual copy constructor
     */
    virtual NodeDescription* copy() const { return new NodeDescription(*this); }
    //@}

    /** @name set/addParameter-methods
     * Methods to set the parameter inside this description
     */
    //@{
    /** sets the name of a node wich is described in this object
     */
    inline void setNodeName       (const string& node_name)     { setObjectName(node_name); }

    /** sets the type of a node wich is described in this object
     */
    inline void setNodeType       (INode::Type node_type)       { __node_type.setValue(node_type); }

    /** sets the URL
     */
    inline void setURL            (const string& url)           { __url.setValue(url); }

    /** sets the node's input property
     */
    inline void setInputProperty  (const Property& property)    { __input_property.setValue(property); }

    /** sets the node's output property
     */
    inline void setOutputProperty (const Property& property)    { __output_property.setValue(property); }

    /** adds a specific format to the input property
     */
    inline void addInputFormat    (const Format& format)        { __input_property.getRValue().addNewFormat(new Format(format)); }

    /** adds a specific format to the output property
     */
    inline void addOutputFormat   (const Format& format )       { __output_property.getRValue().addNewFormat(new Format(format));}

    /** wrapper method to add an input and output format
     */
    inline void addFormat(const Format& input_format, const Format& output_format) { addInputFormat(input_format); addOutputFormat(output_format); }
    //@}

    /** @name getParameter-methods
     * Methods to get the parameter inside this description
     */
    //@{
    /** return the name of a Node inside the plugin
     */
    inline const string& getNodeName()          const   { return getObjectName(); }

    /** returns the type of the described Node
     */
    inline INode::Type getNodeType()            const   { return __node_type.getRValue(); }

    /** returns the URL
     */
    inline string getURL()            const   { return __url.getRValue(); }

    /** This method tests if the given NodeDescrption nd is a subset of this one
     */
    bool isSubset(const NodeDescription& nd)    const;

    /** returns the input property
     */
    inline const Property& getInputProperty()   const   { return __input_property.getRValue(); }

    /** returns the input property
     */
    inline       Property& getInputProperty()           { return __input_property.getRValue(); }

    /** returns the output property
     */
    inline const Property& getOutputProperty()  const   { return __output_property.getRValue(); }

    /** returns the output property
     */
    inline       Property& getOutputProperty()          { return __output_property.getRValue(); }
    //@}

    void setInternalID(int id)                          { __internal_id.setValue(id); }

    int getInternalID()                           const { return __internal_id.getRValue(); }

    NodeDescription::SharingType getSharingType() const { return __sharing_type.getValue(); }

    void setSharingType(NodeDescription::SharingType s_type) { __sharing_type.setValue(s_type); }

    bool allowsSharing() const;

    void addNodeEvent(const Event&, INode::State);
    CEvent* getNodeEvent(INode::State);
    const CEvent* getNodeEvent(INode::State) const;
  
  private:
    /** register all attibutes by the base class SerializedValue
     */
    void registerAttributes();

    /** checks if the second property is a subset of the first
     */
    bool isPropertySubset(const Property&, const Property&) const;

  private:
    TValue<string>                  __url;                  ///< stores an url (needed for graph builder)
    TValue<INode::Type>             __node_type;            ///< stores the type of a Node, the possibe values are defined in INode
    TValue<Property>                __input_property;       ///< stores the Input-Propoerty of a node
    TValue<Property>                __output_property;      ///< stores the Output-Propoerty of a node
    TValue<int>                     __internal_id;          ///< Internal id for the GraphDescription
    TValue<SharingType>             __sharing_type;
    SequenceValue< CEvent >         __config_events;        ///< stores the config Events. Each postion in the array is used for a certain state
  };

  NMM::OValueStream& operator<<(NMM::OValueStream& ov, const NMM::NodeDescription::SharingType& desc);
  NMM::IValueStream& operator>>(NMM::IValueStream& ov, NMM::NodeDescription::SharingType& desc);

  NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<NMM::NodeDescription>& desc);
  NMM::IValueStream& operator>>(NMM::IValueStream& ov, list<NMM::NodeDescription>& desc);

  typedef auto_ptr<NodeDescription> NodeDescription_var;
} //end namespace

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const vector<NMM::NodeDescription>& desc);
NMM::IValueStream& operator>>(NMM::IValueStream& ov, vector<NMM::NodeDescription>& desc);


#endif
