/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "Plugin.hpp"
#include "nmm/base/NMMApplication.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "PluginManager.hpp"

namespace NMM {

Plugin::Plugin(const string& node_name, int max_instance)
  : __nmm_version(VERSION), __nmm_major_nbr(atoi("NMM_VERSION_MAJOR")), __nmm_minor_nbr(atoi("NMM_VERSION_MINOR")),
    __nmm_micro_nbr(atoi("NMM_VERSION_MICRO")), __registry(0), __node_factory(0), 
    __node_name(node_name), __max_instance(max_instance)
{

}

Plugin::~Plugin()
{
  /* If the Plugin will be destroyed we must ensure that the PluginManager will be informed.
   * Because the Plugins and the PluginManager are static objects it might occur that the Plugins
   * are already destroyed. If the PluginManager is alread destroyed it is automatically recreated.
   */
  Plugin** p_array = SPluginManager::getInstance().requestPluginArray(getLibraryName());
  if ( p_array )  {
    int i = 0;
    while( p_array[i] != 0) {
      /* reset the plugins */
      p_array[i] = 0;
      ++i;
    }
  }
}


INode* Plugin::prepareRegistration(Node& node)   {
  INode*  i_node = node.getInterface<INode>();
  /* Ensure that the interface is bind to an object and we receive no
   * configuration events between ProxyNode and Node */
  i_node->establishBinding();

  /* Register the Plugin as EventListener*/
  i_node->getParentObject()->registerEventListener(this);

  return i_node;
}

void Plugin::disableEventLogging(INode& i_node)   {
  /* remove the Plugin object as EventListener*/
  i_node.getParentObject()->removeEventListener(this);
}

Result Plugin::receiveEvent( const CEvent& ce, const IEventSender* )         {
  /* Store each event in the __config_event list*/
  for (unsigned int i = 0; i<ce.size(); ++i)
    __config_events.push_back(ce[i]);

  return SUCCESS;
}

Result Plugin::receiveEvent( const Event& e  , const IEventSender* )         {
  /* Store the event in the __config_event list*/
  __config_events.push_back(e);
  return SUCCESS;
}

Result Plugin::receiveEvent( const CEvent& ce, const IEventSender* )   const {
  /* Store each event in the __config_event list*/
  for (unsigned int i = 0; i<ce.size(); ++i)
    __config_events.push_back(ce[i]);

  return SUCCESS;
}

Result Plugin::receiveEvent( const Event& e  , const IEventSender* )   const {
  /* Store the event in the __config_event list*/
  __config_events.push_back(e);
  return SUCCESS;
}


void Plugin::createNodeDescription(INode& node, NodeDescription& description)
{
  /* Disable the event logging. */
  disableEventLogging(node);

  try {
    node.init();

    /*Begin> create the NodeDescription*/
    description.setNodeName(__node_name);
    description.setLibraryName(getLibraryName());
    description.setNodeType(node.getType());
    description.setLocation(UserInfo().getHostname());

    /* Begin> store the configuration events in the description. So the node can be reset to the
              same configuration*/
    if (getConfigEvents().size() != 0 ) {
      CEvent ce(getConfigEvents().size());
      int i = 0;
      for (list<Event>::iterator it = getConfigEvents().begin(); it != getConfigEvents().end(); ++it, ++i)
        ce.setEvent(i, *it);

      description.getConfigEvent() = ce;
    }
    /* End< */

    /* Begin> Set the input property if available */
    Property_var input_property(node.getInputProperty());
    if( input_property.get() )
      description.setInputProperty(*input_property.get());
    else {
      GLOBAL_DEBUG_STREAM2("No Input Property available" << endl);
    }
    /* End< */

    /* Begin> Set the output property if available */
    Property_var output_property(node.getOutputProperty());
    if(output_property.get())
      description.setOutputProperty(*output_property.get());
    else {
      GLOBAL_DEBUG_STREAM2("No Output Property available" << endl);
    }
    /* End< */

    /* Begin> store the interfaces supported by this node*/
    for (list<Interface*>::const_iterator it = node.getParentObject()->getInterfaceList().begin(); it != node.getParentObject()->getInterfaceList().end(); ++it)
      description.addInterface(**it);
    /*End< */

    /* deinit the node */
    node.deinit();
    /* reset the event list */
    getConfigEvents().clear();   
    /* enable event logging again */
    node.getParentObject()->registerEventListener(this);
  } catch (const Exception& e) {
    /* reset the event list */
    getConfigEvents().clear();
    /* enable event logging again */
    node.getParentObject()->registerEventListener(this);
    /* Print the exception... */
    GLOBAL_ERROR_STREAM( e << endl);
    /* and throw it again*/
    throw e;
  }
}

void Plugin::registerNode(const NodeDescription& description, int max_instance) {
  /* Begin> Register the node with the created description*/
    /* If a specific registry is set use this one */
  if (getRegistry())
    addObjectID(getRegistry()->registerNode(description, max_instance, getPluginIndex()));
    /* Otherwise use the LocalRegistry */
  else
    addObjectID(SLocalRegistry::getInstance().registerNode(description, max_instance, getPluginIndex()));
  /* End< */
}


}//end namespace

