/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_REGISTRYSTRUCTS_HPP
#define NMM_REGISTRYSTRUCTS_HPP

#include <string>

#include "nmm/utils/Singleton.hpp"

#include "nmm/base/Event.hpp"
#include "RegistryTypedef.hpp"

namespace NMM {

class RegistryID;
class UserInfo;

ostream& operator<<(ostream&, const RegistryID& );
ostream& operator<<(ostream&, const UserInfo& );

/** The UserInfo stores information, that can be used to identify a specifc process on
  * in the network. To identify the host, its IP-address is used, which is unique in the network.
  * To identify the process on a host, the process-id is used.
  */
class UserInfo : public SerializedValue {
public:
/** Constructor tag for static constructor.
  */
  enum Static { STATIC };
  
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  UserInfo();

/** Copy-Constructor
  */
  UserInfo(const UserInfo& );

/** virtual Copy-Constructor
  */
  virtual UserInfo*    copy()           const { return new UserInfo(*this); }
//@}

/** @name Access UserInfo information
  */
//@{
/** This method returns the hostname of the host that creates the object.
  */
  inline const string& getHostname()    const { return __hostname.getRValue(); }

/** This method returns the domainname of the host that creates the object.
  */
  inline const string& getDomainname()  const { return __domainname.getRValue(); }

/** This method returns the user name of the user that creates the object.
  */
  inline const string& getUsername()    const { return __username.getRValue(); }

/** This method returns the password
  */
  inline const string& getPasswd()    const { return __passwd.getRValue(); }

/** This method returns the pid of the application that creates the object.
  */
  inline int           getPid()         const { return __pid.getRValue(); }
//@}

/** @name Operators
  */
//@{
/** The assignment operator stores all information from givenm UserInfo in this object
  * \param u_info UserInfo that should be assigned with this instance
  */
  const UserInfo& operator=(const UserInfo& u_info);

/** This operator compares a given UserInfo with this instance and retrns true if
  * they are equal, otherwise false.
  */
  bool operator==(const UserInfo& info) const;

/** This operator compares a given UserInfo with this instance and retrns false if
  * they are equal, otherwise true.
  */
  bool operator!=(const UserInfo& info) const { return !(*this == info);}
//@}
private://method
/** This method adds all attributes by the base class
  */
  void registerAttributes();

/** private assignment method
  */
  void assign(const UserInfo& );

private:
  TValue<string> __hostname;                  ///< stores the hostname
  TValue<string> __domainname;                ///< stores the domainname
  TValue<string> __username;                  ///< stores the username
  TValue<string> __passwd;                    ///< stores the passwd
  TValue<int>    __pid;                       ///< stores the process id
};

  
/** This singleton class stores the UserInfo for this process.
  */
class ProcessUserInfo {
public:
/** @name Access Methods
  */
//@{
/** This method returns the home directory of a user running this process.
  * To determin the home directory the environment variable home is read.
  * If it is not set the home directory is set to /home/<username>/. If the
  * username is root, it is set to /root/ 
  */
  inline const string& getHomeDirectory() const { return __home_directory; }
  
/** This method returns the username of the user running this process.
  */
  inline const string& getUsername()   const { return __username; }

/** This method returns the hostname of the host this process runs on.
  */
  inline const string& getHostname()   const { return __hostname; }

/** This method returns the domainname of the host this process runs on.
  */
  inline const string& getDomainname() const { return __domainname; }

/** This method returns the PID of this process.
  */
  inline int           getPid()        const { return __pid; }
//@}
  
private:
  friend class CreateUsingNew<ProcessUserInfo>;
  
/** @name Constructor & Destructor
  */
//@{
/** Constructor.
  */
  ProcessUserInfo();
//@}
  
private:
  const string __username;    ///< Username of user running this process
  const string __hostname;    ///< Hostname of host this process runs on
  const string __domainname;  ///< Domainname of host this process runs on
  const int    __pid;         ///< PID of this process
  string __home_directory;    ///< home direcotry of the user
};

/** Singleton class for ProcessUserInfo
  */
typedef Singleton<ProcessUserInfo, CreateUsingNew> SProcessUserInfo;

class RegistryException : public Exception {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  RegistryException()
    : Exception("RegistryException") { setTypeName("RegistryException"); setComment("RegistryException");};

/** Constructor
  * \param s The parameter can be used to set an error message
  */
  RegistryException (const string& s)
    : Exception("RegistryException : " + s) {setTypeName("RegistryException"); setComment("RegistryException : " + s);};

/** Constructor
  * \param n Object that should be used for the output
  * \param s The parameter can be used to set an error message
  */
  RegistryException (const NamedObject* n, const string& s = "")
    : Exception(n, ((s == "") ? "RegistryException" : "RegistryException : " + s)) {setTypeName("RegistryException");};

/** Virtual copy constructor
  */
  virtual RegistryException* copy() const {return new RegistryException(*this); }
//@}
};

/** The RegistryID is used to identify a Registry in the Network. To identify the process and
  * the host a UserInfo is used. To identify a Registry in the application its NMMReference is
  * used.
  */
class RegistryID : public SerializedValue {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  RegistryID(Value::MemoryMgm = Value::INVALUE);

/** Constructor
  */
  RegistryID(const string& reg_name, const NMMReference& reference, Value::MemoryMgm = Value::INVALUE);

/** Copy onstructor
  */
  RegistryID(const RegistryID& );

/** Virtual copy constructor
  */
  virtual RegistryID*     copy()                    const { return new RegistryID(*this);}
//@}

/** @name Access RegistryID information
  */
//@{
/** This method returns the name of the Registry
  */
  inline const string&        getRegistryName()     const {return __reg_name.getRValue();}

/** This method returns the NMMReference of the Registry
  */
  inline const NMMReference&  getReference()        const {return __reference.getRValue();}

/** This method returns the used UserInfo
  */
  inline const UserInfo&      getUserInfo()         const {return __user_info;}
//@}

/** @name Operator
  */
//@{
/** This operator compares a given RegistryID with this instance and retrns true if
  * they are equal, otherwise false.
  */
  bool operator==(const RegistryID& id)             const;

/** This operator compares a given RegistryID with this instance and retrns false if
  * they are equal, otherwise true.
  */
  bool operator!=(const RegistryID& id)             const { return !(*this == id); }
//@}
private:
  void allocate();

private:
  TValue<string>            __reg_name;             ///< Stores the name of the Registry
  TValue<NMMReference>      __reference;            ///< Stores the NMMReference of the Registry
  UserInfo                  __user_info;            ///< Stores the UserInfo to identify the Registry
};

class NodeInfo : public SerializedValue {
public:
/** @name Constructor & Destructor
  */
//@{
/** Default Constructor
  */
  NodeInfo();

/** Constructor to set all members
  */
  NodeInfo(const ObjectID& node_id, const string& lib_name, const string& node_name, int inst_count, int max_count, int node_type);

/** Copy-Constructor
  */
  NodeInfo(const NodeInfo&);

/** virtual copy constructor
  */
  virtual NodeInfo* copy()                                const   { return new NodeInfo(*this); }
//@}

/** @name Set NodeInfo parameter
  */
//@{
/** This method sets the ObjectID
  */
  inline void setObjectID(const ObjectID& node_id)                    { __node_id.setValue(node_id); }

/** This method sets the name of the Node
  */
  inline void setNodeName(const string& node_name)                { __node_name.setValue(node_name); }

/** This method sets the name of the library that offers the node
  */
  inline void setLibraryName(const string& lib_name)              { __library_name.setValue(lib_name); }

/** This method sets the current instance count
  */
  inline void setInstanceCount(int count)                         { __node_id.setValue(count); }

/** This method sets the maximum instance count
  */
  inline void setMaxInstanceCount(int count)                      { __node_id.setValue(count); }

/** This method sets the type of the node
  */
  inline void setNodeType(int nt)                                 { __node_type.setValue(nt); }
//@}

/** @name Get NodeInfo parameter
  */
//@{
/** This method returns the ObjectID of the described node
  */
  inline ObjectID getObjectID()                               const   { return __node_id.getValue(); }

/** This method returns the name of the library that offers the Node
  */
  inline const string& getLibraryName()                   const   { return __library_name.getRValue(); }

/** This method returns the name of the Node
  */
  inline const string& getNodeName()                      const   { return __node_name.getRValue(); }

/** This method returns the current instance count of this Node
  */
  inline int getInstanceCount()                           const   { return __instance_count.getRValue(); }

/** This method returns the maximum instance count of this Node
  */
  inline int getMaxInstanceCount()                        const   { return __max_instance_count.getRValue(); }

/** This method returns the type of the Node
  */
  inline int getNodeType()                                const   { return __node_type.getValue(); }
//@}
private:
  void registerAttributes();

private:
  TValue<ObjectID>    __node_id;                             ///< stores the ObjectID of a Node
  TValue<string>    __library_name;                        ///< stores the library name 
  TValue<string>    __node_name;                           ///< stores the name of the Node
  TValue<int>       __instance_count;                      ///< stores the current instance count
  TValue<int>       __max_instance_count;                  ///< stores the maximum instance count
  TValue<int>       __node_type;                           ///< stores the type of the node.
};

/* Begin> operator<< for STL and these structs */
OValueStream& operator<<(OValueStream& ov, const list<Response>& string_list);
OValueStream& operator<<(OValueStream& ov, const list<RegistryID>& string_list);
OValueStream& operator<<(OValueStream& ov, const list<ObjectID>& string_list);
OValueStream& operator<<(OValueStream& ov, const list<NodeInfo>& string_list);
OValueStream& operator<<(OValueStream& ov, const pair<RegistryID, list<ObjectID> >& string_list);
/* End< */

/* Begin> operator>> for STL and these structs */
IValueStream& operator>>(IValueStream& ov, list<Response>& string_list);
IValueStream& operator>>(IValueStream& ov, list<RegistryID>& string_list);
IValueStream& operator>>(IValueStream& ov, list<ObjectID>& string_list);
IValueStream& operator>>(IValueStream& ov, list<NodeInfo>& string_list);
IValueStream& operator>>(IValueStream& ov, pair<RegistryID, list<ObjectID> >& string_list);
/* End< */

}//end namespace
#endif

