/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *                    NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "FormatSizeCalculator.hpp"
#include <string>
#include <iostream>
#include <cmath>

namespace NMM {

FormatSizeCalculator::FormatSizeCalculator() {
  resolutions_4_3.clear();
  resolutions_4_3.push_back(Resolution(1280,960));
  resolutions_4_3.push_back(Resolution(1024,768));
  resolutions_4_3.push_back(Resolution(800,600));
  resolutions_4_3.push_back(Resolution(768,576));
  resolutions_4_3.push_back(Resolution(720,540));
  resolutions_4_3.push_back(Resolution(640,480));
  resolutions_4_3.push_back(Resolution(320,240));
  resolutions_4_3.push_back(Resolution(160,120));
  resolutions_4_3.push_back(Resolution(80,60));
  resolutions_4_3.push_back(Resolution(0,0));
  resolutions_16_9.clear();
  resolutions_16_9.push_back(Resolution(1280,720));
  resolutions_16_9.push_back(Resolution(1024,576));
  resolutions_16_9.push_back(Resolution(800,450));
  resolutions_16_9.push_back(Resolution(640,360));
  resolutions_16_9.push_back(Resolution(320,180));
  resolutions_16_9.push_back(Resolution(160,90));
  resolutions_16_9.push_back(Resolution(80,45));
  resolutions_16_9.push_back(Resolution(0,0));
}

FormatSizeCalculator::~FormatSizeCalculator() {
  resolutions_4_3.clear();
  resolutions_16_9.clear();
}

uint FormatSizeCalculator::getBitsPerPixel(const Format& i_format) {
  unsigned int bpp = 0;
  if (i_format.getFormatType() == string("video/raw")) {
    try {
      bpp = i_format.getIntValue("bitsperpixel");
    } catch (ParameterExceptionNotExist e) {
      if ((i_format.getStringValue(Format::colorspace_param) == string("yv12")) ||
          (i_format.getStringValue(Format::colorspace_param) == string("yuv411"))) {
        bpp = 12;
      }
      if ((i_format.getStringValue(Format::colorspace_param) == string("yuv422")) ||
          (i_format.getStringValue(Format::colorspace_param) == string("rgb565")))   {
        bpp = 16;
      }
      if (i_format.getStringValue(Format::colorspace_param) == string("rgba32")) {
        bpp = 32;
      }
    }
  }
  return bpp;
}

long FormatSizeCalculator::computeBytesPerSecond(const Format& o_format) {
  string format_type = o_format.getFormatType();

  // video formats known to use fixed size
  if (format_type == string("video/raw")) {
    unsigned int x_size = o_format.getIntValue(Format::x_resolution_param);
    unsigned int y_size = o_format.getIntValue(Format::y_resolution_param);
    float fps = o_format.getFloatValue(Format::framerate_param);
    unsigned int bpp = getBitsPerPixel(o_format);
    // bits per pixel * x_resolution * y_resolution * framerate (/8 to get bytes)
    long return_value = long (bpp * x_size * y_size * fps / 8);
    return return_value;
  }

  // audio formats known to use fixed size
  if (format_type == string("audio/raw")) {
    unsigned int channels = o_format.getIntValue("channels");
    unsigned int bps = o_format.getIntValue("bitspersample");
    unsigned int s_rate = o_format.getIntValue(Format::sample_rate_param);

    long return_value = bps * s_rate * channels;
    return return_value;
  }

  // return 0 if there was no match
  return 0;
}

Format FormatSizeCalculator::computeResolution(const Format& i_format, unsigned long bytes_per_second) {
  Format return_format = i_format;

  uint bpp = getBitsPerPixel(i_format);
  float fps = i_format.getFloatValue(Format::framerate_param);

  if (bpp == 0) {
    return_format.addIntParamValue(Format::x_resolution_param,0);
    return_format.addIntParamValue(Format::y_resolution_param,0);
    return return_format;
  }

  unsigned long single_frame_size = (unsigned long) floor (bytes_per_second / fps);
  unsigned long pixel_per_frame = (unsigned long) std::floor ((double)(single_frame_size *8) / bpp);

  try {
    i_format.getIntValue(Format::ratio_x_param);
  } catch (ParameterExceptionNotExist e) {
    // no aspect ratio given, forcing 4:3
    cerr << "asuming aspect ratio of 4:3\n";
    return_format.addIntParamValue(Format::ratio_x_param,4);
    return_format.addIntParamValue(Format::ratio_y_param,3);
  }

  unsigned long number_of_pixels;
  if (return_format.getIntValue(Format::ratio_x_param) == 4) {
    if (return_format.getIntValue(Format::ratio_y_param) == 3) {
      // 4:3 aspect ratio
      list<Resolution>::iterator rl_i = resolutions_4_3.begin();

      number_of_pixels = (rl_i->x_resolution) * (rl_i->y_resolution);
      while (number_of_pixels > pixel_per_frame) {
        rl_i++;
        number_of_pixels = (rl_i->x_resolution) * (rl_i->y_resolution);
      };
      return_format.addIntParamValue(Format::x_resolution_param,rl_i->x_resolution);
      return_format.addIntParamValue(Format::y_resolution_param,rl_i->y_resolution);
      return return_format;
    }
  } else {
    if (return_format.getIntValue(Format::ratio_x_param) == 16) {
      if (return_format.getIntValue(Format::ratio_y_param) == 9) {
        // 16:9 aspect ratio
      list<Resolution>::iterator rl_i = resolutions_16_9.begin();

      number_of_pixels = (rl_i->x_resolution) * (rl_i->y_resolution);
      while (number_of_pixels > pixel_per_frame) {
        rl_i++;
        number_of_pixels = (rl_i->x_resolution) * (rl_i->y_resolution);
      };
      return_format.addIntParamValue(Format::x_resolution_param,rl_i->x_resolution);
      return_format.addIntParamValue(Format::y_resolution_param,rl_i->y_resolution);
      return return_format;
      }
    }
  }
  return_format.addIntParamValue(Format::x_resolution_param,0);
  return_format.addIntParamValue(Format::y_resolution_param,0);
  return return_format;
}

Format FormatSizeCalculator::getRawVideoFormat(const long bytes_per_second) {
  const int x_aspect = 4;
  const int y_aspect = 3; // aspect_ratio 4:3

  int x_size,y_size;
  int min_framerate = 5;
  int max_framerate = 25; // PAL: 25 , NTSC: 30
  int framerate = max_framerate;
  int bpp = 12; // yv12

  long data_rate = bytes_per_second;

  long frame = data_rate / framerate;
  long frame_size = (8 * frame) / bpp;
  double tmp = y_aspect * frame_size;
  double y_res = std::sqrt(tmp / x_aspect);
  double x_res = (x_aspect * y_res) / y_aspect;

  x_size = (int) floor(x_res);
  y_size = (int) ceil(y_res);

  Format return_format("video/raw");
  return_format.addStringParamValue(Format::colorspace_param,"yv12");
  return_format.addIntParamValue(Format::x_resolution_param,x_size);
  return_format.addIntParamValue(Format::y_resolution_param,y_size);
  return_format.addIntParamValue(Format::ratio_x_param,4);
  return_format.addIntParamValue(Format::ratio_y_param,3);

  return return_format;
}

}

