/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "BufferStream.hpp"
#include "nmm/base/BaseExceptions.hpp"

namespace NMM {
  
  BufferOStream::BufferOStream(ssize_t prealloc /*= 0*/)
    : __buffer(),
      __used(0)
  {
    if (prealloc > 0) {
      ssize_t new_size = 1;
      while (new_size < prealloc) {
        new_size *= 2;
      }
      __buffer.resize(new_size);
    }
  }
  
  BufferOStream& BufferOStream::operator<<(const u_int16_t& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const int16_t& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const u_int32_t& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const int32_t& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const u_int64_t& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const int64_t& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const bool& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const unsigned long& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const long& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const double& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferOStream& BufferOStream::operator<<(const float& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }

  BufferOStream& BufferOStream::operator<<(const char& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }

  BufferOStream& BufferOStream::operator<<(const unsigned char& value) {
    writeValue(reinterpret_cast<const char*>(&value), sizeof(value));
    return *this;
  }

  BufferOStream& BufferOStream::operator<<(const string& value) {
    size_t size = value.size();
    writeValue(reinterpret_cast<const char*>(&size), sizeof(size));

    if (size == 0) {
      return *this;
    }

    writeValue(value.c_str(), value.size());
    return *this;
  }

  void BufferOStream::writeData() {}

  void BufferOStream::writeBuffer(const char* buffer, unsigned int size) {
    writeValue(buffer, size);
  }

  void BufferOStream::writeBuffer(const unsigned char* buffer, unsigned int size) {
    writeValue(reinterpret_cast<const char*>(buffer), size);
  }

  void BufferOStream::writeValue(const char* data, ssize_t size) {
    /* Begin> If the buffer is too small, double its size, starting with 8 bytes */
    ssize_t new_size = __buffer.size();
    if (!new_size) {
      new_size = 8;
    }
    while (__used + size > new_size) {
      new_size *= 2;
    }
    if (new_size > static_cast<ssize_t>(__buffer.size())) {
      __buffer.resize(new_size);
    }
    /* End< */

    /* Begin> Copy the data to the buffer and increase the used counter */
    memcpy(&__buffer[__used], data, size);
    __used += size;
    /* End< */
  }
  
  
  BufferIStream::BufferIStream(const vector<char>::const_iterator& begin, const vector<char>::const_iterator& end)
    : __begin(begin),
      __end(end)
  {}
  
  BufferIStream& BufferIStream::operator>>(u_int16_t& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(int16_t& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(u_int32_t& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(int32_t& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(u_int64_t& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(int64_t& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(bool& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(unsigned long& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(long& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(double& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }
  
  BufferIStream& BufferIStream::operator>>(float& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }

  BufferIStream& BufferIStream::operator>>(char& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }

  BufferIStream& BufferIStream::operator>>(unsigned char& value) {
    readValue(reinterpret_cast<char*>(&value), sizeof(value));
    return *this;
  }

  BufferIStream& BufferIStream::operator>>(string& value) {
    size_t size = 0;
    readValue(reinterpret_cast<char*>(&size), sizeof(size));

    if (size != 0) {   //string is not empty
      if (value.size() != size)
        value.resize(size);

      value.resize(size);
      readValue(reinterpret_cast<char*>(&value[0]), size);
    } else {
      value.resize(0);
    }

    return *this;
  }
  
  void BufferIStream::readBuffer(char* buffer, unsigned int size) {
    readValue(buffer, size);
  }
  
  void BufferIStream::readBuffer(unsigned char* buffer, unsigned int size) {
    readValue(reinterpret_cast<char*>(buffer), size);
  }
  
  void BufferIStream::readValue(char* data, ssize_t size) {
    if (size > __end - __begin) {
      throw UnexpectedException("Input past end of buffer");
    }
    
    memcpy(data, &(*__begin), size);
    __begin += size;
  }
}
