/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef NMM_VALUEFACTORY_HPP
#define NMM_VALUEFACTORY_HPP

#include "nmm/utils/Singleton.hpp"
#include "nmm/utils/NamedObject.hpp"
#include "ValueStream.hpp"

#include <map>
#include <string>

namespace NMM {
class Value;
class IValueStream;
class OValueStream;
class TypeInfo;

template<class T>
class TValue;

/** This Factory class creates Values.
  */
class ValueConstructor {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  ValueConstructor(const TypeInfo&);

/** Copy constructor
  */
  ValueConstructor(const ValueConstructor& );

/** Destructor
  */
  virtual ~ValueConstructor();

/** virtual copy constructor
  */
  virtual ValueConstructor* copy() const = 0;
//@}

/** @name Factory method
  */
//@{
/** Creates an instance of the value. The caller of this method is responsible to delete the object
  */
  virtual Value* createValue()     const = 0;

/** Creates an instance of the value with an TValue-Wrapper. The caller of this method is responsible to delete the object
  */
  virtual Value* createTValue()    const = 0;

/** Creates an TValue-Wrapper with a pointer to the object. The caller of this method is responsible to delete the object
  */
  virtual Value* createTPointerValue() const = 0;
//@}
private:
  TypeInfo  __type_info;        ///< Stores the TypeInfo
  bool      __register_value;   ///< is true if this object registert the typeinfo by the ValueFactory
};


class ValueFactory : public NamedObject {
public:
  friend class CreateUsingNew<ValueFactory>;

/** @name Registration methods
  */
//@{
/** This method registers a ValueConstructor with given type-information
  * \param vc ValueConstructor that can creat the value.
  * \param type The type name of the value
  * \param ns The namespace of the type
  */
  void registerValue(const ValueConstructor& vc, const string& type, const string& ns = "");

/** This method registers a ValueConstructor with given type-information
  * \param vc ValueConstructor that can creat the value.
  * \param t_info The type information of the value
  */
  void registerValue(const ValueConstructor& vc, const TypeInfo& t_info);

/** This mehtod removes a Type from the ValueFactory
  */
  void unregisterValue(const TypeInfo& ti);

/** This mehtod returns true if the given type information are known, otherwise false.
  */
  bool isRegisterd (const string& type, const string& ns = "");

/** This mehtod returns true if the given type information are known, otherwise false.
  */
  bool isRegisterd (const TypeInfo&);

/** This mehtod returns true if the given type information are known, otherwise false.
  */
  bool isValueAvailable(const TypeInfo& ti)  const;
//@}


/** @name Factory methods for values
  */
//@{
/** creates a value from given type information
  */
  Value* createValue(const string& type, const string& ns = "");

/** creates a value from given type information
  */
  Value* createValue(const TypeInfo& );

/** creates a value with an TValue-wrapper which is need for the values for events
  */
  Value* createTValue(const string& type, const string& ns = "");
/** creates a value with an TValue-wrapper which is need for the values for events
  */
  Value* createTValue(const TypeInfo&);

/** creates a value with an TValue-wrapper and T is allocated as Pointer which is need for the values for events
  */
  Value* createTPointerValue(const string& type, const string& ns = "");

/** creates a value with an TValue-wrapper and T is allocated as Pointer which is need for the values for events
  */
  Value* createTPointerValue(const TypeInfo&);

/** creates a value from given type information and perfors a dynamic cast to T
  */
  template<class T>
  T* createType(const TypeInfo& ti) {
    Value* tmp_value = createValue(ti);
    T* type = dynamic_cast<T*>(tmp_value);
    if (!type) {
      delete tmp_value;
      return 0;
    }
    return type;
  }

/** creates a value from given type information and perfors a dynamic cast to T
  */
  template<class T>
  TValue<T>* createTValueType(const TypeInfo& ti) {
    Value* tmp_value = createTValue(ti);
    TValue<T>* type = dynamic_cast< TValue<T>* >(tmp_value);
    if (!type) {
      delete tmp_value;
      return 0;
    }
    return type;
  }

/** creates a value with an TValue-wrapper and T is allocated as Pointer which is need for the
  * values for events and perfors a dynamic cast to T.
  */
  template<class T>
  TValue<T*> * createTPointerValueType(const TypeInfo& ti) {
    Value* tmp_value = createTPointerValue(ti);
    TValue<T*> * type = dynamic_cast< TValue<T*>* >(tmp_value);
    if (!type) {
      delete tmp_value;
      return 0;
    }
    return type;
  }
//@}


/** @name Factory methods for ValueStreams
  */
//@{
/** This method creates an IValueStream with a specific type.
  */
  IValueStream* createIValueStream(const string& );

/** This method creates an OValueStream with a specific type.
  */
  OValueStream* createOValueStream(const string& );
//@}

/** @name Methods for ValueStreams
  */
//@{
/** This method returns the corresponding OValueStream from an IValueStream
  */
  string getIValueStreamPartner(const string&  ) const;

/** This method returns the corresponding IValueStream from an OValueStream
  */
  string getOValueStreamPartner(const string&  ) const;

/** This method adds IValueStream with its OValueStream and the corrsponding type name
  */
  void registerValueStream(const string& i_stream, const string& o_stream, const ValueStreamConstructor& );
//@}
protected:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  ValueFactory();

/** Destructor
  */
  ~ValueFactory();

/** Disabled Copy constructor
  */
  ValueFactory(const ValueFactory&);
//@}
  ValueConstructor* findValueConstructor(const string&, const string& ns)  ;

private: //member

  typedef map<string, map<string, ValueConstructor*> >::iterator MapIt;
  typedef map<string, map<string, ValueConstructor*> >::const_iterator CMapIt;

  typedef map<string, ValueConstructor*>::iterator InternalIt;
  typedef map<string, ValueConstructor*>::const_iterator CInternalIt;

  typedef map<string, ValueStreamConstructor*>::iterator ValueStreamIt;
  typedef map<string, ValueStreamConstructor*>::const_iterator CValueStreamIt;

  map<string, map<string, ValueConstructor*> > __value_map;        ///< stores the FactoryObjects
  map<string, ValueStreamConstructor*>  __i_valuestream_map;       ///< stores the IValueConstructorn
  map<string, ValueStreamConstructor*>  __o_valuestream_map;       ///< stores the OValueConstructorn
};

/* Typedef to access the ValueFactory with the method call SValueFactory::getInstance() */
typedef Singleton<ValueFactory, CreateUsingNew> SValueFactory;


}//end namespace NMM
#endif


