/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "XMLValueStream.hpp"
#include "nmm/utils/NamedObject.hpp"
#include <vector>

namespace NMM {


XMLIValueStream::XMLIValueStream(XMLNode* node)
{
  tree.set_root(node);
}

XMLIValueStream::XMLIValueStream(const string& s)
{
  tree.read_buffer(s);
  if (tree.root()->name() != "XMLValueStream") {
    GLOBAL_ERROR_STREAM("Wrong StreamType" << endl);
  }

  parent_stack.push_front(tree.root());
  pos_stack.push_front( tree.root()->children().begin() );
}


XMLIValueStream::~XMLIValueStream()
{

}

const TypeInfo& XMLIValueStream::getCurrentTypeInfo() const {
  __current_type.t_name = getCurrentType();
  __current_type.t_namespace = getCurrentNamespace();

  return __current_type;
}

const string& XMLIValueStream::getCurrentType() const {
  /* If the Stream is empty return an empty string */
  if (isEmpty())
    return __empty_string;
  
  /* Search the property that stores the namespace */
  XMLProperty* property = (**pos_stack.begin())->property("type");
  if ( property )
    return property->value();
  else 
    cerr << "No type property availbale !!!" << endl;
  
  /* If no c_ns entry is available return an empty string */
  return __empty_string;
}

const string& XMLIValueStream::getCurrentNamespace() const {
  /* If the Stream is empty return an empty string */
  if (isEmpty())
    return __empty_string;

  /* Search the property that stores the namespace */
  XMLProperty* property = (**pos_stack.begin())->property("c_ns");
  if ( property )
    return property->value();

  /* If no c_ns entry is available return an empty string */
  return __empty_string;
}

bool XMLIValueStream::isEmpty() const {
  if ((pos_stack.size() == 0) || (parent_stack.size() == 0))
    return false;

  if (*pos_stack.begin() == (*parent_stack.begin())->children().end())
    return true;

  return false;
}


void XMLIValueStream::goUp() {
  /* If the stacks are empty return, because we can not go up */
  if (parent_stack.empty() || pos_stack.empty())
    return ;

  /* pop the first element of the parent stack */
  parent_stack.pop_front();
  /* pop the first element of the position stack */
  pos_stack.pop_front();

  /* read the type information of the next element */
  nextElement();
}


bool XMLIValueStream::goDown() {
  if (parent_stack.empty() || pos_stack.empty())
    return false;

  if ( (**pos_stack.begin())->children().empty())
    return false;

  /* Store the current parent in the parent_stack */
  parent_stack.push_front(**pos_stack.begin());

  /* Store the current position in the position_stack */
  pos_stack.push_front((*parent_stack.begin())->children().begin());
  return true;
}

bool XMLIValueStream::nextElement() {
  if (isEmpty())
    return false;
  XMLNodeIterator it = *pos_stack.begin();
  ++it;
  pos_stack.pop_front();
  pos_stack.push_front(it);
  return true;
}

void XMLIValueStream::readBuffer(char* buffer, unsigned int size) {
  if (isEmpty()) {
    GLOBAL_ERROR_STREAM( "Stream is Empty for type id : " << "char_buffer" << endl);
    return;
  }

  /* Begin> Type checking */
  if (getCurrentType() != "char_buffer") {
    GLOBAL_ERROR_STREAM( "Invalid type conversion from char_buffer to " << getCurrentType()  << endl);
    nextElement();
    return;
  }
  /* End< */

  XMLProperty* prop =  (**pos_stack.begin())->property("Value");
  if (prop) {
    unsigned char* tmp_ptr= reinterpret_cast<unsigned char*>(&prop->value()[0]);
    int new_size = prop->value().size();
    /* convert the buffer from 7 bytes to 8 bytes */
    UTF8Toisolat1(reinterpret_cast<unsigned char*>(buffer), reinterpret_cast<int*>(&size), tmp_ptr, reinterpret_cast<int*>(&new_size));
  }

  nextElement();
}

void XMLIValueStream::readBuffer(unsigned char* buffer, unsigned int size) {
  if (isEmpty()) {
    GLOBAL_ERROR_STREAM("Stream is Empty for type id : " << "unsigned_char_buffer" << endl);
    return;
  }

  /* Begin> Type checking */
  if (getCurrentType() != "unsigned_char_buffer") {
    GLOBAL_ERROR_STREAM("Invalid type conversion from char_buffer to " << getCurrentType()  << endl);
    nextElement();
    return;
  }
  /* End< */

  /* Get the property "Value" that stores the data */
  XMLProperty* prop =  (**pos_stack.begin())->property("Value");
  if (prop) {
    unsigned char* tmp_ptr= reinterpret_cast<unsigned char*>(&prop->value()[0]);
    int new_size = prop->value().size();
    /* convert the buffer from 7 bytes to 8 bytes */
    UTF8Toisolat1(buffer, reinterpret_cast<int*>(&size), tmp_ptr, reinterpret_cast<int*>(&new_size));
  }
  nextElement();
}

void XMLIValueStream::readData() {
  /* Begin> Read the XML string from the partner stream */
  if (getIPartnerStream()) {
    getIPartnerStream()->readData();
    unsigned int size = 0;
    (*getIPartnerStream()) >> size;
    if (size > 0) {
      string s;
      s.resize(size);
      char* buffer = &(s[0]);
      getIPartnerStream()->readBuffer(buffer, size);
      fromString(s);
    }
  }
  /* End< */
}


XMLIValueStream& XMLIValueStream::operator>>(u_int16_t& i) {
  readType<short unsigned int>(i, u_int16_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(int16_t&  i) {
  readType<short int>(i, int16_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(u_int32_t& i) {
  readType<unsigned int>(i, u_int32_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(int32_t&  i) {
  readType<int32_t>(i, int32_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(long& i) {
  readType<long >(i, long_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(unsigned long& i) {
  readType<unsigned long>(i, unsigned_long_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(u_int64_t& i) {
  readType<u_int64_t >(i, u_int64_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(int64_t& i) {
  readType<int64_t >(i, int64_type_info.t_name);
  return *this;
}


XMLIValueStream& XMLIValueStream::operator>>(double& i) {
  readType<double >(i, double_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(float& i) {
  readType<float >(i, float_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(char& i) {
  readType<char >(i, char_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(unsigned char& i) {
  readType<unsigned char>(i, unsigned_char_type_info.t_name);
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(string& i) {
  if (isEmpty()) {
    GLOBAL_ERROR_STREAM("Stream is Empty" << endl);
    return *this;
  }

  if (getCurrentType() != string_type_info.t_name) {
    GLOBAL_ERROR_STREAM("Try type conversion from type " << getCurrentType() << " to " << "string" << endl);
  }

  XMLProperty* prop =  (**pos_stack.begin())->property("Value");
  if (prop) {
    i = prop->value();
  }
  nextElement();
  return *this;
}

XMLIValueStream& XMLIValueStream::operator>>(bool& i) {
  readType<bool >(i, bool_type_info.t_name);
  return *this;
}

void XMLIValueStream::fromString(const string& data) {
  parent_stack.clear();
  pos_stack.clear();
  
  tree.read_buffer(data);
  if (tree.root()->name() != "XMLValueStream")
    GLOBAL_ERROR_STREAM("Wrong StreamType" << endl);

  parent_stack.push_front(tree.root());
  pos_stack.push_front( tree.root()->children().begin() );
}

const string XMLIValueStream::__i_value_stream_type = "XMLIValueStream";
const string XMLIValueStream::__empty_string;

/* XMLOValueStream*/




XMLOValueStream::~XMLOValueStream() {
}


XMLOValueStream::XMLOValueStream() 
{
  XMLNode* root_node = new XMLNode("XMLValueStream");
  tree.set_root(root_node);
  latest_node = root_node;
  parent_stack.push_front(latest_node);
}

void XMLOValueStream::beginType(const string& type, const string& ns) {
  latest_node = latest_node->add_child("Compound");
  latest_node->add_property("type", type);    
  if (!ns.empty()) {
    latest_node->add_property("c_ns", ns);
  }
  parent_stack.push_front( latest_node );
}

void XMLOValueStream::endType() {
  if (parent_stack.empty()) {
    GLOBAL_ERROR_STREAM("XMLOValueStream : end error" << endl);
    return;
  }
  parent_stack.pop_front();
  latest_node = (*parent_stack.begin());
}

string XMLOValueStream::toString() const {
   return tree.write_buffer();
}


void XMLOValueStream::writeBuffer(const char* buffer, unsigned int size) {
  XMLNode* tmp_node = latest_node->add_child("simple");
  tmp_node->add_property("type", "char_buffer");
  string s;
  unsigned int new_size = size * 2;
  s.resize(new_size);
  /* convert the buffer to UTF8 */
  isolat1ToUTF8(reinterpret_cast<unsigned char*>(&s[0]), reinterpret_cast<int*>(&new_size), reinterpret_cast<const unsigned char*>(buffer), reinterpret_cast<int*>(&size));
  s.resize(new_size);
  tmp_node->add_property("Value", s);
}

void XMLOValueStream::writeData()  {
  if (getOPartnerStream()) {
    string s = toString();
    (*getOPartnerStream()) << s.size();
    getOPartnerStream()->writeBuffer(&s[0], s.size());
    getOPartnerStream()->writeData();
  }
  parent_stack.clear();
  XMLNode* root_node = new XMLNode("XMLValueStream");
  tree.set_root(root_node);
  latest_node = root_node;
  parent_stack.push_front(latest_node);
}
void XMLOValueStream::writeBuffer(const unsigned char* buffer, unsigned int size) {
  XMLNode* tmp_node = latest_node->add_child("simple");
  tmp_node->add_property("type", "unsigned_char_buffer");

  XMLProperty* property = tmp_node->add_property("Value", string());
  string& s = property->value();

  unsigned int new_size = size * 2;
  s.resize(new_size);
  /* convert the buffer to UTF8 */
  isolat1ToUTF8(reinterpret_cast<unsigned char*>(&s[0]), reinterpret_cast<int*>(&new_size), reinterpret_cast<const unsigned char*>(buffer), reinterpret_cast<int*>(&size));
  s.resize(new_size);
}


XMLOValueStream& XMLOValueStream::operator<<(const short unsigned int &i) {
  writeType<short unsigned int>(i, u_int16_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const unsigned int &i) {
  writeType<unsigned int>(i, u_int32_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const int32_t&  i) {
  writeType<int>(i, int32_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const short int&  i) {
  writeType<short int>(i, int16_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const unsigned long& i) {
  writeType<unsigned long>(i, unsigned_long_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const u_int64_t& i) {
  writeType<u_int64_t>(i, u_int64_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const int64_t& i) {
  writeType<int64_t>(i, int64_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const long& i) {
  writeType<long>(i, long_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const double& i) {
  writeType<double>(i, double_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const float& i) {
  writeType<float>(i, float_type_info.t_name);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const char& i) {
  writeType<char>(i, char_type_info.t_name);
  return *this;
}
XMLOValueStream& XMLOValueStream::operator<<(const unsigned char& i) {
  writeType<unsigned char>(i, unsigned_char_type_info.t_name);
  return *this;
}


XMLOValueStream& XMLOValueStream::operator<<(const string& i) {
  writeType<string>(i, string_type_info.t_name, string_type_info.t_namespace);
  return *this;
}

XMLOValueStream& XMLOValueStream::operator<<(const bool& i) {
  writeType<bool>(i, bool_type_info.t_name);
  return *this;
}



const string XMLOValueStream::__o_value_stream_type = "XMLOValueStream";
TValueStreamConstructor<XMLIValueStream, XMLOValueStream> xml_value_stream(XMLIValueStream::__i_value_stream_type, XMLOValueStream::__o_value_stream_type);

} //end namespace NMM


