/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_PACKET_STREAM_HPP
#define NMM_PACKET_STREAM_HPP

#include "NetStream.hpp"

#include "nmm/utils/thread/Thread.hpp"

namespace NMM {

  /**
   *  @brief A packet of data for network transmission.
   *
   *  The packet stream delivers only complete messages and splits these messages into one or more packets,
   *  depending on the maximum packet size.
   *
   *  @sa PacketNetStream
   */
  class PacketHeader {
  public:
    /**
     *  @brief Constructor
     *  @param message_id Numeric ID of the complete message this packet belongs to
     *  @param packet_no Packet sequence number in this message
     *  @param packet_count Total number of packets in this message
     *  @param size Payload size of this packet
     */
    PacketHeader(int message_id, int packet_no, int packet_count, ssize_t size);
    
    /** Get packet payload size in bytes.
     */
    ssize_t getPacketSize() const {
      return __size;
    }
    
    /** Get message ID.
     */
    int getMessageID() const {
      return __message_id;
    }
    
    /** Get packet sequence number in the message.
     */
    int getPacketNumber() const {
      return __packet_no;
    }
    
    /** Get number of packets in the message.
     */
    int getPacketCount() const {
      return __packet_count;
    }
    
  private:
    u_int32_t __message_id;        ///< ID of the message, this packet is part of
    u_int16_t __packet_no;         ///< Sequence number, can be between 0 and packet_count-1
    u_int16_t __packet_count;      ///< Number of packets that this message consists of
    u_int32_t __size;              ///< Size of packet data in bytes
  };
  
  
  /**
   *  @brief Packet network stream
   *
   *  Special NetStream which splits data into packets for consistent transmission on unreliable connections.
   */
  class PacketNetStream : public NetStream {
  public:
    /**
     *  @name Constructor & Destructor
     */
    //@{
    /**
     *  @brief Constructor
     *  @param sock The socket use for packet transmission
     */
    PacketNetStream(Sock& sock);
    
    /** Copy Constructor.
     */
    PacketNetStream(const PacketNetStream& other);

    /** Destructor.
     */
    virtual ~PacketNetStream();

    /** virtual Copy constructor.
     */
    virtual PacketNetStream* copy() const;
    //@}
    
    /**
     *  @name IStream methods
     */
    //@{
    /**
     *  @brief This method actually writes the data
     *
     *  The data is split into packets which are then sent over the socket.
     */
    virtual void writeData();

    /** Returns true if the stream does not contain any data
      */
    virtual bool isEmpty() const { return true; }
    //@}    

    /**
     *  @name OStream methods
     */
    //@{
    /**
     *  @brief This method actually reads the data
     *
     *  A new incoming message is constructed from a set of packets that make up a complete message.
     *  This method may only be called after receivePacket returned true.
     */
    virtual void readData();
    //@}
    
    /**
     *  @name NetStream methods
     */
    //@{
    /**
     *  @brief Handle incoming data
     *  @return @c true, if a complete message is waiting in the packet queue
     *
     *  This method reads exactly one packet from the socket, checks if it is a valid packet and
     *  adds it to the incoming packet queue. It then checks whether a complete message can be constructed
     *  from the packet queue, and discards old packets in the queue.
     */
    virtual bool receivePacket();
    //@}
    
    /**
     *  @name Protocol Options
     */
    //@{
    /**
     *  @brief Set the maximum size of a packet
     *  @param size Maximum packet size in bytes
     */
    virtual void setPacketSize(const ssize_t& size);
    
    /**
     *  @brief Get the maximum size of a packet
     *  @return Maximum packet size in bytes
     */
    virtual ssize_t getPacketSize() const;
    
    /**
     *  @brief Set the maximum number of partial messages in the receive queue
     *  @param max_partial Maximum number of partial messages
     *
     *  Packets in the receive queue with a message ID less than the value set by this method are
     *  thrown away.
     */
    virtual void setMaxPartialMessages(const unsigned int& max_partial);
    
    /**
     *  @brief Get the maximum number of partial messages in the receive queue
     *  @return Maximum number of partial messages
     *
     *  Packets in the receive queue with a message ID less than the value returned by this method are
     *  thrown away.
     */
    virtual unsigned int getMaxPartialMessages() const;
    //@}
    
    /**
     *  @name Protocol Statistics
     */
    //@{
    /** This method resets all statistics.
     */
    virtual void resetStatistics();
    
    /** This method returns the total amount of bytes sent.
     */
    virtual ssize_t getTotalSentBytes() const;
    
    /** This method returns the total amount of bytes received.
     */
    virtual ssize_t getTotalReceivedBytes() const;
    
    /** This method returns the total number of packets sent.
     */
    virtual unsigned int getTotalSentPackets() const;
    
    /** This method returns the total number of packets received.
     */
    virtual unsigned int getTotalReceivedPackets() const;
    
    /** This method returns the total number of discarded incoming packets.
     */
    virtual unsigned int getTotalDiscardedPackets() const;
    
    /** This method returns the total number of lost incoming packets.
     */
    virtual unsigned int getTotalLostPackets() const;
    //@}
    
  protected:
    /** Internal method to write the data.
     */
    virtual void writeValue(const void*, size_t);

    /** Internal method to read the data.
     */
    virtual void readValue(void*, size_t);
    
  private:
    /**
     *  @brief A packet is represented as a vector of bytes.
     *
     *  To get the packet header, use a reinterpret_cast to PacketHeader.
     */
    typedef vector<char> Packet;    
    
    /** Write packet to socket.
     */
    void writePacket(const void* packet);
    
    /** Read packet from socket.
     * @return Actual number of bytes received
     */
    ssize_t readPacket(void* packet);
    
    /** Check if we have a complete message.
     */
    bool checkForCompleteMessage();
    
    /** Discard all packets older than the current message.
     */
    void discardOldPackets();
    
  private:
    /* Begin> Packet buffers */
    ThreadMutex __packet_mutex;             ///< Mutex to protect packet buffers
    int __max_packet_size;                  ///< Maximum packet size
    int __max_partial;                      ///< Maximum number of partial messages in queue
    Packet __i_packet;                      ///< Input packet buffer
    Packet __o_packet;                      ///< Output packet buffer
    bool __stream_socket;                   ///< Working on a stream socket
    /* End< Packet buffers */
    
    /* Begin> Incoming packet queue */
    ThreadMutex __packet_queue_mutex;       ///< Mutex to protect packet queue
    list<Packet> __i_packets;               ///< Packets received
    /* End< Incoming packet queue */
    
    /* Begin> Messages */
    vector<char> __i_message;               ///< Current input message buffer
    vector<char> __o_message;               ///< Current output message buffer
    int __i_message_pos;                    ///< Write position for input message
    int __o_message_pos;                    ///< Read position for output message
    int __i_message_id;                     ///< Last input message id
    int __o_message_id;                     ///< Current output message id
    /* End< */
    
    /* Begin> Statistics */
    ssize_t __total_sent_bytes;             ///< Total amount of bytes sent
    ssize_t __total_received_bytes;         ///< Total amount of bytes received
    unsigned int __total_sent_packets;      ///< Total number of packets sent
    unsigned int __total_received_packets;  ///< Total number of packets received
    unsigned int __total_discarded_packets; ///< Total number of packets discarded
    unsigned int __total_lost_packets;      ///< Total number of packets lost
    /* End< */
  };
}

#endif
