/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "GraphURLObject.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/device/cdda/ICDDA.hpp"
#include "nmm/interfaces/device/dvd/IDVD.hpp"
#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/interfaces/video/tv/IIVTV.hpp"

namespace NMM {

  Format* GraphURLObject::getFormatFromExtension() const
  {
    string format_type = "";
   
    if(__url->getURIExtension() == "mp3") {
      format_type = Format::audio_mpeg_type;
    }
    else if((__url->getURIExtension() == "vob") || 
	    (__url->getURIExtension() == "mpg") ||
	    (__url->getURIExtension() == "mpeg") ||
	    (__url->getURIExtension() == "vdr")) {
      format_type = Format::av_mpegps_type;
    }
    else if(__url->getURIExtension() == "wav") {
      format_type = Format::audio_raw_type;
    }
    else if(__url->getURIExtension() == "ac3") {
      format_type = Format::audio_ac3_type;
    }
    else if(__url->getURIExtension() == "ogg") {
      format_type = Format::audio_x_ogg_type;
    }
    else if(__url->getURIExtension() == "ogm") {
      format_type = Format::av_ogm_type;
    }

    if(format_type.empty()) {
      return 0;
    }
    else {
      return new Format(format_type);
    }
  }

  // ****************** FileURLObject ***************************

  string FileURLObject::getNodeNameFromExtension() const
  {
    string node_name = "";

    if(( __url->getURIExtension() == "mpg")  ||
       ( __url->getURIExtension() == "mpeg") ||
       ( __url->getURIExtension() == "vdr")  ||
       ( __url->getURIExtension() == "vob")) {
      node_name = "GenericReadNode";
    }
    else if(__url->getURIExtension() == "avi") {
      node_name = "AVIReadNode";
    }
    return node_name;
  }

  string FileURLObject::getNodeNameFromContent() const
  {
    string node_name = "";
    string mime_type = GraphURL::getMIMETypeFromFile(__url->getURIPath());

    // image data -> MagickReadNode
    if(mime_type.find("image") != string::npos) {
      node_name = "MagickReadNode";
    }
    return node_name;
  }

  Format* FileURLObject::getFormatFromContent() const
  {
    string format_type = "";

    string mime_type = GraphURL::getMIMETypeFromFile(__url->getURIPath());

    // 'file' will return audio/x-ogg for all ogg/ogm -> check again
    if(mime_type == Format::audio_x_ogg_type) {
      string mime_type2 = GraphURL::getMIMETypeFromFile(__url->getURIPath(), false);
      if(mime_type2.find("Vorbis audio") != string::npos) {
	format_type = Format::audio_x_ogg_type;
      }
      else {
	format_type = Format::av_ogm_type;
      }
    }

    // 'file' will return video/mpeg for all mpeg a/v files -> reset to av/mpegps
    else if(mime_type == Format::video_mpeg_type) {
      format_type = Format::av_mpegps_type;
    }

    // audio/x-wav -> audio/raw
    else if(mime_type.find(Format::audio_x_wav_type) != string::npos) {
      format_type = Format::audio_raw_type;
    }

    // avi files will output msvideo -> check again for AVI
    else if(mime_type.find("msvideo") != string::npos) {
      string mime_type2 = GraphURL::getMIMETypeFromFile(__url->getURIPath(), false);
      if(mime_type2.find("AVI") != string::npos) {
	format_type = Format::av_avi_type;
      }
    }

    if(format_type.empty()) {
      return 0;
    }
    else {
      return new Format(format_type);
    }
  }

  list<string> FileURLObject::getInterfaces() const
  {
    list<string> l;
    l.push_back(IFileHandler::IFileHandler_type);
    return l;
  }

  void FileURLObject::configureConstructed(INode* node) 
  {
    IFileHandler_var file_handler(node->getParentObject()->getCheckedInterface<IFileHandler>());
    file_handler->setFilename(__url->getURIPath());
  }

  // ****************** AudioCDURLObject ***************************

  void AudioCDURLObject::configureConstructed(INode* node) 
  {
    string device = __url->getURIPath();
    if (device == "") {
      device = string("/dev/cdrom");
    }
    
    ICDDA_var icdda(node->getParentObject()->getCheckedInterface<ICDDA>());
    icdda->setDevice(device);
  }

  void AudioCDURLObject::configureOutputInitialized(INode* node) 
  {
    unsigned int track = 0;
    string track_str = __url->getOption("track");
    if (track_str == "") {
      track = 0;
    } 
    else {
      track = static_cast<unsigned int>(atoi(track_str.c_str()));
    }
    ICDDA_var icdda(node->getParentObject()->getCheckedInterface<ICDDA>());
    icdda->setTrack(track);
    icdda->setOneTrackOnly(true);
    icdda->setRepeat(false);
  }

  list<string> AudioCDURLObject::getInterfaces() const 
  {
    list<string> l;
    l.push_back(ICDDA::ICDDA_type);
    return l;
  }

  Format* AudioCDURLObject::getFormatFromProtocol() const
  {
    return new Format(Format::audio_raw_type);
  }

  // ****************** DVDURLObject ***************************

  void DVDURLObject::configureConstructed(INode* node) 
  {
    string device = __url->getURIPath();
    if (device == "") {
      device = string("/dev/cdrom");
    }   
    
    IDVD_var idvd(node->getParentObject()->getCheckedInterface<IDVD>());
    idvd->setDevice(device);
  }

  void DVDURLObject::configureOutputInitialized(INode* node) 
  {
    unsigned int title = 0;
    string title_str = __url->getOption("title");
    if (title_str == "") {
      title = 1;
    } 
    else {
      title = static_cast<unsigned int>(atoi(title_str.c_str()));
    }

    unsigned int chapter = 0;
    string chapter_str = __url->getOption("chapter");
    if (chapter_str == "") {
      chapter = 1;
    } 
    else {
      chapter = static_cast<unsigned int>(atoi(chapter_str.c_str()));
    }

    unsigned int angle = 0;
    string angle_str = __url->getOption("angle");
    if (angle_str == "") {
      angle = 1;
    } 
    else {
      angle = static_cast<unsigned int>(atoi(angle_str.c_str()));
    }

    IDVD_var idvd(node->getParentObject()->getCheckedInterface<IDVD>());
    idvd->select(title, chapter, angle);
  }

  list<string> DVDURLObject::getInterfaces() const 
  {
    list<string> l;
    l.push_back(IDVD::IDVD_type);
    return l;
  }

  Format* DVDURLObject::getFormatFromProtocol() const
  {
    return new Format(Format::av_mpegps_type);
  }

  // ******************** GenericTVURLObject ******************************
  void GenericTVURLObject::configureConstructed(INode* node) 
  {
    string device = __url->getURIPath();
    if (device == "") {
      device = string("/dev/video0");
    }   

    ITVCard_var tv(node->getParentObject()->getCheckedInterface<ITVCard>());
    tv->setDevice(device);
  }

  void GenericTVURLObject::configureInitialized(INode* node) 
  {

    unsigned int channel = 0;
    string channel_str = __url->getOption("channel");
    if (channel_str == "") {
      channel = 0;
    }else {
      channel = atoi(channel_str.c_str());
    }
    
    ITVCard_var tv(node->getParentObject()->getCheckedInterface<ITVCard>());
    tv->setChannel(channel);        
  }

  list<string> GenericTVURLObject::getInterfaces() const 
  {
    list<string> l;
    l.push_back(ITVCard::ITVCard_type);
    return l;
  }

  // ******************** TVURLObject ******************************
  string TVURLObject::getNodeNameFromProtocol() const
  {
    return "TVCardReadNode";
  }
  
  // ******************** DVBTVURLObject ******************************
  string DVBTVURLObject::getNodeNameFromProtocol() const
  {
    return "DVBReadNode";
  }

  // ******************** IVTVURLObject ******************************
  string IVTVURLObject::getNodeNameFromProtocol() const
  {
    return "IVTVReadNode";
  }

  list<string> IVTVURLObject::getInterfaces() const 
  {
    list<string> l;
    l.push_back(ITVCard::ITVCard_type);
    l.push_back(IIVTV::IIVTV_type);
    return l;
  }

  void IVTVURLObject::configureInitialized(INode* node) 
  {
    int input = 4;
    string input_option = __url->getOption("input");
    if(input_option == ""){
    }else{
      input = atoi(input_option.c_str());
      ITV_var itv(node->getParentObject()->getCheckedInterface<ITV>());
      itv->setCurrentInput(input);
    }


    string table = "pal-europe";
    string table_option = __url->getOption("freqtable");
    if(table_option == ""){
    }else{
      table = table_option;
    }

    ITV_var freqtab(node->getParentObject()->getCheckedInterface<ITV>());
    freqtab->setFrequencyTable(table);

    unsigned int channel = 0;
    string channel_str = __url->getOption("channel");
    if (channel_str == "") {
      channel = 0;
    }else {
      channel = atoi(channel_str.c_str());
    }
    
    ITVCard_var tv(node->getParentObject()->getCheckedInterface<ITVCard>());
    tv->setChannel(channel);        
  }

  // ********************** MPEGTVURLObject **************************
  void MPEGTVURLObject::configureConstructed(INode* node) 
  {
    IFileHandler_var file(node->getParentObject()->getCheckedInterface<IFileHandler>());
    file->setFilename(__url->getURIPath());
  }

  string MPEGTVURLObject::getNodeNameFromProtocol() const
  {
    return "GenericReadNode";
  }
}
