
#ifndef NMM_INodeImpl_HPP
#define NMM_INodeImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "INode.hpp"

#include "IInputJackImpl.hpp"
#include "IOutputJackImpl.hpp"

namespace NMM
{

  class INodeImpl  : public virtual NMMObject { 
  public:
    INodeImpl() { addInterface(new INode(*this)); registerEvents(); };

    virtual ~INodeImpl() {};

  protected:
    /**
     * This method will try to reach the speficied state by calling the
     * required state transition methods
     */

    virtual void reachState(const INode::State& state) 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method tries to reach the state CONSTRUCTED
     */

    virtual void reachConstructed() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method tries to reach the state INITIALIZED
     */

    virtual void reachInitialized() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method tries to reach the state OUTPUT_INITIALIZED
     */

    virtual void reachOutputInitialized() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method tries to reach the state ACTIAVTED
     */

    virtual void reachActivated() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method tries to reach the state STARTED
     */

    virtual void reachStarted() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method enables the pre-fetch mode and at the same time sets
     * the number of messages to be received before pre-fetching can be
     * finished. This method has to be called in state CONSTRUCTED.
     */

    virtual void setPrefetchQueueSize(const unsigned int& pre_fetch_queue_size) /* limited(INode::CONSTRUCTED) */ = 0;
    /**
     * Request the number of messages to be prefetched. If zero,
     * prefetching is disabled.
     */

    virtual unsigned int getPrefetchQueueSize() = 0;
    /**
     * Blocks until the process of prefetching is finished. This
     * method has to be called in state ACTIVATED.
     */

    virtual void waitPrefetchFinished() /* limited(INode::ACTIVATED) */ = 0;
    /**
     * This method initialize a node and the node enters state INITIALIZED.
     */

    virtual void init() /* limited(INode::CONSTRUCTED) */ 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method initialize a node and the node enters state OUPUT_INITIALIZED.
     */

    virtual void initOutput() /* limited(INode::INITIALIZED) */ 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method activates a node. 
     * The node enters an ACTIVATED state. This does not mean output
     * is generated from any input since the node's play method needs
     * to be called.
     */

    virtual void activate() /* limited(INode::OUTPUT_INITIALIZED) */ 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method causes the node to perform a flush-operation.
     */

    virtual Result flush() /* limited(INode::ACTIVATED) */ /* limited(INode::STARTED) */ = 0;
    /*
     * This method causes the node to start generating output
     * or to generate output from input.
     */

    virtual void start() /* limited(INode::ACTIVATED) */ 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method causes the node to stop generating output
     * or to no longer generate output from input.
     */

    virtual void stop() /* limited(INode::STARTED) */ 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method deactivates a node. 
     */

    virtual void deactivate() /* limited(INode::ACTIVATED) */ 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method deinits the output of a node. 
     */

    virtual void deinitOutput() /* limited(INode::OUTPUT_INITIALIZED) */ 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;
    /**
     * This method deinits a node. 
     */

    virtual void deinit() /* limited(INode::INITIALIZED) */ 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) = 0;

    virtual bool isConstructed() const = 0;

    virtual bool isInitialized() const = 0;

    virtual bool isOutputInitialized() const = 0;

    virtual bool isActivated() const = 0;

    virtual bool isStarted() const = 0;
    /**
     *  This method returns the successor node of this node.
     *  \param out_jack the name of the output.
     *  \return a node or NULL.
     */

    virtual INode* getSuccessorNode(const string& out_jack) = 0;
    /**
     * This method returns the predecessor node of this node.
     * \param in_jack the name of the input.
     * \return a node or NULL.
     */

    virtual INode* getPredecessorNode(const string& in_jack) = 0;
    /**
     * This method returns an input jack with the given format
     * and jack tag.
     * \param format the format for the jack
     * \param jack_tag the name of the jack.
     * \return an input jack or 0 if no input jack with the given parameters could be obtained.
     */

    virtual IInputJack* getInputJack(const Format& format, const string& jack_tag) /* limited(INode::INITIALIZED) */ = 0;
    /**
     * This method returns a connected input jack with the given jack tag.
     * \param jack_tag the name of the jack.
     * \return an input jack or 0 if no input jack with the given parameter could be obtained.
     */

    virtual IInputJack* getInputStream(const string& jack_tag) /* limited(INode::INITIALIZED) */ = 0;
    /**
     * This method returns the list of available intput stream tags.
     * \return a list of input stream tags as strings.
     */

    virtual list<string >  getInputStreamTags() const = 0;
    /**
     * This method can be used to test if a InputJack of a Node is connected.
     * \param jack_tag The name of the Jack to test.
     * \return true if the Jack is connected or false if not or the the Jack does not exist.
     */

    virtual bool isInputConnected(const string& jack_tag = "default") const = 0;
    /**
     * Obtain the format associated with the named input stream.
     *
     * \param jack_tag The tag of the input stream.
     * \returns a pointer to the corresponding format.
     */

    virtual Format* getInputStreamFormat(const string& jack_tag = "default") = 0;
    /**
     * This method returns an output jack with the given format
     * and jack tag.
     * \param format the format for the jack
     * \param jack_tag the name of the jack.
     * \return an output jack or 0 if no output jack with the given parameters could be obtained.
     */

    virtual IOutputJack* getOutputJack(const Format& format, const string& jack_tag) /* limited(INode::OUTPUT_INITIALIZED) */ = 0;
    /**
     * This method returns an output jack with the given format
     * and jack tag.
     * \param format the format for the jack
     * \param jack_tag the name of the jack.
     * \return an output jack or 0 if no output jack with the given parameters could be obtained.
     */

    virtual IOutputJack* getUnintializedOutputJack(const Format& format, const string& jack_tag) /* limited(INode::INITIALIZED) */ = 0;
    /**
     * This method returns a connected output jack with the given jack tag.
     * \param jack_tag the name of the jack.
     * \return an output jack or 0 if no output jack with the given parameter could be obtained.
     */

    virtual IOutputJack* getOutputStream(const string& jack_tag) /* limited(INode::OUTPUT_INITIALIZED) */ = 0;
    /**
     * This method returns the list of available output stream tags.
     * \return a list of output stream tags as strings.
     */

    virtual list<string >  getOutputStreamTags() const = 0;
    /**
     * This method can be used to test if a OutputJack of a Node is connected.
     * \param jack_tag The name of the Jack to test.
     * \return true if the Jack is connected or false if not or the the Jack does not exist.
     */

    virtual bool isOutputConnected(const string& jack_tag = "default") const = 0;
    /**
     * This method returns a cloned output jack and fills the given format with
     * the intersection format
     * \param jack_tag the name of the jack
     * \param intersec_format the format to be filled with the intersection format
     * \return the cloned output jack
     */

    virtual IOutputJack* cloneOutputJack(Format*& intersec_format, const string& jack_tag = "default") /* limited(INode::OUTPUT_INITIALIZED) */ /* limited(INode::ACTIVATED) */ = 0;
    /**
     * Obtain the format associated with the named output stream.
     *
     * \param jack_tag The tag of the output stream.
     * \returns a pointer to the corresponding format.
     */

    virtual Format* getOutputStreamFormat(const string& jack_tag = "default") = 0;
    /**
     * This method returns the cost for the specified jack_tag.
     * \param jack_tag name of the jack.
     * \return the costs for this jack.
     **/

    virtual float getCost(const string& jack_tag) const = 0;
    /**
     * Returns the type of the Node.
     *  \return the type of the node.
     **/

    virtual INode::Type getType() const = 0;
    /** This method returns current state
     */

    virtual INode::State getState() const = 0;
    /**
     * This method returns the InputProperty of the node
     * \return Property *
     */

    virtual Property* getInputProperty() = 0;

    virtual Property* getWorkingInputProperty() = 0;
    /**
     * This method returns the OutputProperty of the node
     * \return Property *
     */

    virtual Property* getOutputProperty() = 0;

    virtual Property* getWorkingOutputProperty() = 0;
     /**
      * This method returns the computation configuration of the node
      */

    virtual ComputationConfig* getComputationConfig() const = 0;

    virtual ComputationConfig* getWorkingComputationConfig() const = 0;
     /**
      * This method sets the computation configuration of the node
      */

    virtual void setComputationConfig(const ComputationConfig* cc) = 0;

    virtual void setWorkingComputationConfig(const ComputationConfig* cc) = 0;
    /**
     * Returns the internal cost for the node. These costs accrue for the basic
     * functionality of the node. The total costs for a node are the sum of the
     * internal cost and the cost of the input- and output-jacks.
     * return the internal cost of the node.
     **/

    virtual float getInternalCost() const = 0;
    /**
     * this method signals instream, that a successing node requests
     * some data to do its "GenericNode::analyseData()".
     * \returns DELETE always.
     */

    virtual Result sendData() /* limited(INode::OUTPUT_INITIALIZED) */ /* limited(INode::ACTIVATED) */ /* limited(INode::STARTED) */ /* in-stream */  = 0;
    /**
     * this method signals instream, that a predecessing node could
     * not provide some data for "GenericNode::analyseData()"
     * \returns DELETE always.
     */

    virtual Result sendDataFailed() /* limited(INode::INITIALIZED) */ /* in-stream */  = 0;


  private:
    void registerEvents();
  };

}

#endif
