
#ifndef NMM_IRTPControl_HPP
#define NMM_IRTPControl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>


namespace NMM
{
  /**
   *  @brief RTP Control interface.
   *
   *  This interface provides method for getting and setting options for the RTP protocol.
   */

  class IRTPControl : virtual public Interface { 
  public:
    IRTPControl() {  setTypeInfo(TypeInfo(IRTPControl::IRTPControl_type, "NMM")); };

    IRTPControl(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(IRTPControl::IRTPControl_type, "NMM")); setEventReceiver(parent); };

    virtual ~IRTPControl() {};

    static const char* IRTPControl_type;

    virtual  string getInterfaceType() const { return IRTPControl_type; }

    Interface* copy() const { return new IRTPControl(*this); }
    static const char* setCNAME_event;
    static const char* getCNAME_event;
    static const char* setTTL_event;
    static const char* getTTL_event;
    static const char* setBufferSize_event;
    static const char* getBufferSize_event;
    static const char* setReceiveBufferSize_event;
    static const char* getReceiveBufferSize_event;
    static const char* setMaxFrameSize_event;
    static const char* getMaxFrameSize_event;
    static const char* setSessionBandwidth_event;
    static const char* getSessionBandwidth_event;
    /**
     *  @name Get/Set SDES Items
     */
//@{
    /**
     *  @brief Set canonical name
     *  @param cname The canonical name
     *
     *  This method sets the canonical name of this RTP session. This is a free-form string
     *  which identifies an RTP session. The default is "NMM-RTP".
     */

    virtual void setCNAME(const string& cname)
      throw(const Exception&);
    /**
     *  @brief Get canonical name
     *  @return The canonical name
     */

    virtual string getCNAME() const
      throw(const Exception&);
//@}
    /**
     *  @name Connection Parameters
     */
//@{
    /**
     *  @brief Set multicast TTL
     *  @param ttl The multicast TTL
     *
     *  This method sets the time to live (TTL) for multicast packets. For unicast sessions, this
     *  parameter has no effect. The integer number is the maximum number of hops until a packet
     *  is deleted.
     */

    virtual void setTTL(const unsigned int& ttl)
      throw(const Exception&);
    /**
     *  @brief Get multicast TTL
     *  @return The multicast TTL
     */

    virtual unsigned int getTTL() const
      throw(const Exception&);
    /**
     *  @brief Set buffer size
     *  @param buffer_size The buffer size
     *
     *  This method sets the size for NMM buffers allocated by the RTP receiver.
     */

    virtual void setBufferSize(const size_t& buffer_size)
      throw(const Exception&);
    /**
     *  @brief Get buffer size
     *  @return The buffer size (in bytes)
     */

    virtual size_t getBufferSize() const
      throw(const Exception&);
    /**
     *  @brief Set RTP socket receive buffer size
     *  @param recv_buffer_size Receive buffer size
     *
     *  This method sets the receive buffer size for receiving RTP sockets
     */

    virtual void setReceiveBufferSize(const size_t& recv_buffer_size)
      throw(const Exception&);
    /**
     *  @brief Get RTP socket receive buffer size
     *  @return The receive buffer size
     */

    virtual size_t getReceiveBufferSize() const
      throw(const Exception&);
    /**
     *  @brief Set maximum frame size
     *  @param max_frame_size The maximum frame size (in bytes)
     *
     *  This method sets the maximum size that a received RTP frame may have. If an incoming frame
     *  is larger than this size, it may be truncated.
     */

    virtual void setMaxFrameSize(const size_t& max_frame_size)
      throw(const Exception&);
    /**
     *  @brief Get maximum frame size
     *  @return The maximum frame size (in bytes)
     */

    virtual size_t getMaxFrameSize() const
      throw(const Exception&);
    /**
     *  @brief Set session bandwidth
     *  @param bandwidth The session bandwidth (in kb/s)
     *  @todo Check if this parameter is really in kb/s
     *
     *  This method sets the session bandwidth. Note that this is only an estimate which
     *  serves as a hint to the RTP library. The actual bandwidth is determined by the
     *  data which is transferred (for example the video bitrate).
     */

    virtual void setSessionBandwidth(const unsigned int& buffer_size)
      throw(const Exception&);
    /**
     *  @brief Get session bandwidth
     *  @retun The session bandwidth (in kb/s)
     */

    virtual unsigned int getSessionBandwidth() const
      throw(const Exception&);

//@}

  };

  typedef auto_ptr<IRTPControl> IRTPControl_var;




}

#endif
