
#ifndef NMM_IUDPControl_HPP
#define NMM_IUDPControl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "IIPControl.hpp"

namespace NMM
{
  /** UDP control interface.
   * This interface is provided by the UDPStrategy for setting options and reading
   * statistics for the UDP transport protocol.
   */

  class IUDPControl : virtual public IIPControl { 
  public:
    IUDPControl() {  setTypeInfo(TypeInfo(IUDPControl::IUDPControl_type, "NMM")); };

    IUDPControl(NMMObject& parent) : IIPControl(parent) {  setTypeInfo(TypeInfo(IUDPControl::IUDPControl_type, "NMM")); setEventReceiver(parent); };

    virtual ~IUDPControl() {};

    static const char* IUDPControl_type;

    virtual  string getInterfaceType() const { return IUDPControl_type; }

    Interface* copy() const { return new IUDPControl(*this); }
    static const char* setUsePacketStream_event;
    static const char* getUsePacketStream_event;
    static const char* setPacketSize_event;
    static const char* getPacketSize_event;
    static const char* setMaxPartialMessages_event;
    static const char* getMaxPartialMessages_event;
    static const char* resetStatistics_event;
    static const char* getTotalSentBytes_event;
    static const char* getTotalReceivedBytes_event;
    static const char* getTotalSentPackets_event;
    static const char* getTotalReceivedPackets_event;
    static const char* getTotalDiscardedPackets_event;
    static const char* getTotalLostPackets_event;
    /** @name Packet Stream Options
     */
    /** This method sets whether the UDPStrategy should use the PacketStream.
     * The packet stream guarantees consistent, yet unreliable, delivery of NMM messages.
     */

    virtual void setUsePacketStream(const bool& use_packet_stream)
      throw(const Exception&);
    /** This method gets whether the UDPStrategy should use the PacketStream.
     * The packet stream guarantees consistent, yet unreliable, delivery of NMM messages.
     */

    virtual bool getUsePacketStream() const
      throw(const Exception&);
    /** This method sets the maximum size of a UDP packet.
     */

    virtual void setPacketSize(const ssize_t& size)
      throw(const Exception&);
    /** This method returns the maximum size of a UDP packet.
     */

    virtual ssize_t getPacketSize() const
      throw(const Exception&);
    /** This method sets the maximum number of partial messages in the receive queue.
     * Packets in the receive queue with a message ID less than the value set by this method are
     * thrown away.
     */

    virtual void setMaxPartialMessages(const unsigned int& max_partial)
      throw(const Exception&);
    /** This method returns the maximum number of partial messages in the receive queue.
     * Packets in the receive queue with a message ID less than the value returned by this method are
     * thrown away.
     */

    virtual unsigned int getMaxPartialMessages() const
      throw(const Exception&);
    /** @name Protocol Statistics
     */
    /** This method resets all statistics in the UDPStrategy
     */

    virtual void resetStatistics()
      throw(const Exception&);
    /** This method returns the total amount of bytes sent.
     */

    virtual ssize_t getTotalSentBytes() const
      throw(const Exception&);
    /** This method returns the total amount of bytes received.
     */

    virtual ssize_t getTotalReceivedBytes() const
      throw(const Exception&);
    /** This method returns the total number of packets sent
     */

    virtual unsigned int getTotalSentPackets() const
      throw(const Exception&);
    /** This method returns the total number of packets received
     */

    virtual unsigned int getTotalReceivedPackets() const
      throw(const Exception&);
    /** This method returns the total number of discarded incoming packets
     */

    virtual unsigned int getTotalDiscardedPackets() const
      throw(const Exception&);
    /** This method returns the total number of lost incoming packets
     */

    virtual unsigned int getTotalLostPackets() const
      throw(const Exception&);


  };

  typedef auto_ptr<IUDPControl> IUDPControl_var;




}

#endif
