
#ifndef NMM_IBuffersPassedEventImpl_HPP
#define NMM_IBuffersPassedEventImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "IBuffersPassedEvent.hpp"


namespace NMM
{
  /**
   * Node that sends the number of buffers passed upstream.
   * This interface is for setting the number of buffers after which to send an upstream event
   * containing the number of buffers passed. The event is received by a node that implements
   * IBuffersPassedEventReceiver.
   */

  class IBuffersPassedEventSenderImpl  : public virtual NMMObject { 
  public:
    IBuffersPassedEventSenderImpl() { addInterface(new IBuffersPassedEventSender(*this)); registerEvents(); };

    virtual ~IBuffersPassedEventSenderImpl() {};

  protected:
    /**
     * Set buffer count interval.
     * Sets the number of buffers after which to send the upstream event.
     * Setting it to zero disables sending the event. Buffers processed while the interval is zero
     * are not counted in upstream events.
     */

    virtual void setBufferCountInterval(const int& value) = 0;
    /**
     * Get buffer count interval.
     * Gets the number of buffers after which to send the upstream event.
     * Zero means no upstream events are sent.
     */

    virtual int getBufferCountInterval() const = 0;


  private:
    void registerEvents();
  };
  /**
   * Node that receives events containing the number of buffers passed.
   * These events are sent upstream from a node which implements IBuffersPassedEventSender.
   */

  class IBuffersPassedEventReceiverImpl  : public virtual NMMObject { 
  public:
    IBuffersPassedEventReceiverImpl() { addInterface(new IBuffersPassedEventReceiver(*this)); registerEvents(); };

    virtual ~IBuffersPassedEventReceiverImpl() {};

  protected:
    /**
     * Receive number of buffers processed recently.
     * This is an upstream event. The value is the number of buffers processed since the last event.
     */

    virtual Result buffersPassed(const int& buffers) /* in-stream */  = 0;
    /**
     * Get number of buffers processed below.
     * Returns the number of buffers processed below in the graph, according to the upstream events
     * we received.
     */

    virtual int getUpstreamBuffersPassed() const = 0;
    /**
     * Reset number of buffers processed below.
     */

    virtual void resetUpstreamBufferCounter() = 0;


  private:
    void registerEvents();
  };

}

#endif
