/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef __NMM_IVTV_STRUCTS_HPP
#define __NMM_IVTV_STRUCTS_HPP

#include "nmm/base/serialize/ComplexValue.hpp"

namespace NMM {  
  
  /** IVTV Codec properties.
   * This structure contains all the codec parameters for an IVTV based TV card.
   */
  class IVTVCodec
    : public SerializedValue
  {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default constructor.
     */
    IVTVCodec();
    
    /** Constructor.
     * @param audio Audio bitmask
     * @param bframes Number of B-Frames
     * @param bitrate_mode Bitrate mode
     * @param bitrate Bitrate
     * @param bitrate_peak Bitrate peak
     * @param dnr_mode
     * @param dnr_spatial
     * @param dnr_temporal
     * @param dnr_type
     * @param framerate Framerate type
     * @param frames_per_gop Number of frames per gop
     * @param gop_closure GOP closure mode
     * @param pulldown
     * @param stream_type Stream type
     */
    IVTVCodec(unsigned int audio, unsigned int bframes,
              unsigned int bitrate_mode, unsigned int bitrate, unsigned int bitrate_peak,
              unsigned int dnr_mode, unsigned int dnr_spatial, unsigned int dnr_temporal, unsigned int dnr_type,
              unsigned int framerate, unsigned int frames_per_gop, unsigned int gop_closure,
              unsigned int pulldown, unsigned int stream_type);
    
    /** virtual copy constructor
     */
    virtual IVTVCodec* copy()                     const { return new IVTVCodec(*this); }
    //@}
    
    /** @name Access Methods
     */
    //@{
    /** This method returns the audio bitmask of the codec.
     */
    inline unsigned int  getAudioBitmask()        const { return __audio.getRValue(); }
    
    /** This method returns the number of B-frames of the codec.
     */
    inline unsigned int  getBFrames()             const { return __bframes.getRValue(); }
    
    /** This method returns the bitrate mode of the codec.
     */
    inline unsigned int  getBitrateMode()         const { return __bitrate_mode.getRValue(); }
    
    /** This method returns the bitrate of the codec.
     */
    inline unsigned int  getBitrate()             const { return __bitrate.getRValue(); }
    
    /** This method returns the peak bitrate of the codec.
     */
    inline unsigned int  getBitratePeak()         const { return __bitrate_peak.getRValue(); }
    
    /** This method returns the DNR mode of the codec.
     */
    inline unsigned int  getDNRMode()             const { return __dnr_mode.getRValue(); }
    
    /**
     */
    inline unsigned int  getDNRSpatial()          const { return __dnr_spatial.getRValue(); }
    
    /**
     */
    inline unsigned int  getDNRTemporal()         const { return __dnr_temporal.getRValue(); }
    
    /**
     */
    inline unsigned int  getDNRType()             const { return __dnr_type.getRValue(); }
    
    /** This method returns the framerate ID of the codec
     */
    inline unsigned int  getFramerate()           const { return __framerate.getRValue(); }
    
    /** This method returns number of frames per GOP of the codec
     */
    inline unsigned int  getFramesPerGOP()        const { return __frames_per_gop.getRValue(); }
    
    /** This method returns the GOP closure mode of the codec
     */
    inline unsigned int  getGOPClosure()          const { return __gop_closure.getRValue(); }
    
    /**
     */
    inline unsigned int  getPulldown()            const { return __pulldown.getRValue(); }
    
    /** This method returns the stream type of the codec
     */
    inline unsigned int  getStreamType()          const { return __stream_type.getRValue(); }

    //////////////////////

    /** This method sets the audio bitmask of the codec.
     */
    inline void  setAudioBitmask(unsigned int value)        {__audio.setValue(value); }
    
    /** This method sets the number of B-frames of the codec.
     */
    inline void  setBFrames(unsigned int value)             {__bframes.setValue(value); }
    
    /** This method sets the bitrate mode of the codec.
     */
    inline void  setBitrateMode(unsigned int value)         {__bitrate_mode.setValue(value); }
    
    /** This method sets the bitrate of the codec.
     */
    inline void  setBitrate(unsigned int value)             {__bitrate.setValue(value); }
    
    /** This method sets the peak bitrate of the codec.
     */
    inline void  setBitratePeak(unsigned int value)         {__bitrate_peak.setValue(value); }
    
    /** This method sets the DNR mode of the codec.
     */
    inline void  setDNRMode(unsigned int value)             {__dnr_mode.setValue(value); }
    
    /**
     */
    inline void  setDNRSpatial(unsigned int value)          {__dnr_spatial.setValue(value); }
    
    /**
     */
    inline void  setDNRTemporal(unsigned int value)         {__dnr_temporal.setValue(value); }
    
    /**
     */
    inline void  setDNRType(unsigned int value)             {__dnr_type.setValue(value); }
    
    /** This method sets the framerate ID of the codec
     */
    inline void  setFramerate(unsigned int value)           {__framerate.setValue(value); }
    
    /** This method sets number of frames per GOP of the codec
     */
    inline void  setFramesPerGOP(unsigned int value)        {__frames_per_gop.setValue(value); }
    
    /** This method sets the GOP closure mode of the codec
     */
    inline void  setGOPClosure(unsigned int value)          {__gop_closure.setValue(value); }
    
    /**
     */
    inline void  setPulldown(unsigned int value)            {__pulldown.setValue(value); }
    
    /** This method sets the stream type of the codec
     */
    inline void  setStreamType(unsigned int value)          {__stream_type.setValue(value); }
    //@}
    
  private:
    /** This method adds all attributes by the base class.
     */
    void registerAttributes();
    
  private:
    TValue<unsigned int> __audio;                 ///< Audio bitmask
    TValue<unsigned int> __bframes;               ///< B-Frames
    TValue<unsigned int> __bitrate_mode;          ///< Bitrate mode
    TValue<unsigned int> __bitrate;               ///< Bitrate
    TValue<unsigned int> __bitrate_peak;          ///< Bitrate peak
    TValue<unsigned int> __dnr_mode;              ///< ???
    TValue<unsigned int> __dnr_spatial;           ///< ???
    TValue<unsigned int> __dnr_temporal;          ///< ???
    TValue<unsigned int> __dnr_type;              ///< ???
    TValue<unsigned int> __framerate;             ///< Framerate id
    TValue<unsigned int> __frames_per_gop;        ///< Number of frames per GOP
    TValue<unsigned int> __gop_closure;           ///< Open/Closed GOP
    TValue<unsigned int> __pulldown;
    TValue<unsigned int> __stream_type;
  };
  
} // namespace NMM

#endif // __NMM_IVTV_STRUCTS_HPP
