/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "TVGuide.hpp"

#include "nmm/interfaces/misc/tvguide/ITVGuideUpdateFinished.hpp"

#include "nmm/base/registry/PluginManager.hpp"

#include "nmm/base/ProxyObject.hpp"

#include "nmm/utils/fork/Fork.hpp"
#include "nmm/utils/utils.hpp"


#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/ServiceProxy.hpp"

//for Benchmarks
#include "nmm/base/sync/LocalTime.hpp"
#include "nmm/base/sync/Clock.hpp"


#include <fstream>
#include <unistd.h>
#include <time.h>
#include <regex.h>

namespace NMM{
  TVGuide::TVGuide()
  {
    
    try{
      XMLPlatformUtils::Initialize();
      
    }catch (const XMLException& toCatch){
      GLOBAL_ERROR_STREAM("Error during initialization! :\n"
           << toCatch.getMessage() << XERCES_STD_QUALIFIER endl);
    }
    
    database = new TVGXercescDatabase(this->getInterface<ITVGuide>());
    peerConnector = new TVGuidePeerConnector(this->getInterface<ITVGuide>());

    SFork::getInstance();

    //Load the plugins
    TVGuidePluginManager& manager = STVGuidePluginManager::getInstance();

    try{
      IPluginManager_var pm(SPluginManager::getInstance().getCheckedInterface<IPluginManager>());
      string pluginDir = getNMMPluginDir() + "/tvguide";
      pm->loadExternalPlugins(pluginDir);
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("Error loading TVGuide-Plugins" << endl << e << endl);
    }

    GLOBAL_DEBUG_STREAM("Found " << manager.getPluginList().size() << " Plugins" << endl);

  }//end Constructor

  TVGuide::~TVGuide(){
    XMLPlatformUtils::Terminate();
    try{
      IPluginManager_var pm(SPluginManager::getInstance().getCheckedInterface<IPluginManager>());
      string pluginDir = getNMMPluginDir() + "/tvguide";
      pm->removeExternalPlugins(pluginDir);
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("Error removing TVGuide-Plugins" << endl << e << endl);
    }

  }

  /******************************
   *      Plugin-handling       *
   ****************************** */
  

  ITVGuidePlugin* TVGuide::getGrabberInterface(const string& type){  
    
    TVGuidePluginManager& manager = STVGuidePluginManager::getInstance();

    list<TVGuidePlugin*> pl = manager.getPluginList();


    for(list<TVGuidePlugin*>::iterator i = pl.begin(); i != pl.end(); i++){

      list<Interface*> il = (*i)->getInterfaceList();
      for(list<Interface*>::iterator j = il.begin(); j != il.end(); j++){
        if((*j)->getInterfaceType() == type){
          updatePlugins.push_back(*i);
          return (*j)->getParentObject()->getInterface<ITVGuidePlugin>();
        }
      }      
    }
 
    return NULL;
  }
  
  Result TVGuide::init(){

    //TVGuidePluginManager& manager = STVGuidePluginManager::getInstance();
    //list<TVGuidePlugin*> pl = manager.getPluginList();

    list<TVGuidePlugin*> pl = updatePlugins;
    for(list<TVGuidePlugin*>::iterator i = pl.begin(); i != pl.end(); i++){
      ITVGuidePlugin * plug = (*i)->getInterface<ITVGuidePlugin>();
      try{
        ITVGuideProducer_var tvgp((*i)->getInterface<ITVGuideProducer>());
        if(tvgp.get()){
          //If we can't register a TVGuide-Objekt, there is no reason to start the plugin
          tvgp->registerTVGuideServer(this->getInterface<ITVGuide>());
        }else{
          break;
        }
        plug->init();
      }catch(const Exception &e){
        //if we can't initialize the plugin, we remove it from list
        updatePlugins.remove(*i);
      }
    }

    return SUCCESS;
  }
  
  Result TVGuide::deinit(){
    //TVGuidePluginManager& manager = STVGuidePluginManager::getInstance();    
    //list<TVGuidePlugin*> pl = manager.getPluginList();

    list<TVGuidePlugin*> pl = updatePlugins;
    for(list<TVGuidePlugin*>::iterator i = pl.begin(); i != pl.end(); i++){
      ITVGuidePlugin * plug = (*i)->getInterface<ITVGuidePlugin>();
      try{
        plug->deinit();
      }catch(const Exception &e){
      }
    }

    return SUCCESS;
  }

  Result TVGuide::start(){
    //TVGuidePluginManager& manager = STVGuidePluginManager::getInstance();
    //list<TVGuidePlugin*> pl = manager.getPluginList();
    
    list<TVGuidePlugin*> pl = updatePlugins;

    for(list<TVGuidePlugin*>::iterator i = pl.begin(); i != pl.end(); i++){
      ITVGuidePlugin * plug = (*i)->getInterface<ITVGuidePlugin>();
      //plug->getParentObject()->registerEventListener(ITVGuideUpdateFinished::partiallyFinished_event, &finished_listener);
      try{
        plug->start();
      }catch(const Exception &e){
        GLOBAL_ERROR_STREAM("error starting "  << plug->getName()<< endl);
      }
      
    }

    return SUCCESS;
  }

  Result TVGuide::stop(){
    //TVGuidePluginManager& manager = STVGuidePluginManager::getInstance();
    //list<TVGuidePlugin*> pl = manager.getPluginList();

    list<TVGuidePlugin*> pl = updatePlugins;

    for(list<TVGuidePlugin*>::iterator i = pl.begin(); i != pl.end(); i++){
      ITVGuidePlugin * plug = (*i)->getInterface<ITVGuidePlugin>();
      //plug->getParentObject()->registerEventListener(ITVGuideUpdateFinished::partiallyFinished_event, &finished_listener);
      try{
        plug->stop();
      }catch(const Exception &e){
        GLOBAL_ERROR_STREAM("error stoping " << plug->getName() << e << endl);
      }
      
    }

    /*
    for(list<const ITVGuidePlugin *>::iterator i = updatePlugins.begin(); i != updatePlugins.end(); i++){
      ITVGuidePlugin * p = const_cast<ITVGuidePlugin *>(*i);
      p->stop();
    }
    */
    return SUCCESS;
  }

  /* *************************************
   *            Data-handling            *
   * ************************************* */
  Result TVGuide::insertChannel(const string &id, const string &name){
    Result res = database->insertChannel(id, name);
    try{
      if(res == SUCCESS)
        this->getCheckedInterface<ITVGuideUpdateFinished>()->newChannel();
    }catch(const Exception &e){}
    return res;
  }
  Result TVGuide::insertChannelDescription(const ChannelDescription &cd){
    Result res = database->insertChannelDescription(cd);
    try{
      if(res == SUCCESS)
        this->getCheckedInterface<ITVGuideUpdateFinished>()->newChannel();
    }catch(const Exception &e){}

    return res;
  }
  void TVGuide::setDirectory(const string &path){
    database->setDirectory(path);
  }
  string TVGuide::getDirectory() throw (const NMM::Exception&) {
    return database->getDirectory();
  }
  
  list<string> TVGuide::getCategories() throw (const NMM::Exception&) {
    return database->getCategories();
  }
  
  list<ChannelDescription> TVGuide::getChannelsList()throw (const NMM::Exception&) {
    peerConnector->getChannels();
    return database->getChannelsList();
  }
  string  TVGuide::getChannelById(const string &id){
    string result = database->getChannelById(id);
    /*
    if(result.empty()){
      result = peerConnector->getChannelById(id);
    }
    */
    return result;
  }
  string TVGuide::getChannelByName(const string &name){
    return database->getChannelByName(name);
  }
  
  Result TVGuide::insertEvent(const ProgramDescription &pd){
    return database->insertEvent(pd);
  }
  ProgramDescription TVGuide::getEvent(const string &start, const string &end, const string &id){
    return database->getEvent(start, end, id);
  }
  list<ProgramDescription> TVGuide::getEventsOnDayOnChannel(const string &day, const string &chan){
     list<ProgramDescription> result  =  database->getEventsOnDayOnChannel(day, chan);
     peerConnector->verify(day, chan, result);
     return result;
  }
  list<ProgramDescription> TVGuide::search(const ProgramFilter &pf, const ITVGuidePeerObject* peer){

    Clock clock;
    Time t0 = clock.getTime();
    list<ProgramDescription> result;
    result = database->search(pf);
    Time t1 = clock.getTime();
    cerr << "Time local search: " << t1 - t0 << endl;

    if(peer != NULL){
      
      list<Interface*> peers;
      peers.push_back(dynamic_cast<ITVGuidePeerObject*>(peer->copy()));
      peerConnector->search(pf, peers);
    }

    return result;
  }
  
  void TVGuide::addExtDescription(const string& start, const string& end, const string& id, const string& desc){
    database->addExtDescription(start, end, id, desc);
  }
  
  ProgramDescription TVGuide::getCurrentEventOnChannel(const string& channel){
    ProgramDescription res = database->getCurrentEventOnChannel(channel);
    if(res.empty()){
      peerConnector->getCurrentEventOnChannel(channel);
    }
    return res;
  }
  ProgramDescription TVGuide::getFollowingEventOnChannel(const string& channel){
    ProgramDescription res = database->getFollowingEventOnChannel(channel);
    if(res.empty()){
      peerConnector->getFollowingEventOnChannel(channel);
    }
    return res;
  }
  
  void TVGuide::cleanDatabase(const int &days){
    string directory = database->getDirectory();

    DirReader dir;
    dir.setRootDir(directory);
    dir.changeDir("");

    vector <DirReader::directory_t> subdirs = dir.getDirsOnlyEntries();
    for(vector<DirReader::directory_t>::iterator i = subdirs.begin(); i != subdirs.end(); i++){
      if(i->name == "." || i->name == "..") continue;

      dir.changeDir(i->name);
      vector<DirReader::directory_t> files = dir.getRegFiles();
      for(vector<DirReader::directory_t>::iterator j = files.begin(); j != files.end(); j++){
        time_t now = time(NULL);
        now -= days * 24 * 60 * 60;
        string filedate = j->name.substr(0, j->name.find_first_of('.'));
        
        if(stringToTime(filedate) < now){
          string command = "rm " + j->path + "/" + j->name;
          //system(command.c_str());
          //cerr << command << endl;
        }

      }//for j
      dir.changeDir("..");
    }//end for i
  }


  void TVGuide::addPeerHost(const string& host, const int& port){
    TCPAddress a(host, port);
    peerConnector->addHost(a);
  }
  void TVGuide::clearPeerHosts(){
    peerConnector->clearHosts();
  }

  Result TVGuide::peerSearch(const ProgramFilter& filter, 
                             const list<Interface*>& history){

    list<ProgramDescription> result;

    Clock clock;
    Time t0 = clock.getTime();
    result = database->search(filter);
    Time t1 = clock.getTime();
    cerr << "Time peerSearch: " << t1 -t0 << endl;

    //notify the last peer with our search result, if not empty() ...
    if(!result.empty()){
      list<Interface*> newHistory = history;
      ITVGuidePeerObject *peer = dynamic_cast<ITVGuidePeerObject*>(newHistory.back());
      newHistory.pop_back();
      peer->peerResult(result, filter.getID(), newHistory);
    }
       
    //... and propagate the request to the next peer
    peerConnector->search(filter, history);  

    return SUCCESS;
  }
  
  Result TVGuide::peerChannelSearch(const list<Interface*>& history){

    try{
    list<ChannelDescription> result = database->getChannelsList();

    //notify the last peer with our search result, if not empty() ...
    if(!result.empty()){
      list<Interface*> newHistory = history;
      ITVGuidePeerObject *peer = dynamic_cast<ITVGuidePeerObject*>(newHistory.back());
      newHistory.pop_back();
      peer->peerChannelResult(result, newHistory);
    }

    //... and propagate the request to the next peer
    peerConnector->getChannels(history);

    }catch(const Exception &e){cerr << e << endl;}
    return SUCCESS;

  }

}
