/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "XMLTVPlugin.hpp"

#include <fstream>
#include <stdio.h>

#include "nmm/interfaces/misc/tvguide/ITVGuideUpdateFinished.hpp"

#include "nmm/utils/SecurityCheck.hpp"

#include "nmm/utils/utils.hpp"

#include "nmm/utils/fork/Fork.hpp"

#include "nmm/base/registry/ReferenceManager.hpp"

//Xerces-C-Header
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMImplementationRegistry.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/dom/DOMError.hpp>
#include <xercesc/dom/DOMLocator.hpp>
#include <xercesc/dom/DOMNamedNodeMap.hpp>
#include <xercesc/dom/DOMAttr.hpp>
#include <xercesc/parsers/AbstractDOMParser.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>

namespace NMM{

  XMLTVPlugin* xmltvplugin = new XMLTVPlugin();

  XMLTVPlugin::XMLTVPlugin() :
    TVGuidePlugin("XMLTVPlugin"),
    filename("/tmp/xmltv.tmp"),
    locale("de"),
    language("de"),
    lastUpdateDate(""),
    lastUpdateChannel(""),
    termUpdate(false),
    temporaryConfigFile("/tmp/xmltv.conf"),
    fillHolesInDB(false),
    period("d1"),
    update_timer(this, &XMLTVPlugin::restartUpdate)
  {
    configFile = getNMMResource("tvguide") + "/XMLTVPluginConf.xml";
    readPluginConfig();
  }


  XMLTVPlugin::~XMLTVPlugin(){
    termUpdate = true;
  }

  /* *********** IXMLTVPlugin-Interface **************** */
  void XMLTVPlugin::setFilename(const string &file){ 
    filename = file;
  }
  string XMLTVPlugin::getFilename(){ 
    return filename;
  }
    
  void XMLTVPlugin::setConfigFile(const string &file){
    configFile = file;
    readPluginConfig();
  }
  string XMLTVPlugin::getConfigFile(){
    return configFile;
  }

  Result XMLTVPlugin::setLocale(const string &loc){
    for(list<pair<string,string> >::iterator i = grabber.begin(); i != grabber.end(); i++){
      if(loc == i->first || loc == i->second){
        locale = loc;
        return SUCCESS;
      }
    }
    GLOBAL_ERROR_STREAM("grabber " << loc << "not supported" << endl);
    locale = loc;
    return FAILURE;
  }
  string XMLTVPlugin::getLocale(){
    return locale;
  }

  void XMLTVPlugin::setPeriod(const string &p){
    period = p;
  }
  string XMLTVPlugin::getPeriod(){
    return period;
  }

  void XMLTVPlugin::setFillHolesInDB(const bool &b){
    fillHolesInDB = b;
  }
  bool XMLTVPlugin::getFillHolesInDB(){
    return fillHolesInDB;
  }


  /* **************** ITVGuidePlugin-Interface **************** */
  Result XMLTVPlugin::doInit(){
    
    try{
      //setup timerproducer for triggered updates
      timer = new TimerProducer();
      timer->connectTo(this); 
      timer->start();
      
      //remove unavailable grabber from list
      for(list<pair<string,string> >::iterator i = grabber.begin(); i != grabber.end(); i++){
        string testCommand = i->second + " -h >/dev/null";
        if(SFork::getInstance().executeCommand(testCommand.c_str()) != 0){  //Test, if The grabber can be executed. if no, remove it from list.
          i->second=""; //blank the command
        }
      }
      
      static const XMLCh gLS[] = { chLatin_L, chLatin_S, chNull };
      impl = DOMImplementationRegistry::getDOMImplementation(gLS);
      parser = NULL;
      
      transcoder = new XercesTranscoder();
      
      registerEvent("NEXTXMLTVUPDATE", &update_timer);
      return SUCCESS;
    }catch(const Exception &e){
      return FAILURE;
    }catch(...){
      return FAILURE;
    }
  }
  Result XMLTVPlugin::doDeinit(){
    //timer->stop();
    //timer->disconnect();
    //removeEvent("NEXTXMLTVUPDATE");
    return SUCCESS;
  }


  Result XMLTVPlugin::doStop(){
    GLOBAL_DEBUG_STREAM("TERMINATING " << locale << endl);
    termUpdate=true;

    for(list<pair<string,string> >::iterator i = grabber.begin(); i != grabber.end(); i++){
      if(locale == i->first || locale == i->second){
        string s = "killall -9 " + i->second + " 2>/dev/null";
        SFork::getInstance().executeCommand(s.c_str());
      } 
   }

    //timer->stop();
    writePluginConfig();
    return SUCCESS;
  }
  
  Result XMLTVPlugin::doStart(){
    termUpdate = false;   
    
    string command;
    //select grabber
    for(list<pair<string,string> >::iterator i = grabber.begin(); i != grabber.end(); i++){
      if(locale == i->first || locale == i->second){
        language = i->first;
        command = i->second;
      }
    }
    if(command.empty()){
      GLOBAL_ERROR_STREAM("Grabber " << locale << " " << command << " not supported! Please see documentation for supported grabbers" << endl);
      return FAILURE;
    }
      

    GLOBAL_DEBUG_STREAM("using grabber " << command << endl);
    string path = SProcessUserInfo::getInstance().getHomeDirectory() + string("/.xmltv/") + command + ".conf";
    bool res = readXMLTVConfig(path);
    GLOBAL_DEBUG_STREAM("result of reading configfile: " << res << " With " <<  xmltvChannels.size() << " channels" << endl);

    //Claculate the beginning of update (DAY)
    int updateOffset = 0;
    if(!lastUpdateDate.empty()){
      updateOffset = (stringToTime(lastUpdateDate) - time(NULL)) / 60 / 60 / 24;
      GLOBAL_DEBUG_STREAM(lastUpdateDate << " Using updateOffset " << updateOffset << endl);
      if(!fillHolesInDB && (updateOffset < 0)){ //do not Fill Holes in DB, when last update lies in past
        updateOffset=0;
      }
    }else{
      char tmpTime[13];
      time_t t = time(NULL);
      strftime(tmpTime, sizeof(tmpTime), "%Y%m%d0000", localtime(&t));
      lastUpdateDate = string(tmpTime);
    }

    //Determine the last channel grabbed
    list<string>::iterator xmltvChannelsOffset = xmltvChannels.begin();
    if(!lastUpdateChannel.empty()){
      if(lastUpdateChannel != xmltvChannels.back()){
        while(xmltvChannelsOffset != xmltvChannels.end()){
          if(*xmltvChannelsOffset == lastUpdateChannel){
            xmltvChannelsOffset++;
            break;
          }
          xmltvChannelsOffset++;
        }
        //if the last channel grabber can't be found, start at front
        if(xmltvChannelsOffset == xmltvChannels.end()) xmltvChannelsOffset = xmltvChannels.begin();
      }
    }

    
    //now, we determined the offsets and can start grabbing
    for(int i = updateOffset; i < 7; i++){

      string options = "--days 1 --quiet";
      
      char tmp[3];
      sprintf(tmp, "%i", i);
      options += " --offset " + string(tmp);

      if(!xmltvChannels.empty()){
        for(list<string>::iterator j = xmltvChannelsOffset; j != xmltvChannels.end(); j ++){
          if(termUpdate) return SUCCESS;
          
          ofstream outfile(temporaryConfigFile.c_str());
          if(!outfile.fail()){
            outfile << "channel " << *j << endl;
          }
          outfile.close();
          
          string grabberCommand = command + " " + options + " -config-file " + temporaryConfigFile + "  > " + filename;
          GLOBAL_DEBUG_STREAM("Fetch data with command: " << grabberCommand << endl);
          
          int systemRes = SFork::getInstance().executeCommand(grabberCommand.c_str());
          GLOBAL_DEBUG_STREAM("result of system-call: " << systemRes << endl);
          
          if(systemRes == 2) return SUCCESS; //Strg-C is pressed
          
          if(systemRes == 0){
            systemRes = readXMLTV(filename);
            GLOBAL_DEBUG_STREAM("Result of reading xmlfile: " << systemRes << endl);
            if(systemRes == 0){
              lastUpdateChannel = *j;
              writePluginConfig();
            }
          }else{
            GLOBAL_ERROR_STREAM("Can't execute grabber " << command << endl);
            return FAILURE;
          }
        }
      
        xmltvChannelsOffset = xmltvChannels.begin(); //now, all channels from this day should be grabbed, so start at front
      }else{
          string grabberCommand = command + " " + options + "  > " + filename;
          GLOBAL_DEBUG_STREAM("Fetch data with command: " << grabberCommand << endl);
          
          int systemRes = SFork::getInstance().executeCommand(grabberCommand.c_str());
          GLOBAL_DEBUG_STREAM("result of system-call: " << systemRes << endl);
          
          if(systemRes == 2) return SUCCESS; //Strg-C is pressed
          
          if(systemRes == 0){
            systemRes = readXMLTV(filename);
            GLOBAL_DEBUG_STREAM("Result of reading xmlfile: " << systemRes << endl);
            if(systemRes == 0){
              writePluginConfig();
            }
          }else{
            GLOBAL_ERROR_STREAM("Can't execute grabber " << command << endl);
            return FAILURE;
          }

      }

      //store the date of last update
      char tmpTime[13];
      time_t t = stringToTime(lastUpdateDate) + 1 * 60 * 60 * 24; //lastUpdate now 1 day later
      strftime(tmpTime, sizeof(tmpTime), "%Y%m%d0000", localtime(&t));
      lastUpdateDate=string(tmpTime);
      writePluginConfig();
    }
    
    //since all the available data is grabbed, we setup a timer to start the next update in the future
    //Setup period
    //compute time for an update
    int updateInterval = 0;

    if(termUpdate) return SUCCESS;
   
    try{
      if(period != ""){
        updateInterval = atoi((period.substr(1).c_str()));
        if(period.substr(0,1) == "d"){
          //days
          updateInterval *= 3600*24;
        }else if (period.substr(0,1) == "h"){
          //hours
          updateInterval*=3600;
        }else if (period.substr(0,1) == "m"){
          //minutes
          updateInterval*=60;
        }else if (period.substr(0,1) == "s"){
          //seconds
        }else{
          //default is days
          updateInterval *= 3600*24;
          GLOBAL_ERROR_STREAM("period-String empty. This should not hapen, since period is initialized with d1" << endl);
        }
      }//end if interval!=""
    }catch(const Exception &e){
      updateInterval *= 3600*24;
      GLOBAL_ERROR_STREAM("Error parsing period" << e << endl);
    }
    

    GLOBAL_DEBUG_STREAM("period: " << period << " updateInterval: " << updateInterval << endl);
    if(termUpdate) return SUCCESS;

    Event* e = new Event("NEXTXMLTVUPDATE");
    timer->registerTimer(time(NULL)+updateInterval, e);


    return stop();
  }  

  /*******************************
   *   EventListener-Function    *
   ******************************* */

  Result XMLTVPlugin::restartUpdate(){
    GLOBAL_DEBUG_STREAM("restart XMLTV-Update" << endl);
    start(); 
    return SUCCESS;
  }

  /* ****************************************
   *           Private Functions            *                     
   ****************************************** */

  /** read XML-Files **/
  int XMLTVPlugin::readXMLTV(const string& file){
    if(termUpdate) return 0;
    
    if(!tvgobject.get()) {
      GLOBAL_ERROR_STREAM("No TVGuide-Object available" << endl);
      return -1;
    }

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(file)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << file << ", check .nmmrc for options" << endl);
      return -1;
    }

    try{
      XMLPlatformUtils::Initialize();
    }catch (const XMLException& toCatch){
      GLOBAL_ERROR_STREAM("Error during initialization! :\n"
           << toCatch.getMessage() << endl);
    }catch(...){
      GLOBAL_ERROR_STREAM("strange things are going on" << endl);

    }

    //read the source-file and build the XML-Tree
    DOMDocument *doc = NULL;
    try{
      parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
      parser->setProperty(XMLUni::fgXercesScannerName, (void*)XMLUni::fgWFXMLScanner);
      parser->resetDocumentPool();
      doc = parser->parseURI(file.c_str());
      if(doc){
        DOMNode *root = dynamic_cast<DOMNode*>(doc->getDocumentElement());
        if(root){
          map<string, string> channelMapping;
          DOMNode* text = NULL; //DOMNode to evalute text-content
          DOMNamedNodeMap *pAttributes = NULL; //attribute-List of a node
          DOMAttr *pAttributeNode = NULL; //1 attribute of a node

          //process channels
          for(DOMNode* child = root->getFirstChild(); child != NULL; child = child->getNextSibling()){
            if(termUpdate) {
              XMLPlatformUtils::Terminate(); 
              return 0;
            }
            if(child->getNodeType() == DOMNode::ELEMENT_NODE){
              char *nodename = XMLString::transcode(child->getNodeName());
              if(XMLString::compareString(nodename, "channel") == 0){
                bool insertChannel = true;
                // get all the attributes of the node
                pAttributes = child->getAttributes();
                int nSize = pAttributes->getLength();
                if(nSize > 0){
                  pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("id"));
                  // get attribute name
                  if(pAttributeNode){
                    char *value = XMLString::transcode(pAttributeNode->getValue());
                    string tmpName = tvgobject->getChannelById(string(value)); //to search DB only once
                    if(tmpName != ""){
                      //channelId already exists in database
                      channelMapping[string(value)] = string(value);
                      continue;
                    }
                    for (DOMNode* i = child->getFirstChild(); i != NULL; i=i->getNextSibling()){
                      if (i->getNodeType() == DOMNode::ELEMENT_NODE){
                        char *displayname = XMLString::transcode(i->getNodeName());
                        if(XMLString::compareString(displayname, "display-name") == 0){
                          //test, if a channel with this name already exists
                          string tmpId = tvgobject->getChannelByName(string(XMLString::transcode(i->getTextContent())));//to search DB only once
                          if(tmpId != ""){
                            channelMapping[string(value)] = tmpId;
                            insertChannel = false;
                            continue;
                          }
                          XMLString::release(&displayname);  
                        }
                        
                        //if we come so far, the channel does not exist in the DB and we have to add it
                        if(insertChannel){
                          string s = string(XMLString::transcode(i->getTextContent()));
                          tvgobject->insertChannel(string(value),s);
                          channelMapping[string(value)] = string(value);
                        }
                      }   
                    }
                    XMLString::release(&value);
                  }
                }
              }else if(XMLString::compareString(nodename, "programme") == 0){
                ProgramDescription *pd = new ProgramDescription(ProgramDescription::XMLTV);
                if(child->hasAttributes()) {
                  // get all the attributes of the node
                  pAttributes = child->getAttributes();
                  int nSize = pAttributes->getLength();
                  if(nSize > 0){
                    pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("start"));
                    char *name = NULL;
                    if(pAttributeNode){
                      name = XMLString::transcode(pAttributeNode->getValue());
                      //XERCES_STD_QUALIFIER cout << "  setStartTime(" << name << ")" << endl;;
                      pd->setStartTime(string(name));
                      XMLString::release(&name);
                    }
                    
                    pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("stop"));
                    if(pAttributeNode){
                      name = XMLString::transcode(pAttributeNode->getValue());
                      //XERCES_STD_QUALIFIER cout << "  setEndTime  (" << name << ")" << endl;;
                      pd->setEndTime(string(name));
                      XMLString::release(&name);
                    }
                    
                    pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("channel"));
                    if(pAttributeNode){
                      name = XMLString::transcode(pAttributeNode->getValue());
                      //XERCES_STD_QUALIFIER cout << "  setChannelID(" << name << ")" << endl;;
                      pd->setChannelID(channelMapping[string(name)]);
                      XMLString::release(&name);
                    }
                    
                    pAttributeNode = (DOMAttr*) pAttributes->getNamedItem(XMLString::transcode("duration"));
                    if(pAttributeNode){
                      name = XMLString::transcode(pAttributeNode->getValue());
                      //XERCES_STD_QUALIFIER cout << "  setDuration(" << name << ")" << endl;;
                      pd->setDuration(atoi(name));
                      XMLString::release(&name);
                    }
                    
                    for (DOMNode* i = child->getFirstChild(); i != NULL; i=i->getNextSibling()){
                      if (child->getNodeType() == DOMNode::ELEMENT_NODE){
                        name = XMLString::transcode(i->getNodeName());
                        
                        if(XMLString::compareString(name, "title") == 0){
                          //evaluate title-tag
                          text = i->getFirstChild();
                          if(text){
                            if(text->getNodeType() == DOMNode::TEXT_NODE){
                              //XERCES_STD_QUALIFIER cout << endl << "  addTitle(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                              pd->addTitle(transcoder->transcodeTo(text->getTextContent()));
                            }
                          }
                        }

                        if(XMLString::compareString(name, "sub-title") == 0){
                          //evaluate title-tag
                          text = i->getFirstChild();
                          if(text){
                            if(text->getNodeType() == DOMNode::TEXT_NODE){
                              //XERCES_STD_QUALIFIER cout << endl << "  addTitle(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                              pd->addTitle(transcoder->transcodeTo(text->getTextContent()));
                            }
                          }
                        }
                        
                        if(XMLString::compareString(name, "desc") == 0){
                          //evaluate description              
                          text = i->getFirstChild();
                          if(text){
                            if(text->getNodeType() == DOMNode::TEXT_NODE){
                              //XERCES_STD_QUALIFIER cout << "  addDescription(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                              pd->setDescription(transcoder->transcodeTo(text->getTextContent()));
                            }
                          }
                          
                        }
                        
                        if(XMLString::compareString(name, "category") == 0){              
                          //evaluate description              
                          text = i->getFirstChild();
                          if(text){
                            if(text->getNodeType() == DOMNode::TEXT_NODE){
                              pd->addCategory(transcoder->transcodeTo(text->getTextContent()));
                              //XERCES_STD_QUALIFIER cout << "  addCategory(" << XMLString::transcode(text->getTextContent()) << ")" << endl;
                            }
                          }
                        }
                        if(XMLString::compareString(name, "duration") == 0){
                          //evaluate duration
                          DOMNamedNodeMap *childAttributes = i->getAttributes();
                          int childSize = childAttributes->getLength();
                          if(childSize > 0){
                            DOMAttr *childAttributeNode =
                              (DOMAttr*) childAttributes->getNamedItem(XMLString::transcode("unit"));
                            if(childAttributeNode){
                              char *lang = xercesc::XMLString::transcode(childAttributeNode->getValue());
                              //XERCES_STD_QUALIFIER cout << "    lang=" << lang << endl;
                              XMLString::release(&lang);
                            }
                          }
                          text = i->getFirstChild();
                          if(text){
                            if(text->getNodeType() == DOMNode::TEXT_NODE){
                              char *duration = XMLString::transcode(text->getNodeValue());
                              //XERCES_STD_QUALIFIER cout << "  addDuration(" << duration << ")" << endl;
                              pd->setDuration(atoi(duration));
                              XMLString::release(&duration);
                            }
                          }
                        }
                        XMLString::release(&nodename);
                      }
                    }
                  }
                }
                //insert event to TVGuide-Object
                tvgobject->insertEvent(*pd);
                delete pd;
                usleep(500);
              }              
            }
            if(termUpdate) {
              XMLPlatformUtils::Terminate(); 
              return 0;
            }
          }
        }//if(root)
      }//end if(doc)

      parser->release();
    }catch(const XMLException &e){
      GLOBAL_ERROR_STREAM("XmlTVPlugin: saveCategories: XMLException: " << XMLString::transcode(e.getMessage()) << endl);
    }catch(const DOMException &e){
       GLOBAL_ERROR_STREAM("XmlTVPlugin: saveCategories: DOMException: " << XMLString::transcode(e.msg) << endl);
    }catch (...){
       GLOBAL_ERROR_STREAM("XmlTVPlugin: saveCategories: general Exception: " << endl);
    }

    if(!termUpdate) termUpdate=false;
    XMLPlatformUtils::Terminate(); 

    return 0;

  }
  
  bool XMLTVPlugin::readXMLTVConfig(string file){

    ifstream infile(file.c_str());
    if(!infile) return false;
  
    string tmp;
    char line[101];

    while(!infile.eof()){
      
      infile >> tmp;
      
      //do not parse comments
      if(tmp[0] == '#') {
        char c;
        c = infile.get();
        //discards chars till end-of-line
        while(c != '\n' && !infile.eof())
          c = infile.get();        
        continue;
      }
      
      if(tmp == "channel"){
        infile.getline(line, 100);
        if(string(line).empty()){
        }else{
          xmltvChannels.push_back(string(line));
        }
	
      }
    }

    return true;
  }

  void XMLTVPlugin::readPluginConfig(){
    GLOBAL_DEBUG_STREAM("Reading XMLTVPlugin-Config " << configFile << endl);


    XMLTree* xml  = new XMLTree(configFile);
    XMLNode* root = NULL;
    if(xml) {
      root = xml->root();
    }else{
      GLOBAL_ERROR_STREAM("Unable to open XMLTVPlugon-Config-File " << configFile << endl);
      return;
    }

    if(root){
      //get lastUpdate
      XMLNodeList lastUpdate = root->children("lastUpdate");
      if(!lastUpdate.empty()){
        XMLNode* update = lastUpdate.front();
        XMLProperty* dateProp = update->property("date");
        if(dateProp) lastUpdateDate = dateProp->value();
        XMLProperty* chanProp = update->property("channel");
        if(chanProp) lastUpdateChannel = chanProp->value();
      }
      //get period
      XMLNodeList periodList = root->children("period");
      if(!periodList.empty()){
        XMLNode* periodNode = periodList.front();
        XMLProperty* periodProp = periodNode->property("data");
        if(periodProp) period = periodProp->value();
      }

      //get grabber
      XMLNodeList grabberList = root->children("grabber");
      for(XMLNodeIterator i = grabberList.begin(); i != grabberList.end(); i++){
        XMLNode* grab = *i;
        XMLProperty* grabProp = grab->property("short");
        if(grabProp){
          bool addGrabber = true;
          for(list<pair<string, string> >::iterator g = grabber.begin(); g != grabber.end(); g++){
            if((g->first == grabProp->value()) && (g->second == grab->getData())){
              addGrabber = false;
            }
          }
          if(addGrabber){ //add the grabber only if not already in list
            grabber.push_back(pair<string, string>(grabProp->value(), grab->getData()));
          }
        }
      }

      //get fillHolesInDB
      XMLNodeList holeList = root->children("fillHolesInDB");
      if(!holeList.empty()){
        XMLNode* holeNode = holeList.front();
        XMLProperty* holeProp = holeNode->property("value");
        if(holeProp){
          if(holeProp->value() == "true"){
            fillHolesInDB = true;
          }else{
            fillHolesInDB = false;
          }
        }
      }

    }else{
      GLOBAL_ERROR_STREAM("Unable to open XMLTVPlugin-Config-File " << configFile << endl);
      return;
    }
   
    delete xml;

  }

  void XMLTVPlugin::writePluginConfig(){
    GLOBAL_DEBUG_STREAM("Writing XMLTVPlugin-Config" << endl);
    
    ofstream outfile(configFile.c_str());
    if(!outfile.fail()){

      outfile << "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>" << endl;
      //outfile << "<!DOCTYPE xmltvplugin SYSTEM \"xmltv.dtd\">" << endl;

      outfile << "<xmltvplugin>" << endl;
      outfile << "<lastUpdate ";
      if(!lastUpdateDate.empty()){
        outfile << "date=\"" << lastUpdateDate << "\" ";
      }
      if(!lastUpdateChannel.empty()){
        outfile << "channel=\"" << lastUpdateChannel << "\"";
      }
      outfile << "> </lastUpdate>" << endl;
      outfile << "<period data=\"" << period << "\"/>" << endl;
      
      for(list<pair<string,string> >::iterator i = grabber.begin(); i != grabber.end(); i++){
        if((!i->first.empty()) && (!i->second.empty())){
          outfile << "<grabber short=\"" << i->first << "\">" << i->second << "</grabber>" << endl;
        }
      }

      if(fillHolesInDB){
        outfile << "<fillHolesInDB value=\"true\"></fillHolesInDB>" << endl;
      }else{
        outfile << "<fillHolesInDB value=\"false\"></fillHolesInDB>" << endl;
      }
      outfile << "</xmltvplugin>" << endl;
    }else{
      GLOBAL_ERROR_STREAM("Could not open XMLTVPlugin-Configuration-File " << configFile << endl);
    }
    outfile.close();
  }

}
