          /* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TVManager.hpp"

#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/interfaces/video/tv/ITV.hpp"
#include "nmm/interfaces/video/tv/IChannelInfo.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/base/IReferenceCount.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuideProducer.hpp"

#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

//to disonnect nodes
#include "nmm/base/connect.hpp"


namespace NMM{
  TVManager::TVManager(IUsage::usage_t u, string filename)
    : usageNotifier(NULL),
      myCurrentUsage(u),
      application(NULL),
      currentChannel(0),
      channelConfig((string(getenv("HOME")) + "/.nmm/tv_channels")),
      currentTVCard(NULL),
      currentTVCardURL(""),
      graph(NULL),
      source(NULL),
      currentChannelOnCurrentTVCard(0),
      video_sink(NULL),
      audio_sink(NULL)
  {
    if(filename.empty()){
      channelConfig = ((string(getenv("HOME")) + "/.nmm/tv_channels"));
    }else{
      channelConfig = filename;
    }
  }

  TVManager::TVManager(const IUsageNotify* usageNotify, IUsage::usage_t u)
    : usageNotifier(usageNotify),
      myCurrentUsage(u),
      application(NULL),
      currentChannel(0),
      channelConfig((string(getenv("HOME")) + "/.nmm/tv_channels")),
      currentTVCard(NULL),
      currentTVCardURL(""),
      graph(NULL),
      source(NULL),
      currentChannelOnCurrentTVCard(0),
      video_sink(NULL),
      audio_sink(NULL)
  {}

  TVManager::TVManager(const IUsageNotify* usageNotify, IUsage::usage_t u, string filename)
    : usageNotifier(usageNotify),
      myCurrentUsage(u),
      application(NULL),
      currentChannel(0),
      channelConfig(filename),
      currentTVCard(NULL),
      currentTVCardURL(""),
      graph(NULL),
      source(NULL),
      currentChannelOnCurrentTVCard(0)
  {}

  TVManager::~TVManager(){
    //    if(currentTVCard) delete currentTVCard;
    if(backupGraph) delete backupGraph;
  }

  void TVManager::setApplication(NMMApplication* app){
    application = app;
  }


  Result TVManager::setChannel(const unsigned int &channel){

    //request the desired channel
    TVChannel c;
    try{
      c = channelConfig.getChannelByNumber(channel);
    }catch(const Exception &e){
      return FAILURE;
    }

    //get TVCards supporting rewuested channel
    TVChannel::cardList_t cards = c.getCards();
    TVChannel::cardList_t::iterator i = cards.begin();


    //if we are Master and the channel is available on the current Card
    if((i = cards.find(currentTVCardURL)) != cards.end()){
      if(isMaster(currentTVCard)){
        if(currentTVCard){ //Just for safety. If the pointer is NULL, we are not Master
          ITVCard_var tvcard(currentTVCard->getParentObject()->getInterface<ITVCard>());
          if(tvcard.get()){
            try{
              if(tvcard->setChannel(i->second) == SUCCESS){
                currentChannelOnCurrentTVCard = i->second;
                currentChannel = channel;
                return SUCCESS;
              }else{
                return FAILURE;
              }
            }catch(const Exception &e){
            }catch(...){
            }
          }
        }
        
      }else{
        //check for the tuned channel. if it is not our channel, we can discard this card
        int cNr; string cName;
        ITVCard_var tvcard(currentTVCard->getParentObject()->getInterface<ITVCard>());
        tvcard->getChannel(cNr, cName);
        //is the requested channel tuned in?
        if(cNr != (int)i->second){
          cards.erase(cards.find(currentTVCardURL));
        }
      }
    }
    //we can't use the current card so we have to request a new one

    try{
      ITVManager_var itvmanager(this->getCheckedInterface<ITVManager>());
      if(itvmanager.get()) itvmanager->switchingSource();
    }catch(const Exception &e){}

    requestedGDs.clear();

    int channelToSet = 0;

    if(requestCardExclusive(cards, channelToSet) == FAILURE) {
      //Ok, we can't get a card EXCLUSIVE, so we try to find a shared one
      if(requestCardShared() == FAILURE) {

        requestedGDs.clear();    

        return FAILURE;
      }else{
        try{
          ITVCard_var itvcard(currentTVCard->getParentObject()->getCheckedInterface<ITVCard>());
          int c; string s;
          itvcard->getChannel(c, s);
          currentChannelOnCurrentTVCard = c;
        }catch(const Exception &e){}
      }
    }


    currentChannel = channel;
    return SUCCESS;
  }
  
  void TVManager::setGraphDescription(GraphDescription *gd, NodeDescription *n){
    graph=gd;
    backupGraph = new GraphDescription(*gd);
    source = n;
  }

  void TVManager::releaseGraphDescription(){

    if(currentTVCard){
      IUsage_var use(currentTVCard->getParentObject()->getInterface<IUsage>());
      if(use.get()){
        use->unsetUsage(myCurrentUsage, usageNotifier);
      }
    }

    if(graph){
      graph=NULL;
    }
    
    currentTVCard   = NULL;
    currentChannel  = 0;
    currentTVCardURL = "";
  }

  Result TVManager::channelUp(){
    unsigned int c = currentChannel; //for counting
    unsigned int ccBackup = currentChannel; //for detecting noChannelFound

    cerr << " channel up" << endl;

    if((int)c < channelConfig.size()-1){
      c++;
    }else{
      c = 0;
    }

    Result res = setChannel(c);

    if(res == FAILURE){
      try{
        ITVManager_var itvmanager(this->getCheckedInterface<ITVManager>());
        if(itvmanager.get()) itvmanager->skippingChannel();
      }catch(const Exception &e){}
      cerr << "FAILURE" << endl;
    }else{
      cerr << " SUCCESS" << endl;
      return SUCCESS;
    }

    while((res == FAILURE) && (c != currentChannel)){
      if((int)c < channelConfig.size()){
        c++;
      }else{
        c = 0;
      }

      if(setChannel(c) == SUCCESS) {
        if(c == ccBackup){
          return FAILURE;
        }
        return SUCCESS;
        
      }
    }//end while

    return FAILURE;
  }

  Result TVManager::channelDown(){
    unsigned int c = currentChannel; //for counting
    unsigned int ccBackup = currentChannel; //for detecting noChannelFound

    if( c > 0){
      c--;
    }else{
      if(!channelConfig.empty()){
        c = channelConfig.size()-1;
      }else{
        c = 0;
      }
    }
    
    Result res = setChannel(c);

    if(res == FAILURE){
      try{
        ITVManager_var itvmanager(this->getCheckedInterface<ITVManager>());
        if(itvmanager.get()) itvmanager->skippingChannel();
      }catch(const Exception &e){}
    }else{
      return SUCCESS;
    }

    while((res != SUCCESS) && (c != currentChannel)){
      if( c > 0){
        c--;
      }else{
        if(!channelConfig.empty()){
          c=channelConfig.size()-1;
        }else{
          c = 0;
        }
      }
      res = setChannel(c);
      if(setChannel(c) == SUCCESS) {
        if(c == ccBackup){
          return FAILURE;
        }
        return SUCCESS;
      }
    }//end while
    return FAILURE;
  }
  
  TVChannelList* TVManager::getChannelConfig(){
    return &channelConfig;
  }

  INode* TVManager::getTVCard(){
    return currentTVCard;
  }

  TVChannel TVManager::getCurrentChannel(){
    return  channelConfig.getChannelByNumber(currentChannel);
  }

  NodeDescription TVManager::urlToNodeDescription(GraphURL url){
    NodeDescription returnValue;
    if(url.getProtocol() == GraphURL::FILE){

      returnValue = url.getNodeDescription();
      returnValue.setNodeName("DVBSimulatorNode");
      returnValue.addNodeEvent(Event(IFileHandler::setFilename_event, 
                                     new TInValue<string>(url.getURIPath())), INode::CONSTRUCTED);

    }else if(url.getProtocol() == GraphURL::IVTV){

      returnValue = url.getNodeDescription();

      /*
      returnValue.addNodeEvent(Event(ITVCard::setResolution_event, 
                                  new TInValue<unsigned int>(720), 
                                  new TInValue<unsigned int>(576)), 
                            INode::INITIALIZED);
      */
      if(!url.getOption("input").empty()){
        returnValue.addNodeEvent(Event(ITV::setCurrentInput_event, 
                                       new TInValue<unsigned int>(atoi(url.getOption("input").c_str()))), 
                                       INode::INITIALIZED);
      }

      if(!url.getOption("freqtable").empty()){
        returnValue.addNodeEvent(Event(ITV::setFrequencyTable_event, 
                                       new TInValue<string>(url.getOption("freqtable"))), INode::INITIALIZED);      
      }
      if(!url.getOption("standard").empty()){
        returnValue.addNodeEvent(Event(ITV::setCurrentStandard_event, 
                                       new TInValue<unsigned int>(atoi(url.getOption("standard").c_str()))), INode::INITIALIZED);      
      }

      returnValue.addNodeEvent(Event(ITVCard::setResolution_event, 
                                     new TInValue<int>(720), new TInValue<int>(576)), INode::INITIALIZED);      
      
    }else{
      returnValue = url.getNodeDescription();
    }

    if(!url.getOption("channelConfig").empty()){
      returnValue.addNodeEvent(Event(IChannelInfo::setChannelConfigFile_event, 
                                     new TInValue<string>(url.getOption("channelConfig"))), INode::CONSTRUCTED);      
    }

    return returnValue;
  }

  Result TVManager::isMaster(INode *card){
    try{
      if(card){
        IUsage_var use(card->getParentObject()->getInterface<IUsage>());
        if(use.get()){
          return use->isMaster(myCurrentUsage, usageNotifier);
        }
      }
    }catch(const Exception &e){
    }

    return FAILURE;
  }

  Result TVManager::requestCardExclusive(TVChannel::cardList_t cards, int &channel){

    TVChannel::cardList_t::iterator i = cards.begin();

    NodeDescription tvsource;

    while(i != cards.end()){     
      if(application){
        GraphURL url(i->first);
        cerr << "Requesting EXCLUSIVE: " << i->first << endl;
        tvsource = urlToNodeDescription(url);        
        tvsource.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);

        GraphDescription *newgraph = new GraphDescription(*backupGraph);

        if(newgraph){
          try{
            
            newgraph->addEdge(tvsource, *source);
            
            application->getRegistry().requestGraph(*newgraph, RegistryName(tvsource.getLocation(), tvsource.getPort()));
            newgraph->realizeGraph();

            IUsage_var use(newgraph->getINode(tvsource)->getParentObject()->getCheckedInterface<IUsage>());
            use->setUsage(myCurrentUsage, usageNotifier);

            
            if(myCurrentUsage == IUsage::DVBEPG){
              //when the TVManger is used for DVBEPG, we need the ITVGuideProducer-interface
              cerr << "TEST OF DVBEPG" << endl;
              ITVGuideProducer_var tvgprod(newgraph->getINode(tvsource)->
                                           getParentObject()->getCheckedInterface<ITVGuideProducer>());
            }
          }catch(const Exception &e){
            try{
              cerr << "1 release Graph " << e << endl;
              application->getRegistry().releaseGraph(*newgraph);
            }catch(const Exception &e){}

            i++;
            delete newgraph; newgraph = NULL;
            continue;
          }catch(...){}

          //since we prefer cards where we are Master, we store this graph and try to find an other card
          if(!isMaster(newgraph->getINode(tvsource))){
            try{

              ITVCard_var determineChannel(newgraph->getINode(tvsource)->getParentObject()->getCheckedInterface<ITVCard>());

              int cNr; string cName;
              determineChannel->getChannel(cNr, cName);
              //is the requested channel tuned in?
              if(cNr == (int)(i->second)){
                //cerr << "Testing channel " << cNr << " " << (i->second) << endl;
                IUsage_var determineUsage(newgraph->getINode(tvsource)->getParentObject()->getCheckedInterface<IUsage>());
                IUsage::usage_t currentUsage = determineUsage->getUsage();

                try{
                  IUsage_var use(newgraph->getINode(tvsource)->getParentObject()->getCheckedInterface<IUsage>());
                  use->unsetUsage(myCurrentUsage, usageNotifier); 
                }catch(const Exception &e){
                }
                //cerr << "2 release Graph " << endl;

                application->getRegistry().releaseGraph(*newgraph);

                //store informations for the shared-request of the graph
                RequestedGD g;
                g.setSource(url);
                g.setURLString(i->first);
                g.setChannel(i->second);
                g.setUsage(currentUsage);
                           
                requestedGDs[currentUsage].push_back(g);
                i++;  
                delete newgraph; newgraph = NULL;
                continue;
              }
              
              //the card is owned by someone else and has not the correct channel tuned
              IUsage_var use(newgraph->getINode(tvsource)->getParentObject()->getCheckedInterface<IUsage>());
              use->unsetUsage(myCurrentUsage, usageNotifier);
              //cerr << "3 release Graph " << endl;

              application->getRegistry().releaseGraph(*newgraph);

            }catch(const Exception &e){
              //cerr << "4 release Graph " << e << endl;

              application->getRegistry().releaseGraph(*newgraph);
            }

            i++;            
            delete newgraph; newgraph = NULL;
            continue;
          }

          cerr << "YES, MASTER !!!" << endl;
          try{

            //set the channel
            ITVCard_var itvcard(newgraph->getINode(tvsource)->getParentObject()->getCheckedInterface<ITVCard>());
            if(itvcard.get()) itvcard->setChannel(i->second);

              
            if(currentTVCard != NULL){
              //Only release graph when there is a requested graph
              releaseGraph();
            }
            
            //newgraph->realizeGraph();

            *graph = *newgraph;

            delete newgraph; newgraph = NULL;
            

            currentTVCardURL = i->first;
            currentTVCard = graph->getINode(tvsource);

            channel = i->second;

            return SUCCESS;
            
          }catch(const Exception &e){
            GLOBAL_WARNING_STREAM("TVManager::requestCardExclusive : Error requesting the EXCLUSIVE graph"  << e << endl);
            releaseGraph();
          }catch(...){}
        }
        
        if(newgraph) delete newgraph; newgraph = NULL;

      }
      i++;
    }//end while(i!=cards.end())

    return FAILURE;
  }

  Result TVManager::requestCardShared(){
    cerr << "try to request SHARED !! have " << requestedGDs.size() << " usages"<< flush;
    INode* iTVCard = NULL; //store this to detect if the graph changed
    for(map<IUsage::usage_t, list<RequestedGD> >::reverse_iterator j = requestedGDs.rbegin(); j != requestedGDs.rend(); j++){
      list<RequestedGD> l = j->second;
      cerr << " and " << l.size() << " GDs for usage " << j->first << endl;
      for(list<RequestedGD >::iterator i = l.begin(); i != l.end(); i++){
        
        GraphDescription *newgraph = new GraphDescription(*backupGraph);
        NodeDescription tvsource_nd  = urlToNodeDescription(i->getSource());
        tvsource_nd.setSharingType(NodeDescription::SHARED);
        newgraph->addEdge(tvsource_nd, *source);        
        try{
          application->getRegistry().requestGraph(*newgraph, RegistryName(tvsource_nd.getLocation(), tvsource_nd.getPort()));
          newgraph->realizeGraph();
          iTVCard = newgraph->getINode(tvsource_nd);

          if(iTVCard){
            ITVCard_var verifyChannel(iTVCard->getParentObject()->getCheckedInterface<ITVCard>());
            int cNr; string cName;
            verifyChannel->getChannel(cNr, cName);
            if(cNr != i->getChannel()){
              //the channel has changed during our last request, so we can't use this card anymore
              application->getRegistry().releaseGraph(*newgraph);
              delete newgraph; newgraph = NULL;
              continue;
            }
          }else{
            //don't have the ITVcard-Interface. we can't query the channle, so we test the next card
            application->getRegistry().releaseGraph(*newgraph);
            delete newgraph; newgraph = NULL;
            continue;
          }
          
        }catch(const Exception &e) {
          //an error occured during shared request. so test the next card since we can't be sure that this card is usable
          cerr << "an error occured during shared request. so test the next card since we can't be sure that this card is usable " << endl;
          try{
            application->getRegistry().releaseGraph(*newgraph);
          }catch(const Exception &e){}
          delete newgraph; newgraph = NULL;
          continue;
        }

        //we can use this tvcard, so use it ;)
        try{

          if(iTVCard && currentTVCard){
            if(iTVCard->getParentObject() == currentTVCard->getParentObject()){
              //we requested our Graph twice, release the new one 
              cerr << "we requested our Graph twice, release the new one " << endl;
              try{
                application->getRegistry().releaseGraph(*newgraph);
              }catch(const Exception &e){}
              delete newgraph; newgraph = NULL;

              try{
                ITVManager_var itvmanager(this->getCheckedInterface<ITVManager>());
                if(itvmanager.get()) itvmanager->sharedSource();
              }catch(const Exception &e){
              }

              return SUCCESS;
            }
          }
                    
          cerr << "we have a new graph" << endl;

          if(currentTVCard != NULL){
            //Only release graph when there is a requested graph
            releaseGraph();        
          }

          *graph = *newgraph;
          
          
          delete newgraph; newgraph = NULL;
          
          currentTVCardURL = i->getURLString();
          currentTVCard = graph->getINode(tvsource_nd);
          
          IUsage_var use(currentTVCard->getParentObject()->getInterface<IUsage>());
          if(use.get()){
            use->setUsage(myCurrentUsage, usageNotifier);
          }else{
            GLOBAL_WARNING_STREAM("TVManager::requestSharedCard : ERROR: cant' set mode" << endl);
          }
          
          
          try{
            ITVManager_var itvmanager(this->getCheckedInterface<ITVManager>());
            if(itvmanager.get()) itvmanager->sharedSource();
          }catch(const Exception &e){
          }
          
          return SUCCESS;
        
        }catch(const Exception &e){

        }catch(...){}
        
        application->getRegistry().releaseGraph(*graph);
        
      }
    }
    return FAILURE;
  }

  Result TVManager::buildOldGraph(string /*cardURL*/){
    NodeDescription tvsource;

    if(application){

      //request the desired channel
      TVChannel c;
      
      try{
        c = channelConfig.getChannelByNumber(currentChannel);
      }catch(const Exception &e){
        return FAILURE;
      }
      TVChannel::cardList_t cards = c.getCards();
      TVChannel::cardList_t::iterator i = cards.find(currentTVCardURL);

      if(i == cards.end()) return FAILURE;

      GraphURL url(i->first);
      tvsource = urlToNodeDescription(url);
      
      tvsource.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
      
      
      try{
          
        *graph = *backupGraph;
        graph->addEdge(tvsource, *source);
        
        application->getRegistry().requestGraph(*graph, RegistryName(tvsource.getLocation(), tvsource.getPort()));

        //synchronize();
        
        graph->realizeGraph();
        currentTVCard = graph->getINode(tvsource);

        
        IUsage_var use(currentTVCard->getParentObject()->getInterface<IUsage>());
        if(use.get()){
          use->setUsage(myCurrentUsage, usageNotifier);
        }else{
        }
        
        requestedGDs.clear();
        return SUCCESS;
      }catch(const Exception &e){
        try{
          application->getRegistry().releaseGraph(*graph);
        }catch(const Exception &e){
        }
     
      }catch(...){}
      
    }
        
    return SUCCESS;
  }

  Result TVManager::realizeFixGraph(NodeDescription *tvsource, string session_location){

    try{

      int port = ClientRegistry::default_port;

      if(session_location.empty()){
        session_location = tvsource->getLocation();
      }else{
        size_t pos = string::npos;
        if((pos = session_location.find_first_of(':')) != string::npos){
          port = atoi(session_location.substr(pos+1).c_str());
        }
      }

      *graph = *backupGraph;

      graph->addEdge(*tvsource, *source);

      cerr << "sessionLocation: " << session_location << endl;

      application->getRegistry().requestGraph(*graph, RegistryName(session_location, port));
      
      graph->realizeGraph();
      currentTVCard = graph->getINode(*tvsource);
      
      
      IUsage_var use(currentTVCard->getParentObject()->getInterface<IUsage>());
      if(use.get()){
        use->setUsage(myCurrentUsage, usageNotifier);
      }else{
      }      
      
      //update internal variables
      int cnr = 0; string s;
      ITVCard_var itv(currentTVCard->getParentObject()->getInterface<ITVCard>());
      if(itv.get()){
        itv->getChannel(cnr, s);
      }      
      try{
        TVChannel c = channelConfig.getChannelByName(s);
        currentChannel = c.getChannelNumber();
      }catch(const Exception &e){
      }
      currentChannelOnCurrentTVCard = cnr;

      return SUCCESS;
    }catch(const Exception &e){
      try{
        application->getRegistry().releaseGraph(*graph);
      }catch(const Exception &e){
      }
      
    }catch(...){}
    
    return FAILURE;
  }


  void TVManager::releaseGraph(){
    try{
      
      cerr << "release the graph" << endl;

      //Disconnect the given sinks from the graph
      if(video_sink){
        if(video_sink->isStarted()){
          video_sink->reachInitialized();
          disconnect_input(video_sink, "stream1");
        }
      }
      
      
      if(audio_sink){
        if(audio_sink->isStarted()){
          audio_sink->reachInitialized();
          disconnect_input(audio_sink);   
        }
      }
      

      if(currentTVCard){
        IUsage_var use(currentTVCard->getParentObject()->getCheckedInterface<IUsage>());
        if(use.get()){
          use->unsetUsage(myCurrentUsage, usageNotifier);
        }else{
        }
      }
          
      if(currentTVCard){
        //TVCard-Interface will be deleted by the Graph-Description
        //delete currentTVCard;
        currentTVCard = NULL;
      }

      if(graph){
        application->getRegistry().releaseGraph(*graph);
      }else{
      }
    }catch(const Exception &e){
      GLOBAL_WARNING_STREAM("TVManager::releaseGraph() " << e << endl);
    }
  }

  Result TVManager::startGraph(){
    graph->startGraph();
    return SUCCESS;
  }

}











