/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_TVMANAGER_HPP
#define NMM_TVMANAGER_HPP NMM_TVMANAGER_HPP

#include "nmm/NMMTypes.hpp"

#include "nmm/interfaces/misc/tvmanager/ITVManagerImpl.hpp"
#include "nmm/interfaces/video/tv/TVChannelList.hpp"
#include "nmm/interfaces/base/IUsage.hpp"
#include "nmm/interfaces/base/IUsageNotify.hpp"

#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/utils/GraphURL.hpp"

#include "nmm/interfaces/base/sync/IMultiAudioVideoSynchronizer.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"


#include <string>

namespace NMM{

  /**
   *This stores the informations about shared sources in TVManger. 
   *With these informations the TVManager requests shared TV-Cards, if now exclusive could be found
   */
  class RequestedGD{
  private:

    GraphURL source;
    int channel;
    IUsage::usage_t mode;
    string urlString;
  public:
    RequestedGD(){}
    ~RequestedGD(){}

    /**
     *Set the GraphURL of the TV-Card so the NodeDescription can be recreated when requesting the card shared
     */
    void setSource(GraphURL n){source=n;}
    GraphURL getSource(){return source;}

    void setURLString(string s){urlString = s;}
    string getURLString(){return urlString;}
    /**
     *Set the channelnumber on the tvcard
     */
    void setChannel(int c){channel = c;}
    int getChannel(){return channel;}

    /**
     *Store the current usage of the requested source
     */
    void setUsage(IUsage::usage_t u){mode = u;}
    IUsage::usage_t getUsage(){return mode;}
  };


  /**
   *The class requests sources for a given graph. It changes the source and configures the graph.
   */
  class TVManager :
    public ITVManagerImpl
  {
  private:

    //store the apllications IUsageNotify-Interface
    const IUsageNotify* usageNotifier;

    IUsage::usage_t myCurrentUsage;

    //application to request registry from
    NMMApplication *application;

    unsigned int currentChannel;
    TVChannelList channelConfig;
    INode *currentTVCard;
    string  currentTVCardURL;
    GraphDescription *graph;

    //store a copy of the TV-Graph. Since ClientRegistry::releaseGraph(...) calls clear() on the GraphDescription, the graph can be recreated
    GraphDescription *backupGraph;

    NodeDescription *source;

    map< IUsage::usage_t, list<RequestedGD> > requestedGDs;

    unsigned int currentChannelOnCurrentTVCard;

  public:
    TVManager(IUsage::usage_t u, string filename = "");

    TVManager(const IUsageNotify* usageNotify, IUsage::usage_t u);
    /**
     *Additionally specify a filename to load a special TVChannelList
     */
    TVManager(const IUsageNotify* usageNotify, IUsage::usage_t u, string filename);
    ~TVManager();
    
    /**
     *Set the NMMApplication to access the registry
     */
    void setApplication(NMMApplication* app);
    /**
     *Set ao pointer to the GraphDescription, which represents the TV-Graph. 
     *Also give a pointer to the first NodeDescription after the Source for easiert change of the source.
     *@param GraphDescription* pointer to the GraphDescription
     *@param NodeDescirption* pointer to the first node after the source
     */
    void setGraphDescription(GraphDescription *gd, NodeDescription *n);

    /**
     *Method to set pointer to the SinkNodes of the graph to connect the Session-Synchronizer
     *INode* audio The audio-sink of the graph
     *INode* video The video-sink of the graph
     */
    void setSinkNodes(INode* audio, INode* video){audio_sink = audio; video_sink = video;}

    /**
     *Release the GraphDescription
     */
    void releaseGraphDescription();

    /**
     *Realise a Graph with the given source. Source-determination will not be done.
     *NodeDescription* Pointer to the source
     *string session_location give an alternative session-location
     */
    Result realizeFixGraph(NodeDescription *source, string session_location = "");

    /**
     *Get the currentChannel tuned from the ChannelList
     *@return TVChannel the current channel
     */
    TVChannel getCurrentChannel();
    /**
     *Get the current channel-number according to the TVChannelList
     *return int current channel-number
     */
    unsigned int getCurrentChannelNumber(){return currentChannel;}
    /**
     *Get the current channel-number on the current TV-Card according to the TV-Card's channel-configuration-file
     *@return int current channel-number on current TV-card
     */
    unsigned int getCurrentChannelOnCurrentTVCard(){return currentChannelOnCurrentTVCard;}
    /**
     *get the current ChannelConfig wich is used
     */
    TVChannelList* getChannelConfig();

  private:
    NodeDescription urlToNodeDescription(GraphURL url);

    Result isMaster(INode *card);
    Result requestCardExclusive(TVChannel::cardList_t cards, int &channel);
    Result requestCardShared();
    Result buildOldGraph(string cardURL);

    //synchronizing
    INode* video_sink;
    INode* audio_sink;

  protected:
    //ITVManager-Interface

    /**
     *Set a channel directly on the current TV-Graph
     */
    Result setChannel(const unsigned int &channel);
    /**
     *Switch one channel up on the current TV-Graph. If the next channel is unavailable, it tunes to the next available channel
     */
    Result channelUp();
    /**
     *Switch one channel down on the current TV-Graph. If the next channel is unavailable, it tunes to the next available channel
     */
    Result channelDown();

    /**
     *get the INode-interface of the current TV-card
     */
    INode* getTVCard();

    Result startGraph();

    void setUsageNotify(const IUsageNotify* u){usageNotifier = dynamic_cast<IUsageNotify*>(u->copy());}

    void releaseGraph();

    Result switchingSource(){return SUCCESS;}
    Result sharedSource(){return SUCCESS;}
    Result skippingChannel(){return SUCCESS;}

  };
}
#endif
