/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TVRecorderManager.hpp"

#include "nmm/misc/tvtimer/TVTimerList.hpp"

#include "nmm/interfaces/file/IFileHandler.hpp"

#include "nmm/base/ProxyObject.hpp"

#include "nmm/interfaces/misc/tvtimer/ITVTimerEvents.hpp"
#include "nmm/interfaces/base/IUsage.hpp"

#include "nmm/utils/utils.hpp"

#define RECORDSTARTEVENT "STARTRECORDING"
#define RECORDSTOPEVENT  "STOPRECORDING"
#define TIMEOFFSET 180

namespace NMM{
  TVRecorderManager::TVRecorderManager(ITVTimer *itvtimer)
    : recordStartTimer(this, &TVRecorderManager::startRecorder),
      recordStopTimer(this, &TVRecorderManager::stopRecorder),
      directory(""), 
      controller(itvtimer)
  {
    timer = new TimerProducer();
    timer->connectTo(this);
    timer->start();

    registerEvent(RECORDSTARTEVENT, &recordStartTimer);
    registerEvent(RECORDSTOPEVENT, &recordStopTimer);
  }

  TVRecorderManager::~TVRecorderManager(){

    removeEvent(RECORDSTARTEVENT);
    removeEvent(RECORDSTOPEVENT);

    timer->disconnect();
    timer->stop();
    delete timer;

  }

  Result TVRecorderManager::record(TVTimerEntry &entry){
    GLOBAL_DEBUG_STREAM("TVRecorderManager::record(TVTimerEntry)" << entry.getID()<< endl);
    entry.setStatus(TVTimerEntry::SLEEPING);
    STVTimerList::getInstance().changeTimer(entry);

    Event* e1 = new Event(RECORDSTARTEVENT, 1);
    e1->setValue(0, new TInValue<unsigned int>(entry.getID()));
    timer->registerTimer(entry.getStartTime(), e1);

    return SUCCESS;
  }

  Result TVRecorderManager::recordShared(TVTimerEntry &entry){
    Result res = FAILURE;
    try{
      unsigned int id = entry.getID();
      TVRecorder *r = new TVRecorder(entry, this);
      if(r) {
        recorder[id] = r;
        r->setDirectory(directory);
        r->setApplication(app);
        if(r->startShared() != SUCCESS){
          //if starting failed, delete the recorder
          delete r;
          recorder.erase(id);
          res = FAILURE;
        }else{
          res = SUCCESS;
        }
      }
    }catch(const NoTimerFoundException &e){
      res = FAILURE;
    }

    if(controller.get()){
      ITVTimerEvents_var itimerEvent(controller->getParentObject()->getInterface<ITVTimerEvents>());
      if(itimerEvent.get()) itimerEvent->timerUpdate();
    }

    return res;
  }

  Result TVRecorderManager::removeRecord(unsigned int id){
    if(recorder.find(id) != recorder.end()){
      recorder[id]->stop();
      delete recorder[id];
      recorder[id] = NULL;
      recorder.erase(id);
    }

    if(controller.get()){
      ITVTimerEvents_var itimerEvent(controller->getParentObject()->getInterface<ITVTimerEvents>());
      if(itimerEvent.get()) itimerEvent->timerUpdate();
    }

    return SUCCESS;
  }

  Result TVRecorderManager::startRecorder(unsigned int &id){
    GLOBAL_DEBUG_STREAM("TVRecorderManager::startRecorder(" << id << ")" << endl);
    try{
      TVTimerEntry entry = STVTimerList::getInstance().getTimer(id);
      TVRecorder *r = new TVRecorder(entry, this);
      if(r) {
        recorder[id] = r;
        r->setDirectory(directory);
        r->setApplication(app);
        if(r->start() != SUCCESS){
          //if starting failed, delete the recorder
          delete r;
          r = NULL;
          recorder.erase(id);

          //initiate a reprogramming of the timer, if the graph can't be build
          try{
            ITVTimerEvents_var itvtimerevent(controller->getParentObject()->getCheckedInterface<ITVTimerEvents>());
            if(itvtimerevent.get()) itvtimerevent->tvcardUnavailable(id);
          }catch(const Exception &e){
            GLOBAL_ERROR_STREAM(e << endl);
          }
          return FAILURE;
        }
      }
    }catch(const NoTimerFoundException &e){
      if(controller.get()){
        ITVTimerEvents_var itimerEvent(controller->getParentObject()->getInterface<ITVTimerEvents>());
        if(itimerEvent.get()) itimerEvent->timerUpdate();
      }
      return FAILURE;
    }

    if(controller.get()){
      ITVTimerEvents_var itimerEvent(controller->getParentObject()->getInterface<ITVTimerEvents>());
      if(itimerEvent.get()) itimerEvent->timerUpdate();
    }

    return SUCCESS;
  }

  Result TVRecorderManager::stopRecorder(unsigned int &id){
    GLOBAL_DEBUG_STREAM("TVRecorderManager::stopRecorder(" << id << ")" << endl);

    if(recorder.find(id) == recorder.end()) return FAILURE;

    recorder[id]->stop();
    delete recorder[id];
    recorder[id] = NULL;
    recorder.erase(id);

    if(controller.get()){
      ITVTimerEvents_var itimerEvent(controller->getParentObject()->getInterface<ITVTimerEvents>());
      if(itimerEvent.get()) itimerEvent->timerUpdate();
    }

    try{
      TVTimerEntry entry = STVTimerList::getInstance().getTimer(id);

      //(do or don't) program the next timer according to the period. 
      
      switch(entry.getPeriod()){
      case (TVTimerEntry::DAILY):
        {
          time_t s = entry.getStartTime();
          struct tm *startTime = localtime(&s);
          startTime->tm_mday += 1;
          entry.setStartTime(mktime(startTime));
          
          time_t e = entry.getEndTime();        
          struct tm *endTime = localtime(&e);
          endTime->tm_mday += 1;
          entry.setEndTime(mktime(endTime));
        }
        break;
      case (TVTimerEntry::WEEKLY):
        {
          time_t s = entry.getStartTime();
          struct tm *startTime = localtime(&s);
          startTime->tm_mday += 7;
          entry.setStartTime(mktime(startTime));
          
          time_t e = entry.getEndTime();        
          struct tm *endTime = localtime(&e);
          endTime->tm_mday += 7;
          entry.setEndTime(mktime(endTime));
        }
        break;
      case (TVTimerEntry::WEEKDAYS):
        {
          time_t s = entry.getStartTime();
          struct tm *startTime = localtime(&s);
          if(startTime->tm_wday == 5){
            startTime->tm_mday += 3;
          }else{
            startTime->tm_mday += 1;
          }
          entry.setStartTime(mktime(startTime));
          
          time_t e = entry.getEndTime();        
          struct tm *endTime = localtime(&e);
          if(endTime->tm_wday == 5){
            endTime->tm_mday += 3;
          }else{
            endTime->tm_mday += 1;
          }
          entry.setEndTime(mktime(endTime));
        }
        break;
      case (TVTimerEntry::WEEKENDS):
        {
          
          time_t s = entry.getStartTime();
          struct tm *startTime = localtime(&s);
          if(startTime->tm_wday == 0){
            startTime->tm_mday += 6;
          }else{
            startTime->tm_mday += 1;
          }
          entry.setStartTime(mktime(startTime));
          
          time_t e = entry.getEndTime();        
          struct tm *endTime = localtime(&e);
          if(endTime->tm_wday == 0){
            endTime->tm_mday += 6;
          }else{
            endTime->tm_mday += 1;
          }
          entry.setEndTime(mktime(endTime));
        }
        break;
      case (TVTimerEntry::MONTHLY):
        {
          time_t s = entry.getStartTime();
          struct tm *startTime = localtime(&s);
          startTime->tm_mon += 1;
          entry.setStartTime(mktime(startTime));
          
          time_t e = entry.getEndTime();        
          struct tm *endTime = localtime(&e);
          endTime->tm_mon += 1;
          entry.setEndTime(mktime(endTime));
        }
        break;
      case(TVTimerEntry::RANDOM):
        {
          list<ProgramDescription> nextTimerList = entry.getNextTimers();         
          if(nextTimerList.empty()){

            //look for new entries
            if(!entry.getTimerInfo().empty()){
              entry.getTimerInfo().localizeTimes();
              ProgramFilter pf;
              pf.setStartTime(entry.getTimerInfo().getEndTime());
              pf.setSearchPhrase(entry.getTimerInfo().getTitles().front());
              if(controller.get()){
                if(controller->getTVGuide()){
                  nextTimerList = controller->getTVGuide()->search(pf);
                  nextTimerList.sort();
                  entry.setNextTimers(nextTimerList);
                }
              }
            }else{              
              return SUCCESS;
            }
          }
        
          if(!nextTimerList.empty()){

            //Program only the first timer with the old channel as this period. 
            //Determine other timers, wich have the same start-time and program them once.
            //remove this values from the next-timers-list
            if(determineNextTimer(entry) != SUCCESS){
             return SUCCESS;
            }

            nextTimerList = entry.getNextTimers(); //actualize nextTimerList
            list<ProgramDescription>::iterator i = nextTimerList.begin();
            while(stringToTime(i->getStartTime()) <= stringToTime(entry.getTimerInfo().getStartTime())){
              i++;
              if(i == nextTimerList.end()) break;
            }

            //set the correct channelname
            if(controller->getTVGuide()){
              string channelname = controller->getTVGuide()->getChannelById(entry.getTimerInfo().getChannelID());
              entry.setChannelName(channelname);
            }
          }else{
            return SUCCESS;
          }
        }
        break;
      case(TVTimerEntry::RANDOM_ONLY_CHANNEL_AND_TIME):
        {
          list<ProgramDescription> nextTimerList = entry.getNextTimers();         
          if(nextTimerList.empty()){
            //look for new entries
            if(!entry.getTimerInfo().empty()){
              entry.getTimerInfo().localizeTimes();
              ProgramFilter pf;
              pf.setStartTime(entry.getTimerInfo().getEndTime());
              pf.addChannel(entry.getTimerInfo().getChannelID());
              pf.setSearchPhrase(entry.getTimerInfo().getTitles().front());
              if(controller.get()){
                if(controller->getTVGuide()){
                  nextTimerList = controller->getTVGuide()->search(pf);
                  nextTimerList.sort();
                  list<ProgramDescription> resList2;
                  
                  for(list<ProgramDescription>::iterator i = nextTimerList.begin(); i != nextTimerList.end(); i++){
                    i->localizeTimes();
                    if(atoi(i->getStartTime().substr(8,4).c_str()) >= atoi(entry.getTimerInfo().getStartTime().substr(8,4).c_str())
                       && atoi(i->getStartTime().substr(8,4).c_str()) <= atoi(entry.getTimerInfo().getEndTime().substr(8,4).c_str())){
                      resList2.push_back(*i);
                    }
                  }
                  
                  entry.setNextTimers(resList2);
                }
              }
            }else{
              return SUCCESS;
            }
          }
        
          if(!nextTimerList.empty()){
            if(determineNextTimer(entry) != SUCCESS){
              return SUCCESS;
            }
          }else{
            return SUCCESS;
          }
        }
        break;
      case(TVTimerEntry::RANDOM_ONLY_CHANNEL):
        {
          list<ProgramDescription> nextTimerList = entry.getNextTimers();         
          if(nextTimerList.empty()){
            //look for new entries
            if(!entry.getTimerInfo().empty()){
              entry.getTimerInfo().localizeTimes();
              ProgramFilter pf;
              pf.setStartTime(entry.getTimerInfo().getEndTime());
              pf.addChannel(entry.getTimerInfo().getChannelID());
              pf.setSearchPhrase(entry.getTimerInfo().getTitles().front());
              if(controller.get()){
                if(controller->getTVGuide()){
                  nextTimerList = controller->getTVGuide()->search(pf);
                  nextTimerList.sort();
                  entry.setNextTimers(nextTimerList);
                }
              }
            }else{
              return SUCCESS;
            }
          }
        
          if(!nextTimerList.empty()){
            if(determineNextTimer(entry) != SUCCESS){
              return SUCCESS;
            }
          }else{
            return SUCCESS;
          }
        }
        break;
      default: //also handles TVTimerEntry::ONCE
        return SUCCESS;
        break;
      }

      controller->changeTimer(entry);
    }catch(const NoTimerFoundException &e){
    }

    return SUCCESS;
  }

  void TVRecorderManager::setDirectory(string dir){
    directory = dir;
  }
  string TVRecorderManager::getDirectory(){
    return directory;
  }

  Result TVRecorderManager::determineNextTimer(TVTimerEntry &entry){
    list<ProgramDescription> nextTimerList = entry.getNextTimers();

    if(nextTimerList.empty()) return FAILURE;

    ProgramDescription currentTimerInfo = entry.getTimerInfo();
    list<ProgramDescription>::iterator nextTimer = nextTimerList.begin();

    ProgramDescription nextTimerInfo = *nextTimer; //because list is sorted

    for(nextTimer = nextTimerList.begin(); nextTimer != nextTimerList.end(); nextTimer++){
      //search the next entry for the last channel recorded.
      if(nextTimer->getChannelID() == currentTimerInfo.getChannelID()){
        nextTimerInfo = *nextTimer;
        //remove this timer from the next-timers-list
        nextTimerList.erase(nextTimer);
        break;
      }
    }

    try{
      struct tm newTime;
      newTime.tm_sec  = atoi(nextTimerInfo.getStartTime().substr(12,2).c_str());
      newTime.tm_min  = atoi(nextTimerInfo.getStartTime().substr(10,2).c_str());
      newTime.tm_hour = atoi(nextTimerInfo.getStartTime().substr(8,2).c_str());
      newTime.tm_mday = atoi(nextTimerInfo.getStartTime().substr(6,2).c_str());
      newTime.tm_mon  = atoi(nextTimerInfo.getStartTime().substr(4,2).c_str())-1;
      newTime.tm_year = atoi(nextTimerInfo.getStartTime().substr(0,4).c_str())-1900;
      time_t startTime = mktime( &newTime );

      newTime.tm_sec  = atoi(nextTimerInfo.getEndTime().substr(12,2).c_str());
      newTime.tm_min  = atoi(nextTimerInfo.getEndTime().substr(10,2).c_str());
      newTime.tm_hour = atoi(nextTimerInfo.getEndTime().substr(8,2).c_str());
      newTime.tm_mday = atoi(nextTimerInfo.getEndTime().substr(6,2).c_str());
      newTime.tm_mon  = atoi(nextTimerInfo.getEndTime().substr(4,2).c_str())-1;
      newTime.tm_year = atoi(nextTimerInfo.getEndTime().substr(0,4).c_str())-1900;
      time_t stopTime =  mktime( &newTime );

      entry.setStartTime(startTime - TIMEOFFSET);
      entry.setEndTime(stopTime + TIMEOFFSET);
      entry.setTimerInfo(nextTimerInfo);
      entry.setNextTimers(nextTimerList);
    }catch(const Exception &e){
      return FAILURE;
    }catch(...){
      return FAILURE;
    }
    return SUCCESS;
  }
    
  //------------ TVRecorder ------------------------------
  TVRecorder::TVRecorder(TVTimerEntry entry, TVRecorderManager *rm)
    : timerEntry(entry),
      manager(rm),
      directory(""),
      filename(""),
      app(NULL),
      channelSwitched_listener(this, &TVRecorder::channelSwitched),
      tvcard(NULL)
  {
    recTimer = new TimerProducer();
    recTimer->connectTo(manager);
    recTimer->start();
  }

  TVRecorder::~TVRecorder(){    

    //recTimer->stop();
    //recTimer->disconnect();
    //I'm not allowed to delete the recTimer. Otherwise I get a segfault with malloc in libLiveMedia ???
    //delete recTimer;    
  }
  
  Result TVRecorder::start(){
    setFilename();
    if(buildGraph(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED) == SUCCESS){
      
      graph.activateGraph();

      graph.startGraph();

      IUsage_var use(tvcard->getParentObject()->getInterface<IUsage>());
      if(use.get()){
        use->setUsage(IUsage::RECORDER, manager->getInterface<IUsageNotify>());
      }else{
        //cerr << "ERROR: cant' set mode" << endl;
      }
      
      //tvcard->getParentObject()->registerEventListener(ITVCard::setChannel_event, &channelSwitched_listener);
    }else{
      GLOBAL_ERROR_STREAM("TVRecorder::start() Building Graph failed" << endl);
      timerEntry.setStatus(TVTimerEntry::RECORDED_WITH_ERROR);
      STVTimerList::getInstance().changeTimer(timerEntry);
      releaseGraph();
      return FAILURE;
    }

    //change the status of the timer, if the graph could be build
    timerEntry.setStatus(TVTimerEntry::RECORDING);
    STVTimerList::getInstance().changeTimer(timerEntry);

    Event* e = new Event(RECORDSTOPEVENT, 1);
    e->setValue(0, new TInValue<unsigned int>(timerEntry.getID()));
    recTimer->registerTimer(timerEntry.getEndTime(), e);

    return SUCCESS;
  }
  void TVRecorder::stop(){
    //if (tvcard) tvcard->getParentObject()->removeEventListener(&channelSwitched_listener);

    try{

      //unset our mode on the source
      IUsage_var use(tvcard->getParentObject()->getInterface<IUsage>());
      if(use.get()){
        use->unsetUsage(IUsage::RECORDER, manager->getInterface<IUsageNotify>());
      }else{
        //cerr << "ERROR: cant' unset mode" << endl;
      }

      releaseGraph();
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("TVRecorder::stop(): " << e << endl);
    }

    
    //change the status of the timer
    timerEntry.setStatus(TVTimerEntry::RECORDED);
    STVTimerList::getInstance().changeTimer(timerEntry);
  }

  Result TVRecorder::channelSwitched(int & /* channel */){
    return SUCCESS;
    try{
      //releaseGraph();
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("TVRecorder::stop(): " << e << endl);
    }
    
    int recordID = timerEntry.getID();

    //change the status of the timer
    timerEntry.setStatus(TVTimerEntry::RECORDED_WITH_ERROR);
    STVTimerList::getInstance().changeTimer(timerEntry);
    
    manager->removeRecord(recordID);

    return SUCCESS;
  }
  
  Result TVRecorder::buildGraph(NodeDescription::SharingType st){
    
    if(app){
      try{
        
        NodeDescription tv_nd = timerEntry.getTVCard();

        tv_nd.setSharingType(st); 

        NodeDescription writer_nd("GenericWriteNode");
        writer_nd.setSharingType(NodeDescription::EXCLUSIVE);
        
        writer_nd.addNodeEvent(Event(IFileHandler::setFilename_event, new TInValue<string>(filename)), INode::INITIALIZED);

        graph.clear();
        graph.addEdge(tv_nd, writer_nd);

        app->getRegistry().requestGraph(graph, RegistryName(tv_nd.getLocation(), 22801));
        tvcard = graph.getINode(tv_nd);

        graph.realizeGraph();

        if(!isMaster(tvcard)){
          //cerr << " **** I'm slave, so check the channel" << endl;
          int channel = -1;
          string s; //don't need this. Just to call getChannel
          if(tvcard){
            ITVCard_var itvcard(tvcard->getParentObject()->getCheckedInterface<ITVCard>());
            itvcard->getChannel(channel,s);
            //cerr << "shared channel " << channel << " need channel " << timerEntry.getChannelNumber() << endl;
            if(channel != timerEntry.getChannelNumber()){
              return FAILURE;
            }
          }
        }else{
          //cerr << "I AM THE MASTER !!!" << endl;
          ITVCard_var itvcard(tvcard->getParentObject()->getInterface<ITVCard>());
          if(itvcard.get()){
           itvcard->setChannel(timerEntry.getChannelNumber());
          }
        }

      }catch(const Exception &e){
        GLOBAL_ERROR_STREAM( "buildGraph: " << e << endl);
        try{
          app->getRegistry().releaseGraph(graph);
        }catch(const Exception &e){}
        return FAILURE;
      }
      
      return SUCCESS;
    }
    return FAILURE;
  }

  Result TVRecorder::releaseGraph(){          

    try{
      if(app){
        app->getRegistry().releaseGraph(graph);
        tvcard = NULL;
      }    
    }catch(const Exception &e){
      return FAILURE;
    }
    return SUCCESS;
  }

  Result TVRecorder::setFilename(){

    if(timerEntry.getTimerInfo().empty()){
      time_t t = timerEntry.getStartTime();
      char tmp[20];
      strftime(tmp, sizeof(tmp), "%Y-%m-%d_%H:%M", localtime(&t));
      
      filename = timerEntry.getChannelName() + "_" + string(tmp) + ".mpeg";
      
      if(!directory.empty()){
        filename = directory + "/" + filename;
      }
    }else{
      ProgramDescription pd = timerEntry.getTimerInfo();

      time_t t = timerEntry.getStartTime();
      char tmp[20];
      strftime(tmp, sizeof(tmp), "%Y-%m-%d_%H:%M", localtime(&t));

      filename = pd.getTitles().front() + "_" + string(tmp) + ".mpeg";

      if(!directory.empty()){
        filename = directory + "/" + filename;
      }
    }
    
    GLOBAL_DEBUG_STREAM("TVRecorder::setFilename(" << filename << ")" << endl);

    timerEntry.setFileURL("file://" + getHostname()+ filename);
    return SUCCESS;

  }

  void TVRecorder::setDirectory(string dir){
    directory = dir;
  }
  string TVRecorder::getDirectory(){
    return directory;
  }

  Result TVRecorder::startShared(){
    setFilename();
    if(app){
      list<TVCardDescription> cards = timerEntry.getTVCards();
      for(list<TVCardDescription>::iterator i = cards.begin(); i != cards.end(); i++){
        try{
          //try every known TVCard for a shared request. this is our last desperate try to record the timer
          NodeDescription tv_nd(*i);
          tv_nd.setSharingType(NodeDescription::SHARED); 
          
          NodeDescription writer_nd("GenericWriteNode");
          writer_nd.setSharingType(NodeDescription::EXCLUSIVE);
          
          writer_nd.addNodeEvent(Event(IFileHandler::setFilename_event, new TInValue<string>(filename)), INode::INITIALIZED);
          
          graph.clear();
          graph.addEdge(tv_nd, writer_nd);
          
          app->getRegistry().requestGraph(graph, RegistryName(tv_nd.getLocation(), 22801));
          tvcard = graph.getINode(tv_nd);
          
          graph.realizeGraph();
        
          int channel = -1;
          string s; //don't need this. Just to call getChannel
          if(tvcard){
            ITVCard_var itvcard(tvcard->getParentObject()->getCheckedInterface<ITVCard>());
            itvcard->getChannel(channel,s);
            if(channel != timerEntry.getChannelNumber()){
              continue;
            }
          }

          graph.startGraph();
          //change the status of the timer, if the graph could be build
          timerEntry.setStatus(TVTimerEntry::RECORDING);
          STVTimerList::getInstance().changeTimer(timerEntry);
          
          Event* e = new Event(RECORDSTOPEVENT, 1);
          e->setValue(0, new TInValue<unsigned int>(timerEntry.getID()));
          recTimer->registerTimer(timerEntry.getEndTime(), e);
        
          return SUCCESS;
        }catch(const Exception &e){
          try{
            app->getRegistry().releaseGraph(graph);
          }catch(const Exception &e){}
        }
      }

      
      GLOBAL_ERROR_STREAM("TVRecorder::startShared() failed" << endl);
      timerEntry.setStatus(TVTimerEntry::RECORDED_WITH_ERROR);
      STVTimerList::getInstance().changeTimer(timerEntry);
      return FAILURE;
    }
        
    return FAILURE;
  }

  Result TVRecorder::isMaster(INode *card){
    try{
      if(card){
        IUsage_var use(card->getParentObject()->getCheckedInterface<IUsage>());
        if(use.get()){
          return use->isMaster(IUsage::RECORDER, manager->getCheckedInterface<IUsageNotify>());
        }
      }
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("TVRecorder: " << e << endl);
    }

    return FAILURE;
  
  }


}
