/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_CDDANODE_HPP
#define NMM_CDDANODE_HPP NMM_CDDANODE_HPP

#include "nmm/base/GenericSourceNode.hpp"

#include "nmm/interfaces/device/cdda/ICDDAImpl.hpp"
#include "nmm/interfaces/device/cdda/ICDDBImpl.hpp"
#include "nmm/interfaces/general/progress/IProgressImpl.hpp"
#include "nmm/interfaces/file/ISeekableImpl.hpp"

extern "C" {
#include <cdda_interface.h>
#include <cdda_paranoia.h>
}

#include "TableOfContents.hpp"

const string DEFAULT_CD_DEVICE = "/dev/cdrom";
const bool DEFAULT_DO_CDDB_QUERY = false;
const bool DEFAULT_CD_REPEAT = true;
const string DEFAULT_CDDB_SERVER = "freedb.freedb.org";
const bool DEFAULT_CDDB_PORT = 8880;

namespace NMM {

  /**
   * This node extracts digital audio from a cdrom by using cdparanoia (libcdda_{interface,paranoia}.so. 
   */

  class CDDANode :
    public GenericSourceNode,
    public ICDDAImpl,
    public ICDDBImpl,
    public IProgressImpl,
    public ISeekableImpl
    //    public ITagInfoImpl
  {
  private:
    enum EventState {TOC, START, INCOMPLETE, SYNC_RESET, DATA, SEEKED, END, FINISHED, FLUSH};
    EventState event;

    //! The structure of a cdrom device as provided by libcdda_interface and needed to initialize libcdda_paranoia
    struct cdrom_drive* drive;
    //! A pointer for libcdda_paranoia, where it saves its configuration after initializing
    cdrom_paranoia* paranoia;
    //! The node's internal property for the cdrom device to use
    string device;
    //! The node's internal property for doing cddb queries or not
    bool do_cddb_query;
    //! The node's internal property for indikating if there is already added a cddb server
    bool cddb_server_added;
    //! verbosity level for the cdparanoia library
    bool repeat;
    //! verbosity level for the cdparanoia library
    int   verbose;
    //! error correction level for the cdparanoia library
    int paranoia_mode;
    //! The current played track number
    unsigned int track_number;
    //! The number of the last track played for internal event sending
    unsigned int last_track_number;
    //! A copy of the sector currently to extract by cdparanoia
    long  current_sector;
    //! The stream counter for the buffer timestamps
    long  stream_counter;
    //! mutex for paranoia seek vs read operations
    ThreadMutex paranoia_seek_mutex;
    //! mutex for internal event handling
    ThreadMutex event_mutex;
    //! The table of contents of the cdrom, including methods to fetch cddb data
    TableOfContents toc;

    bool         send_progress;
    unsigned int buffers_before_progress;
    unsigned int buffers_progress;

    bool one_track_only;
    //! we want numbered tracknames?
    bool numberTrack;

    //! we want detailed TagInfos?
    bool sendtaginfo;

    /**
     * helper method.
     * Used by the different seeking methods to do the real seek.
     */
    Result jumpToSector(int sector);
  protected:
    /**
     * Initializiation of the cdaranoia library and cdrom.
     * This method does the initialization stuff for cdparanoia, like
     * searching the cdrom device if not explicitly specified and opening it for cdparanoia,
     * setting the verbosity and error correction level for cdparanoia,
     * seeking to the begining of the cdrom,
     * initializing the TOC with the track lengths and dummy titles and
     * fetching cddb information into the TOC if selected.
     * \returns FAILURE if there was no problem, SUCCESS otherwise.
     */
    virtual Result doInit();
    
    /**
     * Deinitializiation of the cdaranoia library and cdrom.
     * \returns always SUCCESS as there could not be any problems recognized.
     */
    
    virtual Result doInitOutput();
    virtual Result doDeinitOutput();
    
    /**
     * Main work.
     * This method does the real work:
     * it calls cdparanoia multiple times to produce a bigger buffer than the 2kB provided by cdparanoia and
     * generates events for new tracks, end of tracks and end of cdrom.
     * cdrom access is mutex-locked!
     * \returns NULL if there was no problem, a valid data buffer otherwise.
     */
    virtual Message* processBuffer(Buffer*); // even source nodes are now called with a buffer pointer (which is obviously always NULL) to better match the generic node interface


    // IDevice interface
    virtual void setDevice(const string& device) throw(const DeviceNotFoundException&);
    virtual string getDevice() const;
    
    // ICDDA interface
    virtual Result setTrack(const int& track);
    virtual int getTrack();
    virtual void setRepeat(const bool& repeat);
    virtual bool getRepeat();
    virtual int getTotalTracks();
    virtual bool isFinished();
    virtual void setOneTrackOnly(const bool& enable);
    virtual bool getOneTrackOnly();
    

    // ISeekable interface
    virtual bool hasPercentSeek() const;
    virtual void seekPercent(const Rational& percent);
    virtual void seekPercentTo(const Rational& percent);
    virtual bool hasBytesSeek() const;
    virtual void seekBytes(const int64_t& offset);
    virtual void seekBytesTo(const u_int64_t& bytes);
    virtual bool hasTimeSeek() const;
    virtual void seekTime(const Interval& interval);
    virtual void seekTimeTo(const Time& time);
    virtual void numberTracks(const bool& number);

    // IProgress interface
    virtual void sendProgressInformation(const bool& value);
    virtual void setProgressInterval(const int& buffers);
    
    // ICDDB interface
    virtual Result setCDDBServer(const string& host, const int& port);
    virtual string getTrackName(const int& track);
    virtual string getDiscTitle();
    virtual string getDiscArtist();
    virtual void setCDDB(const bool& enable);
    virtual bool getCDDB();

    //ITrackInfo Interface
    //virtual void sendTagInfo(const bool& _sendinfo);

  public:
    /**
     * The default constructor.
     * \param name: the name oth the node (default is "CDDANode").
     */
    CDDANode(const char* name = "CDDANode",
	     StreamQueue::Mode = StreamQueue::MODE_SUSPEND,
	     int = StreamQueue::DEFAULT_SIZE);
    /**
     * The default destructor.
     * deletes all cdparanoia objects and config data.
     */
    ~CDDANode();
  };
}

#endif //NMM_CDDANODE_HPP
