/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

#ifndef NMM_LADSPANODE_HPP
#define NMM_LADSPANODE_HPP NMM_LADSPANODE_HPP

#include "nmm/base/GenericProcessorNode.hpp"
#include "nmm/interfaces/audio/ladspa/ILADSPAImpl.hpp"
#include "ladspa.h"
#include "ltdl.h"


namespace NMM {
    /**
     *@class LADSPANode
     *@brief a LADSPA-plugin host node
     * this node implemements a LADSPA host
     */
    class LADSPANode :
    public GenericProcessorNode,
    public ILADSPAImpl
  {
    
    struct LADSPA_Buffer_t {
      int port;
      LADSPA_Data* data;
    };
    
    typedef std::vector<LADSPA_Buffer_t*> LADSPA_Buffer;
    typedef std::vector<LADSPA_Data> LADSPA_Vector;

    struct LADSPA_Plugin_t {
      lt_dlhandle libhandle;                      /* filehandle for .so file */
      std::string libname;                        /* filename for .so file (needed for presets) */
      const LADSPA_Descriptor *descriptor;        /* "interface" to the plugin */
      LADSPA_Handle plugin;                       /* instance handle of the plugin */
      unsigned int numControlIn;                  /* number of Control-In ports */
      unsigned int numControlOut;                 /* number of Control-Out ports */
      unsigned int numAudioIn;                    /* number of Audio-In ports */
      unsigned int numAudioOut;                   /* number of Audio-Out port */
      LADSPA_Vector  bufControlIn;                /* data buffer for Control-In ports */
      LADSPA_Data bufControlOut;                  /* data buffer for Control-Out ports */
      LADSPA_Buffer bufAudioIn;                   /* data buffer for Audio-In ports */
      LADSPA_Buffer bufAudioOut;                  /* data buffer for Audio-Out ports */
      LADSPA_Vector baseValues;                   /* preset values for Control-In ports */
    }; // struct LADSPA_Plugin_t

  public:
    /* class construction kit */
    LADSPANode(const char* = "LADSPANode",
               const StreamQueue::Mode qmode= StreamQueue::DEFAULT_MODE,
               const int qize= StreamQueue::DEFAULT_SIZE);
    ~LADSPANode(void);

    /* ILADSPA interface */

    /**
     * adds a new LADSPA plugin to the list;
     * this method assumes that the LADSPA_PATH environment variable is set
     * \param value name or label of a LADSPA plugin
     * \name a string identifying this LADSPA plugin uniquly
     */
    virtual bool addLADSPAPlugin(const string& value,string& name,unsigned int& execpos);

    virtual void removeLADSPAPluginByName(const string& name);
    virtual void removeLADSPAPluginByIndex(const unsigned int& index);

    /**
     * tells the LADSPANode which file to use to initialize the LADSPA plugins
     * \param value filename for a XML file with config-date for LADSPA plugins
     */
    virtual void setConfigFilename(const string& value);
    virtual void setConfigProfileByName(const string& name,const string& profile);
    virtual void setConfigProfileByIndex(const unsigned int& index, const string& profile);

    virtual void resetToBaseValuesByName(const string& name);
    virtual void resetToBaseValuesByIndex(const unsigned int& index);

    virtual void setControlInValueByName(const string& name,const unsigned int& port, const float& value);
    virtual void setControlInValueByIndex(const unsigned int& index,const unsigned int& port, const float& value);

    virtual float getControlInValueByName(const string& name,const unsigned int& port);
    virtual float getControlInValueByIndex(const unsigned int& index,const unsigned int& port);

    virtual string getControlInNameByName(const string& name,const unsigned int& port);
    virtual string getControlInNameByIndex(const unsigned int& index,const unsigned int& port);

    virtual void setControlInBaseByName(const string& name,const unsigned int& port, const float& value);
    virtual void setControlInBaseByIndex(const unsigned int& index,const unsigned int& port, const float& value);

    virtual float getControlInBaseByName(const string& name,const unsigned int& port);
    virtual float getControlInBaseByIndex(const unsigned int& index,const unsigned int& port);

    virtual unsigned int getLADSPAPluginCount();
    virtual unsigned int getLADSPAPluginIndex(const string& name);
    virtual string getLADSPAPluginName(const unsigned int& index);

    virtual unsigned int getControlInCountByName(const string& name);
    virtual unsigned int getControlInCountByIndex(const unsigned int& index);

    virtual bool hasControlInUpperBoundByName(const string& name,const unsigned int& port);
    virtual bool hasControlInUpperBoundByIndex(const unsigned int& index,const unsigned int& port);

    virtual bool hasControlInLowerBoundByName(const string& name,const unsigned int& port);
    virtual bool hasControlInLowerBoundByIndex(const unsigned int& index,const unsigned int& port);

    virtual void getControlInBoundsByName(const string& name,const unsigned int& port, float& lower, float& upper);
    virtual void getControlInBoundsByIndex(const unsigned int& index,const unsigned int& port, float& lower, float& upper);

    virtual string getLADSPAPluginRealNameByName(const string& name);
    virtual string getLADSPAPluginRealNameByIndex(const unsigned int& index);

    virtual string getLADSPAPluginLabelByName(const string& name);
    virtual string getLADSPAPluginLabelByIndex(const unsigned int& index);

    virtual string getLADSPAPluginLibraryNameByName(const string& name);
    virtual string getLADSPAPluginLibraryNameByIndex(const unsigned int& index);

  protected:
    /* INode interface */
    virtual Result doInit(void);
    virtual Result doDeinit(void);
    virtual Result doInitOutput(void);
    virtual Result doDeinitOutput(void);
    virtual Result doActivate(void);
    virtual Result doDeactivate(void);
    virtual Message* processBuffer(Buffer* in_buffer);
    
  private:
    std::string configFilename;                         /* filename of used config file   */
    std::vector<LADSPA_Plugin_t*> LADSPA_Plugins;       /* list with added LADSPA_Plugins */
    std::map<std::string,unsigned int>  LADSPA_NameMap; /* converter exported name->list index */
    
    int samplerate;     /* samplerate of delivered audio-data */
    int numInChannels;  /* number of input channels specified through the input format */
    int numOutChannels; /* number of output channels depending on used LADSPA-Plugins */	

    /**
     * count the number of ports for a specified porttype.
     * \param desc plugin descriptor to enumerate
     * \param porttype the type of ports to enumerate
     * \returns number of ports with specified type
     */
    unsigned int getPortCountByType(const LADSPA_Descriptor *desc,unsigned int porttype);

    LADSPA_Plugin_t* getLADSPAPlugin(const string& name) throw (const Exception&);    
    LADSPA_Plugin_t* getLADSPAPlugin(const unsigned int index) throw (const Exception&);

    LADSPA_Data clampCIValue(const LADSPA_Plugin_t *plugin,const unsigned int port,LADSPA_Data value);
    unsigned int clampCIPort(const LADSPA_Plugin_t *plugin,unsigned int port);

    LADSPA_Plugin_t* loadLADSPAPlugin(const string& value) throw (const Exception&);
    void unloadLADSPAPlugin(LADSPA_Plugin_t *plugin);

    void initLADSPAPlugin(LADSPA_Plugin_t *plugin) throw (const Exception&);
    void deinitLADSPAPlugin(LADSPA_Plugin_t *plugin);

    
    /**
     * load presets for the specified LADSPA plugin
     * \param pidx
     * \returns false if there was an error, true otherwise
     */
    bool loadPresets(LADSPA_Plugin_t *plugin,const string& profile);
    void setControlInBase(LADSPA_Plugin_t *plugin, const unsigned int& port, const float& value);
  }; // class LADSPANode
    
}; // namespace NMM

#endif // NMM_LADSPANODE
