/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2004
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "FFMpegAudioEncodeNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include "ffmpeg/avcodec.h"

NMM_REGISTER_PLUGIN(NMM::FFMpegAudioEncodeNode, "FFMpegAudioEncodeNode")

namespace NMM {

  FFMpegAudioEncodeNode::FFMpegAudioHandle::FFMpegAudioHandle() : ffAVCodec(0), ffAVCC(0) { }

  FFMpegAudioEncodeNode::FFMpegAudioEncodeNode(const char* name /*="FFMpegAudioEncodeNode"*/,
                                               const StreamQueue::Mode queue_mode /*= StreamQueue::MODE_SUSPEND*/,
                                               const int queue_size /*= StreamQueue::DEFAULT_SIZE*/) :
                                               GenericProcessorNode(name,queue_mode,queue_size),
                                               chunk_size(0),
                                               old_size(0),
                                               old_data(NULL)
  {
    mFFMpeg    = new FFMpegAudioHandle;

    // initialize libavcodec
    avcodec_init();
    avcodec_register_all();

    // check if the uses libavcodec has compiled in support for MP3 (via external libmp3lame)
    hasMP3Support = (avcodec_find_encoder(CODEC_ID_MP3)   != NULL);

  } // FFMpegAudioEncodeNode::FFMpegAudioEncodeNode



  FFMpegAudioEncodeNode::~FFMpegAudioEncodeNode() {
    delete mFFMpeg;
  }  // FFMpegAudioEncodeNode::~FFMpegAudioEncodeNode

  Result FFMpegAudioEncodeNode::doInit(void) {

    // input format -> raw audio samples
    Format *in_format = getOwnInputProperty()->addNewFormat(Format::audio_raw_type);
    in_format->addWildcard(Format::sample_rate_param);
    in_format->addWildcard(Format::channels_param);
    in_format->addIntParamValue(Format::bitpersample_param,  8);
    in_format->addIntParamValue(Format::bitpersample_param, 16);
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::encoding_param, "linear");
    in_format->addStringParamValue(Format::sign_param, "signed");
    getOwnInputProperty()->setDefaultFormat(in_format);


    // output format -> mpeg audio layer 2 & 3
    Format *out_format = getOwnOutputProperty()->addNewFormat(Format::audio_mpeg_type);
    out_format = getOwnOutputProperty()->addNewFormat(Format::audio_mpeg_type);
    out_format->addWildcard(Format::bitrate_param);
    out_format->addIntParamValue(Format::layer_param, 2);
    if (hasMP3Support)
      out_format->addIntParamValue(Format::layer_param, 3);


    // output format -> ac3
    out_format = getOwnOutputProperty()->addNewFormat("audio/ac3");
    out_format->addWildcard(Format::bitrate_param);

    return SUCCESS;
  } // FFMpegAudioEncodeNode::doInit


  Result FFMpegAudioEncodeNode::doDeinit() {
    getOwnInputProperty()->erase();
    getOwnOutputProperty()->erase();

    return SUCCESS;
  } // FFMpegAudioEncodeNode::doDeinit


  Result FFMpegAudioEncodeNode::doInitOutput() {
    const Format *in_format = getInputFormat();
    if (!in_format) {
      ERROR_STREAM("no input format specified" << std::endl);
      return FAILURE;
    }

    Format *i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);

    // output format -> mpeg audio layer 2 & 3
    Format *out_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_mpeg_type);
    out_format->addWildcard(Format::bitrate_param);
    out_format->addIntParamValue(Format::layer_param, 2);
    if (hasMP3Support)
      out_format->addIntParamValue(Format::layer_param, 3);


    // output format -> ac3
    out_format = getOwnWorkingOutputProperty()->addNewFormat("audio/ac3");
    out_format->addWildcard(Format::bitrate_param);

    i_format->setIOPartner(out_format);
    out_format->setIOPartner(i_format);

    return SUCCESS;
  } // FFMpegAudioEncodeNode::doInitOutput


  Result FFMpegAudioEncodeNode::doDeinitOutput() {
    if (mFFMpeg->ffAVCC) {
      delete mFFMpeg->ffAVCC;
      mFFMpeg->ffAVCC = 0;
    }

    getWorkingInputProperty()->erase();
    getWorkingOutputProperty()->erase();
    return SUCCESS;
  } // FFMpegAudioEncodeNode::doDeinitOutput


  Result FFMpegAudioEncodeNode::doActivate() {
    const Format *i_format = getInputFormat();
    const Format *o_format = getOutputFormat();
    if (!o_format) {
      ERROR_STREAM("no output format specified" << std::endl);
      return FAILURE;
    }

    std::string of_type = o_format->getFormatType();

    if (Format::audio_mpeg_type == of_type) {
      if (o_format->isParameterIncluded(Format::layer_param)) {
        if (3 == o_format->getIntValue(Format::layer_param)) {
          if (hasMP3Support) {
            mFFMpeg->ffAVCodec = avcodec_find_encoder(CODEC_ID_MP3);
          } else {
            ERROR_STREAM("mpeg layer 3 requested but not compiled in");
            return FAILURE;
          }
        } else {
          mFFMpeg->ffAVCodec = avcodec_find_encoder(CODEC_ID_MP2);
        }
      } else {
	ERROR_STREAM("mpeg layer not specified");
        return FAILURE;
      }
    } else if (Format::audio_ac3_type == of_type) {
      mFFMpeg->ffAVCodec = avcodec_find_encoder(CODEC_ID_AC3);
    }

    if (!mFFMpeg->ffAVCodec) {
      ERROR_STREAM("unsupported Codec for format type " << of_type << std::endl);
      return FAILURE;
    } else {
      DEBUG_STREAM("codec for format type " << of_type << " found" << std::endl);
    }

    mFFMpeg->ffAVCC = avcodec_alloc_context();
    mFFMpeg->ffAVCC->bit_rate    = o_format->getIntValue(Format::bitrate_param);
    mFFMpeg->ffAVCC->sample_rate = i_format->getIntValue(Format::sample_rate_param);
    mFFMpeg->ffAVCC->channels    = i_format->getIntValue(Format::channels_param);

    if (0 > avcodec_open(mFFMpeg->ffAVCC,mFFMpeg->ffAVCodec)) {
      ERROR_STREAM("could not open codec" << std::endl);
      return FAILURE;
    }

    // compute the amount of bytes avcodec_encode_audio will consume in each call
    chunk_size = static_cast<int>(mFFMpeg->ffAVCC->frame_size * mFFMpeg->ffAVCC->channels * (i_format->getIntValue(Format::bitpersample_param) / 8.0f));
    old_data = new char[chunk_size];

    return SUCCESS;
  } // FFMpegAudioEncodeNode::doActivate


  Result FFMpegAudioEncodeNode::doDeactivate(void) {
    avcodec_close(mFFMpeg->ffAVCC);
    delete [] old_data;
    return SUCCESS;
  } //FFMpegAudioEncodeNode::doDeactivate



  Message* FFMpegAudioEncodeNode::processBuffer(Buffer *in_buffer) {
    if (!in_buffer) {
      return NULL;
    }

    Buffer* out = getNewBuffer(in_buffer->getUsed(),in_buffer);

    unsigned out_size(0);
    unsigned pos(0);

    if (old_size) {
      // paranoia check: do we get enough data? - ought to be dead code
      if (chunk_size - old_size > in_buffer->getUsed()) {
        memcpy(old_data + old_size, in_buffer->getData(), in_buffer->getUsed());
        old_size += in_buffer->getUsed();
        in_buffer->release();
        return NULL;
      }

      // encode the remains of the old buffer
      memcpy(old_data + old_size, in_buffer->getData(), chunk_size - old_size);
      pos = chunk_size - old_size;
      old_size = 0;
      out_size += avcodec_encode_audio(mFFMpeg->ffAVCC,reinterpret_cast<unsigned char*>(out->getData()),
                                       chunk_size,     reinterpret_cast<short int*>(old_data));
    }

    // each call to avcodec_encode_audio takes only mFFMpeg->ffAVCC->frame_size samples or chunk_size bytes
    while (pos + chunk_size <= in_buffer->getUsed()) {
      out_size += avcodec_encode_audio(mFFMpeg->ffAVCC, reinterpret_cast<unsigned char*>(out->getData() + out_size),
                                       chunk_size,      reinterpret_cast<short int*>(in_buffer->getData() + pos));
      pos += chunk_size;
    }

    // if the input buffer is not yet completely coded, save the remaining bits
    if (pos < in_buffer->getUsed()) {
      memcpy(old_data, in_buffer->getData() + pos, in_buffer->getUsed() - pos);
      old_size = in_buffer->getUsed() - pos;
    }

    out->setUsed(out_size);

    in_buffer->release();

    return out;
  } // FFMpegAudioEncodeNode::processBuffer

} //namespace NMM
