/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef _NMM_SCOPE_NODE_HPP_
#define _NMM_SCOPE_NODE_HPP_

#include "nmm/base/GenericProcessorNode.hpp"

#include "nmm/interfaces/base/sync/ISyncResetImpl.hpp"
#include "nmm/interfaces/audio/visualization/IScopeImpl.hpp"

namespace NMM {

  //! Oszilloscope Node.
  class ScopeNode :
    public GenericProcessorNode,
    public IScopeImpl,
    public ISyncResetImpl
  {
  private:
    // Input parameters
    int input_channels;    //!< The number of channels in the audio input.
    int sample_rate;       //!< The audio input sample rate.
    int byte_per_sample;
    

    // Output parameters
    int image_height;      //!< The image width and height chosen with the
    int image_width;       //!< output format.
    int range;             //!< The resulting array size for an image.
    double framerate;

    // Internally used variables
    Buffer* old_in_buffer; //!< The last dequeued in_buffer.
    int curr_audio;        //!< The current position in the audio buffer.
    int curr_video;        //!< The current column position in the video
                           //!< frame.
    int gap;               //!< The number of audio bytes to leave out between
                           //!< two frames.
    int audio_end;         //!< The number of bytes in the actual audio buffer.
    char* audio;           //!< The audio data.
    
    unsigned char* image;  //!< Used to compute the next image; it is 
                           //!< necessary to store the image because of
                           //!< the blur effect.

    bool blur_enabled;     //!< True iff the blur effect is enabled
    int blur_step;         //!< The value to subtract from the last frame
                           //!< for blur effect.
    
    Timestamp valid_ts;    //!< Last valid timestamp
    Timestamp in_ts;       //!< Timestamp of current input buffer
    Timestamp out_ts;      //!< Timestamp of next output buffer
    int valid_samples;     //!< Audio samples since last valid timestamp
    int in_samples;        //!< Audio samples since    
    
    //! Sets a point at position (x, y) in the image buffer.
    void setPoint(unsigned char* image, int x, int y);
    
  protected:
    /**
     * Reset synchronization.
     */
    Result syncReset();
    
    /**
     * Enables or disables the blur effect.
     */
    Result setBlurEnabled(const bool& toggle);
    
    /**
     * sets the intensity of the blur effect.
     */
    Result setBlurStep(const int& step);
    
    //! The doInit - method from Node.
    Result doInit();
    
    //! The doDeinit - method from Node.
    Result doDeinit();
    
    //! The doInitOutput - method from Node.
    Result doInitOutput();
    
    //! The doDeinitOutput - method from Node.
    Result doDeinitOutput();
    
    //! The doActivate - method from Node.
    Result doActivate();
    
    //! The doDeactivate - method from Node.
    Result doDeactivate();
    
    //! The processBuffer - method from GenericNode.
    Buffer* processBuffer(Buffer* in_buffer);
    
  public:
    //! The main constructor.
    ScopeNode(const char *name = "Scope Node",
	      const StreamQueue::Mode queue_mode = StreamQueue::DEFAULT_MODE,
	      const int size = StreamQueue::DEFAULT_SIZE);

    //! The main destructor.
    virtual ~ScopeNode();
  };

}

#endif // _NMM_SCOPE_NODE_HPP_


