/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    Network-Multimedia Workgroup,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:	      Roger Dostert <djurig@graphics.cs.uni-sb.de 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#include <string.h>
#include "FFAVFormat.hpp"
#include "ffmpeg/avformat.h"

namespace NMM {
	
	// class implementation
	
	FFAVFormat::FFAVFormat(void) {
	// initialize class members
	
		streamCount   = 0;
		
		v_stream = NULL;
		a_stream = NULL;
		
		buffer        = new unsigned char[4096];
		io_buffer     = NULL;
		io_buffersize = 0;
		io_bufferused = 0;
		
		a_buffers.clear();
		v_buffers.clear();
		
		n_vbuffer = 0;
		n_abuffer = 0;
		g_vbuffer = 0;
		g_abuffer = 0;
		
		// setup format context
		av_register_all();
		
		fmtctx = (AVFormatContext*)(av_malloc(sizeof(AVFormatContext)));
		fmtctx->oformat = NULL;
		
		url_open_buf(&fmtctx->pb,buffer,4096,URL_WRONLY); // the URL_WRONLY is important; otherwise there is no data written
		fmtctx->pb.write_packet = io_writePacket;
		fmtctx->pb.opaque = this;
	} // FFAVFormat::FFAVFormat()
	
	
	FFAVFormat::~FFAVFormat(void) {
	  if (a_stream) {
    	avcodec_close(&a_stream->codec);
	  }
	
		if (v_stream) {
			avcodec_close(&v_stream->codec);
		}
		
		delete[] io_buffer;
		
		if (fmtctx) {
		  url_close_buf(&fmtctx->pb);
		  delete fmtctx->oformat;
			delete fmtctx;
		}
		
		delete[] buffer;
	} // FFAVFormat::~FFAVFormat()
	
	
	void FFAVFormat::addAudioStream(int codecId,int bitrate,int samplerate,int channels) {
		if (a_stream) {
			GLOBAL_ERROR_STREAM("there is already an audio stream" << std::endl);
			return;
  	}
	
		streamCount++;
		a_stream = av_new_stream(fmtctx,streamCount);
		if (!a_stream) { 
			GLOBAL_ERROR_STREAM("could not allocate menory for  audio stream" << std::endl); 
			streamCount--;
			return;
	  }
	
		AVCodecContext* ctx;
		ctx              = &(a_stream->codec);
		ctx->codec_id    = (CodecID)codecId;
		ctx->codec_type  = CODEC_TYPE_AUDIO;
		ctx->bit_rate    = bitrate;
		ctx->sample_rate = samplerate;
		ctx->channels    = channels;
		avcodec_open(ctx,avcodec_find_encoder((CodecID)codecId));
		
		GLOBAL_DEBUG_STREAM("added audio stream" << std::endl);
	} // FFAVFormat::addAudioStream(CodecID,int,int,int)
	
	
	void FFAVFormat::addVideoStream(int codecId,int bitrate,int width,int height,float framerate) {
		if (v_stream) {
			GLOBAL_ERROR_STREAM("there is already a video stream" << std::endl);
			return;
		}
		
		streamCount++;
		v_stream = av_new_stream(fmtctx,streamCount);
		if (!v_stream) { 
			GLOBAL_ERROR_STREAM("could not allocate memory for video stream" << std::endl); 
			streamCount--;
			return;
		}
		
		AVCodecContext* c;
		c                  = &(v_stream->codec);
		c->codec_id        = (CodecID)codecId;
		c->codec_type      = CODEC_TYPE_VIDEO;
		c->bit_rate        = bitrate;
		c->width           = width;
		c->height          = height;
		c->frame_rate      = int(framerate*1000);
		c->frame_rate_base = 1000;
		avcodec_open(c,avcodec_find_encoder((CodecID)codecId));
		
		GLOBAL_DEBUG_STREAM("added video stream" << std::endl);
	} // void FFAVFormat::addVideoStream(CodecID,int,int,int,int)
	
	
	void FFAVFormat::writeHeader(void) {
		av_write_header(fmtctx);
	} // FFAVFormat::writeHeader()
	
	
	void FFAVFormat::writeTrailer(void) {
		av_write_trailer(fmtctx);
	} // FFAVFormat::writeTrailer()
	
	
	int FFAVFormat::writeAudioFrame() {
		if (a_buffers.empty()) {
		  return NB_NEED_A;
	  }
		n_abuffer++;
		Buffer* b = a_buffers.front();
		AVPacket pkt;
		av_init_packet(&pkt);
		pkt.size = b->getUsed();
		pkt.pts = a_stream->pts.val;
		//pkt.flags = ?
		pkt.stream_index = a_stream->index;
		pkt.data = (uint8_t*)b->getData();
		
		if (av_write_frame(fmtctx, &pkt) != 0)
			GLOBAL_ERROR_STREAM("FFAVFormat::writeAudioFrame - Error while writing audio frame\n");
		
		b->release();
		a_buffers.pop_front();
		
		return NB_NONE;
	} // FFAVFormat::writeAudioFrame()
	
	
	int FFAVFormat::writeVideoFrame() {
		if (v_buffers.empty()) {
		  return NB_NEED_V;
	  }
		
		n_vbuffer++;
		Buffer *b = v_buffers.front();

		AVPacket pkt;
		av_init_packet(&pkt);
		pkt.size = b->getUsed();
		pkt.pts = v_stream->pts.val;
		//pkt.flags = ?
		pkt.stream_index = v_stream->index;
		pkt.data = (uint8_t*)b->getData();
		
		if (av_write_frame(fmtctx, &pkt) != 0)
			GLOBAL_ERROR_STREAM("FFAVFormat::writeAudioFrame - Error while writing video frame\n");

				
		b->release();
		v_buffers.pop_front();
		
		
		return NB_NONE;
	} // FFAVFormat::writeVideoFrame(const char*,int)
	
	
	int FFAVFormat::muxFrame(Buffer* buffer,bool video) {
		double a_pts,v_pts;
		int ret = NB_NONE;
		
		if (buffer) {
			if (video) {
				g_vbuffer++;
				v_buffers.push_back(buffer);
			} else {
				g_abuffer++;
				a_buffers.push_back(buffer);
			}
			
			while (NB_NONE == ret) {
				// get current timestamps
				if (v_stream) {
					v_pts = (double)v_stream->pts.val * v_stream->time_base.num / v_stream->time_base.den;
				} else {
					v_pts = 0;
				}
				
								
				if (a_stream) {
					a_pts = (double)a_stream->pts.val * a_stream->time_base.num / a_stream->time_base.den;
				} else {
					a_pts = 0;
				}
				
				// write next buffer
				if (!v_stream || (v_stream && a_stream && (a_pts < v_pts)) ) {
					ret |= writeAudioFrame();
				} else {
					ret |= writeVideoFrame();
				}
			} // while (true)
		} else {
			GLOBAL_DEBUG_STREAM("=====> end of stream reached : write out remaining buffers" << std::endl);
			if (v_buffers.empty() ) { 
				GLOBAL_DEBUG_STREAM("no video buffers left" << std::endl); 
				ret |= NB_NEED_V;
			}
			if (a_buffers.empty() ) { 
				GLOBAL_DEBUG_STREAM("no audio buffers left" << std::endl); 
				ret |= NB_NEED_A;
			}
		
			switch (ret) {
				case NB_NEED_V : // write remaining audio buffers
					while (!a_buffers.empty()) {
						writeAudioFrame();
				  }
					ret |= NB_NEED_A;
					break;
				
				case NB_NEED_A: // write remaining video buffers
					while (!v_buffers.empty()) {
						writeVideoFrame();
				  }
					ret |= NB_NEED_V;
					break;
			
				case NB_NEED_AV: // no buffers left -> all done
					break;
				
				default:
					GLOBAL_ERROR_STREAM("FFAVFormat::muxFrame() : invalid multiplex state" << std::endl);
					return NB_ERROR;
			}
		}
		
		return ret;
	} // FFAVFormat::muxFrame(const char*,int,bool)
	
	
	int FFAVFormat::readBack(char* data,const int size) {
		int len = (io_bufferused > size) ? size : io_bufferused;
		
		// fill target buffer
		memcpy(data,io_buffer,len);
		
		// shift remaining data to front
		memmove(io_buffer,io_buffer+len,io_bufferused-len);
		io_bufferused -= len;
		
		return len;
	} // FFAVFormat::readBack(const char*,const int)
	
	
	void FFAVFormat::setOutputFormat(const char* short_name) {
		delete fmtctx->oformat;
		
		fmtctx->oformat = findOutputFormat(short_name,strlen(short_name));
		if (!fmtctx->oformat) { 
			GLOBAL_ERROR_STREAM("=====>  requested Output Format (" << short_name << ") not found" << std::endl); 
		}
		
		av_set_parameters(fmtctx,NULL);
	}
		
		
  AVOutputFormat* FFAVFormat::findOutputFormat(const char* short_name,int len) {
		AVOutputFormat* ret = NULL;
		AVOutputFormat* of = first_oformat;
		
		while (of) {
			if (0 == strncmp(of->name,short_name,len)) {
				ret = new AVOutputFormat;
				memcpy(ret,of,sizeof(AVOutputFormat));
				ret->next = NULL;
				break;
			} 
			of = of->next;
		}
		
		return ret;
	} // FFAVFormat::getOutputFormat(const char*,int)
	
	
	void FFAVFormat::io_writePacket(void* opaque,unsigned char* buf,int size) {
		unsigned char* b = NULL;
		FFAVFormat* _instance = reinterpret_cast<FFAVFormat*>(opaque);
		
		if ((_instance->io_buffersize - _instance->io_bufferused) < size) {
			_instance->io_buffersize += size;
			b = new unsigned char [_instance->io_buffersize];
			
			if (_instance->io_buffer) {
				memcpy(b,_instance->io_buffer,_instance->io_bufferused);
				delete[] _instance->io_buffer;
			}
		} else {
			b = _instance->io_buffer;
		}
		
		memcpy(b+_instance->io_bufferused,buf,size);
		_instance->io_bufferused += size;
		_instance->io_buffer = b;
	} // FFAVFormat::io_writePacket(void*,unsigned char*,int) {

} // namespace NMM
