/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    Network-Multimedia Workgroup,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:	      Roger Dostert <djurig@graphics.cs.uni-sb.de
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "nmm/plugins/av/mpeg/FFMpegMuxNode.hpp"
#include "nmm/plugins/av/mpeg/FFAVFormat.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include <ffmpeg/avformat.h>

NMM_REGISTER_PLUGIN(NMM::FFMpegMuxNode, "FFMpegMuxNode")

namespace NMM {

  /*
   * FFMpegMuxNode
   */
  

  FFMpegMuxNode::FFMpegMuxNode(const char* name /*="FFMpegMuxNode"*/,
			       StreamQueue::Mode qmode/*= StreamQueue::MODE_SUSPEND*/,
			       int qsize /*= StreamQueue::DEFAULT_SIZE*/) 
    : GenericMultiplexerNode(name,qmode,qsize) {
    
    // initialize members
    ffAVFormat         = NULL;

    firstFrame         = false;
    lastFrame          = false;
    receivedStartTrack = 0;
    receivedEndTrack   = 0;
  } // FFMpegMuxNode::FFMpegMuxNode  


  FFMpegMuxNode::~FFMpegMuxNode(void) {
    if (ffAVFormat) { delete ffAVFormat; }
  }

  
  Result FFMpegMuxNode::doInit(void) {

    // add input streams : one for audio, one for video

    if (FAILURE == addInputStream("audio")) {
      ERROR_STREAM("Failed to add input stream 'audio'" << std::endl);
      return FAILURE;
    }

    if (FAILURE == addInputStream("video")) {
      ERROR_STREAM("Failed to add input stream 'video'" << std::endl);

      return FAILURE;
    }

    // video subformats

    // MPEG-video
    Format v_format1(Format::video_mpeg_type);
    v_format1.addWildcard(Format::version_param);
    v_format1.addWildcard(Format::x_resolution_param);
    v_format1.addWildcard(Format::y_resolution_param);
    v_format1.addWildcard(Format::ratio_x_param);
    v_format1.addWildcard(Format::ratio_y_param);
    v_format1.addWildcard(Format::framerate_param);
    v_format1.addWildcard(Format::bitrate_param);
    // audio subformats

    // MPEG Audio layer 2
    Format a_format1(Format::audio_mpeg_type);
    a_format1.addIntParamValue(Format::audio_codec_param,0x55); // 0x55 stands for MPEG compressed Audio
    a_format1.addIntParamValue(Format::layer_param,2);
    a_format1.addWildcard(Format::bitrate_param);
    a_format1.addWildcard(Format::channels_param);
    a_format1.addWildcard(Format::sample_rate_param);
    a_format1.addWildcard(Format::bitpersample_param);

    // MP2 with bitrate only
    Format a_format2(Format::audio_mpeg_type);
    a_format2.addWildcard(Format::bitrate_param);
    a_format2.addIntParamValue(Format::layer_param,2);

    // MPEG Audio layer 3
    //    Format a_format2(Format::audio_mpeg_type);
    //    a_format1.addIntParamValue(Format::audio_codec_param,0x55); // 0x55 stands for MPEG compressed Audio
    //    a_format1.addIntParamValue(Format::layer_param,3);
    //    a_format1.addWildcard(Format::bitrate_param);
    //    a_format1.addWildcard(Format::channels_param);
    //    a_format1.addWildcard(Format::sample_rate_param);
    //    a_format1.addWildcard(Format::bitpersample_param);



    // video input formats

    Format* in_v_format = new Format(v_format1);
    in_v_format->setJackTag("video");
    getOwnInputProperty()->addNewFormat(in_v_format);
    getOwnInputProperty()->setDefaultFormat(in_v_format,"video");

    // audio input formats

    Format* in_a_format = new Format(a_format1);
    in_a_format->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format);
    getOwnInputProperty()->setDefaultFormat(in_a_format,"audio");

    in_a_format = new Format(a_format2);
    in_a_format->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format);


    // composite output formats

    // Mpeg-Video with MP2 Audio
    Format* out_format = getOwnOutputProperty()->addNewFormat(Format::av_mpeg_type);
    out_format->addFPointerParamValue(Format::video_format_param,new Format(v_format1));
    out_format->addFPointerParamValue(Format::audio_format_param,new Format(a_format1));
    getOwnOutputProperty()->setDefaultFormat(out_format);

    out_format = getOwnOutputProperty()->addNewFormat(Format::av_mpeg_type);
    out_format->addFPointerParamValue(Format::video_format_param,new Format(v_format1));
    out_format->addFPointerParamValue(Format::audio_format_param,new Format(a_format2));

    // Mpeg-Video only
    out_format = getOwnOutputProperty()->addNewFormat(Format::av_mpeg_type);
    out_format->addFPointerParamValue(Format::video_format_param,new Format(v_format1));

    // MP2 only
    out_format = getOwnOutputProperty()->addNewFormat(Format::av_mpeg_type);
    out_format->addFPointerParamValue(Format::audio_format_param,new Format(a_format1));

    // MP3 only
    //    out_format = getOwnOutputProperty()->addNewFormat(Format::av_mpeg_type);
    //    out_format->addFPointerParamValue(Format::audio_format_param,new Format(a_format2));


    return SUCCESS;
  } // FFMpegMuxNode::doInit


  Result FFMpegMuxNode::doDeinit(void) {
    getOwnInputProperty()->erase();
    getOwnOutputProperty()->erase();

    return SUCCESS;
  } // FFMpegMuxNode::doDeinit


  Result FFMpegMuxNode::doInitOutput(void) {
    // set the output format in dependency of the input format

    // get input formats
    const Format* in_v_format = getInputFormat("video");
    const Format* in_a_format = getInputFormat("audio");

    // init working input property
    Format* i_v_format = NULL;
    Format* i_a_format = NULL;

    // initialize libavformat
    if (!ffAVFormat) ffAVFormat = new FFAVFormat;


    if (in_v_format) {
      int w,h,b;
      float f;

      i_v_format = new Format(*in_v_format);
      getOwnWorkingInputProperty()->addNewFormat(i_v_format);
      i_v_format->setJackTag("video");
      getOwnWorkingInputProperty()->setDefaultFormat(i_v_format,"video");

      w = i_v_format->getIntValue(Format::x_resolution_param);
      h = i_v_format->getIntValue(Format::y_resolution_param);
      b = i_v_format->getIntValue(Format::bitrate_param);

      f = i_v_format->getFloatValue(Format::framerate_param);

      ffAVFormat->addVideoStream(CODEC_ID_MPEG2VIDEO,b,w,h,f);
    }



    if (in_a_format) {
      int b,s,c;

      i_a_format = new Format(*in_a_format);
      getOwnWorkingInputProperty()->addNewFormat(i_a_format);
      i_a_format->setJackTag("audio");
      getOwnWorkingInputProperty()->setDefaultFormat(i_a_format,"audio");
      
      b = i_a_format->getIntValue(Format::bitrate_param);
      if (i_a_format->isParameterIncluded(Format::sample_rate_param)) {
        s = i_a_format->getIntValue(Format::sample_rate_param);
      } else {
        s = 44100;
      }

      if (i_a_format->isParameterIncluded(Format::channels_param)) {
        c = i_a_format->getIntValue(Format::channels_param);
      } else {
        c = 2;
      }

      ffAVFormat->addAudioStream(CODEC_ID_MP2,b,s,c);
    }

    // output format
    Format* format = getOwnWorkingOutputProperty()->addNewFormat(Format::av_mpeg_type);
    if (i_v_format) {
      format->addFPointerParamValue(Format::video_format_param,new Format(*i_v_format));
    }
    if (i_a_format) {
      format->addFPointerParamValue(Format::audio_format_param,new Format(*i_a_format));
    }
    getOwnWorkingOutputProperty()->setDefaultFormat(format);


    // configure streams for libavformat
    ffAVFormat->setOutputFormat("vob");

    return SUCCESS;
  } //  FFMpegMuxNode::doInitOutput


  Result FFMpegMuxNode::doDeinitOutput(void) {
    if (ffAVFormat) {
      delete ffAVFormat;
      ffAVFormat = NULL;
    }

    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    return SUCCESS;
  } // FFMpegMuxNode::doDeinitOutput


  Result FFMpegMuxNode::doActivate(void) {
    return SUCCESS;
  } // FFMpegMuxNode::doActivate


  Result FFMpegMuxNode::doDeactivate(void) {
    return SUCCESS;
  } // FFMpegMuxNode::doDeactivate


  Result FFMpegMuxNode::doStart(void) {
    return SUCCESS;
  } // FFMpegMuxNode::doStart


  Result FFMpegMuxNode::doStop(void) {
    return SUCCESS;
  } // FFMpegMuxNode::doStop


  Message* FFMpegMuxNode::processBuffer(Buffer* inbuffer) {
    bool is_video;

    if (lastFrame) {
      setWorkingFlag(false);
      return new CEvent(Event(ITrack::endTrack_event));
    }

    if (inbuffer) {
      if (firstFrame) {
	firstFrame = false;
	ffAVFormat->writeHeader();
      } 
    
      if (0 == strcmp(getCurrentRecvInputStream(),"audio")) {
	is_video = false;
      } else if (0 == strcmp(getCurrentRecvInputStream(),"video")) {
	is_video = true;
      } else {
	ERROR_STREAM("undefined input stream (THIS SHOULD NOT HAPPEN)" << std::endl);
	inbuffer->release();
	return NULL;
      }
      ffAVFormat->muxFrame(inbuffer,is_video);
    } else {
      switch (ffAVFormat->muxFrame(NULL,false) ) { // in the case of a NULL-buffer the second parameter is ignored
	case NB_NEED_AV:
	  lastFrame = true;
	  break;
	  
	default:
	  ERROR_STREAM("invalid result from ffAVFormat->muxframe (THIS SHOULD NOT HAPPEN)" << std::endl);
	  return NULL;
      }
    }

    // read back data
    Buffer* outbuffer = getNewBuffer(ffAVFormat->getIOBufferSize());
    int bytes = ffAVFormat->readBack(outbuffer->getData(),outbuffer->getSize());
    outbuffer->setUsed(bytes);
    ERROR_STREAM("=====> muxed out " << bytes << " bytes" << std::endl);

    return outbuffer;
  } // FFMpegMuxNode::processBuffer


    Result FFMpegMuxNode::startTrack(const std::string& /*filename*/) {
    ++receivedStartTrack;
    if (receivedStartTrack < 2) {
      return DELETE;
    } else {
      firstFrame = true;
      lastFrame = false;
      receivedEndTrack = 0;
      return SUCCESS;
    }

  } // FFMpegMuxNode::startTrack


  Result FFMpegMuxNode::endTrack(void) {
    ++receivedEndTrack;
    if (2 == receivedEndTrack) {
      setWorkingFlag(true);
    } else {
      receivedStartTrack = 0;
    }

    return DELETE;
  } // FFMpegMuxNode::endTrack


  Result FFMpegMuxNode::finished(void) {
    return DELETE;
  } // FFMpegMuxNode::finished

}; // namespace NMM
