/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "WavReadNode.hpp"

#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/base/sync/ISyncReset.hpp"
#include "nmm/interfaces/general/ITrackDuration.hpp"
#include "nmm/interfaces/general/progress/IProgressListenerImpl.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/base/registry/Plugin.hpp"

#include <iomanip>

NMM_REGISTER_PLUGIN(NMM::WavReadNode, "WavReadNode")

namespace NMM {

  WavReadNode::WavReadNode(const char *name /* = "WavReadNode" */,
		   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
		   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size),
      buffer_size(32*1024),
      filename(),
      channels(0),
      bytes_per_sample(0),
      bytes_per_second(0),
      filesize(0),
      filepos(0),
      seekto(0),
      track_duration(0, 0),
      state(START),
      send_format(false),
      send_progress(false),
      buffers_before_progress(5),
      buffers_progress(0) {
  }
  
  WavReadNode::~WavReadNode() {
    in.close();
  }
  
  void WavReadNode::setFilename(const string& name) {
    filename = name;
  }  
  
  void WavReadNode::setBufferSize(const int& size) {
    buffer_size = size;
  }
  
  int WavReadNode::getBufferSize() {
    return buffer_size;
  }

  Result WavReadNode::doInitOutput() {
    unsigned char *b = new unsigned char[160];

    long int sample_rate;
    long int bits_per_sample;//, bitrate;
    long int pcm_length;
    
    /* --- open the file ----------------------------------- */
    in.clear(); 
    
    if ("" == filename) {
      ERROR_STREAM("Filename not set" << endl);
      return FAILURE;
    }
    
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    
    in.open(filename.c_str(), fstream::in | ios::binary);
    
    if (!in.is_open()) {
      ERROR_STREAM("Can\'t open " << filename << endl);
      return FAILURE;
    }
    
    
    // get filesize and seek to beginning again
    in.seekg(0, ios::end);
    filesize = in.tellg();
    in.seekg(0, ios::beg);

    /* --- analyse the header and set the out format ------- */

    /* read the first 160 Byte */
    in.read(reinterpret_cast<char*>(b), 160);

    /* 00 - 03 "RIFF" ASCII characters */
    if (!((b[0] == 'R') && (b[1] == 'I') && (b[2] = 'F') && (b[3] == 'F'))) {
      ERROR_STREAM("Error parsing wav header: can\'t find \'RIFF\' chunk" << endl);
      return FAILURE;
    }

    /* 12 - 15 "fmt_" ASCII characters */
    if (!((b[12] = 'f') && (b[13] = 'm') && (b[14] = 't') && (b[15] = '_'))) {
      ERROR_STREAM("Error parsing wav header: can\'t find \'FORMAT\' chunk" << endl);
      return FAILURE;
    }

    /* Byte 22-23 Channel numbers */
    switch (b[22] + (b[23] << 8)) {
    case 0x01: channels = 1; break; /* mono */
    case 0x02: channels = 2; break; /* stereo */
    default:
      ERROR_STREAM("Error can\'t determinate number of channels" << endl);
      return FAILURE;
    }

    /* 24 - 27 Sample rate (binary in Hz) */
    sample_rate = b[24] + (b[25] << 8) + (b[26] << 16) + (b[27] << 24);

    /* 28 - 31 Bytes per second */
    bytes_per_second = b[28] + (b[29] << 8) + (b[30] << 16) + (b[31] << 24);

    /* 32 - 33 Bytes per sample */
    bytes_per_sample = b[32] + (b[33] << 8);

    /* 34 - 35 Bits per sample */
    bits_per_sample = b[34] + (b[35] << 8);

    /* search 'DATA' Chunk */
    long int sdata = 36;
    while (!((b[sdata+0] == 'd') && (b[sdata+1] == 'a') && (b[sdata+2] == 't') && (b[sdata+3] == 'a'))) {
      sdata += 1;
      if (sdata > 155) {
        ERROR_STREAM("Error parsing wav header: can\'t find \'DATA\' chunk" << endl);
        return FAILURE;
      }
    }

    /* sdata+4 - sdata+7 Length of data to follow */
    pcm_length = b[sdata+4] + (b[sdata+5] << 8) + (b[sdata+6] << 16) + (b[sdata+7] << 24);

    // compute track duration
    track_duration = Interval(pcm_length / bytes_per_second, 0);

    in.seekg(sdata+8, ios::beg);  /* skip the header */

    // check if we are already connected
    if(isOutputConnected("default")) {
      send_format = true;
    }
    
    /* --- now define the formats -------------------------- */

    Format *out_format = getOwnWorkingOutputProperty()->addNewFormat("audio/raw");
    
    DEBUG_STREAM("Channels         : " << channels << endl);
    out_format ->addIntParamValue(Format::channels_param, channels);
    
    DEBUG_STREAM("Sample rate      : " << sample_rate << endl);
    out_format ->addIntParamValue(Format::sample_rate_param, sample_rate);
    
    DEBUG_STREAM("Bytes per second : " << bytes_per_second << endl);
    DEBUG_STREAM("Bytes per sample : " << bytes_per_sample << endl);
    DEBUG_STREAM("Bits per sample  : " << bits_per_sample << endl);
    out_format ->addIntParamValue(Format::bitpersample_param, bits_per_sample);
    
    DEBUG_STREAM("PCM length       : " << pcm_length << endl);
    
    int t = (int) pcm_length / bytes_per_second;
    DEBUG_STREAM("Playtime         : "
                 << setfill('0') << setw(2) << (int) (t / 3600) << ":"
                 << setfill('0') << setw(2) << (int) ((t % 3600) / 60) << ":"
                 << setfill('0') << setw(2) << (int) (t % 3600 % 60)
                 << endl);
    
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::encoding_param, "linear");
    out_format->addStringParamValue(Format::sign_param, "signed");
    
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    
    delete b;

    setProducingFlag(true);
    state = START;
    return SUCCESS;
  }
  
  Message *WavReadNode::processBuffer(Buffer *) {
    Buffer *out_buffer = 0;
    Event e;
    long int bytes_read;
    string event_str = "";
    string::size_type pos1;
    string::size_type pos2;

    if ( send_progress ) {
      Event e;
      if ( buffers_progress > buffers_before_progress ) {
        buffers_progress = 0;
        CEvent* cevent(new CEvent(IProgressListener::create_setProgress(filepos, filesize)));
        cevent->setMandatoryBit(false);
        return cevent;
      }
      buffers_progress++;
    }

    switch (state) {
    case START:
      {
        CEvent* ce = new CEvent();
        pos1 = filename.rfind("/");
        if (filename.npos == pos1) { 
          pos1=0; 
        } else {
          pos1++;
        }
        pos2 = filename.find(".wav",pos1);
        DEBUG_STREAM("Send startTrack event" << endl);
        ce->insertEvent(ITrack::create_startTrack(filename.substr(pos1,pos2-pos1)));
        DEBUG_STREAM("Sending syncReset" << endl);
        ce->insertEvent(ISyncReset::create_syncReset());
        DEBUG_STREAM("Sending trackDuration" << endl);
        ce->insertEvent(ITrackDuration::create_trackDuration(track_duration));

        if(send_format) {
          send_format = false;
         
          //set new format
          setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());

          DEBUG_STREAM("Sending setFormat" << endl);
          ce->insertEvent(IFormatChange::create_setFormat(*getOutputFormat()));
          
        }
        
        state = DATA;
        return ce;
      }
      break;

    case FINISHED:
      DEBUG_STREAM("Send finished event " << endl);
      e = Event(ITrack::finished_event);
      setProducingFlag(false);
      return new CEvent(e);
      break;

    case DATA:
      out_buffer = getNewBuffer(buffer_size);    
      if(!out_buffer) {
        ERROR_STREAM("Error allocating new buffer" << endl);
        return NULL;
      }

      if (seekto != 0) {
        // test if we would get negative
        if (seekto < 44 && labs(seekto) > static_cast<int64_t>(filepos)) {
          // if seek result would be negative go to beginning of file
          in.seekg(44, ios::beg);
          filepos = 44;
          DEBUG_STREAM("Seeked to beginning " << seekto << " from " << filepos << ", filesize " << filesize << endl);
        }
        else if (filepos + seekto > filesize) {
          // set back to short-before-file-end
          in.seekg(0, ios::end);
          DEBUG_STREAM("Seeked to end " << seekto << " from " << filepos << ", filesize " << filesize << endl);
        }
        else {
          // seek normally
          in.seekg(filepos + seekto);
          filepos = filepos+seekto;
          DEBUG_STREAM("Seeked normally " << seekto << " from " << filepos << ", filesize " << filesize << endl);
        }
        seekto = 0;

        CEvent* cevent = new FlushEvent();
        cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());
        return cevent;
      }

      in.read(out_buffer->getData(), buffer_size);
      bytes_read = in.gcount();

      if (0 < bytes_read) {
        out_buffer->setUsed(bytes_read); 
        filepos+=bytes_read;
      } 
      else if (0 == bytes_read) {
        DEBUG_STREAM("Send endTrack event " << endl);
        e = Event(ITrack::endTrack_event);
        state = FINISHED;
        return new CEvent(e);
        break;
      }
      else {
        ERROR_STREAM("Error reading file" << endl);
        out_buffer->setUsed(0); 
      }
      break;
    default:
       ERROR_STREAM("Internal state error" << endl);     
       exit(1);
    }

    return out_buffer;
  } /* produceBuffer */
  
  Result WavReadNode::doInit() {
    /*Get the output property and erase the old formats*/
    Property* property = getOwnOutputProperty();
    
    /*The doInit method stores all formats in the input and output property which are
      supported by this node. These properties are used to register a Node by a registry.  */
    Format* out_format = property->addNewFormat("audio/raw");
    out_format->addWildcard(Format::channels_param);
    out_format->addWildcard(Format::sample_rate_param);
    out_format->addWildcard(Format::bitpersample_param);
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::encoding_param, "linear");
    out_format->addStringParamValue(Format::sign_param, "signed");
    
    property->setDefaultFormat(out_format);
    return SUCCESS;
  }
  
  Result WavReadNode::doDeinitOutput() {
    getOwnWorkingOutputProperty()->erase();
    in.close();
    return SUCCESS;
  }

  Result WavReadNode::doDeinit( ) {
    /*Erase the Formats added in doRegister*/
    getOwnOutputProperty()->erase();
    return SUCCESS;
  } 

  /******************** ISeekable interface ********************/
  bool WavReadNode::hasPercentSeek() const {
    return true;
  }

  void WavReadNode::seekPercent(const Rational& percent) {
    Rational internal_percent(percent);
    if (internal_percent > Rational(1)) {
      ERROR_STREAM("Seeking more than 100% not possible, defaulting to end!" << endl);
      internal_percent = Rational(1);
    }
    else if (internal_percent < Rational(-1)) {
      ERROR_STREAM("Seeking less than -100% not possible, defaulting to beginning!" << endl);
      internal_percent = Rational(-1);
    }

    Rational rat_seekto = Rational(filesize) * internal_percent;
    seekto = rat_seekto.getNumerator() / rat_seekto.getDenominator();

    int modulo(seekto % (channels * bytes_per_sample));
    if (modulo) {
      seekto += (channels * bytes_per_sample) - modulo;
    }

    DEBUG_STREAM(seekto << " = " << internal_percent << " * " << filesize << endl);
  }
   
  void WavReadNode::seekPercentTo(const Rational& percent) {
    seekPercent(percent - Rational(filepos, filesize));
  }

  bool WavReadNode::hasBytesSeek() const {
    return true;
  }

  void WavReadNode::seekBytes(const int64_t& offset) {
    if ((offset < 0) && (labs(offset) > static_cast<int64_t>(filepos))) {
      ERROR_STREAM("Seeking to negativ byte-offset not possible, defaulting to beginning of file!" << endl);
      seekBytesTo(0);
    } else {
      seekBytesTo(filepos + offset);
    }
  }

  void WavReadNode::seekBytesTo(const u_int64_t& bytes) {
    u_int64_t internal_bytes(bytes);

    if (internal_bytes > filesize) {
      ERROR_STREAM("Seeking over EOF not possible, defaulting to EOF!" << endl);
      internal_bytes = filesize;
    }

    seekto = internal_bytes - filepos;

    int modulo(seekto % (channels * bytes_per_sample));
    if (modulo) {
      seekto += (channels * bytes_per_sample) - modulo;
    }

    DEBUG_STREAM(seekto << " = " << internal_bytes << " - " << filepos << endl);
  }

  bool WavReadNode::hasTimeSeek() const {
    return true;
  }

  void WavReadNode::seekTime(const Interval& interval) {
    Rational rat_interval(Rational(interval.sec) + Rational(interval.nsec, nano_per_sec));

    Rational rat_seekto(rat_interval * Rational(bytes_per_second));

    seekto = rat_seekto.getNumerator() / rat_seekto.getDenominator();

    int modulo(seekto % (channels * bytes_per_sample));
    if (modulo) {
      seekto += (channels * bytes_per_sample) - modulo;
    }

    DEBUG_STREAM(seekto << " = " << interval << " - " << filepos << endl);
  }

  void WavReadNode::seekTimeTo(const Time& time) {
    Rational rat_time(Rational(time.sec) + Rational(time.nsec, nano_per_sec));

    Rational rat_pos(rat_time * Rational(bytes_per_second));

    seekto = rat_pos.getNumerator() / rat_pos.getDenominator() - filepos;

    int modulo(seekto % (channels * bytes_per_sample));
    if (modulo) {
      seekto += (channels * bytes_per_sample) - modulo;
    }

    DEBUG_STREAM(seekto << " = " << time << " - " << filepos << endl);
  }
  /******************** ISeekable interface ********************/

  /******************** IProgress interface ********************/
  void WavReadNode::sendProgressInformation(const bool& in) {
    send_progress = in;
  }

  void WavReadNode::setProgressInterval(const int& buffers) {
    buffers_before_progress = buffers;
  }
  /******************** IProgress interface ********************/
}

