/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "SwitchNode.hpp"
#include "nmm/interfaces/base/sync/ISyncReset.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/audio/IMute.hpp"
#include <stdio.h>
#include <iostream>
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::SwitchNode, "SwitchNode")

namespace NMM {

  SwitchNode::SwitchNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericDemultiplexerNode(name, qmode, qsize) 
  {
    num_of_outputs = 2;

    output_connected = new bool[num_of_outputs];
    for(int i = 0; i < num_of_outputs; ++i) {
      output_connected[i] = true;
    }

    current_in = 0;
    switch_from = switch_to = -1;

    // enable internal switching mechanism but disable switching in GenericNode
    // (since the code in GenericNode::run() should not be executed)
    internal_switch_enabled = true;
    GenericNode::setSwitchOutputsEnabled(false);
  }
  
  SwitchNode::~SwitchNode()
  {
    if(output_connected != 0) {
      delete output_connected;
    }
  }

  Result SwitchNode::doInit()
  {
    Format* any_format_i = new Format("*/*");
    getOwnInputProperty()->addNewFormat(any_format_i);
    
    Format* any_format_o = new Format("*/*");
    getOwnOutputProperty()->addNewFormat(any_format_o);
    
    return SUCCESS;
  }
  
  Result SwitchNode::doInitOutput() {
    // get input format
    if (!getInputFormat()) {
      GLOBAL_ERROR_STREAM("No Input Format! FAILURE" << endl);
      return FAILURE;
    }
    Format* i_format = new Format(*getInputFormat());
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    for( unsigned short i = 0; i < num_of_outputs; i++ ) {
      char tag[16];
      sprintf(tag, "%u", i);
      DEBUG_STREAM("Adding stream with name: " << tag << " and with format: " << i_format->getFormatType() << endl);
      if(addOutputStream(tag) == FAILURE) {
	WARNING_STREAM("Could not add stream with name " << tag << endl);
      } else {
        Format* o_format = new Format(*i_format);
        o_format->setJackTag(string(tag));
        getOwnWorkingOutputProperty()->addNewFormat(o_format);
        getOwnWorkingOutputProperty()->setDefaultFormat(o_format, string(tag));
      }
    }
    return SUCCESS;
  }
  
  Result SwitchNode::doDeinitOutput() 
  {
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    
    for(unsigned short i = 0; i < num_of_outputs; i++) {
      char tag[16];
      sprintf(tag, "%u", i);
      DEBUG_STREAM("Removing stream with name " << tag << endl);
      if(removeOutputStream(tag) == FAILURE) {
	WARNING_STREAM("Could not remove stream with name " << tag << endl);
      }
    }
    return SUCCESS;
  }
  
  void SwitchNode::switchOutputs()
  {
    MutexGuard mg(switch_mutex);
    if(internal_switch_enabled) {
      perform_switch = true;
      setWorkingFlag(true);
      DEBUG_STREAM("switchOutputs from " << switch_from << " to " << switch_to << endl);
    }
  }

  void SwitchNode::setSwitchOutputsEnabled(const bool& enable)
  {
    MutexGuard mg(switch_mutex);
    internal_switch_enabled = enable;
  }

  bool SwitchNode::getSwitchOutputsEnabled()
  {
    return internal_switch_enabled;
  }
  
  Message* SwitchNode::processBuffer(Buffer *in) 
  {
    MutexGuard mg(switch_mutex);
    // if one or more outputs
    if(num_of_outputs) {
      if(getWorkingFlag()) {
	// there are two conditions for the working flag to be true:
	// (1) sent syncReset and stored buffer in last iteration of processBuffer -> send this buffer now
	if((current_in != 0) || (perform_switch == false)) { // make sure current_in is really sent!
	  Buffer* tmp = current_in;
	  current_in = 0;
	  setWorkingFlag(false);
	  return tmp;
	}
	// (2) switchOutputs was called out-of-band -> send mute/unmute 
	//     no need to send buffer since processBuffer was called with 0-pointer this time 
	else { 
	  if(switch_from != -1) { // first iteration -> send mute to 'switch_from'
	    char tag[16];
	    sprintf(tag, "%u", switch_from);
	    setCurrentSendOutputStream(tag);
	    switch_from = -1;
	    return new CEvent(IMute::create_mute());
	  }
	  else { // second iteration -> send unumte to 'switch_to' and set working-flag back to false
	    char tag[16];
	    sprintf(tag, "%u", switch_to);
	    setCurrentSendOutputStream(tag);
	    setWorkingFlag(false);
	    perform_switch = false;
	    switch_to = -1;
	    return new CEvent(IMute::create_unmute());
	  }
	}
      }
      else {
	if(internal_switch_enabled) {
	  // test if previously unconnected output is now connected
	  char tag[16];
	  for(int i = 0; i < num_of_outputs; ++i) {
	    sprintf(tag, "%u", i);
	    IOutputJack_var jack(getOutputStream(tag));   
	    if(jack.get()) {
	      if(jack->isConnected()) {
		// if output was *not* connected before, but is connected now -> syncReset
		if(output_connected[i] == false) {
		  output_connected[i] = true;
		  setWorkingFlag(true);
		  current_in = in;
		  // output i is now connected -> this is the output we want to switch to
		  switch_to = i;
		  // and therefore, the other output is the output we want to switch from
		  switch_from = !switch_to;
		  DEBUG_STREAM("Send sync_reset to all connected outputs! from " << switch_from << " to " << switch_to << endl);
		  return new CEvent(ISyncReset::create_syncReset());
		}
	      }
	      else {
		output_connected[i] = false;
	      }
	    }
	    else { 
	      DEBUG_STREAM2("Jack " << i << " : " << tag << " not connected, number of outputs = " << num_of_outputs << endl);
	      output_connected[i] = false;
	    }
	  }
	}

	// any case, send buffer
	return in;
      } // end: else of if(getWorkingFlag())
    }
    else {
      in->release();
      return 0;
    }
  }
}
