/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    Network-Multimedia Workgroup,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Eric Peters <epeters@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "NetSinkNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include "nmm/base/ConnectionManager.hpp"
#include "nmm/base/FormatStrategyMap.hpp"
#include "nmm/base/TSFactory.hpp"

NMM_REGISTER_PLUGIN(NMM::NetSinkNode, "NetSinkNode")

namespace NMM {

  NetSinkNode::NetSinkNode(const char* name) : GenericSinkNode(name, StreamQueue::MODE_SUSPEND, 10), 
					       __transport_strategy(0),
					       __internal_cond(__internal_mutex),
					       __payload_type(0)
  {}

  NetSinkNode::~NetSinkNode() {
    if (__transport_strategy) {
      __transport_strategy->scheduleDestruction();
      __transport_strategy = 0;
    }
  }

  void NetSinkNode::setRTPStrategy(const string& _address, const int& _rtp_port, const int& _rtcp_port) {
    TSAddress* address = new RTPAddress(UnicastIPAddress(_address), _rtp_port, _rtcp_port);
    setTransportStrategy(address);    
  }

  void NetSinkNode::setTransportStrategy(TransportStrategy* transport_strategy) {
    /* Begin> If an old strategy is available, release it. */
    if (__transport_strategy)
      __transport_strategy->scheduleDestruction();
    /* End< */

    /* stores the new strategy */
    __transport_strategy = transport_strategy;

    __transport_strategy->setSignalReceiver(new NetSignalReceiver(this)); 
  }

  void NetSinkNode::setTransportStrategy(const TSAddress* address) throw(const BindingException&) {
    /* If the address is invalid throw an Exception */
    if (!address)
      throw BindingException("Invalid Address (0-ptr)");

    /* Create a TransportStrategy which can connect to the given address */
    TransportStrategy* transport_strategy = STSFactory::getInstance().createStrategy(address->getStrategyType());
    
    /* If it could not be created throw a BindingException */
    if (!transport_strategy)
      throw  BindingException("Can not create transport strategy with type "+ address->getStrategyType());
    
    /* The strategy connects to the given address */
    transport_strategy->setAddress(*address);
    
    setTransportStrategy(transport_strategy);
  }

  Result NetSinkNode::enableInputStream() {
    /* If a FREE_QUEUE_SLOT signal is received the input stream will be enabled */
    MutexGuard mg(__internal_mutex);
    __internal_cond.notify();

    return SUCCESS;
  }

  Message* NetSinkNode::processBuffer(Buffer* buffer) {
    
    if (buffer) {
      __internal_mutex.lock();
      Result status = __transport_strategy->receiveMessage(buffer, __transport_strategy);
      if (status == BLOCKED) __internal_cond.wait();
      __internal_mutex.unlock();
    }

    return NULL;
  }

  Result NetSinkNode::doInit()  {
    string strategy_type;
    if (__transport_strategy) {
      strategy_type = __transport_strategy->getStrategyType();

      list<Format*> formats = SConnectionManager::getInstance().getFormatStrategyMap().getSupportedFormats(strategy_type);
      list<Format*>::iterator i;

      // set static in and out formats in own property
      for (i  = formats.begin();
	   i != formats.end();
	   ++i) {
	Format* inputFormat = (*i)->duplicateFormat();
	getOwnInputProperty()->addNewFormat(inputFormat);
      }

    } else {
      WARNING_STREAM("No transport strategy set. Using default format \"*/*\"." << endl);
      Format* inputFormat = getOwnInputProperty()->addNewFormat("*/*");
      getOwnInputProperty()->setDefaultFormat(inputFormat);
    }

    return SUCCESS;
  }
  
  Result NetSinkNode::doInitOutput() {
    
    const Format* in_format = getInputFormat();
    
    if (in_format == NULL) {
      ERROR_STREAM("Error in doInitOutput: no in_format specified" << endl);
      return FAILURE;
    }
    
    if (!__transport_strategy) {
      WARNING_STREAM("No transport strategy set. You should set a transport strategy before using this node" << endl);
      return SUCCESS;
    }

    __transport_strategy->configureForOutputJack(*in_format);
    
    if (__transport_strategy->getStrategyType() == "RTPStrategy") {
      RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
      __payload_type = (int)rtp_strategy->getPayloadType();
    }
    
    return SUCCESS;
  }

  void NetSinkNode::setPayloadType(const int& type) {
    __payload_type = type;
    DEBUG_STREAM("Now using NetSinkNode with payload type " << __payload_type << endl);
  }

  int NetSinkNode::getPayloadType() const {
    return __payload_type;
  }

  /*****************************************/
  /* methods from IRTPParameters interface */
  /*****************************************/

  string NetSinkNode::getRTPSinkMediaType() const{
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (string(rtp_strategy->getRTPSinkMediaType()));
      }
    }
    return string("");
  }
  
  unsigned char NetSinkNode::getRTPSinkPayloadType() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getPayloadType());
      }
    }
    return 0;
  }
  
  string NetSinkNode::getRTPSinkMapLine() const{
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (string(rtp_strategy->getRTPSinkMapLine()));
      }
    }
    return string("");
  }
  
  string NetSinkNode::getRTPSinkAuxSDPLine() const{
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (string(rtp_strategy->getRTPSinkAuxSDPLine()));
      }
    }
    return string("");
  }

  unsigned short NetSinkNode::getRTPSeqNum() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getRTPSeqNum());
      }
    }
    return 0;
  }
    
  unsigned NetSinkNode::getRTPTimestamp() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getRTPTimestamp());
      }
    }
    return 0;
  }

  unsigned NetSinkNode::getRTPPort() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getRTPPort());
      }
    }
    return 0;
  }

  unsigned NetSinkNode::getRTCPPort() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getRTCPPort());
      }
    }
    return 0;
  }

  /***************************/
  /* Begin NetSignalReceiver */
  /***************************/
  
  NetSignalReceiver::NetSignalReceiver(NetSinkNode* _netsink) {
    __netsink = _netsink;
  }

  NetSignalReceiver::~NetSignalReceiver() {
  }
    
  void NetSignalReceiver::receiveSignal(Signal signal){
    
    switch (signal) {
    case FREE_QUEUE_SLOT:
      GLOBAL_DEBUG_STREAM2("NetSignalReceiver: Received signal FREE_QUEUE_SLOT from send queue" << endl);
      __netsink->enableInputStream();
      break;
    default:
      break;
    } 
  }
  
} //end namespace
