/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#include "SourceNetNode.hpp"
#include "nmm/base/EventDispatcher.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::SourceNetNode, "SourceNetNode")

namespace NMM{
  
  SourceNetNode::SourceNetNode(const char* name /* = "SourceNetNode" */,
		   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
		   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size) {
    s_stream     = 0;
    is_connected = false;
  }
  
  SourceNetNode::~SourceNetNode() {
    delete s_stream;
  }
  
  Buffer* SourceNetNode::processBuffer(Buffer *) {
    if (!is_connected)
      return 0;
    size_t size = 0;
    if (s_stream->read(reinterpret_cast<char*>(&size), sizeof(size_t)) == -1) //read new buffer size
      return 0;

    DEBUG_STREAM("size " << size << endl);

    Buffer* buffer = getNewBuffer(size);
    char* time = new char[sizeof(Timestamp)];
    char* labelstr = new char[sizeof(unsigned long)];
    try {
      s_stream->readStream(time, sizeof(Timestamp));
      s_stream->readStream(labelstr, sizeof(unsigned long));
      s_stream->readStream(buffer->getData(), size);
      buffer->setUsed(size);
    } catch (const TransmissionException& te) {
      ERROR_STREAM(te << endl);
      buffer->release();
      return 0;
    }
    Timestamp timestamp;
    memcpy(&timestamp, time, sizeof(Timestamp));
    delete[] time;
    buffer->setTimestamp(timestamp);
    DEBUG_STREAM("Set Timestamp to " << timestamp << " seconds " << timestamp.sync_time.sec << endl);

    /*if (0 != timestamp.sync_time.sec) {
      ERROR_STREAM("Network lost!!!!" << endl);
      Event e = Event("lost_connection", new TInValue<bool>(true));
      requestSetEventDispatcher()->dispatch(e); // send event to application
      setProducingFlag(false);
      return NULL;
      }*/

    unsigned long label = 0;
    memcpy(&label, labelstr, sizeof(unsigned long));
    DEBUG_STREAM("Got label " << label << endl);
    delete[] labelstr;
    buffer->setLabel(label);

    return buffer;
  }

  void SourceNetNode::setPort(const int& value) {
    addr.setPort(value);
  }

  void SourceNetNode::setHostname(const string& value) {
    addr.setAddress(value.c_str());
  }

  void SourceNetNode::setProtocol(const string& value) {
    addr.setTransportProtocol(value.c_str());
  }
  
  int SourceNetNode::getPort() const {
    return addr.getPort();
  }
  
  string SourceNetNode::getHostname() const {
    return "bugs.bunny";
  }
  
  string SourceNetNode::getProtocol() const {
    return addr.getTransportProtocol();
  }
  
  void SourceNetNode::connectionLost() {
    // maybe clean up something???
  }

  Result SourceNetNode::doInit() {
    try {
      s_stream = s_connect.connect(addr);	
    }
    catch (const ConnectionException& ce) {
      ERROR_STREAM(ce << endl);
      return FAILURE;
    }
    if (s_stream == 0) {
      ERROR_STREAM("Socket is 0" << endl);
      return FAILURE;
    }

    if(addr.getTransportProtocol() != "udp") {
      DEBUG_STREAM("Connected with protocol " << addr.getTransportProtocol() << endl);
    }
    else {
      DEBUG_STREAM("Using connectionless protocol " << addr.getTransportProtocol() << endl);
    }
    
    Format* any_format = new Format("*/*");
    getOwnOutputProperty()->addNewFormat(any_format);
    getOwnOutputProperty()->setDefaultFormat(any_format);
    
    is_connected = true;
    return SUCCESS;
  }
  
  Result SourceNetNode::doInitOutput() {
    Format* any_format = new Format("*/*");
    getOwnOutputProperty()->addNewFormat(any_format);
    getOwnOutputProperty()->setDefaultFormat(any_format);
    return SUCCESS;
  }
  
  /* todo: this should be doDeinit() */
  Result SourceNetNode::doDeactivate() {
    try {
      s_stream->close();
    } catch (const ConnectionException& ce) {
      ERROR_STREAM(ce << endl);
      is_connected = false;
    }
    is_connected = false;
    return SUCCESS;
  }
} //namespace NMM








