/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "FramerateConverterNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include <assert.h>

NMM_REGISTER_PLUGIN(NMM::FramerateConverterNode, "FramerateConverterNode")

namespace NMM {
  
  FramerateConverterNode::FramerateConverterNode(const char *name /*= "FramerateConverterNode"*/,
                                                 const StreamQueue::Mode queue_mode /*= StreamQueue::MODE_SUSPEND*/,
                                                 const int queue_size /*= 1*/ )
    : GenericProcessorNode(name, queue_mode, queue_size),
      first_frame(false),
      out_framerate(0),
      alpha(0.5),
      start_time(0),
      in_buffer1(NULL),
      in_buffer2(NULL),
      time_offset(0),
      num_copies1(0),
      num_copies2(0),
      total_in_buffers(0),
      total_out_buffers(0),
      total_dropped(0),
      total_copied(0)
  {
  }
  
  FramerateConverterNode::~FramerateConverterNode() {
    updateStatistics();
    
    if (in_buffer1) {
      in_buffer1->release();
      in_buffer1 = NULL;
    }
    
    if (in_buffer2) {
      in_buffer2->release();
      in_buffer2 = NULL;
    }
    
    resetStatistics();
  }
  
  Result FramerateConverterNode::doInit() {
    /* Raw video input format */
    Format* i_format = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    i_format->addWildcard(Format::x_resolution_param);
    i_format->addWildcard(Format::y_resolution_param);
    i_format->addWildcard(Format::ratio_x_param);
    i_format->addWildcard(Format::ratio_y_param);
    i_format->addWildcard(Format::framerate_param);
    i_format->addWildcard(Format::channelorder_param);
    i_format->addWildcard(Format::endian_param);
    i_format->addWildcard(Format::colorspace_param);
    i_format->addWildcard(Format::bitperpixel_param);
    i_format->addWildcard(Format::format_param);
    getOwnInputProperty()->setDefaultFormat(i_format);
    
    /* Raw video output format */
    Format* o_format = getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addWildcard(Format::framerate_param);
    o_format->addWildcard(Format::channelorder_param);
    o_format->addWildcard(Format::endian_param);
    o_format->addWildcard(Format::colorspace_param);
    o_format->addWildcard(Format::bitperpixel_param);
    o_format->addWildcard(Format::format_param);
    getOwnOutputProperty()->setDefaultFormat(o_format);
    
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    
    return SUCCESS;
  }
  
  Result FramerateConverterNode::doInitOutput() {
    const Format* in_format = getInputFormat();
    if (!in_format) {
      ERROR_STREAM("No input format!!" << endl);
      return FAILURE;
    }
    Format* i_format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    
    /* Default format with output framerate == input framerate */
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_raw_type);
    o_format->addIntParamValue(Format::x_resolution_param, in_format->getIntValue(Format::x_resolution_param));
    o_format->addIntParamValue(Format::y_resolution_param, in_format->getIntValue(Format::y_resolution_param));
    o_format->addIntParamValue(Format::ratio_x_param, in_format->getIntValue(Format::ratio_x_param));
    o_format->addIntParamValue(Format::ratio_y_param, in_format->getIntValue(Format::ratio_y_param));
    o_format->addFloatParamValue(Format::framerate_param, in_format->getFloatValue(Format::framerate_param));
    o_format->addStringParamValue(Format::channelorder_param, in_format->getStringValue(Format::channelorder_param));
    o_format->addStringParamValue(Format::endian_param, in_format->getStringValue(Format::endian_param));
    o_format->addStringParamValue(Format::colorspace_param, in_format->getStringValue(Format::colorspace_param));
    o_format->addIntParamValue(Format::bitperpixel_param, in_format->getIntValue(Format::bitperpixel_param));
    o_format->addStringParamValue(Format::format_param, in_format->getStringValue(Format::format_param));
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    /* Generic format with framerate wildcard */
    o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_raw_type);
    o_format->addIntParamValue(Format::x_resolution_param, in_format->getIntValue(Format::x_resolution_param));
    o_format->addIntParamValue(Format::y_resolution_param, in_format->getIntValue(Format::y_resolution_param));
    o_format->addIntParamValue(Format::ratio_x_param, in_format->getIntValue(Format::ratio_x_param));
    o_format->addIntParamValue(Format::ratio_y_param, in_format->getIntValue(Format::ratio_y_param));
    o_format->addWildcard(Format::framerate_param);
    o_format->addStringParamValue(Format::channelorder_param, in_format->getStringValue(Format::channelorder_param));
    o_format->addStringParamValue(Format::endian_param, in_format->getStringValue(Format::endian_param));
    o_format->addStringParamValue(Format::colorspace_param, in_format->getStringValue(Format::colorspace_param));
    o_format->addIntParamValue(Format::bitperpixel_param, in_format->getIntValue(Format::bitperpixel_param));
    o_format->addStringParamValue(Format::format_param, in_format->getStringValue(Format::format_param));
    
    return SUCCESS;
  }
  
  Result FramerateConverterNode::doActivate() {
    const Format* out_format = getOutputFormat();
    if (!out_format) {
      ERROR_STREAM("No output format!!" << endl);
      return FAILURE;
    }
    
    out_framerate = out_format->getFloatValue(Format::framerate_param);
    
    /* Calculate tolerance threshold. This is 1/4 output frame rate */
    float inv_out_framerate = 0.25f / out_framerate;
    int sec = static_cast<int>(inv_out_framerate);
    int msec = static_cast<int>((inv_out_framerate - sec) * 1000000000);
    threshold = Interval(sec, msec);
    
    first_frame = true;
    
    resetStatistics();
    
    return SUCCESS;
  }
  
  Message* FramerateConverterNode::processBuffer(Buffer* in_buffer) {
    
    Buffer* out_buffer;
    
    /* This node has two different states, which are determined by the working flag.
     *
     * STATE 0 (working flag NOT set):
     * =======
     *
     * In this state we receive input buffers until we have a pair of subsequent buffers such that
     * the next_out_ts lies between the timestamps of the two input buffers.
     * In this way we get the buffer with the greatest timestamp less than next_out_ts (in_buffer1)
     * and the buffer with the smallest timestamp greater than next_out_ts (in_buffer2)
     * When this condition is reached, we turn on the working flag and switch to STATE 1.
     *
     * In this state we discard input buffers.
     *
     * STATE 1 (working flag set):
     * =======
     *
     * In this state we have a pair of input buffers (in_buffer1, in_buffer2) such that
     * timestamp(in_buffer1) < next_out_ts < timestamp(in_buffer2).
     * In each iteration we create a copy of the input buffer whose timestamp is nearest to next_out_ts,
     * assign next_out_ts to the copy and send it. Then we advance next_out_ts by 1/out_framerate.
     * If the condition about the timestamps given above is no longer true, we shift the input buffers,
     * discarding in_buffer1, clear the working flag and switch to STATE 0.
     */
    
    /* Begin> STATE 0 */
    if (!getWorkingFlag()) {
      
      ++total_in_buffers;
      
      /* INVARIANT: We have no second buffer. */
      assert(!in_buffer2);
      
      /* Begin> Special case for first frame */
      /* This is the case at the beginning of the stream and after a syncReset. This code assures that the invariants are fulfilled. */
      if (first_frame) {
        first_frame = false;
        
        assert(!in_buffer1);
        
        /* Begin> Initialize the output timestamp */
        next_out_ts = in_buffer->getTimestamp();
        if (!next_out_ts.is_valid) {
          ERROR_STREAM("First frame has invalid timestamp!!" << endl);
          // XXX: what to do now ?!
          // XXX: to ensure predictable behaviour, discard the buffer
          in_buffer->release();
          return NULL;
        }
        next_out_ts.stream_counter = 0;
        next_out_ts.is_valid = true;
        next_out_ts.sync_time += time_offset;
        start_time = next_out_ts.sync_time;
        /* End< */
        
        /* The first buffer is used for filling the empty space before the video actually begins.
         * If the time_offset is negative, which means that the output stream should start earlier in
         * time than the input stream, we simply duplicate the first buffer until we have reached the
         * starting time of the input stream.
         * If the time_offset is positive, we will automatically discard buffers until we receive an
         * input buffer whose time stamp is greater than or equal to start_time.
         */
        in_buffer1 = in_buffer->copy();
        num_copies1 = 0;
        in_ts1 = in_buffer1->getTimestamp();
        
      } // if (first_frame)
      
      /* Begin> Get in_buffer timestamp */
      Timestamp in_ts = in_buffer->getTimestamp();
      if (!in_ts.is_valid) {
        // XXX: what to do now ?!
        // XXX: to ensure predictable behaviour, discard the buffer
        in_buffer->release();
        ERROR_STREAM("Frame has invalid timestamp!!" << endl);
        return NULL;
      }
      /* End< */
      
      /* Begin> Check if the condition for STATE 1 is met */
      if (in_ts.sync_time >= next_out_ts.sync_time) {
        /* We found the first buffer with timestamp > next_out_ts */
        in_buffer2 = in_buffer;
        num_copies2 = 0;
        in_ts2 = in_buffer->getTimestamp();
        
        /* Switch to STATE 1 */
        setWorkingFlag(true);
        return NULL;
      } else {
        updateStatistics();
        
        /* Keep only the newest buffer with timestamp < next_out_ts */
        in_buffer1->release();
        in_buffer1 = in_buffer;
        in_ts1 = in_buffer->getTimestamp();
        
        /* Stay in STATE 0 */
        return NULL;
      }
      
    } else { // if (!getWorkingFlag())
      
      /* INVARIANT: We have a first buffer */
      assert(in_buffer1);
      
      /* INVARIANT: We have a second buffer */
      assert(in_buffer2);
      
      /* This is not the first frame */
      assert(!first_frame);
      
      /* Get the time distances of both buffers to next_out_ts */
      // XXX: compute this only when it changes
      Interval i1(in_ts1.sync_time - Time(0)); i1 *= 1-alpha;
      Interval i2(in_ts2.sync_time - Time(0)); i2 *= alpha;
      Time m(i1 + i2);
      if (next_out_ts.sync_time < m-threshold || next_out_ts.sync_time > m+threshold) {
        /* Send out the in_buffer which is nearest to next_out_ts */
        if (next_out_ts.sync_time < m) {
          in_buffer1->increaseCount();
          out_buffer = in_buffer1->getWriteableInstance();
          ++num_copies1;
        } else {
          in_buffer2->increaseCount();
          out_buffer = in_buffer2->getWriteableInstance();
          assert(out_buffer != in_buffer2);
          ++num_copies2;
        }
      } else {
        /* Treat time distances as equal if they are too small and avoid copying and dropping frames */
        if (!num_copies1) {
          in_buffer1->increaseCount();
          out_buffer = in_buffer1->getWriteableInstance();
          ++num_copies1;
        } else {
          in_buffer2->increaseCount();
          out_buffer = in_buffer2->getWriteableInstance();
          assert(out_buffer != in_buffer2);
          ++num_copies2;
        }
      }
      
      /* Set output timestamp */
      out_buffer->setTimestamp(next_out_ts);
      next_out_ts.sync_time = getFrameTime(++total_out_buffers);
      next_out_ts.stream_counter++;
      
      /* Begin> Check if condition for STATE 1 is still met */
      if (in_ts2.sync_time < next_out_ts.sync_time) {
        /* Shift buffers and discard the older one */
        if (in_buffer1) {
          in_buffer1->release();
          in_buffer1 = NULL;
        }
        in_buffer1 = in_buffer2;
        num_copies1 = num_copies2;
        num_copies2 = 0;
        in_ts1 = in_ts2;
        in_buffer2 = NULL;
        
        /* Switch to STATE 0 after sending current buffer */
        setWorkingFlag(false);
      }
      
      return out_buffer;
    }
    
  }
  
  Time FramerateConverterNode::getFrameTime(int frameno) {
    float temp = frameno / out_framerate;
    int sec = static_cast<int>(temp);
    int nsec = static_cast<int>(((temp - sec) * 1000000000));
    return start_time + Interval(sec, nsec);
  }
  
  void FramerateConverterNode::updateStatistics() {
    if (in_buffer1) {
      if (!num_copies1) {
        MESSAGE_STREAM("Discarded buffer with timestamp " << in_ts1 << endl);
        ++total_dropped;
      } else if (num_copies1 > 1) {
        MESSAGE_STREAM("Created " << num_copies1 - 1 << " copies of buffer with timestamp " << in_ts1 << endl);
        total_copied += num_copies1 - 1;
      }
    }
  }
  
  void FramerateConverterNode::resetStatistics() {
    
    if (total_in_buffers) {
      float out_percent = 100.0f * total_out_buffers / total_in_buffers;
      float drop_percent = 100.0f * total_dropped / total_in_buffers;
      float copy_percent = 100.0f * total_copied / total_in_buffers;
      
      MESSAGE_STREAM("Total buffers received: " << total_in_buffers << endl);
      MESSAGE_STREAM("Total buffers sent    : " << total_out_buffers
                     << " (" << out_percent << "%)" << endl);
      MESSAGE_STREAM("Total buffers copied  : " << total_copied
                     << " (" << copy_percent << "%)" << endl);
      MESSAGE_STREAM("Total buffers dropped : " << total_dropped
                     << " (" << drop_percent << "%)" << endl);
    }
    
    num_copies1 = 0;
    num_copies2 = 0;
    total_in_buffers = 0;
    total_out_buffers = 0;
    total_dropped = 0;
    total_copied = 0;
  }
  
  /******************** ITrack interface ********************/
  Result FramerateConverterNode::startTrack(const string& /*filename*/) {
    first_frame = true;
    
    return SUCCESS;
  }
  
  Result FramerateConverterNode::endTrack() {
    // XXX: This behaviour is NOT correct!
    // XXX: The node should send buffers until its output timestamp is greater than the timestamp of in_buffer2.
    // XXX: Also we should make sure that in_buffer2 is sent at least once. That means that we should delay the processing
    // XXX: of endTrack until the node has reached state 0 and then send in_buffer1 and then do the following:
    
    first_frame = true;
    
    updateStatistics();
    
    if (in_buffer1) {
      in_buffer1->release();
      in_buffer1 = NULL;
    }
    
    if (in_buffer2) {
      in_buffer2->release();
      in_buffer2 = NULL;
    }
    
    resetStatistics();
    
    return SUCCESS;
  }
  
  Result FramerateConverterNode::finished() {
    return SUCCESS;
  }
  /******************** ITrack interface ********************/
  
  /******************** ISyncReset interface ********************/
  Result FramerateConverterNode::syncReset() {
    first_frame = true;
    
    updateStatistics();
    
    if (in_buffer1) {
      in_buffer1->release();
      in_buffer1 = NULL;
    }
    
    if (in_buffer2) {
      in_buffer2->release();
      in_buffer2 = NULL;
    }
    
    resetStatistics();
    
    return SUCCESS;
  }
  /******************** ISyncReset interface ********************/

  /******************** IFramerateConverter interface ********************/
  void FramerateConverterNode::setTimeOffsetMillisecs(const int& millisecs)
  {
    int secs = millisecs / 1000;
    int msec = (millisecs % 1000) * 1000000;
    Interval i(secs, msec);
    time_offset = i;
  }

  void FramerateConverterNode::setTimeOffset(const Interval& start) {
    time_offset = start;
  }

  void FramerateConverterNode::setFrameOffset(const float& alpha) {
    this->alpha = alpha;
  }
  
  int FramerateConverterNode::getTotalInBuffers() const {
    return total_in_buffers;
  }
  
  int FramerateConverterNode::getTotalOutBuffers() const {
    return total_out_buffers;
  }
  
  int FramerateConverterNode::getTotalDroppedBuffers() const {
    return total_dropped;
  }
  
  int FramerateConverterNode::getTotalCopiedBuffers() const {
    return total_copied;
  }
  /******************** IFramerateConverter interface ********************/
}
