/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_FRAMERATECONVERTERNODE_HPP_
#define _NMM_FRAMERATECONVERTERNODE_HPP_ _NMM_FRAMERATECONVERTERNODE_HPP_

#include "nmm/base/GenericProcessorNode.hpp"

#include "nmm/interfaces/file/ITrackImpl.hpp"
#include "nmm/interfaces/video/IFramerateConverterImpl.hpp"
#include "nmm/interfaces/base/sync/ISyncResetImpl.hpp"

namespace NMM {
  
  /**
   * Framerate converter node.
   * This node produces a video stream with a consistent framerate that matches the framerate specified
   * in the output format. It can be used for converting between framerate and for correction of errors
   * (missing, duplicate frames) in video streams or for converting from variable frame rate to constant
   * frame rate. The node only works on raw video streams.
   *
   * Formally the node generates a series of timestamps with a constant interval and an offset, both of
   * which can be set with the IFramerateConverter interface. To each generated output timestamp, it
   * associates the input frame with the input timestamp "nearest" to the output timestamp, where nearest
   * means the following:
   *
   * Let @a T1 and @a T2 be the timestamps of two successive input frames. Then an output timestamp @a T will
   * be associated with the frame at @a T1 if T1 < T < M and with the frame at @a T2 if M < T < T2 where
   * M = (T1 * (1-alpha) + T2 * alpha). The @a alpha value is 0.5 by default, so each output frame will
   * be associated with the input frame nearest to the output frame. To have an input frame start exactly
   * at its input timestamp in the output video stream (for example if your input is a stream of slides),
   * you want to set alpha to 0 using the setFrameOffset method.
   */
  class FramerateConverterNode :
    public GenericProcessorNode,
    public ITrackImpl,
    public ISyncResetImpl,
    public IFramerateConverterImpl
  {  
  private:
    bool first_frame;               //!< First frame of a stream
    float out_framerate;            //!< Output framerate
    float alpha;                    //!< Parameter for frame interpolation (see above)
    Time start_time;                //!< First incoming timestamp
    Timestamp next_out_ts;          //!< Timestamp of next output frame
    Timestamp in_ts1;               //!< Timestamp of in_buffer1
    Timestamp in_ts2;               //!< Timestamp of in_buffer2
    Buffer* in_buffer1;             //!< Input frame with highest timestamp <= next_out_ts
    Buffer* in_buffer2;             //!< Input frame with lowest  timestamp >= next_out_ts
    Interval time_offset;           //!< Time offset
    Interval threshold;             //!< Tolerance threshold for timestamp rounding errors
    
    // statistics
    int num_copies1;                //!< Counts how many time in_buffer1 has been copied
    int num_copies2;                //!< Counts how many time in_buffer2 has been copied
    int total_in_buffers;           //!< Total buffers received
    int total_out_buffers;          //!< Total buffers sent
    int total_dropped;              //!< Total buffers dropped
    int total_copied;               //!< Number of copies created
    
    void updateStatistics();
    void resetStatistics();
    
    Time getFrameTime(int frameno);
    
  protected:
    virtual Message* processBuffer(Buffer* in_buffer);
    
    virtual Result doInit();
    virtual Result doInitOutput();
    virtual Result doActivate();
    
    // ITrack interface
    virtual Result startTrack(const string& filename);
    virtual Result endTrack();
    virtual Result finished();
    
    // ISyncReset interface
    virtual Result syncReset();
    
    // IFramerateConverter interface
    virtual void setTimeOffsetMillisecs(const int& millisecs);
    virtual void setTimeOffset(const Interval& start);
    virtual void setFrameOffset(const float& alpha);
    virtual int getTotalInBuffers() const;
    virtual int getTotalOutBuffers() const;
    virtual int getTotalDroppedBuffers() const;
    virtual int getTotalCopiedBuffers() const;
    
  public:
    FramerateConverterNode(const char *name = "FramerateConverterNode",
                           const StreamQueue::Mode queue_mode = StreamQueue::MODE_SUSPEND, 
                           const int queue_size = 1 );
    ~FramerateConverterNode();
  }; 
}

#endif // _NMM_FRAMERATECONVERTERNODE_HPP_
