/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "YUVBluescreen.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

#include <stdio.h>
#include <iostream>
#include <math.h>

#ifndef NULL
#define NULL 0
#endif

NMM_REGISTER_PLUGIN(NMM::YUVBluescreen, "YUVBluescreen")

namespace NMM {
  
  YUVBluescreen::YUVBluescreen(const char *name) : GenericProcessorNode(name) {

    // initialize formats
    NMM::Format *myInputFormat;
    myInputFormat=getOwnInputProperty()->addNewFormat("video/raw");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addStringParamValue(Format::colorspace_param,"yuv422");
    myInputFormat->addStringParamValue(Format::channelorder_param,"uyvy");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,16);
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addWildcard(Format::framerate_param);
 
    NMM::Format *myOutputFormat;
    myOutputFormat=getOwnOutputProperty()->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"rgba");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addWildcard(Format::framerate_param);
 
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
  }
  
  Buffer* YUVBluescreen::processBuffer(Buffer *in_buffer) {
    DEBUG_STREAM("processing buffer" << endl);
    unsigned char *in_data, *out_data;
    Buffer *out_buffer = getNewBuffer(in_buffer->getUsed() * 2, in_buffer);
    in_data = (unsigned char*)in_buffer->getData();
    out_data = (unsigned char*)out_buffer->getData();
    convertKeyYUV2RGBA(in_data, in_buffer->getUsed(), out_data);
    out_buffer->setUsed(in_buffer->getUsed() * 2);
    in_buffer->release();
    return out_buffer;
  }


#define CLIPPED(x) ((x)>255?255:((x)<0?0:(x)))                                  

  void YUVBluescreen::convertKeyYUV2RGBA(unsigned char *src, int n, unsigned char *dst,
                                         bool hard,
                                         int target_u, int target_v,
                                         float edge_factor) {
    unsigned char *end=src+n;
    int r, g, b;
    int r1, g1, b1;
    int r2, g2, b2;
    int u, v, y1, y2, y, a;
    
    a = 255;

    while(src!=end) {
      
      // Generate a1 and a2 as alpha values by some chromakeying
      // algorithm.
 
      u = src[0];  // U/V
      y1 = src[1];  // Y
      v = src[2];  // V/U
      y2 = src[3];  // Y
      y = (y1 + y2) >> 1;
 
      // calc bg factor
      
#if 1
      int x,z;
      x = target_u - u;
      z = target_v - v;
      
      // only filter if in top left square
      if ((x > 0) && (z > 0)) {
 
        // only calculate lot of stuff if we'll use soft edges
        if (!hard) {
 
          int ds = (x>z)?z:x;
          float df = (float)(ds) / edge_factor;
          if (df > 1.0) df = 1.0;
 
          // suppress foreground
 
          if (x > z) {
            u += z;
            v += z;
          } else {
            u += x;
            v += x;
          }
 
          a = (int)(255 * (1.0 - df));
 
        } else {
          // kill color and alpha
          a = 0;
        }
      } else {
        // do nothing;
        a = 255;
      }
#endif 

      if (a > 0) {
        
        r=((((int)u-128)*0x166E9)>>16);
        g=-((((int)v-128)*0x05819)>>16)-((((int)u-128)*0x0B6D1)>>16);
        b=((((int)v-128)*0x1C5A1)>>16);
        
        r1=r+y1;
        g1=g+y1;
        b1=b+y1;
        r2=r+y2;
        g2=g+y2;
        b2=b+y2;

        // filter out remaining "bleed" in RGB space ?
        
        dst[0]=CLIPPED(r1);
        dst[1]=CLIPPED(g1);
        dst[2]=CLIPPED(b1);
        dst[3]=CLIPPED(a);
        dst[4]=CLIPPED(r2);
        dst[5]=CLIPPED(g2);
        dst[6]=CLIPPED(b2);
        dst[7]=CLIPPED(a);
        
      } else {
        // fully transparent pixels
        dst[0]=0;
        dst[1]=0;
        dst[2]=0;
        dst[3]=0;
        dst[4]=0;
        dst[5]=0;
        dst[6]=0;
        dst[7]=0;
        
      }
      
      src+=4;
      dst+=8;
    }
  }
  
}











