/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include <iomanip>
#include <sstream>

#include "GLDisplayNode.hpp"

#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

#include "GLDisplay.hpp"

NMM_REGISTER_PLUGIN(NMM::GLDisplayNode, "GLDisplayNode")

namespace NMM {
  
  GLDisplayNode::GLDisplayNode(const char *name,
                               const StreamQueue::Mode queue_mode,
                               const int size) 
    : GenericUSyncSinkNode(name, queue_mode, size) 
  {
    width = 0;
    height = 0;
    display_width = 0;
    display_height = 0;

    rgb_channel_order = true;
    working_buffer = NULL;

    display = NULL;
    is_GL_context_valid = false;

    benchmark.setNamedObject(this);
    benchmark.setEnabled(true);

    //this->setMaxPresentationTime(20000000);
    //this->setMinPresentationTime(10000000);

    title_update_rate = 10;
    disp_ev_config_file = string("");
    enable_disp_ev = false;
  }

  GLDisplayNode::~GLDisplayNode() {
    if (display) {
      delete display;
    }
  }

  Result GLDisplayNode::doInit() {
    Format* i_format = getOwnInputProperty()->addNewFormat("video/raw");
    i_format->addWildcard(Format::x_resolution_param);
    i_format->addWildcard(Format::y_resolution_param);
    i_format->addWildcard(Format::ratio_x_param);
    i_format->addWildcard(Format::ratio_y_param);
    i_format->addWildcard(Format::framerate_param);
    i_format->addStringParamValue(Format::channelorder_param,"rgb");
    i_format->addStringParamValue(Format::endian_param, "little");
    i_format->addStringParamValue(Format::colorspace_param,"rgb24");
    i_format->addIntParamValue(Format::bitperpixel_param,32);
    i_format->addStringParamValue(Format::format_param, "packet");
    i_format->setIOPartner(0);

    i_format = getOwnInputProperty()->addNewFormat("video/raw");
    i_format->addWildcard(Format::x_resolution_param);
    i_format->addWildcard(Format::y_resolution_param);
    i_format->addWildcard(Format::ratio_x_param);
    i_format->addWildcard(Format::ratio_y_param);
    i_format->addWildcard(Format::framerate_param);
    i_format->addStringParamValue(Format::channelorder_param,"bgra");
    i_format->addStringParamValue(Format::endian_param, "little");
    i_format->addStringParamValue(Format::colorspace_param,"rgb24");
    i_format->addIntParamValue(Format::bitperpixel_param,32);
    i_format->addStringParamValue(Format::format_param, "packet");
    i_format->setIOPartner(0);

    return SUCCESS;
  }

  Result GLDisplayNode::doInitOutput() {
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    Format* format;
    
    format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(format);
    getOwnWorkingInputProperty()->setDefaultFormat(format);
    
    return SUCCESS;
  }

  Result GLDisplayNode::doActivate() {
    const Format* i_format = getInputFormat();
    
    if (i_format == 0) {
      ERROR_STREAM( "Invalid input format (NULL-ptr)" << endl);
      return FAILURE;
    }
    
    width  = i_format->getIntValue(Format::x_resolution_param);
    height = i_format->getIntValue(Format::y_resolution_param);

    string channel_order = i_format->getStringValue(Format::channelorder_param);
    if (channel_order == "bgra") rgb_channel_order = false;

    if ((width >= 1) && (height >= 1)) {
      return SUCCESS;
    } else {
      return FAILURE;
    }
  }

  Result GLDisplayNode::doDeactivate() {
    disableDisplayEvents();

    // Kill the display window here.
    if (display) {
      display->deinit();
      delete display;
      display = NULL;
    }
    is_GL_context_valid = false;

    return SUCCESS;
  }

  void GLDisplayNode::createGLContext() {
    /* Create the window and GL context.
     * This is done in the GLDisplay class.
     */
      
    /*************************
     * GL context MUST be created in same thread! 
     *************************/
    display_size_mutex.lock();
    display_width = width;
    display_height = height;
    
    display = new GLDisplay(this, display_width, display_height, "GLDisplay", rgb_channel_order);
    display_size_mutex.unlock();

    display->init();
    
    is_GL_context_valid = true;
  }

  Result GLDisplayNode::redraw() {
    setWorkingFlag(true);

    return SUCCESS;
  }

  void GLDisplayNode::prepareBuffer(Buffer* in_buffer) {
    if (!is_GL_context_valid) {
      // As all GL functions should be called in one thread,
      // do creation of GL stuff here, but only once
      createGLContext();
      if (enable_disp_ev) {
	display->enableDisplayEvents(disp_ev_config_file);
      }
    }

    // If called with a regular buffer, prepare the output
    // else assume the user resized the window
    if (!getWorkingFlag()) {
      // Swap R and B channels for rgba channelorder (GLX takes bgra)
      if (rgb_channel_order) {
        // convert buffer "in-place"
        working_buffer = in_buffer->getWriteableInstance();
        char tmp(' '); // temporal cache
        for (// loop over the pixels of the image
             unsigned int pixel = 0;
             pixel < working_buffer->getUsed();
             pixel += 4) {
          // one pixel is "RGBA"
          tmp = (working_buffer->getData())[pixel + 2]; // save rgBa value
          (working_buffer->getData())[pixel + 2] = (working_buffer->getData())[pixel + 0]; // move Rgba value to rgRa value
          (working_buffer->getData())[pixel + 0] = tmp; // restore Bgra value
          // now one pixel is "BGRA"
        }
      }
      else {
        // just forward the in_buffer Ptr
        working_buffer = in_buffer;
      }

      display->prepareOutput(working_buffer);
    }

// errror 20030507: as "checkResize() does not work yet, I comment it
//                  out here for the moment
//                  anayway, this code should then come BEFORE
//                  prepareOutput
//     // in any case, check if redrawing the window is needed
//     display_size_mutex.lock();
//     display->checkResize(display_width, display_height);
//     display_size_mutex.unlock();
    
//     cerr << "prepareBuffer: drawImage!!!" << endl;
//     display->drawImage();
  }


  Message* GLDisplayNode::presentBuffer(Buffer* /* see note below */) {
    // do not use given buffer here as we might have requested a
    // writable instance in prepareBuffer yet and should NOT use the
    // given buffer any more. Use working_buffer instead.
    display->update(working_buffer);

    // do benchmarking only for real buffers, otherwise
    // in_buffer would be NULL!
    if (!getWorkingFlag()) {
      if(benchmark.getEnabled()) {
	benchmark.benchmark(working_buffer->getUsed());
	if((benchmark.getFrameCount() % 10) == 0) {
	  ostringstream frame_title_str;
	  frame_title_str << getName() << " " << setw(5) << setfill(' ')
			  << setiosflags(ios::fixed) << setprecision(1) << benchmark.getFPS()
			  << " fps" << endl;
	  display->setTitle(frame_title_str.str());
	}
      }
      working_buffer->release();
      working_buffer = NULL;
    }

    return 0;
  }

  void GLDisplayNode::enableDisplayEvents(const string& _config_file /* = "" */) {
    // have to cache _config_file and start display events in run method because
    // display is created there when first buffer arrives
    disp_ev_config_file = _config_file;
    enable_disp_ev = true;
  }

  void GLDisplayNode::setFullscreen(const bool&) {}
  bool GLDisplayNode::getFullscreen() const {return false;}
  bool GLDisplayNode::hasFullscreen() const {return false;}
  bool GLDisplayNode::isFullscreenOnly() const {return false;}
  bool GLDisplayNode::makeScreenshot(const string &) const {return false;}
  
  void GLDisplayNode::disableDisplayEvents() {
    display->disableDisplayEvents();
  }

}
