/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "nmm/plugins/video/divx4/DivX4DecodeNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::DivX4DecodeNode, "DivX4DecodeNode")

namespace NMM {

  DivX4DecodeNode::DivX4DecodeNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericProcessorNode(name, qmode, qsize)
  {
    // initialization of variables - values taken as advertised by README-File of lib

    x_dim = 0;               // the x dimension of the frames to be decoded
    y_dim = 0;               // the y dimension of the frames to be decoded
    output_format = DEC_YV12;// the format of the image frame
    framerate = 0;           // the frame rate of the sequence to be decoded

    postproc_level = 0;      // level for postprocessing, range from 0 to 100
    handle = (long)this;     // create a unique handle to pass to the library

    timestamp = Timestamp();
    in_buf_ts = Timestamp();

    vid_delay = 0;

    released = true;         // at the beginning, the decoder is in state released

    compressor = "DIVX";     // try DIVX as default compressor

    skip_until_keyframe = false;
  }
  
  DivX4DecodeNode::~DivX4DecodeNode() {
  }
  
  Result DivX4DecodeNode::doInit() {
    // general input format
    Format *in_format;
    in_format = getOwnInputProperty()->addNewFormat("video/mpeg4");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::compressor_param,"div3");
    in_format->addStringParamValue(Format::compressor_param,"DIV3");
    in_format->addStringParamValue(Format::compressor_param,"div4");
    in_format->addStringParamValue(Format::compressor_param,"DIV4");
    in_format->addStringParamValue(Format::compressor_param,"divx");
    in_format->addStringParamValue(Format::compressor_param,"DIVX");
    in_format->addStringParamValue(Format::compressor_param,"dx50");
    in_format->addStringParamValue(Format::compressor_param,"DX50");
    in_format->addStringParamValue(Format::compressor_param,"xvid");
    in_format->addStringParamValue(Format::compressor_param,"XVID");
    in_format->addWildcard(Format::bitrate_param);
    getOwnInputProperty()->setDefaultFormat(in_format);
    
    in_format = getOwnInputProperty()->addNewFormat("video/mpeg4");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::compressor_param,"div3");
    in_format->addStringParamValue(Format::compressor_param,"DIV3");
    in_format->addStringParamValue(Format::compressor_param,"div4");
    in_format->addStringParamValue(Format::compressor_param,"DIV4");
    in_format->addStringParamValue(Format::compressor_param,"divx");
    in_format->addStringParamValue(Format::compressor_param,"DIVX", 0.1);
    in_format->addStringParamValue(Format::compressor_param,"dx50");
    in_format->addStringParamValue(Format::compressor_param,"DX50");
    in_format->addStringParamValue(Format::compressor_param,"xvid");
    in_format->addStringParamValue(Format::compressor_param,"XVID");
    getOwnInputProperty()->setDefaultFormat(in_format);
    
    // general output format
    Format* out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"yvu");
    out_format->addStringParamValue(Format::endian_param,"little");
    out_format->addStringParamValue(Format::colorspace_param,"yv12");
    out_format->addIntParamValue(Format::bitperpixel_param,12);
    out_format->addStringParamValue(Format::format_param,"planar");
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    // set I/O partners
    in_format->setIOPartner(out_format);
    out_format->setIOPartner(in_format);
    
    return SUCCESS;
  }
  
  Result DivX4DecodeNode::doDeinit() {
    if (released) {
      return SUCCESS;
    } else {
      return releaseDecoder(); 
    }
  }
  
  Result DivX4DecodeNode::doInitOutput() {
    // get values from in- and out-formats
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    
    // set internal variables
    x_dim = in_format->getIntValue(Format::x_resolution_param); 
    y_dim = in_format->getIntValue(Format::y_resolution_param);
    framerate = in_format->getFloatValue(Format::framerate_param);
    compressor = in_format->getStringValue(Format::compressor_param);
    int x_ratio = in_format->getIntValue(Format::ratio_x_param);
    int y_ratio = in_format->getIntValue(Format::ratio_y_param);
    
    // setup working input property
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    // setup working output property
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    o_format->addIntParamValue(Format::x_resolution_param, x_dim);
    o_format->addIntParamValue(Format::y_resolution_param, y_dim);
    o_format->addIntParamValue(Format::ratio_x_param, x_ratio);
    o_format->addIntParamValue(Format::ratio_y_param, y_ratio);
    o_format->addFloatParamValue(Format::framerate_param, framerate);
    o_format->addStringParamValue(Format::channelorder_param,"yvu");
    o_format->addStringParamValue(Format::endian_param,"little");
    o_format->addStringParamValue(Format::colorspace_param,"yv12");
    o_format->addIntParamValue(Format::bitperpixel_param, 12);
    o_format->addStringParamValue(Format::format_param,"planar");
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    // set I/O partners
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    
    return SUCCESS;
  }
  
  Result DivX4DecodeNode::doActivate() {
    int res;
    DEBUG_STREAM ("Got parameters: x_dim=" << x_dim << ", y_dim=" << y_dim << endl);

    // initialize the library
    DEC_PARAM divx;
    memset(&divx,0,sizeof(divx));
    divx.output_format = output_format;
    if ("DIV3" == compressor || "div3" == compressor) {
      divx.codec_version = 311; // for divx3.11
    } else if ("DIVX" == compressor || "divx" == compressor ||
               "div4" == compressor || "div4" == compressor) {
      divx.codec_version = 400; // for divx4
    } else {
      divx.codec_version = 500; // 500, divx5.0, try for all other
    }
    divx.build_number = 0;
    divx.x_dim = x_dim;
    divx.y_dim = y_dim;
    if((res = decore(handle, DEC_OPT_INIT, &divx, NULL)) != DEC_OK) {
      DEBUG_STREAM("divx-library open error; result of init:" << res << endl);
      return(FAILURE); 
    }

    // set the postprocessing quality
    DEC_SET dec_set;
    dec_set.postproc_level = postproc_level;
    if((res = decore(handle, DEC_OPT_SETPP, &dec_set, NULL)) != DEC_OK) {
      DEBUG_STREAM("Error while setting postprocessing level; result of init:" << res << endl);
      return(FAILURE); 
    }
    
    // now the decoder is not released any more
    released = false;

    return SUCCESS;
  }
  
  Result DivX4DecodeNode::releaseDecoder() {
    // release the decoder
    int res;
    if ((res = decore(handle, DEC_OPT_RELEASE, NULL, NULL)) != DEC_OK) {
      DEBUG_STREAM("Release of Decoder failed! result of decore: " << res << endl);
      return FAILURE;
    }

    DEBUG_STREAM("Decoder released successfully!" << endl);
    // now the decoder is released
    released = true;
    return SUCCESS;
  }

  Result DivX4DecodeNode::startTrack(const string&) {
    return SUCCESS;
  }

  Result DivX4DecodeNode::doDeactivate() {
    releaseDecoder();
    released = true;
    return SUCCESS;
  }
  
  Result DivX4DecodeNode::endTrack() {
    DEBUG_STREAM("Received Event: endTrack");
    if (released) {
      return SUCCESS;
    } else {
      return releaseDecoder();
    }
  }
  
  Result DivX4DecodeNode::finished() {
    return SUCCESS;
  }
  
  Result DivX4DecodeNode::setVideoDelay(const int& delay) {
    DEBUG_STREAM("Received Event: video_delay. delay is " << delay << endl);
    vid_delay = delay;
    return SUCCESS;
  }
  
  Message* DivX4DecodeNode::processBuffer(Buffer* in_buffer) {
    MutexGuard divx_mg(divx_mutex);
    Buffer *out_buffer;
    
    // request buffer - they have a fix size for a single video stream
    out_buffer = getNewBuffer(x_dim * y_dim * 12/8, in_buffer);

    //set DEC_FRAME values
    decframe.bitstream = in_buffer->getData();
    // stride obviously is the offset to the next line in a picture
    // and obviously, if it is set to 0, the library calcutes it correctly by itself
    // (there was no documentation about that in the libraries description)
    // errror 20030620: divx-lib cannot compute stride correct for
    // div3, setting it to correct value width. Again there was no
    // documentation about this fact, found right value in sources of MPlayer
    decframe.stride = x_dim;
    decframe.bmp = out_buffer->getData();
    decframe.length = in_buffer->getUsed();
    decframe.render_flag = 1;

    if (decore(handle, DEC_OPT_FRAME, &decframe, &frameinfo) != DEC_OK) {
      ERROR_STREAM("FAILURE WHILE DECODING FRAME" << endl);
    }

    // This is strange, because the library should provide us the correct value, but it doesn't...
    // So set size as expected in YV12
    out_buffer->setUsed(static_cast<unsigned int>(x_dim * y_dim * 1.5));

    in_buf_ts = in_buffer->getTimestamp();

    if (!in_buf_ts.is_valid) {
      // Caculate time-stamp for current frame
      timestamp.stream_counter = in_buf_ts.stream_counter;//frame_count;
      save_ts.sync_time += Interval(static_cast<long>((1 / framerate) * nano_per_sec));
      timestamp.sync_time = save_ts.sync_time;//Time(now);
      timestamp.is_valid = true;

      DEBUG_STREAM("Set Timestamp " << timestamp << endl);
      out_buffer->setTimestamp(timestamp);
    }
    else {
      DEBUG_STREAM("Found Timestamp " << in_buf_ts << endl);
      DEBUG_STREAM("Leaving it untouched." << endl);
      save_ts = in_buf_ts;
    }
    
    if (skip_until_keyframe) {
      if (frameinfo.intra) {
        DEBUG_STREAM("KeyFrame: " << out_buffer->getTimestamp().stream_counter << endl);
        skip_until_keyframe = false;
      } else {
        DEBUG_STREAM("Dropping non-keyframe!" << endl);
        out_buffer->release();
        out_buffer = NULL;
      }
    }

    in_buffer->release();
    return out_buffer;
  }  

  Result DivX4DecodeNode::doFlush() {
    skip_until_keyframe = true;
    return SUCCESS;
  }
  
  Result DivX4DecodeNode::setKeyframe() {
    /* there are no keyframes in raw video streams, so delete event */
    return DELETE;
  }
}
