/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "MagickManipulationNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::MagickManipulationNode, "MagickManipulationNode")

namespace NMM {
  MagickManipulationNode::MagickManipulationNode(const char* name)
    : GenericProcessorNode(name), bit_per_pixel(0), border_r(0), border_g(0), border_b(0)
  {}

  MagickManipulationNode::~MagickManipulationNode()
  {
    // Release the manipulators when the node object is destructed.
    clearManipulators();
  }

  Result MagickManipulationNode::doInit()
  {
    // Initialize supported input formats.
    // The formats rgb24 with 24 and 32 bits are supported.
    Format* myInputFormat = getOwnInputProperty()->addNewFormat("video/raw");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addWildcard(Format::ratio_x_param);
    myInputFormat->addWildcard(Format::ratio_y_param);
    myInputFormat->addWildcard(Format::framerate_param);
    myInputFormat->addStringParamValue(Format::channelorder_param, "rgb");
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addStringParamValue(Format::colorspace_param, "rgb24");
    myInputFormat->addIntParamValue(Format::bitperpixel_param, 32);
    myInputFormat->addStringParamValue(Format::format_param, "packet");

    // The default format has 32 bits.
    getOwnInputProperty()->setDefaultFormat(myInputFormat);

    myInputFormat = getOwnInputProperty()->addNewFormat("video/raw");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addWildcard(Format::ratio_x_param);
    myInputFormat->addWildcard(Format::ratio_y_param);
    myInputFormat->addWildcard(Format::framerate_param);
    myInputFormat->addStringParamValue(Format::channelorder_param, "rgb");
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addStringParamValue(Format::colorspace_param, "rgb24");
    myInputFormat->addIntParamValue(Format::bitperpixel_param, 24);
    myInputFormat->addStringParamValue(Format::format_param, "packet");

    // Initialize supported output formats.
    // The formats rgb24 with 24 and 32 bits are supported.
    Format* myOutputFormat = getOwnOutputProperty()->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addWildcard(Format::framerate_param);
    myOutputFormat->addStringParamValue(Format::channelorder_param, "rgb");
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addStringParamValue(Format::colorspace_param, "rgb24");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param, 24);
    myOutputFormat->addStringParamValue(Format::format_param, "packet");

     // The default format has 32 bits.
    getOwnOutputProperty()->setDefaultFormat(myOutputFormat);

    myOutputFormat = getOwnOutputProperty()->addNewFormat("video/raw");

    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addWildcard(Format::framerate_param);
    myOutputFormat->addStringParamValue(Format::channelorder_param, "rgb");
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addStringParamValue(Format::colorspace_param, "rgb24");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param, 24);
    myOutputFormat->addStringParamValue(Format::format_param, "packet");


    return SUCCESS;    
  }

  Result MagickManipulationNode::doInitOutput() {
    const Format* in_format = getInputFormat();
    if (in_format == NULL)
      return FAILURE;

    try {
      // Initialize internal parameters.
      bit_per_pixel    = in_format->getIntValue(Format::bitperpixel_param);
      image_resolution.first   = in_format->getIntValue(Format::x_resolution_param);
      image_resolution.second  = in_format->getIntValue(Format::y_resolution_param);

      new_image_resolution =  image_resolution;

      // Some manipulators might alter the image resolution.
      for (list<Manipulator*>::iterator it = manipulators.begin(); it != manipulators.end(); it++) {
	(*it)->manipulateSize(new_image_resolution.first, new_image_resolution.second);
      }

      // Set the output formats.
      Format* rgb_in_format = getOwnWorkingInputProperty()->addNewFormat("video/raw");
      rgb_in_format->addIntParamValue(Format::x_resolution_param, image_resolution.first);
      rgb_in_format->addIntParamValue(Format::y_resolution_param, image_resolution.second);
      rgb_in_format->addIntParamValue(Format::ratio_x_param, new_image_resolution.first);
      rgb_in_format->addIntParamValue(Format::ratio_y_param, new_image_resolution.second);
      rgb_in_format->addStringParamValue(Format::channelorder_param,"rgb");
      rgb_in_format->addStringParamValue(Format::endian_param, "little");
      rgb_in_format->addStringParamValue(Format::colorspace_param,"rgb24");
      rgb_in_format->addIntParamValue(Format::bitperpixel_param,bit_per_pixel);
      rgb_in_format->addStringParamValue(Format::format_param, "packet");
      rgb_in_format->addFloatParamValue(Format::framerate_param, in_format->getFloatValue(Format::framerate_param));
      getOwnWorkingInputProperty()->setDefaultFormat(rgb_in_format);

      Format* rgb_out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      rgb_out_format->addIntParamValue(Format::x_resolution_param, new_image_resolution.first);
      rgb_out_format->addIntParamValue(Format::y_resolution_param, new_image_resolution.second);
      rgb_out_format->addIntParamValue(Format::ratio_x_param, new_image_resolution.first);
      rgb_out_format->addIntParamValue(Format::ratio_y_param, new_image_resolution.second);
      rgb_out_format->addStringParamValue(Format::channelorder_param,"rgb");
      rgb_out_format->addStringParamValue(Format::endian_param, "little");
      rgb_out_format->addStringParamValue(Format::colorspace_param,"rgb24");
      rgb_out_format->addIntParamValue(Format::bitperpixel_param,bit_per_pixel);
      rgb_out_format->addStringParamValue(Format::format_param, "packet");
      rgb_out_format->addFloatParamValue(Format::framerate_param, in_format->getFloatValue(Format::framerate_param));
      getOwnWorkingOutputProperty()->setDefaultFormat(rgb_out_format);
      rgb_in_format->setIOPartner(rgb_out_format);
      rgb_out_format->setIOPartner(rgb_in_format);
    } catch (...) {
      return FAILURE;
    }

    return SUCCESS;
  }

  Message* MagickManipulationNode::processBuffer(Buffer* buffer)
  {
    // Create new Magick::Image object.
    Magick::Image image;
    
    buffer = buffer->getWriteableInstance();
    char* data = buffer->getData();
    Timestamp ts = buffer->getTimestamp();

    // Read the pixel data from the buffer into the image.
    try
    {
	if (bit_per_pixel == 24) {
	image.read(image_resolution.first, image_resolution.second,
			"BGR", Magick::CharPixel, data);
	}
	else {
	image.read(image_resolution.first, image_resolution.second,
			"BGRA", Magick::CharPixel, data);
	}

	// Set the border color before applying any manipulations.
	image.borderColor(Magick::ColorRGB(border_r, border_g, border_b));

	// Apply one manipulation after another to the image.
	for (list<Manipulator*>::iterator it = manipulators.begin(); it != manipulators.end(); it++) {
	  (*it)->manipulate(image);
	}
    } 
    catch (const Magick::Exception& e) {
      ERROR_STREAM(e.what() << endl);
    }
    catch (...) {
      ERROR_STREAM("Unknown exception." << endl);
    }


    if (new_image_resolution != image_resolution) {
      buffer->release();
      buffer = getNewBuffer(new_image_resolution.first * new_image_resolution.second * (bit_per_pixel >> 3));
      buffer->setTimestamp(ts);
      data = buffer->getData();
    }

    // Write the pixel data from the image back into the buffer.
    try
    {
	if (bit_per_pixel == 24) {
	image.write(0, 0, new_image_resolution.first, new_image_resolution.second,
			"BGR", Magick::CharPixel, data);
	}
	else {
	image.write(0, 0, new_image_resolution.first, new_image_resolution.second,
			"BGRA", Magick::CharPixel, data);
	}
    }
    catch (const Magick::Exception& e)
    {
      ERROR_STREAM(e.what() << endl);
    }
    catch (...) {
      ERROR_STREAM("Unknown exception." << endl);
    }
    
    // Send the buffer.
    return buffer;
  }

  void MagickManipulationNode::clearManipulators() {
    // Delete all manipulators.
    for (list<Manipulator*>::iterator it = manipulators.begin();
	 it != manipulators.end(); it++) {
      delete *it;
    }

    // Clear the list.
    manipulators.clear();
  }

  void MagickManipulationNode::setBorderColor(const double& r, const double& g, const double& b) {
    border_r = r;
    border_g = g;
    border_b = b;
  }

  // The following methods create the corresonding manipulator objects and add them to the list.
  void MagickManipulationNode::adaptiveThreshold(const unsigned int& width,
						    const unsigned int& height,
						    const unsigned int& offset) {
    manipulators.push_back(new AdaptiveThresholdManipulator(width, height, offset));
  }
  void MagickManipulationNode::addNoise(const string& noise_type) {
    manipulators.push_back(new AddNoiseManipulator(noise_type));
  }
  void MagickManipulationNode::annotate(const string& text,
					   const string& bounding_area,
					   const string& gravity,
					   const double& degrees) {
    manipulators.push_back(new AnnotateManipulator(text, bounding_area, gravity, degrees));
  }
  void MagickManipulationNode::blur(const double& radius, const double& sigma) {
    manipulators.push_back(new BlurManipulator(radius, sigma));
  }
  void MagickManipulationNode::border(const string& geometry) {
    manipulators.push_back(new BorderManipulator(geometry));
  }
  void MagickManipulationNode::channel(const string& layer) {
    manipulators.push_back(new ChannelManipulator(layer));
  }
  void MagickManipulationNode::charcoal(const double& radius, const double& sigma) {
    manipulators.push_back(new CharcoalManipulator(radius, sigma));
  }
  void MagickManipulationNode::chop(const string& geometry) {
    manipulators.push_back(new ChopManipulator(geometry));
  }
  void MagickManipulationNode::colorize(const unsigned int& opacity_red,
					   const unsigned int& opacity_green,
					   const unsigned int& opacity_blue,
					   const double& pen_color_r,
					   const double& pen_color_g,
					   const double& pen_color_b) {
    manipulators.push_back(new ColorizeManipulator(opacity_red,
						     opacity_green,
						     opacity_blue,
						     pen_color_r,
						     pen_color_g,
						     pen_color_b));
  }
  void MagickManipulationNode::comment(const string& comment)
  {
    manipulators.push_back(new CommentManipulator(comment));
  }
  void MagickManipulationNode::contrast(const unsigned int& sharpen) {
    manipulators.push_back(new ContrastManipulator(sharpen));
  }
  void MagickManipulationNode::crop(const string& geometry) {
    manipulators.push_back(new CropManipulator(geometry));
  }
  void MagickManipulationNode::cycleColormap(const int& amount) {
    manipulators.push_back(new CycleColormapManipulator(amount));
  }
  void MagickManipulationNode::despeckle() {
    manipulators.push_back(new DespeckleManipulator());
  }
  void MagickManipulationNode::edge(const unsigned int& radius ) {
    manipulators.push_back(new EdgeManipulator(radius));
  }
  void MagickManipulationNode::emboss(const double& radius , const double& sigma) {
    manipulators.push_back(new EmbossManipulator(radius, sigma));
  }
  void MagickManipulationNode::enhance() {
    manipulators.push_back(new EnhanceManipulator());
  }
  void MagickManipulationNode::equalize() {
    manipulators.push_back(new EqualizeManipulator());
  }
  void MagickManipulationNode::erase() {
    manipulators.push_back(new EraseManipulator());
  }
  void MagickManipulationNode::flip() {
    manipulators.push_back(new FlipManipulator());
  }
  void MagickManipulationNode::floodFillColor(const unsigned int& x, const unsigned int& y, const double& fill_color_r, const double& fill_color_g, const double& fill_color_b) {
    manipulators.push_back(new FloodFillColorManipulator(x, y, fill_color_r, fill_color_g, fill_color_b));
  }
  void MagickManipulationNode::floodFillOpacity(const long& x, const long& y, const unsigned int& opacity, const string& method) {
    manipulators.push_back(new FloodFillOpacityManipulator(x, y, opacity, method));
  }
  void MagickManipulationNode::flop() {
    manipulators.push_back(new FlopManipulator());
  }
  void MagickManipulationNode::frame(const unsigned int& width, const unsigned int& height, const int& x, const int& y, const int& innerBevel, const int& outerBevel) {
    manipulators.push_back(new FrameManipulator(width, height, x, y, innerBevel, outerBevel));
  }
  void MagickManipulationNode::gamma(const double& gamma_red, const double& gamma_green, const double& gamma_blue) {
    manipulators.push_back(new GammaManipulator(gamma_red, gamma_green, gamma_blue));
  }
  void MagickManipulationNode::gaussianBlur(const double& width, const double& sigma) {
    manipulators.push_back(new GaussianBlurManipulator(width, sigma));
  }
  void MagickManipulationNode::implode(const double& factor) {
    manipulators.push_back(new ImplodeManipulator(factor));
  }
  void MagickManipulationNode::label(const string& name) {
    manipulators.push_back(new LabelManipulator(name));
  }
  void MagickManipulationNode::magnify() {
    manipulators.push_back(new MagnifyManipulator());
  }
  void MagickManipulationNode::matteFloodfill(const double& target_r, const double& target_g, const double& target_b, const unsigned int& opacity, const int& x, const int& y, const string& method) {
    manipulators.push_back(new MatteFloodfillManipulator(target_r, target_g, target_b, opacity, x, y, method));
  }
  void MagickManipulationNode::medianFilter(const double& radius) {
    manipulators.push_back(new MedianFilterManipulator(radius));
  }
  void MagickManipulationNode::minify() {
    manipulators.push_back(new MinifyManipulator());
  }
  void MagickManipulationNode::modulate(const double& brightness, const double& saturation, const double& hue) {
    manipulators.push_back(new ModulateManipulator(brightness, saturation, hue));
  }
  void MagickManipulationNode::negate() {
    manipulators.push_back(new NegateManipulator());
  }
  void MagickManipulationNode::normalize() {
    manipulators.push_back(new NormalizeManipulator());
  }
  void MagickManipulationNode::oilPaint(const unsigned int& radius) {
    manipulators.push_back(new OilPaintManipulator(radius));
  }
  void MagickManipulationNode::opacity(const unsigned int& opacity) {
    manipulators.push_back(new OpacityManipulator(opacity));
  }
  void MagickManipulationNode::opaque(const double& opaque_color_r, const double& opaque_color_g, const double& opaque_color_b, const double& pen_color_r, const double& pen_color_g, const double& pen_color_b) {
    manipulators.push_back(new OpaqueManipulator(opaque_color_r, opaque_color_g, opaque_color_b, pen_color_r, pen_color_g, pen_color_b));
  }
  void MagickManipulationNode::quantize(const bool& measureError) {
    manipulators.push_back(new QuantizeManipulator(measureError));
  }
  void MagickManipulationNode::raise(const string& geometry, const bool& raisedFlag) {
    manipulators.push_back(new RaiseManipulator(geometry, raisedFlag));
  }
  void MagickManipulationNode::reduceNoise(const unsigned int& order) {
    manipulators.push_back(new ReduceNoiseManipulator(order));
  }
  void MagickManipulationNode::roll(const int& columns, const int& rows) {
    manipulators.push_back(new RollManipulator(columns, rows));
  }
  void MagickManipulationNode::rotate(const double& degrees) {
    manipulators.push_back(new RotateManipulator(degrees));
  }
  void MagickManipulationNode::sample(const string& geometry) {
    manipulators.push_back(new SampleManipulator(geometry));
  }
  void MagickManipulationNode::scale(const string& geometry) {
    manipulators.push_back(new ScaleManipulator(geometry));
  }
  void MagickManipulationNode::segment(const double& cluster_threshold , const double& smoothing_threshold ) {
    manipulators.push_back(new SegmentManipulator(cluster_threshold, smoothing_threshold));
  }
  void MagickManipulationNode::shade(const double& azimuth, const double& elevation, const bool& color_shading) {
    manipulators.push_back(new ShadeManipulator(azimuth, elevation, color_shading));
  }
  void MagickManipulationNode::sharpen(const double& radius, const double& sigma) {
    manipulators.push_back(new SharpenManipulator(radius, sigma));
  }
  void MagickManipulationNode::shave(const string& geometry) {
    manipulators.push_back(new ShaveManipulator(geometry));
  }
  void MagickManipulationNode::shear(const double& x_shear_angle, const double& y_shear_angle) {
    manipulators.push_back(new ShearManipulator(x_shear_angle, y_shear_angle));
  }
  void MagickManipulationNode::solarize(const double& factor ) {
    manipulators.push_back(new SolarizeManipulator(factor));
  }
  void MagickManipulationNode::spread(const unsigned int& amount ) {
    manipulators.push_back(new SpreadManipulator(amount));
  }
  void MagickManipulationNode::swirl(const double& degrees) {
    manipulators.push_back(new SwirlManipulator(degrees));
  }
  void MagickManipulationNode::threshold(const double& threshold) {
    manipulators.push_back(new ThresholdManipulator(threshold));
  }
  void MagickManipulationNode::transform(const string& image_geometry, const string& crop_geometry) {
    manipulators.push_back(new TransformManipulator(image_geometry, crop_geometry));
  }
  void MagickManipulationNode::transparent(const double& color_r, const double& color_g, const double& color_b) {
    manipulators.push_back(new TransparentManipulator(color_r, color_g, color_b));
  }
  void MagickManipulationNode::trim() {
    manipulators.push_back(new TrimManipulator());
  }
  void MagickManipulationNode::unsharpmask(const double& radius, const double& sigma, const double& amount, const double& threshold) {
    manipulators.push_back(new UnsharpmaskManipulator(radius, sigma, amount, threshold));
  }
  void MagickManipulationNode::wave(const double& amplitude, const double& wavelength) {
    manipulators.push_back(new WaveManipulator(amplitude, wavelength));
  }
  void MagickManipulationNode::zoom(const string& geometry){
    manipulators.push_back(new ZoomManipulator(geometry));
  }
}
