/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_WINDOW_HPP_
#define _NMM_WINDOW_HPP_ _NMM_WINDOW_HPP_

#include "nmm/base/NMMObject.hpp"
#include "nmm/misc/scaler/VideoScaler.hpp"
#include "nmm/interfaces/video/osd/manager/IWindowImpl.hpp"

#include "widgets/widgets.hpp"

#include <pthread.h>

namespace NMM {
  
  class Window : public IWindowImpl {
  private:
    OSDManagerNode* manager;
    
  protected:
    pthread_mutex_t mutex;
    
    Widget* vc;			// Needed to call draw function of a Widget
    workspace_t workspace;	// workspace to pass to the draw functions
    
    int         id;             // window ID
    int	        xpos;		// x pos. of upper left corner of window
    int         ypos;		// y pos. of upper left corner of window
    bool        updateflag;	// indicates, if window should be redrawn
    bool        hidden;		// indicates, if window is hidden
    
    /**
    * Creates workspace (in rgba format)
    * \param width width of workspace
    * \param height height of workspace
    * \returns true if workspace could be created, false otherwise
    */
    virtual bool createWorkSpace( int width, int height );

    /**
    * deletes created workspace
    */
    virtual void deleteWorkSpace();

    /**
    * \returns status of update flag (true or false)
    */
    bool getUpdateFlag();

    /**
    * set update flag
    * \param b true, if window should be redrawn within next get-call
    */
    void setUpdateFlag( bool b );
    
  public:
    Window(OSDManagerNode* manager);
    virtual ~Window();
    
    /**
     * IWindow interface.
     * These methods are called locally and are therefore public.
     * (better design would be to have two sets of methods)
     */
    virtual int getWindowID() const;
    virtual void setPosition(const unsigned int& x, const unsigned int& y);
    virtual void getPosition(unsigned int& x, unsigned int& y) const;
    virtual void setDimension(const unsigned int& w, const unsigned int& h);
    virtual void getDimension(unsigned int& w, unsigned int& h) const;
    virtual void update();
    virtual bool getVisible() const;
    virtual void setVisible(const bool& value);
    virtual void setScaleFactor( const double& scale_x, const double& scale_y ) = 0;
    virtual void setContent(const IWidget& value);
    virtual IWidget* getContent() const;
    
    int getID() const { return id; }
    void setID(int id) { this->id = id; }
    
    /**
     * Creates a window.
     * @param w witdh of the window
     * @param h height of the window
     * @return true, if window could be created succesfully
     */
    bool create( int w, int h );

    /**
     * Deletes the created window.
     */
    void remove();
    
    /**
     * Put widget in a window.
     * @param vc pointer to a widget
     */
    void setContentInternal( Widget* vc );
    
    /**
     * Get widget in a window.
     * @param vc pointer to a widget
     */
    Widget* getContentInternal() const;
    
    /**
     * Lock the window, so that no update() can be done
     */
    void lock();

    /**
     * allow update() call again
     */
    void unlock();

    /**
    * test if window is hidden
    * \returns true if window is hidden
    */
    bool isHidden();

    /**
    * hides the window
    */
    void hide();

    /**
    * makes window visible
    */
    void unHide();

    /**
    * \returns pointer to rgba pixel data of the window, x,y contain the x and y position of the window
    *                                                    w, h contains with and height of the window
    * \returns NULL, if window should not be drawn (window deleted or hidden)
    */
    virtual unsigned char* get( int* x, int* y, int* w, int *h );

  };

  class WindowYV12A: public Window {
  protected:
    workspace_t    yv12a_workspace;
    VideoScaler    scaler;
    unsigned char* scaled_data;
    bool           scale_factor_changed;
    double         scale_factor_x;
    double         scale_factor_y;

    /**
    * Creates workspace (in YV12A format)
    * \param width width of workspace
    * \param height height of workspace
    * \returns true if workspace could be created, false otherwise
    */
    bool createWorkSpace( int width, int height );

    /**
    * deletes created workspace
    */
    void deleteWorkSpace();

    /**
    * converts rgba bitmap to yv12a
    * \param yuv_buffer pointer to destination buffer (make sure enough buffer space is allocated)
    * \param rgb_buffer pointer to rgba bitmap
    * \param width width of the bitmap
    * \param height height of the bitmap
    */
    void convertRGBAtoYV12A( unsigned char* yuv_buffer, unsigned char* rgbbuffer, int width, int height );


  public:
    WindowYV12A(OSDManagerNode* manager);
    virtual ~WindowYV12A();

    /**
    *   - if necessary, converts rgba window to yv12a
    *   - returns information about the window
    * \returns pointer to yv12a pixel data, x,y contain the x and y position of the window
    *          w, h contains with and height of the window
    * \returns NULL, if window should not be drawn (window deleted or hidden)
    */
    unsigned char* get( int* x, int* y, int* w, int *h );

    /**
    * set scale factor
    * this has influence to the get method
    * before data is delivered, the workspace is scaled
    * \param scale_x scale factor that is multiplied with width and x coordinate
    * \param scale_y scale factor that is multiplied with height and y coordinate
    */
    void setScaleFactor( const double& scale_x, const double& scale_y );

  };
} // NMM

#endif //_NMM_WINDOW_HPP_
