/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _SINGLETON_HPP
#define _SINGLETON_HPP

#include <stdlib.h>

/**
  */

/* This code is base on the book:
 *  * Alexandrescu, Andrei. "Modern C++ Design: Generic Programming and Design
 *  *    Patterns Applied". Copyright (c) 2001. Addison-Wesley.
 */

template <class T>
class SingleThreaded {
public:
  typedef  T VolatileType;
};

template <class T>
class MultiThreaded {
public:
  typedef T VolatileType;
};

template <class T>
class CreateUsingNew {
public:
  static T* create()           {return new T();}
  static void destroy(T* obj)  {delete obj;}
};

template <class T>
class DefaultLifetime {
public:
  typedef void (*deleteMethod)();
  static void scheduleDestruction( deleteMethod d) { atexit(d); }
  static void onDeadReference()                   { /*throw std::logic_error("Dead Reference Detected");*/ }
};

template <
  class T,
  template <class> class CreationPolicy = CreateUsingNew,
  template <class> class LifetimePolicy = DefaultLifetime,
  template <class> class ThreadingModel = MultiThreaded
>
class Singleton {
public:
  static T& getInstance();

protected:
  Singleton(){};

private:
  static void destroySingleton();
  typedef typename ThreadingModel<T>::VolatileType InstanceType;
  static InstanceType* _instance;
  static  bool destroyed;
};



/** Implementation */

template <
  class T,
  template <class> class CreationPolicy, //creation policy
  template <class> class LifetimePolicy, //Lifetime policy
  template <class> class ThreadingModel  //Threading model
>

T& Singleton<T, CreationPolicy, LifetimePolicy, ThreadingModel>::getInstance() {
  if (!_instance) {
//    typename ThreadingModel::Lock guard;
    if (!_instance) {
      if (destroyed) {
        LifetimePolicy<T>::onDeadReference();
        destroyed = false;
      }
      _instance = CreationPolicy<T>::create();
      LifetimePolicy<T>::scheduleDestruction(&destroySingleton);
    }
  }
  return *_instance;
}

template <
  class T,
  template <class> class CreationPolicy, //creation policy
  template <class> class LifetimePolicy, //Lifetime policy
  template <class> class ThreadingModel  //Threading model
>
void Singleton<T, CreationPolicy, LifetimePolicy, ThreadingModel>::destroySingleton() {
  CreationPolicy<T>::destroy(_instance);
  _instance = 0;
  destroyed = true;
}

template <
  class T,
  template <class> class CreationPolicy, //creation policy
  template <class> class LifetimePolicy, //Lifetime policy
  template <class> class ThreadingModel  //Threading model
>
typename Singleton<T, CreationPolicy, LifetimePolicy, ThreadingModel>::InstanceType*
Singleton<T, CreationPolicy, LifetimePolicy, ThreadingModel>::_instance = 0;

template <
  class T,
  template <class> class CreationPolicy, //creation policy
  template <class> class LifetimePolicy, //Lifetime policy
  template <class> class ThreadingModel  //Threading model
>
bool Singleton<T, CreationPolicy, LifetimePolicy, ThreadingModel>::destroyed = false;




#endif
