/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "Fork.hpp"
#include <iostream>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

#define BUFFERSIZE 300

namespace NMM{

  Fork::Fork()
    : ThreadedObject()
  {
    internalProcessID = 0;
    childPid=-1;
    socketpair(AF_UNIX, SOCK_STREAM, 0, sv);
    stopThread = false;
    start();
    activateThread();
  }
  
  
  Fork::Fork(const Fork&)
    :ThreadedObject()
  {
    /*
    internalProcessID = 0;
    childPid=-1;
    socketpair(AF_UNIX, SOCK_STREAM, 0, sv);
    stopThread = false;
    activateThread();
    start();
    */
  }
  
  Fork::~Fork(){
    stopThread = true;
    //killall running commands
    for(map<int, ProcessDescription*>::iterator i = runningChilds.begin(); i != runningChilds.end(); i++){
      kill(i->second->pid, SIGKILL);
    }

    write(sv[0], "-1 exit",8 );
    close(sv[0]);
    close(sv[1]);
    deactivateThread();

    //cleanup left ProcessDescriptions
    for(map<int, ProcessDescription*>::iterator i = runningChilds.begin(); i != runningChilds.end(); i++){
      if(i->second) delete i->second;
    }
    runningChilds.clear();
  }
  
  int Fork::getReadSocket(){
    return sv[0];
  }
  
  int Fork::getWriteSocket(){
    return sv[1];
  }


  int Fork::killall(){
    stopThread = true;
    for(map<int, ProcessDescription*>::iterator i = runningChilds.begin(); i != runningChilds.end(); i++){
      kill(i->second->pid, SIGKILL);
    }
    kill(childPid, SIGKILL);

    return 0;
  }

  
  /**
   * execute the given command
   */
  int Fork::executeCommand(string command){
    //lock the mutex
    executemutex.lock();
    internalProcessID++; //increment internalProcessID to be safe, that the ID is unique
    
    //For Multithreading: for the Main-Program; we need to know, for wich ID we are waiting, so we can return the proper return-value
    char tmp[BUFFERSIZE];
    memset(&tmp, 0, BUFFERSIZE);
    sprintf(tmp, "%i %s", internalProcessID, command.c_str());
    int waitForID = internalProcessID;

    //cerr << "EXECUTE COMMAND: waiting for " << tmp << endl;
    
    //create our ProcessDescription
    ProcessDescription *pd = new ProcessDescription();
    pd->id = waitForID;
    pd->finished = false;
    pd->getThreadMutex().lock();

    //add the ProcessDescription to the map that the Control-Thread can wake us up
    runningChilds[waitForID] = pd;

    //We send our ID and Command-line to the Main-Child-Process to start the process
    write(sv[0], tmp , sizeof(tmp));

    executemutex.unlock();

    //wait, until the control-Thread wakes us up, when our issued command is finished. 
    while(!stopThread){
      pd->getThreadCondition().wait();
      if(stopThread){
        pd->getThreadMutex().unlock();
        return -1;
      }
      if(pd->finished){
        pd->getThreadMutex().unlock();
        break;
      }else{
      }
    }

    //Just to be sure ... 
    pd->getThreadMutex().unlock();

    //get the return-value of our issued command to return it
    int retval = runningChilds[waitForID]->returnValue;

    //cerr <<"EXECUTECOMMAND: return " << retval << " of " << waitForID 
    //     << " with pid " << runningChilds[waitForID]->pid << " " << strerror(errno) << endl;

    if(runningChilds[waitForID]) delete runningChilds[waitForID];
    runningChilds.erase(waitForID);

    return retval;
  }
  
  /**
   * Control-Thread 
   * waits for notification from the Child-Processes, evaluates the buffer and notifies the waiting executeCommand-Call
   */
  void Fork::run(){
    char recvBuffer[BUFFERSIZE]; 
    
    do{
      
      //wait until a buffer is received
      int len = read(sv[0], recvBuffer, sizeof(recvBuffer));
      recvBuffer[len] = '\0'; //cap buffer

      string gotBuffer(recvBuffer);
      int pos1 = gotBuffer.find_first_of(" ");
      
      //separate the internalProcessID from the buffer to determine, wich executeCommand has issued the command so we can wake him up
      int finishedID = atoi(gotBuffer.substr(0,pos1).c_str());
      
      int pos2 = gotBuffer.find(" ", pos1+1);

      if(finishedID != -1){
        //fill the ProcessDescription with our determined values and wake up the executeCommand-Call
        runningChilds[finishedID]->getThreadMutex().lock();
        runningChilds[finishedID]->pid = atoi(gotBuffer.substr(pos1+1, pos2).c_str());
        runningChilds[finishedID]->returnValue = atoi(gotBuffer.substr(pos2+1).c_str());
        runningChilds[finishedID]->finished = true;
        runningChilds[finishedID]->getThreadCondition().notify();
        runningChilds[finishedID]->getThreadMutex().unlock();
      }
    }while(!stopThread);

  }

  void Fork::start(){
    //Ignore SIGCHLD that childs aren't displayed <defunct> in 'ps aux' since we don't care about their exit-status
    //exit-satus is transmitted by sockets
    //    signal(SIGCHLD,SIG_IGN);
    
    //These 2 variables are needed to catch the exit-message for Main-Cild-Process
    string gotBuffer;
    int pos = 0;

    //Start Main-Child-Process
    switch(childPid = ::fork()){    
    case -1:
      break;
    case 0: 
      //cerr << "  MAINCHILD: started with pid " << getpid() << " from parent " << getppid() << endl;
      do{
        char buffer[BUFFERSIZE];
        //Main-Child waits for incoming messages on the socket
        int len = read(sv[1], buffer,sizeof(buffer));
        buffer[len] = '\0'; //cap the received buffer
        //cerr << "  MAINCHILD got: " << buffer << endl << endl;
        
        //Store buffer for evaluation later
        gotBuffer = string(buffer);
        //Find the first space in the buffer, since our protocol-string is this: "ID COMMAND"
        pos = gotBuffer.find_first_of(" ");

        if(gotBuffer.substr(pos+1) == "exit") break;

        //Here, we can do some error-handling: On error-Messages, just exit and don't start another child.
        int secondChild = -1;
        
        //Start the second child to do the system()-call or exec()
        secondChild = ::fork();
        switch(secondChild){
        case -1:
          break;
        case 0:
          {
            
            //cerr << endl << "    SECONDCHILD: Ok, second child startet with pid " << getpid() << " from " << getppid() << endl;
            //cerr << "    SECONDCHILD: waits for 10 seconds with message " << buffer << endl;
            //This is the received buffer from main-Child-Process. It contains the commandline to execute
            string childBuffer = string(buffer);
            size_t childpos = childBuffer.find_first_of(" ");
            //int childID = atoi(childBuffer.substr(0,childpos).c_str());
            
            //cerr << "    SECONDCHILD: Execute command " << gotBuffer.substr(pos+1) << endl;
            int res = system(childBuffer.substr(childpos+1).c_str());
            //cerr << "    SECONDCHILD: Now writing back to parent of first child" << endl;
            //Send our return-code back to the main-program
            char t[20];
            memset(&t, 0, 20);
            
            sprintf(t, "%s %d %i", childBuffer.substr(0,childpos).c_str(), getpid(), res);
            write(sv[1], t, sizeof(t));
            
            //exit the child. exit(0) suffices since we don't care about return-codes 
            //cerr << "    SECONDCHILD: EXIT" << endl << endl;
            //Yes, it is rude. but as far as I see the only way to terminate the Child since waitpid does not work
            kill(getpid(), SIGKILL);
          }
        default:
          //don't wait for return-value

          waitpid(secondChild, 0, WUNTRACED);
          //This call allows the Mainchild not to wait for the second child, 
          //but processes appear efuncet und 'ps aux'. so it is disabled
          //waitpid(secondChild, 0, WUNTRACED | WNOHANG);
          break;
        }

        //now, we evaluate our Main-Child-Buffer, if we can go on
      }while(gotBuffer.substr(pos+1) != "exit");

      write(sv[1], "-1 exit" , 8);
      
      //cerr << "  MAINCHILD: exiting" << endl;;
      //Yes, it is rude. but as far as I see the only way to terminate the Child since waitpid does not work
      kill(getpid(), SIGKILL);
      exit(0);
      break;
    default:
      
      break;
    }
  } 


}
