/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                    Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "SockDGram.hpp"

#include <errno.h>       // for errnos
#include <netdb.h>       // for gethostbyname, getprotobyname
#include <unistd.h>      // for close
#include <signal.h>      // for signals

namespace NMM {

  /****************** class SockDGram ******************/
  SockDGram::SockDGram(bool owner /* = true */)
    : __socket(-1),
      __own_socket(owner)
  {
    struct protoent* protocol = getprotobyname("udp");
    __socket = ::socket(PF_INET, SOCK_DGRAM, protocol->p_proto);
  }
  
  SockDGram::SockDGram(int socket, bool owner) 
    : __socket(socket),
      __own_socket(owner)
  {
  }

  SockDGram::~SockDGram() {
    if (__own_socket && __socket >= 0) {
      ::close(__socket);
      __socket = -1;
    }
  }
  
  void SockDGram::setPeerAddress(const INetAddr& addr) {
    __peer_addr = addr;
  }
  
  void SockDGram::setSocket(int socket, bool owner /* = false */) {
    if (__own_socket && __socket >= 0) {
      ::close(__socket);
    }
    
    __socket = socket;
    __own_socket = owner;
  }
  
  INetAddr SockDGram::getLocalAddress() const {
    struct sockaddr_in addr;
    socklen_t len_addr = sizeof(addr);
    
    if (getsockname(__socket, reinterpret_cast<struct sockaddr*>(&addr), &len_addr) == 0) {
      INetAddr address;
      address.setAddress(&addr);
      return address;
    } else {
      GLOBAL_ERROR_STREAM("Failed to get datagram socket address" << endl);
      return INetAddr();
    }
  }
  
  ssize_t SockDGram::recv (void* buffer, size_t len, int flags)
    throw (const TransmissionException&) 
  {
    int retval = 0;
    struct sockaddr_in peer_addr;
    socklen_t len_peer_addr = sizeof(peer_addr);
    
    retval = ::recvfrom(__socket, static_cast<char*>(buffer), len, flags,
                        reinterpret_cast<sockaddr*>(&peer_addr), &len_peer_addr);
    if (retval < 0) {
      createReadTransmissionException(errno);
    }
    
    // XXX: fw: check if peer_addr == __peer_addr !
    
    return retval;
  }
  
  ssize_t SockDGram::send (const void* buffer, size_t len, int flags) 
    throw (const TransmissionException&) 
  {
    ssize_t retval = ::sendto(__socket, static_cast<const char*>(buffer), len, flags, __peer_addr.getAddress(), __peer_addr.getSize());
    if (retval < 0) {
      createWriteTransmissionException(errno);
    }
    
    return retval;
  }
  
  ssize_t SockDGram::read (char* buffer, size_t len) 
    throw (const TransmissionException&) {
    int retval = 0;
    struct sockaddr_in peer_addr;
    int len_peer_addr;
    len_peer_addr = sizeof(peer_addr);
    
    retval = ::recvfrom(__socket, buffer, len, 0, (struct sockaddr *) &peer_addr, (socklen_t *) &len_peer_addr);
    if (retval < 0) {
      createWriteTransmissionException(errno);
    }
    
    return retval;
  }

  void SockDGram::readStream(char* data, size_t length, size_t max_packet_size /* = 0 */ ) 
    throw (const TransmissionException&, const ConnectionClosed&)
  {
    // todo: add real checking of max_packet_size
    if(max_packet_size == 0) {
      max_packet_size = 60000;
    }
    
    int received = 0;
    
    while (length > 0)  {
      struct sockaddr_in peer_addr;
      int len_peer_addr;
      len_peer_addr = sizeof(peer_addr);
      if(length < max_packet_size) {
	received = ::recvfrom(__socket, data, length, 0, reinterpret_cast<sockaddr*>(&peer_addr), (socklen_t *) &len_peer_addr);
      }
      else {
	received = ::recvfrom(__socket, data, max_packet_size, 0, reinterpret_cast<sockaddr*>(&peer_addr), (socklen_t *) &len_peer_addr);
      }
      
      length -= received;
      data   += received;
    }
  }
  
  ssize_t SockDGram::write(const char* buffer, size_t len) 
    throw (const TransmissionException&) 
  {
    signal(SIGPIPE, SIG_IGN);
    ssize_t retval = 0;
    retval = ::sendto(__socket, buffer, len, 0, __peer_addr.getAddress(), __peer_addr.getSize());
    signal(SIGPIPE, SIG_DFL);
    if (retval < 0) {
      createWriteTransmissionException(errno);
    }
    return retval;
  }

  void SockDGram::writeStream (const char* data, size_t length, size_t max_packet_size)
    throw (const TransmissionException&, const ConnectionClosed&)
  {
    // todo: add real checking of max_packet_size
    if(max_packet_size == 0) {
      max_packet_size = 60000;
    }
    
    if (max_packet_size == 0)
      max_packet_size = length;
    
    int transmit = 0;
    while (length > 0)  {    
      if (length < max_packet_size) {
	transmit = write(data, length); // calls SockDGram::write
      } else {
	transmit = write(data, max_packet_size); // calls SockDGram::write  
      }
      data   += transmit;
      length -=  transmit;
    }
  }
  
  void SockDGram::close() {
  }
}
